Package: devtools
Title: Tools to Make Developing R Packages Easier
Version: 2.5.2
Authors@R: c(
    person("Hadley", "Wickham", role = "aut"),
    person("Jim", "Hester", role = "aut"),
    person("Winston", "Chang", role = "aut"),
    person("Jennifer", "Bryan", , "jenny@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-6983-2759")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: Collection of package development tools.
License: MIT + file LICENSE
URL: https://devtools.r-lib.org/, https://github.com/r-lib/devtools
BugReports: https://github.com/r-lib/devtools/issues
Depends: R (>= 4.1), usethis (>= 3.2.1)
Imports: cli (>= 3.6.6), desc (>= 1.4.3), ellipsis (>= 0.3.3), fs (>=
        2.1.0), lifecycle (>= 1.0.5), memoise (>= 2.0.1), miniUI (>=
        0.1.2), pak (>= 0.9.5), pkgbuild (>= 1.4.8), pkgdown (>=
        2.2.0), pkgload (>= 1.5.2), profvis (>= 0.4.0), rcmdcheck (>=
        1.4.0), rlang (>= 1.2.0), roxygen2 (>= 7.3.3), rversions (>=
        3.0.0), sessioninfo (>= 1.2.3), testthat (>= 3.3.2), urlchecker
        (>= 1.0.1), withr (>= 3.0.2)
Suggests: BiocManager (>= 1.30.18), callr (>= 3.7.1), covr (>= 3.5.1),
        curl (>= 4.3.2), digest (>= 0.6.29), DT (>= 0.23), foghorn (>=
        1.4.2), gh (>= 1.3.0), httr2 (>= 1.0.0), knitr (>= 1.39), lintr
        (>= 3.0.0), quarto (>= 1.5.1), remotes (>= 2.5.0), rmarkdown
        (>= 2.14), rstudioapi (>= 0.13), spelling (>= 2.2), xml2
VignetteBuilder: knitr, quarto
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/testthat/parallel: true
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-29 15:42:59 UTC; jenny
Author: Hadley Wickham [aut],
  Jim Hester [aut],
  Winston Chang [aut],
  Jennifer Bryan [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6983-2759>),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Jennifer Bryan <jenny@posit.co>
Repository: CRAN
Date/Publication: 2026-04-30 07:40:02 UTC
Built: R 4.6.0; ; 2026-05-01 22:40:30 UTC; unix

Package: testError
Title: Test package to check error message gives file/line info
License: GPL-2
Description:
Author: Hadley <h.wickham@gmail.com>
Maintainer: Hadley <h.wickham@gmail.com>
Version: 0.1

Package: readme
Version: 0.0.0.9000
Title: What the Package Does (One Line, Title Case)
Description: What the package does (one paragraph).
Authors@R: person("First", "Last", , "first.last@example.com", c("aut", "cre"))
License: What license it uses
Encoding: UTF-8
LazyData: true

Package: testMissingNsObject
Title: Tools to make developing R code easier. This package lists 'b' as an
    export in NAMESPACE, but the 'b' object doesn't exist.
License: GPL-2
Description:
Version: 0.1

Package: testHelp
Title: Tools to make developing R code easier
License: GPL-2
Description: Test package for devtools help.
Author: Hadley <h.wickham@gmail.com>
Maintainer: Hadley <h.wickham@gmail.com>
Version: 0.1
RoxygenNote: 6.0.1

Package: testTest
Title: Tools to Make Developing R Code Easier
License: GPL-2
Description: Package description.
Authors@R:
    c(person(given = "Hadley",
             family = "Wickham",
             role = c("aut", "cre"),
             email = "h.wickham@gmail.com"))
Maintainer: Hadley <h.wickham@gmail.com>
Version: 0.1
Suggests:
    testthat
RoxygenNote: 5.0.1

Package: testInstallWithDeps
Title: Test Package With Dependencies
Version: 0.1.0
Authors@R: person("Test", "User", role = c("aut", "cre"), email = "test@example.com")
Description: Minimal package used to test that devtools::install() does not
    reinstall dependencies that are already satisfied in a non-primary library.
License: GPL-2
Imports:
    cli
Encoding: UTF-8

Package: testPkgdown
Title: Tools to make developing R code easier
License: GPL-2
Description:
Authors@R: c(
 person("Hadley", email = "h.wickham@gmail.com", role = c("aut", "cre")))
Maintainer: Hadley <h.wickham@gmail.com>
Version: 0.1
VignetteBuilder: knitr
Suggests: knitr,
  pkgdown
RoxygenNote: 6.1.1

Package: testDllRcpp
Title: Test package for compiling DLLs that link to Rcpp
License: GPL-2
Description:
Author: Hadley <h.wickham@gmail.com>
Maintainer: Hadley <h.wickham@gmail.com>
Version: 0.1
Depends: Rcpp (>= 0.10.0)
LinkingTo: Rcpp
RoxygenNote: 6.0.1

Package: testDependMissing
Title: Tools to make developing R code easier
License: GPL-2
Description:
Author: Hadley <h.wickham@gmail.com>
Maintainer: Hadley <h.wickham@gmail.com>
Version: 0.1
Depends:
    MASS,
    R
Collate: a.r b.r

Package: testS4union
Title: Test package for S4 class unions
License: GPL-2
Description: 
Author: Winston Chang <winston@stdout.org>
Maintainer: Winston Chang <winston@stdout.org>
Version: 0.1
Collate:
    'classes.r'
Imports:
    stats4,
    methods

Package: testS4import
Title: reproduce S4 import bug with devtools
Version: 0.1
Description: reproduce S4 import bug with devtools. See testS4export
Author: Karl Forner <karl.forner@quartzbio.com>
Maintainer: Karl Forner <karl.forner@quartzbio.com>
Depends:
    R (>= 2.15)
Imports:
    methods,
    testS4export
Suggests:
    testthat (>= 0.7.1.99),
License: GPL (>= 2)
Collate: all.r

Package: infrastructure
Title: What the Package Does (one line, title case)
Version: 0.0.0.9000
Authors@R: person("First", "Last", email = "first.last@example.com", role = c("aut", "cre"))
Description: What the package does (one paragraph).
Depends: R (>= 3.3.1)
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Suggests: testthat,
    knitr,
    rmarkdown,
    covr
VignetteBuilder: knitr
LinkingTo: Rcpp
Imports: Rcpp

Package: testDllLoad
Title: Test package for loading and unloading DLLs
License: GPL-2
Description:
Author: Hadley <h.wickham@gmail.com>
Maintainer: Hadley <h.wickham@gmail.com>
Version: 0.1
Collate: a.r

Package: testDataLazy
Title: Tools to make developing R code easier
License: GPL-2
Description:
Author: Hadley <h.wickham@gmail.com>
Maintainer: Hadley <h.wickham@gmail.com>
Version: 0.1
LazyData: true

Package: testS4sort
Title: Test package for sorting S4 classes
License: GPL (>= 2)
Description:
Author: Facundo Muñoz <facundo.munoz@orleans.inra.fr>
Maintainer: Facundo Muñoz <facundo.munoz@orleans.inra.fr>
Version: 0.1
Collate:
    'classes.r'
Imports:
    methods

# Generated by roxygen2: do not edit by hand

S3method(print,dev_sitrep)
S3method(print,maintainers)
export(as.package)
export(bash)
export(build)
export(build_manual)
export(build_readme)
export(build_rmd)
export(build_site)
export(build_vignettes)
export(check)
export(check_built)
export(check_dep_version)
export(check_doc_fields)
export(check_mac_devel)
export(check_mac_release)
export(check_man)
export(check_rhub)
export(check_win_devel)
export(check_win_oldrelease)
export(check_win_release)
export(clean_dll)
export(clean_vignettes)
export(create)
export(dev_mode)
export(dev_package_deps)
export(dev_packages)
export(dev_sitrep)
export(document)
export(find_rtools)
export(github_pull)
export(github_release)
export(has_devel)
export(has_tests)
export(install)
export(install_bioc)
export(install_bitbucket)
export(install_cran)
export(install_deps)
export(install_dev)
export(install_dev_deps)
export(install_git)
export(install_github)
export(install_gitlab)
export(install_local)
export(install_svn)
export(install_url)
export(install_version)
export(is.package)
export(is_loading)
export(lint)
export(load_all)
export(loaded_packages)
export(missing_s3)
export(package_file)
export(package_info)
export(parse_deps)
export(r_env_vars)
export(release)
export(release_checks)
export(reload)
export(revdep)
export(revdep_maintainers)
export(run_examples)
export(session_info)
export(show_news)
export(source_gist)
export(source_url)
export(spell_check)
export(submit_cran)
export(test)
export(test_active_file)
export(test_coverage)
export(test_coverage_active_file)
export(test_coverage_file)
export(test_file)
export(uninstall)
export(unload)
export(update_packages)
export(uses_testthat)
export(wd)
export(with_debug)
import(fs)
import(rlang)
importFrom(ellipsis,check_dots_used)
importFrom(lifecycle,deprecated)
importFrom(memoise,memoise)
importFrom(pkgbuild,clean_dll)
importFrom(pkgbuild,find_rtools)
importFrom(pkgbuild,has_devel)
importFrom(pkgbuild,with_debug)
importFrom(pkgload,check_dep_version)
importFrom(pkgload,is_loading)
importFrom(pkgload,parse_deps)
importFrom(pkgload,unload)
importFrom(sessioninfo,package_info)
importFrom(sessioninfo,session_info)
importFrom(stats,update)
importFrom(usethis,use_test)

# Devtools - development tools for Arch Linux

This repository contains tools for the Arch Linux distribution for building
and maintaining official repository packages.

## Patches

Patches can be send to arch-projects@archlinux.org or via a pull request on
Github. When sending patches to the mailing list make sure to set a valid
subjectprefix otherwise the email is denied by mailman. Git can be configured
as following.

```
git config format.subjectprefix 'devtools] [PATCH'
```

## Building

When building official distro packages the `BUILDTOOLVER` needs to be set to the
exact label of the release package in order to allow to detect the exactly used
devtools version. This is required for reproducible builds to fetch the according
files like `makepkg.conf`.

```sh
BUILDTOOLVER="${pkgver}-${pkgrel}-${arch}" make all
```

## Releasing

1. bump the version in the Makefile
2. Commit everything as  ```Version $(date +"%Y%m%d")```
3. Create a new tag ```git tag -s $(date +"%Y%m%d")```
4. Push changes
5. Upload the source tarball with ```make dist upload```
6. Update the package

## License

Devtools is licensed under the terms of the **GPL-3.0-or-later** (see [LICENSE](LICENSE)).

---
output:
  md_document:
    variant: markdown_github
---

<!-- README.md is generated from README.Rmd. Please edit that file -->

```{r, echo = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.path = "README-"
)
```

# infrastructure

The goal of infrastructure is to ...

## Example

This is a basic example which shows you how to solve a common problem:

```{r example}
## basic example code
```

#' Environment variables to set when calling R
#'
#' Devtools sets a number of environmental variables to ensure consistent
#' between the current R session and the new session, and to ensure that
#' everything behaves the same across systems. It also suppresses a common
#' warning on windows, and sets `NOT_CRAN` so you can tell that your
#' code is not running on CRAN. If `NOT_CRAN` has been set externally, it
#' is not overwritten.
#'
#' @keywords internal
#' @return a named character vector
#' @export
r_env_vars <- function() {
  vars <- c(
    "R_LIBS" = paste(.libPaths(), collapse = .Platform$path.sep),
    "CYGWIN" = "nodosfilewarning",
    # When R CMD check runs tests, it sets R_TESTS. When the tests
    # themselves run R CMD xxxx, as is the case with the tests in
    # devtools, having R_TESTS set causes errors because it confuses
    # the R subprocesses. Un-setting it here avoids those problems.
    "R_TESTS" = "",
    "R_BROWSER" = "false",
    "R_PDFVIEWER" = "false"
  )

  if (is.na(Sys.getenv("NOT_CRAN", unset = NA))) {
    vars[["NOT_CRAN"]] <- "true"
  }

  vars
}

upload_ftp <- function(file, url, verbose = FALSE) {
  check_suggested("curl")

  stopifnot(file.exists(file))
  stopifnot(is.character(url))
  con <- file(file, open = "rb")
  on.exit(close(con))
  h <- curl::new_handle(upload = TRUE, filetime = FALSE)
  curl::handle_setopt(h, readfunction = function(n) {
    readBin(con, raw(), n = n)
  }, verbose = verbose)
  curl::curl_fetch_memory(url, handle = h)
}

#' @section Package options:
#'
#' Devtools uses the following [options()] to configure behaviour:
#' * `devtools.install.args`: a string giving extra arguments passed to
#'   `R CMD install` by [install()].
#' * `devtools.path`: Deprecated. Path used by the now-deprecated [dev_mode()]
#'   function.
#' @docType package
#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom ellipsis check_dots_used
#' @importFrom lifecycle deprecated
#' @import rlang
## usethis namespace: end
NULL

# https://r-pkgs.org/dependencies-in-practice.html#how-to-not-use-a-package-in-imports
ignore_unused_imports <- function() {
  miniUI::miniPage
  profvis::profvis
  urlchecker::url_check
}

#' Return a vector of names of attached packages
#' @export
#' @keywords internal
#' @return A data frame with columns package and path, giving the name of
#'   each package and the path it was loaded from.
loaded_packages <- function() {
  attached <- data.frame(
    package = search(),
    path = searchpaths(),
    stringsAsFactors = FALSE
  )
  packages <- attached[grepl("^package:", attached$package), , drop = FALSE]
  rownames(packages) <- NULL

  packages$package <- sub("^package:", "", packages$package)
  packages
}

#' Return a vector of names of packages loaded by devtools
#' @export
#' @keywords internal
dev_packages <- function() {
  packages <- map_lgl(loadedNamespaces(), \(x) !is.null(pkgload::dev_meta(x)))

  names(packages)[packages]
}

#' @export
#' @importFrom sessioninfo session_info
sessioninfo::session_info

#' @export
#' @importFrom sessioninfo package_info
sessioninfo::package_info

test_that("uninstall() unloads and removes from library", {
  withr::local_temp_libpaths()

  # Install package
  install(test_path("testHelp"), quiet = TRUE)
  expect_true(require(testHelp, quietly = TRUE))
  expect_true("testHelp" %in% loaded_packages()$package)

  # Uninstall package
  uninstall(test_path("testHelp"), quiet = TRUE)
  expect_false("testHelp" %in% loaded_packages()$package)
  suppressWarnings(expect_false(require(testHelp, quietly = TRUE)))
})

context("Metadata")

test_that("devtools metadata for load hooks", {

  # testLoadHooks test package has .onLoad and .onAttach
  load_all("testLoadHooks")
  md <- dev_meta("testLoadHooks")
  expect_true(md$.onLoad)
  expect_true(md$.onAttach)
  unload("testLoadHooks")


  # testNamespace test package doesn't have .onLoad and .onAttach
  load_all("testNamespace")
  md <- dev_meta("testNamespace")
  expect_false(exists("onLoad", envir = md))
  expect_false(exists("onAttach", envir = md))
  unload("testNamespace")
})


test_that("NULL metadata for non-devtools-loaded packages", {
  expect_true(is.null(dev_meta("stats")))
})


test_that("dev_packages() lists devtools-loaded packages", {
  expect_false(any(c("testNamespace", "testLoadHooks") %in% dev_packages()))
  expect_false("testNamespace" %in% dev_packages())
  expect_false("testLoadHooks" %in% dev_packages())

  load_all("testNamespace")
  expect_true("testNamespace" %in% dev_packages())
  expect_false("testLoadHooks" %in% dev_packages())

  load_all("testLoadHooks")
  expect_true("testNamespace" %in% dev_packages())
  expect_true("testLoadHooks" %in% dev_packages())

  unload("testNamespace")
  expect_false("testNamespace" %in% dev_packages())
  expect_true("testLoadHooks" %in% dev_packages())

  unload("testLoadHooks")
  expect_false("testNamespace" %in% dev_packages())
  expect_false("testLoadHooks" %in% dev_packages())


  expect_false("stats" %in% dev_packages())
})

#' @template devtools
#' @param path Path in which to produce package.  If `NULL`, defaults to
#'   the parent directory of the package.
#' @inherit pkgbuild::build
#' @note The default `manual = FALSE` is not suitable for a CRAN
#'   submission, which may require `manual = TRUE`.  Even better, use
#'   [submit_cran()] or [release()].
#' @param ... Additional arguments passed to [pkgbuild::build].
#' @export
build <- function(
  pkg = ".",
  path = NULL,
  binary = FALSE,
  vignettes = TRUE,
  manual = FALSE,
  args = NULL,
  quiet = FALSE,
  ...
) {
  save_all()

  if (!file_exists(pkg)) {
    cli::cli_abort("{.arg pkg} must exist")
  }

  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  pkgbuild::build(
    path = pkg,
    dest_path = path,
    binary = binary,
    vignettes = vignettes,
    manual = manual,
    args = args,
    quiet = quiet,
    ...
  )
}

#' @importFrom pkgbuild with_debug
#' @export
pkgbuild::with_debug

#' @importFrom pkgbuild clean_dll
#' @export
pkgbuild::clean_dll

#' @importFrom pkgbuild has_devel
#' @export
pkgbuild::has_devel

#' @importFrom pkgbuild find_rtools
#' @export
pkgbuild::find_rtools

context("build")

test_that("source builds return correct filenames", {
  path <- devtools::build("testNamespace", path=tempdir(),
                          quiet=TRUE)
  expect_true(file.exists(path))
})

test_that("binary builds return correct filenames", {
  path <- devtools::build("testNamespace", binary=TRUE, path=tempdir(),
                          quiet=TRUE)
  expect_true(file.exists(path))
})

#' Build Rmarkdown files
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `build_rmd()` is deprecated, as it is a low-level helper for internal use. To
#' render your package's `README.qmd` or `README.Rmd`, use [build_readme()]. To
#' preview a vignette or article, use functions like [pkgdown::build_site()] or
#' [pkgdown::build_article()].
#'
#' @param files The Rmarkdown files to be rendered.
#' @param path path to the package to build the readme.
#' @param ...  additional arguments passed to [rmarkdown::render()]
#' @inheritParams install
#' @inheritParams rmarkdown::render
#' @export
#' @keywords internal
build_rmd <- function(
  files,
  path = ".",
  output_options = list(),
  ...,
  quiet = TRUE
) {
  lifecycle::deprecate_soft("2.5.0", "build_rmd()", "build_readme()")
  build_rmd_impl(
    files = files,
    path = path,
    output_options = output_options,
    ...,
    quiet = quiet
  )
}

# Created as part of the deprecation process to de-export build_rmd().
# We still want to use this internally without needing to suppress deprecation
# signals.
build_rmd_impl <- function(
  files,
  path = ".",
  output_options = list(),
  ...,
  quiet = TRUE,
  call = parent.frame()
) {
  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  pkg <- as.package(path)

  check_installed("rmarkdown")
  save_all()

  paths <- files
  abs_files <- is_absolute_path(files)
  paths[!abs_files] <- path(pkg$path, files[!abs_files])

  ok <- file_exists(paths)
  if (any(!ok)) {
    cli::cli_abort("Can't find file{?s}: {.path {files[!ok]}}.")
  }

  local_install(pkg, quiet = quiet, call = call)

  # Ensure rendering github_document() doesn't generate HTML file
  output_options$html_preview <- FALSE

  for (path in paths) {
    callr::r_safe(
      function(...) rmarkdown::render(...),
      args = list(
        input = path,
        ...,
        output_options = output_options,
        quiet = quiet
      ),
      show = TRUE,
      spinner = FALSE,
      stderr = "2>&1"
    )
  }

  invisible(TRUE)
}

#' Build README
#'
#' Renders an executable README, i.e. `README.qmd` or `README.Rmd`, to
#' `README.md`. Specifically, `build_readme()`:
#' * Installs a copy of the package's current source to a temporary library
#' * Renders the README in a clean R session
#'
#' @param path Path to the top-level directory of the source package.
#' @param quiet If `TRUE`, suppresses most output. Set to `FALSE`
#'   if you need to debug.
#' @param ... Additional arguments passed to [rmarkdown::render()], in the
#'   case of `README.Rmd`. Not used for `README.qmd`
#' @export
build_readme <- function(path = ".", quiet = TRUE, ...) {
  pkg <- as.package(path)

  readme_candidates <- c(
    path(pkg$path, "README.qmd"),
    path(pkg$path, "README.Rmd"),
    path(pkg$path, "inst", "README.qmd"),
    path(pkg$path, "inst", "README.Rmd")
  )
  readme_path <- readme_candidates[file_exists(readme_candidates)]

  if (length(readme_path) == 0) {
    cli::cli_abort(
      "Can't find {.file README.qmd} or {.file README.Rmd}, at the top-level or
      below {.file inst/}."
    )
  }
  if (length(readme_path) > 1) {
    rel_paths <- path_rel(readme_path, pkg$path)
    cli::cli_abort(
      "Found multiple executable READMEs: {.file {rel_paths}}. There can only be
      one."
    )
  }

  if (!quiet) {
    cli::cli_inform(c(i = "Building {.path {readme_path}}"))
  }

  if (path_ext(readme_path) == "qmd") {
    build_qmd_readme(readme_path, path = path, quiet = quiet)
  } else {
    build_rmd_impl(readme_path, path = path, quiet = quiet, ...)
  }
}

build_qmd_readme <- function(
  readme_path,
  path = ".",
  quiet = TRUE,
  call = parent.frame()
) {
  pkg <- as.package(path)

  check_installed("quarto")
  save_all()

  local_install(pkg, quiet = quiet, call = call)

  # Quarto spawns its own R process for knitr, which won't inherit .libPaths().

  # Pass library paths via R_LIBS_USER so the quarto subprocess finds the
  # temporarily installed package first, ahead of any user-installed version.
  lib_paths <- paste(.libPaths(), collapse = .Platform$path.sep)

  callr::r_safe(
    function(input, quiet) {
      quarto::quarto_render(input = input, quiet = quiet)
    },
    args = list(input = readme_path, quiet = quiet),
    env = c(callr::rcmd_safe_env(), R_LIBS_USER = lib_paths),
    show = TRUE,
    spinner = FALSE,
    stderr = "2>&1"
  )

  invisible(TRUE)
}

test_that("can build README.Rmd in root directory", {
  skip_on_cran()

  pkg <- local_package_create()
  usethis::ui_silence(
    usethis::with_project(
      pkg,
      use_readme_rmd(open = FALSE)
    )
  )

  build_readme(pkg, quiet = TRUE)
  expect_true(file_exists(path(pkg, "README.md")))
  expect_false(file_exists(path(pkg, "README.html")))
})

test_that("can build README.Rmd in inst/", {
  skip_on_cran()

  pkg <- local_package_create()
  usethis::ui_silence(
    usethis::with_project(
      pkg,
      use_readme_rmd(open = FALSE)
    )
  )
  dir_create(pkg, "inst")
  file_move(
    path(pkg, "README.Rmd"),
    path(pkg, "inst", "README.Rmd")
  )

  build_readme(pkg, quiet = TRUE)
  expect_true(file_exists(path(pkg, "inst", "README.md")))
  expect_false(file_exists(path(pkg, "README.Rmd")))
  expect_false(file_exists(path(pkg, "README.md")))
  expect_false(file_exists(path(pkg, "inst", "README.html")))
})

test_that("can build README.qmd in root directory", {
  skip_on_cran()
  skip_if_not_installed("quarto")
  skip_if_not(quarto::quarto_available(), "quarto cli not available")

  pkg <- local_package_create()
  # TODO: use usethis::use_readme_qmd() once it's in a usethis release
  # https://github.com/r-lib/usethis/pull/2219
  writeLines(
    c(
      "---",
      "format: gfm",
      "---",
      "",
      "# testpkg",
      "",
      "This is a test package."
    ),
    path(pkg, "README.qmd")
  )

  build_readme(pkg, quiet = TRUE)
  expect_true(file_exists(path(pkg, "README.md")))
})

test_that("can build README.qmd in inst/", {
  skip_on_cran()
  skip_if_not_installed("quarto")
  skip_if_not(quarto::quarto_available(), "quarto cli not available")

  pkg <- local_package_create()
  # TODO: use usethis::use_readme_qmd() once it's in a usethis release
  # https://github.com/r-lib/usethis/pull/2219
  dir_create(pkg, "inst")
  writeLines(
    c(
      "---",
      "format: gfm",
      "---",
      "",
      "# testpkg",
      "",
      "This is a test package."
    ),
    path(pkg, "inst", "README.qmd")
  )

  build_readme(pkg, quiet = TRUE)
  expect_true(file_exists(path(pkg, "inst", "README.md")))
  expect_false(file_exists(path(pkg, "README.qmd")))
  expect_false(file_exists(path(pkg, "README.md")))
})

test_that("useful errors if too few or too many", {
  pkg <- local_package_create()
  expect_snapshot(build_readme(pkg), error = TRUE)

  usethis::ui_silence(
    usethis::with_project(
      pkg,
      use_readme_rmd(open = FALSE)
    )
  )
  dir_create(pkg, "inst")
  file_copy(path(pkg, "README.Rmd"), path(pkg, "inst", "README.Rmd"))
  expect_snapshot(build_readme(pkg), error = TRUE)
})

test_that("errors if both README.qmd and README.Rmd exist", {
  pkg <- local_package_create()
  file_create(path(pkg, "README.Rmd"))
  file_create(path(pkg, "README.qmd"))
  expect_snapshot(build_readme(pkg), error = TRUE)
})

test_that("build_readme() aborts when a dep is missing", {
  pkg <- local_package_create()
  pkg_name <- basename(pkg)
  usethis::ui_silence(
    usethis::with_project(pkg, use_readme_rmd(open = FALSE))
  )

  local_mocked_bindings(
    pkg_dep_status = function(...) {
      data.frame(
        package = "missingpkg",
        latest = "1.0.0",
        installed = NA_character_,
        status = "missing"
      )
    }
  )
  expect_snapshot(
    build_readme(pkg),
    error = TRUE,
    transform = function(x) gsub(pkg_name, "{PACKAGE}", x, fixed = TRUE)
  )
})

test_that("don't error for README in another directory", {
  skip_on_cran()

  pkg <- local_package_create()
  usethis::ui_silence(
    usethis::with_project(
      pkg,
      use_readme_rmd(open = FALSE)
    )
  )
  dir_create(pkg, "data-raw")
  file_create(pkg, "data-raw", "README.md")

  expect_no_error(build_readme(pkg, quiet = TRUE))
})

test_that("build_rmd() is deprecated", {
  skip_on_cran()

  pkg <- local_package_create()
  usethis::ui_silence(
    usethis::with_project(
      pkg,
      use_readme_rmd(open = FALSE)
    )
  )

  withr::local_options(lifecycle_verbosity = "warning")
  # it's hard (impossible?) to silence pak (cli, really) so that's what the
  # suppressMessages() is for
  expect_snapshot(suppressMessages(build_rmd(
    "README.Rmd",
    path = pkg,
    quiet = TRUE
  )))
})

#' Spell checking
#'
#' Runs a spell check on text fields in the package description file, manual
#' pages, and optionally vignettes. Wraps the \link[spelling:spell_check_package]{spelling}
#' package.
#'
#' @export
#' @rdname spell_check
#' @template devtools
#' @param vignettes also check all `rmd` and `rnw` files in the pkg `vignettes` folder
#' @param use_wordlist ignore words in the package [WORDLIST][spelling::get_wordlist] file
spell_check <- function(pkg = ".", vignettes = TRUE, use_wordlist = TRUE) {
  check_suggested("spelling")
  pkg <- as.package(pkg)
  spelling::spell_check_package(pkg = pkg, vignettes = vignettes, use_wordlist = use_wordlist)
}

#' Git checks.
#'
#' This function performs Git checks checks prior to release. It is called
#' automatically by [release()].
#'
#' @template devtools
#' @keywords internal
git_checks <- function(pkg = ".") {
  pkg <- as.package(pkg)
  cat_rule(paste0("Running Git checks for ", pkg$package))

  git_report_branch(pkg)
  git_check_uncommitted(pkg)
  cat_rule()
}

git_report_branch <- function(pkg) {
  cat("Current branch:", git_branch(pkg$path), "\n")
}

git_check_uncommitted <- function(pkg) {
  check_status(
    !git_uncommitted(pkg$path),
    "uncommitted files",
    "All files should be tracked and committed before release. Please add and commit."
  )
}

#' Create package pdf manual
#'
#' @template devtools
#' @param path path in which to produce package manual.
#'   If `NULL`, defaults to the parent directory of the package.
#'
#' @seealso [Rd2pdf()]
#' @export
build_manual <- function(pkg = ".", path = NULL) {
  pkg <- as.package(pkg)
  path <- path %||% path_dir(pkg$path)
  name <- paste0(pkg$package, "_", pkg$version, ".pdf")
  output <- file.path(path, name)

  cli::cli_inform("Saving manual to {.file {output}}")
  withCallingHandlers(
    invisible(rd2pdf(pkg$path, output)),
    error = function(e) {
      cli::cli_abort(
        c("Failed to build manual.", no_wrap(e$stderr)),
        call = quote(build_manual()),
        parent = e
      )
    }
  )
}

rd2pdf <- function(pkg_path, output_path) {
  callr::rcmd(
    "Rd2pdf",
    cmdargs = c("--force", paste0("--output=", output_path), pkg_path),
    stdout = "",
    fail_on_status = TRUE,
    spinner = FALSE
  )
}

test_that("build_manual() shows stderr on failure", {
  skip_on_os("windows")

  pkg <- local_package_create()
  pkg <- normalizePath(pkg)

  # Too hard to replicate actual error, so we just simulate
  local_mocked_bindings(rd2pdf = function(...) {
    stderr <- "! LaTeX Error: File `inconsolata.sty' not found."
    abort("System command 'R' failed", stderr = stderr) #nolint
  })

  expect_snapshot(build_manual(pkg), error = TRUE, transform = function(x) {
    x <- gsub(pkg, "<pkgdir>", x, fixed = TRUE)
    x
  })
})

#' Run all examples in a package
#'
#' One of the most frustrating parts of `R CMD check` is getting all of your
#' examples to pass - whenever one fails you need to fix the problem and then
#' restart the whole process.  This function makes it a little easier by
#' making it possible to run all examples from an R function.
#'
#' @template devtools
#' @inheritParams pkgload::run_example
#' @param start Where to start running the examples: this can either be the
#'   name of `Rd` file to start with (with or without extensions), or
#'   a topic name. If omitted, will start with the (lexicographically) first
#'   file. This is useful if you have a lot of examples and don't want to
#'   rerun them every time you fix a problem. To run only one example, use
#'   [pkgload::run_example()].
#' @family example functions
#' @param show DEPRECATED.
#' @param fresh if `TRUE`, will be run in a fresh R session. This has
#'   the advantage that there's no way the examples can depend on anything in
#'   the current session, but interactive code (like [browser()])
#'   won't work.
#' @param document if `TRUE`, [document()] will be run to ensure
#'   examples are updated before running them.
#' @keywords programming
#'
#' @seealso [pkgload::run_example()] to run a single example.
#' @export
run_examples <- function(
  pkg = ".",
  start = NULL,
  show = deprecated(),
  run_donttest = FALSE,
  run_dontrun = FALSE,
  fresh = FALSE,
  document = TRUE,
  run = deprecated(),
  test = deprecated()
) {
  if (!missing(run)) {
    lifecycle::deprecate_warn(
      "2.3.1",
      "run_examples(run)",
      'run_example(run_dontrun)'
    )
    run_dontrun <- run
  }
  if (!missing(test)) {
    lifecycle::deprecate_warn(
      "2.3.1",
      "run_examples(test)",
      'run_example(run_donttest)'
    )
    run_donttest <- test
  }
  if (!missing(show)) {
    lifecycle::deprecate_warn("2.3.1", "run_examples(show)")
  }

  pkg <- as.package(pkg)

  if (fresh) {
    to_run <-
      function(path, start, run_donttest, run_dontrun) {
        devtools::run_examples(
          pkg = path,
          start = start,
          run_donttest = run_donttest,
          run_dontrun = run_dontrun,
          document = FALSE
        )
      }

    callr::r(
      to_run,
      args = list(
        path = pkg$path,
        start = start,
        run_donttest = run_donttest,
        run_dontrun = run_dontrun
      ),
      show = TRUE,
      spinner = FALSE,
      stderr = "2>&1"
    )
    return(invisible())
  }

  if (document) {
    document(pkg)
  }

  files <- rd_files(pkg$path, start = start)
  if (length(files) == 0) {
    return()
  }

  cli::cat_rule(
    left = paste0("Running ", length(files), " example files"),
    right = pkg$package
  )

  load_all(pkg$path, reset = TRUE, export_all = FALSE, helpers = FALSE)
  on.exit(load_all(pkg$path, reset = TRUE))

  walk(
    files,
    pkgload::run_example,
    run_donttest = run_donttest,
    run_dontrun = run_dontrun
  )

  invisible()
}
# If an error occurs, should print out the suspect line of code, and offer
# the following options:
#   * skip to the next example
#   * quit
#   * browser
#   * rerun example and rerun
#   * reload code and rerun

rd_files <- function(pkg = ".", start = NULL) {
  pkg <- as.package(pkg)

  path_man <- path(pkg$path, "man")
  files <- dir_ls(path_man, regexp = "\\.[Rr]d$")
  names(files) <- path_file(files)
  files <- sort_ci(files)

  if (!is.null(start)) {
    topic <- pkgload::dev_help(start, dev_packages = pkg$package)
    start_path <- path_file(topic$path)

    start_pos <- which(names(files) == start_path)
    if (length(start_pos) == 1) {
      files <- files[-seq(1, start_pos - 1)]
    }
  }

  files
}

#' @import fs
NULL

devtools_default_options <- list(
  devtools.path = "~/R-dev",
  devtools.install.args = "",
  devtools.ellipsis_action = warn
)

.onLoad <- function(libname, pkgname) {
  op <- options()
  toset <- !(names(devtools_default_options) %in% names(op))
  if (any(toset)) {
    options(devtools_default_options[toset])
  }

  # On certain linux systems, pak might call `sudo`, as a probe for
  # capabilities. That lays the ground work for a potential need to work with
  # sysreqs, but that's not necessary in this case and CRAN flags the
  # `sudo -s id` as problematic. Setting `pkg.sysreq` to `FALSE` prevents
  # pak from even checking this.
  # https://pak.r-lib.org/reference/pak-config.html#pak-configuration
  if (Sys.getenv("_R_CHECK_PACKAGE_NAME_", "") != "") {
    options(pkg.sysreqs = FALSE)
  }

  invisible()
}

uses_git <- function(path = ".") {
  dir_exists(path(path, ".git"))
}

git_branch <- function(path = ".") {
  withr::local_dir(path)

  system2("git", c("rev-parse", "--abbrev-ref", "HEAD"), stdout = TRUE)
}

git_uncommitted <- function(path = ".") {
  withr::local_dir(path)

  out <- system2("git", c("status", "--porcelain=v1"), stdout = TRUE)
  length(out) > 0
}

#' Open bash shell in package directory
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `bash()` is deprecated because we no longer use or recommend this workflow.
#' Open bash shell in package directory
#'
#' @template devtools
#' @export
#' @keywords internal
bash <- function(pkg = ".") {
  lifecycle::deprecate_warn("2.5.0", "bash()")
  pkg <- as.package(pkg)

  withr::with_dir(pkg$path, system("bash"))
}

# Return the version of a package on CRAN (or other repository)
# @param package The name of the package.
# @param available A matrix of information about packages.
cran_pkg_version <- function(package, available = available.packages()) {
  idx <- available[, "Package"] == package
  if (any(idx)) {
    as.package_version(available[package, "Version"])
  } else {
    NULL
  }
}

#' Run CRAN checks for package on R-hub
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' This function is deprecated since the underlying function
#' `rhub::check_for_cran()` is now deprecated and defunct. See [`rhub::rhubv2`]
#' learn about the new check system, R-hub v2.
#'
#' @param platforms R-hub platforms to run the check on.
#' @param email email address to notify.
#' @param interactive whether to show the status of the build.
#' @param build_args Arguments passed to `R CMD build`.
#' @param ... extra arguments, passed to `rhub::check_for_cran()`.
#' @inheritParams check
#' @return a `rhub_check` object.
#' @keywords internal
#' @export
check_rhub <- function(
  pkg = ".",
  platforms = NULL,
  email = NULL,
  interactive = TRUE,
  build_args = NULL,
  ...
) {
  lifecycle::deprecate_stop("2.4.6", "check_rhub()")
}

# Suppress R CMD check note. memoise is used at build time!
#' @importFrom memoise memoise
NULL

rstudio_version_string <- function() {
  if (!is_rstudio_running()) {
    return(character())
  }
  rvi <- rstudioapi::versionInfo()
  rvi$long_version %||% as.character(rvi$version)
}

check_for_rstudio_updates <- function(
  os = tolower(Sys.info()[["sysname"]]),
  version = rstudio_version_string(),
  in_rstudio = is_rstudio_running()
) {
  if (!in_rstudio) {
    return()
  }

  url <- sprintf(
    "https://www.rstudio.org/links/check_for_update?version=%s&os=%s&format=%s&manual=true",
    utils::URLencode(version, reserved = TRUE),
    os,
    "kvp"
  )

  tmp <- file_temp()
  withr::defer(file_exists(tmp) && nzchar(file_delete(tmp)))
  suppressWarnings(
    download_ok <- tryCatch(
      {
        utils::download.file(url, tmp, quiet = TRUE)
        TRUE
      },
      error = function(e) FALSE
    )
  )
  if (!download_ok) {
    return(
      sprintf("Unable to check for RStudio updates (you're using %s).", version)
    )
  }
  result <- readLines(tmp, warn = FALSE)

  result <- strsplit(result, "&")[[1]]

  result <- strsplit(result, "=")

  # If no values then we are current
  if (length(result[[1]]) == 1) {
    return()
  }

  nms <- map_chr(result, `[[`, 1)
  values <- map_chr(result, function(x) utils::URLdecode(x[[2]]))

  result <- stats::setNames(values, nms)

  if (!nzchar(result[["update-version"]])) {
    return()
  }

  return(
    sprintf(
      "%s.\nDownload at: %s",
      result[["update-message"]],
      result[["update-url"]]
    )
  )
}

r_release <- function() {
  R_system_version(rversions::resolve("release")$version)
}

#' Report package development situation
#'
#' @description
#' Call this function if things seem weird and you're not sure
#' what's wrong or how to fix it. It reports:
#'
#' * If R is up to date.
#' * If RStudio or Positron is up to date.
#' * If compiler build tools are installed and available for use.
#' * If devtools and its dependencies are up to date.
#' * If the package's dependencies are up to date.
#'
#' @return A named list, with S3 class `dev_sitrep` (for printing purposes).
#' @template devtools
#' @inheritParams pkgbuild::has_build_tools
#' @export
#' @examples
#' \dontrun{
#' dev_sitrep()
#' }
dev_sitrep <- function(pkg = ".", debug = FALSE) {
  pkg <- tryCatch(as.package(pkg), error = function(e) NULL)

  has_build_tools <- !is_windows || pkgbuild::has_build_tools(debug = debug)

  new_dev_sitrep(
    pkg = pkg,
    r_version = getRversion(),
    r_path = path_real(R.home()),
    r_release_version = r_release(),
    is_windows = is_windows,
    has_build_tools = has_build_tools,
    rtools_path = if (has_build_tools) pkgbuild::rtools_path(),
    devtools_version = utils::packageVersion("devtools"),
    devtools_cran_version = pak::pkg_deps(
      "devtools",
      dependencies = FALSE
    )$version,
    devtools_deps = pkg_dep_status("devtools", dependencies = NA),
    pkg_deps = if (!is.null(pkg)) pkg_dep_status(pkg, dependencies = TRUE),
    rstudio_version = if (is_rstudio_running()) rstudioapi::getVersion(),
    rstudio_msg = if (!is_positron()) check_for_rstudio_updates()
  )
}

new_dev_sitrep <- function(
  pkg = NULL,
  r_version = getRversion(),
  r_path = path_real(R.home()),
  r_release_version = r_version,
  is_windows = FALSE,
  has_build_tools = TRUE,
  rtools_path = NULL,
  devtools_version = utils::packageVersion("devtools"),
  devtools_cran_version = devtools_version,
  devtools_deps = NULL,
  pkg_deps = NULL,
  rstudio_version = NULL,
  rstudio_msg = NULL
) {
  structure(
    list(
      pkg = pkg,
      r_version = r_version,
      r_path = r_path,
      r_release_version = r_release_version,
      is_windows = is_windows,
      has_build_tools = has_build_tools,
      rtools_path = rtools_path,
      devtools_version = devtools_version,
      devtools_cran_version = devtools_cran_version,
      devtools_deps = devtools_deps,
      pkg_deps = pkg_deps,
      rstudio_version = rstudio_version,
      rstudio_msg = rstudio_msg
    ),
    class = "dev_sitrep"
  )
}

#' @export
print.dev_sitrep <- function(x, ...) {
  all_ok <- TRUE

  cli::cli_rule("R")
  kv_line("version", x$r_version)
  kv_line("path", x$r_path, path = TRUE)
  if (x$r_version < x$r_release_version) {
    all_ok <- FALSE
    cli::cli_bullets(c(
      "!" = "{.field R} is out of date ({.val {x$r_version}} vs {.val {x$r_release_version}})"
    ))
  }

  if (x$is_windows) {
    cli::cli_rule("Rtools")
    if (x$has_build_tools) {
      kv_line("path", x$rtools_path, path = TRUE)
    } else {
      all_ok <- FALSE
      cli::cli_bullets(c(
        "!" = "{.field Rtools} is not installed.",
        " " = "Download and install it from: {.url https://cloud.r-project.org/bin/windows/Rtools/}"
      ))
    }
  }

  if (!is.null(x$rstudio_version)) {
    cli::cli_rule(if (is_positron()) "Positron" else "RStudio")
    kv_line("version", x$rstudio_version)

    if (!is.null(x$rstudio_msg)) {
      all_ok <- FALSE
      cli::cli_bullets(c("!" = "{x$rstudio_msg}"))
    }
  }

  cli::cli_rule("devtools")
  kv_line("version", x$devtools_version)

  devtools_version <- package_version(x$devtools_version)
  devtools_cran_version <- package_version(x$devtools_cran_version)
  if (devtools_version < devtools_cran_version) {
    all_ok <- FALSE
    cli::cli_bullets(c(
      "!" = "{.field devtools} is out of date ({.val {devtools_version}} vs {.val {devtools_cran_version}}).",
      " " = "Update it with {.run pak::pak(\"devtools\")}."
    ))
  } else if (devtools_version > devtools_cran_version) {
    cli::cli_bullets(c(
      "i" = "{.field devtools} is ahead of CRAN ({.val {devtools_version}} vs {.val {devtools_cran_version}})."
    ))
  }

  devtools_not_ok <- any(x$devtools_deps$status != "ok")
  if (devtools_not_ok) {
    all_ok <- FALSE
    report_deps_ahead_behind(
      x$devtools_deps,
      pkg_name = "devtools",
      update_code = 'pak::pak("devtools")'
    )
  }

  cli::cli_rule("dev package")
  kv_line("package", x$pkg$package)
  kv_line("path", x$pkg$path, path = TRUE)

  dev_pkg_not_ok <- any(x$pkg_deps$status != "ok")
  if (dev_pkg_not_ok) {
    all_ok <- FALSE
    report_deps_ahead_behind(
      x$pkg_deps,
      pkg_name = x$pkg$package,
      update_code = "pak::local_install_dev_deps()"
    )
  }

  if (all_ok) {
    cli::cli_bullets(c("v" = "All checks passed"))
  }

  invisible(x)
}

# Helpers -----------------------------------------------------------------

#' Get dependency status for a package, excluding the package itself
#'
#' @param pkg A package name or a package object, as returned by `as.package()`.
#' @param dependencies Which dependency types to include. Passed along to
#'   `pak::pkg_deps()` or `pak::local_dev_deps()`.
#' @returns A data frame with one row per dependency and columns:
#'   * package (name)
#'   * latest (version)
#'   * installed (version)
#'   * status (one of: missing, behind, ok, ahead)
#' @noRd
pkg_dep_status <- function(pkg, dependencies = NA) {
  if (is_string(pkg)) {
    pkg_name <- pkg
    deps <- pak::pkg_deps(pkg, dependencies = dependencies)
  } else if (inherits(pkg, "package")) {
    pkg_name <- pkg$package
    deps <- pak::local_dev_deps(pkg$path, dependencies = dependencies)
  } else {
    cli::cli_abort(
      "{.arg pkg} must be a string package name or a package object."
    )
  }
  deps <- deps[deps$package != pkg_name, ]

  installed <- map_chr(deps$package, function(p) {
    tryCatch(
      as.character(utils::packageVersion(p)),
      error = function(e) NA_character_
    )
  })
  status <- map2_chr(installed, deps$version, function(inst, latest) {
    if (is.na(inst)) {
      return("missing")
    }
    switch(
      as.character(utils::compareVersion(inst, latest)),
      "-1" = "behind",
      "0" = "ok",
      "1" = "ahead"
    )
  })
  data.frame(
    package = deps$package,
    latest = deps$version,
    installed = installed,
    status = status
  )
}

#' Emit cli messages about ahead/behind dependencies
#'
#' @param dep_status A data frame as returned by `compare_deps()`, with
#'   columns `package`, `latest`, `installed`, `status`.
#' @param pkg_name Package name to mention in the message.
#' @param update_code Code suggestion for updating behind deps.
#' @return Called for its side effects.
#' @noRd
report_deps_ahead_behind <- function(dep_status, pkg_name, update_code) {
  missing <- dep_status[dep_status$status == "missing", ]
  if (nrow(missing) > 0) {
    n <- nrow(missing)
    cli::cli_bullets(c(
      "!" = "{n} {.field {pkg_name}} {cli::qty(n)}{?dependency is/dependencies are} not installed.",
      " " = "Install {cli::qty(n)}{?it/them} with {.run {update_code}}."
    ))
    cli::cli_verbatim(paste("  ", dep_labels(missing)))
  }

  behind <- dep_status[dep_status$status == "behind", ]
  if (nrow(behind) > 0) {
    n <- nrow(behind)
    cli::cli_bullets(c(
      "!" = "{n} {.field {pkg_name}} {cli::qty(n)}{?dependency is/dependencies are} out of date.",
      " " = "Update {cli::qty(n)}{?it/them} with {.run {update_code}}."
    ))
    cli::cli_verbatim(paste("  ", dep_labels(behind)))
  }

  ahead <- dep_status[dep_status$status == "ahead", ]
  if (nrow(ahead) > 0) {
    n <- nrow(ahead)
    cli::cli_bullets(c(
      "i" = "{n} {.field {pkg_name}} {cli::qty(n)}{?dependency is/dependencies are} ahead of CRAN."
    ))
    cli::cli_verbatim(paste(" ", dep_labels(ahead)))
  }
}


dep_labels <- function(deps) {
  # helps with readability
  deps$package <- format(deps$package, justify = "left")
  labels <- list_c(pmap(deps, format_dep_line))
  stats::setNames(labels, rep(" ", length(labels)))
}

format_dep_line <- function(package, installed, latest, status) {
  if (is.na(installed)) {
    paste0(package, " (", cli::col_red("missing"), ")")
  } else {
    status_styled <- if (status == "behind") {
      cli::col_red(status)
    } else {
      cli::col_cyan(status)
    }
    paste0(package, " (", status_styled, ": ", installed, " vs ", latest, ")")
  }
}

kv_line <- function(key, value, path = FALSE) {
  if (is.null(value)) {
    cli::cli_inform(c("*" = "{key}: {.silver <unset>}"))
  } else if (path) {
    cli::cli_inform(c("*" = "{key}: {.path {value}}"))
  } else {
    cli::cli_inform(c("*" = "{key}: {.val {value}}"))
  }
}

test_that("print shows all checks passed", {
  local_reproducible_output(width = 60)
  x <- new_dev_sitrep(
    r_version = R_system_version("4.4.0"),
    r_path = "/usr/lib/R",
    devtools_version = package_version("2.4.6")
  )
  expect_snapshot(print(x))
})

test_that("print warns when devtools is out of date", {
  local_reproducible_output(width = 60)
  x <- new_dev_sitrep(
    r_version = R_system_version("4.4.0"),
    r_path = "/usr/lib/R",
    devtools_version = package_version("2.4.6"),
    devtools_cran_version = package_version("2.5.0")
  )
  expect_snapshot(print(x))
})

test_that("print notes when devtools is ahead of CRAN", {
  local_reproducible_output(width = 60)
  x <- new_dev_sitrep(
    r_version = R_system_version("4.4.0"),
    r_path = "/usr/lib/R",
    devtools_version = package_version("2.5.0.9000"),
    devtools_cran_version = package_version("2.5.0")
  )
  expect_snapshot(print(x))
})

test_that("print warns when R is out of date", {
  local_reproducible_output(width = 60)
  x <- new_dev_sitrep(
    r_version = R_system_version("4.3.0"),
    r_path = "/usr/lib/R",
    r_release_version = R_system_version("4.4.0"),
    devtools_version = package_version("2.4.6")
  )
  expect_snapshot(print(x))
})

test_that("print warns about outdated devtools deps", {
  local_reproducible_output(width = 60)
  x <- new_dev_sitrep(
    r_version = R_system_version("4.4.0"),
    r_path = "/usr/lib/R",
    devtools_version = package_version("2.4.6"),
    devtools_deps = data.frame(
      package = c("rlang", "cli"),
      latest = c("1.0.0", "1.0.0"),
      installed = c("1.0.0", "0.5.0"),
      status = c("ok", "behind")
    )
  )
  expect_snapshot(print(x))
})

test_that("print warns about missing devtools deps", {
  local_reproducible_output(width = 60)
  x <- new_dev_sitrep(
    r_version = R_system_version("4.4.0"),
    r_path = "/usr/lib/R",
    devtools_version = package_version("2.4.6"),
    devtools_deps = data.frame(
      package = c("rlang", "somepkg"),
      latest = c("1.0.0", "1.0.0"),
      installed = c("1.0.0", NA_character_),
      status = c("ok", "missing")
    )
  )
  expect_snapshot(print(x))
})

test_that("print notes dev versions of devtools deps", {
  local_reproducible_output(width = 60)
  x <- new_dev_sitrep(
    r_version = R_system_version("4.4.0"),
    r_path = "/usr/lib/R",
    devtools_version = package_version("2.4.6"),
    devtools_deps = data.frame(
      package = "usethis",
      latest = "3.2.1",
      installed = "3.2.1.9000",
      status = "ahead"
    )
  )
  expect_snapshot(print(x))
})

test_that("print warns about outdated package deps", {
  local_reproducible_output(width = 60)
  x <- new_dev_sitrep(
    r_version = R_system_version("4.4.0"),
    r_path = "/usr/lib/R",
    devtools_version = package_version("2.4.6"),
    pkg = list(package = "mypkg", path = "/tmp/mypkg"),
    pkg_deps = data.frame(
      package = c("dplyr", "tidyr"),
      latest = c("1.1.0", "1.1.0"),
      installed = c("1.0.0", "1.0.0"),
      status = c("behind", "behind")
    )
  )
  expect_snapshot(print(x))
})

test_that("print notes dev versions of package deps", {
  local_reproducible_output(width = 60)
  x <- new_dev_sitrep(
    r_version = R_system_version("4.4.0"),
    r_path = "/usr/lib/R",
    devtools_version = package_version("2.4.6"),
    pkg = list(package = "mypkg", path = "/tmp/mypkg"),
    pkg_deps = data.frame(
      package = "usethis",
      latest = "3.2.1",
      installed = "3.2.1.9000",
      status = "ahead"
    )
  )
  expect_snapshot(print(x))
})

test_that("pkg_dep_status detects ahead packages", {
  local_mocked_bindings(
    pkg_deps = function(...) {
      data.frame(
        package = c("rlang", "cli"),
        version = c("0.0.1", "0.0.1")
      )
    },
    .package = "pak"
  )
  result <- pkg_dep_status("SOMEPACKAGE")
  expect_equal(result$status, c("ahead", "ahead"))
  expect_equal(result$latest, c("0.0.1", "0.0.1"))
})

test_that("pkg_dep_status detects ok packages", {
  local_mocked_bindings(
    pkg_deps = function(...) {
      data.frame(
        package = "rlang",
        version = as.character(packageVersion("rlang"))
      )
    },
    .package = "pak"
  )
  result <- pkg_dep_status("SOMEPACKAGE")
  expect_equal(result$status, "ok")
})

test_that("pkg_dep_status detects behind packages", {
  local_mocked_bindings(
    pkg_deps = function(...) {
      data.frame(
        package = "rlang",
        version = "99999.0.0"
      )
    },
    .package = "pak"
  )
  result <- pkg_dep_status("SOMEPACKAGE")
  expect_equal(result$status, "behind")
})

test_that("pkg_dep_status reports missing packages", {
  local_mocked_bindings(
    pkg_deps = function(...) {
      data.frame(
        package = c("rlang", "thereIsNoSuchPackage"),
        version = c("0.0.1", "1.0.0")
      )
    },
    .package = "pak"
  )
  result <- pkg_dep_status("SOMEPACKAGE")
  expect_equal(result$status, c("ahead", "missing"))
  expect_true(is.na(result$installed[[2]]))
})

test_that("pkg_dep_status filters out self", {
  pkg_path <- local_package_create()
  pkg_obj <- as.package(pkg_path)
  pkg_name <- pkg_obj$package

  deps <- data.frame(
    package = c(pkg_name, "rlang"),
    version = c("0.0.1", "99999.0.0")
  )

  local_mocked_bindings(
    local_dev_deps = function(...) deps,
    pkg_deps = function(...) deps,
    .package = "pak"
  )

  result <- pkg_dep_status(pkg_obj)
  expect_false(pkg_name %in% result$package)

  result <- pkg_dep_status(pkg_name)
  expect_false(pkg_name %in% result$package)
})

test_that("print shows RStudio update message", {
  local_reproducible_output(width = 60)
  withr::local_envvar(POSITRON = "")
  x <- new_dev_sitrep(
    r_version = R_system_version("4.4.0"),
    r_path = "/usr/lib/R",
    devtools_version = package_version("2.4.6"),
    rstudio_version = "2024.04.0",
    rstudio_msg = "RStudio is out of date."
  )
  expect_snapshot(print(x))
})

test_that("check_for_rstudio_updates", {
  skip_if_offline()
  skip_on_cran()
  withr::local_envvar(POSITRON = "")

  # the IDE ends up calling this with `os = "mac"` on macOS, but we would send
  # "darwin" in that case, so I test with "darwin"
  # also mix in some "windows"

  # returns nothing rstudio not needed
  expect_null(check_for_rstudio_updates("darwin", "1.0.0", FALSE))

  # returns nothing if the version is ahead of the current version
  expect_null(check_for_rstudio_updates("windows", "2030.12.0+123", TRUE))

  # returns something if ...
  local_edition(3)
  scrub_current_version <- function(message) {
    sub("(?<=^RStudio )[0-9\\.\\+]+", "{VERSION}", message, perl = TRUE)
  }

  # version is not understood by the service
  expect_snapshot(
    writeLines(check_for_rstudio_updates("windows", "haha-no-wut", TRUE))
  )

  # version is behind the current version

  # truly ancient
  expect_snapshot(
    writeLines(check_for_rstudio_updates("darwin", "0.0.1", TRUE)),
    transform = scrub_current_version
  )

  # Juliet Rose, does not have long_version, last before numbering changed
  expect_snapshot(
    writeLines(check_for_rstudio_updates("windows", "1.4.1717", TRUE)),
    transform = scrub_current_version
  )

  # new scheme, introduced 2021-08
  # YYYY.MM.<patch>[-(daily|preview)]+<build number>[.pro<pro suffix>]
  # YYY.MM is th expected date of release for dailies and previews

  # an out-of-date preview
  expect_snapshot(
    writeLines(check_for_rstudio_updates("darwin", "2021.09.1+372", TRUE)),
    transform = scrub_current_version
  )

  # an out-of-date daily
  expect_snapshot(
    writeLines(check_for_rstudio_updates(
      "windows",
      "2021.09.0-daily+328",
      TRUE
    )),
    transform = scrub_current_version
  )
})

#' Execute testthat tests in a package
#'
#' @description
#' * `test()` runs all tests in a package. It's a shortcut for
#'   [testthat::test_dir()]
#' * `test_active_file()` runs `test()` on the active file.
#' * `test_coverage()` computes test coverage for your package. It's a
#'   shortcut for [covr::package_coverage()] plus [covr::report()].
#' * `test_coverage_active_file()` computes test coverage for the active file. It's a
#'   shortcut for [covr::file_coverage()] plus [covr::report()].
#'
#' @template devtools
#' @param ... additional arguments passed to wrapped functions.
#' @param file One or more source or test files. If a source file the
#'   corresponding test file will be run. The default is to use the active file
#'   in RStudio (if available).
#' @inheritParams testthat::test_dir
#' @inheritParams pkgload::load_all
#' @inheritParams run_examples
#' @export
test <- function(
  pkg = ".",
  filter = NULL,
  stop_on_failure = FALSE,
  export_all = TRUE,
  ...
) {
  save_all()
  pkg <- as.package(pkg)

  if (!uses_testthat(pkg)) {
    cli::cli_inform(c(i = "No testing infrastructure found."))
    if (!interactive()) {
      cli::cli_bullets(c(
        "!" = 'Setup testing with {.run usethis::use_testthat()}.'
      ))
      return(invisible())
    }
    if (yesno("Create it?")) {
      return(invisible())
    }
    usethis_use_testthat(pkg)
    return(invisible())
  }

  cli::cli_inform(c(i = "Testing {.pkg {pkg$package}}"))
  withr::local_envvar(r_env_vars())

  load_package <- load_package_for_testing(pkg)
  testthat::test_local(
    pkg$path,
    filter = filter,
    stop_on_failure = stop_on_failure,
    load_package = load_package,
    ...
  )
}

#' @export
#' @rdname test
test_active_file <- function(file = find_active_file(), ...) {
  save_all()
  test_files <- find_test_file(file)
  pkg <- as.package(path_dir(test_files)[[1]])

  withr::local_envvar(r_env_vars())
  if (is_rstudio_running()) {
    rstudioapi::executeCommand("activateConsole", quiet = TRUE)
  }

  load_package <- load_package_for_testing(pkg)
  testthat::test_file(
    test_files,
    package = pkg$package,
    load_package = load_package,
    ...
  )
}

load_package_for_testing <- function(pkg) {
  if (pkg$package == "testthat") {
    # Must load testthat outside of testthat so tests are run with
    # dev testthat
    load_all(pkg$path, quiet = TRUE, helpers = FALSE)
    "none"
  } else {
    "source"
  }
}

#' @param report How to display the coverage report.
#'  * `"html"` opens an interactive report in the browser.
#'  * `"zero"` prints uncovered lines to the console.
#'  * `"silent"` returns the coverage object without display.
#'
#'  Defaults to `"html"` if interactive; otherwise to `"zero"`.
#' @export
#' @rdname test
test_coverage <- function(pkg = ".", report = NULL, ...) {
  check_installed("covr")

  save_all()
  pkg <- as.package(pkg)
  cli::cli_inform(c(i = "Computing test coverage for {.pkg {pkg$package}}"))

  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  withr::local_envvar(r_env_vars())
  coverage <- covr::package_coverage(pkg$path, ...)

  show_report(coverage, report = report, path = pkg$path)
}

#' @rdname test
#' @export
test_coverage_active_file <- function(
  file = find_active_file(),
  filter = TRUE,
  report = NULL,
  export_all = TRUE,
  ...
) {
  check_installed("covr")
  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  test_file <- find_test_file(file)
  test_dir <- path_dir(test_file)
  pkg <- as.package(test_dir)

  env <- load_all(pkg$path, quiet = TRUE, export_all = export_all)$env

  # Must capture before local_test_directory() sets TESTTHAT envvar,
  # which causes rlang::is_interactive() to return FALSE
  report <- report_default(report)

  # this always ends up using the package DESCRIPTION, which will refer
  # to the source package because of the load_all() above
  testthat::local_test_directory(test_dir, pkg$package)

  # To correctly simulate test_file() we need to set up both a temporary
  # snapshotter (with correct directory specification) for snapshot comparisons
  # and a stop reporter to inform the user about test failures
  snap_reporter <- testthat::local_snapshotter(
    snap_dir = file.path(test_dir, "_snaps")
  )
  snap_reporter$start_file(basename(test_file))
  reporter <- testthat::MultiReporter$new(
    reporters = list(
      testthat::StopReporter$new(praise = FALSE),
      snap_reporter
    )
  )

  withr::local_envvar(r_env_vars())
  testthat::with_reporter(reporter, {
    coverage <- covr::environment_coverage(env, test_file, ...)
    reporter$end_file() # needed to write new snapshots
  })

  if (isTRUE(filter)) {
    coverage_name <- name_source(covr::display_name(coverage))
    local_name <- name_test(test_file)
    coverage <- coverage[coverage_name %in% local_name]
  }

  # Use relative paths
  attr(coverage, "relative") <- TRUE
  attr(coverage, "package") <- pkg

  show_report(coverage, report = report, path = pkg$path)
}


#' Does a package use testthat?
#'
#' @export
#' @keywords internal
uses_testthat <- function(pkg = ".") {
  pkg <- as.package(pkg)

  paths <- c(
    path(pkg$path, "inst", "tests"),
    path(pkg$path, "tests", "testthat")
  )

  any(dir_exists(paths))
}

report_default <- function(report, call = caller_env()) {
  if (is.null(report)) {
    if (is_llm() || !is_interactive()) "zero" else "html"
  } else {
    arg_match(report, c("silent", "zero", "html"), error_call = call)
  }
}

show_report <- function(coverage, report, path, call = caller_env()) {
  report <- report_default(report, call = call)

  if (report == "html") {
    check_installed("DT")
    covered <- unique(covr::display_name(coverage))

    if (length(covered) == 1) {
      covr::file_report(coverage)
    } else {
      covr::report(coverage)
    }
  } else if (report == "zero") {
    zero <- covr::zero_coverage(coverage)
    if (nrow(zero) == 0) {
      cli::cli_inform(c(v = "All lines covered!"))
    } else {
      for (file in unique(zero$filename)) {
        file_zero <- zero[zero$filename == file, ]
        lines_by_fun <- split(file_zero$line, file_zero$functions)

        rel_path <- path_rel(file, path)
        cli::cli_inform("Uncovered lines in {.file {rel_path}}:")
        for (fun in names(lines_by_fun)) {
          lines <- paste0(collapse_lines(lines_by_fun[[fun]]), collapse = ", ")
          cli::cli_inform(c("*" = "{.fn {fun}}: {lines}"))
        }
      }
    }
  }
  invisible(coverage)
}

collapse_lines <- function(x) {
  x <- sort(unique(x))
  breaks <- c(0, which(diff(x) != 1), length(x))

  ranges <- character(length(breaks) - 1)
  for (i in seq_along(ranges)) {
    start <- x[breaks[i] + 1]
    end <- x[breaks[i + 1]]
    if (start == end) {
      ranges[i] <- as.character(start)
    } else {
      ranges[i] <- paste0(start, "-", end)
    }
  }
  ranges
}


#' Defunct functions
#'
#' These functions are defunct and will be removed in a future version of
#' devtools.
#' @name devtools-defunct
#' @keywords internal
NULL

#' @rdname devtools-defunct
#' @export
test_file <- function(file = find_active_file(), ...) {
  lifecycle::deprecate_stop("2.5.0", "test_file()", "test_active_file()")
}

#' @rdname devtools-defunct
#' @export
test_coverage_file <- function(file = find_active_file(), ...) {
  lifecycle::deprecate_stop(
    "2.5.0",
    "test_coverage_file()",
    "test_coverage_active_file()"
  )
}

test_test <- function(...) {
  suppressMessages(test(..., reporter = "silent"))
}
test_test_active_file <- function(...) {
  suppressMessages(test_active_file(..., reporter = "silent"))
}

test_that("Package can be tested with testthat not on search path", {
  pkg1 <- test_path("testTest")
  pkg2 <- test_path("testTestWithDepends")

  testthat_pos <- which(search() == "package:testthat")
  if (length(testthat_pos) > 0) {
    testthat_env <- detach(pos = testthat_pos)
    on.exit(attach(testthat_env, testthat_pos), add = TRUE)
  }

  test_test(pkg1)
  expect_true(TRUE)
  test_test(pkg2)
  expect_true(TRUE)
})

test_that("Filtering works with devtools::test", {
  out <- test_test(test_path("testTest"), filter = "dummy")
  expect_equal(length(out), 1)
})

test_that("devtools::test_active_file works", {
  out <- test_test_active_file(test_path(
    "testTest/tests/testthat/test-dummy.R"
  ))
  expect_equal(length(out), 1)
})

test_that("TESTTHAT_PKG environment variable is set", {
  withr::local_envvar("TESTTHAT_PKG" = "incorrect")

  test_test(
    test_path("testTest"),
    filter = "envvar",
    stop_on_failure = TRUE
  )
  test_test_active_file(
    test_path("testTest/tests/testthat/test-envvar.R"),
    stop_on_failure = TRUE
  )

  expect_true(TRUE)
})

test_that("stop_on_failure defaults to FALSE", {
  expect_error(
    test_test(test_path("testTestWithFailure")),
    NA
  )
  expect_error(
    test_test(test_path("testTestWithFailure"), stop_on_failure = TRUE),
    "Test failures"
  )
})

test_that("test_coverage_active_file() computes coverage", {
  skip_on_covr()
  pkg <- local_package_create()
  writeLines(
    c(
      "add <- function(x, y) x + y",
      "multiply <- function(x, y) x * y",
      "compute <- function(x) {",
      "  x + 1",
      "  x + 2",
      "}"
    ),
    file.path(pkg, "R", "math.R")
  )
  dir_create(file.path(pkg, "tests", "testthat"))
  writeLines(
    c(
      "test_that('add works', {",
      "  expect_equal(add(1, 2), 3)",
      "})"
    ),
    file.path(pkg, "tests", "testthat", "test-math.R")
  )

  expect_snapshot(test_coverage_active_file(
    file.path(pkg, "R", "math.R"),
    report = "zero"
  ))
})

test_that("test_coverage_active_file() reports full coverage", {
  skip_on_covr()
  pkg <- local_package_create()
  writeLines(
    "add <- function(x, y) x + y",
    file.path(pkg, "R", "math.R")
  )
  dir_create(file.path(pkg, "tests", "testthat"))
  writeLines(
    c(
      "test_that('add works', {",
      "  expect_equal(add(1, 2), 3)",
      "})"
    ),
    file.path(pkg, "tests", "testthat", "test-math.R")
  )

  expect_snapshot(test_coverage_active_file(
    file.path(pkg, "R", "math.R"),
    report = "zero"
  ))
})

test_that("report_default() does its job", {
  withr::local_options(rlang_interactive = FALSE)
  expect_equal(report_default(NULL), "zero")

  withr::local_options(rlang_interactive = TRUE)
  if (!is_llm()) {
    expect_equal(report_default(NULL), "html")
  }

  withr::local_envvar(AGENT = 1)
  expect_equal(report_default(NULL), "zero")

  expect_equal(report_default("silent"), "silent")
  expect_equal(report_default("zero"), "zero")
  expect_equal(report_default("html"), "html")
  expect_snapshot(report_default("bad"), error = TRUE)
})

context("Test")

test_that("Package can be tested with testthat not on search path", {
  testthat_pos <- which(search() == "package:testthat")
  if (length(testthat_pos) > 0) {
    testthat_env <- detach(pos = testthat_pos)
    on.exit(attach(testthat_env, testthat_pos), add = TRUE)
  }

  test("testTest", reporter = "stop")
  expect_true(TRUE)
  test("testTestWithDepends", reporter = "stop")
  expect_true(TRUE)
})

test_that("Filtering works with devtools::test", {
  test("testTest", filter = "dummy", reporter = "stop")
  expect_true(TRUE)
})

#' Run a script through some protocols such as http, https, ftp, etc
#'
#' If a SHA-1 hash is specified with the `sha1` argument, then this
#' function will check the SHA-1 hash of the downloaded file to make sure it
#' matches the expected value, and throw an error if it does not match. If the
#' SHA-1 hash is not specified, it will print a message displaying the hash of
#' the downloaded file. The purpose of this is to improve security when running
#' remotely-hosted code; if you have a hash of the file, you can be sure that
#' it has not changed. For convenience, it is possible to use a truncated SHA1
#' hash, down to 6 characters, but keep in mind that a truncated hash won't be
#' as secure as the full hash.
#'
#' @param url url
#' @param ... other options passed to [source()]
#' @param sha1 The (prefix of the) SHA-1 hash of the file at the remote URL.
#' @export
#' @seealso [source_gist()]
#' @examples
#' \dontrun{
#'
#' source_url("https://gist.github.com/hadley/6872663/raw/hi.r")
#'
#' # With a hash, to make sure the remote file hasn't changed
#' source_url("https://gist.github.com/hadley/6872663/raw/hi.r",
#'   sha1 = "54f1db27e60bb7e0486d785604909b49e8fef9f9")
#'
#' # With a truncated hash
#' source_url("https://gist.github.com/hadley/6872663/raw/hi.r",
#'   sha1 = "54f1db27e60")
#' }
source_url <- function(url, ..., sha1 = NULL) {
  check_string(url)
  check_installed(c("digest", "httr2"))

  temp_file <- file_temp()
  on.exit(file_delete(temp_file), add = TRUE)

  httr2::req_perform(httr2::request(url), path = temp_file)

  check_sha1(temp_file, sha1)

  check_dots_used(action = getOption("devtools.ellipsis_action", warn))
  source(temp_file, ...)
}

check_sha1 <- function(path, sha1) {
  file_sha1 <- digest::digest(file = path, algo = "sha1")

  if (is.null(sha1)) {
    cli::cli_inform(c(i = "SHA-1 hash of file is {.str {file_sha1}}"))
  } else {
    if (nchar(sha1) < 6) {
      cli::cli_abort(
        "{.arg sha1} must be at least 6 characters, not {nchar(sha1)}."
      )
    }

    # Truncate file_sha1 to length of sha1
    file_sha1 <- substr(file_sha1, 1, nchar(sha1))

    if (!identical(file_sha1, sha1)) {
      cli::cli_abort(
        "{.arg sha1} ({.str {sha1}}) doesn't match SHA-1 hash of downloaded file ({.str {file_sha1}})"
      )
    }
  }
}

#' Run a script on gist
#'
#' \dQuote{Gist is a simple way to share snippets and pastes with others.
#'   All gists are git repositories, so they are automatically versioned,
#'   forkable and usable as a git repository.}
#' <https://gist.github.com/>
#'
#' @param id either full url (character), gist ID (numeric or character of
#'   numeric).
#' @param ... other options passed to [source()]
#' @param filename if there is more than one R file in the gist, which one to
#' source (filename ending in '.R')? Default `NULL` will source the
#' first file.
#' @param sha1 The SHA-1 hash of the file at the remote URL. This is highly
#'   recommend as it prevents you from accidentally running code that's not
#'   what you expect. See [source_url()] for more information on
#'   using a SHA-1 hash.
#' @param quiet if `FALSE`, the default, prints informative messages.
#' @export
#' @seealso [source_url()]
#' @examples
#' \dontrun{
#' # You can run gists given their id
#' source_gist(6872663)
#' source_gist("6872663")
#'
#' # Or their html url
#' source_gist("https://gist.github.com/hadley/6872663")
#' source_gist("gist.github.com/hadley/6872663")
#'
#' # It's highly recommend that you run source_gist with the optional
#' # sha1 argument - this will throw an error if the file has changed since
#' # you first ran it
#' source_gist(6872663, sha1 = "54f1db27e60")
#' # Wrong hash will result in error
#' source_gist(6872663, sha1 = "54f1db27e61")
#'
#' #' # You can speficy a particular R file in the gist
#' source_gist(6872663, filename = "hi.r")
#' source_gist(6872663, filename = "hi.r", sha1 = "54f1db27e60")
#' }
source_gist <- function(id, ..., filename = NULL, sha1 = NULL, quiet = FALSE) {
  check_installed("gh")
  stopifnot(length(id) == 1)

  url_match <- "((^https://)|^)gist.github.com/([^/]+/)?([0-9a-f]+)$"
  if (grepl(url_match, id)) {
    # https://gist.github.com/kohske/1654919, https://gist.github.com/1654919,
    # or gist.github.com/1654919
    id <- regmatches(id, regexec(url_match, id))[[1]][5]
    url <- find_gist(id, filename)
  } else if (is.numeric(id) || grepl("^[0-9a-f]+$", id)) {
    # 1654919 or "1654919"
    url <- find_gist(id, filename)
  } else {
    cli::cli_abort("Invalid gist id specification {.str {id}}")
  }

  if (!quiet) {
    cli::cli_inform(c(i = "Sourcing gist {.str {id}}"))
  }

  check_dots_used(action = getOption("devtools.ellipsis_action", warn))
  source_url(url, ..., sha1 = sha1)
}

find_gist <- function(id, filename = NULL, call = parent.frame()) {
  files <- gh::gh("GET /gists/:id", id = id)$files
  r_files <- files[grepl("\\.[rR]$", names(files))]

  if (length(r_files) == 0) {
    cli::cli_abort("No R files found in gist", call = call)
  }

  if (!is.null(filename)) {
    if (
      !is.character(filename) ||
        length(filename) > 1 ||
        !grepl("\\.[rR]$", filename)
    ) {
      cli::cli_abort(
        "{.arg filename} must be {.code NULL}, or a single filename ending in .R/.r",
        call = call
      )
    }

    which <- match(tolower(filename), tolower(names(r_files)))
    if (is.na(which)) {
      cli::cli_abort("{.path {filename}} not found in gist", call = call)
    }
  } else {
    if (length(r_files) > 1) {
      cli::cli_inform(
        "{length(r_files)} .R files in gist, using first",
        call = call
      )
    }
    which <- 1
  }

  r_files[[which]]$raw_url
}

context("source_gist")

test_that("gist containing single file works unambiguously", {
  skip_on_cran()

  a <- 10
  source_gist("a65ddd06db40213f1a921237c55defbe", local = environment())
  expect_equal(a, 1)
})

test_that("gist with multiple files uses first with warning", {
  skip_on_cran()

  a <- 10
  expect_warning(source_gist("605a984e764f9ed358556b4ce48cbd08", local = environment()), "using first")
  expect_equal(a, 1)
})

test_that("can specify filename", {
  skip_on_cran()

  b <- 10
  source_gist("605a984e764f9ed358556b4ce48cbd08", filename = "b.r", local = environment())
  expect_equal(b, 2)
})

test_that("error if file doesn't exist or no files", {
  skip_on_cran()

  expect_error(
    source_gist("605a984e764f9ed358556b4ce48cbd08", filename = "c.r", local = environment()),
    "not found"
  )

  expect_error(
    source_gist("c535eee2d02e5f47c8e7642811bc327c"),
    "No R files found"
  )
})

test_that("gist containing single file works unambiguously", {
  skip_if_offline()
  skip_on_cran()
  skip_on_ci()

  a <- 10
  source_gist(
    "a65ddd06db40213f1a921237c55defbe",
    sha1 = "f176f5e1fe05b69b1ef799fdd1e4bac6341aff51",
    local = environment(),
    quiet = TRUE
  )
  expect_equal(a, 1)
})

test_that("gist with multiple files uses first with warning", {
  skip_if_offline()
  skip_on_cran()
  skip_on_ci()

  a <- 10
  expect_snapshot(
    source_gist(
      "605a984e764f9ed358556b4ce48cbd08",
      sha1 = "f176f5e1fe0",
      local = environment()
    )
  )
  expect_equal(a, 1)
})

test_that("errors with bad id", {
  expect_snapshot(source_gist("xxxx"), error = TRUE)
})

test_that("can specify filename", {
  skip_if_offline()
  skip_on_cran()
  skip_on_ci()

  b <- 10
  source_gist(
    "605a984e764f9ed358556b4ce48cbd08",
    filename = "b.r",
    sha1 = "8d1c53241c425a9a52700726809b7f2c164bde72",
    local = environment(),
    quiet = TRUE
  )
  expect_equal(b, 2)
})

test_that("error if file doesn't exist or no files", {
  skip_if_offline()
  skip_on_cran()
  skip_on_ci()

  expect_snapshot(error = TRUE, {
    find_gist("605a984e764f9ed358556b4ce48cbd08", 1)
    find_gist("605a984e764f9ed358556b4ce48cbd08", "c.r")
    find_gist("c535eee2d02e5f47c8e7642811bc327c")
  })
})

test_that("check_sha1() checks or reports sha1 as needed", {
  path <- withr::local_tempfile()
  writeBin("abc\n", path)

  expect_snapshot(error = TRUE, {
    check_sha1(path, NULL)
    check_sha1(path, "f")
    check_sha1(path, "ffffff")
  })
})

#' Reverse dependency tools.
#'
#' Tools to check and notify maintainers of all CRAN and Bioconductor
#' packages that depend on the specified package.
#'
#' The first run in a session will be time-consuming because it must download
#' all package metadata from CRAN and Bioconductor. Subsequent runs will
#' be faster.
#'
#' @param pkg Package name. This is unlike most devtools packages which
#'   take a path because you might want to determine dependencies for a package
#'   that you don't have installed. If omitted, defaults to the name of the
#'   current package.
#' @param ignore A character vector of package names to ignore. These packages
#'   will not appear in returned vector.
#' @param dependencies A character vector listing the types of dependencies
#'   to follow.
#' @param bioconductor If `TRUE` also look for dependencies amongst
#'   Bioconductor packages.
#' @param recursive If `TRUE` look for full set of recursive dependencies.
#' @seealso The [revdepcheck](https://github.com/r-lib/revdepcheck) package can
#'  be used to run R CMD check on all reverse dependencies.
#' @export
#' @examples
#' \dontrun{
#' revdep("ggplot2")
#'
#' revdep("ggplot2", ignore = c("xkcd", "zoo"))
#' }
revdep <- function(pkg,
                   dependencies = c("Depends", "Imports", "Suggests", "LinkingTo"),
                   recursive = FALSE, ignore = NULL,
                   bioconductor = FALSE) {
  if (missing(pkg)) pkg <- as.package(".")$package

  all <- if (bioconductor) packages() else cran_packages()

  deps <- tools::dependsOnPkgs(pkg, dependencies, recursive, installed = all)
  deps <- setdiff(deps, ignore)
  sort_ci(deps)
}

#' @rdname revdep
#' @export
revdep_maintainers <- function(pkg = ".") {
  if (missing(pkg)) pkg <- as.package(".")$package

  maintainers <- unique(packages()[revdep(pkg), "Maintainer"])
  class(maintainers) <- "maintainers"

  maintainers
}

#' @export
print.maintainers <- function(x, ...) {
  x <- gsub("\n", " ", x)
  cat(x, sep = ",\n")
  cat("\n")
}

# Package caches ----------------------------------------------------------

cran_packages <- memoise::memoise(
  function() {
    local <- file.path(tempdir(), "packages.rds")
    utils::download.file("https://cran.R-project.org/web/packages/packages.rds", local,
      mode = "wb", quiet = TRUE
    )
    on.exit(unlink(local))
    cp <- readRDS(local)
    rownames(cp) <- unname(cp[, 1])
    cp
  },
  ~memoise::timeout(30 * 60)
)

bioc_packages <- memoise::memoise(
  function(views = paste(BiocManager::repositories()[["BioCsoft"]], "VIEWS", sep = "/")) {
    con <- url(views)
    on.exit(close(con))
    bioc <- read.dcf(con)
    rownames(bioc) <- bioc[, 1]
    bioc
  },
  ~memoise::timeout(30 * 60)
)

packages <- function() {
  bioc <- bioc_packages()
  cran <- cran_packages()
  cols <- intersect(colnames(cran), colnames(bioc))
  rbind(cran[, cols], bioc[, cols])
}

#' Run `pkgdown::build_site()`
#'
#' This is a thin wrapper around [pkgdown::build_site()], used for generating
#' static HTML documentation. Learn more at <https://pkgdown.r-lib.org>.
#'
#' @param path Path to the package to build the static HTML.
#' @param ...  Additional arguments passed to [pkgdown::build_site()].
#' @export
build_site <- function(path = ".", ...) {
  check_installed("pkgdown")

  save_all()
  pkgdown::build_site(pkg = path, ...)
}

context("Build Site")

test_that("Package pkgdown site can be built ", {
  # This test requires internet
  skip_on_cran()

  destination <- file.path(tempdir(), "testPkgdown", "docs")

  build_output <- capture.output({
    build_site(
      path = "testPkgdown",
      override = list(destination = destination)
    )
  }, type = c("output"))

  build_output <- paste(build_output, collapse = "\n")

  expect_true(file.exists(file.path(destination, "index.html")),
    info = build_output,
    label = "created site index"
  )
  expect_true(file.exists(file.path(destination, "reference", "index.html")),
    info = build_output,
    label = "created reference index"
  )
  expect_true(file.exists(file.path(destination, "articles", "index.html")),
    info = build_output,
    label = "created articles index"
  )
  expect_true(file.exists(file.path(destination, "articles", "test.html")),
    info = build_output,
    label = "created articles index"
  )
})

#' Build and check a package
#'
#' @description
#' `check()` automatically builds and checks a source package, using all known
#' best practices. `check_built()` checks an already-built package.
#'
#' Passing `R CMD check` is essential if you want to submit your package to
#' CRAN: you must not have any ERRORs or WARNINGs, and you want to ensure that
#' there are as few NOTEs as possible.  If you are not submitting to CRAN, at
#' least ensure that there are no ERRORs or WARNINGs: these typically represent
#' serious problems.
#'
#' `check()` automatically builds a package before calling `check_built()`, as
#' this is the recommended way to check packages.  Note that this process runs
#' in an independent R session, so nothing in your current workspace will affect
#' the process. Under-the-hood, `check()` and `check_built()` rely on
#' [pkgbuild::build()] and [rcmdcheck::rcmdcheck()].
#'
#' @section Environment variables:
#'
#' Devtools does its best to set up an environment that combines best practices
#' with how check works on CRAN. This includes:
#'
#' * The standard environment variables set by devtools:
#' [r_env_vars()]. Of particular note for package tests is the `NOT_CRAN` env
#' var, which lets you know that your tests are running somewhere other than
#' CRAN, and hence can take a reasonable amount of time.
#'
#' * Debugging flags for the compiler, set by
#' [`compiler_flags(FALSE)`][compiler_flags()].
#'
#' * If `aspell` is found, `_R_CHECK_CRAN_INCOMING_USE_ASPELL_`
#' is set to `TRUE`. If no spell checker is installed, a warning is issued.
#'
#' * Environment variables, controlled by arguments `incoming`, `remote` and
#' `force_suggests`.
#'
#' @return An object containing errors, warnings, notes, and more.
#' @template devtools
#' @inheritParams rcmdcheck::rcmdcheck
#' @param document By default (`NULL`) will document if your installed
#'   roxygen2 version matches the version declared in the `DESCRIPTION`
#'   file. Use `TRUE` or `FALSE` to override the default.
#' @param build_args Additional arguments passed to `R CMD build`
#' @param ... Additional arguments passed on to [pkgbuild::build()].
#' @param vignettes If `FALSE`, do not build or check vignettes, equivalent to
#'   using `args = '--ignore-vignettes'` and `build_args = '--no-build-vignettes'`.
#' @param cleanup `r lifecycle::badge("deprecated")` See `check_dir` for details.
#' @seealso [release()] if you want to send the checked package to
#'   CRAN.
#' @export
check <- function(
  pkg = ".",
  document = NULL,
  build_args = NULL,
  ...,
  manual = FALSE,
  cran = TRUE,
  remote = FALSE,
  incoming = remote,
  force_suggests = FALSE,
  run_dont_test = FALSE,
  args = "--timings",
  env_vars = c(NOT_CRAN = "true"),
  quiet = FALSE,
  check_dir = NULL,
  cleanup = deprecated(),
  vignettes = TRUE,
  error_on = c("never", "error", "warning", "note")
) {
  pkg <- as.package(pkg)
  withr::local_options(list(warn = 1))

  save_all()

  if (lifecycle::is_present(cleanup)) {
    lifecycle::deprecate_stop("1.11.0", "check(cleanup = )")
  }

  if (missing(error_on) && !interactive()) {
    error_on <- "warning"
  }
  error_on <- match.arg(error_on)

  document <- document %||% can_document(pkg$roxygennote)
  if (document) {
    if (!quiet) {
      cli::cat_rule("Documenting", col = "cyan", line = 2)
    }
    document(pkg, quiet = quiet)
    if (!quiet) {
      cli::cat_line()
    }
  }

  if (!quiet) {
    cli::cat_rule("Building", col = "cyan", line = 2)
    show_env_vars(pkgbuild::compiler_flags(FALSE))
  }

  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  if (identical(vignettes, FALSE)) {
    args <- union(args, "--ignore-vignettes")
  }

  withr::with_envvar(pkgbuild::compiler_flags(FALSE), action = "prefix", {
    built_path <- pkgbuild::build(
      pkg$path,
      tempdir(),
      args = build_args,
      quiet = quiet,
      manual = manual,
      vignettes = vignettes,
      ...
    )
    on.exit(file_delete(built_path), add = TRUE)
  })

  check_built(
    built_path,
    cran = cran,
    remote = remote,
    incoming = incoming,
    force_suggests = force_suggests,
    run_dont_test = run_dont_test,
    manual = manual,
    args = args,
    env_vars = env_vars,
    quiet = quiet,
    check_dir = check_dir,
    error_on = error_on
  )
}

can_document <- function(
  required,
  installed = utils::packageVersion("roxygen2")
) {
  if (is.null(required)) {
    return(FALSE)
  }

  if (required != installed) {
    cli::cat_rule("Documenting", col = "red", line = 2)
    cli::cli_inform(c(
      i = "Installed roxygen2 version ({installed}) doesn't match declared ({required})",
      x = "{.fun check} will not re-document this package.",
      i = "Do you need to re-run {.fun document}?"
    ))
    FALSE
  } else {
    TRUE
  }
}

#' @export
#' @rdname check
#' @param path Path to built package.
#' @param cran if `TRUE` (the default), check using the same settings as CRAN
#'   uses. Because this is a moving target and is not uniform across all of
#'   CRAN's machine, this is on a "best effort" basis. It is more complicated
#'   than simply setting `--as-cran`.
#' @param remote Sets `_R_CHECK_CRAN_INCOMING_REMOTE_` env var. If `TRUE`,
#'   performs a number of CRAN incoming checks that require remote access.
#' @param incoming Sets `_R_CHECK_CRAN_INCOMING_` env var. If `TRUE`, performs a
#'   number of CRAN incoming checks.
#' @param force_suggests Sets `_R_CHECK_FORCE_SUGGESTS_`. If `FALSE` (the
#'   default), check will proceed even if all suggested packages aren't found.
#' @param run_dont_test Sets `--run-donttest` so that examples surrounded in
#'   `\donttest{}` are also run. When `cran = TRUE`, this only affects R 3.6 and
#'   earlier; in R 4.0, code in `\donttest{}` is always run as part of CRAN
#'   submission.
#' @param manual If `FALSE`, don't build and check manual (`--no-manual`).
#' @param env_vars Environment variables set during `R CMD check`
#' @param quiet if `TRUE` suppresses output from this function.
#' @param error_on Whether to throw an error on `R CMD check` failures. Note
#'   that the check is always completed (unless a timeout happens), and the
#'   error is only thrown after completion.
#'
#'   `error_on` is passed through to [rcmdcheck::rcmdcheck()], which is the
#'   definitive source for what the different values mean. If not specified by
#'   the user, both `check()` and `check_built()` default to `error_on =
#'   "never"` in interactive use and `"warning"` in a non-interactive setting.
check_built <- function(
  path = NULL,
  cran = TRUE,
  remote = FALSE,
  incoming = remote,
  force_suggests = FALSE,
  run_dont_test = FALSE,
  manual = FALSE,
  args = "--timings",
  env_vars = NULL,
  check_dir = tempdir(),
  quiet = FALSE,
  error_on = c("never", "error", "warning", "note")
) {
  if (missing(error_on) && !interactive()) {
    error_on <- "warning"
  }
  error_on <- match.arg(error_on)

  pkgname <- gsub("_.*?$", "", path_file(path))

  if (cran) {
    args <- c("--as-cran", args)
    env_vars <- c(
      "_R_CHECK_PACKAGES_USED_IGNORE_UNUSED_IMPORTS_" = as.character(FALSE),
      env_vars
    )
  }
  if (run_dont_test) {
    args <- c("--run-donttest", args)
  }

  if (manual && !pkgbuild::has_latex()) {
    cli::cli_inform(c(x = "pdflatex not found! Not building PDF manual"))
    manual <- FALSE
  }

  if (!manual) {
    args <- c(args, "--no-manual")
  }

  env_vars <- check_env_vars(cran, remote, incoming, force_suggests, env_vars)
  if (!quiet) {
    cli::cat_rule("Checking", col = "cyan", line = 2)
    show_env_vars(env_vars)
  }

  withr::with_envvar(env_vars, action = "replace", {
    rcmdcheck::rcmdcheck(
      path,
      quiet = quiet,
      args = args,
      check_dir = check_dir,
      error_on = error_on
    )
  })
}

check_env_vars <- function(
  cran = FALSE,
  remote = FALSE,
  incoming = remote,
  force_suggests = TRUE,
  env_vars = character()
) {
  c(
    aspell_env_var(),
    # Switch off expensive check for package version
    # https://github.com/r-lib/devtools/issues/1271
    if (
      getRversion() >= "3.4.0" && as.numeric(R.version[["svn rev"]]) >= 70944
    ) {
      c("_R_CHECK_CRAN_INCOMING_REMOTE_" = as.character(remote))
    },
    "_R_CHECK_CRAN_INCOMING_" = as.character(incoming),
    "_R_CHECK_FORCE_SUGGESTS_" = as.character(force_suggests),
    env_vars
  )
}

aspell_env_var <- function() {
  tryCatch(
    {
      utils::aspell(NULL, program = "aspell")
      c("_R_CHECK_CRAN_INCOMING_USE_ASPELL_" = "TRUE")
    },
    error = function(e) character()
  )
}

show_env_vars <- function(env_vars) {
  cli::cat_line("Setting env vars:", col = "darkgrey")
  cli::cat_bullet(
    paste0(format(names(env_vars)), ": ", unname(env_vars)),
    col = "darkgrey"
  )
}

#' Uninstall a local development package
#'
#' Uses `remove.packages()` to uninstall the package. To uninstall a package
#' from a non-default library, use in combination with [withr::with_libpaths()].
#'
#' @inheritParams install
#' @param unload if `TRUE` (the default), ensures the package is unloaded, prior
#'   to uninstalling.
#' @inheritParams utils::remove.packages
#' @export
#' @family package installation
#' @seealso [with_debug()] to install packages with debugging flags set.
uninstall <- function(
  pkg = ".",
  unload = TRUE,
  quiet = FALSE,
  lib = .libPaths()[[1]]
) {
  pkg <- as.package(pkg)

  if (unload && pkg$package %in% loaded_packages()$package) {
    pkgload::unload(pkg$package)
  }

  if (!quiet) {
    cli::cli_inform(c(i = "Uninstalling {.pkg {pkg$package}}"))
  }

  utils::remove.packages(pkg$package, .libPaths()[[1]])

  invisible(TRUE)
}

"%:::%" <- function(p, f) {
  get(f, envir = asNamespace(p))
}

is_windows <- isTRUE(.Platform$OS.type == "windows")
is_macos <- isTRUE(tolower(Sys.info()[["sysname"]]) == "darwin")

sort_ci <- function(x) {
  withr::with_collate("C", x[order(tolower(x), x)])
}

is_loaded <- function(pkg = ".") {
  pkg <- as.package(pkg)
  pkg$package %in% loadedNamespaces()
}

is_attached <- function(pkg = ".") {
  pkg <- as.package(pkg)

  !is.null(pkgload::pkg_env(pkg$package))
}

release_bullets <- function() {
  c(
    '`usethis::use_latest_dependencies(TRUE, "CRAN")`',
    NULL
  )
}

is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

is_positron <- function() {
  Sys.getenv("POSITRON") == "1"
}

is_rstudio_running <- function() {
  !is_testing() && rstudioapi::isAvailable()
}

# Copied from testthat:::is_llm()
is_llm <- function() {
  nzchar(Sys.getenv("AGENT")) ||
    nzchar(Sys.getenv("CLAUDECODE")) ||
    nzchar(Sys.getenv("GEMINI_CLI")) ||
    nzchar(Sys.getenv("CURSOR_AGENT"))
}

# Suppress cli wrapping
no_wrap <- function(x) {
  x <- gsub("{", "{{", x, fixed = TRUE)
  x <- gsub("}", "}}", x, fixed = TRUE)
  x <- gsub(" ", "\u00a0", x, fixed = TRUE)
  x <- gsub("\n", "\f", x, fixed = TRUE)
  x
}

context("sort")

test_that("case-insensitive sort order", {
  expect_equal(sort_ci(rev(letters)), letters)
  expect_equal(sort_ci(rev(LETTERS)), LETTERS)
  expect_equal(sort_ci(c(letters[1:3], LETTERS[1:3])), c("A", "a", "B", "b", "C", "c"))
})

test_that("reload works", {
  withr::local_temp_libpaths()

  pkg <- as.package(test_path("testTest"))

  install(pkg, quiet = TRUE)
  on.exit(unload(pkg$package), add = TRUE)

  expect_false(is_loaded(pkg))

  # Do nothing if the package is not loaded
  expect_error(reload(pkg), NA)
  expect_false(is_loaded(pkg))

  # Reload if loaded
  requireNamespace(pkg$package, quietly = TRUE)
  expect_true(is_loaded(pkg))
  reload(pkg)
  expect_true(is_loaded(pkg))

  # Re-attach if attached
  unload(pkg$package, quiet = TRUE)
  library(pkg$package, character.only = TRUE, quietly = TRUE)
  expect_true(is_loaded(pkg))
  expect_true(is_attached(pkg))
  reload(pkg)
  expect_true(is_loaded(pkg))
  expect_true(is_attached(pkg))
})

#' Save all documents in an active IDE session
#'
#' Helper function wrapping IDE-specific calls to save all documents in the
#' active session. In this form, callers of `save_all()` don't need to
#' execute any IDE-specific code. This function can be extended to include
#' other IDE implementations of their equivalent
#' `rstudioapi::documentSaveAll()` methods.
#' @return NULL
save_all <- function() {
  if (rstudioapi::hasFun("documentSaveAll")) {
    rstudioapi::documentSaveAll()
  }
}

# Some helpers around usethis functions

# we need to import some usethis function so the namespace is loaded when
# devtools is loaded, but not attached.
#' @importFrom usethis use_test
NULL

usethis_use_testthat <- function(pkg) {
  usethis::local_project(pkg$path, quiet = TRUE)
  usethis::use_testthat()
}

usethis_use_directory <- function(pkg, path, ignore = FALSE) {
  usethis::local_project(pkg$path, quiet = TRUE)
  usethis::use_directory(path, ignore)
}

usethis_use_git_ignore <- function(pkg, ignores, ignore = FALSE) {
  usethis::local_project(pkg$path, quiet = TRUE)
  usethis::use_git_ignore(ignores)
}

#' Check documentation, as `R CMD check` does
#'
#' This function attempts to run the documentation related checks in the
#' same way that `R CMD check` does. Unfortunately it can't run them
#' all because some tests require the package to be loaded, and the way
#' they attempt to load the code conflicts with how devtools does it.
#'
#' @template devtools
#' @return Nothing. This function is called purely for it's side effects: if
#' no errors there will be no output.
#' @export
#' @examples
#' \dontrun{
#' check_man("mypkg")
#' }
check_man <- function(pkg = ".") {
  pkg <- as.package(pkg)
  document(pkg)

  old <- options(warn = -1)
  on.exit(options(old))

  cli::cli_inform(c(i = "Checking documentation..."))

  check_Rd_contents <- if (getRversion() < "4.1") {
    asNamespace("tools")$.check_Rd_contents
  } else {
    asNamespace("tools")$checkRdContents
  }

  ok <-
    all(
      man_message(("tools" %:::% ".check_package_parseRd")(dir = pkg$path)),
      man_message(("tools" %:::% ".check_Rd_metadata")(dir = pkg$path)),
      man_message(("tools" %:::% ".check_Rd_xrefs")(dir = pkg$path)),
      man_message(check_Rd_contents(dir = pkg$path)),
      man_message(tools::checkDocFiles(dir = pkg$path)),
      man_message(tools::checkDocStyle(dir = pkg$path)),
      man_message(tools::checkReplaceFuns(dir = pkg$path)),
      man_message(tools::checkS3methods(dir = pkg$path)),
      man_message(tools::undoc(dir = pkg$path))
    )

  if (ok) {
    cli::cli_inform(c(v = "No issues detected"))
  }

  invisible()
}

man_message <- function(x) {
  if (inherits(x, "undoc") && length(x$code) == 0) {
    # Returned by tools::undoc()
    TRUE
  } else if ("bad" %in% names(x) && length(x$bad) == 0) {
    # Returned by check_Rd_xrefs()
    TRUE
  } else if (length(x) == 0) {
    TRUE
  } else {
    print(x)
    FALSE
  }
}

#' Check for missing documentation fields
#'
#' Checks all Rd files in `man/` and looks for any that have a `\usage` section
#' (i.e. a function) but that *don't* have `\value` and `\examples` sections.
#' These missing fields are flagged by CRAN on initial submission.
#'
#' @template devtools
#' @param fields A character vector of Rd field names to check for.
#' @returns A named list of character vectors, one for each field, containing
#'   the names of Rd files missing that field. Returned invisibly.
#' @export
#' @examples
#' \dontrun{
#' check_doc_fields(".")
#' }
check_doc_fields <- function(pkg = ".", fields = c("value", "examples")) {
  pkg <- as.package(pkg)
  fields <- stats::setNames(fields, fields)

  paths <- dir_ls(path(pkg$path, "man"), regexp = "\\.Rd$")
  names(paths) <- path_rel(paths, pkg$path)
  rd <- map(paths, tools::parse_Rd, permissive = TRUE)
  rd_tags <- map(rd, \(x) unlist(map(x, attr, "Rd_tag")))

  has_tag <- function(tags, this) {
    any(paste0("\\", this) %in% tags)
  }

  has_usage <- map_lgl(rd_tags, has_tag, this = "usage")
  rd_tags <- rd_tags[has_usage]

  results <- map(fields, function(field) {
    missing <- !map_lgl(rd_tags, has_tag, this = field)
    names(rd_tags)[missing]
  })

  for (field in fields) {
    missing <- results[[field]]
    if (length(missing) > 0) {
      cli::cli_inform(c(
        "!" = "Missing {.code \\{field}} section in {length(missing)} file{?s}:",
        stats::setNames(missing, rep("*", length(missing)))
      ))
    } else {
      cli::cli_inform(c("v" = "All Rd files have a {.code \\{field}} section."))
    }
  }

  invisible(results)
}

test_that("check_doc_fields output - missing fields", {
  pkg <- local_package_create()
  dir.create(file.path(pkg, "man"))

  writeLines(
    "\\name{foo}\\title{Foo}\\usage{foo()}\\description{A function.}\\examples{foo()}",
    file.path(pkg, "man", "foo.Rd")
  )
  writeLines(
    "\\name{bar}\\title{Bar}\\usage{foo()}\\description{A function.}\\value{Something.}",
    file.path(pkg, "man", "bar.Rd")
  )

  expect_snapshot(check_doc_fields(pkg))
})

test_that("check_doc_fields output - all present", {
  pkg <- local_package_create()
  dir.create(file.path(pkg, "man"))

  writeLines(
    "\\name{foo}\\title{Foo}\\usage{foo()}\\description{A function.}\\value{A value.}\\examples{foo()}",
    file.path(pkg, "man", "foo.Rd")
  )

  expect_snapshot(check_doc_fields(pkg))
})

test_that("check_man works", {
  # tools:::.check_Rd_xrefs which is called by `check_man()` assumes the base
  # and recommended packages will all be in the library path, which is not the
  # case during R CMD check, so we only run these tests interactively

  skip_if_not(interactive())

  pkg <- local_package_create()
  dir.create(file.path(pkg, "man"))
  writeLines(
    c(
      "
\\name{foo}
\\title{Foo bar}
\\usage{
foo(x)
}
\\arguments{\\item{foo}{}}
"
    ),
    file.path(pkg, "man", "foo.Rd")
  )

  expect_snapshot(
    check_man(pkg),
    transform = \(x) gsub(basename(pkg), "{package}", x, fixed = TRUE)
  )
})

#' Show package news
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `show_news()` is deprecated because we no longer use or recommend this
#' workflow.
#'
#' @template devtools
#' @param latest if `TRUE`, only show the news for the most recent
#'   version.
#' @param ... other arguments passed on to `news`
#' @export
#' @keywords internal
show_news <- function(pkg = ".", latest = TRUE, ...) {
  lifecycle::deprecate_warn("2.5.0", "show_news()")
  pkg <- as.package(pkg)
  news_path <- find_news(pkg$path)

  withCallingHandlers(
    {
      news_db <- switch(
        path_ext(news_path),
        Rd = ("tools" %:::% ".build_news_db_from_package_NEWS_Rd")(news_path),
        md = ("tools" %:::% ".build_news_db_from_package_NEWS_md")(news_path),
        ("tools" %:::% ".news_reader_default")(news_path)
      )
    },
    error = function(e) {
      cli::cli_abort("Failed to parse NEWS file.", parent = e)
    }
  )

  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  out <- utils::news(..., db = news_db)
  if (latest) {
    ver <- numeric_version(out$Version)
    recent <- ver == max(ver)
    structure(out[recent, ], class = class(out), bad = attr(out, "bad")[recent])
  } else {
    out
  }
}

find_news <- function(path) {
  news_paths <- c(
    file.path(path, "inst", "NEWS.Rd"),
    file.path(path, "NEWS.md"),
    file.path(path, "NEWS"),
    file.path(path, "inst", "NEWS")
  )
  news_path <- news_paths[file_exists(news_paths)]

  if (length(news_path) == 0) {
    cli::cli_abort("Failed to find NEWS file.")
  }

  path_abs(news_path[[1]])
}

test_that("find_news() finds NEWS in all expected locations", {
  pkg <- local_package_create()

  dir_create(pkg, "inst")
  file_create(pkg, "inst", "NEWS")
  expect_equal(path_rel(find_news(pkg), pkg), path("inst/NEWS"))

  file_create(pkg, "NEWS")
  expect_equal(path_rel(find_news(pkg), pkg), path("NEWS"))

  file_create(pkg, "NEWS.md")
  expect_equal(path_rel(find_news(pkg), pkg), path("NEWS.md"))

  file_create(pkg, "inst", "NEWS.Rd")
  expect_equal(path_rel(find_news(pkg), pkg), path("inst/NEWS.Rd"))
})

test_that("fails when NEWS is missing or improperly formatted", {
  withr::local_options(lifecycle_verbosity = "quiet")
  skip_on_cran()
  skip_unless_r(">= 4.2.0") # different error message

  pkg <- local_package_create()
  expect_snapshot(show_news(pkg), error = TRUE)

  dir_create(pkg, "inst")
  file_create(pkg, "inst", "NEWS.Rd")

  expect_snapshot(show_news(pkg), error = TRUE)
})

#' Check a package on macOS
#'
#' Check on either the released or development versions of R, using
#' <https://mac.r-project.org/macbuilder/submit.html>.
#'
#'
#' @template devtools
#' @inheritParams check_win
#' @param dep_pkgs Additional custom dependencies to install prior to checking
#'   the package.
#' @family build functions
#' @return The url with the check results (invisibly)
#' @export
check_mac_release <- function(
  pkg = ".",
  dep_pkgs = character(),
  args = NULL,
  manual = TRUE,
  quiet = FALSE,
  ...
) {
  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  check_mac(
    pkg = pkg,
    version = "R-release",
    dep_pkgs = dep_pkgs,
    args = args,
    manual = manual,
    quiet = quiet,
    ...
  )
}

#' @rdname check_mac_release
#' @export
check_mac_devel <- function(
  pkg = ".",
  dep_pkgs = character(),
  args = NULL,
  manual = TRUE,
  quiet = FALSE,
  ...
) {
  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  check_mac(
    pkg = pkg,
    version = "R-devel",
    dep_pkgs = dep_pkgs,
    args = args,
    manual = manual,
    quiet = quiet,
    ...
  )
}

check_mac <- function(
  pkg = ".",
  version = c("R-devel", "R-release"),
  dep_pkgs = character(),
  args = NULL,
  manual = TRUE,
  quiet = FALSE,
  ...
) {
  pkg <- as.package(pkg)

  version <- match.arg(version, several.ok = FALSE)

  if (!quiet) {
    cli::cli_inform(c(
      "Checking macOS version of {.pkg {pkg$package}} ({pkg$version})",
      i = "Using https://mac.r-project.org/macbuilder/submit.html."
    ))
  }

  built_path <- pkgbuild::build(
    pkg$path,
    tempdir(),
    args = args,
    manual = manual,
    quiet = quiet,
    ...
  )

  dep_built_paths <- character()
  for (i in seq_along(dep_pkgs)) {
    dep_pkg <- as.package(dep_pkgs[[i]])$path
    dep_built_paths[[i]] <- pkgbuild::build(
      dep_pkg,
      tempdir(),
      args = args,
      manual = manual,
      quiet = quiet,
      ...
    )
  }
  on.exit(file_delete(c(built_path, dep_built_paths)), add = TRUE)

  url <- "https://mac.r-project.org/macbuilder/v1/submit"

  check_installed("httr2")
  body <- list(
    pkgfile = curl::form_file(built_path),
    rflavor = tolower(version)
  )

  if (length(dep_built_paths) > 0) {
    uploads <- map(dep_built_paths, curl::form_file)
    names(uploads) <- rep("depfiles", length(uploads))
    body <- append(body, uploads)
  }

  req <- httr2::request(url) |>
    httr2::req_body_multipart(!!!body) |>
    httr2::req_headers(Accept = "application/json")
  resp <- httr2::req_perform(req)

  response_url <- httr2::resp_body_json(resp)$url

  if (!quiet) {
    time <- strftime(Sys.time() + 10 * 60, "%I:%M %p")

    cli::cat_rule(col = "cyan")
    cli::cli_inform(c(
      i = "Check {.url {response_url}} for the results in 5-10 mins (~{time})."
    ))
  }

  invisible(response_url)
}

#' Find missing s3 exports
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `missing_s3()` is deprecated because roxygen2 now provides the same
#' functionality. Run `devtools::document()` and look for
#' `"Missing documentation for S3 method"` warnings.
#'
#' @template devtools
#' @export
#' @keywords internal
missing_s3 <- function(pkg = ".") {
  lifecycle::deprecate_warn("2.5.0", "missing_s3()")
  pkg <- as.package(pkg)
  loaded <- load_all(pkg$path)

  # Find all S3 methods in package
  objs <- ls(envir = loaded$env)
  is_s3 <- function(x) roxygen2::is_s3_method(x, env = loaded$env)
  s3_objs <- keep(objs, is_s3)

  # Find all S3 methods in NAMESPACE
  ns <- pkgload::parse_ns_file(pkg$path)
  exports <- paste(ns$S3methods[, 1], ns$S3methods[, 2], sep = ".")

  setdiff(s3_objs, exports)
}

#' Release package to CRAN
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `release()` is deprecated in favour of [usethis::use_release_issue()].
#' We no longer feel confident recommending `release()` because we don't use it
#' ourselves, so there's no guarantee that it will track best practices as
#' they evolve over time.
#'
#' If you want to programmatical submit to CRAN, you can continue to use
#' [submit_cran()].
#'
#' @template devtools
#' @param check if `TRUE`, run checking, otherwise omit it.  This
#'   is useful if you've just checked your package and you're ready to
#'   release it.
#' @param args An optional character vector of additional command
#'   line arguments to be passed to `R CMD build`.
#' @seealso [usethis::use_release_issue()] to create a checklist of release
#'   tasks that you can use in addition to or in place of `release`.
#' @export
release <- function(pkg = ".", check = FALSE, args = NULL) {
  lifecycle::deprecate_warn(
    "2.5.0",
    "release()",
    "usethis::use_release_issue()"
  )
  if (!is_interactive()) {
    cli::cli_abort("Interactive session required.")
  }

  pkg <- as.package(pkg)
  # Figure out if this is a new package
  cran_version <- cran_pkg_version(pkg$package)
  new_pkg <- is.null(cran_version)

  if (yesno("Have you checked for spelling errors (with `spell_check()`)?")) {
    return(invisible())
  }

  if (check) {
    cli::cat_rule(
      left = "Building and checking",
      right = pkg$package,
      line = 2
    )
    check(
      pkg,
      cran = TRUE,
      remote = TRUE,
      manual = TRUE,
      build_args = args,
      run_dont_test = TRUE
    )
  }
  if (yesno("Have you run `R CMD check` locally?")) {
    return(invisible())
  }

  release_checks(pkg)
  if (yesno("Were devtool's checks successful?")) {
    return(invisible())
  }

  if (!new_pkg) {
    show_cran_check <- TRUE
    cran_details <- NULL
    end_sentence <- " ?"
    if (requireNamespace("foghorn", quietly = TRUE)) {
      show_cran_check <- has_cran_results(pkg$package)
      cran_details <- foghorn::cran_details(pkg = pkg$package)
    }
    if (show_cran_check) {
      if (!is.null(cran_details)) {
        end_sentence <- "\n shown above?"
        cli::cat_rule(paste0(
          "Details of the CRAN check results for ",
          pkg$package
        ))
        summary(cran_details)
        cli::cat_rule()
      }
      cran_url <- paste0(
        cran_mirror(),
        "/web/checks/check_results_",
        pkg$package,
        ".html"
      )
      if (
        yesno(
          "Have you fixed all existing problems at \n{cran_url}{end_sentence}"
        )
      ) {
        return(invisible())
      }
    }
  }

  if (yesno("Have you checked on R-hub (see `?rhub::rhubv2`)?")) {
    return(invisible())
  }

  if (yesno("Have you checked on win-builder (with `check_win_devel()`)?")) {
    return(invisible())
  }

  deps <- if (new_pkg) 0 else length(revdep(pkg$package))
  if (deps > 0) {
    msg <- paste0(
      "Have you checked the ",
      deps,
      " reverse dependencies ",
      "(with the revdepcheck package)?"
    )
    if (yesno(msg)) {
      return(invisible())
    }
  }

  questions <- c(
    "Have you updated `NEWS.md` file?",
    "Have you updated `DESCRIPTION`?",
    "Have you updated `cran-comments.md?`",
    if (file_exists("codemeta.json")) {
      "Have you updated codemeta.json with codemetar::write_codemeta()?"
    },
    find_release_questions(pkg)
  )
  for (question in questions) {
    if (yesno(question)) return(invisible())
  }

  if (uses_git(pkg$path)) {
    git_checks(pkg)
    if (yesno("Were Git checks successful?")) {
      return(invisible())
    }
  }

  submit_cran(pkg, args = args)

  invisible(TRUE)
}

has_cran_results <- function(pkg) {
  cran_res <- foghorn::cran_results(
    pkg = pkg,
    show = c("error", "fail", "warn", "note")
  )
  sum(cran_res[, -1]) > 0
}

find_release_questions <- function(pkg = ".") {
  pkg <- as.package(pkg)

  q_fun <- pkgload::ns_env(pkg$package)$release_questions
  if (is.null(q_fun)) {
    character()
  } else {
    q_fun()
  }
}

yesno <- function(msg, .envir = parent.frame()) {
  if (!is_interactive()) {
    cli::cli_abort("Called from non-interactive context.")
  }

  yeses <- c(
    "Yes",
    "Definitely",
    "For sure",
    "Yup",
    "Yeah",
    "Of course",
    "Absolutely"
  )
  nos <- c("No way", "Not yet", "I forget", "No", "Nope", "Uhhhh... Maybe?")

  cli::cli_inform(msg, .envir = .envir)
  qs <- c(sample(yeses, 1), sample(nos, 2))
  rand <- sample(length(qs))

  utils::menu(qs[rand]) != which(rand == 1)
}

maintainer <- function(pkg = ".") {
  pkg <- as.package(pkg)

  authors <- pkg$`authors@r`
  if (!is.null(authors)) {
    people <- eval(parse(text = authors))
    if (is.character(people)) {
      maintainer <- utils::as.person(people)
    } else {
      maintainer <- Find(function(x) "cre" %in% x$role, people)
    }
  } else {
    maintainer <- pkg$maintainer
    if (is.null(maintainer)) {
      cli::cli_abort("No maintainer defined in package.")
    }
    maintainer <- utils::as.person(maintainer)
  }

  list(
    name = paste(maintainer$given, maintainer$family),
    email = maintainer$email
  )
}

cran_comments <- function(pkg = ".", call = parent.frame()) {
  pkg <- as.package(pkg)

  path <- path(pkg$path, "cran-comments.md")
  if (!file_exists(path)) {
    cli::cli_warn(
      c(
        x = "Can't find {.file cran-comments.md}.",
        i = "This file is used to communicate your release process to the CRAN team.",
        i = "Create it with {.run usethis::use_cran_comments()}."
      ),
      call = call
    )
    return(character())
  }

  paste0(readLines(path, warn = FALSE), collapse = "\n")
}

cran_submission_url <- "https://xmpalantir.wu.ac.at/cransubmit/index2.php"

#' Submit a package to CRAN
#'
#' @description
#' This submits your package to CRAN using the web-form submission process.
#' To complete the submission you will need respond to the email sent to the
#' maintainer email address.
#'
#' We generally recommend using this part of the process defined by
#' [use_release_issue()]; this process maximizes the chances of a successful
#' submission.
#'
#' @template devtools
#' @inheritParams release
#' @export
#' @keywords internal
submit_cran <- function(pkg = ".", args = NULL) {
  if (yesno("Is your email address {maintainer(pkg)$email}?")) {
    return(invisible())
  }

  pkg <- as.package(pkg)

  built_path <- pkgbuild::build(pkg$path, tempdir(), manual = TRUE, args = args)

  size <- format(as.object_size(file_info(built_path)$size), units = "auto")
  cli::cat_rule("Submitting", col = "cyan")
  cli::cli_inform(c(
    "i" = "Path {.file {built_path}}",
    "i" = "File size: {size}"
  ))
  cli::cat_line()

  if (yesno("Ready to submit {pkg$package} ({pkg$version}) to CRAN?")) {
    return(invisible())
  }

  upload_cran(pkg, built_path)

  usethis::with_project(pkg$path, flag_release(pkg))
}

extract_cran_msg <- function(msg) {
  # Remove "CRAN package Submission" and "Submit package to CRAN"
  msg <- gsub("CRAN package Submission|Submit package to CRAN", "", msg)

  # remove all html tags
  msg <- gsub("<[^>]+>", "", msg)

  # remove tabs
  msg <- gsub("\t+", "", msg)

  # Remove extra newlines
  msg <- gsub("\n+", "\n", msg)

  msg
}

upload_cran <- function(pkg, built_path, call = parent.frame()) {
  check_installed("httr2")

  pkg <- as.package(pkg)
  maint <- maintainer(pkg)
  comments <- cran_comments(pkg, call = call)

  # Initial upload ---------
  cli::cli_inform(c(i = "Uploading package & comments"))
  req <- httr2::request(cran_submission_url)
  req <- httr2::req_body_multipart(
    req,
    pkg_id = "",
    name = maint$name,
    email = maint$email,
    uploaded_file = curl::form_file(built_path, "application/x-gzip"),
    comment = comments,
    upload = "Upload package"
  )
  resp <- httr2::req_perform(req)
  new_url <- httr2::url_parse(httr2::resp_url(resp))

  # Confirmation -----------
  cli::cli_inform(c(i = "Confirming submission"))

  req <- httr2::request(cran_submission_url)
  req <- httr2::req_body_multipart(
    req,
    pkg_id = new_url$query$pkg_id,
    name = maint$name,
    email = maint$email,
    policy_check = "1/",
    submit = "Submit package"
  )

  resp <- httr2::req_perform(req)
  new_url <- httr2::url_parse(httr2::resp_url(resp))
  if (new_url$query$submit == "1") {
    cli::cli_inform(c(
      "v" = "Package submission successful",
      "i" = "Check your email for confirmation link."
    ))
  } else {
    cli::cli_abort("Package failed to upload.", call = call)
  }

  invisible(TRUE)
}

as.object_size <- function(x) structure(x, class = "object_size")

flag_release <- function(pkg = ".") {
  pkg <- as.package(pkg)
  if (!uses_git(pkg$path)) {
    return(invisible())
  }

  cli::cli_inform(c(
    "!" = "Don't forget to tag this release once accepted by CRAN"
  ))

  withr::with_dir(pkg$path, {
    sha <- system2("git", c("rev-parse", "HEAD"), stdout = TRUE)
  })

  dat <- list(
    Version = pkg$version,
    Date = format(Sys.time(), tz = "UTC", usetz = TRUE),
    SHA = sha
  )

  write.dcf(dat, file = path(pkg$path, "CRAN-SUBMISSION"))
  usethis::use_build_ignore("CRAN-SUBMISSION")
}

cran_mirror <- function(repos = getOption("repos")) {
  repos[repos == "@CRAN@"] <- "https://cloud.r-project.org"

  if (is.null(names(repos))) {
    names(repos) <- "CRAN"
  }

  repos[["CRAN"]]
}

# Return the version of a package on CRAN (or other repository)
# @param package The name of the package.
# @param available A matrix of information about packages.
cran_pkg_version <- function(package, available = utils::available.packages()) {
  idx <- available[, "Package"] == package
  if (any(idx)) {
    as.package_version(available[package, "Version"])
  } else {
    NULL
  }
}

#' Lint all source files in a package
#'
#' The default linters correspond to the style guide at
#' <https://style.tidyverse.org/>, however it is possible to override any or all
#' of them using the `linters` parameter.
#'
#' @template devtools
#' @param cache Store the lint results so repeated lints of the same content use
#'   the previous results. Consult the lintr package to learn more about its
#'   caching behaviour.
#' @param ... Additional arguments passed to [lintr::lint_package()].
#' @seealso [lintr::lint_package()], [lintr::lint()]
#' @export
lint <- function(pkg = ".", cache = TRUE, ...) {
  check_installed("lintr")
  pkg <- as.package(pkg)

  cli::cli_inform(c(i = "Linting {.pkg {pkg$package}}"))

  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  lintr::lint_package(pkg$path, cache = cache, ...)
}

#' Create a package
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `create()` is deprecated. Please use [usethis::create_package()] directly
#' instead.
#'
#' @param path A path. If it exists, it is used. If it does not exist, it is
#'   created, provided that the parent path exists.
#' @param ... Additional arguments passed to [usethis::create_package()]
#' @inheritParams usethis::create_package
#' @return The path to the created package, invisibly.
#' @export
#' @keywords internal
create <- function(path, ..., open = FALSE) {
  lifecycle::deprecate_warn("2.5.0", "create()", "usethis::create_package()")
  usethis::create_package(path, ..., open = open)
}

#' Coerce input to a package
#'
#' Possible specifications of package:
#' \itemize{
#'   \item path
#'   \item package object
#' }
#' @param x object to coerce to a package
#' @param create `r lifecycle::badge("deprecated")` Hasn't worked for some time.
#' @export
#' @keywords internal
as.package <- function(x = NULL, create = deprecated()) {
  if (is.package(x)) {
    return(x)
  }
  if (lifecycle::is_present(create)) {
    lifecycle::deprecate_warn("2.5.0", "as.package(create = )")
  }

  x <- package_file(path = x)
  load_pkg_description(x)
}

#' Find file in a package
#'
#' It always starts by walking up the path until it finds the root directory,
#' i.e. a directory containing `DESCRIPTION`. If it cannot find the root
#' directory, or it can't find the specified path, it will throw an error.
#'
#' @param ... Components of the path.
#' @param path Place to start search for package directory.
#' @keywords internal
#' @export
#' @examples
#' \dontrun{
#' package_file("figures", "figure_1")
#' }
package_file <- function(..., path = ".") {
  check_string(path)
  if (!dir_exists(path)) {
    cli::cli_abort("{.path {path}} is not a directory.")
  }

  base_path <- path
  path <- strip_slashes(path_real(path))

  # Walk up to root directory
  while (!has_description(path)) {
    path <- path_dir(path)

    if (is_root(path)) {
      cli::cli_abort(c(
        "Could not find package root.",
        i = "Is {.path {base_path}} inside a package?"
      ))
    }
  }

  path(path, ...)
}

has_description <- function(path) {
  file_exists(path(path, "DESCRIPTION"))
}

is_root <- function(path) {
  identical(path, path_dir(path))
}

strip_slashes <- function(x) {
  x <- sub("/*$", "", x)
  x
}

# Load package DESCRIPTION into convenient form.
load_pkg_description <- function(path) {
  path_desc <- path(path, "DESCRIPTION")

  info <- read.dcf(path_desc)[1, ]
  Encoding(info) <- 'UTF-8'
  desc <- as.list(info)
  names(desc) <- tolower(names(desc))
  desc$path <- path

  structure(desc, class = "package")
}

#' Is the object a package?
#'
#' @keywords internal
#' @export
is.package <- function(x) inherits(x, "package")

# Mockable variant of interactive
interactive <- function() .Primitive("interactive")()

test_that("package_file() gives useful errors", {
  expect_snapshot(error = TRUE, {
    package_file(path = 1)
    package_file(path = "doesntexist")
    package_file(path = "/")
  })
})

test_that("create argument is deprecated", {
  path <- local_package_create()
  expect_snapshot(x <- as.package(path, create = TRUE))
})

#' Activate and deactivate development mode
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' We no longer recommend `dev_mode()` and it will be removed in a future
#' release of devtools. Instead, we now rely on [load_all()] to test drive an
#' in-development package. If you really like the idea of corralling
#' experimental packages in a special library, you might enjoy
#' `withr::local_libpaths()`. If you are concerned about different projects
#' interfering with each other through the use of a shared library, consider
#' using the [renv package](https://rstudio.github.io/renv/).
#'
#' Original description: When activated, `dev_mode` creates a new library for
#' storing installed packages. This new library is automatically created when
#' `dev_mode` is activated if it does not already exist. This allows you to test
#' development packages in a sandbox, without interfering with the other
#' packages you have installed.
#'
#' @param on turn dev mode on (`TRUE`) or off (`FALSE`).  If omitted will guess
#'   based on whether or not `path` is in [.libPaths()]
#' @param path directory to library.
#' @export
#' @keywords internal
#' @examples
#' \dontrun{
#' dev_mode()
#' dev_mode()
#' }
dev_mode <- local({
  .prompt <- NULL

  function(on = NULL, path = getOption("devtools.path")) {
    lifecycle::deprecate_warn("2.4.6", "dev_mode()")
    lib_paths <- .libPaths()

    path <- path_real(path)
    if (is.null(on)) {
      on <- !(path %in% lib_paths)
    }

    if (on) {
      if (!file_exists(path)) {
        dir_create(path)
      }
      if (!file_exists(path)) {
        cli::cli_abort("Failed to create {.path {path}}")
      }

      if (!is_library(path)) {
        cli::cli_warn(c(
          "{.path {path}} does not appear to be a library.",
          "Are sure you specified the correct directory?"
        ))
      }

      cli::cli_inform(c(v = "Dev mode: ON"))
      options(dev_path = path)

      if (is.null(.prompt)) {
        .prompt <<- getOption("prompt")
      }
      options(prompt = paste("d> "))

      .libPaths(c(path, lib_paths))
    } else {
      cli::cli_inform(c(v = "Dev mode: OFF"))
      options(dev_path = NULL)

      if (!is.null(.prompt)) {
        options(prompt = .prompt)
      }
      .prompt <<- NULL

      .libPaths(setdiff(lib_paths, path))
    }
  }
})

is_library <- function(path) {
  # empty directories can be libraries
  if (length(dir_ls(path)) == 0) {
    return(TRUE)
  }

  # otherwise check that the directories are compiled R directories -
  # i.e. that they contain a Meta directory
  dirs <- dir_ls(path, type = "directory")

  has_pkg_dir <- function(path) length(dir_ls(path, regexp = "Meta")) > 0
  help_dirs <- map_lgl(dirs, has_pkg_dir)

  all(help_dirs)
}

#' Install a local development package
#'
#' @description
#' Uses `R CMD INSTALL` to install the package, after installing needed
#' dependencies with [pak::local_install_deps()].
#'
#' To install to a non-default library, use [withr::with_libpaths()].
#'
#' @template devtools
#' @param reload if `TRUE` (the default), will automatically attempt to reload
#'   the package after installing. Reloading is not always completely possible
#'   so see [pkgload::unregister()] for caveats.
#' @param quick if `TRUE`, skips some optional steps (e.g. help
#'   pre-rendering and multi-arch builds) to make installation as fast
#'   as possible.
#' @param build If `TRUE` (the default), first [pkgbuild::build()]s the
#'   package. This ensures that the installation is completely clean, and
#'   prevents any binary artefacts (like \file{.o}, `.so`) from appearing
#'   in your local package directory, but is considerably slower, because
#'   every compile has to start from scratch.
#'
#'   One downside of installing from a built tarball is that the package is
#'   installed from a temporary location. This means that any source references
#'   will point to dangling locations and debuggers won't have direct access to
#'   the source for step-debugging. For development purposes, `build = FALSE` is
#'   often the better choice.
#'
#'   If `FALSE`, the package is installed directly from its source
#'   directory. This is faster and can be favorable for preserving source
#'   references for debugging (see `keep_source`).
#' @param args An optional character vector of additional command line
#'   arguments to be passed to `R CMD INSTALL`. This defaults to the
#'   value of the option `"devtools.install.args"`.
#' @param build_vignettes if `TRUE`, will build vignettes. Normally it is
#'   `build` that's responsible for creating vignettes; this argument makes
#'   sure vignettes are built even if a build never happens (i.e. because
#'   `build = FALSE`).
#' @param keep_source If `TRUE` will keep the srcrefs from an installed
#'   package. This is useful for debugging (especially inside of RStudio or
#'   Positron). Defaults to `getOption("keep.source.pkgs") || !build`,
#'   since srcrefs are most useful when the package is installed from its
#'   source directory, i.e. when `build = FALSE`.
#' @param quiet If `TRUE`, suppress output.
#' @param force `r lifecycle::badge("deprecated")` No longer used.
#' @family package installation
#' @inheritParams pak::local_install_deps
#' @seealso [with_debug()] to install packages with debugging flags set.
#' @export
install <- function(
  pkg = ".",
  reload = TRUE,
  quick = FALSE,
  build = !quick,
  args = getOption("devtools.install.args"),
  quiet = FALSE,
  dependencies = NA,
  upgrade = FALSE,
  build_vignettes = FALSE,
  keep_source = getOption("keep.source.pkgs") || !build,
  force = deprecated()
) {
  if (!is.logical(upgrade) || length(upgrade) != 1) {
    cli::cli_abort("{.arg upgrade} must be a single TRUE, FALSE, or NA")
  }
  if (lifecycle::is_present(force)) {
    lifecycle::deprecate_warn("2.5.0", "install(force)")
  }

  pkg <- as.package(pkg)

  # Forcing all of the promises for the current namespace now will avoid lazy-load
  # errors when the new package is installed overtop the old one.
  # https://stat.ethz.ch/pipermail/r-devel/2015-December/072150.html
  if (reload && is_loaded(pkg)) {
    eapply(pkgload::ns_env(pkg$package), base::force, all.names = TRUE)
  }

  # Install dependencies
  local({
    if (quiet) {
      withr::local_output_sink(withr::local_tempfile())
      withr::local_message_sink(withr::local_tempfile())
      # pak uses callr internally. When install() itself runs inside a callr
      # worker (e.g. testthat parallel tests), pak's callr_message conditions
      # leak to the parent process. Muffle them here; tryInvokeRestart() is a
      # safe no-op when the muffleMessage restart doesn't exist.
      withCallingHandlers(
        pak::local_install_deps(
          pkg$path,
          upgrade = upgrade,
          dependencies = dependencies || isTRUE(build_vignettes)
        ),
        callr_message = function(cnd) tryInvokeRestart("muffleMessage")
      )
    } else {
      cli::cat_rule("pak::local_install_deps()", col = "cyan")
      pak::local_install_deps(
        pkg$path,
        upgrade = upgrade,
        dependencies = dependencies || isTRUE(build_vignettes)
      )
    }
  })

  if (isTRUE(build_vignettes)) {
    build_opts <- c("--no-resave-data", "--no-manual")
  } else {
    build_opts <- c("--no-resave-data", "--no-manual", "--no-build-vignettes")
  }
  opts <- c(
    if (keep_source) "--with-keep.source",
    "--install-tests"
  )
  if (quick) {
    opts <- c(opts, "--no-docs", "--no-multiarch", "--no-demo")
  }
  opts <- c(opts, args)

  if (build) {
    install_path <- pkgbuild::build(
      pkg$path,
      dest_path = tempdir(),
      args = build_opts,
      quiet = quiet
    )
    on.exit(file_delete(install_path), add = TRUE)
  } else {
    install_path <- pkg$path
  }

  was_loaded <- is_loaded(pkg)
  was_attached <- is_attached(pkg)

  if (reload && was_loaded) {
    pkgload::unregister(pkg$package)
  }

  if (!quiet) {
    cli::cat_rule("R CMD INSTALL", col = "cyan")
  }
  pkgbuild::with_build_tools(
    required = FALSE,
    callr::rcmd(
      "INSTALL",
      c(install_path, opts),
      echo = !quiet,
      show = !quiet,
      spinner = FALSE,
      stderr = "2>&1",
      fail_on_status = TRUE
    )
  )

  if (reload && was_loaded) {
    if (was_attached) {
      require(pkg$package, quietly = TRUE, character.only = TRUE)
    } else {
      requireNamespace(pkg$package, quietly = TRUE)
    }
  }
}

#' Install package dependencies if needed
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' These functions are deprecated. Better alternatives:
#' * `pak::local_install_deps()` instead of `install_deps()`
#' * `pak::local_install_dev_deps()` instead of `install_dev_deps()`
#'
#' @inherit remotes::install_deps
#' @inheritParams install
#' @param ... Additional arguments passed to [remotes::install_deps()].
#' @export
install_deps <- function(
  pkg = ".",
  dependencies = NA,
  repos = getOption("repos"),
  type = getOption("pkgType"),
  upgrade = c("default", "ask", "always", "never"),
  quiet = FALSE,
  build = TRUE,
  build_opts = c("--no-resave-data", "--no-manual", " --no-build-vignettes"),
  ...
) {
  lifecycle::deprecate_warn(
    "2.5.0",
    "install_deps()",
    "pak::local_install_deps()"
  )
  pkg <- as.package(pkg)

  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  check_installed("remotes")
  remotes::install_deps(
    pkg$path,
    dependencies = dependencies,
    repos = repos,
    type = type,
    upgrade = upgrade,
    quiet = quiet,
    build = build,
    build_opts = build_opts,
    ...
  )
}

#' @rdname install_deps
#' @export
install_dev_deps <- function(
  pkg = ".",
  dependencies = TRUE,
  repos = getOption("repos"),
  type = getOption("pkgType"),
  upgrade = c("default", "ask", "always", "never"),
  quiet = FALSE,
  build = TRUE,
  build_opts = c("--no-resave-data", "--no-manual", " --no-build-vignettes"),
  ...
) {
  lifecycle::deprecate_warn(
    "2.5.0",
    "install_dev_deps()",
    "pak::local_install_dev_deps()"
  )
  check_installed("remotes")
  remotes::update_packages("roxygen2")

  pkg <- as.package(pkg)

  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  remotes::install_deps(
    pkg$path,
    dependencies = dependencies,
    repos = repos,
    type = type,
    upgrade = upgrade,
    quiet = quiet,
    build = build,
    build_opts = build_opts,
    ...
  )
}

abort_for_missing_deps <- function(dep_status, pkg, call = parent.frame()) {
  missing <- dep_status[dep_status$status == "missing", ]
  if (nrow(missing) == 0) {
    return(invisible())
  }

  n <- nrow(missing)
  cli::cli_abort(
    c(
      "{n} {.pkg {pkg$package}} {cli::qty(n)}{?dependency is/dependencies are} not installed.",
      "i" = "Install {cli::qty(n)}{?it/them} with {.run pak::local_install_dev_deps()}.",
      dep_labels(missing)
    ),
    call = call
  )
}

local_install <- function(
  pkg = ".",
  quiet = TRUE,
  env = parent.frame(),
  call = parent.frame()
) {
  pkg <- as.package(pkg)

  dep_status <- pkg_dep_status(pkg, dependencies = NA)
  abort_for_missing_deps(dep_status, pkg, call = call)
  report_deps_ahead_behind(
    dep_status,
    pkg_name = pkg$package,
    update_code = "pak::local_install_dev_deps()"
  )

  if (!quiet) {
    cli::cli_inform(c(
      i = "Installing {.pkg {pkg$package}} in temporary library"
    ))
  }
  withr::local_temp_libpaths(.local_envir = env)
  install(
    pkg,
    upgrade = FALSE,
    reload = FALSE,
    quick = TRUE,
    dependencies = FALSE,
    quiet = quiet
  )
}

show_headers <- function(lines) {
  lines <- cli::ansi_strip(lines)
  lines <- lines[grepl("^--", lines)]
  lines
}

test_that("install reports stages", {
  skip_on_cran()
  local_reproducible_output(width = 60)

  pkg <- local_package_copy(test_path("testTest"))
  withr::local_temp_libpaths()

  expect_snapshot(
    install(pkg, reload = FALSE, build = FALSE),
    transform = show_headers
  )
  expect_snapshot(
    install(pkg, reload = FALSE, build = TRUE),
    transform = show_headers
  )
})

test_that("vignettes built on install", {
  skip_on_cran()

  if (!pkgbuild::has_latex()) {
    skip("pdflatex not available")
  }

  pkg <- local_package_copy(test_path("testVignettesBuilt"))

  withr::local_temp_libpaths()
  install(pkg, reload = FALSE, quiet = TRUE, build_vignettes = TRUE)

  vigs <- vignette(package = "testVignettesBuilt")$results
  expect_equal(nrow(vigs), 1)
  expect_equal(vigs[3], "new")
})

context("shim")

# Utility functions -----------------------------
# Take file paths and split them into pieces
expand_path <- function(path) {
  strsplit(path, .Platform$file.sep)
}

# Return the last n elements of vector x
last_n <- function(x, n = 1) {
  len <- length(x)
  x[(len-n+1):len]
}


# Tests -----------------------------------------

test_that("system.file returns correct values when used with load_all", {
  load_all("testShim")
  shim_ns <- ns_env("testShim")

  # The devtools::system.file function should return modified values.
  files <- shim_system.file(c("A.txt", "B.txt", "C.txt", "D.txt"),
                            package = "testShim")
  files <- expand_path(files)

  expect_true(all(last_n(files[[1]], 3) == c("testShim", "inst", "A.txt")))
  expect_true(all(last_n(files[[2]], 3) == c("testShim", "inst", "B.txt")))
  # Note that C.txt wouldn't be returned by base::system.file (see comments
  # in shim_system.file for explanation)
  expect_true(all(last_n(files[[3]], 2) == c("testShim", "C.txt")))
  # D.txt should be dropped
  expect_equal(length(files), 3)

  # If all files are not present, return ""
  files <- shim_system.file("nonexistent", package = "testShim")
  expect_equal(files, "")

  # Test packages loaded the usual way - should just pass through to
  # base::system.file
  expect_identical(base::system.file("Meta", "Rd.rds", package = "stats"),
    shim_system.file("Meta", "Rd.rds", package = "stats"))
  expect_identical(base::system.file("INDEX", package = "stats"),
    shim_system.file("INDEX", package = "stats"))
  expect_identical(base::system.file("nonexistent", package = "stats"),
    shim_system.file("nonexistent", package = "stats"))

  unload("testShim")
})

test_that("shimmed system.file respects mustWork", {
  load_all("testShim")
  find_missing <- function(mustWork) {
    shim_system.file("missing.txt", package = "testShim", mustWork = mustWork)
  }

  expect_equal(find_missing(FALSE), "")
  expect_error(find_missing(TRUE), "No file found")
})


test_that("Shimmed system.file returns correct values when used with load_all", {
  load_all("testShim")
  shim_ns <- ns_env("testShim")

  # Make sure the version of system.file inserted into the namespace's imports
  # is the same as devtools::system.file
  expect_identical(get("system.file", envir = shim_ns), shim_system.file)

  # Another check
  expect_identical(get_system.file(), shim_system.file)

  unload("testShim")
})

test_that("Replacement system.file returns correct values when installed", {
  # This set of tests is mostly a sanity check - it doesn't use the special
  # version of system.file, but it's useful to make sure we know what to look
  # for in the other tests.

  # Make a temp lib directory to install test package into
  old_libpaths <- .libPaths()
  tmp_libpath = file.path(tempdir(), "devtools_test")
  if (!dir.exists(tmp_libpath)) dir.create(tmp_libpath)
  .libPaths(c(tmp_libpath, .libPaths()))

  install("testShim", quiet = TRUE)
  expect_true(require(testShim))

  # The special version of system.file shouldn't exist - this get() will fall
  # through to the base namespace
  expect_identical(get("system.file", pos = asNamespace("testShim")),
    base::system.file)

  # Test within package testShim
  files <- get_system.file()(c("A.txt", "B.txt", "C.txt", "D.txt"),
    package = "testShim")
  files <- expand_path(files)
  expect_true(all(last_n(files[[1]], 2) == c("testShim", "A.txt")))
  expect_true(all(last_n(files[[2]], 2) == c("testShim", "B.txt")))
  expect_equal(length(files), 2)  # Third and fourth should be dropped

  # If all files are not present, return ""
  files <- get_system.file()("nonexistent", package = "testShim")
  expect_equal(files, "")

  detach("package:testShim", unload = TRUE)

  # Reset the libpath
  .libPaths(old_libpaths)
})

context("Compiled DLLs")

test_that("unload() unloads DLLs from packages loaded with library()", {

  # Make a temp lib directory to install test package into
  old_libpaths <- .libPaths()
  tmp_libpath = file.path(tempdir(), "devtools_test")
  if (!dir.exists(tmp_libpath)) dir.create(tmp_libpath)
  .libPaths(c(tmp_libpath, .libPaths()))

  # Reset the libpath on exit
  on.exit(.libPaths(old_libpaths), add = TRUE)

  # Install package
  install("testDllLoad", quiet = TRUE, args = "--no-multiarch")
  expect_true(require(testDllLoad))

  # Check that it's loaded properly, by running a function from the package.
  # nulltest() calls a C function which returns null.
  expect_true(is.null(nulltest()))

  # DLL should be listed in .dynLibs()
  dynlibs <- vapply(.dynLibs(), `[[`, "name", FUN.VALUE = character(1))
  expect_true(any(grepl("testDllLoad", dynlibs)))

  unload("testDllLoad")

  # DLL should not be listed in .dynLibs()
  dynlibs <- vapply(.dynLibs(), `[[`, "name", FUN.VALUE = character(1))
  expect_false(any(grepl("testDllLoad", dynlibs)))



  # Clean out compiled objects
  clean_dll("testDllLoad")
})


test_that("load_all() compiles and loads DLLs", {

  clean_dll("testDllLoad")

  load_all("testDllLoad", reset = TRUE, quiet = TRUE)

  # Check that it's loaded properly, by running a function from the package.
  # nulltest() calls a C function which returns null.
  expect_true(is.null(nulltest()))

  # DLL should be listed in .dynLibs()
  dynlibs <- vapply(.dynLibs(), `[[`, "name", FUN.VALUE = character(1))
  expect_true(any(grepl("testDllLoad", dynlibs)))

  unload("testDllLoad")

  # DLL should not be listed in .dynLibs()
  dynlibs <- vapply(.dynLibs(), `[[`, "name", FUN.VALUE = character(1))
  expect_false(any(grepl("testDllLoad", dynlibs)))


  # Loading again, and reloading
  # Should not re-compile (don't have a proper test for this)
  load_all("testDllLoad", quiet = TRUE)
  expect_true(is.null(nulltest()))

  # load_all when already loaded
  # Should not re-compile (don't have a proper test for this)
  load_all("testDllLoad", quiet = TRUE)
  expect_true(is.null(nulltest()))

  # Should re-compile (don't have a proper test for this)
  load_all("testDllLoad", recompile = TRUE, quiet = TRUE)
  expect_true(is.null(nulltest()))
  unload("testDllLoad")

  # Clean out compiled objects
  clean_dll("testDllLoad")
})


test_that("Specific functions from DLLs listed in NAMESPACE can be called", {
  load_all("testDllLoad", quiet = TRUE)

  # nulltest() uses the calling convention:
  # .Call("null_test", PACKAGE = "testDllLoad")
  expect_true(is.null(nulltest()))

  # nulltest2() uses a specific C function listed in NAMESPACE, null_test2
  # null_test2 is an object in the packg_env
  # It uses this calling convention:
  # .Call(null_test2)
  expect_true(is.null(nulltest2()))
  nt2 <- ns_env("testDllLoad")$null_test2
  expect_equal(class(nt2), "NativeSymbolInfo")
  expect_equal(nt2$name, "null_test2")

  unload("testDllLoad")

  # Clean out compiled objects
  clean_dll("testDllLoad")
})


test_that("load_all() can compile and load DLLs linked to Rcpp", {

  skip_on_os("solaris")

  clean_dll("testDllRcpp")

  load_all("testDllRcpp", reset = TRUE, quiet = TRUE)

  # Check that it's loaded properly by calling the hello world function
  # which returns a list
  expect_true(is.list(rcpp_hello_world()))

  # Check whether attribute compilation occurred and that exported
  # names are available from load_all
  expect_true(rcpp_test_attributes())

  # Unload and clean out compiled objects
  unload("testDllRcpp")
  clean_dll("testDllRcpp")
})

#' Was devtools installed with tests?
#'
#' @keywords internal
#' @export
has_tests <- function() {
  test_path <- tryCatch(
    path_package("devtools", "tests"),
    error = function(e) NULL
  )

  !is.null(test_path)
}

#' Deprecated package installation functions
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' These functions have been deprecated in favor of [pak](https://pak.r-lib.org/),
#' as that is what we now recommend for package installation. There are a few
#' functions which have no pak equivalent, where you can instead call the
#' old remotes functions directly.
#'
#' ## Migration guide
#'
#' | devtools function | Replacement |
#' |-------------------|-------------|
#' | `install_bioc("pkg")` | `pak::pak("bioc::pkg")` |
#' | `install_bitbucket("user/repo")` | `remotes::install_bitbucket("user/repo")` |
#' | `install_cran("pkg")` | `pak::pak("pkg")` |
#' | `install_dev("pkg")` | `remotes::install_dev("pkg")` |
#' | `install_git("url")` | `pak::pak("git::url")` |
#' | `install_github("user/repo")` | `pak::pak("user/repo")` |
#' | `install_gitlab("user/repo")` | `pak::pak("gitlab::user/repo")` |
#' | `install_local("path")` | `pak::pak("local::path")` |
#' | `install_svn("url")` | `remotes::install_svn("url")` |
#' | `install_url("url")` | `pak::pak("url::url")` |
#' | `install_version("pkg", "1.0.0")` | `pak::pak("pkg@1.0.0")` |
#' | `update_packages("pkg")` | `pak::pak("pkg")` |
#' | `dev_package_deps()` | `pak::local_dev_deps()` |
#' | `github_pull("123")` | `remotes::github_pull("123")` |
#' | `github_release()` | `remotes::github_release()` |
#'
#' @name install-deprecated
#' @keywords internal
NULL

#' @rdname install-deprecated
#' @export
install_bioc <- function(...) {
  lifecycle::deprecate_warn(
    "2.5.0",
    "install_bioc()",
    I('pak::pak("bioc::pkg")')
  )
  check_installed("remotes")
  pkgbuild::with_build_tools(remotes::install_bioc(...), required = FALSE)
}

#' @rdname install-deprecated
#' @export
install_bitbucket <- function(...) {
  lifecycle::deprecate_warn(
    "2.5.0",
    "install_bitbucket()",
    "remotes::install_bitbucket()"
  )
  check_installed("remotes")
  pkgbuild::with_build_tools(remotes::install_bitbucket(...), required = FALSE)
}

#' @rdname install-deprecated
#' @export
install_cran <- function(...) {
  lifecycle::deprecate_warn("2.5.0", "install_cran()", "pak::pak()")
  check_installed("remotes")
  pkgbuild::with_build_tools(remotes::install_cran(...), required = FALSE)
}

#' @rdname install-deprecated
#' @export
install_dev <- function(...) {
  lifecycle::deprecate_warn("2.5.0", "install_dev()", "remotes::install_dev()")
  check_installed("remotes")
  pkgbuild::with_build_tools(remotes::install_dev(...), required = FALSE)
}

#' @rdname install-deprecated
#' @export
install_git <- function(...) {
  lifecycle::deprecate_warn("2.5.0", "install_git()", I('pak::pak("git::url")'))
  check_installed("remotes")
  pkgbuild::with_build_tools(remotes::install_git(...), required = FALSE)
}

#' @rdname install-deprecated
#' @export
install_github <- function(...) {
  lifecycle::deprecate_warn(
    "2.5.0",
    "install_github()",
    I('pak::pak("user/repo")')
  )
  check_installed("remotes")
  pkgbuild::with_build_tools(remotes::install_github(...), required = FALSE)
}

#' @rdname install-deprecated
#' @export
install_gitlab <- function(...) {
  lifecycle::deprecate_warn(
    "2.5.0",
    "install_gitlab()",
    I('pak::pak("gitlab::user/repo")')
  )
  check_installed("remotes")
  pkgbuild::with_build_tools(remotes::install_gitlab(...), required = FALSE)
}

#' @rdname install-deprecated
#' @export
install_local <- function(...) {
  lifecycle::deprecate_warn(
    "2.5.0",
    "install_local()",
    I('pak::pak("local::path")')
  )
  check_installed("remotes")
  pkgbuild::with_build_tools(remotes::install_local(...), required = FALSE)
}

#' @rdname install-deprecated
#' @export
install_svn <- function(...) {
  lifecycle::deprecate_warn("2.5.0", "install_svn()", "remotes::install_svn()")
  check_installed("remotes")
  pkgbuild::with_build_tools(remotes::install_svn(...), required = FALSE)
}

#' @rdname install-deprecated
#' @export
install_url <- function(...) {
  lifecycle::deprecate_warn("2.5.0", "install_url()", I('pak::pak("url::url")'))
  check_installed("remotes")
  pkgbuild::with_build_tools(remotes::install_url(...), required = FALSE)
}

#' @rdname install-deprecated
#' @export
install_version <- function(...) {
  lifecycle::deprecate_warn(
    "2.5.0",
    "install_version()",
    I('pak::pak("pkg@version")')
  )
  check_installed("remotes")
  pkgbuild::with_build_tools(remotes::install_version(...), required = FALSE)
}

#' @rdname install-deprecated
#' @export
update_packages <- function(...) {
  lifecycle::deprecate_warn("2.5.0", "update_packages()", "pak::pak()")
  check_installed("remotes")
  pkgbuild::with_build_tools(remotes::update_packages(...), required = FALSE)
}

#' @rdname install-deprecated
#' @export
dev_package_deps <- function(...) {
  lifecycle::deprecate_warn(
    "2.5.0",
    "dev_package_deps()",
    "pak::local_dev_deps()"
  )
  check_installed("remotes")
  pkgbuild::with_build_tools(remotes::dev_package_deps(...), required = FALSE)
}

#' @rdname install-deprecated
#' @export
github_pull <- function(...) {
  lifecycle::deprecate_warn("2.5.0", "github_pull()", "remotes::github_pull()")
  check_installed("remotes")
  remotes::github_pull(...)
}

#' @rdname install-deprecated
#' @export
github_release <- function(...) {
  lifecycle::deprecate_warn(
    "2.5.0",
    "github_release()",
    "remotes::github_release()"
  )
  check_installed("remotes")
  remotes::github_release(...)
}

context("Dependencies")

test_that("Warned about dependency versions", {
  # Should give a warning about grid version
  expect_warning(load_all("testImportVersion"), "Need grid >=")
  unload("testImportVersion")

  # TODO: Add check for NOT giving a warning about compiler version
  # Not possible with testthat?
})


test_that("Error on missing dependencies", {
  # Should give a warning about missing package
  expect_error(load_all("testImportMissing"), "missingpackage not available")

  # Loading process will be partially done; unload it
  unload("testImportMissing")
})

test_that("Packages in depends are required", {
  load_all("testDependMissing")
  expect_true("package:MASS" %in% search())
  unload("testDependMissing")
  detach("package:MASS", unload = TRUE)
})

test_that("Parse dependencies", {
  deps <- parse_deps("\nhttr (< 2.1),\nRCurl (>= 3),\nutils (== 2.12.1),\ntools,\nR (>= 2.10),\nmemoise")
  expect_equal(nrow(deps), 5)
  expect_false("R" %in% deps$name)

  expect_equal(deps$compare, c("<", ">=", "==", NA, NA))
  expect_equal(deps$version, c("2.1", "3", "2.12.1", NA, NA))


  # Invalid version specifications
  expect_error(parse_deps("\nhttr (< 2.1),\nRCurl (3.0)"))
  expect_error(parse_deps("\nhttr (< 2.1),\nRCurl ( 3.0)"))
  expect_error(parse_deps("\nhttr (< 2.1),\nRCurl (==3.0)"))
  expect_error(parse_deps("\nhttr (< 2.1),\nRCurl (==3.0 )"))
  expect_error(parse_deps("\nhttr (< 2.1),\nRCurl ( ==3.0)"))

  # This should be OK (no error)
  deps <- parse_deps("\nhttr (< 2.1),\nRCurl (==  3.0.1)")
  expect_equal(deps$compare, c("<", "=="))
  expect_equal(deps$version, c("2.1", "3.0.1"))
})

test_that("Dependencies of development package include direct dependencies", {
  deps <- dev_package_deps("testNamespace")
  expect_equal(deps$package, "bitops")
})

#' Build package vignettes
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `build_vignettes()` is deprecated because we no longer recommend that you
#' build vignettes in this way, because it leaves build artifacts in your
#' development directory. Instead, use [pkgdown::build_article()] to
#' render articles locally for preview and polishing.
#'
#' @template devtools
#' @param quiet If `TRUE`, suppresses most output. Set to `FALSE`
#'   if you need to debug.
#' @param install If `TRUE`, install the package before building
#'   vignettes.
#' @param keep_md If `TRUE`, move md intermediates as well as rendered
#'   outputs. Most useful when using the `keep_md` YAML option for Rmarkdown
#'   outputs.
#' @inheritParams tools::buildVignettes
#' @inheritParams remotes::install_deps
#' @importFrom stats update
#' @keywords internal
#' @export
build_vignettes <- function(
  pkg = ".",
  dependencies = "VignetteBuilder",
  clean = TRUE,
  upgrade = "never",
  quiet = FALSE,
  install = TRUE,
  keep_md = TRUE
) {
  lifecycle::deprecate_warn("2.5.0", "build_vignettes()")
  pkg <- as.package(pkg)
  save_all()

  vigns <- tools::pkgVignettes(dir = pkg$path)
  if (length(vigns$docs) == 0) {
    return(invisible())
  }

  check_installed("remotes")
  deps <- remotes::dev_package_deps(pkg$path, dependencies)
  update(deps, upgrade = upgrade)

  if (isTRUE(install)) {
    local_install(pkg, quiet = TRUE)
  }

  cli::cli_inform(c(i = "Building vignettes for {.pkg {pkg$package}}"))
  callr::r(
    function(...) tools::buildVignettes(...),
    args = list(
      dir = pkg$path,
      clean = clean,
      tangle = TRUE,
      quiet = quiet
    ),
    show = !quiet,
    spinner = FALSE
  )

  # We need to re-run pkgVignettes now that they are built to get the output
  # files as well
  cli::cli_inform(c(i = "Copying vignettes"))
  vigns <- tools::pkgVignettes(dir = pkg$path, source = TRUE, output = TRUE)
  copy_vignettes(pkg, keep_md)
  create_vignette_index(pkg, vigns)

  invisible(TRUE)
}

create_vignette_index <- function(pkg, vigns) {
  cli::cli_inform(c(i = "Building vignette index"))

  usethis_use_directory(pkg, "Meta", ignore = TRUE)
  usethis_use_git_ignore(pkg, "/Meta/")

  vignette_index <- ("tools" %:::% ".build_vignette_index")(vigns)
  vignette_index_path <- path(pkg$path, "Meta", "vignette.rds")

  saveRDS(vignette_index, vignette_index_path, version = 2L)
}

#' Clean built vignettes
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `clean_vignettes()` is deprecated because [build_vignettes()] is deprecated.
#'
#' @template devtools
#' @keywords internal
#' @export
clean_vignettes <- function(pkg = ".") {
  lifecycle::deprecate_warn("2.5.0", "clean_vignettes()")
  pkg <- as.package(pkg)
  vigns <- tools::pkgVignettes(dir = pkg$path)
  if (length(vigns$docs) == 0 || path_file(vigns$dir) != "vignettes") {
    return(invisible())
  }

  cli::cli_inform(c(
    i = "Cleaning built vignettes and index from {.pkg {pkg$package}}"
  ))

  doc_path <- path(pkg$path, "doc")

  vig_candidates <- if (dir_exists(doc_path)) dir_ls(doc_path) else character()
  vig_rm <- vig_candidates[file_name(vig_candidates) %in% file_name(vigns$docs)]

  extra_candidates <- path(doc_path, path_file(find_vignette_extras(pkg)))
  extra_rm <- extra_candidates[file_exists(extra_candidates)]

  meta_path <- path(pkg$path, "Meta")
  vig_index_path <- path(meta_path, "vignette.rds")
  vig_index_rm <- if (file_exists(vig_index_path)) vig_index_path

  to_remove <- c(vig_rm, extra_rm, vig_index_rm)
  if (length(to_remove) > 0) {
    cli::cli_inform(c(x = "Removing {.file {path_file(to_remove)}}"))
    file_delete(to_remove)
  }

  walk(c(doc_path, meta_path), dir_delete_if_empty)

  invisible(TRUE)
}

dir_delete_if_empty <- function(x) {
  if (dir_exists(x) && is_empty(dir_ls(x))) {
    dir_delete(x)
    cli::cli_inform(c(x = "Removing {.file {path_file(x)}}"))
  }
}

file_name <- function(x) {
  if (length(x) == 0) {
    return(NULL)
  }
  path_ext_remove(path_file(x))
}

test_that("Sweave vignettes copied into doc", {
  withr::local_options(lifecycle_verbosity = "quiet")
  if (!pkgbuild::has_latex()) {
    skip("pdflatex not available")
  }

  pkg <- local_package_copy(test_path("testVignettes"))

  suppressMessages(build_vignettes(pkg, quiet = TRUE))
  expect_setequal(
    path_file(dir_ls(path(pkg, "doc"))),
    c("new.pdf", "new.R", "new.Rnw")
  )
})

test_that("Built files are updated", {
  withr::local_options(lifecycle_verbosity = "quiet")
  # This test is time dependant and sometimes fails on CRAN because the systems are under heavy load.
  skip_on_cran()
  pkg <- local_package_copy(test_path("testMarkdownVignettes"))

  suppressMessages(build_vignettes(pkg, quiet = TRUE))
  output <- dir_ls(path(pkg, "doc"), regexp = "new")
  first <- file_info(output)$modification_time

  Sys.sleep(.01)
  suppressMessages(build_vignettes(pkg, quiet = TRUE))
  second <- file_info(output)$modification_time

  expect_true(all(second > first))
})

test_that("Rmarkdown vignettes copied into doc", {
  withr::local_options(lifecycle_verbosity = "quiet")
  pkg <- local_package_copy(test_path("testMarkdownVignettes"))
  doc <- path(pkg, "doc")

  suppressMessages(build_vignettes(pkg, quiet = TRUE))
  expect_setequal(path_file(dir_ls(doc)), c("test.html", "test.R", "test.Rmd"))
})

test_that("extra files copied and removed", {
  withr::local_options(lifecycle_verbosity = "quiet")
  pkg <- local_package_copy(test_path("testMarkdownVignettes"))
  writeLines("a <- 1", path(pkg, "vignettes", "a.R"))

  extras_path <- path(pkg, "vignettes", ".install_extras")
  writeLines("a.R", extras_path)

  suppressMessages(build_vignettes(pkg, quiet = TRUE))
  expect_true(file_exists(path(pkg, "doc", "a.R")))

  suppressMessages(clean_vignettes(pkg))
  expect_false(file_exists(path(pkg, "doc", "a.R")))
})

test_that(".gitignore updated when building vignettes", {
  withr::local_options(lifecycle_verbosity = "quiet")
  pkg <- local_package_copy(test_path("testMarkdownVignettes"))
  gitignore <- path(pkg, ".gitignore")

  suppressMessages(build_vignettes(pkg, quiet = TRUE))
  expect_true(all(c("/Meta/", "/doc/") %in% readLines(gitignore)))
})

test_that("build_vignettes() and clean_vignettes() are deprecated", {
  pkg <- local_package_create()
  expect_snapshot({
    build_vignettes(pkg)
    clean_vignettes(pkg)
  })
})

context("Vignettes")

test_that("Sweave vignettes copied into inst/doc", {
  if (!has_latex()) {
    skip("pdflatex not available")
  }

  clean_vignettes("testVignettes")
  expect_false("new.pdf" %in% dir("testVignettes/inst/doc"))
  expect_false("new.R" %in% dir("testVignettes/inst/doc"))
  expect_false("new.Rnw" %in% dir("testVignettes/inst/doc"))

  build_vignettes("testVignettes")
  expect_true("new.pdf" %in% dir("testVignettes/inst/doc"))
  expect_true("new.R" %in% dir("testVignettes/inst/doc"))
  expect_true("new.Rnw" %in% dir("testVignettes/inst/doc"))

  clean_vignettes("testVignettes")
  expect_false("new.pdf" %in% dir("testVignettes/inst/doc"))
  expect_false("new.R" %in% dir("testVignettes/inst/doc"))
  expect_false("new.Rnw" %in% dir("testVignettes/inst/doc"))
})

test_that("Built files are updated", {
  if (!has_latex()) {
    skip("pdflatex not available")
  }

  clean_vignettes("testVignettes")
  build_vignettes("testVignettes")
  on.exit(clean_vignettes("testVignettes"))

  output <- dir("testVignettes/inst/doc", "new", full.names = TRUE)
  first <- file.info(output)$mtime

  Sys.sleep(1)
  build_vignettes("testVignettes")
  second <- file.info(output)$mtime

  expect_true(all(second > first))
})

if (packageVersion("knitr") >= 1.2) {
  test_that("Rmarkdown vignettes copied into inst/doc", {
    pkg <- as.package("testMarkdownVignettes")
    doc_path <- file.path(pkg$path, "inst", "doc")

    clean_vignettes(pkg)
    expect_false("test.html" %in% dir(doc_path))
    expect_false("test.R" %in% dir(doc_path))
    expect_false("test.Rmd" %in% dir(doc_path))

    build_vignettes(pkg)
    expect_true("test.html" %in% dir(doc_path))
    expect_true("test.R" %in% dir(doc_path))
    expect_true("test.Rmd" %in% dir(doc_path))

    clean_vignettes(pkg)
    expect_false("test.html" %in% dir(doc_path))
    expect_false("test.R" %in% dir(doc_path))
    expect_false("test.Rmd" %in% dir(doc_path))
  })

  test_that("dependencies argument", {
    pkg <- as.package("testMarkdownVignettes")
    doc_path <- file.path(pkg$path, "inst", "doc")

    clean_vignettes(pkg)
    on.exit(clean_vignettes(pkg), add = TRUE)
    installed_deps <- NULL
    with_mock(
      install_deps = function(pkg, dependencies, ...) installed_deps <<- dependencies,
      build_vignettes(pkg, FALSE)
    )
    expect_false(installed_deps)
  })
}


test_that("Extra files copied and removed", {
  if (!has_latex()) {
    skip("pdflatex not available")
  }

  pkg <- as.package("testVignetteExtras")
  doc_path <- file.path(pkg$path, "inst", "doc")

  extras_path <- file.path("testVignetteExtras", "vignettes",
    ".install_extras")
  writeLines("a.r", extras_path)
  on.exit(unlink(extras_path))

  clean_vignettes(pkg)
  expect_false("a.r" %in% dir(doc_path))

  build_vignettes(pkg)
  expect_true("a.r" %in% dir(doc_path))

  clean_vignettes(pkg)
  expect_false("a.r" %in% dir(doc_path))
})


test_that("vignettes built on install", {
  if (!has_latex()) {
    skip("pdflatex not available")
  }

  # Make sure it fails if we build without installing
  expect_error(build_vignettes("testVignettesBuilt"),
    "there is no package called")

  install("testVignettesBuilt", reload = FALSE, quiet = TRUE,
    build_vignettes = TRUE)
  unlink("testVignettesBuilt/vignettes/new.tex")
  unlink("testVignettesBuilt/vignettes/.build.timestamp")

  vigs <- vignette(package = "testVignettesBuilt")$results
  expect_equal(nrow(vigs), 1)
  expect_equal(vigs[3], "new")

  suppressMessages(remove.packages("testVignettesBuilt"))
})

find_active_file <- function(arg = "file", call = parent.frame()) {
  if (!is_rstudio_running()) {
    cli::cli_abort("{.arg {arg}} is absent but must be supplied.", call = call)
  }
  normalizePath(rstudioapi::getSourceEditorContext()$path)
}

find_test_file <- function(path, call = parent.frame()) {
  type <- test_file_type(path)
  if (anyNA(type)) {
    file <- path_file(path[is.na(type)])
    cli::cli_abort(
      "Don't know how to find tests associated with the active file {.file {file}}",
      call = call
    )
  }

  pkg <- as.package(dirname(path))

  is_test <- type == "test"
  path[!is_test] <- paste0(
    pkg$path,
    "/tests/testthat/test-",
    name_source(path[!is_test]),
    ".R"
  )
  path <- unique(path[file_exists(path)])

  if (length(path) == 0) {
    cli::cli_abort("No test files found", call = call)
  }
  path
}

test_file_type <- function(path) {
  dir <- path_file(path_dir(path))
  # this accounts for snapshot files in a variant subfolder
  parent_dir <- path_file(path_dir(path_dir(path)))
  name <- path_file(path)
  ext <- tolower(path_ext(path))

  src_ext <- c("c", "cc", "cpp", "cxx", "h", "hpp", "hxx")

  type <- rep(NA_character_, length(path))
  type[dir == "R" & ext == "r"] <- "R"
  type[dir == "testthat" & ext == "r" & grepl("^test", name)] <- "test"
  type[dir == "src" & ext %in% src_ext] <- "src"
  type[dir == "_snaps" & ext == "md"] <- "snap"
  type[parent_dir == "_snaps" & ext == "md"] <- "snap"
  type
}

# Figure out "name" of a test or source file
name_test <- function(path) {
  gsub("^test[-_]", "", name_source(path))
}
name_source <- function(path) {
  path_ext_remove(path_file(path))
}

test_that("find_active_file() gives useful error if no RStudio", {
  expect_snapshot(find_active_file(), error = TRUE)
})

test_that("fails if can't find tests", {
  dir <- local_package_create()
  withr::local_dir(dir)

  expect_snapshot(error = TRUE, {
    find_test_file("R/foo.blah")
    find_test_file("R/foo.R")
  })
})

test_that("find_test_file() works with snapshot files", {
  dir <- local_package_create()
  withr::local_dir(dir)

  dir_create("tests/testthat/_snaps")
  file_create("tests/testthat/test-foo.R")
  file_create("tests/testthat/_snaps/foo.md")

  path <- find_test_file("tests/testthat/_snaps/foo.md")
  expect_equal(path_file(path), "test-foo.R")
})

test_that("find_test_file() works with snapshot variant files", {
  dir <- local_package_create()
  withr::local_dir(dir)

  dir_create("tests/testthat/_snaps/variant")
  file_create("tests/testthat/test-foo.R")
  file_create("tests/testthat/_snaps/variant/foo.md")

  path <- find_test_file("tests/testthat/_snaps/variant/foo.md")
  expect_equal(path_file(path), "test-foo.R")
})

test_that("can determine file type", {
  expect_equal(test_file_type("R/foo.R"), "R")
  expect_equal(test_file_type("R/foo.c"), NA_character_)

  expect_equal(test_file_type("src/foo.c"), "src")
  expect_equal(test_file_type("src/foo.R"), NA_character_)

  expect_equal(test_file_type("tests/testthat/test-foo.R"), "test")
  expect_equal(test_file_type("tests/testthat/test-foo.c"), NA_character_)
  expect_equal(test_file_type("tests/testthat/foo.R"), NA_character_)

  expect_equal(test_file_type("tests/testthat/_snaps/foo.md"), "snap")
  expect_equal(test_file_type("tests/testthat/_snaps/variant/foo.md"), "snap")
  expect_equal(test_file_type("tests/testthat/_snaps/foo.R"), NA_character_)

  expect_equal(test_file_type("DESCRIPTION"), NA_character_)
})

#' Test function for help
#'
#' The purpose of this function is to test out \code{help} and \code{?} from
#' devtools.
#'
#' @examples
#' stopifnot(foofoo() == 'You called foofoo.')
#' @export
foofoo <- function() "You called foofoo."

#' Unload and reload package
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `reload()` is deprecated because we no longer use or recommend this
#' workflow. Instead, we recommend [load_all()] to load a package for
#' interactive development.
#'
#' @template devtools
#' @param quiet if `TRUE` suppresses output from this function.
#' @seealso [load_all()] to load a package for interactive development.
#' @export
#' @keywords internal
reload <- function(pkg = ".", quiet = FALSE) {
  lifecycle::deprecate_warn("2.5.0", "reload()", "load_all()")
  pkg <- as.package(pkg)

  if (is_attached(pkg)) {
    if (!quiet) {
      cli::cli_inform(c(i = "Reloading attached {.pkg {pkg$package}}"))
    }
    pkgload::unload(pkg$package)
    require(pkg$package, character.only = TRUE, quietly = TRUE)
  } else if (is_loaded(pkg)) {
    if (!quiet) {
      cli::cli_inform(c(i = "Reloading loaded {.pkg {pkg$package}}"))
    }
    pkgload::unload(pkg$package)
    requireNamespace(pkg$package, quietly = TRUE)
  }
}

#' @inherit pkgload::load_all
#' @param ... Additional arguments passed to [pkgload::load_all()].
#' @export
load_all <- function(
  path = ".",
  reset = TRUE,
  recompile = FALSE,
  export_all = TRUE,
  helpers = TRUE,
  quiet = FALSE,
  ...
) {
  if (inherits(path, "package")) {
    path <- path$path
  }

  if (is_loading()) {
    cli::cli_abort(c(
      "Recursive loading detected.",
      i = "Did you accidentally include {.fn load_all} in an R source file?"
    ))
  }

  save_all()

  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  pkgload::load_all(
    path = path,
    reset = reset,
    recompile = recompile,
    export_all = export_all,
    helpers = helpers,
    quiet = quiet,
    ...
  )
}

#' @importFrom pkgload is_loading
#' @export
pkgload::is_loading

#' @importFrom pkgload unload
#' @export
pkgload::unload

context("s4-unload")

# Returns a named vector of this class's superclasses.
# Results are sorted so they can be compared easily to a vector.
# A contains B  ==  A is a superclass of B
get_superclasses <- function(class) {
  superclasses <- vapply(getClass(class)@contains, slot, "superClass",
    FUN.VALUE = character(1))

  sort(unname(superclasses))
}

# Returns a named vector of this class's subclasses
# Results are sorted so they can be compared easily to a vector.
# A extends B  ==  A is a subclass of B
get_subclasses <- function(class) {
  subclasses <- vapply(getClass(class)@subclasses, slot, "subClass",
    FUN.VALUE = character(1))

  sort(unname(subclasses))
}


test_that("loading and reloading s4 classes", {
  load_all("testS4union")

  # Check class hierarchy
  expect_equal(get_superclasses("A"), c("AB", "mle2A", "mleA"))
  expect_equal(get_subclasses("AB"), c("A", "B"))
  expect_equal(get_superclasses("mle2"), c("mle", "mle2A", "mleA"))
  expect_equal(get_subclasses("mleA"), c("A", "mle", "mle2"))
  expect_equal(get_subclasses("mle2A"), c("A", "mle2"))

  # Check that package is registered correctly
  expect_equal(getClassDef("A")@package, "testS4union")
  expect_equal(getClassDef("AB")@package, "testS4union")
  expect_equal(getClassDef("mle2")@package, "testS4union")

  # Unloading shouldn't result in any errors or warnings
  if (packageVersion("testthat") >= "0.7.1.99") {
    expect_that(unload("testS4union"), not(gives_warning()))
  } else {
    unload("testS4union")
  }

  # Check that classes are unregistered
  expect_true(is.null(getClassDef("A")))
  expect_true(is.null(getClassDef("B")))
  expect_true(is.null(getClassDef("AB")))


  # Load again and repeat tests --------------------------------------------

  # Loading again shouldn't result in any errors or warnings
  if (packageVersion("testthat") >= "0.7.1.99") {
    expect_that(load_all("testS4union", reset = FALSE), not(gives_warning()))
  } else {
    load_all("testS4union", reset = FALSE)
  }

  # Check class hierarchy
  expect_equal(get_superclasses("A"), c("AB", "mle2A", "mleA"))
  expect_equal(get_subclasses("AB"), c("A", "B"))
  expect_equal(get_superclasses("mle2"), c("mle", "mle2A", "mleA"))
  expect_equal(get_subclasses("mleA"), c("A", "mle", "mle2"))
  expect_equal(get_subclasses("mle2A"), c("A", "mle2"))

  # Check that package is registered correctly
  expect_equal(getClassDef("A")@package, "testS4union")
  expect_equal(getClassDef("AB")@package, "testS4union")
  expect_equal(getClassDef("mle2")@package, "testS4union")

  unload("testS4union")
  unloadNamespace("stats4")   # This was imported by testS4union

  # Check that classes are unregistered
  # This test on A fails for some bizarre reason - bug in R? But it doesn't
  # to cause any practical problems.
  # expect_true(is.null(getClassDef("A")))
  expect_true(is.null(getClassDef("B")))
  expect_true(is.null(getClassDef("AB")))
})

context("Load: collate")

test_that("If collate absent, load in alphabetical order", {
  load_all("testCollateAbsent")

  expect_equal(a, 3)

  unload("testCollateAbsent")
})

test_that("Warned about files missing from collate, but they're still loaded", {
  expect_message(load_all("testCollateMissing"), "a.r")

  expect_equal(a, 1)
  expect_equal(b, 2)

  unload("testCollateMissing")
})

test_that("Extra files in collate don't error, but warn", {
  expect_message(load_all("testCollateExtra"), "b.r")

  expect_equal(a, 1)

  unload("testCollateExtra")
})

temp_copy_pkg <- function(pkg) {
  file.copy(normalizePath(pkg), tempdir(), recursive = TRUE)
  normalizePath(file.path(tempdir(), pkg))
}

test_that("DESCRIPTION Collate field, with latest @includes, is recognised by load_all", {
  # Make a temporary copy of the package for this test,
  # since update_collate (in load_all) may have permanent side effects,
  # namely changing the collate field in the DESCRIPTION file
  test_pkg <- temp_copy_pkg('testCollateOrder')
  on.exit(unlink(test_pkg, recursive = TRUE))

  expect_output(
    expect_message(load_all(test_pkg), "Loading testCollateOrder"),
    "Updating collate directive"
  )

  expect_equal(a, 1) #even though b.r set it to 2

  unload(test_pkg)
})

context("Imports")

test_that("Imported objects are copied to package environment", {
  load_all("testNamespace")
  # This package imports the whole 'compiler' package, bitops::bitAnd, and
  # bitops::bitOr.
  imp_env <- imports_env("testNamespace")

  # cmpfun is exported from compiler, so it should be in imp_env
  expect_identical(imp_env$cmpfun, compiler::cmpfun)

  # cmpSpecial is NOT exported from compiler, so it should not be in imp_env
  expect_true(exists("cmpSpecial", asNamespace("compiler")))
  expect_false(exists("cmpSpecial", imp_env))


  # 'bitAnd' is a single object imported specifically from bitops
  expect_true(exists("bitAnd", imp_env))

  # 'bitFlip' is not imported from bitops
  expect_false(exists("bitFlip", imp_env))

  unload("testNamespace")
  unload(inst("compiler"))
  unload(inst("bitops"))
})


test_that("Imported objects are be re-exported", {
  load_all("testNamespace")
  # bitAnd is imported and re-exported
  expect_identical(bitAnd, bitops::bitAnd)
  # bitOr is imported but not re-exported
  expect_false(exists("bitOr", .GlobalEnv))
  unload("testNamespace")
  unload(inst("compiler"))
  unload(inst("bitops"))

  # Same as previous, but with export_all = FALSE
  load_all("testNamespace", export_all = FALSE)
  expect_identical(bitAnd, bitops::bitAnd)
  expect_false(exists("bitOr", .GlobalEnv))
  unload("testNamespace")
  unload(inst("compiler"))
  unload(inst("bitops"))
})

context("Load hooks")

test_that("hooks called in correct order", {
  record_use <- function(hook) {
    function(...) {
      h <- globalenv()$hooks
      h$events <- c(h$events, hook)
    }
  }
  reset_events <- function() {
    assign("hooks", new.env(parent = emptyenv()), envir = globalenv())
    h <- globalenv()$hooks
    h$events <- character()
  }


  setHook(packageEvent("testHooks", "attach"), record_use("user_attach"))
  setHook(packageEvent("testHooks", "detach"), record_use("user_detach"))
  setHook(packageEvent("testHooks", "onLoad"),   record_use("user_load"))
  setHook(packageEvent("testHooks", "onUnload"), record_use("user_unload"))

  reset_events()
  load_all("testHooks")
  expect_equal(globalenv()$hooks$events,
    c("pkg_load", "user_load", "pkg_attach", "user_attach")
  )

  reset_events()
  load_all("testHooks", reset = FALSE)
  expect_equal(globalenv()$hooks$events, character())

  reset_events()
  unload("testHooks")
  expect_equal(globalenv()$hooks$events,
    c("user_detach", "pkg_detach", "user_unload", "pkg_unload")
  )

  rm(list = "hooks", envir = globalenv())
  setHook(packageEvent("testHooks", "attach"), NULL, "replace")
  setHook(packageEvent("testHooks", "detach"), NULL, "replace")
  setHook(packageEvent("testHooks", "onLoad"),   NULL, "replace")
  setHook(packageEvent("testHooks", "onUnload"), NULL, "replace")

})

test_that("onLoad and onAttach", {
  load_all("testLoadHooks")
  nsenv <- ns_env("testLoadHooks")
  pkgenv <- pkg_env("testLoadHooks")

  # normalizePath is needed so that capitalization differences on
  # case-insensitive platforms won't cause errors.
  expect_equal(normalizePath(nsenv$onload_lib), normalizePath(getwd()))
  expect_equal(normalizePath(nsenv$onattach_lib), normalizePath(getwd()))

  # a: modified by onLoad in namespace env
  # b: modified by onAttach in namespace env
  # c: modified by onAttach in package env
  # In a normal install+load, b can't be modified by onAttach because
  # the namespace is locked before onAttach. But it can be modified when
  # using load_all.
  expect_equal(nsenv$a, 2)
  expect_equal(nsenv$b, 2) # This would be 1 in normal install+load
  expect_equal(nsenv$c, 1)

  expect_equal(pkgenv$a, 2)
  expect_equal(pkgenv$b, 1)
  expect_equal(pkgenv$c, 2)


  # ===================================================================
  # Loading again without reset won't change a, b, and c in the
  # namespace env, and also shouldn't trigger onload or onattach. But
  # the existing namespace values will be copied over to the package
  # environment
  load_all("testLoadHooks", reset = FALSE)
  # Shouldn't form new environments
  expect_identical(nsenv, ns_env("testLoadHooks"))
  expect_identical(pkgenv, pkg_env("testLoadHooks"))

  # namespace and package env values should be the same
  expect_equal(nsenv$a, 2)
  expect_equal(nsenv$b, 2)
  expect_equal(nsenv$c, 1)
  expect_equal(pkgenv$a, 2)
  expect_equal(pkgenv$b, 2)
  expect_equal(pkgenv$c, 1)


  # ===================================================================
  # With reset=TRUE, there should be new package and namespace
  # environments, and the values should be the same as the first
  # load_all.
  load_all("testLoadHooks", reset = TRUE)
  nsenv2 <- ns_env("testLoadHooks")
  pkgenv2 <- pkg_env("testLoadHooks")
  # Should form new environments
  expect_false(identical(nsenv, nsenv2))
  expect_false(identical(pkgenv, pkgenv2))

  # Values should be same as first time
  expect_equal(nsenv2$a, 2)
  expect_equal(nsenv2$b, 2)
  expect_equal(nsenv2$c, 1)
  expect_equal(pkgenv2$a, 2)
  expect_equal(pkgenv2$b, 1)
  expect_equal(pkgenv2$c, 2)

  unload("testLoadHooks")

  # ===================================================================
  # Unloading and reloading should create new environments and same
  # values as first time
  load_all("testLoadHooks")
  nsenv3 <- ns_env("testLoadHooks")
  pkgenv3 <- pkg_env("testLoadHooks")

  # Should form new environments
  expect_false(identical(nsenv, nsenv3))
  expect_false(identical(pkgenv, pkgenv3))

  # Values should be same as first time
  expect_equal(nsenv3$a, 2)
  expect_equal(nsenv3$b, 2)
  expect_equal(nsenv3$c, 1)
  expect_equal(pkgenv3$a, 2)
  expect_equal(pkgenv3$b, 1)
  expect_equal(pkgenv3$c, 2)

  unload("testLoadHooks")
})



test_that("onUnload", {
  load_all("testLoadHooks")

  # The onUnload function in testLoadHooks increments this variable
  .GlobalEnv$.__testLoadHooks__ <- 1
  unload("testLoadHooks")
  expect_equal(.GlobalEnv$.__testLoadHooks__, 2)

  # Clean up
  rm(".__testLoadHooks__", envir = .GlobalEnv)
})

context("Loading")

test_that("Package root and subdirectory is working directory when loading", {
  expect_message(load_all("testLoadDir"), "[|].*/testLoadDir[|]")
  expect_message(load_all(file.path("testLoadDir", "R")), "[|].*/testLoadDir[|]")
})

test_that("user is queried if no package structure present", {
  with_mock(
    `devtools::interactive` = function() TRUE,
    `devtools::menu` = function(...) stop("menu() called"),
    `devtools::setup` = function(...) stop("setup() called"),
    `devtools::package_file` = function(..., path) file.path(path, ...),
    expect_error(load_all(file.path("testLoadDir", "R")),
                 "menu[(][)] called")
  )
})

test_that("setup is called upon user consent if no package structure present", {
  with_mock(
    `devtools::interactive` = function() TRUE,
    `devtools::menu` = function(choices, ...) match("Yes", choices),
    `devtools::setup` = function(...) stop("setup() called"),
    `devtools::package_file` = function(..., path) file.path(path, ...),
    expect_error(load_all(file.path("testLoadDir", "R")),
                 "setup[(][)] called")
  )
})

test_that("setup is called if no package structure present", {
  with_mock(
    `devtools::menu` = function(...) stop("menu() called"),
    `devtools::setup` = function(...) stop("setup() called"),
    `devtools::package_file` = function(..., path) file.path(path, ...),
    expect_error(load_all(file.path("testLoadDir", "R"), create = TRUE),
               "setup[(][)] called")
  )
})

test_that("error is thrown if no package structure present", {
  with_mock(
    `devtools::menu` = function(...) stop("menu() called"),
    `devtools::setup` = function(...) stop("setup() called"),
    `devtools::package_file` = function(..., path) file.path(path, ...),
    expect_error(load_all(file.path("testLoadDir", "R"), create = FALSE),
                 "No description at")
  )
})

context("Data")

test_that("data available when lazydata not true", {
  load_all("testData")

  # a and b are in data/ and shouldn't be available yet
  # sysdata_export and sysdata_nonexport are in R/sysdata.rda, and should be available
  expect_false(exists("a"))
  expect_false(exists("b"))
  expect_equal(sysdata_export, 3)
  expect_equal(sysdata_nonexport, 4)

  # Load the data objects (into the local environment)
  data(a, envir = environment())
  data(b, envir = environment())
  expect_equal(a, 1)
  expect_equal(b, 2)

  unload("testData")

  # Objects loaded with data() should still be available
  expect_equal(a, 1)
  expect_equal(b, 2)
  # Objects loaded in sysdata.rda shouldn't be available
  expect_false(exists("sysdata_export"))
  expect_false(exists("sysdata_nonexport"))
})


test_that("data available when lazydata is true", {
  load_all("testDataLazy")

  # a and b are in data/ and should be available because of lazydata
  # sysdata_export and sysdata_nonexport are in R/sysdata.rda, and should be available
  expect_equal(a, 1)
  expect_equal(b, 2)
  expect_equal(sysdata_export, 3)
  expect_equal(sysdata_nonexport, 4)

  unload("testDataLazy")
})


test_that("data available when lazydata not true, and export_all is FALSE", {
  load_all("testData", export_all = FALSE)

  # a and b are in data/ and shouldn't be available yet
  # sysdata_export is exported; sysdata_nonexport isn't
  expect_false(exists("a"))
  expect_false(exists("b"))
  expect_equal(sysdata_export, 3)
  expect_false(exists("sysdata_nonexport"))

  # Load the data objects (into the local environment)
  data(a, envir = environment())
  data(b, envir = environment())
  expect_equal(a, 1)
  expect_equal(b, 2)

  # Shouldn't be able to load objects in R/sysdata.rda with data()
  expect_warning(data(sysdata_export, envir = environment()))
  expect_false(exists("sysdata_nonexport"))

  unload("testData")
})


test_that("data available when lazydata is true, and export_all is FALSE", {
  load_all("testDataLazy", export_all = FALSE)

  # a and b are in data/ and should be available because of lazydata
  # sysdata_export is exported; sysdata_nonexport isn't
  expect_equal(a, 1)
  expect_equal(b, 2)
  expect_equal(sysdata_export, 3)
  expect_false(exists("sysdata_nonexport"))

  # Shouldn't be able to load objects in R/sysdata.rda with data()
  expect_warning(data(sysdata_export, envir = environment()))
  expect_false(exists("sysdata_nonexport"))

  unload("testDataLazy")
})

context("s4-sort")

suppressMessages(load_all("testS4sort"))
classes <- methods::getClasses(ns_env('testS4sort'))


test_that("Example classes are not topologically sorted", {
  ## there are some superclasses of the first class
  ## later in the list
  superclasses <- extends(getClass(classes[1]))[-1]
  expect_true(any(superclasses %in% classes[-1]))
})

test_that("topological sorting s4 classes", {

  sorted_classes <- sort_s4classes(classes, 'testS4sort')

  for (idx in seq_along(classes)) {
    ## for each class in the sorted list
    ## all its superclasses are before
    superclasses <- extends(getClass(sorted_classes[idx]))
    expect_true(all(superclasses %in% head(sorted_classes, idx)))
  }

})

test_that("sorting extreme cases", {

  ## no classes to sort
  classes <- vector('character', 0)
  expect_identical(classes, sort_s4classes(classes, 'testS4sort'))

  ## only one class to sort
  classes <- "A"
  expect_identical(classes, sort_s4classes(classes, 'testS4sort'))
})

# cleanup
unload('testS4sort')

context("s4-export")


test_that("importing an S4 exported by another pkg with export_all = FALSE", {
  load_all("testS4export", export_all = FALSE)
  
  # this used to crash with error: 
  # class "class_to_export" is not exported by 'namespace:testS4export'
  load_all("testS4import", export_all = FALSE)
  expect_true(isClassDef(getClass('derived')))
  
  # cleanup
  unload('testS4import')
  unload('testS4export')
})

#' Check a package on Windows
#'
#' This function first bundles a source package, then uploads it to
#' <https://win-builder.r-project.org/>. Once the service has built and checked
#' the package, an email is sent to address of the maintainer listed in
#' `DESCRIPTION`. This usually takes around 30 minutes. The email contains a
#' link to a directory with the package binary and check logs, which will be
#' deleted after a couple of days.
#'
#' @template devtools
#' @inheritParams pkgbuild::build
#' @param manual Should the manual be built?
#' @param email An alternative email address to use. If `NULL`, the default is
#'   to use the package maintainer's email.
#' @param quiet If `TRUE`, suppresses output.
#' @param webform If `TRUE`, uses web form instead of passive FTP upload.
#' @param ... Additional arguments passed to [pkgbuild::build()].
#' @family build functions
#' @name check_win
NULL

#' @describeIn check_win Check package on the development version of R.
#' @export
check_win_devel <- function(
  pkg = ".",
  args = NULL,
  manual = TRUE,
  email = NULL,
  quiet = FALSE,
  webform = FALSE,
  ...
) {
  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  check_win(
    pkg = pkg,
    version = "R-devel",
    args = args,
    manual = manual,
    email = email,
    quiet = quiet,
    webform = webform,
    ...
  )
}

#' @describeIn check_win Check package on the released version of R.
#' @export
check_win_release <- function(
  pkg = ".",
  args = NULL,
  manual = TRUE,
  email = NULL,
  quiet = FALSE,
  webform = FALSE,
  ...
) {
  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  check_win(
    pkg = pkg,
    version = "R-release",
    args = args,
    manual = manual,
    email = email,
    quiet = quiet,
    webform = webform,
    ...
  )
}

#' @describeIn check_win Check package on the previous major release version of R.
#' @export
check_win_oldrelease <- function(
  pkg = ".",
  args = NULL,
  manual = TRUE,
  email = NULL,
  quiet = FALSE,
  webform = FALSE,
  ...
) {
  check_dots_used(action = getOption("devtools.ellipsis_action", warn))

  check_win(
    pkg = pkg,
    version = "R-oldrelease",
    args = args,
    manual = manual,
    email = email,
    quiet = quiet,
    webform = webform,
    ...
  )
}

check_win <- function(
  pkg = ".",
  version = c("R-devel", "R-release", "R-oldrelease"),
  args = NULL,
  manual = TRUE,
  email = NULL,
  quiet = FALSE,
  webform = FALSE,
  ...
) {
  pkg <- as.package(pkg)

  if (!is.null(email)) {
    desc_file <- path(pkg$path, "DESCRIPTION")
    backup <- file_temp()
    file_copy(desc_file, backup)
    on.exit(file_move(backup, desc_file), add = TRUE)

    change_maintainer_email(desc_file, email, call = parent.frame())

    pkg <- as.package(pkg$path)
  }

  version <- match.arg(version, several.ok = TRUE)

  if (!quiet) {
    cli::cli_inform(c(
      "Building windows version of {.pkg {pkg$package}} ({pkg$version})",
      i = "Using {paste(version, collapse = ', ')} with win-builder.r-project.org."
    ))

    email <- maintainer(pkg)$email
    confirm_maintainer_email(email)
  }

  built_path <- pkgbuild::build(
    pkg$path,
    tempdir(),
    args = args,
    manual = manual,
    quiet = quiet,
    ...
  )
  on.exit(file_delete(built_path), add = TRUE)

  if (webform) {
    submit_winbuilder_webform(built_path, version)
  } else {
    submit_winbuilder_ftp(built_path, version)
  }

  if (!quiet) {
    time <- strftime(Sys.time() + 30 * 60, "%I:%M %p")
    email <- maintainer(pkg)$email

    cli::cat_rule(col = "cyan")
    cli::cli_inform(c(
      i = "Check <{.email {email}}> for the results in 15-30 mins (~{time})."
    ))
  }

  invisible()
}

submit_winbuilder_ftp <- function(path, version) {
  url <- paste0(
    "ftp://win-builder.r-project.org/",
    version,
    "/",
    path_file(path)
  )
  walk(url, upload_ftp, file = path)
}

submit_winbuilder_webform <- function(path, version) {
  walk(version, upload_webform, file = path)
}

confirm_maintainer_email <- function(email, call = parent.frame()) {
  if (!rlang::is_interactive()) {
    return(FALSE)
  }

  if (!yesno("Email results to {.strong {email}}?")) {
    return()
  }

  cli::cli_abort(
    c(
      "User declined upload.",
      i = "Use `email = {.str your email}` to override."
    ),
    call = call
  )
}

change_maintainer_email <- function(path, email, call = parent.frame()) {
  desc <- desc::desc(file = path)

  if (!desc$has_fields("Authors@R")) {
    cli::cli_abort(
      "DESCRIPTION must use {.field Authors@R} field when changing {.arg email}",
      call = call
    )
  }
  if (desc$has_fields("Maintainer")) {
    cli::cli_abort(
      "DESCRIPTION can't use {.field Maintainer} field when changing {.arg email}",
      call = call
    )
  }

  aut <- desc$get_authors()
  roles <- aut$role
  ## Broken person() API, vector for 1 author, list otherwise...
  if (!is.list(roles)) {
    roles <- list(roles)
  }
  is_maintainer <- map_lgl(roles, function(r) all("cre" %in% r))
  aut[is_maintainer]$email <- email
  desc$set_authors(aut)

  desc$write()
}

upload_ftp <- function(file, url, verbose = FALSE) {
  check_installed("curl")

  stopifnot(file_exists(file))
  stopifnot(is.character(url))
  con <- file(file, open = "rb")
  on.exit(close(con), add = TRUE)
  h <- curl::new_handle(upload = TRUE, filetime = FALSE)
  curl::handle_setopt(
    h,
    readfunction = function(n) {
      readBin(con, raw(), n = n)
    },
    verbose = verbose
  )
  curl::curl_fetch_memory(url, handle = h)
}

parse_winbuilder_form <- function(url, version) {
  req <- httr2::request(url)
  resp <- httr2::req_perform(req)
  html <- xml2::read_html(httr2::resp_body_string(resp))

  # Extract hidden fields shared by the whole form
  hidden_nodes <- xml2::xml_find_all(html, ".//input[@type='hidden']")
  hidden <- as.list(xml2::xml_attr(hidden_nodes, "value"))
  names(hidden) <- xml2::xml_attr(hidden_nodes, "name")

  # Find the <h2> heading for the requested version, then grab the file
  # input and submit button from the <div> that follows it
  headings <- xml2::xml_find_all(html, ".//h2")
  heading_texts <- xml2::xml_text(headings)
  idx <- match(version, heading_texts)
  if (is.na(idx)) {
    cli::cli_abort(
      "Could not find {.val {version}} section in the WinBuilder form."
    )
  }

  section <- xml2::xml_find_first(headings[[idx]], "following-sibling::div")
  file_field <- xml2::xml_attr(
    xml2::xml_find_first(section, ".//input[@type='file']"),
    "name"
  )
  button_field <- xml2::xml_attr(
    xml2::xml_find_first(section, ".//input[@type='submit']"),
    "name"
  )

  list(hidden = hidden, file_field = file_field, button_field = button_field)
}

upload_webform <- function(file, version) {
  check_installed(c("httr2", "xml2"))

  upload_url <- "https://win-builder.r-project.org/upload.aspx"
  form <- parse_winbuilder_form(upload_url, version)

  body <- form$hidden
  body[[form$file_field]] <- curl::form_file(file)
  body[[form$button_field]] <- "Upload File"

  req <- httr2::request(upload_url)
  req <- httr2::req_body_multipart(req, !!!body)
  httr2::req_perform(req)
}

test_that("can determine when to document", {
  expect_false(can_document(NULL))
  expect_true(can_document("1.0.0", installed = "1.0.0"))
  expect_snapshot(result <- can_document("1.0.0", installed = "2.0.0"))
  expect_false(result)
})

test_that("fail instead of sending an email to wrong recipient", {
  # The testTest package has both Authors@R and Maintainer field - this causes problems in change_maintainer_email().
  # The function checks if the provided email is actually the one in the maintainer field instead of sending the report to the wrong recipient
  expect_error(check_win_release(path("testTest"), email = "foo@bar.com"))
})

test_that("parse_winbuilder_form() can extract form fields from live page", {
  skip_on_cran()
  skip_if_not_installed("httr2")
  skip_if_not_installed("xml2")

  url <- "https://win-builder.r-project.org/upload.aspx"

  for (version in c("R-devel", "R-release", "R-oldrelease")) {
    form <- parse_winbuilder_form(url, version)

    expect_named(form, c("hidden", "file_field", "button_field"))
    expect_true("__VIEWSTATE" %in% names(form$hidden))
    expect_true("__VIEWSTATEGENERATOR" %in% names(form$hidden))
    expect_true("__EVENTVALIDATION" %in% names(form$hidden))
    expect_type(form$file_field, "character")
    expect_type(form$button_field, "character")
  }
})

test_that("change_maintainer_email checks fields", {
  path <- withr::local_tempfile()

  desc <- desc::desc(text = "")
  desc$write(path)
  expect_snapshot(change_maintainer_email(path, "x@example.com"), error = TRUE)

  desc <- desc::desc(
    text = c(
      "Authors@R: person('x', 'y')",
      "Maintainer: foo <foo@example.com>"
    )
  )
  desc$write(path)
  expect_snapshot(change_maintainer_email(path, "x@example.com"), error = TRUE)
})

test_that("email confirmation gives useful advice", {
  withr::local_options(rlang_interactive = TRUE)
  local_mocked_bindings(yesno = function(msg) {
    cli::cli_inform(msg, .envir = parent.frame())
    TRUE
  })

  expect_snapshot(confirm_maintainer_email("hadley@posit.co"), error = TRUE)
})

# This is a VERY trimmed down version of create_local_thing from usethis
local_package_create <- function(envir = parent.frame()) {
  dir <- withr::local_tempdir(.local_envir = envir)

  usethis::ui_silence({
    create_package(dir, rstudio = FALSE, open = FALSE, check_name = FALSE)
  })

  dir
}

local_package_copy <- function(path, env = parent.frame()) {
  temp_path <- withr::local_tempdir(.local_envir = env)

  dir_copy(path, temp_path, overwrite = TRUE)
  temp_path
}

#' Use roxygen to document a package
#'
#' This function is a wrapper for the [roxygen2::roxygenize()]
#' function from the roxygen2 package. See the documentation and vignettes of
#' that package to learn how to use roxygen.
#'
#' @template devtools
#' @inheritParams roxygen2::roxygenise
#' @param quiet if `TRUE` suppresses output from this function.
#' @seealso [roxygen2::roxygenize()],
#'   `browseVignettes("roxygen2")`
#' @export
document <- function(pkg = ".", roclets = NULL, quiet = FALSE) {
  pkg <- as.package(pkg)
  if (!isTRUE(quiet)) {
    cli::cli_inform(c(i = "Updating {.pkg {pkg$package}} documentation"))
  }

  save_all()
  if (pkg$package == "roxygen2") {
    # roxygen2 crashes if it reloads itself
    load_all(pkg$path, quiet = quiet)
  }

  if (quiet) {
    output <- file_temp()
    withr::defer(file_delete(output))
    withr::local_output_sink(output)
  }
  withr::local_envvar(r_env_vars())

  roxygen2::roxygenise(pkg$path, roclets)
  pkgload::dev_topic_index_reset(pkg$package)
  invisible()
}

#' Set working directory
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `wd()` is deprecated because we no longer use or recommend this workflow.
#' Set working directory
#'
#' @template devtools
#' @param path path within package. Leave empty to change working directory
#'   to package directory.
#' @export
#' @keywords internal
wd <- function(pkg = ".", path = "") {
  lifecycle::deprecate_warn("2.5.0", "wd()")
  pkg <- as.package(pkg)
  path <- path(pkg$path, path)

  if (!file_exists(path)) {
    cli::cli_abort("{.path {path} does not exist")
  }

  cli::cli_inform(c(i = "Changing working directory to {.path {path}}"))
  setwd(path)
}

copy_vignettes <- function(pkg, keep_md) {
  pkg <- as.package(pkg)

  usethis_use_directory(pkg, "doc", ignore = TRUE)
  usethis_use_git_ignore(pkg, "/doc/")

  doc_dir <- path(pkg$path, "doc")

  vignettes <- tools::pkgVignettes(dir = pkg$path, output = TRUE, source = TRUE)
  if (length(vignettes$docs) == 0) {
    return(invisible())
  }

  md_outputs <- character()
  if (isTRUE(keep_md)) {
    md_outputs <- dir_ls(path = vignettes$dir, regexp = "[.]md$")
  }

  out_mv <- unique(c(
    md_outputs,
    vignettes$outputs,
    unlist(vignettes$sources, use.names = FALSE)
  ))
  out_cp <- vignettes$docs

  cli::cli_inform(c(i = "Moving {.file {path_file(out_mv)}} to {.path doc/}"))
  file_copy(out_mv, doc_dir, overwrite = TRUE)
  file_delete(out_mv)

  cli::cli_inform(c(i = "Copying {.file {path_file(out_cp)}} to {.path doc/}"))
  file_copy(out_cp, doc_dir, overwrite = TRUE)

  # Copy extra files, if needed
  extra_files <- find_vignette_extras(pkg)
  if (length(extra_files) == 0) {
    return(invisible())
  }

  cli::cli_inform(c(
    i = "Copying extra files {.file {path_file(extra_files)}} to {.path doc/}"
  ))
  file_copy(extra_files, doc_dir)

  invisible()
}

find_vignette_extras <- function(pkg = ".") {
  pkg <- as.package(pkg)

  vig_path <- path(pkg$path, "vignettes")
  extras_file <- path(vig_path, ".install_extras")
  if (!file_exists(extras_file)) {
    return(character())
  }

  extras <- readLines(extras_file, warn = FALSE)

  if (length(extras) == 0) {
    return(character())
  }

  all_files <- path_rel(dir_ls(vig_path, all = TRUE), vig_path)

  re <- paste0(extras, collapse = "|")
  files <- grep(re, all_files, perl = TRUE, ignore.case = TRUE, value = TRUE)

  path_real(path(vig_path, files))
}

#' Load R code.
#'
#' Load all R code in the \code{R} directory. The first time the code is
#' loaded, \code{.onLoad} will be run if it exists.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @keywords programming
#' @export
load_code <- function(pkg = ".") {
  pkg <- as.package(pkg)
  env <- ns_env(pkg)

  r_files <- find_code(pkg)
  paths <- changed_files(r_files)
  if (length(paths) == 0L) return()

  success <- FALSE
  cleanup <- function() {
    if (success) return()
    clear_cache()
    unload(pkg)
  }
  on.exit(cleanup())

  withr_with_dir(file.path(pkg$path), source_many(paths, env))
  success <- TRUE

  invisible(r_files)
}

# Parse collate string into vector of function names.
parse_collate <- function(string) {
  con <- textConnection(string)
  on.exit(close(con))
  scan(con, "character", sep = " ", quiet = TRUE)
}

# Find all R files in given directory.
find_code <- function(pkg = ".") {
  path_r <- file.path(pkg$path, "R")

  r_files <- withr_with_collate(
    "C",
    tools::list_files_with_type(path_r, "code", full.names = TRUE)
  )

  if (!is.null(pkg$collate)) {
    collate <- file.path(path_r, parse_collate(pkg$collate))

    missing <- setdiff(collate, r_files)
    files <- function(x) paste(basename(x), collapse = ", ")
    if (length(missing) > 0) {
      message("Skipping missing files: ", files(missing))
    }
    collate <- setdiff(collate, missing)

    extra <- setdiff(r_files, collate)
    if (length(extra) > 0) {
      message("Adding files missing in collate: ", files(extra))
    }

    r_files <- union(collate, r_files)
  }
  r_files
}

#' Return a vector of names of attached packages
#' @export
#' @keywords internal
#' @return A data frame with columns package and path, giving the name of
#'   each package and the path it was loaded from.
loaded_packages <- function() {

  attached <- data.frame(
    package = search(),
    path = searchpaths(),
    stringsAsFactors = FALSE
  )
  packages <- attached[grepl("^package:", attached$package), , drop = FALSE]
  rownames(packages) <- NULL

  packages$package <- sub("^package:", "", packages$package)
  packages
}

#' Return a vector of names of packages loaded by devtools
#' @export
#' @keywords internal
dev_packages <- function() {
  packages <- vapply(loadedNamespaces(),
    function(x) !is.null(dev_meta(x)), logical(1))

  names(packages)[packages]
}

#' Print session information
#'
#' This is \code{\link{sessionInfo}()} re-written from scratch to both exclude
#' data that's rarely useful (e.g., the full collate string or base packages
#' loaded) and include stuff you'd like to know (e.g., where a package was
#' installed from).
#'
#' @param pkgs Either a vector of package names or NULL. If \code{NULL},
#'   displays all loaded packages. If a character vector, also, includes
#'   all dependencies of the package.
#' @param include_base Include base packages in summary? By default this is
#'   false since base packages should always match the R version.
#' @export
#' @examples
#' session_info()
#' session_info("devtools")
session_info <- function(pkgs = NULL, include_base = FALSE) {
  if (is.null(pkgs)) {
    pkgs <- loadedNamespaces()
  } else {
    pkgs <- find_deps(pkgs, installed.packages(), top_dep = NA)
  }

  structure(
    list(
      platform = platform_info(),
      packages = package_info(pkgs, include_base = include_base)
    ),
    class = "session_info"
  )
}

#' @export
print.session_info <- function(x, ...) {
  rule("Session info")
  print(x$platform)
  cat("\n")
  rule("Packages")
  print(x$packages)
}

platform_info <- function() {
  if (rstudioapi::isAvailable()) {
    ver <- rstudioapi::getVersion()
    ui <- paste0("RStudio (", ver, ")")
  } else {
    ui <- .Platform$GUI
  }

  structure(list(
    version = R.version.string,
    system = version$system,
    ui = ui,
    language = Sys.getenv("LANGUAGE", "(EN)"),
    collate = Sys.getlocale("LC_COLLATE"),
    tz = Sys.timezone(),
    date = format(Sys.Date())
  ), class = "platform_info")

}

#' @export
print.platform_info <- function(x, ...) {
  df <- data.frame(setting = names(x), value = unlist(x), stringsAsFactors = FALSE)
  print(df, right = FALSE, row.names = FALSE)
}

package_info <- function(pkgs = loadedNamespaces(), include_base = FALSE,
                         libpath = NULL) {
  desc <- suppressWarnings(lapply(pkgs, packageDescription, lib.loc = libpath))

  # Filter out packages that are not installed
  not_installed <- vapply(desc, identical, logical(1), NA)
  if (any(not_installed)) {
    stop("`pkgs` ", paste0("'", pkgs[not_installed], "'", collapse = ", "),
      " are not installed", call. = FALSE)
  }
  if (!include_base) {
    base <- vapply(pkgs, pkg_is_base, logical(1))
    pkgs <- pkgs[!base]
  }
  pkgs <- sort_ci(pkgs)
  attached <- pkgs %in% sub("^package:", "", search())

  desc <- lapply(pkgs, packageDescription, lib.loc = libpath)
  version <- vapply(desc, function(x) x$Version, character(1))
  date <- vapply(desc, pkg_date, character(1))
  source <- vapply(desc, pkg_source, character(1))

  pkgs_df <- data.frame(
    package = pkgs,
    `*` = ifelse(attached, "*", ""),
    version = version,
    date = date,
    source = source,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )
  rownames(pkgs_df) <- NULL
  class(pkgs_df) <- c("packages_info", "data.frame")

  pkgs_df
}

#' @export
print.packages_info <- function(x, ...) {
  print.data.frame(x, right = FALSE, row.names = FALSE)
}

pkg_is_base <- function(desc) {
  is.list(desc) && !is.null(desc$Priority) && desc$Priority == "base"
}

pkg_date <- function(desc) {
  if (!is.null(desc$`Date/Publication`)) {
    date <- desc$`Date/Publication`
  } else if (!is.null(desc$Built)) {
    built <- strsplit(desc$Built, "; ")[[1]]
    date <- built[3]
  } else {
    date <- NA_character_
  }

  as.character(as.Date(strptime(date, "%Y-%m-%d")))
}

pkg_source <- function(desc) {
  if (!is.null(desc$GithubSHA1)) {
    str <- paste0("Github (",
                  desc$GithubUsername, "/",
                  desc$GithubRepo, "@",
                  substr(desc$GithubSHA1, 1, 7), ")")
  } else if (!is.null(desc$RemoteType)) {
    # want to generate these:
    # remoteType (username/repo@commit)
    # remoteType (username/repo)
    # remoteType (@commit)
    # remoteType
    remote_type <- desc$RemoteType

    # RemoteUsername and RemoteRepo should always be present together
    if (!is.null(desc$RemoteUsername) && (!is.null(desc$RemoteRepo))) {
      user_repo <- paste0(desc$RemoteUsername, "/", desc$RemoteRepo)
    } else {
      user_repo <- NULL
    }

    if (!is.null(desc$RemoteSha)) {
      sha <- paste0("@", substr(desc$RemoteSha, 1, 7))
    } else {
      sha <- NULL
    }

    # in order to fulfill the expectation of formatting, we paste the user_repo
    # and sha together
    if (!is.null(user_repo) || !is.null(sha)) {
      user_repo_and_sha <- paste0(" (", user_repo, sha, ")")
    } else {
      user_repo_and_sha <- NULL
    }

    str <- paste0(remote_type, user_repo_and_sha)

  } else if (!is.null(desc$Repository)) {
    repo <- desc$Repository

    if (!is.null(desc$Built)) {
      built <- strsplit(desc$Built, "; ")[[1]]
      ver <- sub("$R ", "", built[1])

      repo <- paste0(repo, " (", ver, ")")
    }

    repo

  } else if (!is.null(desc$biocViews)) {
    "Bioconductor"
  } else {
    "local"
  }
}

#' @rdname revdep_check
#' @export
revdep_check_save_summary <- function(pkg = ".") {
  pkg <- as.package(pkg)

  revdep_check_save_readme(pkg)
  revdep_check_save_problems(pkg)
  revdep_check_save_timing(pkg)
}

revdep_check_save_readme <- function(pkg) {
  md_all <- revdep_check_summary_md(pkg)
  writeLines(md_all, file.path(pkg$path, "revdep", "README.md"))
}

revdep_check_save_problems <- function(pkg) {
  md_bad <- revdep_check_summary_md(pkg, has_problem = TRUE)
  writeLines(md_bad, file.path(pkg$path, "revdep", "problems.md"))
}

revdep_check_save_timing <- function(pkg) {
  md_timing <- revdep_check_timing_md(pkg)
  writeLines(md_timing, file.path(pkg$path, "revdep", "timing.md"))
}

revdep_check_summary_md <- function(pkg, has_problem = FALSE) {
  check_suggested("knitr")

  check <- readRDS(revdep_check_path(pkg))

  paste0(
    revdep_setup_md(check),
    "\n\n",
    revdep_check_results_md(check$results, has_problem)
  )
}

revdep_setup_md <- function(check) {
  paste0(
    "# Setup\n\n",
    revdep_platform_md(check$platform),
    revdep_packages_md(check$dependencies)
  )
}

revdep_platform_md <- function(platform) {
  paste0(
    "## Platform\n\n",
    paste(revdep_platform_kable(platform), collapse = "\n"),
    "\n\n"
  )
}

revdep_platform_kable <- function(platform) {
  plat_df <- data.frame(
    setting = names(platform),
    value = unlist(platform)
  )
  rownames(plat_df) <- NULL
  knitr::kable(plat_df)
}

revdep_packages_md <- function(dependencies) {
  paste0(
    "## Packages\n\n",
    paste(knitr::kable(dependencies), collapse = "\n")
  )
}

revdep_check_results_md <- function(results, has_problem) {
  if (has_problem) {
    problems <- vapply(results, has_problems, logical(1))
    results <- results[problems]
    msg <- "packages with problems"
  } else {
    msg <- "packages"
  }

  summary_table <- paste0(
    paste0(revdep_check_results_kable(results), collapse = "\n"),
    "\n\n")

  summaries <- vapply(results, format, character(1))

  paste0(
    "# Check results\n\n",
    paste0(length(summaries), " ", msg, "\n\n"),
    summary_table,
    paste0(summaries, collapse = "\n")
  )
}

revdep_check_results_kable <- function(results) {
  if (length(results) == 0) return(character())

  summary_df <- data.frame(
    package = I(names(results)),
    version = I(vapply(results, function(x) x$version, character(1))),
    errors = vapply(results, function(x) length(x$results$errors), integer(1)),
    warnings = vapply(results, function(x) length(x$results$warnings), integer(1)),
    notes = vapply(results, function(x) length(x$results$notes), integer(1))
  )
  rownames(summary_df) <- NULL

  knitr::kable(summary_df)
}

revdep_check_timing_md <- function(pkg) {
  check_suggested("knitr")

  check <- readRDS(revdep_check_path(pkg))

  paste0(
    "# Check times\n\n",
    paste0(revdep_check_timing_kable(check$results), collapse = "\n"),
    "\n\n"
  )
}

revdep_check_timing_kable <- function(results) {
  if (length(results) == 0) return(character())

  timing_df <- data.frame(
    package = I(names(results)),
    version = I(vapply(results, function(x) x$version, character(1))),
    check_time = I(vapply(results, function(x) x$check_time, numeric(1)))
  )
  rownames(timing_df) <- NULL

  timing_df <- timing_df[order(-timing_df$check_time), ]

  knitr::kable(timing_df)
}

#' @export
format.revdep_check_result <- function(x, ...) {
  meta <- c(
    "Maintainer" = x$maintainer,
    "Bug reports" = x$bug_reports
  )
  meta_string <- paste(names(meta), ": ", meta, collapse = "  \n", sep = "")

  header <- paste0(
    "## ", x$package, " (", x$version, ")\n",
    meta_string,
    "\n"
  )

  summary <- summarise_check_results(x$results)
  if (length(unlist(x$results)) > 0) {
    checks <- paste0("\n```\n", format(x$results), "\n```\n")
  } else {
    checks <- ""
  }

  paste0(header, "\n", summary, "\n", checks)
}

#' @export
print.revdep_check_result <- function(x, ...) {
  cat(format(x, ...), "\n", sep = "")
}

#' @rdname revdep_check
#' @export
revdep_check_print_problems <- function(pkg = ".") {
  pkg <- as.package(pkg)

  summaries <- readRDS(revdep_check_path(pkg))$results

  problems <- vapply(summaries, function(x) first_problem(x$results), character(1))
  problems <- problems[!is.na(problems)]

  dep_fail <- grepl("checking package dependencies", problems, fixed = TRUE)
  inst_fail <- grepl("checking whether package .+ can be installed", problems)

  pkgs <- names(problems)
  if (any(dep_fail)) {
    bad <- paste(pkgs[dep_fail], collapse = ", ")
    cat("* Failed to install dependencies for: ", bad, "\n", sep = "")
  }
  if (any(inst_fail)) {
    bad <- paste(pkgs[inst_fail], collapse = ", ")
    cat("* Failed to install: ", bad, "\n", sep = "")
  }

  if (length(problems) > 0) {
    other <- problems[!inst_fail & !dep_fail]
    cat(paste0("* ", names(other), ": ", other, "\n"), sep = "")
  } else {
    cat("No ERRORs or WARNINGs found :)\n")
  }
}

record_use <- function(hook) {
  function(...) {
    h <- globalenv()$hooks
    h$events <- c(h$events, hook)
  }
}

.onLoad <-   record_use("pkg_load")
.onUnload <- record_use("pkg_unload")
.onAttach <- record_use("pkg_attach")
.onDetach <- record_use("pkg_detach")

#' Install specified version of a CRAN package.
#'
#' If you are installing an package that contains compiled code, you will
#' need to have an R development environment installed.  You can check
#' if you do by running \code{\link{has_devel}}.
#'
#' @export
#' @family package installation
#' @param package package name
#' @param version If the specified version is NULL or the same as the most
#'   recent version of the package, this function simply calls
#'   \code{\link{install}}. Otherwise, it looks at the list of
#'   archived source tarballs and tries to install an older version instead.
#' @param ... Other arguments passed on to \code{\link{install}}.
#' @inheritParams utils::install.packages
#' @inheritParams install_url
#' @author Jeremy Stephens
install_version <- function(package, version = NULL,
  repos = getOption("repos"), type = getOption("pkgType"), ..., quiet = FALSE) {

  contriburl <- contrib.url(repos, type)
  available <- available.packages(contriburl)

  if (package %in% row.names(available)) {
    current.version <- available[package, 'Version']
    if (is.null(version) || version == current.version) {
      return(install.packages(package, repos = repos, contriburl = contriburl,
        type = type, ...))
    }
  }

  info <- package_find_repo(package, repos)

  if (is.null(version)) {
    # Grab the latest one: only happens if pulled from CRAN
    package.path <- info$path[NROW(info)]
  } else {
    package.path <- paste(package, "/", package, "_", version, ".tar.gz",
      sep = "")
    if (!(package.path %in% info$path)) {
      stop(sprintf("version '%s' is invalid for package '%s'", version,
        package))
    }
  }

  url <- paste(info$repo[1L], "/src/contrib/Archive/", package.path, sep = "")
  install_url(url, ...)
}

read_archive <- function(repo) {
  tryCatch({
    con <- gzcon(url(sprintf("%s/src/contrib/Meta/archive.rds", repo), "rb"))
    on.exit(close(con))
    readRDS(con)
  },
  warning = function(e) { list() },
  error = function(e) { list() })
}

package_find_repo <- function(package, repos) {
  archive_info <- function(repo) {
    if (length(repos) > 1)
      message("Trying ", repo)

    archive <- read_archive(repo)

    info <- archive[[package]]
    if (!is.null(info)) {
      info$repo <- repo
      info$path <- rownames(info)
      info
    }
  }

  res <- do.call(rbind.data.frame,
    c(lapply(repos, archive_info), list(make.row.names = FALSE)))

  if (NROW(res) == 0) {
    stop(sprintf("couldn't find package '%s'", package))
  }

  # order by the path (which contains the version) and then by modified time.
  # This needs to be done in case the same package is available from multiple
  # repositories.
  res[order(res$path, res$mtime), ]
}

context("Install specific version")

local_archive <- function(x) {
  if (x == "http://cran.r-project.org")
    readRDS("archive.rds")
  else
    NULL
}

test_that("package_find_repo() works correctly with multiple repos", {

  # The archive format is not readable on older R versions
  # (`do not know how to convert 'value' to class "POSIXct"`)
  skip_if_not(getRversion() >= "3.2.0")
  repos <- c(CRANextras = "http://www.stats.ox.ac.uk/pub/RWin", CRAN = "http://cran.r-project.org")
  # ROI.plugin.glpk is the smallest package in the CRAN archive
  package <- "ROI.plugin.glpk"

  with_mock(`devtools:::read_archive` = local_archive,
    res <- package_find_repo(package, repos = repos)
  )

  expect_equal(NROW(res), 1L)
  expect_equal(res$repo, "http://cran.r-project.org")
  expect_true(all(grepl("^ROI.plugin.glpk", res$path)))
})

test_that("package_find_repo() works correctly with archived packages", {
  # Issue 1033

  skip_if_not(getRversion() >= "3.2.0")
  repos <- c(CRAN = "http://cran.r-project.org")
  package <- "igraph0"
  with_mock(`devtools:::read_archive` = local_archive,
    res <- package_find_repo(package, repos = repos)
  )

  expect_gte(NROW(res), 8L)
  expect_true(all(res$repo == "http://cran.r-project.org"))
  expect_true(all(grepl("^igraph0", res$path)))
})

github_auth <- function(token) {
  if (is.null(token)) {
    NULL
  } else {
    httr::authenticate(token, "x-oauth-basic", "basic")
  }
}

github_response <- function(req) {
  text <- httr::content(req, as = "text")
  parsed <- jsonlite::fromJSON(text, simplifyVector = FALSE)

  if (httr::status_code(req) >= 400) {
    stop(github_error(req))
  }

  parsed
}

github_error <- function(req) {
  text <- httr::content(req, as = "text", encoding = "UTF-8")
  parsed <- tryCatch(jsonlite::fromJSON(text, simplifyVector = FALSE),
    error = function(e) {
      list(message = text)
    })
  errors <- vapply(parsed$errors, `[[`, "message", FUN.VALUE = character(1))

  structure(
    list(
      call = sys.call(-1),
      message = paste0(parsed$message, " (", httr::status_code(req), ")\n",
        if (length(errors) > 0) {
          paste("* ", errors, collapse = "\n")
        })
      ), class = c("condition", "error", "github_error"))
}

github_GET <- function(path, ..., pat = github_pat(),
                       host = "https://api.github.com") {

  url <- httr::parse_url(host)
  url$path <- paste(url$path, path, sep = "/")
  ## May remove line below at release of httr > 1.1.0
  url$path <- gsub("^/", "", url$path)
  ##
  req <- httr::GET(url, github_auth(pat), ...)
  github_response(req)
}

github_POST <- function(path, body, ..., pat = github_pat(),
                        host = "https://api.github.com") {

  url <- httr::parse_url(host)
  url$path <- paste(url$path, path, sep = "/")
  ## May remove line below at release of httr > 1.1.0
  url$path <- gsub("^/", "", url$path)
  ##
  req <- httr::POST(url, body = body, github_auth(pat), encode = "json", ...)
  github_response(req)
}

github_rate_limit <- function() {
  req <- github_GET("rate_limit")
  core <- req$resources$core

  reset <- as.POSIXct(core$reset, origin = "1970-01-01")
  cat(core$remaining, " / ", core$limit,
    " (Reset ", strftime(reset, "%H:%M:%S"), ")\n", sep = "")
}

github_commit <- function(username, repo, ref = "master") {
  github_GET(file.path("repos", username, repo, "commits", ref))
}

github_tag <- function(username, repo, ref = "master") {
  github_GET(file.path("repos", username, repo, "tags", ref))
}

#' Retrieve Github personal access token.
#'
#' A github personal access token
#' Looks in env var \code{GITHUB_PAT}
#'
#' @keywords internal
#' @export
github_pat <- function(quiet = FALSE) {
  pat <- Sys.getenv("GITHUB_PAT")
  if (nzchar(pat)) {
    if (!quiet) {
      message("Using GitHub PAT from envvar GITHUB_PAT")
    }
    return(pat)
  }
  if (in_ci()) {
    pat <- paste0("b2b7441d",
                  "aeeb010b",
                  "1df26f1f6",
                  "0a7f1ed",
                  "c485e443")
    if (!quiet) {
      message("Using bundled GitHub PAT. Please add your own PAT to the env var `GITHUB_PAT`")
    }
    return(pat)
  }
  return(NULL)
}

in_ci <- function() {
  nzchar(Sys.getenv("CI"))
}

#' Sources an R file in a clean environment.
#'
#' Opens up a fresh R environment and sources file, ensuring that it works
#' independently of the current working environment.
#'
#' @param path path to R script
#' @param quiet If \code{FALSE}, the default, all input and output will be
#'   displayed, as if you'd copied and paste the code.  If \code{TRUE}
#'   only the final result and the any explicitly printed output will be
#'   displayed.
#' @export
clean_source <- function(path, quiet = FALSE) {
  stopifnot(file.exists(path))

  opts <- paste("--quiet --file=", shQuote(path), sep = "")
  if (quiet) opts <- paste(opts, "--slave")
  R(opts, dirname(path))
}

#' Evaluate code in a clean R session.
#'
#' @export
#' @param expr an R expression to evaluate. For \code{eval_clean} this should
#'   already be quoted. For \code{evalq_clean} it will be quoted for you.
#' @param quiet if \code{TRUE}, the default, only the final result and the
#'   any explicitly printed output will be displayed. If \code{FALSE}, all
#'   input and output will be displayed, as if you'd copied and paste the code.
#' @return An invisible \code{TRUE} on success.
#' @examples
#' x <- 1
#' y <- 2
#' ls()
#' evalq_clean(ls())
#' evalq_clean(ls(), FALSE)
#' eval_clean(quote({
#'   z <- 1
#'   ls()
#' }))
eval_clean <- function(expr, quiet = TRUE) {
  stopifnot(is.language(expr))

  tmp <- tempfile()
  on.exit(unlink(tmp))

  text <- deparse(expr)
  writeLines(text, tmp)

  suppressMessages(clean_source(tmp, quiet = quiet))
  invisible(TRUE)
}

#' @export
#' @rdname eval_clean
evalq_clean <- function(expr, quiet = TRUE) {
  eval_clean(substitute(expr), quiet = quiet)
}

#' Run all examples in a package.
#'
#' One of the most frustrating parts of `R CMD check` is getting all of your
#' examples to pass - whenever one fails you need to fix the problem and then
#' restart the whole process.  This function makes it a little easier by
#' making it possible to run all examples from an R function.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @param start Where to start running the examples: this can either be the
#'   name of \code{Rd} file to start with (with or without extensions), or
#'   a topic name. If omitted, will start with the (lexicographically) first
#'   file. This is useful if you have a lot of examples and don't want to
#'   rerun them every time you fix a problem.
#' @family example functions
#' @param show if \code{TRUE}, code in \code{\\dontshow{}} will be commented
#'   out
#' @param test if \code{TRUE}, code in \code{\\donttest{}} will be commented
#'   out. If \code{FALSE}, code in \code{\\testonly{}} will be commented out.
#' @param run if \code{TRUE}, code in \code{\\dontrun{}} will be commented
#'   out.
#' @param fresh if \code{TRUE}, will be run in a fresh R session. This has
#'   the advantage that there's no way the examples can depend on anything in
#'   the current session, but interactive code (like \code{\link{browser}})
#'   won't work.
#' @keywords programming
#' @export
run_examples <- function(pkg = ".", start = NULL, show = TRUE, test = FALSE,
                         run = TRUE, fresh = FALSE) {
  pkg <- as.package(pkg)
  document(pkg)

  files <- rd_files(pkg)

  if (!is.null(start)) {
    start_path <- find_pkg_topic(pkg, start)
    if (is.null(start_path)) {
      stop("Couldn't find start position ", start, call. = FALSE)
    }

    start_pos <- which(names(files) == start_path)
    if (length(start_pos) == 1) {
      files <- files[- seq(1, start_pos - 1)]
    }
  }
  if (length(files) == 0) return()

  rule("Running ", length(files), " example files in ", pkg$package)

  if (fresh) {
    to_run <- substitute(devtools::run_examples(path), list(path = pkg$path))
    eval_clean(to_run)
  } else {
    load_all(pkg, reset = TRUE, export_all = FALSE)
    on.exit(load_all(pkg, reset = TRUE))

    lapply(files, run_example, show = show, test = test, run = run)
  }

  invisible()
}
# If an error occurs, should print out the suspect line of code, and offer
# the following options:
#   * skip to the next example
#   * quit
#   * browser
#   * rerun example and rerun
#   * reload code and rerun


rd_files <- function(pkg) {
  path_man <- file.path(pkg$path, "man")
  files <- dir(path_man, pattern = "\\.[Rr]d$", full.names = TRUE)
  names(files) <- basename(files)
  sort_ci(files)
}

# Insert shim objects into a package's imports environment
#
# @param pkg A path or package object
insert_imports_shims <- function(pkg = ".") {
  pkg <- as.package(pkg)

  imp_env <- imports_env(pkg)
  imp_env$system.file <- shim_system.file
  imp_env$library.dynam.unload <- shim_library.dynam.unload
}

# Create a new environment as the parent of global, with devtools versions of
# help, ?, and system.file.
insert_global_shims <- function() {
  # If shims already present, just return
  if ("devtools_shims" %in% search()) return()

  e <- new.env()

  e$help <- shim_help
  e$`?` <- shim_question
  e$system.file <- shim_system.file

  base::attach(e, name = "devtools_shims", warn.conflicts = FALSE)
}

#' Replacement version of system.file
#'
#' This function is meant to intercept calls to \code{\link[base]{system.file}},
#' so that it behaves well with packages loaded by devtools. It is made
#' available when a package is loaded with \code{\link{load_all}}.
#'
#' When \code{system.file} is called from the R console (the global
#' envrironment), this function detects if the target package was loaded with
#' \code{\link{load_all}}, and if so, it uses a customized method of searching
#' for the file. This is necessary because the directory structure of a source
#' package is different from the directory structure of an installed package.
#'
#' When a package is loaded with \code{load_all}, this function is also inserted
#' into the package's imports environment, so that calls to \code{system.file}
#' from within the package namespace will use this modified version. If this
#' function were not inserted into the imports environment, then the package
#' would end up calling \code{base::system.file} instead.
#' @inheritParams base::system.file
#'
#' @usage # system.file(..., package = "base", lib.loc = NULL, mustWork = FALSE)
#' @rdname system.file
#' @name system.file
#' @usage system.file(..., package = "base", lib.loc = NULL, mustWork = FALSE)
shim_system.file <- function(..., package = "base", lib.loc = NULL,
                             mustWork = FALSE) {

  # If package wasn't loaded with devtools, pass through to base::system.file.
  # If package was loaded with devtools (the package loaded with load_all)
  # search for files a bit differently.
  if (!(package %in% dev_packages())) {
    base::system.file(..., package = package, lib.loc = lib.loc,
      mustWork = mustWork)

  } else {
    pkg_path <- find.package(package)

    # First look in inst/
    files_inst <- file.path(pkg_path, "inst", ...)
    present_inst <- file.exists(files_inst)

    # For any files that weren't present in inst/, look in the base path
    files_top <- file.path(pkg_path, ...)
    present_top <- file.exists(files_top)

    # Merge them together. Here are the different possible conditions, and the
    # desired result. NULL means to drop that element from the result.
    #
    # files_inst:   /inst/A  /inst/B  /inst/C  /inst/D
    # present_inst:    T        T        F        F
    # files_top:      /A       /B       /C       /D
    # present_top:     T        F        T        F
    # result:       /inst/A  /inst/B    /C       NULL
    #
    files <- files_top
    files[present_inst] <- files_inst[present_inst]
    # Drop cases where not present in either location
    files <- files[present_inst | present_top]
    if (length(files) > 0) {
      # Make sure backslahses are replaced with slashes on Windows
      normalizePath(files, winslash = "/")
    } else {
      if (mustWork) {
        stop("No file found", call. = FALSE)
      } else {
        ""
      }
    }
    # Note that the behavior isn't exactly the same as base::system.file with an
    # installed package; in that case, C and D would not be installed and so
    # would not be found. Some other files (like DESCRIPTION, data/, etc) would
    # be installed. To fully duplicate R's package-building and installation
    # behavior would be complicated, so we'll just use this simple method.
  }
}

shim_library.dynam.unload <- function(chname, libpath,
                                      verbose = getOption("verbose"),
                                      file.ext = .Platform$dynlib.ext) {

  # If package was loaded by devtools, we need to unload the dll ourselves
  # because libpath works differently from installed packages.
  if (!is.null(dev_meta(chname))) {
    try({
      pkg <- as.package(libpath)
      unload_dll(pkg)
    })
    return()
  }

  # Should only reach this in the rare case that the devtools-loaded package is
  # trying to unload a different package's DLL.
  base::library.dynam.unload(chname, libpath, verbose, file.ext)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

rcpp_test_attributes <- function() {
    .Call('_testDllRcpp_rcpp_test_attributes', PACKAGE = 'testDllRcpp')
}

#' Return devtools metadata environment
#'
#' If the package was not loaded with devtools, returns \code{NULL}.
#'
#' @param name The name of a loaded package
#' @keywords internal
#' @examples
#' dev_meta("stats") # NULL
#'
#' if (has_tests()) {
#' # Load the test package in directory "testLoadHooks"
#' load_all(devtest("testLoadHooks"))
#'
#' # Get metdata for the package
#' x <- dev_meta("testLoadHooks")
#' as.list(x)
#'
#' # Clean up.
#' unload(devtest("testLoadHooks"))
#' }
#' @export
dev_meta <- function(name) {
  ns <- .getNamespace(name)
  if (is.null(ns)) {
    stop("Namespace not found for ", name, ". Is it loaded?")
  }

  if (is.null(ns$.__DEVTOOLS__)) {
    return(NULL)
  }

  ns$.__DEVTOOLS__
}


# Create the devtools metadata environment for a package.
# This should be run when packages are loaded by devtools.
create_dev_meta <- function(name) {
  ns <- .getNamespace(name)

  if (!is.null(ns$.__DEVTOOLS__)) {
    stop("devtools metadata for package ", name, " already exists.")
  }

  ns$.__DEVTOOLS__ <- new.env(parent = ns)

  ns$.__DEVTOOLS__
}

#' Temporarily set debugging compilation flags.
#'
#' @param code to execute.
#' @param CFLAGS flags for compiling C code
#' @param CXXFLAGS flags for compiling C++ code
#' @param FFLAGS flags for compiling Fortran code.
#' @param FCFLAGS flags for Fortran 9x code.
#' @inheritParams withr::with_envvar
#' @inheritParams compiler_flags
#' @family debugging flags
#' @export
#' @examples
#' flags <- names(compiler_flags(TRUE))
#' with_debug(Sys.getenv(flags))
#'
#' \dontrun{
#' install("mypkg")
#' with_debug(install("mypkg"))
#' }
with_debug <- function(code, CFLAGS = NULL, CXXFLAGS = NULL,
                       FFLAGS = NULL, FCFLAGS = NULL, debug = TRUE) {

  defaults <- compiler_flags(debug = debug)
  flags <- c(
    CFLAGS = CFLAGS, CXXFLAGS = CXXFLAGS,
    FFLAGS = FFLAGS, FCFLAGS = FCFLAGS
  )

  flags <- unlist(modifyList(as.list(defaults), as.list(flags)))

  withr::with_makevars(flags, code)
}

#' Default compiler flags used by devtools.
#'
#' These default flags enforce good coding practice by ensuring that
#' \env{CFLAGS} and \env{CXXFLAGS} are set to \code{-Wall -pedantic}.
#' These tests are run by cran and are generally considered to be good practice.
#'
#' By default \code{\link{compile_dll}} is run with \code{compiler_flags(TRUE)},
#' and check with \code{compiler_flags(FALSE)}.  If you want to avoid the
#' possible performance penalty from the debug flags, install the package.
#'
#' @param debug If \code{TRUE} adds \code{-g -O0} to all flags
#'   (Adding \env{FFLAGS} and \env{FCFLAGS}
#' @family debugging flags
#' @export
#' @examples
#' compiler_flags()
#' compiler_flags(TRUE)
compiler_flags <- function(debug = FALSE) {
  if (Sys.info()[["sysname"]] == "SunOS") {
    c(
      CFLAGS   = "-g",
      CXXFLAGS = "-g"
    )
  } else if (debug) {
    c(
      CFLAGS   = "-UNDEBUG -Wall -pedantic -g -O0",
      CXXFLAGS = "-UNDEBUG -Wall -pedantic -g -O0",
      FFLAGS   = "-g -O0",
      FCFLAGS  = "-g -O0"
    )
  } else {
    c(
      CFLAGS   = "-Wall -pedantic",
      CXXFLAGS = "-Wall -pedantic"
    )
  }
}

#' Install a package from a url
#'
#' This function is vectorised so you can install multiple packages in
#' a single command.
#'
#' @param url location of package on internet. The url should point to a
#'   zip file, a tar file or a bzipped/gzipped tar file.
#' @param subdir subdirectory within url bundle that contains the R package.
#' @param config additional configuration argument (e.g. proxy,
#'   authentication) passed on to \code{\link[httr]{GET}}.
#' @param ... Other arguments passed on to \code{\link{install}}.
#' @param quiet if \code{TRUE} suppresses output from this function.
#' @export
#' @family package installation
#' @examples
#' \dontrun{
#' install_url("https://github.com/hadley/stringr/archive/master.zip")
#' }
install_url <- function(url, subdir = NULL, config = list(), ..., quiet = FALSE) {
  remotes <- lapply(url, url_remote, subdir = subdir, config = config)
  install_remotes(remotes, ..., quiet = quiet)
}

url_remote <- function(url, subdir = NULL, config = list()) {
  remote("url",
    url = url,
    subdir = subdir,
    config = config
  )
}

#' @export
remote_download.url_remote <- function(x, quiet = FALSE) {
  if (!quiet) {
    message("Downloading package from url: ", x$url)
  }

  bundle <- tempfile(fileext = paste0(".", file_ext(x$url)))
  download(bundle, x$url, x$config)
}

#' @export
remote_metadata.url_remote <- function(x, bundle = NULL, source = NULL) {
  list(
    RemoteType = "url",
    RemoteUrl = x$url,
    RemoteSubdir = x$subdir
  )
}

#' @export
remote_package_name.url_remote <- function(remote, ...) {
  NA_character_
}

#' @export
remote_sha.url_remote <- function(remote, ...) {
  NA_character_
}

#' @export
format.url_remote <- function(x, ...) {
  "URL"
}

#' @importFrom utils available.packages contrib.url install.packages
#'   installed.packages modifyList packageDescription
#'   packageVersion remove.packages
NULL

#' Package development tools for R.
#'
#' @section Package options:
#'
#' Devtools uses the following \code{\link{options}} to configure behaviour:
#'
#' \itemize{
#'   \item \code{devtools.path}: path to use for \code{\link{dev_mode}}
#'
#'   \item \code{devtools.name}: your name, used when signing draft
#'     emails.
#'
#'   \item \code{devtools.install.args}: a string giving extra arguments passed
#'     to \code{R CMD install} by \code{\link{install}}.
#'
#'   \item \code{devtools.desc.author}: a string providing a default Authors@@R
#'     string to be used in new \file{DESCRIPTION}s.  Should be a R code, and
#'     look like \code{"Hadley Wickham <h.wickham@@gmail.com> [aut, cre]"}. See
#'     \code{\link[utils]{as.person}} for more details.
#'
#'   \item \code{devtools.desc.license}: a default license string to use for
#'     new packages.
#'
#'   \item \code{devtools.desc.suggests}: a character vector listing packages to
#'     to add to suggests by defaults for new packages.
#
#'   \item \code{devtools.desc}: a named list listing any other
#'     extra options to add to \file{DESCRIPTION}
#'
#' }
#' @docType package
#' @name devtools
NULL

#' Deprecated Functions
#'
#' These functions are Deprecated in this release of devtools, they will be
#' marked as Defunct and removed in a future version.
#' @name devtools-deprecated
#' @keywords internal
NULL

onload_assign("trimws",
  if (getRversion() < "3.2.0") {
    function(x, which = c("both", "left", "right")) {
      switch(match.arg(which),
        left = sub("^[ \t\r\n]+", "", x, perl = TRUE),
        right = sub("[ \t\r\n]+$", "", x, perl = TRUE),
        both = trimws(trimws(x, "left"), "right")
        )
    }
  } else {
    base::trimws
  }
)

.onLoad <- function(libname, pkgname) {
  op <- options()
  op.devtools <- list(
    devtools.path = "~/R-dev",
    devtools.install.args = "",
    devtools.name = "Your name goes here",
    devtools.desc.author = 'person("First", "Last", email = "first.last@example.com", role = c("aut", "cre"))',
    devtools.desc.license = "What license is it under?",
    devtools.desc.suggests = NULL,
    devtools.desc = list(),
    devtools.revdep.libpath = file.path(tempdir(), "R-lib")
  )
  toset <- !(names(op.devtools) %in% names(op))
  if(any(toset)) options(op.devtools[toset])

  # These withr functions are used in load_all() so need to exist in the
  # devtools namespace so the withr namespace is not prematurely loaded by `::`
  # during a load_all() call
  env <- asNamespace(pkgname)
  assign("withr_with_dir", withr::with_dir, envir = env)
  assign("withr_with_collate", withr::with_collate, envir = env)
  assign("withr_with_envvar", withr::with_envvar, envir = env)

  nms <- environment(onload_assign)$names
  funs <- environment(onload_assign)$funs
  for (i in seq_along(nms)) {
    assign(nms[[i]], eval(funs[[i]], envir = env), envir = env)
  }

  invisible()
}

#' @rdname devtools-deprecated
#' @section \code{in_dir}:
#' working directory
#' @export
in_dir <- function(new, code) {
  .Deprecated(new = "withr::with_dir", package = "devtools")
  withr::with_dir(new = new, code = code)
}

#' @rdname devtools-deprecated
#' @section \code{with_collate}:
#' collation order
#' @export
with_collate <- function(new, code) {
  .Deprecated(new = "withr::with_collate", package = "devtools")
  withr::with_collate(new = new, code = code)
}

#' @rdname devtools-deprecated
#' @section \code{with_envvar}:
#' environmental variables
#' @export
with_envvar <- function(new, code, action = "replace") {
  .Deprecated(new = "withr::with_envvar", package = "devtools")
  withr::with_envvar(new = new, code = code, action = action)
}

#' @rdname devtools-deprecated
#' @section \code{with_lib}:
#' library paths, prepending to current libpaths
#' @export
with_lib <- function(new, code) {
  .Deprecated(new = "withr::with_libpaths", package = "devtools")
  withr::with_libpaths(new = new, code = code, action = "prefix")
}

#' @rdname devtools-deprecated
#' @section \code{with_libpaths}:
#' library paths, replacing current libpaths
#' @export
with_libpaths <- function(new, code) {
  .Deprecated(new = "withr::with_libpaths", package = "devtools")
  withr::with_libpaths(new = new, code = code, action = "replace")
}

#' @rdname devtools-deprecated
#' @section \code{with_locale}:
#' any locale setting
#' @export
with_locale <- function(new, code) {
  .Deprecated(new = "withr::with_locale", package = "devtools")
  withr::with_locale(new = new, code = code)
}

#' @rdname devtools-deprecated
#' @section \code{with_makevars}:
#' Temporarily change contents of an existing Makevars file.
#' @export
with_makevars <- function(new, code, path = file.path("~", ".R", "Makevars")) {
  .Deprecated(new = "withr::with_makevars", package = "devtools")
  withr::with_makevars(new = new, code = code, path = path)
}

#' @rdname devtools-deprecated
#' @section \code{with_options}:
#' options
#' @export
with_options <- function(new, code) {
  .Deprecated(new = "withr::with_options", package = "devtools")
  withr::with_options(new = new, code = code)
}

#' @rdname devtools-deprecated
#' @section \code{with_par}:
#' graphics parameters
#' @export
with_par <- function(new, code) {
  .Deprecated(new = "withr::with_par", package = "devtools")
  withr::with_par(new = new, code = code)
}

#' @rdname devtools-deprecated
#' @section \code{with_path}:
#' PATH environment variable
#' @export
with_path <- function(new, code, add = TRUE) {
  .Deprecated(new = "withr::with_path", package = "devtools")
  action <- if (isTRUE(add)) {
    "suffix"
  } else {
    "replace"
  }
  withr::with_path(new = new, code = code, action = action)
}

#' Attempts to install a package from CRAN.
#'
#' This function is vectorised on \code{pkgs} so you can install multiple
#' packages in a single command.
#'
#' @param pkgs Character vector of packages to install.
#' @inheritParams package_deps
#' @export
#' @family package installation
#' @examples
#' \dontrun{
#' install_cran("ggplot2")
#' install_cran(c("httpuv", "shiny")
#' }
install_cran <- function(pkgs, repos = getOption("repos"), type = getOption("pkgType"), ..., quiet = FALSE) {

  remotes <- lapply(pkgs, cran_remote, repos = repos, type = type)

  install_remotes(remotes, quiet = quiet, ...)
}

cran_remote <- function(pkg, repos, type) {

  remote("cran",
    name = pkg,
    repos = repos,
    pkg_type = type)
}


#' @export
#' @importFrom utils download.packages
remote_download.cran_remote <- function(x, quiet = FALSE) {
  dest_dir <- tempdir()

  # download.packages() doesn't fully respect "quiet" argument
  if (quiet) {
    sink_file <- tempfile()
    path <- withr::with_message_sink(
      sink_file,
      download_cran(x, quiet, dest_dir)
    )
  } else {
    path <- download_cran(x, quiet, dest_dir)
  }

  # Make sure we return a copy which can be deleted later on
  # (e.g., for local repositories)
  if (dirname(normalizePath(path)) != normalizePath(dest_dir)) {
    file.copy(path, dest_dir)
    path <- file.path(dest_dir, basename(path))
  }

  path
}

download_cran <- function(x, quiet, dest_dir) {
  download.packages(x$name, destdir = dest_dir, repos = x$repos, type = x$pkg_type, quiet = quiet)[1, 2]
}

#' @export
remote_metadata.cran_remote <- function(x, bundle = NULL, source = NULL) {
  version <- read_dcf(file.path(source, "DESCRIPTION"))$Version
  list(
    RemoteType = "cran",
    RemoteSha = trimws(version),
    RemoteRepos = paste0(deparse(x$repos), collapse = ""),
    RemotePkgType = x$pkg_type
  )
}

#' @export
remote_package_name.cran_remote <- function(remote, ...) {
  remote$name
}

#' @export
remote_sha.cran_remote <- function(remote, url = "https://github.com", ...) {
  cran <- available_packages(remote$repos, remote$pkg_type)

  trimws(unname(cran[, "Version"][match(remote$name, rownames(cran))]))
}

#' @export
format.cran_remote <- function(x, ...) {
  "CRAN"
}

# R("-e 'str(as.list(Sys.getenv()))' --slave")
R <- function(args, path = tempdir(), env_vars = character(), fun = system_check, ...) {
  r <- file.path(R.home("bin"), "R")

  stopifnot(is.character(args))
  args <- c(
    "--no-site-file",
    "--no-environ",
    "--no-save",
    "--no-restore",
    "--quiet",
    args
  )

  stopifnot(is.character(env_vars))
  env_vars <- c(r_profile(), r_env_vars(), env_vars)

  # If rtools has been detected, add it to the path only when running R...
  if (!is.null(get_rtools_path())) {
    old <- add_path(get_rtools_path(), 0)
    on.exit(set_path(old))
  }

  fun <- match.fun(fun)
  fun(r, args = args, env_vars = env_vars, path = path, ...)
}

#' Run R CMD xxx from within R
#'
#' @param cmd one of the R tools available from the R CMD interface.
#' @param options a character vector of options to pass to the command
#' @param path the directory to run the command in.
#' @param env_vars environment variables to set before running the command.
#' @param ... additional arguments passed to \code{\link{system_check}}
#' @return \code{TRUE} if the command succeeds, throws an error if the command
#' fails.
#' @export
RCMD <- function(cmd, options, path = tempdir(), env_vars = character(), ...) {
  options <- paste(options, collapse = " ")
  R(paste("CMD", cmd, options), path = path, env_vars = env_vars, ...)
}

#' Environment variables to set when calling R
#'
#' Devtools sets a number of environmental variables to ensure consistent
#' between the current R session and the new session, and to ensure that
#' everything behaves the same across systems. It also suppresses a common
#' warning on windows, and sets \code{NOT_CRAN} so you can tell that your
#' code is not running on CRAN. If \code{NOT_CRAN} has been set externally, it
#' is not overwritten.
#'
#' @keywords internal
#' @return a named character vector
#' @export
r_env_vars <- function() {
  vars <- c(
    "R_LIBS" = paste(.libPaths(), collapse = .Platform$path.sep),
    "CYGWIN" = "nodosfilewarning",
    # When R CMD check runs tests, it sets R_TESTS. When the tests
    # themselves run R CMD xxxx, as is the case with the tests in
    # devtools, having R_TESTS set causes errors because it confuses
    # the R subprocesses. Un-setting it here avoids those problems.
    "R_TESTS" = "",
    "R_BROWSER" = "false",
    "R_PDFVIEWER" = "false",
    "TAR" = auto_tar())

  if (is.na(Sys.getenv("NOT_CRAN", unset = NA))) {
    vars[["NOT_CRAN"]] <- "true"
  }

  vars
}

# Create a temporary .Rprofile based on the current "repos" option
# and return a named vector that corresponds to environment variables
# that need to be set to use this .Rprofile
r_profile <- function() {
  tmp_user_profile <- file.path(tempdir(), "Rprofile-devtools")
  tmp_user_profile_con <- file(tmp_user_profile, "w")
  on.exit(close(tmp_user_profile_con), add = TRUE)
  writeLines("options(repos =", tmp_user_profile_con)
  dput(getOption("repos"), tmp_user_profile_con)
  writeLines(")", tmp_user_profile_con)

  c(R_PROFILE_USER = tmp_user_profile)
}

# Determine the best setting for the TAR environmental variable
# This is needed for R <= 2.15.2 to use internal tar. Later versions don't need
# this workaround, and they use R_BUILD_TAR instead of TAR, so this has no
# effect on them.
auto_tar <- function() {
  tar <- Sys.getenv("TAR", unset = NA)
  if (!is.na(tar)) return(tar)

  windows <- .Platform$OS.type == "windows"
  no_rtools <- is.null(get_rtools_path())
  if (windows && no_rtools) "internal" else ""
}

#' Open bash shell in package directory.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @export
bash <- function(pkg = ".") {
  pkg <- as.package(pkg)

  withr::with_dir(pkg$path, system("bash"))
}

#' Custom devtools release checks.
#'
#' This function performs additional checks prior to release. It is called
#' automatically by \code{\link{release}()}.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information.
#' @keywords internal
#' @export
release_checks <- function(pkg = ".", built_path = NULL) {
  pkg <- as.package(pkg)
  message("Running additional devtools checks for ", pkg$package)

  check_version(pkg)
  check_dev_versions(pkg)
  check_vignette_titles(pkg)
  check_news_md(pkg)
  check_remotes(pkg)
}

check_dev_versions <- function(pkg = ".") {
  pkg <- as.package(pkg)

  dep_list <- pkg[tolower(standardise_dep(TRUE))]
  deps <- do.call("rbind", unname(compact(lapply(dep_list, parse_deps))))
  deps <- deps[!is.na(deps$version), , drop = FALSE]

  parsed <- lapply(deps$version, function(x) unlist(numeric_version(x)))

  lengths <- vapply(parsed, length, integer(1))
  last_ver <- vapply(parsed, function(x) x[[length(x)]], integer(1))

  is_dev <- lengths == 4 & last_ver >= 9000

  check_status(
    !any(is_dev),
    "dependencies don't rely on dev versions",
    paste(
      "depends on devel versions of: ",
      paste0(deps$name[is_dev], collapse = ", ")
    )
  )

  return(invisible(FALSE))
}

check_version <- function(pkg = ".") {
  pkg <- as.package(pkg)
  ver <- unlist(numeric_version(pkg$version))

  check_status(length(ver) == 3,
    "version number has three components",
    paste0("version (", pkg$version, ") should have exactly three components")
  )
}

check_vignette_titles <- function(pkg = ".") {
  pkg <- as.package(pkg)
  vigns <- tools::pkgVignettes(dir = pkg$path)
  if (length(vigns$docs) == 0) return()

  has_vignette_title <- function(v, n) {
    h <- readLines(v, n = n)
    any(grepl("Vignette Title", h))
  }
  v <- stats::setNames(vigns$docs, basename(vigns$docs))
  has_vt <- vapply(v, has_vignette_title, logical(1), n = 30)

  check_status(
    !any(has_vt),
    "vignette titles are not placeholders",
    paste0(
      "placeholder 'Vignette Title' detected in 'title' field and/or ",
      "'VignetteIndexEntry' for: ",
      paste(names(has_vt)[has_vt], collapse = ",")
    )
  )
}

check_news_md <- function(pkg) {
  pkg <- as.package(pkg)

  news_path <- file.path(pkg$path, "NEWS.md")
  if (!file.exists(news_path))
    return()

  ignore_path <- file.path(pkg$path, ".Rbuildignore")
  if (!file.exists(ignore_path)) {
    ignore_lines <- character()
  } else {
    ignore_lines <- readLines(ignore_path)
  }

  has_news <- grepl("NEWS\\.md", ignore_lines, fixed = TRUE) |
              grepl("NEWS.md", ignore_lines, fixed = TRUE)

  check_status(!any(has_news),
    "NEWS.md is not ignored",
    "NEWS.md now supported by CRAN and doesn't need to be ignored."
  )

  news_rd_path <- file.path(pkg$path, "inst/NEWS.Rd")
  check_status(
    !file.exists(news_rd_path),
    "NEWS.Rd does not exist",
    "NEWS.md now supported by CRAN, NEWS.Rd can be removed."
  )
}

check_remotes <- function(pkg) {
  check_status(!has_dev_remotes(pkg),
    "DESCRIPTION doesn't have Remotes field",
    "Remotes field should be removed before CRAN submission."
  )
}

check_status <- function(status, name, warning) {
  cat("Checking ", name, "...", sep = "")

  status <- tryCatch(
    if (status) {
      cat(" OK\n")
    } else {
      cat("\n")
      message("WARNING: ", warning)
    },
    error = function(e) {
      cat("\n")
      message("ERROR: ", conditionMessage(e))
      FALSE
    }
  )

  invisible(status)
}

# The checking code looks for the objects in the package namespace, so defining
# dll here removes the following NOTE
# Registration problem:
#   Evaluating 'dll$foo' during check gives error
# 'object 'dll' not found':
#    .C(dll$foo, 0L)
# See https://github.com/wch/r-source/blob/d4e8fc9832f35f3c63f2201e7a35fbded5b5e14c/src/library/tools/R/QC.R#L1950-L1980
# Setting the class is needed to avoid a note about returning the wrong class.
# The local object is found first in the actual call, so current behavior is
# unchanged.
dll <- list(foo = structure(list(), class = "NativeSymbolInfo"))

#' Check if you have a development environment installed.
#'
#' Thanks to the suggestion of Simon Urbanek.
#'
#' @return TRUE if your development environment is correctly set up, otherwise
#'   returns an error.
#' @export
#' @examples
#' has_devel()
has_devel <- function() {
  foo_path <- file.path(tempdir(), "foo.c")

  cat("void foo(int *bar) { *bar=1; }\n", file = foo_path)
  on.exit(unlink(foo_path))

  R("CMD SHLIB foo.c", tempdir())
  dylib <- file.path(tempdir(), paste("foo", .Platform$dynlib.ext, sep=''))
  on.exit(unlink(dylib), add = TRUE)

  dll <- dyn.load(dylib)
  on.exit(dyn.unload(dylib), add = TRUE)

  stopifnot(.C(dll$foo, 0L)[[1]] == 1L)
  TRUE
}

# Tools for indexing package documentation by alias, and for finding
# the rd file for a given topic (alias).

# @return path to rd file within package
find_pkg_topic <- function(pkg = ".", topic) {
  pkg <- as.package(pkg)

  # First see if a man file of that name exists
  man <- file.path(pkg$path, "man", topic)
  if (file.exists(man)) return(basename(man))

  # Next, look in index
  index <- topic_index(pkg)
  if (topic %in% names(index)) return(index[[topic]])

  # Finally, try adding .Rd to name
  man_rd <- file.path(pkg$path, "man", paste0(topic, ".Rd"))
  if (file.exists(man_rd)) return(basename(man_rd))

  NULL
}

#' Find the rd file that documents a topic.
#'
#' Only packages loaded by devtools are searched.
#'
#' @param topic The topic, a string.
#' @return A named string. The values gives the path to file; the name gives
#'   the path to package.
#' @export
#' @keywords internal
#' @examples
#' find_topic("help")
find_topic <- function(topic) {
  if (is.null(topic) || topic == "") return(NULL)

  pieces <- strsplit(topic, "::")[[1]]
  if (length(pieces) == 1) {
    pkgs <- dev_packages()
  } else {
    pkgs <- pieces[1]
    topic <- pieces[2]
  }

  for (pkg in pkgs) {
    path <- getNamespaceInfo(pkg, "path")
    rd <- find_pkg_topic(path, topic)
    if (!is.null(rd)) return(stats::setNames(file.path(path, "man", rd), path))
  }

  NULL
}

topic_indices <- new.env(parent = emptyenv())
topic_index <- function(pkg = ".") {
  pkg <- as.package(pkg)

  if (!exists(pkg$package, topic_indices)) {
    topic_indices[[pkg$package]] <- build_topic_index(pkg)
  }
  topic_indices[[pkg$package]]
}

clear_topic_index <- function(pkg = ".") {
  pkg <- as.package(pkg)
  if (exists(pkg$package, topic_indices)) {
    rm(list = pkg$package, envir = topic_indices)
  }

  invisible(TRUE)
}

build_topic_index <- function(pkg = ".") {
  pkg <- as.package(pkg)
  rds <- rd_files(pkg)

  aliases <- function(path) {
    parsed <- tools::parse_Rd(path)
    tags <- vapply(parsed, function(x) attr(x, "Rd_tag")[[1]], character(1))
    unlist(parsed[tags == "\\alias"])
  }

  invert(lapply(rds, aliases))
}

invert <- function(L) {
  if (length(L) == 0) return(L)
  t1 <- unlist(L)
  names(t1) <- rep(names(L), lapply(L, length))
  tapply(names(t1), t1, c)
}

# Generate checksums for a vector of file paths.
# @keywords internal
md5 <- function(paths) {
  unlist(lapply(paths, tools::md5sum))
}

make_cache <- function() {
  .file_cache <- character()

  make <- function(paths) {
    paths <- path.expand(paths)
    new_hash <- md5(paths)
    old_hash <- .file_cache[paths]

    changed <- is.na(old_hash) | new_hash != old_hash
    .file_cache[paths[changed]] <<- new_hash[changed]

    paths[changed]
  }

  clear <- function() {
    .file_cache <<- character()
  }

  list(make = make, clear = clear)
}
.cache <- make_cache()

# Given vector of paths, return only those paths that have changed since the
# last invocation.
# @keywords internal
changed_files <- .cache$make

# Clear file cache.
# @keywords internal
clear_cache <- .cache$clear

#' Find all dependencies of a CRAN or dev package.
#'
#' Find all the dependencies of a package and determine whether they are ahead
#' or behind CRAN. A \code{print()} method identifies mismatches (if any)
#' between local and CRAN versions of each dependent package; an
#' \code{update()} method installs outdated or missing packages from CRAN.
#'
#' @param pkg A character vector of package names. If missing, defaults to
#'   the name of the package in the current directory.
#' @param dependencies Which dependencies do you want to check?
#'   Can be a character vector (selecting from "Depends", "Imports",
#'    "LinkingTo", "Suggests", or "Enhances"), or a logical vector.
#'
#'   \code{TRUE} is shorthand for "Depends", "Imports", "LinkingTo" and
#'   "Suggests". \code{NA} is shorthand for "Depends", "Imports" and "LinkingTo"
#'   and is the default. \code{FALSE} is shorthand for no dependencies (i.e.
#'   just check this package, not its dependencies).
#' @param quiet If \code{TRUE}, suppress output.
#' @param upgrade If \code{TRUE}, also upgrade any of out date dependencies.
#' @param repos A character vector giving repositories to use.
#' @param type Type of package to \code{update}.  If "both", will switch
#'   automatically to "binary" to avoid interactive prompts during package
#'   installation.
#'
#' @param object A \code{package_deps} object.
#' @param bioconductor Install Bioconductor dependencies if the package has a
#' BiocViews field in the DESCRIPTION.
#' @param ... Additional arguments passed to \code{install_packages}.
#'
#' @return
#'
#' A \code{data.frame} with columns:
#'
#' \tabular{ll}{
#' \code{package} \tab The dependent package's name,\cr
#' \code{installed} \tab The currently installed version,\cr
#' \code{available} \tab The version available on CRAN,\cr
#' \code{diff} \tab An integer denoting whether the locally installed version
#'   of the package is newer (1), the same (0) or older (-1) than the version
#'   currently available on CRAN.\cr
#' }
#'
#' @export
#' @examples
#' \dontrun{
#' package_deps("devtools")
#' # Use update to update any out-of-date dependencies
#' update(package_deps("devtools"))
#' }
package_deps <- function(pkg, dependencies = NA, repos = getOption("repos"),
                         type = getOption("pkgType")) {
  if (identical(type, "both")) {
    type <- "binary"
  }

  if (length(repos) == 0)
    repos <- character()

  repos[repos == "@CRAN@"] <- cran_mirror()

  if (missing(pkg)) {
    pkg <- as.package(".")$package
  }

  # It is important to not extract available_packages() to a variable,
  # for the case when pkg is empty (e.g., install(dependencies = FALSE) ).
  deps <- sort_ci(find_deps(pkg, available_packages(repos, type), top_dep = dependencies))

  # Remove base packages
  inst <- installed.packages()
  base <- unname(inst[inst[, "Priority"] %in% c("base", "recommended"), "Package"])
  deps <- setdiff(deps, base)

  # get remote types
  remote <- structure(lapply(deps, package2remote, repos = repos, type = type), class = "remotes")

  inst_ver <- vapply(deps, local_sha, character(1))
  cran_ver <- vapply(remote, remote_sha, character(1))

  cran_remote <- vapply(remote, inherits, logical(1), "cran_remote")
  diff <- compare_versions(inst_ver, cran_ver, cran_remote)

  res <- structure(
    data.frame(
      package = deps,
      installed = inst_ver,
      available = cran_ver,
      diff = diff,
      stringsAsFactors = FALSE
    ),
    class = c("package_deps", "data.frame")
  )
  res$remote <- remote

  res
}

#' @export
#' @rdname package_deps
dev_package_deps <- function(pkg = ".", dependencies = NA,
                             repos = getOption("repos"),
                             type = getOption("pkgType"),
                             bioconductor = TRUE) {
  pkg <- as.package(pkg)

  repos <- c(repos, parse_additional_repositories(pkg))

  dependencies <- tolower(standardise_dep(dependencies))
  dependencies <- intersect(dependencies, names(pkg))

  parsed <- lapply(pkg[tolower(dependencies)], parse_deps)
  deps <- unlist(lapply(parsed, `[[`, "name"), use.names = FALSE)

  if (isTRUE(bioconductor) && is_bioconductor(pkg)) {
    check_bioconductor()
    bioc_repos <- BiocInstaller::biocinstallRepos()

    missing_repos <- setdiff(names(bioc_repos), names(repos))

    if (length(missing_repos) > 0)
      repos[missing_repos] <- bioc_repos[missing_repos]
  }

  res <- filter_duplicate_deps(
    package_deps(deps, repos = repos, type = type),

    # We set this cache in install() so we can run install_deps() twice without
    # having to re-query the remotes
    installing$remote_deps %||% remote_deps(pkg))

  # Only keep dependencies we actually want to use
  res[res$package %in% deps, ]
}

filter_duplicate_deps <- function(cran_deps, remote_deps, dependencies) {
  deps <- rbind(cran_deps, remote_deps)

  # Only keep the remotes that are specified in the cran_deps
  # Keep only the Non-CRAN remotes if there are duplicates as we want to install
  # the development version rather than the CRAN version. The remotes will
  # always be specified after the CRAN dependencies, so using fromLast will
  # filter out the CRAN dependencies.
  deps[!duplicated(deps$package, fromLast = TRUE), ]
}

## -2 = not installed, but available on CRAN
## -1 = installed, but out of date
##  0 = installed, most recent version
##  1 = installed, version ahead of CRAN
##  2 = package not on CRAN

UNINSTALLED <- -2L
BEHIND <- -1L
CURRENT <- 0L
AHEAD <- 1L
UNAVAILABLE <- 2L

compare_versions <- function(inst, remote, is_cran) {
  stopifnot(length(inst) == length(remote) && length(inst) == length(is_cran))

  compare_var <- function(i, c, cran) {
    if (!cran) {
      if (identical(i, c)) {
        return(CURRENT)
      } else {
        return(BEHIND)
      }
    }
    if (is.na(c)) return(UNAVAILABLE)           # not on CRAN
    if (is.na(i)) return(UNINSTALLED)           # not installed, but on CRAN

    i <- package_version(i)
    c <- package_version(c)

    if (i < c) {
      BEHIND                               # out of date
    } else if (i > c) {
      AHEAD                                # ahead of CRAN
    } else {
      CURRENT                              # most recent CRAN version
    }
  }

  vapply(seq_along(inst),
    function(i) compare_var(inst[[i]], remote[[i]], is_cran[[i]]),
    integer(1))
}

parse_one_remote <- function(x) {
  pieces <- strsplit(x, "::", fixed = TRUE)[[1]]

  if (length(pieces) == 1) {
    type <- "github"
    repo <- pieces
  } else if (length(pieces) == 2) {
    type <- pieces[1]
    repo <- pieces[2]
  } else {
    stop("Malformed remote specification '", x, "'", call. = FALSE)
  }
  fun <- tryCatch(get(paste0(tolower(type), "_remote"),
      envir = asNamespace("devtools"), mode = "function", inherits = FALSE),
    error = function(e) stop("Unknown remote type: ", type, call. = FALSE))

  fun(repo)
}

split_remotes <- function(x) {
  trim_ws(unlist(strsplit(x, ",[[:space:]]*")))
}

remote_deps <- function(pkg) {
  pkg <- as.package(pkg)

  if (!has_dev_remotes(pkg)) {
    return(NULL)
  }

  dev_packages <- split_remotes(pkg[["remotes"]])
  remote <- lapply(dev_packages, parse_one_remote)

  package <- vapply(remote, remote_package_name, character(1), USE.NAMES = FALSE)
  installed <- vapply(package, local_sha, character(1), USE.NAMES = FALSE)
  available <- vapply(remote, remote_sha, character(1), USE.NAMES = FALSE)
  diff <- installed == available
  diff <- ifelse(!is.na(diff) & diff, CURRENT, BEHIND)

  res <- structure(
    data.frame(
      package = package,
      installed = installed,
      available = available,
      diff = diff,
      stringsAsFactors = FALSE
      ),
    class = c("package_deps", "data.frame"))
  res$remote <- structure(remote, class = "remotes")

  res
}

has_dev_remotes <- function(pkg) {
  pkg <- as.package(pkg)

  !is.null(pkg[["remotes"]])
}


#' @export
print.package_deps <- function(x, show_ok = FALSE, ...) {
  class(x) <- "data.frame"

  ahead <- x$diff > CURRENT
  behind <- x$diff < CURRENT
  same_ver <- x$diff == CURRENT

  x$diff <- NULL
  x[] <- lapply(x, format)

  if (any(behind)) {
    cat("Needs update -----------------------------\n")
    print(x[behind, , drop = FALSE], row.names = FALSE, right = FALSE)
  }

  if (any(ahead)) {
    cat("Not on CRAN ----------------------------\n")
    print(x[ahead, , drop = FALSE], row.names = FALSE, right = FALSE)
  }

  if (show_ok && any(same_ver)) {
    cat("OK ---------------------------------------\n")
    print(x[same_ver, , drop = FALSE], row.names = FALSE, right = FALSE)
  }
}

#' @export
#' @rdname package_deps
#' @importFrom stats update
update.package_deps <- function(object, ..., quiet = FALSE, upgrade = TRUE) {
  non_cran <- !vapply(object$remote, inherits, logical(1), "cran_remote")
  unavailable <- object$diff == UNAVAILABLE & non_cran
  if (any(unavailable)) {
    if (upgrade) {
      install_remotes(object$remote[unavailable], ..., quiet = quiet)
    } else if (!quiet) {
      message(sprintf(ngettext(sum(unavailable),
            "Skipping %d unavailable package: %s",
            "Skipping %d unavailable packages: %s"
            ), sum(unavailable), paste(object$package[unavailable], collapse = ", ")))
    }
  }

  ahead <- object$diff == AHEAD & non_cran
  if (any(ahead)) {
    if (upgrade) {
      install_remotes(object$remote[ahead], ..., quiet = quiet)
    } else if (!quiet) {
      message(sprintf(ngettext(sum(ahead),
            "Skipping %d package ahead of CRAN: %s",
            "Skipping %d packages ahead of CRAN: %s"
            ), sum(ahead), paste(object$package[ahead], collapse = ", ")))
    }
  }

  if (upgrade) {
    behind <- object$diff < CURRENT
  } else {
    behind <- is.na(object$installed)
  }

  install_remotes(object$remote[behind], ..., quiet = quiet)
}

install_packages <- function(pkgs, repos = getOption("repos"),
                             type = getOption("pkgType"), ...,
                             dependencies = FALSE, quiet = NULL) {
  if (identical(type, "both"))
    type <- "binary"
  if (is.null(quiet))
    quiet <- !identical(type, "source")

    message(sprintf(ngettext(length(pkgs),
      "Installing %d package: %s",
      "Installing %d packages: %s"
    ), length(pkgs), paste(pkgs, collapse = ", ")))

  # if type is 'source' and on windows add Rtools to the path this assumes
  # setup_rtools() has already run and set the rtools path
  if (type == "source" && !is.null(get_rtools_path())) {
    old <- add_path(get_rtools_path(), 0)
    on.exit(set_path(old))
  }
  utils::install.packages(pkgs, repos = repos, type = type,
    dependencies = dependencies, quiet = quiet)
}

find_deps <- function(pkgs, available = available.packages(), top_dep = TRUE,
                      rec_dep = NA, include_pkgs = TRUE) {
  if (length(pkgs) == 0 || identical(top_dep, FALSE))
    return(character())

  top_dep <- standardise_dep(top_dep)
  rec_dep <- standardise_dep(rec_dep)

  top <- tools::package_dependencies(pkgs, db = available, which = top_dep)
  top_flat <- unlist(top, use.names = FALSE)

  if (length(rec_dep) != 0 && length(top_flat) > 0) {
    rec <- tools::package_dependencies(top_flat, db = available, which = rec_dep,
      recursive = TRUE)
    rec_flat <- unlist(rec, use.names = FALSE)
  } else {
    rec_flat <- character()
  }

  unique(c(if (include_pkgs) pkgs, top_flat, rec_flat))
}

standardise_dep <- function(x) {
  if (identical(x, NA)) {
    c("Depends", "Imports", "LinkingTo")
  } else if (isTRUE(x)) {
    c("Depends", "Imports", "LinkingTo", "Suggests")
  } else if (identical(x, FALSE)) {
    character(0)
  } else if (is.character(x)) {
    x
  } else {
    stop("Dependencies must be a boolean or a character vector", call. = FALSE)
  }
}

#' Update packages that are missing or out-of-date.
#'
#' Works similarly to \code{install.packages()} but doesn't install packages
#' that are already installed, and also upgrades out dated dependencies.
#'
#' @param pkgs Character vector of packages to update. IF \code{TRUE} all
#'   installed packages are updated. If \code{NULL} user is prompted to
#'   confirm update of all installed packages.
#' @inheritParams package_deps
#' @seealso \code{\link{package_deps}} to see which packages are out of date/
#'   missing.
#' @export
#' @examples
#' \dontrun{
#' update_packages("ggplot2")
#' update_packages(c("plyr", "ggplot2"))
#' }
update_packages <- function(pkgs = NULL, dependencies = NA,
                            repos = getOption("repos"),
                            type = getOption("pkgType")) {

  if (isTRUE(pkgs)) {
    pkgs <- installed.packages()[, "Package"]
  }
  else if (is.null(pkgs)) {
    if (!yesno("Are you sure you want to update all installed packages?")) {
      pkgs <- installed.packages()[, "Package"]
    } else {
      return(invisible())
    }
  }

  pkgs <- package_deps(pkgs,
    dependencies = dependencies,
    repos = repos,
    type = type)

  update(pkgs)
}

has_additional_repositories <- function(pkg) {
  pkg <- as.package(pkg)

  "additional_repositories" %in% names(pkg)
}

parse_additional_repositories <- function(pkg) {
  pkg <- as.package(pkg)

  if (has_additional_repositories(pkg)) {
    strsplit(pkg[["additional_repositories"]], "[,[:space:]]+")[[1]]
  }
}

#' @export
`[.remotes` <- function(x,i,...) {
  r <- NextMethod("[")
  mostattributes(r) <- attributes(x)
  r
}

#' Build the development version of devtools from GitHub.
#'
#' This function is especially useful for Windows users who want to upgrade
#' their version of devtools to the development version hosted on on GitHub.
#' In Windows, it's not possible to upgrade devtools while the package is loaded
#' because there is an open DLL, which in Windows can't be overwritten. This
#' function allows you to build a binary package of the development version of
#' devtools; then you can restart R (so that devtools isn't loaded) and install
#' the package.
#'
#' Mac and Linux users don't need this function; they can use
#' \code{\link{install_github}} to install devtools directly, without going
#' through the separate build-restart-install steps.
#'
#' This function requires a working development environment. On Windows, it
#' needs \url{https://cran.r-project.org/bin/windows/Rtools/}.
#'
#' @param outfile The name of the output file. If NULL (the default), it uses
#'   ./devtools.tgz (Mac and Linux), or ./devtools.zip (Windows).
#' @return a string giving the location (including file name) of the built
#'  package
#' @examples
#' \dontrun{
#' library(devtools)
#' build_github_devtools()
#'
#' #### Restart R before continuing ####
#' install.packages("./devtools.zip", repos = NULL)
#'
#' # Remove the package after installation
#' unlink("./devtools.zip")
#' }
#' @export
build_github_devtools <- function(outfile = NULL) {
  .Deprecated(msg = "`build_github_devtools()` is deprecated, you can simply use `install_github(\"hadley/devtools\")`")
  if (!has_devel()) {
    stop("This requires a working development environment.")
  }

  ext <- if (.Platform$OS.type == "windows") "zip" else "tgz"
  outfile <- paste0("./devtools.", ext)

  url <- "https://github.com/hadley/devtools/archive/master.zip"
  message("Downloading devtools from ", url)
  bundle <- file.path(tempdir(), "devtools-master.zip")

  # Download package file
  request <- httr::GET(url)
  httr::stop_for_status(request)
  writeBin(httr::content(request, "raw"), bundle)
  on.exit(unlink(bundle))

  utils::unzip(bundle, exdir = tempdir())

  # Build binary package
  pkgdir <- file.path(tempdir(), "devtools-master")
  built_pkg <- devtools::build(pkgdir, binary = TRUE)

  message("Renaming file to ", outfile)
  file.rename(built_pkg, outfile)

  invisible(outfile)
}

#' Git checks.
#'
#' This function performs Git checks checks prior to release. It is called
#' automatically by \code{\link{release}()}.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information.
#' @keywords internal
git_checks <- function(pkg = ".") {
  pkg <- as.package(pkg)
  message("Running Git checks for ", pkg$package)

  check_uncommitted(pkg)
  check_sync_status(pkg)
}

check_uncommitted <- function(pkg) {
  check_status(!git_uncommitted(pkg$path),
               "uncommitted files",
               "All files should be committed before release. Please add and commit."
  )
}

check_sync_status <- function(pkg) {
  check_status(!git_sync_status(pkg$path, check_ahead = FALSE),
               "if synched with remote branch",
               "Local branch should contain all commits of remote branch before release. Please pull."
  )
}

#' Install a package from a local file
#'
#' This function is vectorised so you can install multiple packages in
#' a single command.
#'
#' @param path path to local directory, or compressed file (tar, zip, tar.gz
#'   tar.bz2, tgz2 or tbz)
#' @inheritParams install_url
#' @export
#' @examples
#' \dontrun{
#' dir <- tempfile()
#' dir.create(dir)
#' pkg <- download.packages("testthat", dir, type = "source")
#' install_local(pkg[, 2])
#' }
install_local <- function(path, subdir = NULL, ..., quiet = FALSE) {
  remotes <- lapply(path, local_remote, subdir = subdir)
  install_remotes(remotes, ..., quiet = quiet)
}

local_remote <- function(path, subdir = NULL, branch = NULL, args = character(0)) {
  remote("local",
    path = path,
    subdir = subdir
  )
}

#' @export
remote_download.local_remote <- function(x, quiet = FALSE) {
  # Already downloaded - just need to copy to tempdir()
  bundle <- tempfile()
  dir.create(bundle)
  file.copy(x$path, bundle, recursive = TRUE)

  # file.copy() creates directory inside of bundle
  dir(bundle, full.names = TRUE)[1]
}

#' @export
remote_metadata.local_remote <- function(x, bundle = NULL, source = NULL) {
  res <- list(
    RemoteType = "local",
    RemoteUrl = x$path
  )

  res$RemoteSha <- remote_sha.local_remote(x)
  if (uses_git(x$path)) {
    res$RemoteBranch <- git_branch(path = x$path)
  }
  if (uses_github(x$path)) {
    info <- github_info(x$path)
    res$RemoteUsername <- info$username
    res$RemoteRepo <- info$repo
  }
  res
}

#' @export
remote_metadata.package <- remote_metadata.local_remote

#' @export
remote_package_name.local_remote <- function(remote, ...) {
  description_path <- file.path(remote$path, "DESCRIPTION")

  read_dcf(description_path)$Package
}

#' @export
remote_sha.local_remote <- function(remote, ...) {
  if (uses_git(remote$path)) {
    if (git_uncommitted(remote$path)) {
      return(NA_character_)
    }
    tryCatch({
      git_sha1(path = remote$path)
    }, error = function(e) NA_character_)
  } else {
    read_dcf(file.path(remote$path, "DESCRIPTION"))$Version
  }
}

#' @export
format.local_remote <- function(x, ...) {
  "local"
}

has_src <- function(pkg = ".") {
  pkg <- as.package(pkg)

  src_path <- file.path(pkg$path, "src")
  file.exists(src_path)
}

check_build_tools <- function(pkg = ".") {
  if (!has_src(pkg)) {
    return(TRUE)
  }

  # RStudio provides a dialog that can prompt the user to install the tools.
  check <- getOption("buildtools.check", NULL)
  if (!is.null(check)) {
    can_build <- check("Building R package from source")
    setup_rtools()
  } else {
    # Outside of Rstudio, check on Windows, otherwise assume they're present
    can_build <- setup_rtools()
  }

  if (!can_build) {
    stop("Could not find build tools necessary to build ", pkg$package,
      call. = FALSE)
  }
}

has_latex <- function(verbose = FALSE) {
  has <- nzchar(Sys.which("pdflatex"))
  if (!has && verbose) {
    message("pdflatex not found! Not building PDF manual or vignettes.\n",
            "If you are planning to release this package, please run a check with ",
            "manual and vignettes beforehand.\n")
  }
  has
}

#' Load a compiled DLL
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @keywords programming
#' @name load_dll
#' @usage load_dll(pkg = ".")
#' @export
onload_assign("load_dll",
  make_function(alist(pkg = "."),
    bquote({
      pkg <- as.package(pkg)
      env <- ns_env(pkg)
      nsInfo <- parse_ns_file(pkg)

      dlls <- list()
      dynLibs <- nsInfo$dynlibs

      ## The code below taken directly from base::loadNamespace
      ## https://github.com/wch/r-source/blob/tags/R-3-3-0/src/library/base/R/namespace.R#L466-L485
      ## except for the call to library.dynam2, which is a special version of
      ## library.dynam
      .(for_loop)
      addNamespaceDynLibs(env, nsInfo$dynlibs)

      invisible(dlls)
    },
    list(for_loop =
      modify_lang(
        f = function(x)
          if (comp_lang(x, quote(library.dynam()), 1)) {
            quote(library.dynam2(pkg, lib))
          } else {
            x
          },

          extract_lang(body(loadNamespace),
            comp_lang, y = quote(for (i in seq_along(dynLibs)) NULL), idx = 1:3)
      )))))

# Return a list of currently loaded DLLs from the package
loaded_dlls <- function(pkg = ".") {
  pkg <- as.package(pkg)
  libs <- .dynLibs()
  matchidx <- vapply(libs, "[[", character(1), "name") == pkg$package
  libs[matchidx]
}

# This is a replacement for base::library.dynam, with a slightly different
# call interface. The original requires that the name of the package is the
# same as the directory name, which isn't always the case when loading with
# devtools. This version allows them to be different, and also searches in
# the src/ directory for the DLLs, instead of the libs/$R_ARCH/ directory.
library.dynam2 <- function(pkg = ".", lib = "") {
  pkg <- as.package(pkg)

  dllname <- paste(lib, .Platform$dynlib.ext, sep = "")
  dllfile <- file.path(pkg$path, "src", dllname)

  if (!file.exists(dllfile))
    return(invisible())

  # # The loading and registering of the dll is similar to how it's done
  # # in library.dynam.
  dllinfo <- dyn.load(dllfile)

  # Register dll info so it can be unloaded with library.dynam.unload
  .dynLibs(c(.dynLibs(), list(dllinfo)))

  return(dllinfo)
}


# This is taken directly from base::loadNamespace()
# https://github.com/wch/r-source/blob/tags/R-3-3-0/src/library/base/R/namespace.R#L270-L273
onload_assign("addNamespaceDynLibs",
  eval(extract_lang(body(loadNamespace),
      comp_lang, y = quote(addNamespaceDynLibs <- NULL), idx = 1:2)[[3]]))

# This is taken directly from base::loadNamespace
# https://github.com/wch/r-source/blob/tags/R-3-3-0/src/library/base/R/namespace.R#L287-L308
# The only change is the line used get the package name
onload_assign("assignNativeRoutines", {
  f <- eval(
    extract_lang(body(loadNamespace),
      comp_lang, y = quote(assignNativeRoutines <- NULL), idx = 1:2)[[3]])
  body(f) <- as.call(append(after = 1,
      as.list(body(f)),
      quote(package <- methods::getPackageName(env))))
  f
})

#' Execute all \pkg{test_that} tests in a package.
#'
#' Tests are assumed to be located in either the \code{inst/tests/} or
#' \code{tests/testthat} directory (the latter is recommended).
#' See \code{\link[testthat]{test_dir}} for the naming convention of test
#' scripts within one of those directories and
#' \code{\link[testthat]{test_check}} for the folder structure conventions.
#'
#' If no testing infrastructure is present
#' (detected by the \code{uses_testthat} function), you'll be asked if you want
#' devtools to create it for you (in interactive sessions only). See
#' \code{\link{use_test}} for more details.
#'
#' @param pkg package description, can be path or package name. See
#'   \code{\link{as.package}} for more information
#' @param ... additional arguments passed to \code{\link[testthat]{test_dir}}
#' @inheritParams testthat::test_dir
#' @inheritParams run_examples
#' @export
test <- function(pkg = ".", filter = NULL, ...) {
  check_suggested("testthat")
  pkg <- as.package(pkg)

  if (!uses_testthat(pkg) && interactive()) {
    message("No testing infrastructure found. Create it?")
    if (menu(c("Yes", "No")) == 1) {
      use_testthat(pkg)
    }
    return(invisible())
  }

  test_path <- find_test_dir(pkg$path)
  test_files <- dir(test_path, "^test.*\\.[rR]$")
  if (length(test_files) == 0) {
    message("No tests: no files in ", test_path, " match '^test.*\\.[rR]$'")
    return(invisible())
  }

  # Need to attach testthat so that (e.g.) context() is available
  # Update package dependency to avoid explicit require() call (#798)
  if (pkg$package != "testthat") {
    pkg$depends <- paste0("testthat, ", pkg$depends)
    if (grepl("^testthat, *$", pkg$depends))
      pkg$depends <- "testthat"
  }

  # Run tests in a child of the namespace environment, like
  # testthat::test_package
  message("Loading ", pkg$package)
  ns_env <- load_all(pkg, quiet = TRUE)$env

  message("Testing ", pkg$package)
  Sys.sleep(0.05); utils::flush.console() # Avoid misordered output in RStudio

  env <- new.env(parent = ns_env)

  testthat_args <- list(test_path, filter = filter, env = env, ... = ...)
  if (packageVersion("testthat") > "1.0.2") {
    testthat_args <- c(testthat_args, load_helpers = FALSE, encoding = pkg$encoding %||% "unknown")
  }

  withr::with_envvar(r_env_vars(),
    do.call(testthat::test_dir, testthat_args))
}

find_test_dir <- function(path) {
  testthat <- file.path(path, "tests", "testthat")
  if (dir.exists(testthat)) return(testthat)

  inst <- file.path(path, "inst", "tests")
  if (dir.exists(inst)) return(inst)

  stop("No testthat directories found in ", path, call. = FALSE)
}

#' Return the path to one of the packages in the devtools test dir
#'
#' Devtools comes with some simple packages for testing. This function
#' returns the path to them.
#'
#' @param package Name of the test package.
#' @keywords internal
#' @examples
#' if (has_tests()) {
#' devtest("testData")
#' }
#' @export
devtest <- function(package) {
  stopifnot(has_tests())

  path <- system.file(package = "devtools", "tests", "testthat", package)
  if (path == "") stop(package, " not found", call. = FALSE)

  path
}

#' @inheritParams test
#' @rdname test
#' @export
uses_testthat <- function(pkg = ".") {
  pkg <- as.package(pkg)

  paths <- c(
    file.path(pkg$path, "inst", "tests"),
    file.path(pkg$path, "tests", "testthat")
  )

  any(dir.exists(paths))
}

# Supress R CMD check note
#' @importFrom memoise memoise
NULL

.rstudio_release <- function() {
  url <- "http://s3.amazonaws.com/rstudio-server/current.ver"
  numeric_version(readLines(url, warn = FALSE))
}

rstudio_release <- memoise::memoise(.rstudio_release)

.r_release <- function() {
  check_suggested("rversions")
  R_system_version(rversions::r_release()$version)
}

r_release <- memoise::memoise(.r_release)


#' Diagnose potential devtools issues
#'
#' This checks to make sure you're using the latest release of R,
#' the released version of RStudio (if you're using it as your gui),
#' and the latest version of devtools and its dependencies.
#'
#' @family doctors
#' @export
#' @examples
#' \dontrun{
#' dr_devtools()
#' }
dr_devtools <- function() {
  msg <- character()

  if (getRversion() < r_release()) {
    msg[["R"]] <- paste0(
      "* R is out of date (", getRversion(), " vs ", r_release(), ")"
    )
  }

  deps <- package_deps("devtools", dependencies = NA)
  old <- deps$diff < 0
  if (any(old)) {
    msg[["devtools"]] <- paste0(
      "* Devtools or dependencies out of date: \n",
      paste(deps$package[old], collapse = ", ")
    )
  }

  if (rstudioapi::isAvailable()) {
    rel <- rstudio_release()
    cur <- rstudioapi::getVersion()

    if (cur < rel) {
      msg[["rstudio"]] <- paste0(
        "* RStudio is out of date (", cur, " vs ", rel, ")"
      )
    }
  }

  doctor("devtools", msg)
}

#' Diagnose potential GitHub issues
#'
#' @param path Path to repository to check. Defaults to current working
#'   directory
#' @family doctors
#' @export
#' @examples
#' \donttest{
#' dr_github()
#' }
dr_github <- function(path = ".") {
  if (!uses_git(path)) {
    return(doctor("github", "Path is not a git repository"))
  }
  if (!uses_github(path)) {
    return(doctor("github", "Path is not a GitHub repository"))
  }

  msg <- character()
  r <- git2r::repository(path, discover = TRUE)

  config <- git2r::config(r)
  config_names <- names(modifyList(config$global, config$local))

  if (!uses_github(path))
    msg[["github"]] <- " * cannot detect that this repo is connected to GitHub"
  if (!("user.name" %in% config_names))
    msg[["name"]] <- "* user.name config option not set"
  if (!("user.email" %in% config_names))
    msg[["user"]] <- "* user.email config option not set"

  if (!file.exists("~/.ssh/id_rsa"))
    msg[["ssh"]] <- "* SSH private key not found"

  if (identical(Sys.getenv("GITHUB_PAT"), ""))
    msg[["PAT"]] <- paste("* GITHUB_PAT environment variable not set",
      "(this is not necessary unless you want to install private repos",
      "or connect local repos to GitHub)")

  desc_path <- file.path(path, "DESCRIPTION")
  desc <- read_dcf(desc_path)
  field_empty <- function(d, f) is.null(d[[f]]) || identical(d[[f]], "")
  field_no_re <- function(d, f, re) !grepl(re, d[[f]])

  re <- "https://github.com/(.*?)/(.*)"
  if (field_empty(desc, "URL")) {
    msg[["URL_empty"]] <-"* empty URL field in DESCRIPTION"
  } else if (field_no_re(desc, "URL", re)) {
    msg[["URL"]] <-"* no GitHub repo link in URL field in DESCRIPTION"
  }

  re <- paste0(re, "/issues")
  if (field_empty(desc, "BugReports")) {
    msg[["BugReports_empty"]] <-"* empty BugReports field in DESCRIPTION"
  } else if (field_no_re(desc, "BugReports", re)) {
    msg[["BugReports"]] <-"* no GitHub Issues link in URL field in DESCRIPTION"
  }

  doctor("github", msg)
}

# Doctor class ------------------------------------------------------------

doctor <- function(name, messages) {
  structure(
    length(messages) == 0,
    doctor = paste0("DR_", toupper(name)),
    messages = messages,
    class = "doctor"
  )
}

#' @export
print.doctor <- function(x, ...) {
  if (x) {
    message(attr(x, "doctor"), " SAYS YOU LOOK HEALTHY")
    return()
  }

  warning(attr(x, "doctor"), " FOUND PROBLEMS", call. = FALSE, immediate. = TRUE)
  messages <- strwrap(attr(x, "messages"), exdent = 2)
  message(paste(messages, collapse = "\n"))
}

#' Build and check a package, cleaning up automatically on success.
#'
#' \code{check} automatically builds and checks a source package, using all
#' known best practices. \code{check_built} checks an already built package.
#'
#' Passing \code{R CMD check} is essential if you want to submit your package
#' to CRAN: you must not have any ERRORs or WARNINGs, and you want to ensure
#' that there are as few NOTEs as possible.  If you are not submitting to CRAN,
#' at least ensure that there are no ERRORs or WARNINGs: these typically
#' represent serious problems.
#'
#' \code{check} automatically builds a package before calling \code{check_built}
#' as this is the recommended way to check packages.  Note that this process
#' runs in an independent realisation of R, so nothing in your current
#' workspace will affect the process.
#'
#' @section Environment variables:
#'
#' Devtools does its best to set up an environment that combines best practices
#' with how check works on CRAN. This includes:
#'
#' \itemize{
#'
#'  \item The standard environment variables set by devtools:
#'    \code{\link{r_env_vars}}. Of particular note for package tests is the
#'    \code{NOT_CRAN} env var which lets you know that your tests are not
#'    running on cran, and hence can take a reasonable amount of time.
#'
#'  \item Debugging flags for the compiler, set by
#'    \code{\link{compiler_flags}(FALSE)}.
#'
#'  \item If \code{aspell} is found \code{_R_CHECK_CRAN_INCOMING_USE_ASPELL_}
#'   is set to \code{TRUE}. If no spell checker is installed, a warning is
#'   issued.)
#'
#'  \item env vars set by arguments \code{check_version} and
#'    \code{force_suggests}
#' }
#'
#' @return An object containing errors, warnings, and notes.
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @param document if \code{TRUE} (the default), will update and check
#'   documentation before running formal check.
#' @param build_args Additional arguments passed to \code{R CMD build}
#' @param ... Additional arguments passed on to \code{\link{build}()}.
#' @param cleanup Deprecated.
#' @seealso \code{\link{release}} if you want to send the checked package to
#'   CRAN.
#' @export
check <- function(pkg = ".",
                  document = TRUE,
                  build_args = NULL,
                  ...,
                  manual = FALSE,
                  cran = TRUE,
                  check_version = FALSE,
                  force_suggests = FALSE,
                  run_dont_test = FALSE,
                  args = NULL,
                  env_vars = NULL,
                  quiet = FALSE,
                  check_dir = tempdir(),
                  cleanup = TRUE) {
  pkg <- as.package(pkg)
  if (!missing(cleanup)) {
    warning("`cleanup` is deprecated", call. = FALSE)
  }

  if (document) {
    document(pkg)
  }

  if (!quiet) {
    show_env_vars(compiler_flags(FALSE))
    rule("Building ", pkg$package)
  }

  withr::with_envvar(compiler_flags(FALSE), action = "prefix", {
    built_path <- build(pkg, tempdir(), quiet = quiet, args = build_args,
                        manual = manual, ...)
    on.exit(unlink(built_path), add = TRUE)
  })

  check_built(
    built_path,
    cran = cran,
    check_version = check_version,
    force_suggests = force_suggests,
    run_dont_test = run_dont_test,
    manual = manual,
    args = args,
    env_vars = env_vars,
    quiet = quiet,
    check_dir = check_dir
  )
}

#' @export
#' @rdname check
#' @param path Path to built package.
#' @param cran if \code{TRUE} (the default), check using the same settings as
#'   CRAN uses.
#' @param check_version Sets \code{_R_CHECK_CRAN_INCOMING_} env var.
#'   If \code{TRUE}, performs a number of checked related
#'   to version numbers of packages on CRAN.
#' @param force_suggests Sets \code{_R_CHECK_FORCE_SUGGESTS_}. If
#'   \code{FALSE} (the default), check will proceed even if all suggested
#'   packages aren't found.
#' @param run_dont_test Sets \code{--run-donttest} so that tests surrounded in
#'   \code{\\dontest\{\}} are also tested. This is important for CRAN
#'   submission.
#' @param manual If \code{FALSE}, don't build and check manual
#'   (\code{--no-manual}).
#' @param args Additional arguments passed to \code{R CMD check}
#' @param env_vars Environment variables set during \code{R CMD check}
#' @param check_dir the directory in which the package is checked
#' @param quiet if \code{TRUE} suppresses output from this function.
check_built <- function(path = NULL, cran = TRUE,
                        check_version = FALSE, force_suggests = FALSE,
                        run_dont_test = FALSE, manual = FALSE, args = NULL,
                        env_vars = NULL,
                        check_dir = tempdir(), quiet = FALSE) {

  pkgname <- gsub("_.*?$", "", basename(path))

  args <- c("--timings", args)
  if (cran) {
    args <- c("--as-cran", args)
  }
  if (run_dont_test) {
    args <- c("--run-donttest", args)
  }

  if (manual && !has_latex(verbose = TRUE)) {
    manual <- FALSE
  }

  if (!manual) {
    args <- c(args, "--no-manual")
  }

  env_vars <- check_env_vars(cran, check_version, force_suggests, env_vars)
  if (!quiet) {
    show_env_vars(env_vars)
    rule("Checking ", pkgname)
  }

  R(c(paste("CMD check", shQuote(path)), args),
    path = check_dir,
    env_vars = env_vars,
    quiet = quiet,
    throw = FALSE
  )

  package_path <- file.path(
    normalizePath(check_dir),
    paste(pkgname, ".Rcheck", sep = "")
  )
  if (!file.exists(package_path)) {
    stop("Check failed: '", package_path, "' doesn't exist", call. = FALSE)
  }

  # Record successful completion
  writeLines("OK", file.path(package_path, "COMPLETE"))

  log_path <- file.path(package_path, "00check.log")
  parse_check_results(log_path)
}

check_env_vars <- function(cran = FALSE, check_version = FALSE,
                           force_suggests = TRUE, env_vars) {
  c(
    aspell_env_var(),
    "_R_CHECK_CRAN_INCOMING_" = as.character(check_version),
    "_R_CHECK_FORCE_SUGGESTS_" = as.character(force_suggests),
    env_vars
  )
}

aspell_env_var <- function() {
  tryCatch({
    utils::aspell(NULL)
    c("_R_CHECK_CRAN_INCOMING_USE_ASPELL_" = "TRUE")
  }, error = function(e) character())
}

show_env_vars <- function(env_vars) {
  rule("Setting env vars")
  message(paste0(format(names(env_vars)), ": ", unname(env_vars), collapse = "\n"))
}

#' Install a remote package.
#'
#' This:
#' \enumerate{
#'   \item checks if source bundle is different from installed version
#'   \item checks if log of a past installation failure exists
#'   \item downloads source bundle
#'   \item decompresses & checks that it's a package
#'   \item adds metadata to DESCRIPTION
#'   \item calls install
#' }
#' @noRd
install_remote <- function(remote, ..., force = FALSE, quiet = FALSE,
                           out_dir = NULL, skip_if_log_exists = FALSE) {
  stopifnot(is.remote(remote))

  remote_sha <- remote_sha(remote)
  package_name <- remote_package_name(remote)
  local_sha <- local_sha(package_name)

  if (!isTRUE(force) &&
    !different_sha(remote_sha = remote_sha, local_sha = local_sha)) {

    if (!quiet) {
      message(
        "Skipping install of '", package_name, "' from a ", sub("_remote", "", class(remote)[1L]), " remote,",
        " the SHA1 (", substr(remote_sha, 1L, 8L), ") has not changed since last install.\n",
        "  Use `force = TRUE` to force installation")
    }
    return(invisible(FALSE))
  }

  if (!is.null(out_dir)) {
    out_file <- file.path(out_dir, paste0(package_name, ".out"))
    if (skip_if_log_exists && file.exists(out_file)) {
      message("Skipping ", package_name, ", installation failed before, see log in ", out_file)
      return(invisible(FALSE))
    }
  }

  if (is_windows && inherits(remote, "cran_remote")) {
    install_packages(package_name, repos = remote$repos, type = remote$pkg_type, dependencies = NA, ..., quiet = quiet,
                     out_dir = out_dir, skip_if_log_exists = skip_if_log_exists)
    return(invisible(TRUE))
  }

  bundle <- remote_download(remote, quiet = quiet)
  on.exit(unlink(bundle), add = TRUE)

  source <- source_pkg(bundle, subdir = remote$subdir)
  on.exit(unlink(source, recursive = TRUE), add = TRUE)

  metadata <- remote_metadata(remote, bundle, source)

  install(source, ..., quiet = quiet, metadata = metadata,
          out_dir = out_dir, skip_if_log_exists = skip_if_log_exists)
}

try_install_remote <- function(..., quiet) {
  tryCatch(
    install_remote(..., quiet = quiet),
    error = function(e) {
      if (!quiet) {
        message("Installation failed: ", conditionMessage(e))
      }
      FALSE
    }
  )
}

install_remotes <- function(remotes, ...) {
  invisible(vapply(remotes, try_install_remote, ..., FUN.VALUE = logical(1)))
}

# Add metadata
add_metadata <- function(pkg_path, meta) {
  # During installation, the DESCRIPTION file is read and an package.rds file
  # created with most of the information from the DESCRIPTION file. Functions
  # that read package metadata may use either the DESCRIPTION file or the
  # package.rds file, therefore we attempt to modify both of them, and return an
  # error if neither one exists.
  source_desc <- file.path(pkg_path, "DESCRIPTION")
  binary_desc <- file.path(pkg_path, "Meta", "package.rds")
  if (file.exists(source_desc)) {
    desc <- read_dcf(source_desc)

    desc <- modifyList(desc, meta)

    write_dcf(source_desc, desc)
  }

  if (file.exists(binary_desc)) {
    pkg_desc <- base::readRDS(binary_desc)
    desc <- as.list(pkg_desc$DESCRIPTION)
    desc <- modifyList(desc, meta)
    pkg_desc$DESCRIPTION <- stats::setNames(as.character(desc), names(desc))
    base::saveRDS(pkg_desc, binary_desc)
  }

  if (!file.exists(source_desc) && !file.exists(binary_desc)) {
    stop("No DESCRIPTION found!", call. = FALSE)
  }

}

# Modify the MD5 file - remove the line for DESCRIPTION
clear_description_md5 <- function(pkg_path) {
  path <- file.path(pkg_path, "MD5")

  if (file.exists(path)) {
    text <- readLines(path)
    text <- text[!grepl(".*\\*DESCRIPTION$", text)]

    writeLines(text, path)
  }
}

remote <- function(type, ...) {
  structure(list(...), class = c(paste0(type, "_remote"), "remote"))
}
is.remote <- function(x) inherits(x, "remote")

different_sha <- function(remote_sha = NULL,
                          local_sha = NULL) {
  if (is.null(remote_sha)) {
    remote_sha <- remote_sha(remote)
  }

  if (is.null(local_sha)) {
    local_sha <- local_sha(remote_package_name(remote))
  }

  same <- remote_sha == local_sha
  same <- isTRUE(same) && !is.na(same)
  !same
}

local_sha <- function(name) {
  if (!is_installed(name)) {
    return(NA_character_)
  }
  package2remote(name)$sha %||% NA_character_
}

remote_download <- function(x, quiet = FALSE) UseMethod("remote_download")
remote_metadata <- function(x, bundle = NULL, source = NULL) UseMethod("remote_metadata")
remote_package_name <- function(remote, ...) UseMethod("remote_package_name")
remote_sha <- function(remote, ...) UseMethod("remote_sha")

package2remote <- function(name, repos = getOption("repos"), type = getOption("pkgType")) {

  x <- tryCatch(packageDescription(name, lib.loc = .libPaths()), error = function(e) NA, warning = function(e) NA)

  # will be NA if not installed
  if (identical(x, NA)) {
    return(remote("cran",
        name = name,
        repos = repos,
        pkg_type = type,
        sha = NA_character_))
  }

  if (is.null(x$RemoteType)) {

    # Packages installed with install.packages() or locally without devtools
    return(remote("cran",
        name = x$Package,
        repos = repos,
        pkg_type = type,
        sha = x$Version))
  }

  switch(x$RemoteType,
    github = remote("github",
      host = x$RemoteHost,
      repo = x$RemoteRepo,
      subdir = x$RemoteSubdir,
      username = x$RemoteUsername,
      ref = x$RemoteRef,
      sha = x$RemoteSha),
    git = remote("git",
      url = x$RemoteUrl,
      ref = x$RemoteRef,
      sha = x$RemoteSha,
      subdir = x$RemoteSubdir),
    bitbucket = remote("bitbucket",
      host = x$RemoteHost,
      repo = x$RemoteRepo,
      username = x$RemoteUsername,
      ref = x$RemoteRef,
      sha = x$RemoteSha,
      subdir = x$RemoteSubdir),
    svn = remote("svn",
      url = x$RemoteUrl,
      svn_subdir = x$RemoteSvnSubdir,
      branch = x$RemoteBranch,
      sha = x$RemoteRevision,
      args = x$RemoteArgs),
    local = remote("local",
      path = x$RemoteUrl,
      branch = x$RemoteBranch,
      subdir = x$RemoteSubdir,
      sha = x$RemoteSha %||% x$Version,
      username = x$RemoteUsername,
      repo = x$RemoteRepo),
    url = remote("url",
      url = x$RemoteUrl,
      subdir = x$RemoteSubdir,
      config = x$RemoteConfig),
    bioc = remote("bioc",
      repo = x$RemoteRepo,
      mirror = x$RemoteMirror,
      release = x$RemoteRelease,
      username = x$RemoteUsername,
      password = x$RemotePassword,
      revision = x$RemoteRevision,
      sha = x$RemoteSha),

    # packages installed with install_cran
    cran = remote("cran",
      name = x$Package,
      repos = eval(parse(text = x$RemoteRepos)),
      pkg_type = x$RemotePkgType,
      sha = x$RemoteSha))
}

#' @export
format.remotes <- function(x, ...) {
  vapply(x, format, character(1))
}

context("remote_deps")
test_that("remote_deps returns NULL if no remotes specified", {
  expect_equal(remote_deps("testTest"), NULL)
})


test_that("remote_deps returns works with implicit types", {
  with_mock(`devtools::package2remote` = function(...) NULL, {
    expect_equal(parse_one_remote("hadley/testthat"),
      github_remote("hadley/testthat"))

    expect_equal(parse_one_remote("klutometis/roxygen"),
      github_remote("klutometis/roxygen"))
  })

  expect_equal(split_remotes("hadley/testthat,klutometis/roxygen"),
      c("hadley/testthat", "klutometis/roxygen"))

  expect_equal(split_remotes("hadley/testthat,\n  klutometis/roxygen"),
    c("hadley/testthat", "klutometis/roxygen"))

  expect_equal(split_remotes("hadley/testthat,\n\t klutometis/roxygen"),
    c("hadley/testthat", "klutometis/roxygen"))
})

test_that("dev_remote_type errors", {
  expect_error(parse_one_remote(""),
    "Malformed remote specification ''")

  expect_error(parse_one_remote("git::testthat::blah"),
    "Malformed remote specification 'git::testthat::blah'")
  expect_error(parse_one_remote("hadley::testthat"),
    "Unknown remote type: hadley")
  expect_error(parse_one_remote("SVN2::testthat"),
    "Unknown remote type: SVN2")
})

test_that("dev_remote_type works with explicit types", {
  with_mock(`devtools::package2remote` = function(...) NULL, {
    expect_equal(parse_one_remote("github::hadley/testthat"),
      github_remote("hadley/testthat"))
  })

  expect_equal(split_remotes("github::hadley/testthat,klutometis/roxygen"),
    c("github::hadley/testthat", "klutometis/roxygen"))

  expect_equal(split_remotes("hadley/testthat,github::klutometis/roxygen"),
    c("hadley/testthat", "github::klutometis/roxygen"))

  expect_equal(split_remotes("github::hadley/testthat,github::klutometis/roxygen"),
    c("github::hadley/testthat", "github::klutometis/roxygen"))

  expect_equal(split_remotes("bioc::user:password@release/Biobase#12345,github::klutometis/roxygen"),
    c("bioc::user:password@release/Biobase#12345", "github::klutometis/roxygen"))
})

test_that("different_sha returns TRUE if remote or local sha is NA not found", {
  expect_true(different_sha(remote_sha = NA, local_sha = "4a2ea2"))
  expect_true(different_sha(remote_sha = "4a2ea2", local_sha = NA))
  expect_true(different_sha(remote_sha = NA, local_sha = NA))
})
test_that("different_sha returns TRUE if remote_sha and local_sha are different", {
  expect_true(different_sha(remote_sha = "5b3fb3", local_sha = "4a2ea2"))
})
test_that("different_sha returns FALSE if remote_sha and local_sha are the same", {
  expect_false(different_sha(remote_sha = "4a2ea2", local_sha = "4a2ea2"))
})
test_that("local_sha returns NA if package is not installed", {
  expect_equal(local_sha("tsrtarst"), NA_character_)
})
test_that("remote_sha.github_remote returns NA if remote doesn't exist", {
  expect_equal(remote_sha(github_remote("arst/arst")), NA_character_)
})
test_that("remote_sha.github_remote returns expected value if remote does exist", {
  expect_equal(remote_sha(github_remote("hadley/devtools@v1.8.0")), "ad9aac7b9a522354e1ff363a86f389e32cec181b")
})

test_that("package2remotes looks for the DESCRIPTION in .libPaths", {
   expect_equal(package2remote("testTest")$sha, NA_character_)
   withr::with_temp_libpaths({
     expect_equal(package2remote("testTest")$sha, NA_character_)
     install("testTest", quiet = TRUE)
     expect_equal(package2remote("testTest")$sha, "0.1")

     # Load the namespace, as packageDescription looks in loaded namespaces
     # first.
     loadNamespace("testTest")
    })
   expect_equal(package2remote("testTest")$sha, NA_character_)
})

#' Add useful infrastructure to a package.
#'
#' @param pkg package description, can be path or package name. See
#'   \code{\link{as.package}} for more information.
#' @name infrastructure
#' @family infrastructure
NULL

#' @section \code{use_testthat}:
#' Add testing infrastructure to a package that does not already have it.
#' This will create \file{tests/testthat.R}, \file{tests/testthat/} and
#' add \pkg{testthat} to the suggested packages. This is called
#' automatically from \code{\link{test}} if needed.
#' @rdname infrastructure
#' @export
use_testthat <- function(pkg = ".") {
  pkg <- as.package(pkg)

  check_suggested("testthat")
  if (uses_testthat(pkg = pkg)) {
    message("* testthat is already initialized")
    return(invisible(TRUE))
  }

  message("* Adding testthat to Suggests")
  add_desc_package(pkg, "Suggests", "testthat")

  use_directory("tests/testthat", pkg = pkg)
  use_template(
    "testthat.R",
    "tests/testthat.R",
    data = list(name = pkg$package),
    pkg = pkg
  )

  invisible(TRUE)
}

#' @section \code{use_test}:
#' Add a test file, also add testing infrastructure if necessary.
#' This will create \file{tests/testthat/test-<name>.R} with a user-specified
#' name for the test.  Will fail if the file exists.
#' @rdname infrastructure
#' @export
use_test <- function(name, pkg = ".") {
  pkg <- as.package(pkg)

  check_suggested("testthat")
  if (!uses_testthat(pkg = pkg)) {
    use_testthat(pkg = pkg)
  }

  use_template("test-example.R",
    sprintf("tests/testthat/test-%s.R", name),
    data = list(test_name = name),
    open = TRUE,
    pkg = pkg
  )

  invisible(TRUE)
}

#' @export
#' @rdname infrastructure
use_rstudio <- function(pkg = ".") {
  pkg <- as.package(pkg)

  use_template(
    "template.Rproj",
    paste0(pkg$package, ".Rproj"),
    pkg = pkg
  )

  use_git_ignore(c(".Rproj.user", ".Rhistory", ".RData"), pkg = pkg)
  use_build_ignore(c("^.*\\.Rproj$", "^\\.Rproj\\.user$"), escape = FALSE, pkg = pkg)

  invisible(TRUE)
}

#' @section \code{use_vignette}:
#' Adds needed packages to \code{DESCRIPTION}, and creates draft vignette
#' in \code{vignettes/}. It adds \code{inst/doc} to \code{.gitignore}
#' so you don't accidentally check in the built vignettes.
#' @param name File name to use for new vignette. Should consist only of
#'   numbers, letters, _ and -. I recommend using lower case.
#' @export
#' @rdname infrastructure
use_vignette <- function(name, pkg = ".") {
  pkg <- as.package(pkg)
  check_suggested("rmarkdown")

  add_desc_package(pkg, "Suggests", "knitr")
  add_desc_package(pkg, "Suggests", "rmarkdown")
  add_desc_package(pkg, "VignetteBuilder", "knitr")

  use_directory("vignettes", pkg = pkg)
  use_git_ignore("inst/doc", pkg = pkg)

  path <- file.path(pkg$path, "vignettes", paste0(name, ".Rmd"))
  rmarkdown::draft(path, "html_vignette", "rmarkdown",
    create_dir = FALSE, edit = FALSE)

  open_in_rstudio(path)
}

#' @section \code{use_rcpp}:
#' Creates \code{src/} and adds needed packages to \code{DESCRIPTION}.
#' @export
#' @rdname infrastructure
use_rcpp <- function(pkg = ".") {
  pkg <- as.package(pkg)
  check_suggested("Rcpp")

  message("Adding Rcpp to LinkingTo and Imports")
  add_desc_package(pkg, "LinkingTo", "Rcpp")
  add_desc_package(pkg, "Imports", "Rcpp")

  use_directory("src/", pkg = pkg)

  message("* Ignoring generated binary files.")
  ignore_path <- file.path(pkg$path, "src", ".gitignore")
  union_write(ignore_path, c("*.o", "*.so", "*.dll"))

  message(
    "Next, include the following roxygen tags somewhere in your package:\n\n",
    "#' @useDynLib ", pkg$package, "\n",
    "#' @importFrom Rcpp sourceCpp\n",
    "NULL\n\n",
    "Then run document()"
  )
}

#' @rdname infrastructure
#' @section \code{use_travis}:
#' Add basic travis template to a package. Also adds \code{.travis.yml} to
#' \code{.Rbuildignore} so it isn't included in the built package.
#' @param browse open a browser window to enable Travis builds for the package
#' automatically.
#' @export
#' @aliases add_travis
use_travis <- function(pkg = ".", browse = interactive()) {
  pkg <- as.package(pkg)

  use_template("travis.yml", ".travis.yml", ignore = TRUE, pkg = pkg)

  gh <- github_info(pkg$path)
  travis_url <- file.path("https://travis-ci.org", gh$fullname)

  message("Next: \n",
    " * Add a travis shield to your README.md:\n",
    "[![Travis-CI Build Status]",
       "(https://travis-ci.org/", gh$fullname, ".svg?branch=master)]",
       "(https://travis-ci.org/", gh$fullname, ")\n",
    " * Turn on travis for your repo at ", travis_url, "\n"
  )
  if (browse) {
    utils::browseURL(travis_url)
  }

  invisible(TRUE)
}

#' @rdname infrastructure
#' @param type CI tool to use. Currently supports codecov and coverall.
#' @section \code{use_coverage}:
#' Add test code coverage to basic travis template to a package.
#' @export
use_coverage <- function(pkg = ".", type = c("codecov", "coveralls")) {
  pkg <- as.package(pkg)
  check_suggested("covr")

  path <- file.path(pkg$path, ".travis.yml")
  if (!file.exists(path)) {
    use_travis()
  }

  message("* Adding covr to Suggests")
  add_desc_package(pkg, "Suggests", "covr")

  gh <- github_info(pkg$path)
  type <- match.arg(type)

  message("Next:")
  switch(type,
    codecov = {
      use_template("codecov.yml", "codecov.yml", ignore = TRUE, pkg = pkg)
      message("* Add to `README.md`: \n",
        "[![Coverage Status]",
        "(https://img.shields.io/codecov/c/github/", gh$fullname, "/master.svg)]",
        "(https://codecov.io/github/", gh$fullname, "?branch=master)"
      )
      message("* Add to `.travis.yml`:\n",
        "after_success:\n",
        "  - Rscript -e 'covr::codecov()'"
      )
    },

    coveralls = {
      message("* Turn on coveralls for this repo at https://coveralls.io/repos/new")
      message("* Add to `README.md`: \n",
        "[![Coverage Status]",
        "(https://img.shields.io/coveralls/", gh$fullname, ".svg)]",
        "(https://coveralls.io/r/", gh$fullname, "?branch=master)"
      )
      message("* Add to `.travis.yml`:\n",
        "after_success:\n",
        "  - Rscript -e 'covr::coveralls()'"
      )
    })

  invisible(TRUE)
}

#' @rdname infrastructure
#' @section \code{use_appveyor}:
#' Add basic AppVeyor template to a package. Also adds \code{appveyor.yml} to
#' \code{.Rbuildignore} so it isn't included in the built package.
#' @export
use_appveyor <- function(pkg = ".") {
  pkg <- as.package(pkg)

  use_template("appveyor.yml", ignore = TRUE, pkg = pkg)

  gh <- github_info(pkg$path)
  message("Next: \n",
          " * Turn on AppVeyor for this repo at https://ci.appveyor.com/projects\n",
          " * Add an AppVeyor shield to your README.md:\n",
          "[![AppVeyor Build Status]",
          "(https://ci.appveyor.com/api/projects/status/github/", gh$username, "/", gh$repo, "?branch=master&svg=true)]",
          "(https://ci.appveyor.com/project/", gh$username, "/", gh$repo, ")"
  )

  invisible(TRUE)
}

#' @rdname infrastructure
#' @section \code{use_package_doc}:
#' Adds a roxygen template for package documentation
#' @export
use_package_doc <- function(pkg = ".") {
  pkg <- as.package(pkg)

  use_template(
    "packagename-package.r",
    file.path("R", paste(pkg$package, "-package.r", sep = "")),
    data = list(name = pkg$package),
    open = TRUE,
    pkg = pkg
  )
}

#' Use specified package.
#'
#' This adds a dependency to DESCRIPTION and offers a little advice
#' about how to best use it.
#'
#' @param package Name of package to depend on.
#' @param type Type of dependency: must be one of "Imports", "Depends", 
#'   "Suggests", "Enhances", or "LinkingTo" (or unique abbreviation)
#' @param pkg package description, can be path or package name. See
#'   \code{\link{as.package}} for more information.
#' @family infrastructure
#' @export
#' @examples
#' \dontrun{
#' use_package("ggplot2")
#' use_package("dplyr", "suggests")
#'
#' }
use_package <- function(package, type = "Imports", pkg = ".") {
  stopifnot(is.character(package), length(package) == 1)
  stopifnot(is.character(type), length(type) == 1)

  if (!is_installed(package)) {
    stop(package, " must be installed before you can take a dependency on it",
      call. = FALSE)
  }

  types <- c("Imports", "Depends", "Suggests", "Enhances", "LinkingTo")
  names(types) <- tolower(types)

  type <- types[[match.arg(tolower(type), names(types))]]

  message("* Adding ", package, " to ", type)
  add_desc_package(pkg, type, package)

  msg <- switch(type,
    Imports = paste0("Refer to functions with ", package, "::fun()"),
    Depends = paste0("Are you sure you want Depends? Imports is almost always",
      " the better choice."),
    Suggests = paste0("Use requireNamespace(\"", package, "\", quietly = TRUE)",
      " to test if package is installed,\n",
      "then use ", package, "::fun() to refer to functions."),
    Enhances = "",
    LinkingTo = show_includes(package)
  )
  message("Next: ")
  message(msg)
  invisible()
}

show_includes <- function(package) {
  incl <- system.file("include", package = package)
  h <- dir(incl, "\\.(h|hpp)$")
  if (length(h) == 0) return()

  message("Possible includes are:\n",
    paste0("#include <", h, ">", collapse = "\n"))

}

add_desc_package <- function(pkg = ".", field, name) {
  pkg <- as.package(pkg)
  desc_path <- file.path(pkg$path, "DESCRIPTION")

  desc <- read_dcf(desc_path)
  old <- desc[[field]]
  if (is.null(old)) {
    new <- name
    changed <- TRUE
  } else {
    if (!grepl(paste0('\\b', name, '\\b'), old)) {
      new <- paste0(old, ",\n    ", name)
      changed <- TRUE
    } else {
      changed <- FALSE
    }
  }
  if (changed) {
    desc[[field]] <- new
    write_dcf(desc_path, desc)
  }
  invisible(changed)
}

#' Use data in a package.
#'
#' This function makes it easy to save package data in the correct format.
#'
#' @param ... Unquoted names of existing objects to save.
#' @param pkg Package where to store data. Defaults to package in working
#'   directory.
#' @param internal If \code{FALSE}, saves each object in individual
#'   \code{.rda} files in the \code{data/} directory. These are available
#'   whenever the package is loaded. If \code{TRUE}, stores all objects in
#'   a single \code{R/sysdata.rda} file. These objects are only available
#'   within the package.
#' @param overwrite By default, \code{use_data} will not overwrite existing
#'   files. If you really want to do so, set this to \code{TRUE}.
#' @param compress Choose the type of compression used by \code{\link{save}}.
#'   Should be one of "gzip", "bzip2" or "xz".
#' @export
#' @family infrastructure
#' @examples
#' \dontrun{
#' x <- 1:10
#' y <- 1:100
#'
#' use_data(x, y) # For external use
#' use_data(x, y, internal = TRUE) # For internal use
#' }
use_data <- function(..., pkg = ".", internal = FALSE, overwrite = FALSE,
                     compress = "bzip2") {
  pkg <- as.package(pkg)

  objs <- get_objs_from_dots(dots(...))

  if (internal) {
    dir_name <- file.path(pkg$path, "R")
    paths <- file.path(dir_name, "sysdata.rda")
    objs <- list(objs)
  } else {
    dir_name <- file.path(pkg$path, "data")
    paths <- file.path(dir_name, paste0(objs, ".rda"))
  }

  check_data_paths(paths, overwrite)

  message("Saving ", paste(unlist(objs), collapse = ", "),
          " as ", paste(basename(paths), collapse = ", "),
          " to ", dir_name)
  envir <- parent.frame()
  mapply(save, list = objs, file = paths,
         MoreArgs = list(envir = envir, compress = compress))

  invisible()
}

get_objs_from_dots <- function(.dots) {
  if (length(.dots) == 0L) {
    stop("Nothing to save", call. = FALSE)
  }

  is_name <- vapply(.dots, is.symbol, logical(1))
  if (any(!is_name)) {
    stop("Can only save existing named objects", call. = FALSE)
  }

  objs <- vapply(.dots, as.character, character(1))
  duplicated_objs <- which(stats::setNames(duplicated(objs), objs))
  if (length(duplicated_objs) > 0L) {
    objs <- unique(objs)
    warning("Saving duplicates only once: ",
            paste(names(duplicated_objs), collapse = ", "),
            call. = FALSE)
  }
  objs
}

check_data_paths <- function(paths, overwrite) {
  data_path <- dirname(paths[[1]])
  if (!file.exists(data_path)) dir.create(data_path)

  if (!overwrite) {
    paths_exist <- which(stats::setNames(file.exists(paths), paths))

    if (length(paths_exist) > 0L) {
      paths_exist <- unique(names(paths_exist))
      existing_names <- basename(paths_exist)
      stop(paste(existing_names, collapse = ", "), " already exists in ",
           dirname(paths_exist[[1L]]),
           ". ",
           "Use overwrite = TRUE to overwrite", call. = FALSE)
    }
  }
}

#' Use \code{data-raw} to compute package datasets.
#'
#' @param pkg Package where to create \code{data-raw}. Defaults to package in
#'   working directory.
#' @export
#' @family infrastructure
use_data_raw <- function(pkg = ".") {
  pkg <- as.package(pkg)

  use_directory("data-raw", ignore = TRUE, pkg = pkg)

  message("Next: \n",
    "* Add data creation scripts in data-raw\n",
    "* Use devtools::use_data() to add data to package")
}

#' Add a file to \code{.Rbuildignore}
#'
#' \code{.Rbuildignore} has a regular expression on each line, but it's
#' usually easier to work with specific file names. By default, will (crudely)
#' turn a filename into a regular expression that will only match that
#' path. Repeated entries will be silently removed.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @param files Name of file.
#' @param escape If \code{TRUE}, the default, will escape \code{.} to
#'   \code{\\.} and surround with \code{^} and \code{$}.
#' @return Nothing, called for its side effect.
#' @export
#' @aliases add_build_ignore
#' @family infrastructure
#' @keywords internal
use_build_ignore <- function(files, escape = TRUE, pkg = ".") {
  pkg <- as.package(pkg)

  if (escape) {
    files <- paste0("^", gsub("\\.", "\\\\.", files), "$")
  }

  path <- file.path(pkg$path, ".Rbuildignore")
  union_write(path, files)

  invisible(TRUE)
}

#' Create README files.
#'
#' Creates skeleton README files with sections for
#' \itemize{
#' \item a high-level description of the package and its goals
#' \item R code to install from GitHub, if GitHub usage detected
#' \item a basic example
#' }
#' Use \code{Rmd} if you want a rich intermingling of code and data. Use
#' \code{md} for a basic README. \code{README.Rmd} will be automatically
#' added to \code{.Rbuildignore}. The resulting README is populated with default
#' YAML frontmatter and R fenced code blocks (\code{md}) or chunks (\code{Rmd}).
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @export
#' @examples
#' \dontrun{
#' use_readme_rmd()
#' use_readme_md()
#' }
#' @family infrastructure
use_readme_rmd <- function(pkg = ".") {
  pkg <- as.package(pkg)

  if (uses_github(pkg$path)) {
    pkg$github <- github_info(pkg$path)
  }
  pkg$Rmd <- TRUE

  use_template("omni-README", save_as = "README.Rmd", data = pkg,
               ignore = TRUE, open = TRUE, pkg = pkg)
  use_build_ignore("^README-.*\\.png$", escape = FALSE, pkg = pkg)

  if (uses_git(pkg$path) && !file.exists(pkg$path, ".git", "hooks", "pre-commit")) {
    message("* Adding pre-commit hook")
    use_git_hook("pre-commit", render_template("readme-rmd-pre-commit.sh"),
      pkg = pkg)
  }

  invisible(TRUE)
}

#' @export
#' @rdname use_readme_rmd
use_readme_md <- function(pkg = ".") {
  pkg <- as.package(pkg)
  if (uses_github(pkg$path)) {
    pkg$github <- github_info(pkg$path)
  }

  use_template("omni-README", save_as = "README.md",
               data = pkg, open = TRUE, pkg = pkg)
}

#' Use NEWS.md
#'
#' This creates \code{NEWS.md} from a template.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @export
#' @family infrastructure
use_news_md <- function(pkg = ".") {
  pkg <- as.package(pkg)

  use_template("NEWS.md", data = pkg, open = TRUE, pkg = pkg)
}

#' @rdname infrastructure
#' @section \code{use_revdep}:
#' Add \code{revdep} directory and basic check template.
#' @export
#' @aliases add_travis
use_revdep <- function(pkg = ".") {
  pkg <- as.package(pkg)

  use_directory("revdep", ignore = TRUE, pkg = pkg)
  use_template(
    "revdep.R",
    "revdep/check.R",
    data = list(name = pkg$package),
    pkg = pkg
  )
  use_git_ignore(revdep_cache_path_raw(""), pkg = pkg)
}

#' @rdname infrastructure
#' @section \code{use_cran_comments}:
#' Add \code{cran-comments.md} template.
#' @export
#' @aliases add_travis
use_cran_comments <- function(pkg = ".") {
  pkg <- as.package(pkg)

  use_template(
    "cran-comments.md",
    data = list(rversion = paste0(version$major, ".", version$minor)),
    ignore = TRUE,
    open = TRUE,
    pkg = pkg
  )

  invisible()
}

#' @rdname infrastructure
#' @section \code{use_code_of_conduct}:
#' Add a code of conduct to from \url{http://contributor-covenant.org}.
#'
#' @export
#' @aliases add_travis
use_code_of_conduct <- function(pkg = ".") {
  pkg <- as.package(pkg)

  use_template(
    "CONDUCT.md",
    ignore = TRUE,
    pkg = pkg
  )

  message("* Don't forget to describe the code of conduct in your README.md:")
  message("Please note that this project is released with a ",
    "[Contributor Code of Conduct](CONDUCT.md). ", "By participating in this ",
    "project you agree to abide by its terms.")
}

add_build_ignore <- function(pkg = ".", files, escape = TRUE) {
  use_build_ignore(files, escape = escape, pkg = pkg)
}

union_write <- function(path, new_lines) {
  if (file.exists(path)) {
    lines <- readLines(path, warn = FALSE)
  } else {
    lines <- character()
  }

  all <- union(lines, new_lines)
  writeLines(all, path)
}


#' @rdname infrastructure
#' @section \code{use_cran_badge}:
#' Add a badge to show CRAN status and version number on the README
#' @export
use_cran_badge <- function(pkg = ".") {
  pkg <- as.package(pkg)
  message(
    " * Add a CRAN status shield by adding the following line to your README:\n",
    "[![CRAN_Status_Badge](http://www.r-pkg.org/badges/version/", pkg$package, ")](https://cran.r-project.org/package=", pkg$package, ")"
  )
  invisible(TRUE)
}

#' @rdname infrastructure
#' @section \code{use_mit_license}:
#' Adds the necessary infrastructure to declare your package as
#' distributed under the MIT license.
#' @param copyright_holder The copyright holder for this package. Defaults to
#'   \code{getOption("devtools.name")}.
#' @export
use_mit_license <- function(pkg = ".", copyright_holder = getOption("devtools.name", "<Author>")) {
  pkg <- as.package(pkg)

  # Update the DESCRIPTION
  message("* Updating license field in DESCRIPTION.")
  descPath <- file.path(pkg$path, "DESCRIPTION")
  DESCRIPTION <- read_dcf(descPath)
  DESCRIPTION$License <- "MIT + file LICENSE"
  write_dcf(descPath, DESCRIPTION)

  use_template(
    "mit-license.txt",
    "LICENSE",
    data = list(
      year = format(Sys.Date(), "%Y"),
      copyright_holder = copyright_holder
    ),
    open = identical(copyright_holder, "<Author>"),
    pkg = pkg
  )
}


#' @rdname infrastructure
#' @section \code{use_gpl3_license}:
#' Adds the necessary infrastructure to declare your package as
#' distributed under the GPL v3.
#' @export
use_gpl3_license <- function(pkg = ".") {
  pkg <- as.package(pkg)

  # Update the DESCRIPTION
  message("* Updating license field in DESCRIPTION.")
  descPath <- file.path(pkg$path, "DESCRIPTION")
  DESCRIPTION <- read_dcf(descPath)
  DESCRIPTION$License <- "GPL-3 + file LICENSE"
  write_dcf(descPath, DESCRIPTION)

  use_template("gpl-v3.md", "LICENSE", pkg = pkg)
}

#' @section \code{use_dev_version}:
#' This adds ".9000" to the package \code{DESCRIPTION}, adds a new heading to
#' \code{NEWS.md} (if it exists), and then checks the result into git.
#' @rdname infrastructure
#' @export
use_dev_version <- function(pkg = ".") {
  pkg <- as.package(pkg)

  if (uses_git(pkg$path) && git_uncommitted(pkg$path)) {
    stop(
      "Uncommited changes. Please commit to git before continuing",
      call. = FALSE
    )
  }

  message("* Adding .9000 to version")
  desc_path <- file.path(pkg$path, "DESCRIPTION")
  DESCRIPTION <- read_dcf(desc_path)
  if (length(unlist(package_version(DESCRIPTION$Version))) > 3) {
    stop("Already has development version", call. = FALSE)
  }
  DESCRIPTION$Version <- paste0(DESCRIPTION$Version, ".9000")
  write_dcf(desc_path, DESCRIPTION)

  news_path <- file.path(pkg$path, "news.md")
  if (file.exists(news_path)) {
    message("* Adding new heading to NEWS.md")

    news <- readLines(news_path)
    news <- c(
      paste0("# ", pkg$package, " ", DESCRIPTION$Version),
      "",
      news
    )
    writeLines(news, news_path)
  }

  if (uses_git(pkg$path)) {
    message("* Checking into git")
    r <- git2r::init(pkg$path)
    paths <- unlist(git2r::status(r))
    git2r::add(r, paths)
    git2r::commit(r, "Use development version")
  }

  invisible(TRUE)
}

# Utilities ---------------------------------------------------------------

use_directory <- function(path, ignore = FALSE, pkg = ".") {
  pkg <- as.package(pkg)
  pkg_path <- file.path(pkg$path, path)

  if (file.exists(pkg_path)) {
    if (!is_dir(pkg_path)) {
      stop("`", path, "` exists but is not a directory.", call. = FALSE)
    }
  } else {
    message("* Creating `", path, "`.")
    dir.create(pkg_path, showWarnings = FALSE, recursive = TRUE)
  }

  if (ignore) {
    message("* Adding `", path, "` to `.Rbuildignore`.")
    use_build_ignore(path, pkg = pkg)
  }

  invisible(TRUE)
}

use_template <- function(template, save_as = template, data = list(),
                         ignore = FALSE, open = FALSE, pkg = ".") {
  pkg <- as.package(pkg)

  path <- file.path(pkg$path, save_as)
  if (!can_overwrite(path)) {
    stop("`", save_as, "` already exists.", call. = FALSE)
  }

  template_path <- system.file("templates", template, package = "devtools",
    mustWork = TRUE)
  template_out <- whisker::whisker.render(readLines(template_path), data)

  message("* Creating `", save_as, "` from template.")
  writeLines(template_out, path)

  if (ignore) {
    message("* Adding `", save_as, "` to `.Rbuildignore`.")
    use_build_ignore(save_as, pkg = pkg)
  }

  if (open) {
    message("* Modify `", save_as, "`.")
    open_in_rstudio(path)
  }

  invisible(TRUE)
}

open_in_rstudio <- function(path) {
  if (!rstudioapi::isAvailable())
    return()

  if (!rstudioapi::hasFun("navigateToFile"))
    return()

  rstudioapi::navigateToFile(path)

}

can_overwrite <- function(path) {
  name <- basename(path)

  if (!file.exists(path)) {
    TRUE
  } else if (interactive() && !yesno("Overwrite `", name, "`?")) {
    TRUE
  } else {
    FALSE
  }
}

context("Infrastructure")

test_that("use_* functions consistently", {
  pkg <- "infrastructure"
  unlink(pkg, recursive = TRUE)
  withr::with_output_sink(tempfile(), create(pkg))

  use_test("test1", pkg = pkg)
  use_package_doc(pkg = pkg)
  use_vignette("test2", pkg = pkg)

  use_rcpp(pkg = pkg)
  use_travis(pkg = pkg, browse = FALSE)
  use_coverage(pkg = pkg)
  use_appveyor(pkg = pkg)

  x <- 1:100
  use_data(x, pkg = pkg)
  use_data_raw(pkg = pkg)

  use_readme_rmd(pkg = pkg)
  use_readme_md(pkg = pkg)
  use_news_md(pkg = pkg)

  use_revdep(pkg = pkg)
  use_cran_comments(pkg = pkg)
  use_code_of_conduct(pkg = pkg)

  use_mit_license(pkg = pkg)

  # Suppress R CMD check note
  file.rename("infrastructure/.travis.yml", "infrastructure/travis.yml")
  file.rename("infrastructure/.Rbuildignore", "infrastructure/Rbuildignore")
})

test_that("use_data", {
  on.exit(unlink(c("testUseData/data", "testUseData/R/sysdata.rda"),
                 recursive = TRUE, force = TRUE), add = TRUE)

  # Add data to package
  local({
    expect_false(exists("global_test_data_item_to_save", .GlobalEnv))
    .GlobalEnv$global_test_data_item_to_save <- 42L
    on.exit(rm(list = "global_test_data_item_to_save", pos = .GlobalEnv), add = TRUE)

    local_test_data_item_to_save <- global_test_data_item_to_save
    system_test_data_item_to_save <- global_test_data_item_to_save

    expect_message(use_data(global_test_data_item_to_save, pkg = "testUseData"),
                   "Saving")
    expect_message(use_data(local_test_data_item_to_save, pkg = "testUseData"),
                   "Saving")
    expect_message(use_data(system_test_data_item_to_save, pkg = "testUseData",
                            internal = TRUE),
                   "Saving")
    expect_error(use_data(global_test_data_item_to_save, pkg = "testUseData"),
                 "overwrite = TRUE")
    expect_error(use_data(local_test_data_item_to_save, pkg = "testUseData"),
                 "overwrite = TRUE")
    expect_error(use_data(system_test_data_item_to_save, pkg = "testUseData", internal = TRUE),
                 "overwrite = TRUE")

    expect_message(use_data(global_test_data_item_to_save, pkg = "testUseData",
                            overwrite = TRUE),
                   "Saving")
    expect_message(use_data(local_test_data_item_to_save, pkg = "testUseData",
                            overwrite = TRUE),
                   "Saving")
    expect_message(use_data(system_test_data_item_to_save, pkg = "testUseData",
                            internal = TRUE, overwrite = TRUE),
                   "Saving")
  })

  # Test data is in package
  local({
    expect_false(exists("global_test_data_item_to_save"))
    expect_false(exists("local_test_data_item_to_save"))
    expect_false(exists("system_test_data_item_to_save"))

    expect_warning(load_all("testUseData"),
      "Objects listed as exports, but not present in namespace: sysdata_export")

    on.exit(unload("testUseData"), add = TRUE)

    expect_false(exists("global_test_data_item_to_save"))
    expect_false(exists("local_test_data_item_to_save"))
    expect_equal(system_test_data_item_to_save, 42L)

    data(global_test_data_item_to_save, envir = environment())
    data(local_test_data_item_to_save, envir = environment())
    expect_equal(global_test_data_item_to_save, 42L)
    expect_equal(local_test_data_item_to_save, 42L)
  })
})

#' Uninstall a local development package.
#'
#' Uses \code{remove.package} to uninstall the package.
#' To uninstall a package from a non-default library,
#' use \code{\link[withr]{with_libpaths}}.
#'
#' @inheritParams install
#' @param unload if \code{TRUE} (the default), will automatically unload the
#'   package prior to uninstalling.
#' @param ... additional arguments passed to \code{\link{remove.packages}}.
#' @export
#' @family package installation
#' @seealso \code{\link{with_debug}} to install packages with debugging flags
#'   set.
uninstall <- function(pkg = ".", unload = TRUE, quiet = FALSE, ...) {

  pkg <- as.package(pkg)

  if (unload && pkg$package %in% loaded_packages()$package) {
    unload(pkg)
  }

  if (!quiet) {
    message("Uninstalling ", pkg$package)
  }

  remove.packages(pkg$package)

  invisible(TRUE)
}

# Given the name or vector of names, returns a named vector reporting
# whether each exists and is a directory.
dir.exists <- function(x) {
  res <- file.exists(x) & file.info(x)$isdir
  stats::setNames(res, x)
}

compact <- function(x) {
  is_empty <- vapply(x, function(x) length(x) == 0, logical(1))
  x[!is_empty]
}

"%||%" <- function(a, b) if (!is.null(a)) a else b

"%:::%" <- function(p, f) {
  get(f, envir = asNamespace(p))
}

rule <- function(..., pad = "-") {
  if (nargs() == 0) {
    title <- ""
  } else {
    title <- paste0(..., " ")
  }
  width <- max(getOption("width") - nchar(title) - 1, 0)
  message(title, paste(rep(pad, width, collapse = "")))
}

# check whether the specified file ends with newline
ends_with_newline <- function(path) {
  conn <- file(path, open = "rb", raw = TRUE)
  on.exit(close(conn))
  seek(conn, where = -1, origin = "end")
  lastByte <- readBin(conn, "raw", n = 1)
  lastByte == 0x0a
}

render_template <- function(name, data = list()) {
  path <- system.file("templates", name, package = "devtools")
  template <- readLines(path)
  whisker::whisker.render(template, data)
}

is_installed <- function(pkg, version = 0) {
  installed_version <- tryCatch(utils::packageVersion(pkg), error = function(e) NA)
  !is.na(installed_version) && installed_version >= version
}

check_bioconductor <- function() {
  if (is_installed("BiocInstaller")) {
    return()
  }

  msg <- paste0("'BiocInstaller' must be installed to install Bioconductor packages")
  if (!interactive()) {
    stop(msg, call. = FALSE)
  }

  message(
    msg, ".\n",
    "Would you like to install it? ",
    "This will source <https://bioconductor.org/biocLite.R>."
  )

  if (menu(c("Yes", "No")) != 1) {
    stop("'BiocInstaller' not installed", call. = FALSE)
  }

  suppressMessages(
    source("https://bioconductor.org/biocLite.R")
  )
}

check_suggested <- function(pkg, version = NULL, compare = NA) {

  if (is.null(version)) {
    if (!is.na(compare)) {
      stop("Cannot set ", sQuote(compare), " without setting ",
           sQuote(version), call. = FALSE)
    }

    dep <- suggests_dep(pkg)

    version <- dep$version
    compare <- dep$compare
  }

  if (!is_installed(pkg) || !check_dep_version(pkg, version, compare)) {
    msg <- paste0(sQuote(pkg),
      if (is.na(version)) "" else paste0(" >= ", version),
      " must be installed for this functionality.")

    if (interactive()) {
      message(msg, "\nWould you like to install it?")
      if (menu(c("Yes", "No")) == 1) {
        install.packages(pkg)
      } else {
        stop(msg, call. = FALSE)
      }
    } else {
      stop(msg, call. = FALSE)
    }
  }
}

suggests_dep <- function(pkg) {

  suggests <- read_dcf(system.file("DESCRIPTION", package = "devtools"))$Suggests
  deps <- parse_deps(suggests)

  found <- which(deps$name == pkg)[1L]

  if (!length(found)) {
     stop(sQuote(pkg), " is not in Suggests: for devtools!", call. = FALSE)
  }
  deps[found, ]
}

read_dcf <- function(path) {
  fields <- colnames(read.dcf(path))
  as.list(read.dcf(path, keep.white = fields)[1, ])
}

write_dcf <- function(path, desc) {
  desc <- unlist(desc)
  # Add back in continuation characters
  desc <- gsub("\n[ \t]*\n", "\n .\n ", desc, perl = TRUE, useBytes = TRUE)
  desc <- gsub("\n \\.([^\n])", "\n  .\\1", desc, perl = TRUE, useBytes = TRUE)

  starts_with_whitespace <- grepl("^\\s", desc, perl = TRUE, useBytes = TRUE)
  delimiters <- ifelse(starts_with_whitespace, ":", ": ")
  text <- paste0(names(desc), delimiters, desc, collapse = "\n")

  # If the description file has a declared encoding, set it so nchar() works
  # properly.
  if ("Encoding" %in% names(desc)) {
    Encoding(text) <- desc[["Encoding"]]
  }

  if (substr(text, nchar(text), 1) != "\n") {
    text <- paste0(text, "\n")
  }

  cat(text, file = path)
}

dots <- function(...) {
  eval(substitute(alist(...)))
}

first_upper <- function(x) {
  substr(x, 1, 1) <- toupper(substr(x, 1, 1))
  x
}

download <- function(path, url, ...) {
  request <- httr::GET(url, ...)
  httr::stop_for_status(request)
  writeBin(httr::content(request, "raw"), path)
  path
}

download_text <- function(url, ...) {
  request <- httr::GET(url, ...)
  httr::stop_for_status(request)
  httr::content(request, "text")
}

last <- function(x) x[length(x)]

# Modified version of utils::file_ext. Instead of always returning the text
# after the last '.', as in "foo.tar.gz" => ".gz", if the text that directly
# precedes the last '.' is ".tar", it will include also, so
# "foo.tar.gz" => ".tar.gz"
file_ext <- function (x) {
    pos <- regexpr("\\.((tar\\.)?[[:alnum:]]+)$", x)
    ifelse(pos > -1L, substring(x, pos + 1L), "")
}

is_bioconductor <- function(x) {
  x$package != "BiocInstaller" && !is.null(x$biocviews)
}

trim_ws <- function(x) {
  gsub("^[[:space:]]+|[[:space:]]+$", "", x)
}

is_dir <- function(x) file.info(x)$isdir

indent <- function(x, spaces = 4) {
  ind <- paste(rep(" ", spaces), collapse = "")
  paste0(ind, gsub("\n", paste0("\n", ind), x, fixed = TRUE))
}

is_windows <- isTRUE(.Platform$OS.type == "windows")

all_named <- function (x) {
  if (length(x) == 0) return(TRUE)
  !is.null(names(x)) && all(names(x) != "")
}

make_function <- function (args, body, env = parent.frame()) {
  args <- as.pairlist(args)
  stopifnot(all_named(args), is.language(body))
  eval(call("function", args, body), env)
}

comp_lang <- function(x, y, idx = seq_along(y)) {
  if (is.symbol(x) || is.symbol(y)) {
    return(identical(x, y))
  }

  if (length(x) < length(idx)) return(FALSE)

  identical(x[idx], y[idx])
}

extract_lang <- function(x, f, ...) {
  recurse <- function(y) {
    unlist(compact(lapply(y, extract_lang, f = f, ...)), recursive = FALSE)
  }

  # if x matches predicate return it
  if (isTRUE(f(x, ...))) {
    return(x)
  }

  if (is.call(x)) {
    res <- recurse(x)[[1]]
    if (top_level_call <- identical(sys.call()[[1]], as.symbol("extract_lang"))
        && is.null(res)) {
      warning("Devtools is incompatible with the current version of R. `load_all()` may function incorrectly.")
    }
    return(res)
  }

  NULL
}

modify_lang <- function(x, f, ...) {
  recurse <- function(x) {
    lapply(x, modify_lang, f = f, ...)
  }

  x <- f(x, ...)

  if (is.call(x)) {
    as.call(recurse(x))
  } else if (is.function(x)) {
     formals(x) <- modify_lang(formals(x), f, ...)
     body(x) <- modify_lang(body(x), f, ...)
  } else {
    x
  }
}

strip_internal_calls <- function(x, package) {
  if (is.call(x) && identical(x[[1L]], as.name(":::")) && identical(x[[2L]], as.name(package))) {
    x[[3L]]
  } else {
    x
  }
}

sort_ci <- function(x) {
  withr::with_collate("C", x[order(tolower(x), x)])
}

comma <- function(x, at_most = 20) {
  if (length(x) > at_most) {
    x <- c(x[seq_len(at_most)], "...")
  }
  paste(x, collapse = ", ")
}

menu <- function(...) {
  utils::menu(...)
}

context("extract_lang")

f <- function(x) {
  a <- 1:10
  for (i in seq_along(a)) {
    print(i)
  }
}

test_that("extract_lang issues warning if nothing found", {

  expect_warning(extract_lang(body(f), comp_lang, quote(j)),
    "Devtools is incompatible")
})

test_that("extract_lang and comp_lang finds full statements", {

  expect_equal(extract_lang(body(f), comp_lang, quote(a <- 1:10)),
    quote(a <- 1:10))
})

test_that("extract_lang and comp_lang find child calls", {

  expect_equal(extract_lang(body(f), comp_lang, quote(seq_along(a))),
    quote(seq_along(a)))
})

test_that("extract_lang and comp_lang finds partial statements", {

  expect_equal(extract_lang(body(f), comp_lang, quote(a <- NULL), 1:2),
    quote(a <- 1:10))
})

test_that("extract_lang and comp_lang finds partial statements from for conditionals", {

  expect_equal(extract_lang(body(f), comp_lang, quote(for (i in seq_along(a)) NULL), 1:3),
    quote(for (i in seq_along(a)) { print(i) }))
})

test_that("modify_lang modifies properly", {
  expect_equal(modify_lang(quote(a <- 1:10), function(x) if (comp_lang(x, quote(a))) quote(b) else x),
      quote(b <- 1:10))
})

#' Attempts to install a package directly from GitHub.
#'
#' This function is vectorised on \code{repo} so you can install multiple
#' packages in a single command.
#'
#' @param repo Repository address in the format
#'   \code{username/repo[/subdir][@@ref|#pull]}. Alternatively, you can
#'   specify \code{subdir} and/or \code{ref} using the respective parameters
#'   (see below); if both are specified, the values in \code{repo} take
#'   precedence.
#' @param username User name. Deprecated: please include username in the
#'   \code{repo}
#' @param ref Desired git reference. Could be a commit, tag, or branch
#'   name, or a call to \code{\link{github_pull}}. Defaults to \code{"master"}.
#' @param subdir subdirectory within repo that contains the R package.
#' @param auth_token To install from a private repo, generate a personal
#'   access token (PAT) in \url{https://github.com/settings/tokens} and
#'   supply to this argument. This is safer than using a password because
#'   you can easily delete a PAT without affecting any others. Defaults to
#'   the \code{GITHUB_PAT} environment variable.
#' @param host GitHub API host to use. Override with your GitHub enterprise
#'   hostname, for example, \code{"github.hostname.com/api/v3"}.
#' @param quiet if \code{TRUE} suppresses output from this function.
#' @param ... Other arguments passed on to \code{\link{install}}.
#' @details
#' Attempting to install from a source repository that uses submodules
#' raises a warning. Because the zipped sources provided by GitHub do not
#' include submodules, this may lead to unexpected behaviour or compilation
#' failure in source packages. In this case, cloning the repository manually
#' using \code{\link{install_git}} with \code{args="--recursive"} may yield
#' better results.
#' @export
#' @family package installation
#' @seealso \code{\link{github_pull}}
#' @examples
#' \dontrun{
#' install_github("klutometis/roxygen")
#' install_github("wch/ggplot2")
#' install_github(c("rstudio/httpuv", "rstudio/shiny"))
#' install_github(c("hadley/httr@@v0.4", "klutometis/roxygen#142",
#'   "mfrasca/r-logging/pkg"))
#'
#' # Update devtools to the latest version, on Linux and Mac
#' # On Windows, this won't work - see ?build_github_devtools
#' install_github("hadley/devtools")
#'
#' # To install from a private repo, use auth_token with a token
#' # from https://github.com/settings/tokens. You only need the
#' # repo scope. Best practice is to save your PAT in env var called
#' # GITHUB_PAT.
#' install_github("hadley/private", auth_token = "abc")
#'
#' }
install_github <- function(repo, username = NULL,
                           ref = "master", subdir = NULL,
                           auth_token = github_pat(quiet),
                           host = "https://api.github.com", quiet = FALSE,
                           ...) {

  remotes <- lapply(repo, github_remote, username = username, ref = ref,
    subdir = subdir, auth_token = auth_token, host = host)

  install_remotes(remotes, quiet = quiet, ...)
}

github_remote <- function(repo, username = NULL, ref = NULL, subdir = NULL,
                       auth_token = github_pat(), sha = NULL,
                       host = "https://api.github.com") {

  meta <- parse_git_repo(repo)
  meta <- github_resolve_ref(meta$ref %||% ref, meta)

  if (is.null(meta$username)) {
    meta$username <- username %||% getOption("github.user") %||%
      stop("Unknown username.")
    warning("Username parameter is deprecated. Please use ",
      username, "/", repo, call. = FALSE)
  }

  remote("github",
    host = host,
    repo = meta$repo,
    subdir = meta$subdir %||% subdir,
    username = meta$username,
    ref = meta$ref,
    sha = sha,
    auth_token = auth_token
  )
}

#' @export
remote_download.github_remote <- function(x, quiet = FALSE) {
  dest <- tempfile(fileext = paste0(".zip"))

  if (missing_protocol <- !grepl("^[^:]+?://", x$host)) {
    x$host <- paste0("https://", x$host)
  }

  src_root <- paste0(x$host, "/repos/", x$username, "/", x$repo)
  src <- paste0(src_root, "/zipball/", x$ref)

  if (!quiet) {
    message("Downloading GitHub repo ", x$username, "/", x$repo, "@", x$ref,
            "\nfrom URL ", src)
  }

  if (!is.null(x$auth_token)) {
    auth <- httr::authenticate(
      user = x$auth_token,
      password = "x-oauth-basic",
      type = "basic"
    )
  } else {
    auth <- NULL
  }

  if (github_has_remotes(x, auth))
    warning("GitHub repo contains submodules, may not function as expected!",
      call. = FALSE)

  download_github(dest, src, auth)
}

github_has_remotes <- function(x, auth = NULL) {
  src_root <- paste0(x$host, "/repos/", x$username, "/", x$repo)
  src_submodules <- paste0(src_root, "/contents/.gitmodules?ref=", x$ref)
  response <- httr::HEAD(src_submodules, , auth)
  identical(httr::status_code(response), 200L)
}

#' @export
remote_metadata.github_remote <- function(x, bundle = NULL, source = NULL) {
  # Determine sha as efficiently as possible
  if (!is.null(bundle)) {
    # Might be able to get from zip archive
    sha <- git_extract_sha1(bundle)
  } else {
    # Otherwise can lookup with remote_ls
    sha <- remote_sha(x)
  }

  list(
    RemoteType = "github",
    RemoteHost = x$host,
    RemoteRepo = x$repo,
    RemoteUsername = x$username,
    RemoteRef = x$ref,
    RemoteSha = sha,
    RemoteSubdir = x$subdir,
    # Backward compatibility for packrat etc.
    GithubRepo = x$repo,
    GithubUsername = x$username,
    GithubRef = x$ref,
    GithubSHA1 = sha,
    GithubSubdir = x$subdir
  )
}

#' GitHub references
#'
#' Use as \code{ref} parameter to \code{\link{install_github}}.
#' Allows installing a specific pull request or the latest release.
#'
#' @param pull The pull request to install
#' @seealso \code{\link{install_github}}
#' @rdname github_refs
#' @export
github_pull <- function(pull) structure(pull, class = "github_pull")

#' @rdname github_refs
#' @export
github_release <- function() structure(NA_integer_, class = "github_release")

github_resolve_ref <- function(x, params) UseMethod("github_resolve_ref")

#' @export
github_resolve_ref.default <- function(x, params) {
  params$ref <- x
  params
}

#' @export
github_resolve_ref.NULL <- function(x, params) {
  params$ref <- "master"
  params
}

#' @export
github_resolve_ref.github_pull <- function(x, params) {
  # GET /repos/:user/:repo/pulls/:number
  path <- file.path("repos", params$username, params$repo, "pulls", x)
  response <- github_GET(path)

  params$username <- response$head$user$login
  params$ref <- response$head$ref
  params
}

# Retrieve the ref for the latest release
#' @export
github_resolve_ref.github_release <- function(x, params) {
  # GET /repos/:user/:repo/releases
  path <- paste("repos", params$username, params$repo, "releases", sep = "/")
  response <- github_GET(path)
  if (length(response) == 0L)
    stop("No releases found for repo ", params$username, "/", params$repo, ".")

  params$ref <- response[[1L]]$tag_name
  params
}

# Parse concise git repo specification: [username/]repo[/subdir][#pull|@ref|@*release]
# (the *release suffix represents the latest release)
parse_git_repo <- function(path) {
  username_rx <- "(?:([^/]+)/)?"
  repo_rx <- "([^/@#]+)"
  subdir_rx <- "(?:/([^@#]*[^@#/]))?"
  ref_rx <- "(?:@([^*].*))"
  pull_rx <- "(?:#([0-9]+))"
  release_rx <- "(?:@([*]release))"
  ref_or_pull_or_release_rx <- sprintf("(?:%s|%s|%s)?", ref_rx, pull_rx, release_rx)
  github_rx <- sprintf("^(?:%s%s%s%s|(.*))$",
    username_rx, repo_rx, subdir_rx, ref_or_pull_or_release_rx)

  param_names <- c("username", "repo", "subdir", "ref", "pull", "release", "invalid")
  replace <- stats::setNames(sprintf("\\%d", seq_along(param_names)), param_names)
  params <- lapply(replace, function(r) gsub(github_rx, r, path, perl = TRUE))
  if (params$invalid != "")
    stop(sprintf("Invalid git repo: %s", path))
  params <- params[sapply(params, nchar) > 0]

  if (!is.null(params$pull)) {
    params$ref <- github_pull(params$pull)
    params$pull <- NULL
  }

  if (!is.null(params$release)) {
    params$ref <- github_release()
    params$release <- NULL
  }

  params
}

#' @export
remote_package_name.github_remote <- function(remote, url = "https://raw.githubusercontent.com", ...) {

  tmp <- tempfile()
  path <- paste(c(
      remote$username,
      remote$repo,
      remote$ref,
      remote$subdir,
      "DESCRIPTION"), collapse = "/")

  if (!is.null(remote$auth_token)) {
    auth <- httr::authenticate(
      user = remote$auth_token,
      password = "x-oauth-basic",
      type = "basic"
    )
  } else {
    auth <- NULL
  }

  req <- httr::GET(url, path = path, httr::write_disk(path = tmp), auth)

  if (httr::status_code(req) >= 400) {
    return(NA_character_)
  }

  read_dcf(tmp)$Package
}

#' @export
remote_sha.github_remote <- function(remote, url = "https://github.com", ...) {
  # If the remote ref is the same as the sha it is a pinned commit so just
  # return that.
  if (!is.null(remote$ref) && !is.null(remote$sha) &&
    grepl(paste0("^", remote$ref), remote$sha)) {
    return(remote$sha)
  }

  tryCatch({
    res <- git2r::remote_ls(
      paste0(url, "/", remote$username, "/", remote$repo, ".git"),
      ...)

    found <- grep(pattern = paste0("/", remote$ref), x = names(res))

    if (length(found) == 0) {
      return(NA_character_)
    }

    unname(res[found[1]])
  }, error = function(e) NA_character_)
}

#' @export
format.github_remote <- function(x, ...) {
  "GitHub"
}

download_github <- function(path, url, ...) {
  request <- httr::GET(url, ...)

  if (httr::status_code(request) >= 400) {
     stop(github_error(request))
  }

  writeBin(httr::content(request, "raw"), path)
  path
}

context("GitHub")

with_mock <- function(name, value, code) {
  env <- asNamespace("devtools")
  orig_value <- env[[name]]
  unlockBinding(name, env)
  env[[name]] <- value
  on.exit(env[[name]] <- orig_value)

  force(code)
}

test_that("GitHub repo paths are parsed correctly", {
  expect_equal(parse_git_repo("devtools"), list(repo="devtools"))
  expect_equal(parse_git_repo("krlmlr/kimisc"), list(username="krlmlr", repo="kimisc"))
  expect_equal(parse_git_repo("my/test/pkg"), list(username="my", repo="test", subdir="pkg"))
  expect_equal(parse_git_repo("devtools@devtools-1.4"), list(repo="devtools", ref="devtools-1.4"))
  expect_equal(parse_git_repo("yihui/tikzDevice#23"), list(username="yihui", repo="tikzDevice", ref=github_pull("23")))
  expect_equal(parse_git_repo("my/test/pkg@ref"), list(username="my", repo="test", subdir="pkg", ref="ref"))
  expect_equal(parse_git_repo("my/test/pkg#1"), list(username="my", repo="test", subdir="pkg", ref=github_pull("1")))
  expect_error(parse_git_repo("test#6@123"), "Invalid git repo")
  expect_error(parse_git_repo("Teradata/teradataR/"), "Invalid git repo")
  expect_error(parse_git_repo("test@*unsupported-release"), "Invalid git repo")
})

# Mock github_resolve_ref.github_pull so that GitHub API is not queried for this test
mock_github_resolve_ref.github_pull <- function(x, params) {
  params$username <- sprintf("user-%s", x)
  params$ref <- sprintf("pull-%s", x)
  params
}

# Mock github_resolve_ref.github_release so that GitHub API is not queried for this test
mock_github_resolve_ref.github_release <- function(x, param) {
  param$ref="latest-release"
  param
}

test_that("GitHub parameters are returned correctly", {
  with_mock("github_resolve_ref.github_pull", mock_github_resolve_ref.github_pull, {
    expect_equal(github_remote("hadley/devtools")$repo, "devtools")
    expect_equal(github_remote("krlmlr/kimisc")$username, "krlmlr")
    expect_equal(github_remote("my/test/pkg")$subdir, "pkg")
    expect_equal(github_remote("hadley/devtools@devtools-1.4")$ref, "devtools-1.4")
    expect_equal(github_remote("yihui/tikzDevice#23")$ref, "pull-23")
  })

  with_mock("github_resolve_ref.github_release", mock_github_resolve_ref.github_release, {
    expect_equal(github_remote("yihui/tikzDevice@*release")$ref, "latest-release")
    expect_equal(github_remote("my/test/pkg@*release")$ref, "latest-release")
  })
})

mock_github_GET <- function(path) {
  if (grepl("^repos/.*/pulls/.*$", path)) {
    list(head=list(user=list(login="username"), ref="some-pull-request"))
  } else if (grepl("^repos/.*/releases$", path)) {
    list(list(tag_name="some-release"))
  } else
    stop("unexpected path: ", path)
}

test_that("GitHub references are resolved correctly", {
  default_params <- as.list(stats::setNames(nm=c("repo", "username")))
  with_mock("github_GET", mock_github_GET, {
    expect_equal(github_resolve_ref(NULL, list())$ref, "master")
    expect_equal(github_resolve_ref("some-ref", list())$ref, "some-ref")
    expect_equal(github_resolve_ref(github_pull(123), default_params)$username, "username")
    expect_equal(github_resolve_ref(github_pull(123), default_params)$ref, "some-pull-request")
    expect_equal(github_resolve_ref(github_release(), default_params)$ref, "some-release")
  })
})

test_that("Github repos with submodules are identified correctly", {
  # Appveyor has a very low GitHub rate limit which causes this to fail often, so
  # skip these tests
  skip_on_appveyor()
  skip_on_travis()

  expect_equal(github_has_remotes(github_remote("hadley/devtools")), FALSE)
  ## a r package repo known to use submodules
  expect_equal(github_has_remotes(github_remote("armstrtw/rzmq")), TRUE)
})

#' Get the installation path of a package
#'
#' Given the name of a package, this returns a path to the installed
#' copy of the package, which can be passed to other devtools functions.
#'
#' It searches for the package in \code{\link{.libPaths}()}. If multiple
#' dirs are found, it will return the first one.
#'
#' @param name the name of a package.
#'
#' @examples
#' inst("devtools")
#' inst("grid")
#' \dontrun{
#' # Can be passed to other devtools functions
#' unload(inst("ggplot2"))
#' }
#' @export
inst <- function(name) {
  # It would be nice to use find.package or system.file, but they
  # also search in the directory in the 'path' attribute of the
  # package environment.

  # Look in the library paths
  paths <- file.path(.libPaths(), name)
  paths <- paths[dir.exists(paths)]

  if (length(paths) > 0) {
    # If multiple matches, return the first one
    return(normalizePath(paths[1]))
  } else {
    return(NULL)
  }
}

available_packages <- memoise::memoise(function(repos, type) {
  suppressWarnings(available.packages(contrib.url(repos, type), type = type))
})

package_url <- function(package, repos,
                        available = available_packages(repos, "source")) {
  ok <- (available[, "Package"] == package)
  ok <- ok & !is.na(ok)

  if (!any(ok)) {
    return(list(name = NA_character_, url = NA_character_))
  }

  vers <- package_version(available[ok, "Version"])
  keep <- vers == max(vers)
  keep[duplicated(keep)] <- FALSE
  ok[ok][!keep] <- FALSE

  name <- paste(package, "_", available[ok, "Version"], ".tar.gz", sep = "")
  url <- file.path(available[ok, "Repository"], name)

  list(name = name, url = url)
}


# Return the version of a package on CRAN (or other repository)
# @param package The name of the package.
# @param available A matrix of information about packages.
cran_pkg_version <- function(package, available = available.packages()) {

  idx <- available[, "Package"] == package
  if (any(idx)) {
    as.package_version(available[package, "Version"])
  } else {
    NULL
  }
}

#' Load complete package.
#'
#' \code{load_all} loads a package. It roughly simulates what happens
#' when a package is installed and loaded with \code{\link{library}}.
#'
#' Currently \code{load_all}:
#'
#' \itemize{
#'   \item Loads all data files in \code{data/}.  See \code{\link{load_data}}
#'     for more details.
#'
#'   \item Sources all R files in the R directory, storing results in
#'     environment that behaves like a regular package namespace. See
#'     below and \code{\link{load_code}} for more details.
#'
#'   \item Compiles any C, C++, or Fortran code in the \code{src/} directory
#'     and connects the generated DLL into R. See \code{\link{compile_dll}}
#'     for more details.
#'
#'   \item Runs \code{.onAttach()}, \code{.onLoad()} and \code{.onUnload()}
#'     functions at the correct times.
#'
#'   \item If you use \pkg{testthat}, will load all test helpers so you
#'     can access them interactively.
#'
#' }
#'
#' @section Namespaces:
#' The namespace environment \code{<namespace:pkgname>}, is a child of
#' the imports environment, which has the name attribute
#' \code{imports:pkgname}. It is in turn is a child of
#' \code{<namespace:base>}, which is a child of the global environment.
#' (There is also a copy of the base namespace that is a child of the empty
#' environment.)
#'
#' The package environment \code{<package:pkgname>} is an ancestor of the
#' global environment. Normally when loading a package, the objects
#' listed as exports in the NAMESPACE file are copied from the namespace
#' to the package environment. However, \code{load_all} by default will
#' copy all objects (not just the ones listed as exports) to the package
#' environment. This is useful during development because it makes all
#' objects easy to access.
#'
#' To export only the objects listed as exports, use
#' \code{export_all=FALSE}. This more closely simulates behavior when
#' loading an installed package with \code{\link{library}}, and can be
#' useful for checking for missing exports.
#'
#' @section Shim files:
#' \code{load_all} also inserts shim functions into the imports environment
#' of the laded package. It presently adds a replacement version of
#' \code{system.file} which returns different paths from
#' \code{base::system.file}. This is needed because installed and uninstalled
#' package sources have different directory structures. Note that this is not
#' a perfect replacement for \code{base::system.file}.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information.
#' @param reset clear package environment and reset file cache before loading
#'   any pieces of the package. This is equivalent to running
#'   \code{\link{unload}} and is the default. Use \code{reset = FALSE} may be
#'   faster for large code bases, but is a significantly less accurate
#'   approximation.
#' @param recompile force a recompile of DLL from source code, if present.
#'   This is equivalent to running \code{\link{clean_dll}} before
#'   \code{load_all}
#' @param export_all If \code{TRUE} (the default), export all objects.
#'   If \code{FALSE}, export only the objects that are listed as exports
#'   in the NAMESPACE file.
#' @param quiet if \code{TRUE} suppresses output from this function.
#' @inheritParams as.package
#' @keywords programming
#' @examples
#' \dontrun{
#' # Load the package in the current directory
#' load_all("./")
#'
#' # Running again loads changed files
#' load_all("./")
#'
#' # With reset=TRUE, unload and reload the package for a clean start
#' load_all("./", TRUE)
#'
#' # With export_all=FALSE, only objects listed as exports in NAMESPACE
#' # are exported
#' load_all("./", export_all = FALSE)
#' }
#' @export
load_all <- function(pkg = ".", reset = TRUE, recompile = FALSE,
  export_all = TRUE, quiet = FALSE, create = NA) {
  pkg <- as.package(pkg, create = create)
  check_suggested("roxygen2")

  if (!quiet) message("Loading ", pkg$package)

  if (pkg$package == "compiler") {
    # Disable JIT while loading the compiler package to avoid interference
    # (otherwise the compiler package would be loaded as a side effect of
    # JIT compilation and it would be locked before we can insert shims into
    # it).
    oldEnabled <- compiler::enableJIT(0)
    on.exit(compiler::enableJIT(oldEnabled), TRUE)
  }

  roxygen2::update_collate(pkg$path)
  # Refresh the pkg structure with any updates to the Collate entry
  # in the DESCRIPTION file
  pkg$collate <- as.package(pkg$path)$collate

  # Forcing all of the promises for the loaded namespace now will avoid lazy-load
  # errors when the new package is loaded overtop the old one.
  #
  # Reloading devtools is a special case. Normally, objects in the
  # namespace become inaccessible if the namespace is unloaded before the
  # object has been accessed. Instead we force the object so they will still be
  # accessible.
  if (is_loaded(pkg)) {
    eapply(ns_env(pkg), force, all.names = TRUE)
  }

  # Check description file is ok
  check <- ("tools" %:::% ".check_package_description")(
    file.path(pkg$path, "DESCRIPTION"))
  if (length(check) > 0) {
    msg <- utils::capture.output(("tools" %:::% "print.check_package_description")(check))
    message("Invalid DESCRIPTION:\n", paste(msg, collapse = "\n"))
  }

  # If installed version of package loaded, unload it
  if (is_loaded(pkg) && is.null(dev_meta(pkg$package))) {
    unload(pkg)
  }

  # Unload dlls
  unload_dll(pkg)

  if (reset) {
    clear_cache()
    if (is_loaded(pkg)) unload(pkg)
  }

  if (recompile) clean_dll(pkg)

  # Compile dll if it exists
  compile_dll(pkg, quiet = quiet)


  # Set up the namespace environment ----------------------------------
  # This mimics the procedure in loadNamespace

  if (!is_loaded(pkg)) create_ns_env(pkg)

  out <- list(env = ns_env(pkg))

  # Load dependencies
  load_depends(pkg)
  load_imports(pkg)
  # Add shim objects to imports environment
  insert_imports_shims(pkg)

  out$data <- load_data(pkg)

  out$code <- load_code(pkg)
  register_s3(pkg)
  out$dll <- load_dll(pkg)

  # Run namespace load hooks
  run_pkg_hook(pkg, "load")
  run_ns_load_actions(pkg)
  run_user_hook(pkg, "load")

  # Set up the exports in the namespace metadata (this must happen after
  # the objects are loaded)
  setup_ns_exports(pkg, export_all)

  # Set up the package environment ------------------------------------
  # Create the package environment if needed
  if (!is_attached(pkg)) attach_ns(pkg)

  # Copy over objects from the namespace environment
  export_ns(pkg)

  # Run hooks
  run_pkg_hook(pkg, "attach")
  run_user_hook(pkg, "attach")

  # Source test helpers into package environment
  if (uses_testthat(pkg)) {
    testthat::source_test_helpers(find_test_dir(pkg$path), env = pkg_env(pkg))
  }


  # Replace help and ? in utils package environment
  insert_global_shims()

  invisible(out)
}


#' Create a default DESCRIPTION file for a package.
#'
#' @details
#' To set the default author and licenses, set \code{options}
#' \code{devtools.desc.author} and \code{devtools.desc.license}.  I use
#' \code{options(devtools.desc.author = '"Hadley Wickham <h.wickham@@gmail.com> [aut,cre]"',
#'   devtools.desc.license = "GPL-3")}.
#' @param path path to package root directory
#' @param extra a named list of extra options to add to \file{DESCRIPTION}.
#'   Arguments that take a list
#' @param quiet if \code{TRUE}, suppresses output from this function.
#' @export
create_description <- function(path = ".", extra = getOption("devtools.desc"),
                               quiet = FALSE) {
  # Don't call check_dir(path) here (#803)
  desc_path <- file.path(path, "DESCRIPTION")

  if (file.exists(desc_path)) return(FALSE)

  subdir <- file.path(path, c("R", "src", "data"))
  if (!any(file.exists(subdir))) {
    stop("'", path, "' does not look like a package: no R/, src/ or data directories",
      call. = FALSE)
  }

  desc <- build_description(extract_package_name(path), extra)

  if (!quiet) {
    message("No DESCRIPTION found. Creating with values:\n\n")
    write_dcf("", desc)
  }

  write_dcf(desc_path, desc)

  TRUE
}

build_description <- function(name, extra = list()) {
  check_package_name(name)

  defaults <- compact(list(
    Package = name,
    Title = "What the Package Does (one line, title case)",
    Version = "0.0.0.9000",
    "Authors@R" = getOption("devtools.desc.author"),
    Description = "What the package does (one paragraph).",
    Depends = paste0("R (>= ", as.character(getRversion()) ,")"),
    License = getOption("devtools.desc.license"),
    Suggests = getOption("devtools.desc.suggests"),
    Encoding = "UTF-8",
    LazyData = "true"
  ))

  # Override defaults with user supplied options
  desc <- modifyList(defaults, extra)
  # Collapse all vector arguments to single strings
  desc <- lapply(desc, function(x) paste(x, collapse = ", "))

  desc
}

#' Run a system command and check if it succeeds.
#'
#' @param cmd Command to run. Will be quoted by \code{\link{shQuote}()}.
#' @param args A character vector of arguments.
#' @param env_vars A named character vector of environment variables.
#' @param path Path in which to execute the command
#' @param quiet If \code{FALSE}, the command to be run will be echoed.
#' @param throw If \code{TRUE}, will throw an error if the command fails
#'   (i.e. the return value is not 0).
#' @param ... additional arguments passed to \code{\link[base]{system}}
#' @keywords internal
#' @export
#' @return The exit status of the command, invisibly.
system_check <- function(cmd, args = character(), env_vars = character(),
                         path = ".", quiet = FALSE, throw = TRUE,
                         ...) {
  full <- paste(shQuote(cmd), " ", paste(args, collapse = " "), sep = "")

  if (!quiet) {
    message(wrap_command(full))
    message()
  }

  result <- suppressWarnings(withr::with_dir(path, withr::with_envvar(env_vars,
    system(full, intern = quiet, ignore.stderr = quiet, ...)
  )))

  if (quiet) {
    status <- attr(result, "status") %||% 0L
  } else {
    status <- result
  }

  ok <- identical(as.character(status), "0")
  if (throw && !ok) {
    stop("Command failed (", status, ")", call. = FALSE)
  }

  invisible(status)
}

#' @noRd
#' @param out_file Path of file to which output is written if \code{quiet} is
#'   \code{TRUE}
system2_check <- function(cmd, args = character(), env_vars = character(),
                          path = ".", quiet = FALSE, throw = TRUE,
                          out_file = NULL, ...) {
  full <- paste(shQuote(cmd), " ", paste(args, collapse = " "), sep = "")

  if (!quiet) {
    message(wrap_command(full))
    message()
  }

  if (quiet)
    std <- TRUE
  else
    std <- ""

  result <- suppressWarnings(withr::with_dir(path, withr::with_envvar(env_vars,
    system2(cmd, args, stdout = std, stderr = std, ...)
  )))

  if (quiet) {
    if (!is.null(out_file)) {
      writeLines(result, out_file)
    }

    status <- attr(result, "status") %||% 0L
  } else {
    status <- result
  }

  ok <- identical(as.character(status), "0")
  if (throw && !ok) {
    stop("Command failed (", status, ")", call. = FALSE)
  }

  invisible(status)
}

#' Run a system command and capture the output.
#'
#' @inheritParams system_check
#' @param ... additional arguments passed to \code{\link[base]{system}}
#' @return command output if the command succeeds, an error will be thrown if
#'   the command fails.
#' @keywords internal
#' @export
system_output <- function(cmd, args = character(), env_vars = character(),
                          path = ".", quiet = FALSE, ...) {
  full <- paste(shQuote(cmd), " ", paste(args, collapse = " "), sep = "")

  if (!quiet) {
    message(wrap_command(full), "\n")
  }

  result <- withCallingHandlers(withr::with_dir(path,
      withr::with_envvar(env_vars,
        system(full, intern = TRUE, ignore.stderr = quiet, ...)
        )), warning = function(w) stop(w))

  result
}

wrap_command <- function(x) {
  lines <- strwrap(x, getOption("width") - 2, exdent = 2)
  continue <- c(rep(" \\", length(lines) - 1), "")
  paste(lines, continue, collapse = "\n")
}

#' Parse package dependency strings.
#'
#' @param string to parse. Should look like \code{"R (>= 3.0), ggplot2"} etc.
#' @return list of two character vectors: \code{name} package names,
#'   and \code{version} package versions. If version is not specified,
#'   it will be stored as NA.
#' @keywords internal
#' @export
#' @examples
#' parse_deps("httr (< 2.1),\nRCurl (>= 3)")
#' # only package dependencies are returned
#' parse_deps("utils (== 2.12.1),\ntools,\nR (>= 2.10),\nmemoise")
parse_deps <- function(string) {
  if (is.null(string)) return()
  stopifnot(is.character(string), length(string) == 1)
  if (grepl("^\\s*$", string)) return()

  pieces <- strsplit(string, "[[:space:]]*,[[:space:]]*")[[1]]

  # Get the names
  names <- gsub("\\s*\\(.*?\\)", "", pieces)
  names <- gsub("^\\s+|\\s+$", "", names)

  # Get the versions and comparison operators
  versions_str <- pieces
  have_version <- grepl("\\(.*\\)", versions_str)
  versions_str[!have_version] <- NA

  compare  <- sub(".*\\((\\S+)\\s+.*\\)", "\\1", versions_str)
  versions <- sub(".*\\(\\S+\\s+(.*)\\)", "\\1", versions_str)

  # Check that non-NA comparison operators are valid
  compare_nna   <- compare[!is.na(compare)]
  compare_valid <- compare_nna %in% c(">", ">=", "==", "<=", "<")
  if(!all(compare_valid)) {
    stop("Invalid comparison operator in dependency: ",
      paste(compare_nna[!compare_valid], collapse = ", "))
  }

  deps <- data.frame(name = names, compare = compare,
    version = versions, stringsAsFactors = FALSE)

  # Remove R dependency
  deps[names != "R", ]
}


#' Check that the version of an imported package satisfies the requirements
#'
#' @param dep_name The name of the package with objects to import
#' @param dep_ver The version of the package
#' @param dep_compare The comparison operator to use to check the version
#' @keywords internal
check_dep_version <- function(dep_name, dep_ver = NA, dep_compare = NA) {
  if (!requireNamespace(dep_name, quietly = TRUE)) {
    stop("Dependency package ", dep_name, " not available.")
  }

  if (xor(is.na(dep_ver), is.na(dep_compare))) {
    stop("dep_ver and dep_compare must be both NA or both non-NA")

  } else if(!is.na(dep_ver) && !is.na(dep_compare)) {

    compare <- match.fun(dep_compare)
    if (!compare(
      as.numeric_version(getNamespaceVersion(dep_name)),
      as.numeric_version(dep_ver))) {

      warning("Need ", dep_name, " ", dep_compare,
        " ", dep_ver,
        " but loaded version is ", getNamespaceVersion(dep_name))
    }
  }
  return(TRUE)
}

#' Find missing s3 exports.
#'
#' The method is heuristic - looking for objs with a period in their name.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @export
missing_s3 <- function(pkg = ".") {
  pkg <- as.package(pkg)
  check_suggested("roxygen2")
  loaded <- load_all(pkg)

  # Find all S3 methods in package
  objs <- ls(envir = loaded$env)
  is_s3 <- function(x) roxygen2::is_s3_method(x, env = loaded$env)
  s3_objs <- Filter(is_s3, objs)

  # Find all S3 methods in NAMESPACE
  ns <- parse_ns_file(pkg)
  exports <- paste(ns$S3methods[, 1], ns$S3methods[, 2], sep = ".")

  setdiff(s3_objs, exports)
}

#' Release package to CRAN.
#'
#' Run automated and manual tests, then ftp to CRAN.
#'
#' The package release process will:
#'
#' \itemize{
#'
#'   \item Confirm that the package passes \code{R CMD check}
#'   \item Ask if you've checked your code on win-builder
#'   \item Confirm that news is up-to-date
#'   \item Confirm that DESCRIPTION is ok
#'   \item Ask if you've checked packages that depend on your package
#'   \item Build the package
#'   \item Submit the package to CRAN, using comments in "cran-comments.md"
#' }
#'
#' You can also add arbitrary extra questions by defining an (un-exported)
#' function called \code{release_questions()} that returns a character vector
#' of additional questions to ask.
#'
#' You also need to read the CRAN repository policy at
#' \url{https://cran.r-project.org/web/packages/policies.html} and make
#' sure you're in line with the policies. \code{release} tries to automate as
#' many of polices as possible, but it's impossible to be completely
#' comprehensive, and they do change in between releases of devtools.
#'
#' @section Guarantee:
#'
#' If a devtools bug causes one of the CRAN maintainers to treat you
#' impolitely, I will personally send you a handwritten apology note.
#' Please forward me the email and your address, and I'll get a card in
#' the mail.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @param check if \code{TRUE}, run checking, otherwise omit it.  This
#'   is useful if you've just checked your package and you're ready to
#'   release it.
#' @param args An optional character vector of additional command
#'   line arguments to be passed to \code{R CMD build}.
#' @param spelling language or dictionary file to spell check documentation.
#' See \code{\link{spell_check}}. Set to \code{NULL} to skip spell checking.
#' @export
release <- function(pkg = ".", check = TRUE, args = NULL, spelling = "en_US") {
  pkg <- as.package(pkg)
  # Figure out if this is a new package
  cran_version <- cran_pkg_version(pkg$package)
  new_pkg <- is.null(cran_version)

  dr_d <- dr_devtools()
  if (!dr_d) {
    print(dr_d)

    if (yesno("Proceed anyway?"))
      return(invisible())
  }

  if (uses_git(pkg$path)) {
    git_checks(pkg)
    if (yesno("Were Git checks successful?"))
      return(invisible())
  }

  if (length(spelling)) {
    cat("Spell checking documentation...\n")
    print(spell_check(pkg, dict = spelling))
    cat("\n")
    if (yesno("Is documentation free of spelling errors? (you can ignore false positives)"))
      return(invisible())
  }

  if (check) {
    rule("Building and checking ", pkg$package, pad = "=")
    check(pkg, cran = TRUE, check_version = TRUE, manual = TRUE,
          build_args = args, run_dont_test = TRUE)
  }
  if (yesno("Was R CMD check successful?"))
    return(invisible())

  release_checks(pkg)
  if (yesno("Were devtool's checks successful?"))
    return(invisible())

  if (!new_pkg) {
    cran_url <- paste0(cran_mirror(), "/web/checks/check_results_",
      pkg$package, ".html")
    if (yesno("Have you fixed all existing problems at \n", cran_url, " ?"))
      return(invisible())
  }

  if (has_src(pkg)) {
    if (yesno("Have you run R CMD check with valgrind?"))
      return(invisible())
  }

  deps <- if (new_pkg) 0 else length(revdep(pkg$package))
  if (deps > 0) {
    msg <- paste0("Have you checked the ", deps ," packages that depend on ",
      "this package (with revdep_check())?")

    if (yesno(msg))
      return(invisible())
  }

  if (yesno("Have you checked on win-builder (with build_win())?"))
    return(invisible())

  if (yesno("Have you updated your NEWS file?"))
    return(invisible())

  rule("DESCRIPTION")
  cat(readLines(file.path(pkg$path, "DESCRIPTION")), sep = "\n")
  cat("\n")
  if (yesno("Is DESCRIPTION up-to-date?"))
    return(invisible())

  release_questions <- pkg_env(pkg)$release_questions
  if (!is.null(release_questions)) {
    questions <- release_questions()
    for (question in questions) {
      if (yesno(question)) return(invisible())
    }
  }

  rule("cran-comments.md")
  cat(cran_comments(pkg), "\n\n")
  if (yesno("Are the CRAN submission comments correct?"))
    return(invisible())

  if (yesno("Is your email address ", maintainer(pkg)$email, "?"))
    return(invisible())

  built_path <- build_cran(pkg, args = args)
  if (yesno("Ready to submit?"))
    return(invisible())

  upload_cran(pkg, built_path)

  if (uses_git(pkg$path)) {
    message("Don't forget to tag the release when the package is accepted!")
  }
  invisible(TRUE)
}

release_email <- function(name, new_pkg) {
  paste(
    "Dear CRAN maintainers,\n",
    "\n",
    if (new_pkg) {
      paste("I have uploaded a new package, ", name, ", to CRAN. ",
        "I have read and agree to the CRAN policies.\n", sep = "")
    } else {
      paste("I have just uploaded a new version of ", name, " to CRAN.\n",
        sep = "")
    },
    "\n",
    "Thanks!\n",
    "\n",
    getOption("devtools.name"), "\n",
    sep = "")
}

yesno <- function(...) {
  yeses <- c("Yes", "Definitely", "For sure", "Yup", "Yeah", "I agree", "Absolutely")
  nos <- c("No way", "Not yet", "I forget", "No", "Nope", "Uhhhh... Maybe?")

  cat(paste0(..., collapse = ""))
  qs <- c(sample(yeses, 1), sample(nos, 2))
  rand <- sample(length(qs))

  menu(qs[rand]) != which(rand == 1)
}

# http://tools.ietf.org/html/rfc2368
email <- function(address, subject, body) {
  url <- paste(
    "mailto:",
    utils::URLencode(address),
    "?subject=", utils::URLencode(subject),
    "&body=", utils::URLencode(body),
    sep = ""
  )

  tryCatch({
    utils::browseURL(url, browser = email_browser())},
    error = function(e) {
      message("Sending failed with error: ", e$message)
      cat("To: ", address, "\n", sep = "")
      cat("Subject: ", subject, "\n", sep = "")
      cat("\n")
      cat(body, "\n", sep = "")
    }
  )

  invisible(TRUE)
}

email_browser <- function() {
  if (!identical(.Platform$GUI, "RStudio"))
    return (getOption("browser"))

  # Use default browser, even if RStudio running
  if (.Platform$OS.type == "windows")
    return (NULL)

  browser <- Sys.which(c("xdg-open", "open"))
  browser[nchar(browser) > 0][[1]]
}


maintainer <- function(pkg = ".") {
  pkg <- as.package(pkg)

  authors <- pkg$`authors@r`
  if (!is.null(authors)) {
    people <- eval(parse(text = authors))
    if (is.character(people)) {
      maintainer <- utils::as.person(people)
    } else {
      maintainer <- Find(function(x) "cre" %in% x$role, people)
    }
  } else {
    maintainer <- pkg$maintainer
    if (is.null(maintainer)) {
      stop("No maintainer defined in package.", call. = FALSE)
    }
    maintainer <- utils::as.person(maintainer)
  }

  list(
    name = paste(maintainer$given, maintainer$family),
    email = maintainer$email
  )
}

cran_comments <- function(pkg = ".") {
  pkg <- as.package(pkg)

  path <- file.path(pkg$path, "cran-comments.md")
  if (!file.exists(path)) {
    warning("Can't find cran-comments.md.\n",
      "This file gives CRAN volunteers comments about the submission,\n",
      "and it must exist. Create it with use_cran_comments().\n",
      call. = FALSE)
    return(character())
  }

  paste0(readLines(path, warn = FALSE), collapse = "\n")
}

cran_submission_url <- "http://xmpalantir.wu.ac.at/cransubmit/index2.php"

#' Submit a package to CRAN.
#'
#' This uses the new CRAN web-form submission process. After submission, you
#' will receive an email asking you to confirm submission - this is used
#' to check that the package is submitted by the maintainer.
#'
#' It's recommended that you use \code{\link{release}()} rather than this
#' function as it performs more checks prior to submission.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @inheritParams release
#' @export
#' @keywords internal
submit_cran <- function(pkg = ".", args = NULL) {
  built_path <- build_cran(pkg, args = args)
  upload_cran(pkg, built_path)
}

build_cran <- function(pkg, args) {
  message("Building")
  built_path <- build(pkg, tempdir(), manual = TRUE, args = args)
  message("Submitting file: ", built_path)
  message("File size: ",
          format(as.object_size(file.info(built_path)$size), units = "auto"))
  built_path
}

upload_cran <- function(pkg, built_path) {
  pkg <- as.package(pkg)
  maint <- maintainer(pkg)
  comments <- cran_comments(pkg)

  # Initial upload ---------
  message("Uploading package & comments")
  body <- list(
    pkg_id = "",
    name = maint$name,
    email = maint$email,
    uploaded_file = httr::upload_file(built_path, "application/x-gzip"),
    comment = comments,
    upload = "Upload package"
  )
  r <- httr::POST(cran_submission_url, body = body)
  httr::stop_for_status(r)
  new_url <- httr::parse_url(r$url)
  new_url$query$strErr

  # Confirmation -----------
  message("Confirming submission")
  body <- list(
    pkg_id = new_url$query$pkg_id,
    name = maint$name,
    email = maint$email,
    policy_check = "1/",
    submit = "Submit package"
  )
  r <- httr::POST(cran_submission_url, body = body)
  httr::stop_for_status(r)
  new_url <- httr::parse_url(r$url)
  if (new_url$query$submit == "1") {
    message("Package submission successful.\n",
      "Check your email for confirmation link.")
  } else {
    stop("Package failed to upload.", call. = FALSE)
  }

  invisible(TRUE)
}

as.object_size <- function(x) structure(x, class = "object_size")

#' Lint all source files in a package.
#'
#' The default lintings correspond to the style guide at
#' \url{http://r-pkgs.had.co.nz/r.html#style}, however it is possible to
#' override any or all of them using the \code{linters} parameter.
#' @param pkg package description, can be path or package name. See
#'   \code{\link{as.package}} for more information
#' @param cache store the lint results so repeated lints of the same content
#' use the previous results.
#' @param ... additional arguments passed to \code{\link[lintr]{lint_package}}
#' @seealso \code{\link[lintr]{lint_package}}, \code{\link[lintr]{lint}}
#' @details
#' The lintr cache is by default stored in \code{~/.R/lintr_cache/} (this can
#' be configured by setting \code{options(lintr.cache_directory)}).  It can be
#' cleared by calling \code{\link[lintr]{clear_cache}}.
#' @export
lint <- function(pkg = ".", cache = TRUE, ...) {
  check_suggested("lintr")
  pkg <- as.package(pkg)

  message("Linting ", pkg$package, appendLF = FALSE)
  lintr::lint_package(pkg$path, cache = cache, ...)
}

#' Creates a new package, following all devtools package conventions.
#'
#' Similar to \code{\link{package.skeleton}}, except that it only creates
#' the standard devtools directory structures; it doesn't try and create
#' source code and data files by inspecting the global environment.
#'
#' \code{create} requires that the directory doesn't exist yet; it will be
#' created but deleted upon failure. \code{setup} assumes an existing
#' directory from which it will infer the package name.
#'
#' @param path location to create new package.  The last component of the path
#'   will be used as the package name.
#' @param description list of description values to override default values or
#'   add additional values.
#' @param check if \code{TRUE}, will automatically run \code{\link{check}}
#' @param rstudio Create an RStudio project file?
#'   (with \code{\link{use_rstudio}})
#' @param quiet if \code{FALSE}, the default, prints informative messages.
#' @seealso Text with \code{\link{package.skeleton}}
#' @export
#' @examples
#' \dontrun{
#' # Create a package using all defaults:
#' path <- file.path(tempdir(), "myDefaultPackage")
#' create(path)
#'
#' # Override a description attribute.
#' path <- file.path(tempdir(), "myCustomPackage")
#' my_description <- list("Maintainer" =
#'   "'Yoni Ben-Meshulam' <yoni@@opower.com>")
#' create(path, my_description)
#' }
create <- function(path, description = getOption("devtools.desc"),
                   check = FALSE, rstudio = TRUE, quiet = FALSE) {
  check_package_name(path)

  # ensure the parent directory exists
  parent_dir <- normalizePath(dirname(path), winslash = "/", mustWork = FALSE)
  if (!file.exists(parent_dir)) {
    stop("Parent directory '", parent_dir, "' does not exist", call. = FALSE)
  }

  # allow 'create' to create a new directory, or populate
  # an empty directory, as long as the parent directory exists
  if (!file.exists(path)) {
    if (!dir.create(path)) {
      stop("Failed to create package directory '", basename(path), "'",
           call. = FALSE)
    }
  }

  # if the directory exists but is not empty, bail
  files <- list.files(path)
  if (length(files)) {

    valid <- length(files) == 1 && tools::file_ext(files) == "Rproj"
    if (!valid)
      stop("Directory exists and is not empty", call. = FALSE)
  }

  path <- normalizePath(path, winslash = "/", mustWork = TRUE)
  setup(path = path, description = description, rstudio = rstudio,
        check = check, quiet = quiet)

  invisible(TRUE)
}

#' @rdname create
#' @export
setup <- function(path = ".", description = getOption("devtools.desc"),
                  check = FALSE, rstudio = TRUE, quiet = FALSE) {
  check_package_name(path)

  parent_dir <- normalizePath(dirname(path), winslash = "/", mustWork = TRUE)
  if (!quiet) {
    message("Creating package '", extract_package_name(path), "' in '", parent_dir, "'")
  }

  dir.create(file.path(path, "R"), showWarnings = FALSE)
  create_description(path, extra = description, quiet = quiet)
  create_namespace(path)

  if (rstudio) use_rstudio(path)

  if (check) check(path)
  invisible(TRUE)
}

extract_package_name <- function(path) {
  basename(normalizePath(path, mustWork = FALSE))
}

check_package_name <- function(path) {
  name <- extract_package_name(path)
  if (!valid_name(name)) {
    stop(
      name, " is not a valid package name: it should contain only\n",
      "ASCII letters, numbers and dot, have at least two characters\n",
      "and start with a letter and not end in a dot.",
      call. = FALSE
    )
  }
}


valid_name <- function(x) {
  grepl("^[[:alpha:]][[:alnum:].]+$", x) && !grepl("\\.$", x)
}

create_namespace <- function(path) {
  ns_path <- file.path(path, "NAMESPACE")
  if (file.exists(ns_path)) return()

  cat(
    '# Generated by roxygen2: fake comment so roxygen2 overwrites silently.\n',
    'exportPattern("^[^\\\\.]")\n',
    sep = "",
    file = ns_path
  )
}

#' Install a package from a SVN repository
#'
#' This function requires \code{svn} to be installed on your system in order to
#' be used.
#'
#' It is vectorised so you can install multiple packages with
#' a single command.
#'
#' @inheritParams install_git
#' @param subdir A sub-directory withing a svn repository that may contain the
#'   package we are interested in installing. By default, this
#'   points to the 'trunk' directory.
#' @param args A character vector providing extra arguments to pass on to
#    svn.
#' @param revision svn revision, if omitted updates to latest
#' @param branch Name of branch or tag to use, if not trunk.
#' @param ... Other arguments passed on to \code{\link{install}}
#' @export
#' @family package installation
#' @examples
#' \dontrun{
#' install_svn("https://github.com/hadley/stringr")
#' install_svn("https://github.com/hadley/httr", branch = "oauth")
#'}
install_svn <- function(url, subdir = NULL, branch = NULL, args = character(0),
  ..., revision = NULL, quiet = FALSE) {

  remotes <- lapply(url, svn_remote, svn_subdir = subdir, branch = branch,
    revision = revision, args = args)

  install_remotes(remotes, ..., quiet = quiet)
}

svn_remote <- function(url, svn_subdir = NULL, branch = NULL, revision = NULL,
  args = character(0)) {
  remote("svn",
    url = url,
    svn_subdir = svn_subdir,
    branch = branch,
    revision = revision,
    args = args
  )
}

#' @export
remote_download.svn_remote <- function(x, quiet = FALSE) {
  if (!quiet) {
    message("Downloading svn repo ", x$url)
  }

  bundle <- tempfile()
  svn_binary_path <- svn_path()

  args <- c("co", x$args, full_svn_url(x), bundle)

  message(shQuote(svn_binary_path), " ", paste0(args, collapse = " "))
  request <- system2(svn_binary_path, args, stdout = FALSE, stderr = FALSE)

  # This is only looking for an error code above 0-success
  if (request > 0) {
    stop("There seems to be a problem retrieving this SVN-URL.", call. = FALSE)
  }

  withr::with_dir(bundle, {
    if (!is.null(x$revision)) {
      request <- system2(svn_binary_path, paste("update -r", x$revision))
      if (request > 0) {
        stop("There was a problem switching to the requested SVN revision", call. = FALSE)
      }
    }
  })

  bundle
}

#' @export
remote_metadata.svn_remote <- function(x, bundle = NULL, source = NULL) {

  if (!is.null(bundle)) {
    withr::with_dir(bundle, {
      revision <- svn_revision()
    })
  } else {
    revision <- NULL
  }

  list(
    RemoteType = "svn",
    RemoteUrl = x$url,
    RemoteSvnSubdir = x$svn_subdir,
    RemoteBranch = x$branch,
    RemoteArgs = if (length(x$args) > 0) paste0(deparse(x$args), collapse = " "),
    RemoteRevision = revision,
    RemoteSha = revision # for compatibility with other remotes
  )
}

svn_path <- function(svn_binary_name = NULL) {
  # Use user supplied path
  if (!is.null(svn_binary_name)) {
    if (!file.exists(svn_binary_name)) {
      stop("Path ", svn_binary_name, " does not exist", .call = FALSE)
    }
    return(svn_binary_name)
  }

  # Look on path
  svn_path <- Sys.which("svn")[[1]]
  if (svn_path != "") return(svn_path)

  # On Windows, look in common locations
  if (.Platform$OS.type == "windows") {
    look_in <- c(
      "C:/Program Files/Svn/bin/svn.exe",
      "C:/Program Files (x86)/Svn/bin/svn.exe"
    )
    found <- file.exists(look_in)
    if (any(found)) return(look_in[found][1])
  }

  stop("SVN does not seem to be installed on your system.", call. = FALSE)
}

#' @export
remote_package_name.svn_remote <- function(remote, ...) {
  description_url <- file.path(full_svn_url(remote), "DESCRIPTION")
  tmp_file <- tempfile()
  on.exit(rm(tmp_file))
  response <- system2(svn_path(), paste("cat", description_url), stdout = tmp_file)
  if (!identical(response, 0L)) {
    stop("There was a problem retrieving the current SVN revision", call. = FALSE)
  }
  read_dcf(tmp_file)$Package
}

#' @export
remote_sha.svn_remote <- function(remote, ...) {
  svn_revision(full_svn_url(remote))
}

svn_revision <- function(url = NULL, svn_binary_path = svn_path()) {
  request <- system2(svn_binary_path, paste("info --xml", url), stdout = TRUE)
  if (!is.null(attr(request, "status")) && !identical(attr(request, "status"), 0L)) {
    stop("There was a problem retrieving the current SVN revision", call. = FALSE)
  }
  gsub(".*<commit[[:space:]]+revision=\"([[:digit:]]+)\">.*", "\\1", paste(collapse = "\n", request))
}

full_svn_url <- function(x) {
  if (!is.null(x$branch)) {
    url <- file.path(x$url, "branches", x$branch)
  } else {
    url <- file.path(x$url, "trunk")
  }
  if (!is.null(x$svn_subdir)) {
    url <- file.path(url, x$svn_subdir)
  }

  url
}

format.svn_remote <- function(x, ...) {
  "SVN"
}

#' Coerce input to a package.
#'
#' Possible specifications of package:
#' \itemize{
#'   \item path
#'   \item package object
#' }
#' @param x object to coerce to a package
#' @param create only relevant if a package structure does not exist yet: if
#'   \code{TRUE}, create a package structure; if \code{NA}, ask the user
#'   (in interactive mode only)
#' @export
#' @keywords internal
as.package <- function(x = NULL, create = NA) {
  if (is.package(x)) return(x)

  x <- package_file(path = x)
  load_pkg_description(x, create = create)
}

#' Find file in a package.
#'
#' It always starts by finding by walking up the path until it finds the
#' root directory, i.e. a directory containing \code{DESCRIPTION}. If it
#' cannot find the root directory, or it can't find the specified path, it
#' will throw an error.
#'
#' @param ... Components of the path.
#' @param path Place to start search for package directory.
#' @export
#' @examples
#' \dontrun{
#' package_file("figures", "figure_1")
#' }
package_file <- function(..., path = ".") {
  if (!is.character(path) || length(path) != 1) {
    stop("`path` must be a string.", call. = FALSE)
  }
  path <- strip_slashes(normalizePath(path, mustWork = FALSE))

  if (!file.exists(path)) {
    stop("Can't find '", path, "'.", call. = FALSE)
  }
  if (!file.info(path)$isdir) {
    stop("'", path, "' is not a directory.", call. = FALSE)
  }

  # Walk up to root directory
  while (!has_description(path)) {
    path <- dirname(path)

    if (is_root(path)) {
      stop("Could not find package root.", call. = FALSE)
    }
  }

  file.path(path, ...)
}

has_description <- function(path) {
  file.exists(file.path(path, 'DESCRIPTION'))
}

is_root <- function(path) {
  identical(path, dirname(path))
}

strip_slashes <- function(x) {
  x <- sub("/*$", "", x)
  x
}

# Load package DESCRIPTION into convenient form.
load_pkg_description <- function(path, create) {
  path_desc <- file.path(path, "DESCRIPTION")

  if (!file.exists(path_desc)) {
    if (is.na(create)) {
      if (interactive()) {
        message("No package infrastructure found in ", path, ". Create it?")
        create <- (menu(c("Yes", "No")) == 1)
      } else {
        create <- FALSE
      }
    }

    if (create) {
      setup(path = path)
    } else {
      stop("No description at ", path_desc, call. = FALSE)
    }
  }

  desc <- as.list(read.dcf(path_desc)[1, ])
  names(desc) <- tolower(names(desc))
  desc$path <- path

  structure(desc, class = "package")
}


#' Is the object a package?
#'
#' @keywords internal
#' @export
is.package <- function(x) inherits(x, "package")

# Mockable variant of interactive
interactive <- function() .Primitive("interactive")()

#' Get/set the PATH variable.
#'
#' @param path character vector of paths
#' @return \code{set_path} invisibly returns the old path.
#' @name path
#' @family path
#' @seealso \code{\link[withr]{with_path}} to temporarily set the path for a block
#'   of code
#' @examples
#' path <- get_path()
#' length(path)
#' old <- add_path(".")
#' length(get_path())
#' set_path(old)
#' length(get_path())
NULL

#' @export
#' @rdname path
get_path <- function() {
  strsplit(Sys.getenv("PATH"), .Platform$path.sep)[[1]]
}

#' @export
#' @rdname path
set_path <- function(path) {
  path <- normalizePath(path, mustWork = FALSE)

  old <- get_path()
  path <- paste(path, collapse = .Platform$path.sep)
  Sys.setenv(PATH = path)
  invisible(old)
}

#' @export
#' @rdname path
#' @param after for \code{add_path}, the place on the PATH where the new paths
#'   should be added
add_path <- function(path, after = Inf) {
  set_path(append(get_path(), path, after))
}


#' Test if an object is on the path.
#'
#' @param ... Strings indicating the executables to check for on the path.
#' @family path
#' @keywords internal
#' @export
#' @examples
#' on_path("R")
#' on_path("gcc")
#' on_path("foo", "bar")  # FALSE in most cases
#' withr::with_path(tempdir(), on_path("gcc"))
on_path <- function(...) {
  commands <- c(...)
  stopifnot(is.character(commands))
  unname(Sys.which(commands) != "")
}

#' Build package.
#'
#' Building converts a package source directory into a single bundled file.
#' If \code{binary = FALSE} this creates a \code{tar.gz} package that can
#' be installed on any platform, provided they have a full development
#' environment (although packages without source code can typically be
#' install out of the box). If \code{binary = TRUE}, the package will have
#' a platform specific extension (e.g. \code{.zip} for windows), and will
#' only be installable on the current platform, but no development
#' environment is needed.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @param path path in which to produce package.  If \code{NULL}, defaults to
#'   the parent directory of the package.
#' @param binary Produce a binary (\code{--binary}) or source (
#'   \code{--no-manual --no-resave-data}) version of the package.
#' @param vignettes,manual For source packages: if \code{FALSE}, don't build PDF
#'   vignettes (\code{--no-build-vignettes}) or manual (\code{--no-manual}).
#' @param args An optional character vector of additional command
#'   line arguments to be passed to \code{R CMD build} if \code{binary = FALSE},
#'   or \code{R CMD install} if \code{binary = TRUE}.
#' @param quiet if \code{TRUE} suppresses output from this function.
#' @export
#' @family build functions
#' @return a string giving the location (including file name) of the built
#'  package
build <- function(pkg = ".", path = NULL, binary = FALSE, vignettes = TRUE,
                  manual = FALSE, args = NULL, quiet = FALSE) {
  pkg <- as.package(pkg)
  if (is.null(path)) {
    path <- dirname(pkg$path)
  }

  check_build_tools(pkg)
  compile_rcpp_attributes(pkg)

  if (binary) {
    args <- c("--build", args)
    cmd <- paste0("CMD INSTALL ", shQuote(pkg$path), " ",
      paste0(args, collapse = " "))
    if (.Platform$OS.type == "windows") {
      ext <- ".zip"
    } else if (grepl("darwin", R.version$os)) {
      ext <- ".tgz"
    } else {
      ext <- paste0("_R_", Sys.getenv("R_PLATFORM"), ".tar.gz")
    }
  } else {
    args <- c(args, "--no-resave-data")

    if (manual && !has_latex(verbose = TRUE)) {
      manual <- FALSE
    }

    if (!manual) {
      args <- c(args, "--no-manual")
    }

    if (!vignettes) {
      args <- c(args, "--no-build-vignettes")
    }

    cmd <- paste0("CMD build ", shQuote(pkg$path), " ",
      paste0(args, collapse = " "))

    ext <- ".tar.gz"
  }

  # Run in temporary library to ensure that default library doesn't get
  # contaminated
  withr::with_temp_libpaths(R(cmd, path, quiet = quiet))
  targz <- paste0(pkg$package, "_", pkg$version, ext)

  file.path(path, targz)
}


#' Build windows binary package.
#'
#' This function works by bundling source package, and then uploading to
#' \url{http://win-builder.r-project.org/}.  Once building is complete you'll
#' receive a link to the built package in the email address listed in the
#' maintainer field.  It usually takes around 30 minutes. As a side effect,
#' win-build also runs \code{R CMD check} on the package, so \code{build_win}
#' is also useful to check that your package is ok on windows.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @inheritParams build
#' @param version directory to upload to on the win-builder, controlling
#'   which version of R is used to build the package. Possible options are
#'   listed on \url{http://win-builder.r-project.org/}. Defaults to R-devel.
#' @export
#' @family build functions
build_win <- function(pkg = ".", version = c("R-release", "R-devel"),
                      args = NULL, quiet = FALSE) {
  pkg <- as.package(pkg)

  if (missing(version)) {
    version <- "R-devel"
  } else {
    version <- match.arg(version, several.ok = TRUE)
  }

  if (!quiet) {
    message("Building windows version of ", pkg$package,
            " for ", paste(version, collapse = ", "),
            " with win-builder.r-project.org.\n")
    if (interactive() && yesno("Email results to ", maintainer(pkg)$email, "?")) {
      return(invisible())
    }
  }

  built_path <- build(pkg, tempdir(), args = args, quiet = quiet)
  on.exit(unlink(built_path))

  url <- paste0("ftp://win-builder.r-project.org/", version, "/",
                basename(built_path))
  lapply(url, upload_ftp, file = built_path)

  if (!quiet) {
    message("Check ", maintainer(pkg)$email, " for a link to the built package",
            if (length(version) > 1) "s" else "",
            " in 30-60 mins.")
  }

  invisible()
}

#' Activate and deactivate development mode.
#'
#' When activated, \code{dev_mode} creates a new library for storing installed
#' packages. This new library is automatically created when \code{dev_mode} is
#' activated if it does not already exist.
#' This allows you to test development packages in a sandbox, without
#' interfering with the other packages you have installed.
#'
#' @param on turn dev mode on (\code{TRUE}) or off (\code{FALSE}).  If omitted
#'  will guess based on whether or not \code{path} is in
#'  \code{\link{.libPaths}}
#' @param path directory to library.
#' @export
#' @examples
#' \dontrun{
#' dev_mode()
#' dev_mode()
#' }
dev_mode <- local({
  .prompt <- NULL

  function(on = NULL, path = getOption("devtools.path")) {
    lib_paths <- .libPaths()

    path <- normalizePath(path, winslash = "/", mustWork = FALSE)
    if (is.null(on)) {
      on <- !(path %in% lib_paths)
    }

    if (on) {
      if (!file.exists(path)) {
        dir.create(path, recursive = TRUE, showWarnings = FALSE)
      }
      if (!file.exists(path)) {
        stop("Failed to create ", path, call. = FALSE)
      }

      if (!is_library(path)) {
        warning(path, " does not appear to be a library. ",
          "Are sure you specified the correct directory?", call. = FALSE)
      }

      message("Dev mode: ON")
      options(dev_path = path)

      if (is.null(.prompt)) .prompt <<- getOption("prompt")
      options(prompt = paste("d> "))

      .libPaths(c(path, lib_paths))
    } else {

      message("Dev mode: OFF")
      options(dev_path = NULL)

      if (!is.null(.prompt)) options(prompt = .prompt)
      .prompt <<- NULL

      .libPaths(setdiff(lib_paths, path))
    }
  }
})

is_library <- function(path) {
  # empty directories can be libraries
  if (length(dir(path)) == 0) return (TRUE)

  # otherwise check that the directories are compiled R directories -
  # i.e. that they contain a Meta directory
  dirs <- dir(path, full.names = TRUE)
  dirs <- dirs[utils::file_test("-d", dirs)]

  has_pkg_dir <- function(path) length(dir(path, pattern = "Meta")) > 0
  help_dirs <- vapply(dirs, has_pkg_dir, logical(1))

  all(help_dirs)
}

#' Install a local development package.
#'
#' Uses \code{R CMD INSTALL} to install the package. Will also try to install
#' dependencies of the package from CRAN, if they're not already installed.
#'
#' By default, installation takes place using the current package directory.
#' If you have compiled code, this means that artefacts of compilation will be
#' created in the \code{src/} directory. If you want to avoid this, you can
#' use \code{local = FALSE} to first build a package bundle and then install
#' it from a temporary directory. This is slower, but keeps the source
#' directory pristine.
#'
#' If the package is loaded, it will be reloaded after installation. This is
#' not always completely possible, see \code{\link{reload}} for caveats.
#'
#' To install a package in a non-default library, use \code{\link[withr]{with_libpaths}}.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @param reload if \code{TRUE} (the default), will automatically reload the
#'   package after installing.
#' @param quick if \code{TRUE} skips docs, multiple-architectures,
#'   demos, and vignettes, to make installation as fast as possible.
#' @param local if \code{FALSE} \code{\link{build}}s the package first:
#'   this ensures that the installation is completely clean, and prevents any
#'   binary artefacts (like \file{.o}, \code{.so}) from appearing in your local
#'   package directory, but is considerably slower, because every compile has
#'   to start from scratch.
#' @param args An optional character vector of additional command line
#'   arguments to be passed to \code{R CMD install}. This defaults to the
#'   value of the option \code{"devtools.install.args"}.
#' @param quiet if \code{TRUE} suppresses output from this function.
#' @param dependencies \code{logical} indicating to also install uninstalled
#'   packages which this \code{pkg} depends on/links to/suggests. See
#'   argument \code{dependencies} of \code{\link{install.packages}}.
#' @param upgrade_dependencies If \code{TRUE}, the default, will also update
#'   any out of date dependencies.
#' @param build_vignettes if \code{TRUE}, will build vignettes. Normally it is
#'   \code{build} that's responsible for creating vignettes; this argument makes
#'   sure vignettes are built even if a build never happens (i.e. because
#'   \code{local = TRUE}).
#' @param keep_source If \code{TRUE} will keep the srcrefs from an installed
#'   package. This is useful for debugging (especially inside of RStudio).
#'   It defaults to the option \code{"keep.source.pkgs"}.
#' @param threads number of concurrent threads to use for installing
#'   dependencies.
#'   It defaults to the option \code{"Ncpus"} or \code{1} if unset.
#' @param force_deps whether to force installation of dependencies even if their
#'   SHA1 reference hasn't changed from the currently installed version.
#' @param metadata Named list of metadata entries to be added to the
#'   \code{DESCRIPTION} after installation.
#' @param out_dir Directory to store installation output in case of failure.
#' @param skip_if_log_exists If the \code{out_dir} is defined and contains
#'   a file named \code{package.out}, no installation is attempted.
#' @param ... additional arguments passed to \code{\link{install.packages}}
#'   when installing dependencies. \code{pkg} is installed with
#'   \code{R CMD INSTALL}.
#' @export
#' @family package installation
#' @seealso \code{\link{with_debug}} to install packages with debugging flags
#'   set.
install <-
  function(pkg = ".", reload = TRUE, quick = FALSE, local = TRUE,
           args = getOption("devtools.install.args"), quiet = FALSE,
           dependencies = NA, upgrade_dependencies = TRUE,
           build_vignettes = FALSE,
           keep_source = getOption("keep.source.pkgs"),
           threads = getOption("Ncpus", 1),
           force_deps = FALSE,
           metadata = remote_metadata(as.package(pkg)),
           out_dir = NULL,
           skip_if_log_exists = FALSE,
           ...) {

  pkg <- as.package(pkg)
  check_build_tools(pkg)

  # Forcing all of the promises for the current namespace now will avoid lazy-load
  # errors when the new package is installed overtop the old one.
  # https://stat.ethz.ch/pipermail/r-devel/2015-December/072150.html
  if (is_loaded(pkg)) {
    eapply(ns_env(pkg), force, all.names = TRUE)
  }

  root_install <- is.null(installing$packages)
  if (root_install) {
    on.exit(installing$packages <- NULL, add = TRUE)
  }

  if (pkg$package %in% installing$packages) {
    if (!quiet) {
      message("Skipping ", pkg$package, ", it is already being installed.")
    }
    return(invisible(FALSE))
  }

  if (!is.null(out_dir)) {
    out_file <- file.path(out_dir, paste0(pkg$package, ".out"))
    if (skip_if_log_exists && file.exists(out_file)) {
      message("Skipping ", pkg$package, ", installation failed before, see log in ", out_file)
      return(invisible(FALSE))
    }
  } else {
    out_file <- NULL
  }

  installing$packages <- c(installing$packages, pkg$package)
  if (!quiet) {
    message("Installing ", pkg$package)
  }

  # If building vignettes, make sure we have all suggested packages too.
  if (build_vignettes && missing(dependencies)) {
    dependencies <- standardise_dep(TRUE)
  } else {
    dependencies <- standardise_dep(dependencies)
  }

  initial_deps <- dependencies[dependencies != "Suggests"]
  final_deps <- dependencies[dependencies == "Suggests"]

  # cache the Remote: dependencies here so we don't have to query them each
  # time we call install_deps
  installing$remote_deps <- remote_deps(pkg)
  on.exit(installing$remote_deps <- NULL, add = TRUE)

  install_deps(pkg, dependencies = initial_deps, upgrade = upgrade_dependencies,
    threads = threads, force_deps = force_deps, quiet = quiet, ...,
    out_dir = out_dir, skip_if_log_exists = skip_if_log_exists)

  # Build the package. Only build locally if it doesn't have vignettes
  has_vignettes <- length(tools::pkgVignettes(dir = pkg$path)$docs > 0)
  if (local && !(has_vignettes && build_vignettes)) {
    built_path <- pkg$path
  } else {
    built_path <- build(pkg, tempdir(), vignettes = build_vignettes, quiet = quiet)
    on.exit(unlink(built_path), add = TRUE)
  }

  opts <- c(
    paste("--library=", shQuote(.libPaths()[1]), sep = ""),
    if (keep_source) "--with-keep.source",
    "--install-tests"
  )
  if (quick) {
    opts <- c(opts, "--no-docs", "--no-multiarch", "--no-demo")
  }
  opts <- paste(paste(opts, collapse = " "), paste(args, collapse = " "))

  built_path <- normalizePath(built_path, winslash = "/")
  R(paste("CMD INSTALL ", shQuote(built_path), " ", opts, sep = ""),
    fun = system2_check,
    quiet = quiet || !is.null(out_file), out_file = out_file)

  # Remove immediately upon success
  unlink(out_file)

  install_deps(pkg, dependencies = final_deps, upgrade = upgrade_dependencies,
    threads = threads, force_deps = force_deps, quiet = quiet, ...,
    out_dir = out_dir, skip_if_log_exists = skip_if_log_exists)

  if (length(metadata) > 0) {
    add_metadata(inst(pkg$package), metadata)
  }

  if (reload) {
    reload(pkg, quiet = quiet)
  }
  invisible(TRUE)
}

# A environment to hold which packages are being installed so packages with
# circular dependencies can be skipped the second time.
installing <- new.env(parent = emptyenv())

#' Install package dependencies if needed.
#'
#' \code{install_deps} is used by \code{install_*} to make sure you have
#' all the dependencies for a package. \code{install_dev_deps()} is useful
#' if you have a source version of the package and want to be able to
#' develop with it: it installs all dependencies of the package, and it
#' also installs roxygen2.
#'
#' @inheritParams install
#' @inheritParams package_deps
#' @param ... additional arguments passed to \code{\link{install.packages}}.
#' @export
#' @examples
#' \dontrun{install_deps(".")}
install_deps <- function(pkg = ".", dependencies = NA,
                         threads = getOption("Ncpus", 1),
                         repos = getOption("repos"),
                         type = getOption("pkgType"),
                         ...,
                         upgrade = TRUE,
                         quiet = FALSE,
                         force_deps = FALSE) {

  pkg <- dev_package_deps(pkg, repos = repos, dependencies = dependencies,
    type = type)
  update(pkg, ..., Ncpus = threads, quiet = quiet, upgrade = upgrade)
  invisible()
}

#' @rdname install_deps
#' @export
install_dev_deps <- function(pkg = ".", ...) {
  update_packages("roxygen2")
  install_deps(pkg, ..., dependencies = TRUE, upgrade = FALSE,
    bioc_packages = TRUE)
}

#' Run a examples for an in-development function.
#'
#' @inheritParams run_examples
#' @param topic Name or topic (or name of Rd) file to run examples for
#' @export
#' @family example functions
#' @examples
#' \dontrun{
#' # Runs installed example:
#' library("ggplot2")
#' example("ggplot")
#'
#' # Runs develoment example:
#' load_all("ggplot2")
#' dev_example("ggplot")
#' }
dev_example <- function(topic) {
  path <- find_topic(topic)

  if (is.null(path)) {
    stop("Can't find development example for topic ", topic, call. = FALSE)
  }

  pkg <- as.package(names(path)[[1]])
  load_all(pkg)

  run_example(path)
}

#' Install a package directly from bitbucket
#'
#' This function is vectorised so you can install multiple packages in
#' a single command.
#'
#' @inheritParams install_github
#' @param auth_user your account username if you're attempting to install
#'   a package hosted in a private repository (and your username is different
#'   to \code{username})
#' @param password your password
#' @param ref Desired git reference; could be a commit, tag, or branch name.
#'   Defaults to master.
#' @seealso Bitbucket API docs:
#'   \url{https://confluence.atlassian.com/bitbucket/use-the-bitbucket-cloud-rest-apis-222724129.html}
#' @family package installation
#' @export
#' @examples
#' \dontrun{
#' install_bitbucket("sulab/mygene.r@@default")
#' install_bitbucket("dannavarro/lsr-package")
#' }
install_bitbucket <- function(repo, username, ref = "master",
                              subdir = NULL, quiet = FALSE,
                              auth_user = NULL, password = NULL, ...) {

  remotes <- lapply(repo, bitbucket_remote, username = username, ref = ref,
    subdir = subdir, auth_user = auth_user, password = password)

  install_remotes(remotes, ..., quiet = quiet)
}

bitbucket_remote <- function(repo, username = NULL, ref = NULL, subdir = NULL,
                              auth_user = NULL, password = NULL, sha = NULL) {

  meta <- parse_git_repo(repo)
  meta$ref <- meta$ref %||% ref %||% "master"

  if (is.null(meta$username)) {
    meta$username <- username %||% stop("Unknown username.")
    warning("Username parameter is deprecated. Please use ",
      username, "/", repo, call. = FALSE)
  }

  remote("bitbucket",
    repo = meta$repo,
    subdir = meta$subdir %||% subdir,
    username = meta$username,
    ref = meta$ref %||% ref,
    sha = sha,
    auth_user = auth_user,
    password = password
  )
}

#' @export
remote_download.bitbucket_remote <- function(x, quiet = FALSE) {
  if (!quiet) {
    message("Downloading bitbucket repo ", x$username, "/", x$repo, "@", x$ref)
  }

  dest <- tempfile(fileext = paste0(".zip"))
  src <- paste("https://bitbucket.org/", x$username, "/", tolower(x$repo), "/get/",
    x$ref, ".zip", sep = "")

  if (!is.null(x$password)) {
    auth <- httr::authenticate(
      user = x$auth_user %||% x$username,
      password = x$password,
      type = "basic")
  } else {
    auth <- NULL
  }

  download(dest, src, auth)
}

#' @export
remote_metadata.bitbucket_remote <- function(x, bundle = NULL, source = NULL) {
  if (!is.null(bundle)) {
    # Might be able to get from zip archive
    sha <- git_extract_sha1(bundle)
  } else {
    # Otherwise can lookup with remote_ls
    sha <- remote_sha(x)
  }

  list(
    RemoteType = "bitbucket",
    RemoteRepo = x$repo,
    RemoteUsername = x$username,
    RemoteRef = x$ref,
    RemoteSha = sha,
    RemoteSubdir = x$subdir
  )
}

#' @export
remote_package_name.bitbucket_remote <- function(remote, ...) {
  remote_package_name.github_remote(remote, url = "https://bitbucket.org", ...)
}

#' @export
remote_sha.bitbucket_remote <-function(remote, ...) {
  remote_sha.github_remote(remote, url = "https://bitbucket.org", ...)
}

#' @export
format.bitbucket_remote <- function(x, ...) {
  "Bitbucket"
}

# Remove s4 classes created by this package.
# This is only necessary if the package was loaded with devtools. If the
# package was NOT loaded by devtools, it's not necessary to remove the
# classes this way, and attempting to do so will result in errors.
remove_s4_classes <- function(pkg = ".") {
  pkg <- as.package(pkg)

  classes <- methods::getClasses(ns_env(pkg))
  lapply(sort_s4classes(classes, pkg), remove_s4_class, pkg)
}

# Sort S4 classes for hierarchical removal
# Derived classes must be removed **after** their parents.
# This reduces to a topological sorting on the S4 dependency class
# https://en.wikipedia.org/wiki/Topological_sorting
sort_s4classes <- function(classes, pkg) {
  pkg <- as.package(pkg)
  nsenv <- ns_env(pkg)

  sorted_classes <- vector(mode = 'character', length = 0)

  ## Return the parent class, if any within domestic classes
  extends_first <- function(x, classes) {
    ext <- methods::extends(methods::getClass(x, where = nsenv))
    parent <- ext[2]
    classes %in% parent
  }

  ## Matrix of classes in columns, extending classes in rows
  extended_classes <- vapply(
    classes,
    extends_first,
    rep(TRUE, length(classes)),
    classes
  )

  if (!is.matrix(extended_classes))
    extended_classes <- as.matrix(extended_classes)

  ## Dynamic set of orphan classes (safe to remove)
  start_idx <- which(apply(extended_classes, 2, sum) == 0)

  while (length(start_idx) > 0) {
    ## add node to sorted list (and remove from pending list)
    i <- start_idx[1]
    start_idx <- utils::tail(start_idx, -1)
    sorted_classes <- c(sorted_classes, classes[i])

    ## check its derived classes if any
    for (j in which(extended_classes[i, ])) {
      extended_classes[i, j] <- FALSE
      if (sum(extended_classes[, j]) == 0) {
        start_idx <- c(start_idx, j)
      }
    }
  }
  if (any(extended_classes)) {
    ## Graph has a cycle. This should not happen
    ## Stop or try to continue?
    idx <- !classes %in% sorted_classes
    sorted_classes <- c(sorted_classes, classes[idx])
  }
  return(sorted_classes)

}

# Remove an s4 class from a package loaded by devtools
#
# For classes loaded with devtools, this is necessary so that R doesn't try to
# modify superclasses that don't have references to this class. For example,
# suppose you have package pkgA with class A, and pkgB with class B, which
# contains A. If pkgB is loaded with load_all(), then class B will have a
# reference to class A, and unloading pkgB the normal way, with
# unloadNamespace("pkgB"), will result in some errors. They happen because R
# will look at B, see that it is a superclass of A, then it will try to modify
# A by removing subclass references to B.
#
# This function sidesteps the problem by modifying B. It finds all the classes
# in B@contains which also have references back to B, then modifes B to keep
# references to those classes, but remove references to all other classes.
# Finally, it removes B. Calling removeClass("B") tells the classes referred to
# in B@contains to remove their references back to B.
#
# It is entirely possible that this code is necessary only because of bugs in
# R's S4 implementation.
#
# @param classname The name of the class.
# @param pkg The package object which contains the class.
remove_s4_class <- function(classname, pkg) {
  pkg <- as.package(pkg)
  nsenv <- ns_env(pkg)

  # Make a copy of the class
  class <- methods::getClassDef(classname, package = pkg$package, inherits = FALSE)

  # Find all the references to classes that (this one contains/extends AND
  # have backreferences to this class) so that R doesn't try to modify them.
  keep_idx <- contains_backrefs(classname, pkg$package, class@contains)
  class@contains <- class@contains[keep_idx]

  # Assign the modified class back into the package
  methods::assignClassDef(classname, class, where = nsenv)

  # Remove the class.
  methods::removeClass(classname, where = nsenv)
}


# Given a list of SClassExtension objects, this returns a logical vector of the
# same length. Each element is TRUE if the corresponding object has a reference
# to this class, FALSE otherwise.
contains_backrefs <- function(classname, pkgname, contains) {

  # If class_a in pkg_a has class_b in pkg_b as a subclass, return TRUE,
  # otherwise FALSE.
  has_subclass_ref <- function(class_a, pkg_a, class_b, pkg_b) {
    x <- methods::getClassDef(class_a, package = pkg_a)
    if (is.null(x)) return(FALSE)

    subclass_ref <- x@subclasses[[class_b]]

    if (!is.null(subclass_ref) && subclass_ref@package == pkg_b) {
      return(TRUE)
    }

    FALSE
  }

  if (length(contains) == 0) {
    return()
  }

  # Get a named vector of 'contains', where each item's name is the class,
  # and the value is the package.
  contain_pkgs <- sapply(contains, "slot", "package")

  mapply(has_subclass_ref, names(contain_pkgs), contain_pkgs, classname, pkgname)
}

source_many <- function(files, envir = parent.frame()) {
  stopifnot(is.character(files))
  stopifnot(is.environment(envir))

  oop <- options(
    keep.source = TRUE,
    show.error.locations = TRUE,
    topLevelEnvironment = as.environment(envir))
  on.exit(options(oop))

  for (file in files) {
    source_one(file, envir = envir)
  }
  invisible()
}

source_one <- function(file, envir = parent.frame()) {
  stopifnot(file.exists(file))
  stopifnot(is.environment(envir))

  lines <- readLines(file, warn = FALSE)
  srcfile <- srcfilecopy(file, lines, file.info(file)[1, "mtime"],
    isFile = TRUE)
  exprs <- parse(text = lines, n = -1, srcfile = srcfile)

  n <- length(exprs)
  if (n == 0L) return(invisible())

  for (i in seq_len(n)) {
    eval(exprs[i], envir)
  }
  invisible()
}

#' Build package vignettes.
#'
#' Builds package vignettes using the same algorithm that \code{R CMD build}
#' does. This means including non-Sweave vignettes, using makefiles (if
#' present), and copying over extra files. You need to ensure that these
#' files are not included in the built package - ideally they should not
#' be checked into source, or at least excluded with \code{.Rbuildignore}
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @inheritParams install_deps
#' @keywords programming
#' @seealso \code{\link{clean_vignettes}} to remove the pdfs in
#'   \file{inst/doc} created from vignettes
#' @export
#' @seealso \code{\link{clean_vignettes}} to remove build tex/pdf files.
build_vignettes <- function(pkg = ".", dependencies = "VignetteBuilder") {
  pkg <- as.package(pkg)
  vigns <- tools::pkgVignettes(dir = pkg$path)
  if (length(vigns$docs) == 0) return()

  install_deps(pkg, dependencies, upgrade = FALSE)

  message("Building ", pkg$package, " vignettes")
  tools::buildVignettes(dir = pkg$path, tangle = TRUE)
  copy_vignettes(pkg)

  invisible(TRUE)
}

#' Clean built vignettes.
#'
#' This uses a fairly rudimentary algorithm where any files in \file{inst/doc}
#' with a name that exists in \file{vignettes} are removed.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @export
clean_vignettes <- function(pkg = ".") {
  pkg <- as.package(pkg)
  vigns <- tools::pkgVignettes(dir = pkg$path)
  if (basename(vigns$dir) != "vignettes") return()

  message("Cleaning built vignettes from ", pkg$package)

  doc_path <- file.path(pkg$path, "inst", "doc")

  vig_candidates <- dir(doc_path, full.names = TRUE)
  vig_rm <- vig_candidates[file_name(vig_candidates) %in% file_name(vigns$docs)]

  extra_candidates <- file.path(doc_path, basename(find_vignette_extras(pkg)))
  extra_rm <- extra_candidates[file.exists(extra_candidates)]

  to_remove <- c(vig_rm, extra_rm)
  if (length(to_remove) > 0) {
    message("Removing ", paste(basename(to_remove), collapse = ", "))
    file.remove(to_remove)
  }

  invisible(TRUE)
}

ext_variations <- function(path, valid_ext) {
  unique(c(outer(file_name(path), valid_ext, FUN = paste, sep = ".")))
}

file_name <- function(x) {
  if (length(x) == 0) return(NULL)
  tools::file_path_sans_ext(basename(x))
}

#' Install a package from a Bioconductor repository
#'
#' This function requires \code{svn} to be installed on your system in order to
#' be used.
#'
#' It is vectorised so you can install multiple packages with
#' a single command.
#'
#' '
#' @inheritParams install_git
#' @param repo Repository address in the format
#'   \code{[username:password@@][release/]repo[#revision]}. Valid values for
#'   the release are \sQuote{devel} (the default if none specified),
#'   \sQuote{release} or numeric release numbers (e.g. \sQuote{3.3}).
#' @param mirror The bioconductor SVN mirror to use
#' @param ... Other arguments passed on to \code{\link{install}}
#' @export
#' @family package installation
#' @examples
#' \dontrun{
#' install_bioc("SummarizedExperiment")
#' install_bioc("user@SummarizedExperiment")
#' install_bioc("user:password@release/SummarizedExperiment")
#' install_bioc("user:password@3.3/SummarizedExperiment")
#' install_bioc("user:password@3.3/SummarizedExperiment#117513")
#'}
install_bioc <- function(repo, mirror = getOption("BioC_svn", "https://hedgehog.fhcrc.org/bioconductor"), ..., quiet = FALSE) {

  remotes <- lapply(repo, bioc_remote, mirror = mirror)

  install_remotes(remotes, ..., quiet = quiet)
}

# Parse concise SVN repo specification: [username[:password]@][branch/]repo[#revision]
parse_bioc_repo <- function(path) {
  user_pass_rx <- "(?:(?:([^:]+):)?([^:@]+)@)?"
  release_rx <- "(?:(devel|release|[0-9.]+)/)?"
  repo_rx <- "([^/@#]+)"
  revision_rx <- "(?:[#][Rr]?([0-9]+))?"
  bioc_rx <- sprintf("^(?:%s%s%s%s|(.*))$", user_pass_rx, release_rx, repo_rx, revision_rx)

  param_names <- c("username", "password", "release", "repo", "revision", "invalid")
  replace <- stats::setNames(sprintf("\\%d", seq_along(param_names)), param_names)
  params <- lapply(replace, function(r) gsub(bioc_rx, r, path, perl = TRUE))
  if (params$invalid != "")
    stop(sprintf("Invalid bioc repo: %s", path))
  params <- params[sapply(params, nchar) > 0]

  if (!is.null(params$password) && is.null(params$username)) {
    params$username <- params$password
    params$password <- NULL
  }

  params
}

bioc_remote <- function(repo, mirror = getOption("BioC_svn", "https://hedgehog.fhcrc.org/bioconductor")) {
  meta <- parse_bioc_repo(repo)

  meta$username <- meta$username %||% "readonly"

  if (meta$username == "readonly") {
    meta$password <- "readonly"
  }

  remote("bioc",
    mirror = mirror,
    username = meta$username,
    password = meta$password,
    repo = meta$repo,
    release = meta$release %||% "devel",
    revision = meta$revision
  )
}

#' @export
remote_download.bioc_remote <- function(x, quiet = FALSE) {
  if (!quiet) {
    message("Downloading Bioconductor repo ", x$repo)
  }

  bundle <- tempfile()
  svn_binary_path <- svn_path()

  args <- c(
    "co",
    bioc_args(x),
    bioc_url(x),
    bundle)

  if (!quiet) {
    message(svn_binary_path, " ", paste0(args, collapse = " "))
  }
  request <- system2(svn_binary_path, args, stdout = FALSE, stderr = FALSE)

  # This is only looking for an error code above 0-success
  if (request > 0) {
    stop("Error retrieving Bioc Remote `", x$repo, "`", call. = FALSE)
  }

  bundle
}

#' @export
remote_metadata.bioc_remote <- function(x, bundle = NULL, source = NULL) {

  if (!is.null(bundle)) {
    withr::with_dir(bundle, {
      revision <- svn_revision()
    })
  } else {
    revision <- NULL
  }

  list(
    RemoteType = "bioc",
    RemoteRepo = x$repo,
    RemoteMirror = x$mirror,
    RemoteRelease = x$release,
    RemoteUsername = x$username,
    RemotePassword = x$password,
    RemoteRevision = revision,
    RemoteSha = revision # for compatibility with other remotes
  )
}

#' @export
remote_package_name.bioc_remote <- function(remote, ...) {
  remote$repo
}

#' @export
remote_sha.bioc_remote <- function(remote, ...) {
  svn_revision(paste(c(bioc_args(remote), bioc_url(remote)), collapse = " "))
}

bioc_args <- function(x) {
  args <- c(
    if (!interactive()) {
      "--non-interactive"
    },
    if (!is.null(x$revision)) {
      c("--revision", x$revision)
    },
    "--username", x$username,
    if (!is.null(x$password)) {
      c("--password", x$password)
    })
}

bioc_url <- function(x) {
  to_svn_release <- function(x) {
    sprintf("RELEASE_%s", sub("[.]", "_", x))
  }

  switch(tolower(x$release),
    devel = file.path(x$mirror, "trunk", "madman", "Rpacks", x$repo),
    release = file.path(x$mirror, "branches", to_svn_release(bioconductor_release()), "madman", "Rpacks", x$repo),
    file.path(x$mirror, "branches", to_svn_release(x$release), "madman", "Rpacks", x$repo))
}

bioconductor_release <- memoise::memoise(function() {
  tmp <- tempfile()
  download.file("http://bioconductor.org/config.yaml", tmp, quiet = TRUE)

  gsub("release_version:[[:space:]]+\"([[:digit:].]+)\"", "\\1",
    grep("release_version:", readLines(tmp), value = TRUE))
})

format.bioc_remote <- function(x, ...) {
  "Bioc"
}

context("bioc")

test_that("bioc repo paths are parsed correctly", {
  expect_equal(parse_bioc_repo("devtools"), list(repo="devtools"))
  expect_equal(parse_bioc_repo("devtools#12345"), list(repo="devtools", revision="12345"))
  expect_equal(parse_bioc_repo("user@devtools"), list(repo="devtools", username = "user"))
  expect_equal(parse_bioc_repo("user:pass@devtools"), list(username = "user", password = "pass", repo="devtools"))
  expect_equal(parse_bioc_repo("devel/devtools"), list(release = "devel", repo="devtools"))
  expect_equal(parse_bioc_repo("3.1/devtools"), list(release = "3.1", repo="devtools"))
  expect_equal(parse_bioc_repo("release/devtools"), list(release = "release", repo="devtools"))
  expect_equal(parse_bioc_repo("user:pass@3.1/devtools#123"), list(username = "user", password = "pass", release = "3.1", repo="devtools", revision = "123"))
  expect_error(parse_bioc_repo("user:@devtools"), "Invalid bioc repo")
  expect_error(parse_bioc_repo("@devtools"), "Invalid bioc repo")
  expect_error(parse_bioc_repo("devtools/"), "Invalid bioc repo")
  expect_error(parse_bioc_repo("junk/devtools"), "Invalid bioc repo")
})

test_that("install_bioc", {
  skip_on_cran()

  lib <- tempfile()
  on.exit(unlink(lib, recursive = TRUE), add = TRUE)
  dir.create(lib)
  libpath <- .libPaths()
  on.exit(.libPaths(libpath), add = TRUE)
  .libPaths(lib)

  # unload BiocInstaller if it is already loaded, unload it after this function
  # finishes as well
  unloadNamespace("BiocInstaller")
  on.exit(unloadNamespace("BiocInstaller"), add = TRUE)

  # Install BiocInstaller to the new library
  source("https://bioconductor.org/biocLite.R")

  # This package has no dependencies or compiled code and is old
  install_bioc("MeasurementError.cor", quiet = TRUE)

  expect_silent(packageDescription("MeasurementError.cor"))
  expect_equal(packageDescription("MeasurementError.cor")$RemoteType, "bioc")
})

## set-up and tear-down
create_in_temp <- function(pkg) {
  temp_path <- tempfile(pattern="devtools-test-")
  dir.create(temp_path)
  test_pkg <- file.path(temp_path, pkg)
  capture.output(suppressMessages(create(test_pkg, description = list())))
  test_pkg
}

erase <- function(path) unlink(path, recursive = TRUE)

## fake GitHub connectivity: set a GitHub remote and add GitHub links
mock_use_github <- function(pkg) {
  use_git_with_config(message = "initial", pkg = pkg, add_user_config = TRUE, quiet = TRUE)
  r <- git2r::repository(pkg)
  git2r::remote_add(r, "origin", "https://github.com/hadley/devtools.git")
  use_github_links(pkg)
  git2r::add(r, "DESCRIPTION")
  git2r::commit(r, "Add GitHub links to DESCRIPTION")
  invisible(NULL)
}

context("remote-metadata")

test_that("pkg_sha metadata from shallow clone", {

    pkg_sha <- git_sha1(path='shallowRepo')
    expect_equal(pkg_sha, '21d5d94011')

})


test_that("install on packages adds metadata", {
  skip_on_cran()

  # temp libPaths
  withr::with_temp_libpaths({
    test_pkg <- create_in_temp("testMetadataInstall")
    mock_use_github(test_pkg)

    # first do metadata = NULL
    install(test_pkg, quiet = TRUE, metadata = NULL)


    # cleanup code for when we are all finished
    on.exit(unload(test_pkg), add = TRUE)
    on.exit(erase(test_pkg), add = TRUE)

    # first time loading the package
    library("testMetadataInstall")

    pkg_info <- session_info()$packages
    expect_equal(pkg_info[pkg_info[, "package"] %in% "testMetadataInstall", "source"],
                 "local")

    # now use default
    r <- git2r::repository(test_pkg)

    # then use metadata
    install(test_pkg, quiet = TRUE)
    library("testMetadataInstall")
    pkg_info <- session_info()$packages
    pkg_source <- pkg_info[pkg_info[, "package"] %in% "testMetadataInstall", "source"]
    pkg_sha <- substring(git2r::commits(r)[[1]]@sha, 1, 7)
    expect_match(pkg_source, pkg_sha)

    # dirty the repo
    cat("just a test", file = file.path(test_pkg, "test.txt"))
    install(test_pkg, quiet = TRUE)
    pkg_info <- session_info()$packages
    pkg_source <- pkg_info[pkg_info[, "package"] %in% "testMetadataInstall", "source"]
    expect_match(pkg_source, "local")

    # use load_all() and reinstall
    git2r::add(r, file.path(test_pkg, "test.txt"))
    git2r::commit(r, "adding test.txt")
    load_all(test_pkg, quiet = TRUE)
    install(test_pkg, quiet = TRUE)
    pkg_info <- session_info()$packages
    pkg_source <- pkg_info[pkg_info[, "package"] %in% "testMetadataInstall", "source"]
    pkg_sha <- substring(git2r::commits(r)[[1]]@sha, 1, 7)
    expect_match(pkg_source, pkg_sha)

  })
})

context("git usage and GitHub connections")

test_that("git (non-)usage is detected, diagnosed, and can be added", {
  skip_on_cran()

  test_pkg <- create_in_temp("testNoGit")

  expect_false(uses_git(test_pkg))
  expect_warning(expect_message(print(dr_github(test_pkg)),
                                "not a git repository"),
                 'DR_GITHUB FOUND PROBLEMS')

  expect_message(use_git_with_config(message = "initial", pkg = test_pkg, add_user_config = TRUE),
                 "Initialising repo")
  expect_true(uses_git(test_pkg))

  erase(test_pkg)

})

test_that("GitHub non-usage is handled", {
  skip_on_cran()

  test_pkg <- create_in_temp("testNoGitHub")
  use_git_with_config(message = "initial", pkg = test_pkg, add_user_config = TRUE, quiet = TRUE)

  expect_true(uses_git(test_pkg))
  expect_false(uses_github(test_pkg))
  expect_warning(expect_message(print(dr_github(test_pkg)),
                                "not a GitHub repository"),
                 "DR_GITHUB FOUND PROBLEMS")

  expect_identical(github_dummy, github_info(test_pkg))

  expect_error(use_github_links(test_pkg), "Cannot detect .* GitHub")

  erase(test_pkg)

})

## If env var GITHUB_PAT exists and there's willingness to call GitHub
## use_github() could be tested right around here.
## As it stands, that function is not under automated testing.

test_that("github info and links can be queried and manipulated", {
  skip_on_cran()

  test_pkg <- create_in_temp("testGitHub")
  mock_use_github(test_pkg)

  expect_true(uses_github(test_pkg))

  gh_info <- github_info(test_pkg)
  expect_equal(gh_info$username, "hadley")
  expect_equal(gh_info$repo, "devtools")

  desc_path <- file.path(test_pkg, "DESCRIPTION")
  desc <- read_dcf(desc_path)

  ## default GitHub links created by use_github_links() via use_github()
  expect_identical(desc[["URL"]],
                   file.path("https://github.com",
                             gh_info$username, gh_info$repo))
  expect_identical(desc[["BugReports"]],
                   file.path("https://github.com",
                             gh_info$username, gh_info$repo, "issues"))

  ## make sure we don't clobber existing links
  mtime_before <- file.info(desc_path)$mtime
  expect_message(use_github_links(test_pkg), "found and preserved")
  mtime_after <- file.info(desc_path)$mtime
  expect_identical(mtime_before, mtime_after)

  ## make sure we diagnose lack of GitHub links
  desc$URL <- "http://www.example.com"
  desc$BugReports <- "http://www.example.com/issues"
  write_dcf(desc_path, desc)
  expect_warning(expect_message(print(dr_github(test_pkg)),
                                "no GitHub repo link"),
                 "DR_GITHUB FOUND PROBLEMS")
  expect_warning(expect_message(print(dr_github(test_pkg)),
                                "no GitHub Issues"),
                 "DR_GITHUB FOUND PROBLEMS")

  erase(test_pkg)

})

test_that("github_info() prefers, but doesn't require, remote named 'origin'", {
  skip_on_cran()

  test_pkg <- create_in_temp("testGitHubInfo")
  mock_use_github(test_pkg)

  r <- git2r::repository(test_pkg, discover = TRUE)
  git2r::remote_add(r, "anomaly",
    "https://github.com/twitter/AnomalyDetection.git")

  ## defaults to "origin"
  expect_equal(github_info(test_pkg)$username, "hadley")
  expect_equal(github_info(test_pkg)$repo, "devtools")

  ## another remote will be used if no "origin"
  git2r::remote_rename(r, "origin", "zzz")
  gh_info <- github_info(test_pkg)
  expect_equal(gh_info$username, "twitter")
  expect_equal(gh_info$repo, "AnomalyDetection")
  git2r::remote_rename(r, "zzz", "origin")

  ## another remote can be requested by name
  gh_info <- github_info(test_pkg, remote_name = "anomaly")
  expect_equal(gh_info$username, "twitter")
  expect_equal(gh_info$repo, "AnomalyDetection")

  ## error if nonexistent remote requested by name
  expect_error(github_info(test_pkg, remote_name = "nope"))

  erase(test_pkg)

})

test_that("username and repo are extracted from github remote URL", {
  gh_info <- list(username = "hadley", repo = "devtools",
                  fullname = "hadley/devtools")
  expect_identical(github_remote_parse("https://github.com/hadley/devtools.git"), gh_info)
  expect_identical(github_remote_parse("https://github.com/hadley/devtools"), gh_info)
  expect_identical(github_remote_parse("git@github.com:hadley/devtools.git"), gh_info)
})

# Adapted from:
# https://github.com/rstudio/rstudio/blob/master/src/cpp/session/modules/SessionPackages.R
#
# Copyright (C) 2009-12 by RStudio, Inc.

download_method <- function(x) {
  if (identical(x, "auto")) {
    auto_download_method()
  } else {
    x
  }
}

auto_download_method <- function() {
  if (isTRUE(capabilities("libcurl"))) {
    "libcurl"
  } else if (isTRUE(capabilities("http/ftp"))) {
    "internal"
  } else if (nzchar(Sys.which("wget"))) {
    "wget"
  } else if (nzchar(Sys.which("curl"))) {
    "curl"
  } else {
    ""
  }
}

download_method_secure <- function(method = getOption("download.file.method", "auto")) {
  method <- download_method(method)

  if (method %in% c("wininet", "libcurl", "wget", "curl")) {
    # known good methods
    TRUE
  } else if (identical(method, "internal")) {
    # if internal then see if were using windows internal with inet2
    identical(Sys.info()[["sysname"]], "Windows") && utils::setInternet2(NA)
  } else {
    # method with unknown properties (e.g. "lynx") or unresolved auto
    FALSE
  }
}


cran_mirror <- function() {
  if (download_method_secure()) {
    "https://cran.rstudio.com"
  } else {
    "http://cran.rstudio.com"
  }
}

#' Read the in-development help for a package loaded with devtools.
#'
#' Note that this only renders a single documentation file, so that links
#' to other files within the package won't work.
#'
#' @param topic name of help to search for.
#' @param stage at which stage ("build", "install", or "render") should
#'   \\Sexpr macros be executed? This is only important if you're using
#'   \\Sexpr macro's in your Rd files.
#' @param type of html to produce: \code{"html"} or \code{"text"}. Defaults to
#'   your default documentation type.
#' @export
#' @examples
#' \dontrun{
#' library("ggplot2")
#' help("ggplot") # loads installed documentation for ggplot
#'
#' load_all("ggplot2")
#' dev_help("ggplot") # loads development documentation for ggplot
#' }
dev_help <- function(topic, stage = "render", type = getOption("help_type")) {
  message("Using development documentation for ", topic)
  path <- find_topic(topic)
  if (is.null(path)) {
    dev <- paste(dev_packages(), collapse = ", ")
    stop("Could not find topic ", topic, " in: ", dev)
  }

  pkg <- basename(names(path)[1])
  path <- normalizePath(path, winslash = "/")
  if (rstudioapi::hasFun("previewRd")) {
    rstudioapi::callFun("previewRd", path)
  } else {
    view_rd(path, pkg, stage = stage, type = type)
  }

}

view_rd <- function(path, package, stage = "render", type = getOption("help_type")) {
  if (is.null(type)) type <- "text"
  type <- match.arg(type, c("text", "html"))

  out_path <- paste(tempfile("Rtxt"), type, sep = ".")

  if (type == "text") {
    tools::Rd2txt(path, out = out_path, package = package, stages = stage)
    file.show(out_path, title = paste(package, basename(path), sep = ":"))
  } else if (type == "html") {
    tools::Rd2HTML(path, out = out_path, package = package, stages = stage,
      no_links = TRUE)

    css_path <- file.path(tempdir(), "R.css")
    if (!file.exists(css_path)) {
      file.copy(file.path(R.home("doc"), "html", "R.css"), css_path)
    }

    utils::browseURL(out_path)
  }
}


#' Drop-in replacements for help and ? functions
#'
#' The \code{?} and \code{help} functions are replacements for functions of the
#' same name in the utils package. They are made available when a package is
#' loaded with \code{\link{load_all}}.
#'
#' The \code{?} function is a replacement for \code{\link[utils]{?}} from the
#' utils package. It will search for help in devtools-loaded packages first,
#' then in regular packages.
#'
#' The \code{help} function is a replacement for \code{\link[utils]{help}} from
#' the utils package. If \code{package} is not specified, it will search for
#' help in devtools-loaded packages first, then in regular packages. If
#' \code{package} is specified, then it will search for help in devtools-loaded
#' packages or regular packages, as appropriate.
#'
#' @inheritParams utils::help utils::`?`
#' @param topic A name or character string specifying the help topic.
#' @param package A name or character string specifying the package in which
#'   to search for the help topic. If NULL, search all packages.
#' @param e1 First argument to pass along to \code{utils::`?`}.
#' @param e2 Second argument to pass along to \code{utils::`?`}.
#' @param ... Additional arguments to pass to \code{\link[utils]{help}}.
#'
#' @rdname help
#' @name help
#' @usage # help(topic, package = NULL, ...)
#'
#' @examples
#' \dontrun{
#' # This would load devtools and look at the help for load_all, if currently
#' # in the devtools source directory.
#' load_all()
#' ?load_all
#' help("load_all")
#' }
#'
#' # To see the help pages for utils::help and utils::`?`:
#' help("help", "utils")
#' help("?", "utils")
#'
#' \dontrun{
#' # Examples demonstrating the multiple ways of supplying arguments
#' # NB: you can't do pkg <- "ggplot2"; help("ggplot2", pkg)
#' help(lm)
#' help(lm, stats)
#' help(lm, 'stats')
#' help('lm')
#' help('lm', stats)
#' help('lm', 'stats')
#' help(package = stats)
#' help(package = 'stats')
#' topic <- "lm"
#' help(topic)
#' help(topic, stats)
#' help(topic, 'stats')
#' }
shim_help <- function(topic, package = NULL, ...) {
  # Reproduce help's NSE for topic - try to eval it and see if it's a string
  topic_name <- substitute(topic)
  is_char <- FALSE
  try(is_char <- is.character(topic) && length(topic) == 1L, silent = TRUE)
  if (is_char) {
    topic_str <- topic
    topic_name <- as.name(topic)
  } else if (missing(topic_name)) {
    # Leave the vars missing
  } else if (is.null(topic_name)) {
    topic_str <- NULL
    topic_name <- NULL
  } else {
    topic_str <- deparse(substitute(topic))
  }

  # help's NSE for package is slightly simpler
  package_name <- substitute(package)
  if (is.name(package_name)) {
    package_str <- as.character(package_name)
  } else if (is.null(package_name)) {
    package_str <- NULL
  } else {
    package_str <- package
    package_name <- as.name(package)
  }

  use_dev <- (!is.null(package_str) && package_str %in% dev_packages()) ||
    (is.null(package_str) && !is.null(find_topic(topic_str)))
  if (use_dev) {
    dev_help(topic_str)
  } else {
    # This is similar to list(), except that one of the args is a missing var,
    # it will replace it with an empty symbol instead of trying to evaluate it.
    as_list <- function(..., .env = parent.frame()) {
      dots <- match.call(expand.dots = FALSE)$`...`

      lapply(dots, function(var) {
        is_missing <- eval(substitute(missing(x), list(x = var)), .env)
        if (is_missing) {
          quote(expr=)
        } else {
          eval(var, .env)
        }
      })
    }

    call <- substitute(
      utils::help(topic, package, ...),
      as_list(topic = topic_name, package = package_name)
    )
    eval(call)
  }
}


#' @usage
#' # ?e2
#' # e1?e2
#'
#' @rdname help
#' @name ?
shim_question <- function(e1, e2) {
  # Get string version of e1, for find_topic
  e1_expr <- substitute(e1)
  if (is.name(e1_expr)) {
    # Called with a bare symbol, like ?foo
    e1_str <- deparse(e1_expr)

  } else if (is.call(e1_expr)) {
    if (e1_expr[[1]] == "?") {
      # Double question mark, like ??foo
      e1_str <- NULL
    } else {
      # Called with function arguments, like ?foo(12)
      e1_str <- deparse(e1_expr[[1]])
    }

  } else {
    # If we got here, it's probably a string
    e1_str <- e1
  }

  # Search for the topic in devtools-loaded packages.
  # If not found, call utils::`?`.
  if (!is.null(find_topic(e1_str))) {
    dev_help(e1_str)
  } else {
    eval(as.call(list(utils::`?`, substitute(e1), substitute(e2))))
  }
}

context("help")

test_that("shim_help behaves the same as utils::help for non-devtools-loaded packages", {
  # stats wasn't loaded with devtools. There are many combinations of calling
  # with quotes and without; make sure they're the same both ways. Need to index
  # in using [1] to drop attributes for which there are unimportant differences.
  expect_identical(shim_help(lm)[1],            utils::help(lm)[1])
  expect_identical(shim_help(lm, stats)[1],     utils::help(lm, stats)[1])
  expect_identical(shim_help(lm, 'stats')[1],   utils::help(lm, 'stats')[1])
  expect_identical(shim_help('lm')[1],          utils::help('lm')[1])
  expect_identical(shim_help('lm', stats)[1],   utils::help('lm', stats)[1])
  expect_identical(shim_help('lm', 'stats')[1], utils::help('lm', 'stats')[1])
})

test_that("shim_help behaves the same as utils::help for nonexistent objects", {
  expect_equal(length(shim_help(foofoo)), 0)
  expect_equal(length(shim_help("foofoo")), 0)
})


test_that("shim_question behaves the same as utils::? for non-devtools-loaded packages", {
  expect_identical(shim_question(lm)[1], utils::`?`(lm)[1])
  expect_identical(shim_question(lm(123))[1], utils::`?`(lm(123))[1])
  expect_identical(shim_question(`lm`)[1], utils::`?`(`lm`)[1])
  expect_identical(shim_question('lm')[1], utils::`?`('lm')[1])
})

test_that("shim_question behaves the same as utils::? for nonexistent objects", {
  expect_equal(length(shim_question(foofoo)), 0)
  expect_equal(length(shim_question(`foofoo`)), 0)
  expect_equal(length(shim_question("foofoo")), 0)

  # If given a function call with nonexistent function, error
  expect_error(utils::`?`(foofoo(123)))
  expect_error(shim_question(foofoo(123)))
})


test_that("help and ? find files for devtools-loaded packages", {
  load_all('testHelp')

  # We can't test dev_help or help directly, because instead of returning an
  # object, they display the Rd file directly. But dev_help uses find_topic,
  # and we can test that.
  expect_true(!is.null(find_topic('foofoo')))
  expect_null(find_topic('bad_value'))
  unload('testHelp')
})

load_depends <- function(pkg = ".") {
  pkg <- as.package(pkg)

  # Get data frame of dependency names and versions
  deps <- parse_deps(pkg$depends)
  if (is.null(deps) || nrow(deps) == 0) return(invisible())

  mapply(check_dep_version, deps$name, deps$version, deps$compare)
  lapply(deps$name, require, character.only = TRUE)

  invisible(deps)
}

## Define a graph of classes with complex inheritance pattern
## example taken from wikipedia:
## https://en.wikipedia.org/wiki/Topological_sorting#Examples

setClass("A")
setClass("B")
setClass("C")
setClassUnion("D", members = c("A", "B", "C"))
setClass("E")
setIs("B", "E")
setClassUnion("F", members = c("D", "E"))
setClass("G")
setIs("D", "G")
setClassUnion("H", members = c("C", "E"))

#' Load data.
#'
#' Loads all \code{.RData} files in the data subdirectory.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @keywords programming
#' @export
load_data <- function(pkg = ".") {
  # Note that this simulates normal R package loading by placing the data sets
  # in the .__NAMESPACE__.$lazydata environment, but unlike with proper lazy
  # loading via lazyLoad(), we'll need to manually copy these objects over to
  # the package environment later.

  pkg <- as.package(pkg)
  nsenv <- ns_env(pkg)
  lazydata_env <- nsenv$.__NAMESPACE__.$lazydata
  objs <- character()

  sysdata <- file.path(pkg$path, "R", "sysdata.rda")
  if (file.exists(sysdata)) {
    objs <- c(objs, load(sysdata, envir = nsenv))
  }

  path_data <- file.path(pkg$path, "data")
  if (file.exists(path_data)) {
    paths <- dir(path_data, "\\.[rR][dD]a(ta)?$", full.names = TRUE)
    paths <- changed_files(paths)
    objs <- c(objs, unlist(lapply(paths, load, envir = lazydata_env)))

    paths <- dir(path_data, "\\.[rR]$", full.names = TRUE)
    paths <- changed_files(paths)
    objs <- c(objs, unlist(lapply(paths, sys.source, envir = lazydata_env,
      chdir = TRUE, keep.source = TRUE)))
  }

  invisible(objs)
}

a <- 1
b <- 1
c <- 1

onload_lib <- ""
onattach_lib <- ""

.onLoad <- function(lib, pkg) {
  onload_lib <<- lib
  a <<- a + 1
}

.onAttach <- function(lib, pkg) {
  onattach_lib <<- lib

  # Attempt to modify b in namespace. This should throw an error
  # in a real install+load because namespace is locked. But with
  # load_all, it will work because the namespace doesn't get locked.
  try(b <<- b + 1, silent = TRUE)

  # Now modify c in package environment
  env <- as.environment("package:testLoadHooks")
  env$c <- env$c + 1
}

.onUnload <- function(libpath) {
  # Increment this variable if it exists in the global env
  if (exists(".__testLoadHooks__", .GlobalEnv)) {
    .GlobalEnv$.__testLoadHooks__ <- .GlobalEnv$.__testLoadHooks__ + 1
  }
}

using_gcc49 <- function() {
  isTRUE(sub("^gcc[^[:digit:]]+", "", Sys.getenv("R_COMPILED_BY")) >= "4.9.3")
}

gcc_arch <- function() if (Sys.getenv("R_ARCH") == "/i386") "32" else "64"

# Need to check for existence so load_all doesn't override known rtools location
if (!exists("set_rtools_path")) {
  set_rtools_path <- NULL
  get_rtools_path <- NULL
  local({
    rtools_paths <- NULL
    set_rtools_path <<- function(rtools) {
      if (is.null(rtools)) {
        rtools_paths <<- NULL
        return()
      }

      stopifnot(is.rtools(rtools))
      path <- file.path(rtools$path, version_info[[rtools$version]]$path)

      # If using gcc49 and _without_ a valid BINPREF already set
      if (using_gcc49() && is.null(rtools$valid_binpref)) {
        Sys.setenv(BINPREF = file.path(rtools$path, "mingw_$(WIN)", "bin", "/"))
      }
      rtools_paths <<- path
    }

    get_rtools_path <<- function() {
      rtools_paths
    }
  })
}

#' Find rtools.
#'
#' To build binary packages on windows, Rtools (found at
#' \url{https://cran.r-project.org/bin/windows/Rtools/}) needs to be on
#' the path. The default installation process does not add it, so this
#' script finds it (looking first on the path, then in the registry).
#' It also checks that the version of rtools matches the version of R.
#'
#' @section Acknowledgements:
#'   This code borrows heavily from RStudio's code for finding rtools.
#'   Thanks JJ!
#' @param cache if \code{TRUE} will used cached version of RTools.
#' @param debug if \code{TRUE} prints a lot of additional information to
#'   help in debugging.
#' @return Either a visible \code{TRUE} if rtools is found, or an invisible
#'   \code{FALSE} with a diagnostic \code{\link{message}}.
#'   As a side-effect the internal package variable \code{rtools_path} is
#'   updated to the paths to rtools binaries.
#' @keywords internal
#' @export
setup_rtools <- function(cache = TRUE, debug = FALSE) {
  # Non-windows users don't need rtools
  if (.Platform$OS.type != "windows") return(TRUE)
  # Don't look again, if we've already found it

  if (!cache) {
    set_rtools_path(NULL)
  }
  if (!is.null(get_rtools_path())) return(TRUE)

  # First try the path
  from_path <- scan_path_for_rtools(debug)
  if (is_compatible(from_path)) {
    set_rtools_path(from_path)
    return(TRUE)
  }

  if (!is.null(from_path)) {
    # Installed
    if (is.null(from_path$version)) {
      # but not from rtools
      if (debug) "gcc and ls on path, assuming set up is correct\n"
      return(TRUE)
    } else {
    # Installed, but not compatible
      message("WARNING: Rtools ", from_path$version, " found on the path",
        " at ", from_path$path, " is not compatible with R ", getRversion(), ".\n\n",
        "Please download and install ", rtools_needed(), " from ", rtools_url,
        ", remove the incompatible version from your PATH.")
      return(invisible(FALSE))
    }
  }

  # Not on path, so try registry
  registry_candidates <- scan_registry_for_rtools(debug)

  if (length(registry_candidates) == 0) {
    # Not on path or in registry, so not installled
    message("WARNING: Rtools is required to build R packages, but is not ",
      "currently installed.\n\n",
      "Please download and install ", rtools_needed(), " from ", rtools_url, ".")
    return(invisible(FALSE))
  }

  from_registry <- Find(is_compatible, registry_candidates, right = TRUE)
  if (is.null(from_registry)) {
    # In registry, but not compatible.
    versions <- vapply(registry_candidates, function(x) x$version, character(1))
    message("WARNING: Rtools is required to build R packages, but no version ",
      "of Rtools compatible with R ", getRversion(), " was found. ",
      "(Only the following incompatible version(s) of Rtools were found:",
      paste(versions, collapse = ","), ")\n\n",
      "Please download and install ", rtools_needed(), " from ", rtools_url, ".")
    return(invisible(FALSE))
  }

  installed_ver <- installed_version(from_registry$path, debug = debug)
  if (is.null(installed_ver)) {
    # Previously installed version now deleted
    message("WARNING: Rtools is required to build R packages, but the ",
      "version of Rtools previously installed in ", from_registry$path,
      " has been deleted.\n\n",
      "Please download and install ", rtools_needed(), " from ", rtools_url, ".")
    return(invisible(FALSE))
  }

  if (installed_ver != from_registry$version) {
    # Installed version doesn't match registry version
    message("WARNING: Rtools is required to build R packages, but no version ",
      "of Rtools compatible with R ", getRversion(), " was found. ",
      "Rtools ", from_registry$version, " was previously installed in ",
      from_registry$path, " but now that directory contains Rtools ",
      installed_ver, ".\n\n",
      "Please download and install ", rtools_needed(), " from ", rtools_url, ".")
    return(invisible(FALSE))
  }

  # Otherwise it must be ok :)
  set_rtools_path(from_registry)
  TRUE
}

scan_path_for_rtools <- function(debug = FALSE,
                                 gcc49 = using_gcc49(),
                                 arch = gcc_arch()) {
  if (debug) cat("Scanning path...\n")

  # First look for ls and gcc
  ls_path <- Sys.which("ls")
  if (ls_path == "") return(NULL)
  if (debug) cat("ls :", ls_path, "\n")

  # We have a candidate installPath
  install_path <- dirname(dirname(ls_path))

  if (gcc49) {
    find_gcc49 <- function(path) {
      if (!file.exists(path)) {
        path <- paste0(path, ".exe")
      }
      file_info <- file.info(path)

      # file_info$exe should be win32 or win64 respectively
      if (!file.exists(path) || file_info$exe != paste0("win", arch)) {
        return(character(1))
      }
      path
    }


    # First check if gcc set by BINPREF/CC is valid and use that is so
    cc_path <- RCMD("config", "CC", fun = system_output, quiet = !debug)

    # remove '-m64' from tail if it exists
    cc_path <- sub("[[:space:]]+-m[[:digit:]]+$", "", cc_path)

    gcc_path <- find_gcc49(cc_path)
    if (nzchar(gcc_path)) {
      return(rtools(install_path, NULL, valid_binpref = TRUE))
    }

    # if not check default location Rtools/mingw_{32,64}/bin/gcc.exe
    gcc_path <- find_gcc49(file.path(install_path, paste0("mingw_", arch), "bin", "gcc.exe"))
    if (!nzchar(gcc_path)) {
      return(NULL)
    }
  } else {
    gcc_path <- Sys.which("gcc")
    if (gcc_path == "") return(NULL)
  }
  if (debug) cat("gcc:", gcc_path, "\n")

  install_path2 <- dirname(dirname(dirname(gcc_path)))

  # If both install_paths are not equal
  if (tolower(install_path2) != tolower(install_path)) return(NULL)

  version <- installed_version(install_path, debug = debug)
  if (debug) cat("Version:", version, "\n")

  rtools(install_path, version)
}

scan_registry_for_rtools <- function(debug = FALSE) {
  if (debug) cat("Scanning registry...\n")

  keys <- NULL
  try(keys <- utils::readRegistry("SOFTWARE\\R-core\\Rtools",
    hive = "HCU", view = "32-bit", maxdepth = 2), silent = TRUE)
  if (is.null(keys))
    try(keys <- utils::readRegistry("SOFTWARE\\R-core\\Rtools",
      hive = "HLM", view = "32-bit", maxdepth = 2), silent = TRUE)
  if (is.null(keys)) return(NULL)

  rts <- vector("list", length(keys))

  for (i in seq_along(keys)) {
    version <- names(keys)[[i]]
    key <- keys[[version]]
    if (!is.list(key) || is.null(key$InstallPath)) next;
    install_path <- normalizePath(key$InstallPath,
      mustWork = FALSE, winslash = "/")

    if (debug) cat("Found", install_path, "for", version, "\n")
    rts[[i]] <- rtools(install_path, version)
  }

  Filter(Negate(is.null), rts)
}

installed_version <- function(path, debug) {
  if (!file.exists(file.path(path, "Rtools.txt"))) return(NULL)

  # Find the version path
  version_path <- file.path(path, "VERSION.txt")
  if (debug) {
    cat("VERSION.txt\n")
    cat(readLines(version_path), "\n")
  }
  if (!file.exists(version_path)) return(NULL)

  # Rtools is in the path -- now crack the VERSION file
  contents <- NULL
  try(contents <- readLines(version_path), silent = TRUE)
  if (is.null(contents)) return(NULL)

  # Extract the version
  contents <- gsub("^\\s+|\\s+$", "", contents)
  version_re <- "Rtools version (\\d\\.\\d+)\\.[0-9.]+$"

  if (!grepl(version_re, contents)) return(NULL)

  m <- regexec(version_re, contents)
  regmatches(contents, m)[[1]][2]
}

is_compatible <- function(rtools) {
  if (is.null(rtools)) return(FALSE)
  if (is.null(rtools$version)) return(FALSE)

  stopifnot(is.rtools(rtools))
  info <- version_info[[rtools$version]]
  if (is.null(info)) return(FALSE)

  r_version <- getRversion()
  r_version >= info$version_min && r_version <= info$version_max
}

rtools <- function(path, version, ...) {
  structure(list(version = version, path = path, ...), class = "rtools")
}
is.rtools <- function(x) inherits(x, "rtools")

# Rtools metadata --------------------------------------------------------------
rtools_url <- "http://cran.r-project.org/bin/windows/Rtools/"
version_info <- list(
  "2.11" = list(
    version_min = "2.10.0",
    version_max = "2.11.1",
    path = c("bin", "perl/bin", "MinGW/bin")
  ),
  "2.12" = list(
    version_min = "2.12.0",
    version_max = "2.12.2",
    path = c("bin", "perl/bin", "MinGW/bin", "MinGW64/bin")
  ),
  "2.13" = list(
    version_min = "2.13.0",
    version_max = "2.13.2",
    path = c("bin", "MinGW/bin", "MinGW64/bin")
  ),
  "2.14" = list(
    version_min = "2.13.0",
    version_max = "2.14.2",
    path = c("bin", "MinGW/bin", "MinGW64/bin")
  ),
  "2.15" = list(
    version_min = "2.14.2",
    version_max = "2.15.1",
    path = c("bin", "gcc-4.6.3/bin")
  ),
  "2.16" = list(
    version_min = "2.15.2",
    version_max = "3.0.0",
    path = c("bin", "gcc-4.6.3/bin")
  ),
  "3.0" = list(
    version_min = "2.15.2",
    version_max = "3.0.99",
    path = c("bin", "gcc-4.6.3/bin")
  ),
  "3.1" = list(
    version_min = "3.0.0",
    version_max = "3.1.99",
    path = c("bin", "gcc-4.6.3/bin")
  ),
  "3.2" = list(
    version_min = "3.1.0",
    version_max = "3.2.99",
    path = c("bin", "gcc-4.6.3/bin")
  ),
  "3.3" = list(
    version_min = "3.2.0",
    version_max = "3.3.99",
    path = if (using_gcc49()) {
      "bin"
    } else {
      c("bin", "gcc-4.6.3/bin")
    }
  ),
  "3.4" = list(
    version_min = "3.3.0",
    version_max = "3.4.99",
    path = if (using_gcc49()) {
      "bin"
    } else {
      c("bin", "gcc-4.6.3/bin")
    }
  )
)

rtools_needed <- function() {
  r_version <- getRversion()

  for(i in rev(seq_along(version_info))) {
    version <- names(version_info)[i]
    info <- version_info[[i]]
    ok <- r_version >= info$version_min && r_version <= info$version_max
    if (ok) return(paste("Rtools", version))
  }
  "the appropriate version of Rtools"
}

#' @rdname setup_rtools
#' @usage NULL
#' @export
find_rtools <- setup_rtools

if (.Platform$OS.type == "windows") {

context("Rtools tests")

with_rtools_path <- function(path, code) {
  bin <- c(file.path(path, "bin"), file.path(path, "gcc-4.6.3", "bin"))
  old <- set_path(bin)
  on.exit(set_path(old))
  force(code)
}

test_that("rtools found on path if present", {
  with_rtools_path("rtools-2.15", {
    rt <- scan_path_for_rtools(gcc49 = FALSE)
    expect_equal(rt$version, "2.15")
  })
})

test_that("out of date rtools is not compatible", {
  with_rtools_path("rtools-2.15", {
    rt <- scan_path_for_rtools(gcc49 = FALSE)
    expect_false(is_compatible(rt))
  })
})

test_that("rtools must be complete to be located", {
  with_rtools_path("rtools-no-gcc", {
    rt <- scan_path_for_rtools(gcc49 = FALSE)
    expect_equal(rt, NULL)
  })
})

test_that("correct path doesn't need fixing", {
  with_rtools_path("rtools-manual", {
    rt <- scan_path_for_rtools(gcc49 = FALSE)
    expect_equal(rt$version, NULL)
  })
})

with_mock(
  `devtools:::RCMD` = function(...) "invalid_path",
  `base::file.info` = function(x) data.frame(exe = if (grepl("32", x)) "win32" else "win64"), {

  test_that("gcc-493 directory structure is found on 32 bit", {
    withr::with_path(file.path("rtools-gcc493", "bin"), {
      rt <- scan_path_for_rtools(gcc49 = TRUE, arch = "32")
      expect_equal(rt$version, "3.3")
    })
  })

  test_that("gcc-493 directory structure is found on 64 bit", {
    withr::with_path(file.path("rtools-gcc493", "bin"), {
      rt <- scan_path_for_rtools(gcc49 = TRUE, arch = "64")
      expect_equal(rt$version, "3.3")
    })
  })

  test_that("gcc-493 directory structure like winbuilder is found on 32 bit", {
    withr::with_path(file.path("rtools-gcc493-winbuilder", "bin"), {
      rt <- scan_path_for_rtools(gcc49 = TRUE, arch = "32")
      expect_equal(rt$version, "3.3")
    })
  })
})}

# Call the Rcpp::compileAttributes function for a package (only do so if the
# package links to Rcpp and a recent enough version of Rcpp in installed).
compile_rcpp_attributes <- function(pkg) {

  # Only scan for attributes in packages explicitly linking to Rcpp
  if (links_to_rcpp(pkg)) {
    check_suggested("Rcpp")

    # We need to use use with_dir here to work around a regression in Rcpp
    # 0.12.11 (https://github.com/RcppCore/Rcpp/pull/697)
    withr_with_dir(pkg$path, {
      Rcpp::compileAttributes()
    })
  }
}

# Does this package have a compilation dependency on Rcpp?
links_to_rcpp <- function(pkg) {
  "Rcpp" %in% pkg_linking_to(pkg)
}

# Get the LinkingTo field of a package as a character vector
pkg_linking_to <- function(pkg) {
  parse_deps(pkg$linkingto)$name
}

# Create the package environment where exported objects will be copied to
attach_ns <- function(pkg = ".") {
  pkg <- as.package(pkg)
  nsenv <- ns_env(pkg)

  if (is_attached(pkg)) {
    stop("Package ", pkg$package, " is already attached.")
  }

  # This should be similar to attachNamespace
  pkgenv <- base::attach(NULL, name = pkg_env_name(pkg))
  attr(pkgenv, "path") <- getNamespaceInfo(nsenv, "path")
}

# Invoke namespace load actions. According to the documentation for setLoadActions
# these actions should be called at the end of processing of S4 metadata, after
# dynamically linking any libraries, the call to .onLoad(), if any, and caching
# method and class definitions, but before the namespace is sealed
run_ns_load_actions <- function(pkg = ".") {
  nsenv <- ns_env(pkg)
  actions <- methods::getLoadActions(nsenv)
  for (action in actions)
    action(nsenv)
}

# Copy over the objects from the namespace env to the package env
export_ns <- function(pkg = ".") {
  pkg <- as.package(pkg)
  nsenv <- ns_env(pkg)
  pkgenv <- pkg_env(pkg)
  nsInfo <- parse_ns_file(pkg)

  exports <- getNamespaceExports(nsenv)
  importIntoEnv(pkgenv, exports, nsenv, exports)

  # If lazydata is true, manually copy data objects in $lazydata to package
  # environment
  if (!is.null(pkg$lazydata) &&
      tolower(pkg$lazydata) %in% c("true", "yes")) {
    copy_env(src = nsenv$.__NAMESPACE__.$lazydata, dest = pkgenv)
  }
}


#' Return package environment
#'
#' This is an environment like \code{<package:pkg>}. The package
#' environment contains the exported objects from a package. It is
#' attached, so it is an ancestor of \code{R_GlobalEnv}.
#'
#' When a package is loaded the normal way, using \code{\link{library}},
#' this environment contains only the exported objects from the
#' namespace. However, when loaded with \code{\link{load_all}}, this
#' environment will contain all the objects from the namespace, unless
#' \code{load_all} is used with \code{export_all=FALSE}.
#'
#' If the package is not attached, this function returns \code{NULL}.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @keywords internal
#' @seealso \code{\link{ns_env}} for the namespace environment that
#'   all the objects (exported and not exported).
#' @seealso \code{\link{imports_env}} for the environment that contains
#'   imported objects for the package.
#' @export
pkg_env <- function(pkg = ".") {
  pkg <- as.package(pkg)
  name <- pkg_env_name(pkg)

  if (!is_attached(pkg)) return(NULL)

  as.environment(name)
}


# Generate name of package environment
# Contains exported objects
pkg_env_name <- function(pkg = ".") {
  pkg <- as.package(pkg)
  paste("package:", pkg$package, sep = "")
}


# Reports whether a package is loaded and attached
is_attached <- function(pkg = ".") {
  pkg_env_name(pkg) %in% search()
}

context("Namespace")

# Is e an ancestor environment of x?
is_ancestor_env <- function(e, x) {
  if (identical(e, x))
    return(TRUE)
  else if (identical(x, emptyenv()))
    return(FALSE)
  else
    is_ancestor_env(e, parent.env(x))
}

# Get parent environment n steps deep
parent_env <- function(e, n = 1) {
  if (n == 0)
    e
  else
    parent_env(parent.env(e), n-1)
}

test_that("Loaded namespaces have correct version", {
  load_all("testNamespace")
  expect_identical(c(version="0.1"), getNamespaceVersion("testNamespace"))
  unload("testNamespace")
})

test_that("Exported objects are visible from global environment", {

  # a is listed as an export in NAMESPACE, b is not. But with load_all(),
  # they should both be visible in the global env.
  load_all("testNamespace")
  expect_equal(a, 1)
  expect_equal(b, 2)
  unload("testNamespace")


  # With export_all = FALSE, only the listed export should be visible
  # in the global env.
  load_all("testNamespace", export_all = FALSE)
  expect_equal(a, 1)
  expect_false(exists("b"))
  unload("testNamespace")
})


test_that("Missing exports don't result in error", {
  expect_warning(load_all("testMissingNsObject"))
  nsenv <- ns_env("testMissingNsObject")
  expect_equal(nsenv$a, 1)
  unload("testMissingNsObject")
})


test_that("All objects are loaded into namespace environment", {
  load_all("testNamespace")
  nsenv <- ns_env("testNamespace")
  expect_equal(nsenv$a, 1)
  expect_equal(nsenv$b, 2)
  unload("testNamespace")
})


test_that("All objects are copied to package environment", {
  load_all("testNamespace")
  pkgenv <- pkg_env("testNamespace")
  expect_equal(pkgenv$a, 1)
  expect_equal(pkgenv$b, 2)
  unload("testNamespace")

  # With export_all = FALSE, only the listed export should be copied
  load_all("testNamespace", export_all = FALSE)
  pkgenv <- pkg_env("testNamespace")
  expect_equal(pkgenv$a, 1)
  expect_false(exists("b", envir = pkgenv))
  unload("testNamespace")
})


test_that("Unloading and reloading a package works", {
  load_all("testNamespace")
  expect_equal(a, 1)

  # A load_all() again without unloading shouldn't change things
  load_all("testNamespace")
  expect_equal(a, 1)

  # Unloading should remove objects
  unload("testNamespace")
  expect_false(exists('a'))

  # Loading again should work
  load_all("testNamespace")
  expect_equal(a, 1)

  # Loading with reset should work
  load_all("testNamespace", reset = TRUE)
  expect_equal(a, 1)

  unload("testNamespace")
})

test_that("Namespace, imports, and package environments have correct hierarchy", {
  load_all("testNamespace")

  pkgenv <- pkg_env("testNamespace")
  nsenv  <- ns_env("testNamespace")
  impenv <- imports_env("testNamespace")

  expect_identical(parent_env(nsenv, 1), impenv)
  expect_identical(parent_env(nsenv, 2), .BaseNamespaceEnv)
  expect_identical(parent_env(nsenv, 3), .GlobalEnv)

  # pkgenv should be an ancestor of the global environment
  expect_true(is_ancestor_env(pkgenv, .GlobalEnv))

  unload("testNamespace")
})


test_that("unload() removes package environments from search", {
  load_all("testNamespace")
  pkgenv <- pkg_env("testNamespace")
  nsenv   <- ns_env("testNamespace")
  unload("testNamespace")
  unload(inst("compiler"))
  unload(inst("bitops"))

  # Should report not loaded for package and namespace environments
  expect_false(is_attached("testNamespace"))
  expect_false(is_loaded("testNamespace"))

  # R's asNamespace function should error
  expect_error(asNamespace("testNamespace"))

  # pkgenv should NOT be an ancestor of the global environment
  # This is what makes the objects inaccessible from global env
  expect_false(is_ancestor_env(pkgenv, .GlobalEnv))
  # Another check of same thing
  expect_false(pkg_env_name("testNamespace") %in% search())

})


test_that("Environments have the correct attributes", {
  load_all("testNamespace")
  pkgenv <- pkg_env("testNamespace")
  impenv <- imports_env("testNamespace")

  # as.environment finds the same package environment
  expect_identical(pkgenv, as.environment("package:testNamespace"))

  # Check name attribute of package environment
  expect_identical(attr(pkgenv, "name"), "package:testNamespace")

  # Check path attribute of package environment
  if (has_tests()) {
    wd <- normalizePath(devtest("testNamespace"))
    expect_identical(wd, attr(pkgenv, "path"))
  }

  # Check name attribute of imports environment
  expect_identical(attr(impenv, "name"), "imports:testNamespace")

  unload("testNamespace")
})

parse_check_results <- function(path) {
  lines <- paste(readLines(path, warn = FALSE), collapse = "\n")

  # Strip off trailing NOTE and WARNING messages
  lines <- gsub("^NOTE: There was .*\n$", "", lines)
  lines <- gsub("^WARNING: There was .*\n$", "", lines)

  pieces <- strsplit(lines, "\n\\* ")[[1]]

  structure(
    list(
      errors =   pieces[grepl("... ERROR", pieces, fixed = TRUE)],
      warnings = pieces[grepl("... WARN", pieces, fixed = TRUE)],
      notes =    pieces[grepl("... NOTE", pieces, fixed = TRUE)]
    ),
    path = path,
    class = "check_results"
  )
}

signal_check_results <- function(x, on = c("none", "error", "warning", "note")) {
  has <- lapply(x, function(x) length(x) > 0)

  on <- match.arg(on)
  has_problem <- switch(on,
    none  =   FALSE,
    error =   has$errors,
    warning = has$errors | has$warnings,
    note =    has$errors | has$warnings | has$notes
  )

  if (has_problem) {
    stop(summarise_check_results(x), call. = FALSE)
  }
  invisible(TRUE)
}

#' @export
print.check_results <- function(x, ...) {
  message("R CMD check results")
  message(summarise_check_results(x))

  cat(format(x), "\n", sep = "")
  invisible(x)
}

#' @export
format.check_results <- function(x, ...) {
  checks <- trunc_middle(unlist(x))

  paste0(checks, collapse = "\n\n")
}

summarise_check_results <- function(x, colour = FALSE) {
  n <- lapply(x, length)
  paste0(
    show_count(n$errors, "error ", "errors", colour && n$errors > 0), " | ",
    show_count(n$warnings, "warning ", "warnings", colour && n$warnings > 0), " | ",
    show_count(n$notes, "note ", "notes")
  )
}

show_count <- function(n, singular, plural, is_error = FALSE) {
  out <- paste0(n, " ", ngettext(n, singular, plural))
  if (is_error && requireNamespace("crayon", quietly = TRUE)) {
    out <- crayon::red(out)
  }
  out
}


has_problems <- function(x) {
  length(x$results$errors) > 0 || length(x$results$warnings) > 0
}

first_problem <- function(x) {
  if (length(x$errors) > 0) {
    problem <- x$errors[[1]]
  } else if (length(x$warnings) > 0) {
    problem <- x$warnings[[1]]
  } else {
    return(NA_character_)
  }

  strsplit(problem, "\n", fixed = TRUE)[[1]][1]
}

trunc_middle <- function(x, n_max = 25, n_top = 10, n_bottom = 10) {
  trunc_middle_one <- function(x) {
    lines <- strsplit(x, "\n", fixed = TRUE)[[1]]
    nlines <- length(lines)

    if (nlines <= n_max)
      return(x)

    paste(c(
      lines[1:n_top],
      paste0("... ", length(lines) - n_top - n_bottom, " lines ..."),
      lines[(nlines - n_bottom):nlines]
    ), collapse = "\n")
  }

  vapply(x, trunc_middle_one, character(1), USE.NAMES = FALSE)
}

#' Parses R CMD check log file for ERRORs, WARNINGs and NOTEs
#'
#' Extracts check messages from the \code{00check.log} file generated by
#' \code{R CMD check}.
#'
#' @param path check path, e.g., value of the \code{check_dir} argument in a
#'   call to \code{\link{check}}
#' @param error,warning,note logical, indicates if errors, warnings and/or
#'   notes should be returned
#' @return a character vector with the relevant messages, can have length zero
#'   if no messages are found
#'
#' @seealso \code{\link{check}}, \code{\link{revdep_check}}
#' @export
check_failures <- function(path, error = TRUE, warning = TRUE, note = TRUE) {
  check_dir <- file.path(path, "00check.log")

  results <- parse_check_results(check_dir)

  c(
    if (error) results$errors,
    if (warning) results$warnings,
    if (note) results$notes
  )
}

context("Check")

test_that("sucessful check doesn't trigger error", {
  skip_on_cran()
  results <- check("testTest", quiet = TRUE)

  expect_error(signal_check_results(results), NA)
  expect_equal(
    summarise_check_results(results),
    "0 errors | 0 warnings | 0 notes",
    fixed = TRUE
  )
})

test_that("check with NOTES captured", {
  skip_on_cran()

  results <- parse_check_results("check-results-note.log")

  expect_error(signal_check_results(results), NA)
  expect_error(
    signal_check_results(results, "note"),
    "0 errors | 0 warnings | 2 notes",
    fixed = TRUE
  )
})

#' Unload and reload package.
#'
#' This attempts to unload and reload a package. If the package is not loaded
#' already, it does nothing. It's not always possible to cleanly unload a
#' package: see the caveats in \code{\link{unload}} for some of the
#' potential failure points. If in doubt, restart R and reload the package
#' with \code{\link{library}}.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @param quiet if \code{TRUE} suppresses output from this function.
#' @examples
#' \dontrun{
#' # Reload package that is in current directory
#' reload(".")
#'
#' # Reload package that is in ./ggplot2/
#' reload("ggplot2/")
#'
#' # Can use inst() to find the package path
#' # This will reload the installed ggplot2 package
#' reload(inst("ggplot2"))
#' }
#' @export
reload <- function(pkg = ".", quiet = FALSE) {
  pkg <- as.package(pkg)

  if (is_attached(pkg)) {
    if (!quiet) message("Reloading installed ", pkg$package)
    unload(pkg)
    require(pkg$package, character.only = TRUE, quietly = TRUE)
  }
}

# Copy all objects from one environment to another.
#
# Returns the destination environment
#
# @param dest Destination environment. If not specified, create a new
#   environment.
# @param ignore Names of objects that should not be copied.
copy_env <- function(src, dest = new.env(parent = emptyenv()),
  ignore = NULL) {

  srclist <- as.list(src, all.names = TRUE)
  srclist <- srclist[ !(names(srclist) %in% ignore) ]
  list2env(srclist, envir = dest)

  dest
}

#' Return imports environment for a package
#'
#' Contains objects imported from other packages. Is the parent of the
#' package namespace environment, and is a child of <namespace:base>,
#' which is a child of R_GlobalEnv.
#' @keywords internal
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information.
#' @seealso \code{\link{ns_env}} for the namespace environment that
#'   all the objects (exported and not exported).
#' @seealso \code{\link{pkg_env}} for the attached environment that contains
#'   the exported objects.
#' @export
imports_env <- function(pkg = ".") {
  pkg <- as.package(pkg)

  if (!is_loaded(pkg)) {
    stop("Namespace environment must be created before accessing imports environment.")
  }

  env <- parent.env(ns_env(pkg))

  if (attr(env, 'name') != imports_env_name(pkg)) {
    stop("Imports environment does not have attribute 'name' with value ",
      imports_env_name(pkg),
      ". This probably means that the namespace environment was not created correctly.")
  }

  env
}


# Generate name of package imports environment
# Contains exported objects
imports_env_name <- function(pkg = ".") {
  pkg <- as.package(pkg)
  paste("imports:", pkg$package, sep = "")
}


#' Load all of the imports for a package
#'
#' The imported objects are copied to the imports environment, and are not
#' visible from R_GlobalEnv. This will automatically load (but not attach)
#' the dependency packages.
#'
#' @keywords internal
load_imports <- function(pkg = ".") {
  pkg <- as.package(pkg)

  # Get data frame of dependency names and versions
  deps <- parse_deps(pkg$imports)
  if (is.null(deps) || nrow(deps) == 0) return(invisible())

  # If we've already loaded imports, don't load again (until load_all
  # is run with reset=TRUE). This is to avoid warnings when running
  # process_imports()
  if (length(ls(imports_env(pkg))) > 0) return(invisible(deps))

  mapply(check_dep_version, deps$name, deps$version, deps$compare)

  process_imports(pkg)

  invisible(deps)
}

# Load imported objects
# The code in this function is taken and adapted from base::loadNamespace
# Setup variables were added and the for loops put in a tryCatch block
# https://github.com/wch/r-source/blob/tags/R-3-3-0/src/library/base/R/namespace.R#L397-L427

# This wraps the inner for loop iterations in a tryCatch
wrap_inner_loop <- function(x) {
  inner <- x[[4]]
  x[[4]] <- call("tryCatch", error = quote(warning), inner)
  x
}

onload_assign("process_imports", {
  make_function(alist(pkg = "."),
    bquote({
      package <- pkg$name
      vI <- ("tools" %:::% ".split_description")(("tools" %:::% ".read_description")(file.path(pkg$path, "DESCRIPTION")))$Imports
      nsInfo <- parse_ns_file(pkg)
      ns <- ns_env(pkg)
      lib.loc <- NULL
      .(for1)
      .(for2)
      .(for3)
    }, list(
        for1 = wrap_inner_loop(
          extract_lang(body(loadNamespace), comp_lang, y = quote(for(i in nsInfo$imports) NULL), idx = 1:3)),

        for2 = wrap_inner_loop(extract_lang(body(loadNamespace),
          comp_lang, y = quote(for(imp in nsInfo$importClasses) NULL), idx = 1:3)),

        for3 = wrap_inner_loop(extract_lang(body(loadNamespace),
          comp_lang, y = quote(for(imp in nsInfo$importMethods) NULL), idx = 1:3))
        )), asNamespace("devtools"))
})

#' Unload a package
#'
#' This function attempts to cleanly unload a package, including unloading
#' its namespace, deleting S4 class definitions and unloading any loaded
#' DLLs. Unfortunately S4 classes are not really designed to be cleanly
#' unloaded, and so we have to manually modify the class dependency graph in
#' order for it to work - this works on the cases for which we have tested
#' but there may be others.  Similarly, automated DLL unloading is best tested
#' for simple scenarios (particularly with \code{useDynLib(pkgname)} and may
#' fail in other cases. If you do encounter a failure, please file a bug report
#' at \url{http://github.com/hadley/devtools/issues}.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#'
#' @examples
#' \dontrun{
#' # Unload package that is in current directory
#' unload(".")
#'
#' # Unload package that is in ./ggplot2/
#' unload("ggplot2/")
#'
#' # Can use inst() to find the path of an installed package
#' # This will load and unload the installed ggplot2 package
#' library(ggplot2)
#' unload(inst("ggplot2"))
#' }
#' @export
unload <- function(pkg = ".") {

  pkg <- as.package(pkg)

  if (pkg$package == "compiler") {
    # Disable JIT compilation as it could interfere with the compiler
    # unloading. Also, if the JIT was kept enabled, it would cause the
    # compiler package to be loaded again soon, anyway. Note if we
    # restored the JIT level after the unloading, the call to
    # enableJIT itself would load the compiler again.
    oldEnable <- compiler::enableJIT(0)
    if (oldEnable != 0) {
      warning("JIT automatically disabled when unloading the compiler.")
    }
  }

  # This is a hack to work around unloading devtools itself. The unloading
  # process normally makes other devtools functions inaccessible,
  # resulting in "Error in unload(pkg) : internal error -3 in R_decompress1".
  # If we simply force them first, then they will remain available for use
  # later.
  if (pkg$package == "devtools") {
    eapply(ns_env(pkg), force, all.names = TRUE)
  }

  # If the package was loaded with devtools, any s4 classes that were created
  # by the package need to be removed in a special way.
  if (!is.null(dev_meta(pkg$package))) {
    remove_s4_classes(pkg)
  }


  if (pkg$package %in% loadedNamespaces()) {
    # unloadNamespace will throw an error if it has trouble unloading.
    # This can happen when there's another package that depends on the
    # namespace.
    # unloadNamespace will also detach the package if it's attached.
    #
    # unloadNamespace calls onUnload hook and .onUnload
    try(unloadNamespace(pkg$package), silent = TRUE)

  } else {
    stop("Package ", pkg$package, " not found in loaded packages or namespaces")
  }

  # Sometimes the namespace won't unload with detach(), like when there's
  # another package that depends on it. If it's still around, force it
  # to go away.
  # loadedNamespaces() and unloadNamespace() often don't work here
  # because things can be in a weird state.
  if (!is.null(.getNamespace(pkg$package))) {
    message("unloadNamespace(\"", pkg$package,
      "\") not successful, probably because another loaded package depends on it.",
      "Forcing unload. If you encounter problems, please restart R.")
    unregister_namespace(pkg$package)
  }

  # Clear so that loading the package again will re-read all files
  clear_cache()

  # Do this after detach, so that packages that have an .onUnload function
  # which unloads DLLs (like MASS) won't try to unload the DLL twice.
  unload_dll(pkg)
}

# This unloads dlls loaded by either library() or load_all()
unload_dll <- function(pkg = ".") {
  pkg <- as.package(pkg)

  # Always run garbage collector to force any deleted external pointers to
  # finalise
  gc()

  # Special case for devtools - don't unload DLL because we need to be able
  # to access nsreg() in the DLL in order to run makeNamespace. This means
  # that changes to compiled code in devtools can't be reloaded with
  # load_all -- it requires a reinstallation.
  if (pkg$package == "devtools") {
    return(invisible())
  }

  pkglibs <- loaded_dlls(pkg)

  for (lib in pkglibs) {
    dyn.unload(lib[["path"]])
  }

  # Remove the unloaded dlls from .dynLibs()
  libs <- .dynLibs()
  .dynLibs(libs[!(libs %in% pkglibs)])

  invisible()
}

#' Initialise a git repository.
#'
#' @param message Message to use for first commit.
#' @param pkg Path to package. See \code{\link{as.package}} for more
#'   information.
#' @family git infrastructure
#' @export
#' @examples
#' \dontrun{use_git()}
use_git <- function(message = "Initial commit", pkg = ".") {
  use_git_with_config(message = message, pkg = pkg)
}

use_git_with_config <- function(message, pkg, add_user_config = FALSE, quiet = FALSE) {
  pkg <- as.package(pkg)

  if (uses_git(pkg$path)) {
    message("* Git is already initialized")
    return(invisible())
  }

  if (!quiet) {
    message("* Initialising repo")
  }
  r <- git2r::init(pkg$path)

  if (add_user_config) {
    git2r::config(r, global = FALSE, user.name = "user", user.email = "user@email.xx")
  }

  use_git_ignore(c(".Rproj.user", ".Rhistory", ".RData"), pkg = pkg, quiet = quiet)

  if (!quiet) {
    message("* Adding files and committing")
  }
  paths <- unlist(git2r::status(r))
  git2r::add(r, paths)
  git2r::commit(r, message)

  invisible()
}

#' Connect a local repo with GitHub.
#'
#' If the current repo does not use git, calls \code{\link{use_git}}
#' automatically. \code{\link{use_github_links}} is called to populate the
#' \code{URL} and \code{BugReports} fields of DESCRIPTION.
#'
#' @section Authentication:
#'
#'   A new GitHub repo will be created via the GitHub API, therefore you must
#'   provide a GitHub personal access token (PAT) via the argument
#'   \code{auth_token}, which defaults to the value of the \code{GITHUB_PAT}
#'   environment variable. Obtain a PAT from
#'   \url{https://github.com/settings/tokens}. The "repo" scope is required
#'   which is one of the default scopes for a new PAT.
#'
#'   The argument \code{protocol} reflects how you wish to authenticate with
#'   GitHub for this repo in the long run. For either \code{protocol}, a remote
#'   named "origin" is created, an initial push is made using the specified
#'   \code{protocol}, and a remote tracking branch is set. The URL of the
#'   "origin" remote has the form \code{git@@github.com:<USERNAME>/<REPO>.git}
#'   (\code{protocol = "ssh"}, the default) or
#'   \code{https://github.com/<USERNAME>/<REPO>.git} (\code{protocol =
#'   "https"}). For \code{protocol = "ssh"}, it is assumed that public and
#'   private keys are in the default locations, \code{~/.ssh/id_rsa.pub} and
#'   \code{~/.ssh/id_rsa}, respectively, and that \code{ssh-agent} is configured
#'   to manage any associated passphrase.  Alternatively, specify a
#'   \code{\link[git2r]{cred_ssh_key}} object via the \code{credentials}
#'   parameter.
#'
#' @inheritParams use_git
#' @param auth_token Provide a personal access token (PAT) from
#'   \url{https://github.com/settings/tokens}. Defaults to the \code{GITHUB_PAT}
#'   environment variable.
#' @param private If \code{TRUE}, creates a private repository.
#' @param host GitHub API host to use. Override with the endpoint-root for your
#'   GitHub enterprise instance, for example,
#'   "https://github.hostname.com/api/v3".
#' @param protocol transfer protocol, either "ssh" (the default) or "https"
#' @param credentials A \code{\link[git2r]{cred_ssh_key}} specifying specific
#' ssh credentials or NULL for default ssh key and ssh-agent behaviour.
#' Default is NULL.
#' @family git infrastructure
#' @export
#' @examples
#' \dontrun{
#' ## to use default ssh protocol
#' create("testpkg")
#' use_github(pkg = "testpkg")
#'
#' ## or use https
#' create("testpkg2")
#' use_github(pkg = "testpkg2", protocol = "https")
#' }
use_github <- function(auth_token = github_pat(), private = FALSE, pkg = ".",
                       host = "https://api.github.com",
                       protocol = c("ssh", "https"), credentials = NULL) {

  if (is.null(auth_token)) {
    stop("GITHUB_PAT required to create new repo")
  }

  protocol <- match.arg(protocol)

  pkg <- as.package(pkg)
  use_git(pkg = pkg)

  if (uses_github(pkg$path)) {
    message("* GitHub is already initialized")
    return(invisible())
  }

  message("* Checking title and description")
  message("  Title: ", pkg$title)
  message("  Description: ", pkg$description)
  if (yesno("Are title and description ok?")) {
    return(invisible())
  }

  message("* Creating GitHub repository")
  create <-
    github_POST(
      "user/repos",
      pat = auth_token,
      body = list(
        name = jsonlite::unbox(pkg$package),
        description = jsonlite::unbox(gsub("\n", " ", pkg$title)),
        private = jsonlite::unbox(private)
      ),
      host = host
    )

  message("* Adding GitHub remote")
  r <- git2r::repository(pkg$path)
  origin_url <- switch(protocol, https = create$clone_url, ssh = create$ssh_url)
  git2r::remote_add(r, "origin", origin_url)

  message("* Adding GitHub links to DESCRIPTION")
  use_github_links(pkg$path, auth_token = auth_token, host = host)
  if (git_uncommitted(pkg$path)) {
    git2r::add(r, "DESCRIPTION")
    git2r::commit(r, "Add GitHub links to DESCRIPTION")
  }

  message("* Pushing to GitHub and setting remote tracking branch")
  if (protocol == "ssh") {
    ## [1] push via ssh required for success setting remote tracking branch
    ## [2] to get passphrase from ssh-agent, you must use NULL credentials
    git2r::push(r, "origin", "refs/heads/master", credentials = credentials)
  } else { ## protocol == "https"
    ## in https case, when GITHUB_PAT is passed as password,
    ## the username is immaterial, but git2r doesn't know that
    cred <- git2r::cred_user_pass("EMAIL", auth_token)
    git2r::push(r, "origin", "refs/heads/master", credentials = cred)
  }
  git2r::branch_set_upstream(git2r::head(r), "origin/master")

  message("* View repo at ", create$html_url)

  invisible(NULL)
}


#' Add a git hook.
#'
#' @param hook Hook name. One of "pre-commit", "prepare-commit-msg",
#'   "commit-msg", "post-commit", "applypatch-msg", "pre-applypatch",
#'   "post-applypatch", "pre-rebase", "post-rewrite", "post-checkout",
#'   "post-merge", "pre-push", "pre-auto-gc".
#' @param script Text of script to run
#' @inheritParams use_git
#' @export
#' @family git infrastructure
#' @keywords internal
use_git_hook <- function(hook, script, pkg = ".") {
  pkg <- as.package(pkg)

  git_dir <- file.path(pkg$path, ".git")
  if (!file.exists(git_dir)) {
    stop("This project doesn't use git", call. = FALSE)
  }

  hook_dir <- file.path(git_dir, "hooks")
  if (!file.exists(hook_dir)) {
    dir.create(hook_dir)
  }

  hook_path <- file.path(hook_dir, hook)
  writeLines(script, hook_path)
  Sys.chmod(hook_path, "0744")
}


use_git_ignore <- function(ignores, directory = ".", pkg = ".", quiet = FALSE) {
  pkg <- as.package(pkg)

  paths <- paste0("`", ignores, "`", collapse = ", ")
  if (!quiet) {
    message("* Adding ", paths, " to ", file.path(directory, ".gitignore"))
  }

  path <- file.path(pkg$path, directory, ".gitignore")
  union_write(path, ignores)

  invisible(TRUE)
}

#' Add GitHub links to DESCRIPTION.
#'
#' Populates the URL and BugReports fields of DESCRIPTION with
#' \code{https://github.com/<USERNAME>/<REPO>} AND
#' \code{https://github.com/<USERNAME>/<REPO>/issues}, respectively, unless
#' those fields already exist.
#'
#' @inheritParams use_git
#' @param auth_token Provide a personal access token (PAT) from
#'   \url{https://github.com/settings/tokens}. Defaults to the \code{GITHUB_PAT}
#'   environment variable.
#' @param host GitHub API host to use. Override with the endpoint-root for your
#'   GitHub enterprise instance, for example,
#'   "https://github.hostname.com/api/v3".
#' @family git infrastructure
#' @keywords internal
#' @export
use_github_links <- function(pkg = ".", auth_token = github_pat(),
                             host = "https://api.github.com") {

  if (!uses_github(pkg)) {
    stop("Cannot detect that package already uses GitHub.\n",
         "You might want to run use_github().")
  }

  gh_info <- github_info(pkg)
  pkg <- as.package(pkg)

  desc_path <- file.path(pkg$path, "DESCRIPTION")
  desc <- new_desc <- read_dcf(desc_path)

  path_to_repo <- paste("repos", gh_info$fullname, sep = "/")
  res <- github_GET(path = path_to_repo, pat = auth_token, host = host)
  github_URL <- res$html_url

  fill <- function(d, f, filler) {
    if (is.null(d[[f]]) || identical(d[[f]], "")) {
      d[[f]] <- filler
    } else {
      message("Existing ", f, " field found and preserved")
    }
    d
  }
  new_desc <- fill(new_desc, "URL", github_URL)
  new_desc <- fill(new_desc, "BugReports", file.path(github_URL, "issues"))

  if (!identical(desc, new_desc))
    write_dcf(desc_path, new_desc)

  new_desc[c("URL", "BugReports")]
}

#' Return the namespace environment for a package.
#'
#' Contains all (exported and non-exported) objects, and is a descendent of
#' \code{R_GlobalEnv}. The hierarchy is \code{<namespace:pkg>},
#' \code{<imports:pkg>}, \code{<namespace:base>}, and then
#' \code{R_GlobalEnv}.
#'
#' If the package is not loaded, this function returns \code{NULL}.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @keywords internal
#' @seealso \code{\link{pkg_env}} for the attached environment that
#'   contains the exported objects.
#' @seealso \code{\link{imports_env}} for the environment that contains
#'   imported objects for the package.
#' @export
ns_env <- function(pkg = ".") {
  pkg <- as.package(pkg)

  if (!is_loaded(pkg)) return(NULL)

  asNamespace(pkg$package)
}


# Create the namespace environment for a package
create_ns_env <- function(pkg = ".") {
  pkg <- as.package(pkg)

  if (is_loaded(pkg)) {
    stop("Namespace for ", pkg$package, " already exists.")
  }

  env <- makeNamespace(pkg$package, pkg$version)
  methods::setPackageName(pkg$package, env)
  # Create devtools metadata in namespace
  create_dev_meta(pkg$package)

  setNamespaceInfo(env, "path", pkg$path)
  setup_ns_imports(pkg)

  env
}

# This is taken directly from base::loadNamespace()
# https://github.com/wch/r-source/blob/tags/R-3-3-0/src/library/base/R/namespace.R#L235-L258
onload_assign("makeNamespace",
  eval(
    modify_lang(
      extract_lang(body(loadNamespace),

        # Find makeNamespace definition
        comp_lang, y = quote(makeNamespace <- NULL), idx = 1:2)[[3]],

      # Replace call to .Internal(registerNamespace()) is replaced by a call to
      # register_namespace
      function(x) {
        if (comp_lang(x, quote(.Internal(registerNamespace(name, env))))) {
          quote(register_namespace(name, env))
        } else {
          x
        }
      }))
)

# Read the NAMESPACE file and set up the imports metdata.
# (which is stored in .__NAMESPACE__.)
setup_ns_imports <- function(pkg = ".") {
  pkg <- as.package(pkg)
  nsInfo <- parse_ns_file(pkg)
  setNamespaceInfo(pkg$package, "imports", nsInfo$imports)
}


# Read the NAMESPACE file and set up the exports metdata. This must be
# run after all the objects are loaded into the namespace because
# namespaceExport throw errors if the objects are not present.
setup_ns_exports <- function(pkg = ".", export_all = FALSE) {
  pkg <- as.package(pkg)
  nsInfo <- parse_ns_file(pkg)
  nsenv <- ns_env(pkg)

  if (export_all) {
    exports <- ls(nsenv, all.names = TRUE)
    # Make sure to re-export objects that are imported from other packages but
    # not copied.
    exports <- union(exports, nsInfo$exports)

    # List of things to ignore is from loadNamespace. There are also a
    # couple things to ignore from devtools.
    ignoreidx <- exports %in% c( ".__NAMESPACE__.",
      ".__S3MethodsTable__.", ".packageName", ".First.lib", ".onLoad",
      ".onAttach", ".conflicts.OK", ".noGenerics",
      ".__DEVTOOLS__", ".cache")
    exports <- exports[!ignoreidx]

  } else {
    # This code is from base::loadNamespace
    exports <- nsInfo$exports
    for (p in nsInfo$exportPatterns)
      exports <- c(ls(nsenv, pattern = p, all.names = TRUE), exports)
    exports <- add_classes_to_exports(ns = nsenv, package = pkg$package,
      exports = exports, nsInfo = nsInfo)
  }

  # Don't try to export objects that are missing from the namespace and imports
  ns_and_imports <- c(ls(nsenv, all.names = TRUE),
                      ls(imports_env(pkg), all.names = TRUE))
  extra_exports <- setdiff(exports, ns_and_imports)

  if (length(extra_exports) > 0) {
    warning("Objects listed as exports, but not present in namespace: ",
            paste(extra_exports, collapse = ", "))
    exports <- intersect(ns_and_imports, exports)
  }
  # Update the exports metadata for the namespace with base::namespaceExport
  # It will throw warnings if objects are already listed in the exports
  # metadata, so catch those warnings and ignore them.
  suppressWarnings(namespaceExport(nsenv, exports))

  invisible()
}

# Lookup S4 classes for export
#
# This function uses code from base::loadNamespace. Previously this code was
# copied directly, now it is dynamically looked up instead, to prevent drift as
# base::loadNamespace changes.
onload_assign("add_classes_to_exports",
  make_function(alist(ns =, package =, exports =, nsInfo =),
    call("{",
      extract_lang(
        f = comp_lang,
        y = quote(if (.isMethodsDispatchOn() && .hasS4MetaData(ns) && !identical(package, "methods")) NULL),
        idx = 1:2,
        modify_lang(body(base::loadNamespace), strip_internal_calls, "methods")),
      quote(exports)),
    asNamespace("methods")))

#' Parses the NAMESPACE file for a package
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @examples
#' if (has_tests()) {
#' parse_ns_file(devtest("testLoadHooks"))
#' }
#' @keywords internal
#' @export
parse_ns_file <- function(pkg = ".") {
  pkg <- as.package(pkg)

  parseNamespaceFile(basename(pkg$path), dirname(pkg$path),
    mustExist = FALSE)
}


# Register the S3 methods for this package
register_s3 <- function(pkg = ".") {
  pkg <- as.package(pkg)
  nsInfo <- parse_ns_file(pkg)

  # Adapted from loadNamespace
  registerS3methods(nsInfo$S3methods, pkg$package, ns_env(pkg))
}


# Reports whether a package is loaded into a namespace. It may be
# attached or not attached.
is_loaded <- function(pkg = ".") {
  pkg <- as.package(pkg)
  pkg$package %in% loadedNamespaces()
}


# Returns the namespace registry
ns_registry <- function() {
  (get(".Internal", envir = baseenv(), mode = "function"))(getNamespaceRegistry())
}
# To avoid a note about getNamespaceRegistry being missing
utils::globalVariables("getNamespaceRegistry")

# Register a namespace
register_namespace <- function(name = NULL, env = NULL) {
  # Be careful about what we allow
  if (!is.character(name) || name == "" || length(name) != 1)
    stop("'name' must be a non-empty character string.")

  if (!is.environment(env))
    stop("'env' must be an environment.")

  if (name %in% loadedNamespaces())
    stop("Namespace ", name, " is already registered.")

  # Add the environment to the registry
  nsr <- ns_registry()
  nsr[[name]] <- env

  env
}


# unregister a namespace - should be used only if unloadNamespace()
# fails for some reason
unregister_namespace <- function(name = NULL) {
  # Be careful about what we allow
  if (!is.character(name) || name == "" || length(name) != 1)
    stop("'name' must be a non-empty character string.")

  if (!(name %in% loadedNamespaces()))
    stop(name, " is not a registered namespace.")

  # Remove the item from the registry
  do.call(rm, args = list(name, envir = ns_registry()))
  invisible()
}

#' Install a package from a git repository
#'
#' It is vectorised so you can install multiple packages with
#' a single command. You do not need to have git installed.
#'
#' @param url Location of package. The url should point to a public or
#'   private repository.
#' @param subdir A sub-directory within a git repository that may
#'   contain the package we are interested in installing.
#' @param branch Name of branch or tag to use, if not master.
#' @param credentials A git2r credentials object passed through
#'   to \code{\link[git2r]{clone}}.
#' @param ... passed on to \code{\link{install}}
#' @inheritParams install_url
#' @export
#' @family package installation
#' @examples
#' \dontrun{
#' install_git("git://github.com/hadley/stringr.git")
#' install_git("git://github.com/hadley/stringr.git", branch = "stringr-0.2")
#'}
install_git <- function(url, subdir = NULL, branch = NULL, credentials = NULL, quiet=FALSE, ...) {

  remotes <- lapply(url, git_remote, subdir = subdir,
                    branch = branch, credentials=credentials)

  install_remotes(remotes, quiet = quiet, ...)
}

git_remote <- function(url, subdir = NULL, branch = NULL, credentials=NULL) {
  remote("git",
    url = url,
    subdir = subdir,
    branch = branch,
    credentials = credentials
  )
}

#' @export
remote_download.git_remote <- function(x, quiet = FALSE) {
  if (!quiet) {
    message("Downloading git repo ", x$url)
  }

  bundle <- tempfile()
  git2r::clone(x$url, bundle, credentials=x$credentials, progress = FALSE)

  if (!is.null(x$branch)) {
    r <- git2r::repository(bundle)
    git2r::checkout(r, x$branch)
  }

  bundle
}

#' @export
remote_metadata.git_remote <- function(x, bundle = NULL, source = NULL) {
  if (!is.null(bundle)) {
    r <- git2r::repository(bundle)
    sha <- git_repo_sha1(r)
  } else {
    sha <- NULL
  }

  list(
    RemoteType = "git",
    RemoteUrl = x$url,
    RemoteSubdir = x$subdir,
    RemoteRef = x$ref,
    RemoteSha = sha
  )
}

#' @export
remote_package_name.git_remote <- function(remote, ...) {

  tmp <- tempfile()
  on.exit(unlink(tmp))
  description_path <- paste0(collapse = "/", c(remote$subdir, "DESCRIPTION"))

  # Try using git archive --remote to retrieve the DESCRIPTION, if the protocol
  # or server doesn't support that return NULL
  res <- try(silent = TRUE,
    system_check(git_path(),
      args = c("archive", "-o", tmp, "--remote", remote$url,
        if (is.null(remote$branch)) "HEAD" else remote$branch,
        description_path),
      quiet = TRUE))

  if (inherits(res, "try-error")) {
    return(NA)
  }

  # git archive return a tar file, so extract it to tempdir and read the DCF
  utils::untar(tmp, files = description_path, exdir = tempdir())

  read_dcf(file.path(tempdir(), description_path))$Package
}

#' @export
remote_sha.git_remote <- function(remote, ...) {
  # If the remote ref is the same as the sha it is a pinned commit so just
  # return that.
  if (!is.null(remote$ref) &&
    grepl(paste0("^", remote$ref), remote$sha)) {
    return(remote$sha)
  }

  tryCatch({
    res <- git2r::remote_ls(remote$url, credentials=remote$credentials, ...)

    branch <- remote$branch %||% "master"

    found <- grep(pattern = paste0("/", branch), x = names(res))

    if (length(found) == 0) {
      return(NA)
    }

    unname(res[found[1]])
  }, error = function(e) NA_character_)
}

#' @export
format.git_remote <- function(x, ...) {
  "Git"
}

install_min <- function(pkg = ".", dest, components = NULL, args = NULL, quiet = FALSE) {
  pkg <- as.package(pkg)
  stopifnot(is.character(dest), length(dest) == 1, file.exists(dest))

  poss <- c("R", "data", "help", "demo", "inst", "docs", "exec", "libs")
  if (!is.null(components)) {
    components <- match.arg(components, poss, several.ok = TRUE)
  }
  no <- setdiff(poss, components)
  no_args <- paste0("--no-", no)

  RCMD("INSTALL", c(
    shQuote(pkg$path),
    paste("--library=", shQuote(dest), sep = ""),
    no_args,
    "--no-multiarch",
    "--no-test-load",
    args
  ), quiet = quiet)

  invisible(file.path(dest, pkg$package))
}

#' Use roxygen to document a package.
#'
#' This function is a wrapper for the \code{\link[roxygen2]{roxygenize}()}
#' function from the roxygen2 package. See the documentation and vignettes of
#' that package to learn how to use roxygen.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @param clean,reload Deprecated.
#' @inheritParams roxygen2::roxygenise
#' @seealso \code{\link[roxygen2]{roxygenize}},
#'   \code{browseVignettes("roxygen2")}
#' @export
document <- function(pkg = ".", clean = NULL, roclets = NULL, reload = TRUE) {
  check_suggested("roxygen2")
  if (!missing(clean)) {
    warning("`clean` is deprecated: roxygen2 now automatically cleans up",
      call. = FALSE)
  }
  if (!missing(reload)) {
    warning("`reload` is deprecated: code is now always reloaded", call. = FALSE)
  }

  pkg <- as.package(pkg)
  message("Updating ", pkg$package, " documentation")

  load_all(pkg)

  if (packageVersion("roxygen2") > "4.1.1") {
    roclets <- roclets %||% roxygen2::load_options(pkg$path)$roclets
    # collate updated by load_all()
    roclets <- setdiff(roclets, "collate")
  }

  withr::with_envvar(r_env_vars(),
    roxygen2::roxygenise(pkg$path, roclets = roclets, load_code = ns_env)
  )

  clear_topic_index(pkg)
  invisible()
}

run_example <- function(path, show = TRUE, test = FALSE, run = TRUE, env = new.env(parent = globalenv())) {
  check_suggested("evaluate")
  rd <- tools::parse_Rd(path)

  ex <- rd[rd_tags(rd) == "examples"]
  code <- process_ex(ex, show = show, test = test, run = run)
  if (is.null(code)) return()

  rule("Running examples in ", basename(path))

  code <- paste(code, collapse = "")
  results <- evaluate::evaluate(code, env)
  replay_stop(results)
}

process_ex <- function(rd, show = TRUE, test = FALSE, run = TRUE) {
  tag <- rd_tag(rd)

  recurse <- function(rd) {
    unlist(lapply(rd, process_ex, show = show, test = test, run = run))
  }

  if (is.null(tag) || tag == "examples") {
    return(recurse(rd))
  }

  # Base case
  if (tag %in% c("RCODE", "COMMENT", "TEXT", "VERB")) {
    return(rd[[1]])
  }

  # Conditional execution
  if (tag %in% c("dontshow", "dontrun", "donttest", "testonly")) {
    out <- recurse(rd)

    if ((tag == "dontshow" && show) ||
        (tag == "dontrun" && run) ||
        (tag == "donttest" && test) ||
        (tag == "testonly" && !test)) {
      type <- paste("\n# ", toupper(tag), "\n", sep = "")
      out <- c(type, out)
      out <- gsub("\n", "\n# ", out)
    }
    return(out)
  }

  if (tag %in% c("dots", "ldots")) {
    return("...")
  }

  warning("Unknown tag ", tag, call. = FALSE)
  tag
}


rd_tag <- function(x) {
  tag <- attr(x, "Rd_tag")
  if (is.null(tag)) return()

  gsub("\\", "", tag, fixed = TRUE)
}

rd_tags <- function(x) {
  vapply(x, function(x) rd_tag(x) %||% "", character(1))
}

remove_tag <- function(x) {
  attr(x, "Rd_tag") <- NULL
  x
}

replay_stop <- function(x) UseMethod("replay_stop", x)
#' @export
replay_stop.error <- function(x) {
  stop(quiet_error(x$message, x$call))
}
#' @export
replay_stop.default <- function(x) evaluate::replay(x)
#' @export
replay_stop.list <- function(x) {
  invisible(lapply(x, replay_stop))
}

quiet_error <- function(message, call = NULL) {
  structure(list(message = as.character(message), call = call),
    class = c("quietError", "error", "condition"))
}
#' @export
as.character.quietError <- function(x, ...) {
  if (is.null(x$call)) {
    paste("Error: ", x$message, sep = "")
  } else {
    call <- deparse(x$call)
    paste("Error in ", call, ": ", x$message, sep = "")
  }
}

#' Check a package from CRAN.
#'
#' Internal function used to power \code{\link{revdep_check}()}.
#'
#' This function does not clean up after itself, but does work in a
#' session-specific temporary directory, so all files will be removed
#' when your current R session ends.
#'
#' @param pkgs Vector of package names - note that unlike other \pkg{devtools}
#'   functions this is the name of a CRAN package, not a path.
#' @param libpath Path to library to store dependencies packages - if you
#'   you're doing this a lot it's a good idea to pick a directory and stick
#'   with it so you don't have to download all the packages every time.
#' @param srcpath Path to directory to store source versions of dependent
#'   packages - again, this saves a lot of time because you don't need to
#'   redownload the packages every time you run the package.
#' @param check_libpath Path to library used for checking, should contain
#'   the top-level library from \code{libpath}.
#' @param bioconductor Include bioconductor packages in checking?
#' @param type binary Package type to test (source, mac.binary etc). Defaults
#'   to the same type as \code{\link{install.packages}()}.
#' @param threads Number of concurrent threads to use for checking.
#'   It defaults to the option \code{"Ncpus"} or \code{1} if unset.
#' @param check_dir,install_dir Directory to store check and installation
#'   results.
#' @param quiet_check If \code{TRUE}, suppresses individual \code{R CMD
#'   check} output and only prints summaries. Set to \code{FALSE} for
#'   debugging.
#' @return Returns (invisibly) the directory where check results are stored.
#' @keywords internal
#' @inheritParams check
#' @export
check_cran <- function(pkgs, libpath = file.path(tempdir(), "R-lib"),
                       srcpath = libpath, check_libpath = libpath,
                       bioconductor = FALSE,
                       type = getOption("pkgType"),
                       threads = getOption("Ncpus", 1),
                       check_dir = tempfile("check_cran"),
                       install_dir = tempfile("check_cran_install"),
                       env_vars = NULL,
                       quiet_check = TRUE) {

  stopifnot(is.character(pkgs))
  if (length(pkgs) == 0) return()

  # For installing from GitHub, if git2r is not installed in the check_libpath
  requireNamespace("git2r", quietly = TRUE)

  rule("Checking ", length(pkgs), " CRAN packages", pad = "=")
  if (!file.exists(check_dir)) dir.create(check_dir)
  message("Results saved in ", check_dir)

  old <- options(warn = 1)
  on.exit(options(old), add = TRUE)

  # Create and use temporary library
  lapply(libpath, dir.create, showWarnings = FALSE, recursive = TRUE)
  libpath <- normalizePath(libpath)

  # Add the temporary library and remove on exit
  withr::with_libpaths(libpath, {

    message("Package library: ", paste(.libPaths(), collapse = ", "))

    repos <- c(CRAN = cran_mirror())
    if (bioconductor) {
      check_bioconductor()
      repos <- c(repos, BiocInstaller::biocinstallRepos())
    }

    rule("Installing dependencies") # ------------------------------------------

    deps <- package_deps(pkgs, repos = repos, type = type, dependencies = TRUE)

    dir.create(install_dir, showWarnings = FALSE, recursive = TRUE)

    needed <- deps$diff != CURRENT
    if (any(needed)) {
      message("Installing ", sum(needed), " packages: ", comma(pkgs))
      update(deps, Ncpus = threads, quiet = FALSE,
             out_dir = install_dir, skip_if_log_exists = TRUE)
    }

    # Download source packages
    available_src <- available_packages(repos, "source")
    urls <- lapply(pkgs, package_url, repos = repos, available = available_src)
    ok <- vapply(urls, function(x) !is.na(x$name), logical(1))
    if (any(!ok)) {
      message(
        "Skipping ", sum(!ok), " packages without source:",
        comma(pkgs[!ok])
      )
      urls <- urls[ok]
      pkgs <- pkgs[ok]
    }

    local_urls <- file.path(srcpath, vapply(urls, `[[`, "name", FUN.VALUE = character(1)))
    remote_urls <- vapply(urls, `[[`, "url", FUN.VALUE = character(1))

    needs_download <- !vapply(local_urls, is_source_pkg, logical(1))
    if (any(needs_download)) {
      message(
        "Downloading ", sum(needs_download), " source packages: ",
        comma(pkgs[needs_download])
      )
      Map(utils::download.file, remote_urls[needs_download],
        local_urls[needs_download], quiet = TRUE)
    }
  })

  # Add the temporary library and remove on exit
  withr::with_libpaths(check_libpath, {

    rule("Checking packages") # --------------------------------------------------
    message("Checking ", length(pkgs), " packages: ", comma(pkgs))
    message("Package library: ", paste(.libPaths(), collapse = ", "))

    check_start <- Sys.time()
    pkg_names <- format(pkgs)
    check_pkg <- function(i) {
      start_time <- Sys.time()
      res <- check_built(
        local_urls[i],
        args = "--no-multiarch --no-manual --no-codoc",
        env_vars = env_vars,
        check_dir = check_dir,
        quiet = quiet_check
      )
      end_time <- Sys.time()

      message("Checked ", pkg_names[i], ": ", summarise_check_results(res, colour = TRUE))
      status_update(i, length(pkgs), check_start)

      write_check_time(
        i, pkgs[i], elapsed_time = as.numeric(end_time - start_time, units = "secs"),
        file.path(check_dir, paste0(pkgs[i], ".Rcheck"), "check-time.txt"))

      NULL
    }

    if (length(pkgs) == 0)
      return()

    if (identical(as.integer(threads), 1L)) {
      lapply(seq_along(pkgs), check_pkg)
    } else {
      parallel::mclapply(seq_along(pkgs), check_pkg, mc.preschedule = FALSE,
        mc.cores = threads)
    }
  })

  invisible(check_dir)
}

status_update <- function(i, n, start_time) {
  if (i %% 10 != 0)
    return()

  hm <- function(x) {
    sprintf("%02i:%02i", x %/% 3600, x %% 3600 %/% 60)
  }

  elapsed <- as.numeric(Sys.time() - start_time, units = "secs")
  estimated <- elapsed / i * (n - i)

  msg <- sprintf(
    "Checked %i/%i. Elapsed %s. Remaining ~%s",
    i, n, hm(elapsed), hm(estimated)
  )
  message(msg)
}

write_check_time <- function(i, pkgs, elapsed_time, path) {
  writeLines(sprintf("%d  %s  %.1f", i, pkgs, elapsed_time), path)
}

parse_check_time <- function(path) {
  utils::read.table(file = path, header = FALSE)[[3]]
}

setClass("A")
setClass("B")
setClassUnion("AB", members = c("A", "B"))


setClass("mle2", contains = "mle")
setClassUnion("mleA", members = c("mle", "A"))
setClassUnion("mle2A", members = c("mle2", "A"))

#' Experimental email notification system.
#'
#' This currently assumes that you use github and gmail, and you have a
#' \code{revdep/email.md} email template.
#'
#' @inheritParams revdep_check
#' @param date Date package will be submitted to CRAN
#' @param version Version which will be used for the CRAN submission (usually
#'   different from the current package version)
#' @param author Name used to sign email
#' @param draft If \code{TRUE}, creates as draft email; if \code{FALSE},
#'   sends immediately.
#' @param template Path of template to use
#' @param only_problems Only inform authors with problems?
#' @param unsent If some emails fail to send, in a previous
#' @keywords internal
#' @export
revdep_email <- function(pkg = ".", date,
                         version,
                         author = getOption("devtools.name"),
                         draft = TRUE,
                         unsent = NULL,
                         template = "revdep/email.md",
                         only_problems = TRUE) {

  pkg <- as.package(pkg)
  force(date)
  force(version)

  if (is.null(author)) {
    stop("Please supply `author`", call. = FALSE)
  }

  if (is.null(unsent)) {
    results <- readRDS(revdep_check_path(pkg))$results
  } else {
    results <- unsent
  }

  if (only_problems) {
    results <- Filter(has_problems, results)
  }

  if (length(results) == 0) {
    message("No emails to send")
    return(invisible())
  }

  template_path <- file.path(pkg$path, template)
  if (!file.exists(template_path)) {
    stop("`", template, "` does not exist", call. = FALSE)
  }
  template <- readLines(template_path)

  maintainers <- vapply(results, function(x) x$maintainer, character(1))
  orphaned <- grepl("ORPHAN", maintainers)
  if (any(orphaned)) {
    orphans <- paste(names(results)[orphaned], collapse = ", ")
    message("Dropping ", sum(orphaned), " orphaned packages: ", orphans)

    results <- results[!orphaned]
    maintainers <- maintainers[!orphaned]
  }

  gh <- github_info(pkg$path)
  data <- lapply(results, maintainer_data, pkg = pkg, version = version,
                 gh = gh, date = date, author = author)
  bodies <- lapply(data, whisker::whisker.render, template = template)
  subjects <- lapply(data, function(x) {
    paste0(x$your_package, " and " , x$my_package, " ", x$my_version, " release")
  })

  emails <- Map(maintainer_email, maintainers, bodies, subjects)

  message("Testing first email")
  send_email(emails[[1]], draft = TRUE)

  if (yesno("Did first draft email look ok?"))
    return(invisible())

  sent <- vapply(emails, send_email, draft = draft, FUN.VALUE = logical(1))

  if (all(sent)) {
    message("All emails successfully sent")
  } else {
    message(sum(!sent), " failed. Call again with unsent = .Last.value")
  }

  results <- results[!sent]
  invisible(results)
}

send_email <- function(email, draft = TRUE) {
  send <- if (draft) gmailr::create_draft else gmailr::send_message
  msg <- if (draft) "Drafting" else "Sending"
  tryCatch(
    {
      message(msg, ": ", gmailr::subject(email))
      send(email)
      TRUE
    },
    interrupt = function(e) {
      message("Aborted by user")
      invokeRestart("abort")
    },
    error = function(e) {
      message("Failed")
      FALSE
    }
  )
}

maintainer_data <- function(result, pkg, version, gh, date, author) {
  problems <- result$results

  summary <- indent(paste(trunc_middle(unlist(problems)), collapse = "\n\n"))
  list(
    your_package = result$package,
    your_version = result$version,
    your_summary = summarise_check_results(problems),
    your_results = summary,

    you_have_problems = length(unlist(problems)) > 0,
    you_cant_install = any(grepl("Rcheck/00install[.]out", problems$errors)),

    me = author,
    date = date,
    my_package = pkg$package,
    my_version = version,
    my_github = gh$fullname
  )
}

maintainer_email <- function(to, body, subject) {
  gmailr::mime(To = to, Subject = subject, body = body)
}

a <- 1

# When this package is loaded with load_all, devtools should add a
# replacement system.file function.
# When the package is loaded with load_all, this returns devtools::system.file
# When installed and loaded, this returns base:system.file.
get_system.file <- function(...) {
  system.file
}

#' Compile a .dll/.so from source.
#'
#' \code{compile_dll} performs a fake R CMD install so code that
#' works here should work with a regular install (and vice versa).
#'
#' During compilation, debug flags are set with
#' \code{\link{compiler_flags}(TRUE)}.
#'
#' Invisibly returns the names of the DLL.
#'
#' @note If this is used to compile code that uses Rcpp, you will need to
#'   add the following line to your \code{Makevars} file so that it
#'   knows where to find the Rcpp headers:
#'   \code{PKG_CPPFLAGS=`$(R_HOME)/bin/Rscript -e 'Rcpp:::CxxFlags()'`}
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @param quiet if \code{TRUE} suppresses output from this function.
#' @seealso \code{\link{clean_dll}} to delete the compiled files.
#' @export
compile_dll <- function(pkg = ".", quiet = FALSE) {
  pkg <- as.package(pkg)

  old <- withr_with_envvar(compiler_flags(TRUE), {

    if (!needs_compile(pkg)) return(invisible())
    compile_rcpp_attributes(pkg)

    # Mock install the package to generate the DLL
    if (!quiet) message("Re-compiling ", pkg$package)
    install_dir <- tempfile("devtools_install_")
    dir.create(install_dir)
    inst <- install_min(pkg, install_dir, components = "libs",
      args = if (needs_clean(pkg)) "--preclean",
      quiet = quiet)

    invisible(dll_path(pkg))
  }, "prefix")
}

#' Remove compiled objects from /src/ directory
#'
#' Invisibly returns the names of the deleted files.
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @seealso \code{\link{compile_dll}}
#' @export
clean_dll <- function(pkg = ".") {
  pkg <- as.package(pkg)

  # Clean out the /src/ directory
  files <- dir(file.path(pkg$path, "src"),
    pattern = sprintf("\\.(o|sl|so|dylib|a|dll)$|(%s\\.def)$", pkg$package),
    full.names = TRUE,
    recursive = TRUE)
  unlink(files)

  invisible(files)
}

# Returns the full path and name of the DLL file
dll_path <- function(pkg = ".") {
  pkg <- as.package(pkg)

  name <- paste(pkg$package, .Platform$dynlib.ext, sep = "")
  file.path(pkg$path, "src", name)
}

mtime <- function(x) {
  x <- x[file.exists(x)]
  if (length(x) == 0) return(NULL)
  max(file.info(x)$mtime)
}

# List all source files in the package
sources <- function(pkg = ".") {
  pkg <- as.package(pkg)

  srcdir <- file.path(pkg$path, "src")
  dir(srcdir, "\\.(c.*|f)$", recursive = TRUE, full.names = TRUE)
}

# List all header files in the package
headers <- function(pkg = ".") {
  pkg <- as.package(pkg)

  incldir <- file.path(pkg$path, "inst", "include")
  srcdir <- file.path(pkg$path, "src")

  c(
    dir(srcdir, "^Makevars.*$", recursive = TRUE, full.names = TRUE),
    dir(srcdir, "\\.h.*$", recursive = TRUE, full.names = TRUE),
    dir(incldir, "\\.h.*$", recursive = TRUE, full.names = TRUE)
  )
}

# Does the package need recompiling?
# (i.e. is there a source or header file newer than the dll)
needs_compile <- function(pkg = ".") {
  pkg <- as.package(pkg)

  source <- mtime(c(sources(pkg), headers(pkg)))
  # no source files, so doesn't need compile
  if (is.null(source)) return(FALSE)

  dll <- mtime(dll_path(pkg))
  # no dll, so needs compile
  if (is.null(dll)) return(TRUE)

  source > dll
}

# Does the package need a clean compile?
# (i.e. is there a header or Makevars newer than the dll)
needs_clean <- function(pkg = ".") {
  pkg <- as.package(pkg)

  headers <- mtime(headers(pkg))
  # no headers, so never needs clean compile
  if (is.null(headers)) return(FALSE)

  dll <- mtime(dll_path(pkg))
  # no dll, so needs compile
  if (is.null(dll)) return(TRUE)

  headers > dll
}

#' Run user and package hooks.
#'
#'
#' @param pkg package description, can be path or package name.  See
#'   \code{\link{as.package}} for more information
#' @param hook hook name: one of "load", "unload", "attach", or "detach"
#' @keywords internal
run_pkg_hook <- function(pkg, hook) {
  pkg <- as.package(pkg)

  trans <- c(
    "load"   = ".onLoad",
    "unload" = ".onUnload",
    "attach" = ".onAttach",
    "detach" = ".onDetach")
  hook <- match.arg(hook, names(trans))
  f_name <- trans[hook]

  metadata <- dev_meta(pkg$package)
  if (isTRUE(metadata[[f_name]])) return(FALSE)

  # Run hook function if defined, and not already run
  nsenv <- ns_env(pkg)
  if (!exists(f_name, nsenv, inherits = FALSE)) return(FALSE)

  if (hook %in% c("load", "attach")) {
    nsenv[[f_name]](dirname(pkg$path), pkg$package)
  } else {
    nsenv[[f_name]](dirname(pkg$path))
  }
  metadata[[f_name]] <- TRUE

  TRUE
}

#' @rdname run_pkg_hook
run_user_hook <- function(pkg, hook) {
  pkg <- as.package(pkg)
  nsenv <- ns_env(pkg)

  trans <- c(
    "load"   = "onLoad",
    "unload" = "onUnload",
    "attach" = "attach",
    "detach" = "detach")
  hook <- match.arg(hook, names(trans))
  hook_name <- trans[hook]

  metadata <- dev_meta(pkg$package)
  if (isTRUE(metadata[[hook_name]])) return(FALSE)

  hooks <- getHook(packageEvent(pkg$package, hook_name))
  if (length(hooks) == 0) return(FALSE)

  for(fun in rev(hooks)) try(fun(pkg$package))
  metadata[[hook_name]] <- TRUE
  invisible(TRUE)
}

hook_warning <- function(pkg) {
  if (basename(pkg$path) == pkg$package) return()

  metadata <- dev_meta(pkg$package)
  if (isTRUE(metadata$hook_warning)) return()

  metadata$hook_warning <- TRUE
  warning(
    'Package path "', basename(pkg$path),
    '" does not match package name "', pkg$package, '". ',
    'This can result in problems when calling package hooks. ',
    'Please rename the directory to match the package name.',
    call. = FALSE)
}

#  Modified from src/library/tools/R/build.R
#
#  Copyright (C) 1995-2013 The R Core Team
#  Copyright (C) 2013 Hadley Wickham
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/

copy_vignettes <- function(pkg) {
  pkg <- as.package(pkg)

  doc_dir <- file.path(pkg$path, "inst", "doc")
  if (!file.exists(doc_dir)) {
    dir.create(doc_dir, recursive = TRUE, showWarnings = FALSE)
  }

  vigns <- tools::pkgVignettes(dir = pkg$path, output = TRUE, source = TRUE)
  if (length(vigns$docs) == 0) return(invisible())

  out_mv <- c(vigns$outputs, unique(unlist(vigns$sources, use.names = FALSE)))
  out_cp <- vigns$docs

  message("Moving ", paste(basename(out_mv), collapse = ", "), " to inst/doc/")
  file.copy(out_mv, doc_dir, overwrite = TRUE)
  file.remove(out_mv)

  message("Copying ", paste(basename(out_cp), collapse = ", "), " to inst/doc/")
  file.copy(out_cp, doc_dir, overwrite = TRUE)

  # Copy extra files, if needed
  extra_files <- find_vignette_extras(pkg)
  if (length(extra_files) == 0) return(invisible())

  message("Copying extra files ", paste(basename(extra_files), collapse = ", "),
    " to inst/doc/")
  file.copy(extra_files, doc_dir, recursive = TRUE)

  invisible()
}

find_vignette_extras <- function(pkg = ".") {
  pkg <- as.package(pkg)

  vig_path <- file.path(pkg$path, "vignettes")
  extras_file <- file.path(vig_path, ".install_extras")
  if (!file.exists(extras_file)) return(character())
  extras <- readLines(extras_file, warn = FALSE)
  if (length(extras) == 0) return(character())

  withr::with_dir(vig_path, {
    allfiles <- dir(all.files = TRUE, full.names = TRUE, recursive = TRUE,
      include.dirs = TRUE)
  })

  inst <- rep(FALSE, length(allfiles))
  for (e in extras) {
    inst <- inst | grepl(e, allfiles, perl = TRUE, ignore.case = TRUE)
  }

  normalizePath(file.path(vig_path, allfiles[inst]))
}

# Decompress pkg, if needed
source_pkg_info <- function(path, subdir = NULL) {
  if (!file.info(path)$isdir) {
    bundle <- path
    outdir <- tempfile(pattern = "devtools")
    dir.create(outdir)

    path <- decompress(path, outdir)
  } else {
    bundle <- NULL
  }

  pkg_path <- if (is.null(subdir)) path else file.path(path, subdir)

  # Check it's an R package
  if (!file.exists(file.path(pkg_path, "DESCRIPTION"))) {
    stop("Does not appear to be an R package (no DESCRIPTION)", call. = FALSE)
  }

  list(pkg_path = pkg_path, bundle = bundle)
}

is_source_pkg <- function(path, subdir = NULL) {
  tryCatch({
    source_pkg_info(path = path, subdir = subdir)
    TRUE
  }, error = function(e) return(FALSE))
}

source_pkg <- function(path, subdir = NULL, before_install = NULL) {
  info <- source_pkg_info(path = path, subdir = subdir)

  # Check configure is executable if present
  config_path <- file.path(info$pkg_path, "configure")
  if (file.exists(config_path)) {
    Sys.chmod(config_path, "777")
  }

  # Call before_install for bundles (if provided)
  if (!is.null(info$bundle) && !is.null(before_install))
    before_install(info$bundle, info$pkg_path)

  info$pkg_path
}


decompress <- function(src, target) {
  stopifnot(file.exists(src))

  if (grepl("\\.zip$", src)) {
    my_unzip(src, target)
    outdir <- getrootdir(as.vector(utils::unzip(src, list = TRUE)$Name))

  } else if (grepl("\\.tar$", src)) {
    utils::untar(src, exdir = target)
    outdir <- getrootdir(utils::untar(src, list = TRUE))

  } else if (grepl("\\.(tar\\.gz|tgz)$", src)) {
    utils::untar(src, exdir = target, compressed = "gzip")
    outdir <- getrootdir(utils::untar(src, compressed = "gzip", list = TRUE))

  } else if (grepl("\\.(tar\\.bz2|tbz)$", src)) {
    utils::untar(src, exdir = target, compressed = "bzip2")
    outdir <- getrootdir(utils::untar(src, compressed = "bzip2", list = TRUE))

  } else {
    ext <- gsub("^[^.]*\\.", "", src)
    stop("Don't know how to decompress files with extension ", ext,
      call. = FALSE)
  }

  file.path(target, outdir)
}


# Returns everything before the last slash in a filename
# getdir("path/to/file") returns "path/to"
# getdir("path/to/dir/") returns "path/to/dir"
getdir <- function(path)  sub("/[^/]*$", "", path)

# Given a list of files, returns the root (the topmost folder)
# getrootdir(c("path/to/file", "path/to/other/thing")) returns "path/to"
getrootdir <- function(file_list) {
  slashes <- nchar(gsub("[^/]", "", file_list))
  if (min(slashes) == 0) return("")

  getdir(file_list[which.min(slashes)])
}

my_unzip <- function(src, target, unzip = getOption("unzip")) {
  if (unzip == "internal") {
    return(utils::unzip(src, exdir = target))
  }

  args <- paste(
    "-oq", shQuote(src),
    "-d", shQuote(target)
  )

  system_check(unzip, args, quiet = TRUE)
}

context("getrootdir")

test_that("finds common prefix", {
  expect_equal(getrootdir(c("x/a", "x/b", "x/c")), "x")
})


test_that("returns empty string when all paths in current directory (#537)", {
  expect_equal(getrootdir(c("a", "b", "c", "d/e")), "")
})

context("git")

git_test_repo <- function() {
  d <- tempfile("")
  dir.create(d)
  r <- git2r::init(d)
  git2r::config(r, user.name = "user", user.email = "user@email.com")
  writeLines(character(), file.path(d, ".gitignore"))
  git2r::add(r, ".gitignore")
  git2r::commit(r, "initial")
  r
}

test_that("SHA for regular repository", {
  r <- git_test_repo()
  commit <- git2r::commits(r)[[1]]
  expect_false(git2r::is_commit(git2r::head(r)))
  expect_equal(git_repo_sha1(r), commit@sha)
})

test_that("SHA for detached head", {
  skip_on_cran()

  r <- git_test_repo()
  commit <- git2r::commits(r)[[1]]
  git2r::checkout(commit)
  expect_true(git2r::is_commit(git2r::head(r)))
  expect_equal(git_repo_sha1(r), commit@sha)
})

context("update.package_deps")

## -2 = not installed, but available on CRAN
## -1 = installed, but out of date
##  0 = installed, most recent version
##  1 = installed, version ahead of CRAN
##  2 = package not on CRAN

test_that("update.package_deps", {

  object <- data.frame(
    stringsAsFactors = FALSE,
    package = c("dotenv", "falsy", "magrittr"),
    installed = c("1.0", "1.0", "1.0"),
    available = c("1.0", NA, "1.0"),
    diff = c(0L, 2L, 0L),
    remote = c("dotenv", "falsy", "magrittr") # these are not actual remotes
  )
  class(object) <- c("package_deps", "data.frame")

  expect_message(
    update(object, quiet = FALSE, upgrade = FALSE),
    "Skipping 1 unavailable package: falsy"
  )

  object <- data.frame(
    stringsAsFactors = FALSE,
    package = c("dotenv", "falsy", "magrittr"),
    installed = c("1.0", "1.1", "1.0"),
    available = c("1.0", "1.0", "1.0"),
    diff = c(0L, 1L, 0L),
    remote = c("dotenv", "falsy", "magrittr") # these are not actual remotes
  )
  class(object) <- c("package_deps", "data.frame")

  expect_message(
    update(object, quiet = FALSE, upgrade = FALSE),
    "Skipping 1 package ahead of CRAN: falsy"
  )

  object <- data.frame(
    stringsAsFactors = FALSE,
    package = c("dotenv", "falsy", "magrittr"),
    installed = c("1.0", "1.0", NA),
    available = c("1.0", "1.1", "1.0"),
    diff = c(0L, 1L, -2L),
    remote = c("dotenv", "falsy", "magrittr") # these are not actual remotes
  )
  class(object) <- c("package_deps", "data.frame")

  with_mock(
    `devtools::install_remotes` = function(packages, ...) packages,
    expect_equal(
      update(object, upgrade = FALSE),
      "magrittr"
    )
  )

})

#!/hint/bash
# License: Unspecified
:

# shellcheck disable=2034
CHROOT_VERSION='v4'

##
#  usage : check_root
##
orig_argv=("$0" "$@")
check_root() {
	(( EUID == 0 )) && return
	if type -P sudo >/dev/null; then
		exec sudo -- "${orig_argv[@]}"
	else
		exec su root -c "$(printf ' %q' "${orig_argv[@]}")"
	fi
}

##
#  usage : is_btrfs( $path )
# return : whether $path is on a btrfs
##
is_btrfs() {
	[[ -e "$1" && "$(stat -f -c %T "$1")" == btrfs ]]
}

##
#  usage : is_subvolume( $path )
# return : whether $path is a the root of a btrfs subvolume (including
#          the top-level subvolume).
##
is_subvolume() {
	[[ -e "$1" && "$(stat -f -c %T "$1")" == btrfs && "$(stat -c %i "$1")" == 256 ]]
}

##
#  usage : is_same_fs( $path_a, $path_b )
# return : whether $path_a and $path_b are on the same filesystem
##
is_same_fs() {
	[[ "$(stat -c %d "$1")" == "$(stat -c %d "$1")" ]]
}

##
#  usage : subvolume_delete_recursive( $path )
#
#    Find all btrfs subvolumes under and including $path and delete them.
##
subvolume_delete_recursive() {
	local subvol

	is_subvolume "$1" || return 0

	while IFS= read -d $'\0' -r subvol; do
		if ! subvolume_delete_recursive "$subvol"; then
			return 1
		fi
	done < <(find "$1" -mindepth 1 -xdev -depth -inum 256 -print0)
	if ! btrfs subvolume delete "$1" &>/dev/null; then
		error "Unable to delete subvolume %s" "$subvol"
		return 1
	fi

	return 0
}

#!/hint/bash
# License: Unspecified
:

# shellcheck disable=2034
_arch=(
	x86_64
	any
)

# shellcheck disable=2034
_tags=(
	core-x86_64 core-any
	extra-x86_64 extra-any
	multilib-x86_64
	staging-x86_64 staging-any
	testing-x86_64 testing-any
	multilib-testing-x86_64
	multilib-staging-x86_64
	community-x86_64 community-any
	community-staging-x86_64 community-staging-any
	community-testing-x86_64 community-testing-any
	kde-unstable-x86_64 kde-unstable-any
	gnome-unstable-x86_64 gnome-unstable-any
)

#!/hint/bash
# License: Unspecified
:

# shellcheck disable=2034
_repos=(
	staging
	testing
	core
	extra
	community-staging
	community-testing
	community
	multilib-staging
	multilib-testing
	multilib
	gnome-unstable
	kde-unstable
)

# shellcheck disable=2034
_build_repos=(
	staging
	testing
	extra
	multilib-staging
	multilib-testing
	multilib
	gnome-unstable
	kde-unstable
)

#!/hint/bash
#
# This may be included with or without `set -euE`
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${_INCLUDE_COMMON_SH:-} ]] || return 0
_INCLUDE_COMMON_SH="$(set +o|grep nounset)"

set +u +o posix
# shellcheck disable=1091
. /usr/share/makepkg/util.sh
$_INCLUDE_COMMON_SH

# Avoid any encoding problems
export LANG=C

# Set buildtool properties
export BUILDTOOL=devtools
export BUILDTOOLVER=m4_devtools_version

shopt -s extglob

# check if messages are to be printed using color
if [[ -t 2 && "$TERM" != dumb ]]; then
	colorize
else
	# shellcheck disable=2034
	declare -gr ALL_OFF='' BOLD='' BLUE='' GREEN='' RED='' YELLOW=''
fi

stat_busy() {
	local mesg=$1; shift
	# shellcheck disable=2059
	printf "${GREEN}==>${ALL_OFF}${BOLD} ${mesg}...${ALL_OFF}" "$@" >&2
}

stat_done() {
	# shellcheck disable=2059
	printf "${BOLD}done${ALL_OFF}\n" >&2
}

_setup_workdir=false
setup_workdir() {
	[[ -z ${WORKDIR:-} ]] && WORKDIR=$(mktemp -d --tmpdir "${0##*/}.XXXXXXXXXX")
	_setup_workdir=true
	trap 'trap_abort' INT QUIT TERM HUP
	trap 'trap_exit' EXIT
}

cleanup() {
	if [[ -n ${WORKDIR:-} ]] && $_setup_workdir; then
		rm -rf "$WORKDIR"
	fi
	exit "${1:-0}"
}

abort() {
	error 'Aborting...'
	cleanup 255
}

trap_abort() {
	trap - EXIT INT QUIT TERM HUP
	abort
}

trap_exit() {
	local r=$?
	trap - EXIT INT QUIT TERM HUP
	cleanup $r
}

die() {
	(( $# )) && error "$@"
	cleanup 255
}

##
#  usage : lock( $fd, $file, $message, [ $message_arguments... ] )
##
lock() {
	# Only reopen the FD if it wasn't handed to us
	if ! [[ "/dev/fd/$1" -ef "$2" ]]; then
		mkdir -p -- "$(dirname -- "$2")"
		eval "exec $1>"'"$2"'
	fi

	if ! flock -n "$1"; then
		stat_busy "${@:3}"
		flock "$1"
		stat_done
	fi
}

##
#  usage : slock( $fd, $file, $message, [ $message_arguments... ] )
##
slock() {
	# Only reopen the FD if it wasn't handed to us
	if ! [[ "/dev/fd/$1" -ef "$2" ]]; then
		mkdir -p -- "$(dirname -- "$2")"
		eval "exec $1>"'"$2"'
	fi

	if ! flock -sn "$1"; then
		stat_busy "${@:3}"
		flock -s "$1"
		stat_done
	fi
}

##
#  usage : lock_close( $fd )
##
lock_close() {
	local fd=$1
	# https://github.com/koalaman/shellcheck/issues/862
	# shellcheck disable=2034
	exec {fd}>&-
}

##
# usage: pkgver_equal( $pkgver1, $pkgver2 )
##
pkgver_equal() {
	if [[ $1 = *-* && $2 = *-* ]]; then
		# if both versions have a pkgrel, then they must be an exact match
		[[ $1 = "$2" ]]
	else
		# otherwise, trim any pkgrel and compare the bare version.
		[[ ${1%%-*} = "${2%%-*}" ]]
	fi
}

##
#  usage: find_cached_package( $pkgname, $pkgver, $arch )
#
#    $pkgver can be supplied with or without a pkgrel appended.
#    If not supplied, any pkgrel will be matched.
##
find_cached_package() {
	local searchdirs=("$PWD" "$PKGDEST") results=()
	local targetname=$1 targetver=$2 targetarch=$3
	local dir pkg pkgbasename name ver rel arch r results

	for dir in "${searchdirs[@]}"; do
		[[ -d $dir ]] || continue

		for pkg in "$dir"/*.pkg.tar?(.!(sig|*.*)); do
			[[ -f $pkg ]] || continue

			# avoid adding duplicates of the same inode
			for r in "${results[@]}"; do
				[[ $r -ef $pkg ]] && continue 2
			done

			# split apart package filename into parts
			pkgbasename=${pkg##*/}
			pkgbasename=${pkgbasename%.pkg.tar*}

			arch=${pkgbasename##*-}
			pkgbasename=${pkgbasename%-"$arch"}

			rel=${pkgbasename##*-}
			pkgbasename=${pkgbasename%-"$rel"}

			ver=${pkgbasename##*-}
			name=${pkgbasename%-"$ver"}

			if [[ $targetname = "$name" && $targetarch = "$arch" ]] &&
					pkgver_equal "$targetver" "$ver-$rel"; then
				results+=("$pkg")
			fi
		done
	done

	case ${#results[*]} in
		0)
			return 1
			;;
		1)
			printf '%s\n' "${results[0]}"
			return 0
			;;
		*)
			error 'Multiple packages found:'
			printf '\t%s\n' "${results[@]}" >&2
			return 1
	esac
}


check_package_validity(){
	local pkgfile=$1
	if grep -q "packager = Unknown Packager" <(bsdtar -xOqf "$pkgfile" .PKGINFO); then
		die "PACKAGER was not set when building package"
	fi
	hashsum=sha256sum
	pkgbuild_hash=$(awk -v"hashsum=$hashsum" -F' = ' '$1 == "pkgbuild_"hashsum {print $2}' <(bsdtar -xOqf "$pkgfile" .BUILDINFO))
	if [[ "$pkgbuild_hash" != "$($hashsum PKGBUILD|cut -d' ' -f1)" ]]; then
		die "PKGBUILD $hashsum mismatch: expected $pkgbuild_hash"
	fi
}


# usage: grep_pkginfo pkgfile pattern
grep_pkginfo() {
	local _ret=()
	mapfile -t _ret < <(bsdtar -xOqf "$1" ".PKGINFO" | grep "^${2} = ")
	printf '%s\n' "${_ret[@]#${2} = }"
}


# Get the package name
getpkgname() {
	local _name

	_name="$(grep_pkginfo "$1" "pkgname")"
	if [[ -z $_name ]]; then
		error "Package '%s' has no pkgname in the PKGINFO. Fail!" "$1"
		exit 1
	fi

	echo "$_name"
}


# Get the package base or name as fallback
getpkgbase() {
	local _base

	_base="$(grep_pkginfo "$1" "pkgbase")"
	if [[ -z $_base ]]; then
		getpkgname "$1"
	else
		echo "$_base"
	fi
}


getpkgdesc() {
	local _desc

	_desc="$(grep_pkginfo "$1" "pkgdesc")"
	if [[ -z $_desc ]]; then
		error "Package '%s' has no pkgdesc in the PKGINFO. Fail!" "$1"
		exit 1
	fi

	echo "$_desc"
}


is_debug_package() {
	local pkgfile=${1} pkgbase pkgname pkgdesc
	pkgbase="$(getpkgbase "${pkgfile}")"
	pkgname="$(getpkgname "${pkgfile}")"
	pkgdesc="$(getpkgdesc "${pkgfile}")"
	[[ ${pkgdesc} == "Detached debugging symbols for "* && ${pkgbase}-debug = "${pkgname}" ]]
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_ISSUE_MOVE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_ISSUE_MOVE_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/cache.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/cache.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh

set -eo pipefail


pkgctl_issue_move_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [IID] [DESTINATION_PACKAGE]

		The move command allows users to transfer an issue from one project to another
		within the Arch Linux packaging group. This is useful when an issue is
		identified to be more relevant or better handled in a different project.

		By default, the command operates within the current directory, but users can
		specify a different package base from which to move the issue.

		Users must specify the issue ID (IID) and the destination package to which the
		issue should be moved. A comment message explaining the reason for the move can
		be provided directly through the command line.

		OPTIONS
		    -p, --package PKGBASE  Move from <pkgbase> instead of the current directory
		    -m, --message MSG      Use the provided message as the comment
		    -e, --edit             Edit the comment using an editor
		    -h, --help             Show this help text

		EXAMPLES
		    $ ${COMMAND} 42 to-bar
		    $ ${COMMAND} --package from-foo 42 to-bar
_EOF_
}

pkgctl_issue_move() {
	if (( $# < 1 )); then
		pkgctl_issue_move_usage
		exit 0
	fi

	local iid=""
	local pkgbase=""
	local message=""
	local edit=0

	local to_project_name to_project_id project_path issue_url to_iid result

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_issue_move_usage
				exit 0
				;;
			-p|--package)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				pkgbase=$2
				shift 2
				;;
			-m|--message)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				message=$2
				shift 2
				;;
			-e|--edit)
				edit=1
				shift
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				break
				;;
		esac
	done

	if [[ -z ${pkgbase} ]]; then
		if ! [[ -f PKGBUILD ]]; then
			die "missing --package option or PKGBUILD in current directory"
		fi
		pkgbase=$(realpath --canonicalize-existing .)
	fi
	pkgbase=$(basename "${pkgbase}")

	if (( $# < 2 )); then
		pkgctl_issue_move_usage
		exit 1
	fi

	iid=$1
	to_project_name=$(basename "$2")

	# spawn editor
	if (( edit )); then
		msgfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-issue-note.XXXXXXXXXX.md)
		printf "%s\n" "${message}" >> "${msgfile}"
		if [[ -n $VISUAL ]]; then
			$VISUAL "${msgfile}" || die
		elif [[ -n $EDITOR ]]; then
			$EDITOR "${msgfile}" || die
		else
			die "No usable editor found (tried \$VISUAL, \$EDITOR)."
		fi
		message=$(cat "${msgfile}")
	fi

	if ! result=$(gitlab_project "${to_project_name}"); then
		msg_error "Failed to query target project ${BOLD}${to_project_name}${ALL_OFF}"
		exit 1
	fi

	if ! to_project_id=$(jq --raw-output ".id" <<< "${result}"); then
		msg_error "Failed to query project id for ${BOLD}${to_project_name}${ALL_OFF}"
		exit 1
	fi

	# comment on issue
	if [[ -n ${message} ]]; then
		if ! result=$(gitlab_create_project_issue_note "${pkgbase}" "${iid}" "${message}"); then
			msg_error "Failed to comment on issue ${BOLD}#${iid}${ALL_OFF}"
			exit 1
		fi
		msg_success "Commented on issue ${BOLD}#${iid}${ALL_OFF}"
	fi

	if ! result=$(gitlab_project_issue_move "${pkgbase}" "${iid}" "${to_project_id}"); then
		msg_error "Failed to move issue ${BOLD}#${iid}${ALL_OFF} to ${BOLD}${to_project_name}${ALL_OFF}"
		exit 1
	fi

	if ! to_iid=$(jq --raw-output ".iid" <<< "${result}"); then
		msg_error "Failed to query issue id for ${BOLD}${to_project_name}${ALL_OFF}"
		exit 1
	fi

	project_path=$(gitlab_project_name_to_path "${to_project_name}")
	issue_url="${GIT_PACKAGING_URL_HTTPS}/${project_path}/-/issues/${to_iid}"

	msg_success "Moved issue ${BOLD}${pkgbase}${ALL_OFF} ${BOLD}#${iid}${ALL_OFF} to ${BOLD}${to_project_name}${ALL_OFF} ${BOLD}#${to_iid}${ALL_OFF}"
	echo
	printf "%sView this issue on GitLab: %s%s\n" "${GRAY}" "${issue_url}" "${ALL_OFF}"
}

#!/bin/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_REPO_WEB_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_REPO_WEB_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh

set -e


pkgctl_repo_web_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [PKGBASE]...

		Open the packaging repository's website via xdg-open. If called with
		no arguments, open the package cloned in the current working directory.

		OPTIONS
		    --print       Print the url instead of opening it with xdg-open
		    -h, --help    Show this help text

		EXAMPLES
		    $ ${COMMAND} linux
_EOF_
}

pkgctl_repo_web() {
	local pkgbases=()
	local path giturl pkgbase url
	local mode=open

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_repo_web_usage
				exit 0
				;;
			--print)
				mode=print
				shift
				;;
			--)
				shift
				break
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				pkgbases=("$@")
				break
				;;
		esac
	done

	# Check if web mode has xdg-open
	if [[ ${mode} == open ]] && ! command -v xdg-open &>/dev/null; then
		die "The web command requires 'xdg-open'"
	fi

	# Check if used without pkgnames in a packaging directory
	if (( ! $# )); then
		path=${PWD}
		if [[ ! -d "${path}/.git" ]]; then
			die "Not a Git repository: ${path}"
		fi

		giturl=$(git -C "${path}" remote get-url origin)
		if [[ ${giturl} != *${GIT_PACKAGING_NAMESPACE}* ]]; then
			die "Not a packaging repository: ${path}"
		fi

		pkgbase=$(basename "${giturl}")
		pkgbase=${pkgbase%.git}
		pkgbases=("${pkgbase}")
	fi

	for pkgbase in "${pkgbases[@]}"; do
		pkgbase=$(basename "${pkgbase}")
		path=$(gitlab_project_name_to_path "${pkgbase}")
		url="${GIT_PACKAGING_URL_HTTPS}/${path}"
		case ${mode} in
			open)
				xdg-open "${url}"
				;;
			print)
				printf "%s\n" "${url}"
				;;
			*)
				die "Unknown mode: ${mode}"
		esac
	done
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_LICENSE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_LICENSE_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

set -e


pkgctl_license_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
    cat <<- _EOF_
		Usage: ${COMMAND} [COMMAND] [OPTIONS]

		Check and manage package license compliance.

		COMMANDS
		    check      Check package license compliance
		    setup      Automatically detect and setup a basic REUSE config

		OPTIONS
		    -h, --help    Show this help text

		EXAMPLES
		    $ ${COMMAND} check libfoo linux libbar
		    $ ${COMMAND} setup libfoo
_EOF_
}

pkgctl_license() {
	if (( $# < 1 )); then
		pkgctl_license_usage
		exit 0
	fi

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_license_usage
				exit 0
				;;
			check)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/license/check.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/license/check.sh
				pkgctl_license_check "$@"
				exit $?
				;;
			setup)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/license/setup.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/license/setup.sh
				pkgctl_license_setup "$@"
				exit 0
				;;
			*)
				die "invalid argument: %s" "$1"
				;;
		esac
	done
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_DB_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_DB_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

set -e


pkgctl_db_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [COMMAND] [OPTIONS]

		Pacman database modification for package update, move etc

		COMMANDS
		    move      Move packages between pacman repositories
		    remove    Remove packages from pacman repositories
		    update    Update the pacman database as final release step

		OPTIONS
		    -h, --help    Show this help text

		EXAMPLES
		    $ ${COMMAND} move extra-staging extra-testing libfoo libbar
		    $ ${COMMAND} remove core-testing libfoo libbar
		    $ ${COMMAND} update
_EOF_
}

pkgctl_db() {
	if (( $# < 1 )); then
		pkgctl_db_usage
		exit 0
	fi

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_db_usage
				exit 0
				;;
			move)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/db/move.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/db/move.sh
				pkgctl_db_move "$@"
				exit 0
				;;
			remove)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/db/remove.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/db/remove.sh
				pkgctl_db_remove "$@"
				exit 0
				;;
			update)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/db/update.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/db/update.sh
				pkgctl_db_update "$@"
				exit 0
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				die "invalid command: %s" "$1"
				;;
		esac
	done
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

# shellcheck disable=2034
DEVTOOLS_VALID_ISSUE_SEVERITY=(
	lowest
	low
	medium
	high
	critical
)

# shellcheck disable=2034
DEVTOOLS_VALID_ISSUE_PRIORITY=(
	low
	normal
	high
	urgent
)

# shellcheck disable=2034
DEVTOOLS_VALID_ISSUE_STATUS=(
	confirmed
	in-progress
	in-review
	on-hold
	unconfirmed
	waiting-input
	waiting-upstream
)

# shellcheck disable=2034
DEVTOOLS_VALID_ISSUE_SCOPE=(
	bug
	feature
	security
	question
	regression
	enhancement
	documentation
	reproducibility
	out-of-date
)

# shellcheck disable=2034
DEVTOOLS_VALID_ISSUE_SEARCH_LOCATION=(
	title
	description
	all
)

# shellcheck disable=2034
DEVTOOLS_VALID_ISSUE_RESOLUTION=(
	cant-reproduce
	completed
	duplicate
	invalid
	not-a-bug
	upstream
	wont-fix
)

# shellcheck disable=2034
DEVTOOLS_VALID_ISSUE_CONFIDENTIALITY=(
	confidential
	public
)

#!/hint/bash

# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_BUILD_OFFLOAD_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_BUILD_OFFLOAD_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/util/makepkg.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/makepkg.sh

source /usr/share/makepkg/util/message.sh

set -eo pipefail


PKGCTL_OFFLOAD_CACHE_HOME="${XDG_CACHE_HOME:-$HOME/.cache}/pkgctl/offload"

pkgctl_build_offload_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [COMMAND] [OPTIONS]...

		Server commands to build packages remotely by offloading the job.

		For internal use only!
_EOF_
}

pkgctl_build_offload() {
	if (( $# < 1 )); then
		pkgctl_build_offload_usage
		exit 1
	fi

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_build_offload_usage
				exit 0
				;;
			create-builddir)
				shift
				pkgctl_build_offload_server_create_builddir "$@"
				exit 0
				;;
			clean-builddir)
				shift
				pkgctl_build_offload_server_clean_builddir "$@"
				exit 0
				;;
			build)
				shift
				pkgctl_build_offload_server_build "$@"
				exit 0
				;;
			collect-files)
				shift
				pkgctl_build_offload_server_collect_files "$@"
				exit 0
				;;
			collect-logs)
				shift
				pkgctl_build_offload_server_collect_logs "$@"
				exit 0
				;;
			*)
				die "invalid argument: %s" "$1"
				;;
		esac
	done
}

pkgctl_build_offload_client() {
	local pkgbase=$1
	local pkgrepo=$2
	local pkgarch=$3
	local server=$4
	shift 4
	# shellcheck disable=SC2031
	local working_dir=$PWD
	local _srcpkg srcpkg files

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	TEMPDIR=$(mktemp --tmpdir="${WORKDIR}" --directory "offload.${pkgbase}.${pkgrepo}-${pkgarch}XXXXXXXXXX")

	# Load makepkg.conf variables to be available
	makepkg_load_config

	# Use a source-only tarball as an intermediate to transfer files. This
	# guarantees the checksums are okay, and guarantees that all needed files are
	# transferred, including local sources, install scripts, and changelogs.
	export SRCPKGDEST="${TEMPDIR}"
	if ! makepkg_source_package; then
		die "unable to make source package"
		return 1
	fi

	# Temporary cosmetic workaround makepkg if SRCDEST is set somewhere else
	# but an empty src dir is created in PWD. Remove once fixed in makepkg.
	rmdir --ignore-fail-on-non-empty src 2>/dev/null || true

	local builddir
	builddir=$(
		ssh "${SSH_OPTS[@]}" -- "$server" pkgctl offload create-builddir "${pkgbase@Q}" "${pkgrepo@Q}" "${pkgarch@Q}"
	)

	# Transfer the srcpkg to the server
	msg "Transferring source package to the server..."
	_srcpkg=("$SRCPKGDEST"/*"$SRCEXT")
	srcpkg="${_srcpkg[0]}"
	if ! rsync "${RSYNC_OPTS[@]}" -- "$srcpkg" "$server":"${builddir}"; then
		die "failed to rsync sources to offload server"
		return 1
	fi

	# Execute build
	if ssh "${SSH_OPTS[@]}" -t -- "$server" pkgctl offload build "${builddir@Q}" "${srcpkg@Q}" "${pkgrepo@Q}" "${pkgarch@Q}" "${@@Q}"; then
		# Get an array of files that should be downloaded from the server
		mapfile -t files < <(
			ssh "${SSH_OPTS[@]}" -- "$server" pkgctl offload collect-files "${builddir@Q}" "${pkgrepo@Q}" "${pkgarch@Q}"
		)
	else
		# Build failed, only the logs should be downloaded from the server
		mapfile -t files < <(
			ssh "${SSH_OPTS[@]}" -- "$server" pkgctl offload collect-logs "${builddir@Q}"
		)
	fi

	# Check if we collected any files to download
	if (( ${#files[@]} == 0 )); then
		die "failed to collect files to download"
		return 1
	fi

	msg 'Downloading files...'
	rsync "${RSYNC_OPTS[@]}" -- "${files[@]/#/$server:}" "${TEMPDIR}/"

	# Clean remote build dir
	ssh "${SSH_OPTS[@]}" -- "$server" pkgctl offload clean-builddir "${builddir@Q}"

	# Move all log files to LOGDEST
	if is_globfile "${TEMPDIR}"/*.log; then
		mkdir --parents "${LOGDEST:-${working_dir}}/"
		mv "${TEMPDIR}"/*.log "${LOGDEST:-${working_dir}}/"
	fi

	# Assume build failed if we didn't download any package files
	if ! is_globfile "${TEMPDIR}"/*.pkg.tar*; then
		error "Build failed, check logs in ${LOGDEST:-${working_dir}}"
		return 1
	fi

	# Building a package may change the PKGBUILD during update_pkgver
	mv "${TEMPDIR}/PKGBUILD" "${working_dir}/"

	# Download package files
	mkdir --parents "${PKGDEST:-${working_dir}}/"
	mv "${TEMPDIR}"/*.pkg.tar* "${PKGDEST:-${working_dir}}/"

	return 0
}

pkgctl_build_offload_server_build() {
	local builddir=$1
	local srcpkg=$2
	local pkgrepo=$3
	local pkgarch=$4
	shift 4
	local buildtool

	if [[ -n $pkgarch ]]; then
		buildtool="${pkgrepo}-${pkgarch}-build"
	else
		buildtool="${pkgrepo}-build"
	fi

	cd "${builddir}"
	bsdtar --strip-components 1 -xvf "$(basename "$srcpkg")"
	LOGDEST="" "${buildtool}" "$@"
}

pkgctl_build_offload_server_create_builddir() {
	local pkgbase=$1
	local pkgrepo=$2
	local pkgarch=$3
	mkdir --parents "${PKGCTL_OFFLOAD_CACHE_HOME}"
	mktemp --directory --tmpdir="${PKGCTL_OFFLOAD_CACHE_HOME}" "${pkgbase}.${pkgrepo}-${pkgarch}XXXXXXXXXX"
}

pkgctl_build_offload_server_clean_builddir() {
	local builddir=$1
	rm --recursive --force -- "${builddir}"
}

pkgctl_build_offload_server_collect_files() {
	local builddir=$1
	local pkgrepo=$2
	local pkgarch=$3

	local makepkg_config
	local makepkg_user_config

	# fallback config for multilib
	if [[ ${pkgrepo} == multilib* ]] && [[ -z ${pkgarch} ]]; then
		pkgarch=x86_64
	fi

	cd "${builddir}"
	makepkg_user_config="${XDG_CONFIG_HOME:-$HOME/.config}/pacman/makepkg.conf"
	makepkg_config="${_DEVTOOLS_LIBRARY_DIR}/makepkg.conf.d/${pkgarch}.conf"
	if [[ -f ${_DEVTOOLS_LIBRARY_DIR}/makepkg.conf.d/${pkgrepo}-${pkgarch}.conf ]]; then
		makepkg_config="${_DEVTOOLS_LIBRARY_DIR}/makepkg.conf.d/${pkgrepo}-${pkgarch}.conf"
	fi
	while read -r file; do
		if [[ -f "${file}" ]]; then
			printf "%s\n" "${file}"
		fi
	done < <(makepkg --config <(cat "${makepkg_user_config}" "${makepkg_config}" 2>/dev/null) --packagelist)

	printf "%s\n" "${builddir}/PKGBUILD"

	pkgctl_build_offload_server_collect_logs "${builddir}"
}

pkgctl_build_offload_server_collect_logs() {
	local builddir=$1
	find "${builddir}" -name "*.log"
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_UTIL_MACHINE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_UTIL_MACHINE_SH=1


# Usage: machine_name(name)
machine_name() {
	local name=$1

	local tool="${0##*/}"
	local machine="${tool}-${name}"
	local max_hostname=64  # see gethostname(2)
	local max_pid_digits=7  # ceil(log(2^22, 10))

	# Normalize the package name so it doubles as a valid hostname
	# https://github.com/systemd/systemd/blob/v256/src/basic/hostname-util.c#L83-L136
	machine="$(
		tr --squeeze-repeats --complement 'a-z0-9.' - <<< "${machine}" | \
			tr --squeeze-repeats '.' | \
			head --bytes=$(( max_hostname - max_pid_digits - 1 ))
	)"
	machine=${machine%%.}
	machine=${machine%%-}
	printf "%s.%s" "${machine}" "$$"
}

#!/bin/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_REPO_SWITCH_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_REPO_SWITCH_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh

source /usr/share/makepkg/util/message.sh

set -e


pkgctl_repo_switch_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [VERSION] [PKGBASE]...

		Switch a package source repository to a specified version, tag or
		branch. The working tree and the index are updated to match the
		specified ref.

		If a version identifier is specified in the pacman version format, that
		identifier is automatically translated to the Git tag name accordingly.

		The current working directory is used if no PKGBASE is specified.

		OPTIONS
		    --discard-changes   Discard changes if index or working tree is dirty
		    -f, --force         An alias for --discard-changes
		    -h, --help          Show this help text

		EXAMPLES
		    $ ${COMMAND} 1.14.6-1 gopass gopass-jsonapi
		    $ ${COMMAND} --force 2:1.19.5-1
		    $ ${COMMAND} main
_EOF_
}

pkgctl_repo_switch() {
	if (( $# < 1 )); then
		pkgctl_repo_switch_usage
		exit 0
	fi

	# options
	local VERSION
	local GIT_REF
	local GIT_CHECKOUT_OPTIONS=()
	local paths path realpath pkgbase

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_repo_switch_usage
				exit 0
				;;
			-f|--force|--discard-changes)
				GIT_CHECKOUT_OPTIONS+=("--force")
				shift
				;;
			--)
				shift
				break
				;;
			-*)
				# - is special to switch back to previous version
				if [[ $1 != - ]]; then
					die "invalid argument: %s" "$1"
				fi
				;;&
			*)
				if [[ -n ${VERSION} ]]; then
					break
				fi
				VERSION=$1
				shift
				;;
		esac
	done

	if [[ -z ${VERSION} ]]; then
		error "missing positional argument 'VERSION'"
		pkgctl_repo_switch_usage
		exit 1
	fi

	GIT_REF="$(get_tag_from_pkgver "${VERSION}")"
	paths=("$@")

	# check if invoked without any path from within a packaging repo
	if (( ${#paths[@]} == 0 )); then
		if [[ -f PKGBUILD ]]; then
			paths=(".")
		else
			die "Not a package repository: $(realpath -- .)"
		fi
	fi

	for path in "${paths[@]}"; do
		# resolve symlink for basename
		if ! realpath=$(realpath --canonicalize-existing -- "${path}"); then
			die "No such directory: ${path}"
		fi
		# skip paths that are not directories
		if [[ ! -d "${realpath}" ]]; then
			continue
		fi
		# skip paths that are not git repositories
		if [[ ! -d "${realpath}/.git" ]]; then
			error "Not a Git repository: ${path}"
			continue
		fi

		pkgbase=$(basename "${realpath}")
		if ! git -C "${path}" checkout "${GIT_CHECKOUT_OPTIONS[@]}" "${GIT_REF}"; then
			die "Failed to switch ${pkgbase} to version ${VERSION}"
		fi
		msg "Successfully switched ${pkgbase} to version ${VERSION}"
	done
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_CACHE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_CACHE_SH=1

set -e

readonly XDG_DEVTOOLS_CACHE_DIR="${XDG_CACHE_HOME:-$HOME/.cache}/devtools"

get_cache_file() {
	local filename=$1
	local path="${XDG_DEVTOOLS_CACHE_DIR}/${filename}"

	mkdir --parents -- "$(dirname -- "$path")"
	if [[ ! -f ${path} ]]; then
		touch -- "${path}"
	fi

	printf '%s' "${path}"
}

#!/bin/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
[[ -z ${DEVTOOLS_INCLUDE_LICENSE_CHECK_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_LICENSE_CHECK_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh

source /usr/share/makepkg/util/message.sh

set -eo pipefail

readonly PKGCTL_LICENSE_CHECK_EXIT_COMPLIANT=0
export PKGCTL_LICENSE_CHECK_EXIT_COMPLIANT
readonly PKGCTL_LICENSE_CHECK_EXIT_FAILURE=2
export PKGCTL_LICENSE_CHECK_EXIT_FAILURE


pkgctl_license_check_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [PKGBASE]...

		Checks package licensing compliance using REUSE and also verifies
		whether a LICENSE file with the expected Arch Linux-specific 0BSD
		license text exists.

		Upon execution, it runs 'reuse lint'.

		OPTIONS
		    -h, --help   Show this help text

		EXAMPLES
		    $ ${COMMAND} neovim vim
_EOF_
}

pkgctl_license_check() {
	local pkgbases=()
	local verbose=0

	local license_text
	license_text=$(< "${_DEVTOOLS_LIBRARY_DIR}"/data/LICENSE)

	local exit_code=${PKGCTL_LICENSE_CHECK_EXIT_COMPLIANT}

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_license_check_usage
				exit 0
				;;
			--)
				shift
				break
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				pkgbases=("$@")
				break
				;;
		esac
	done

	if ! command -v reuse &>/dev/null; then
		die "The \"$_DEVTOOLS_COMMAND\" command requires the 'reuse' CLI tool"
	fi

	# Check if used without pkgbases in a packaging directory
	if (( ${#pkgbases[@]} == 0 )); then
		if [[ -f PKGBUILD ]]; then
			pkgbases=(".")
		else
			pkgctl_license_check_usage
			exit 1
		fi
	fi

	# enable verbose mode when we only have a single item to check
	if (( ${#pkgbases[@]} == 1 )); then
		verbose=1
	fi

	for path in "${pkgbases[@]}"; do
		# skip paths that are not directories
		if [[ ! -d "${path}" ]]; then
			continue
		fi
		pushd "${path}" >/dev/null

		if [[ ! -f PKGBUILD ]]; then
			msg_error "${BOLD}${path}:${ALL_OFF} no PKGBUILD found"
			return 1
		fi

		if [[ ! -f .SRCINFO ]]; then
			msg_error "${BOLD}${path}:${ALL_OFF} no .SRCINFO found"
			return 1
		fi

		if ! pkgbase=$(grep --max-count=1 --extended-regexp "pkgbase = (.+)" .SRCINFO | awk '{print $3}'); then
			msg_error "${BOLD}${path}:${ALL_OFF} pkgbase not found in .SRCINFO"
			return 1
		fi

		if [[ ! -e LICENSE ]]; then
			msg_error "${BOLD}${pkgbase}:${ALL_OFF} is missing the LICENSE file"
			return "${PKGCTL_LICENSE_CHECK_EXIT_FAILURE}"
		fi

		if [[ ! -L LICENSES/0BSD.txt ]]; then
			msg_error "${BOLD}${pkgbase}:${ALL_OFF} LICENSES/0BSD should be a symlink to LICENSE but it isn't"
			return "${PKGCTL_LICENSE_CHECK_EXIT_FAILURE}"
		fi

		# Check if the local LICENSE file mismatches our expectations
		if [[ $license_text != $(< LICENSE) ]]; then
			msg_error "${BOLD}${pkgbase}:${ALL_OFF} LICENSE file doesn't have the expected Arch Linux-specific license text"
			return "${PKGCTL_LICENSE_CHECK_EXIT_FAILURE}"
		fi

		# Check for REUSE compliance
		if ! reuse lint --json | jq --exit-status '.summary.compliant' &>/dev/null; then
			msg_error "${BOLD}${pkgbase}:${ALL_OFF} repository is not REUSE compliant"
			exit_code=${PKGCTL_LICENSE_CHECK_EXIT_FAILURE}

			# re-execute reuse lint for human readable output
			if (( verbose )); then
				reuse lint
			fi

			popd >/dev/null
			continue
		fi

		msg_success "${BOLD}${pkgbase}:${ALL_OFF} repository is REUSE compliant"
		popd >/dev/null
	done

	# return status based on results
	return "${exit_code}"
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_REPO_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_REPO_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

set -e


pkgctl_repo_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [COMMAND] [OPTIONS]

		Manage Git packaging repositories and helps with their configuration
		according to distro specs.

		Git author information and the used signing key is set up from
		makepkg.conf read from any valid location like /etc or XDG_CONFIG_HOME.
		The configure command can be used to synchronize the distro specs and
		makepkg.conf settings for previously cloned repositories.

		The unprivileged option can be used for cloning packaging repositories
		without SSH access using read-only HTTPS.

		COMMANDS
		    clean          Remove untracked files from the working tree
		    clone          Clone a package repository
		    configure      Configure a clone according to distro specs
		    create         Create a new GitLab package repository
		    switch         Switch a package repository to a specified version
		    web            Open the packaging repository's website

		OPTIONS
		    -h, --help     Show this help text

		EXAMPLES
		    $ ${COMMAND} clean --interactive *
		    $ ${COMMAND} clone libfoo linux libbar
		    $ ${COMMAND} clone --maintainer mynickname
		    $ ${COMMAND} configure *
		    $ ${COMMAND} create libfoo
		    $ ${COMMAND} switch 2:1.19.5-1 libfoo
		    $ ${COMMAND} web linux
_EOF_
}

pkgctl_repo() {
	if (( $# < 1 )); then
		pkgctl_repo_usage
		exit 0
	fi

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_repo_usage
				exit 0
				;;
			clean)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/repo/clean.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/repo/clean.sh
				pkgctl_repo_clean "$@"
				exit 0
				;;
			clone)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/repo/clone.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/repo/clone.sh
				pkgctl_repo_clone "$@"
				exit 0
				;;
			configure)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/repo/configure.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/repo/configure.sh
				pkgctl_repo_configure "$@"
				exit 0
				;;
			create)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/repo/create.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/repo/create.sh
				pkgctl_repo_create "$@"
				exit 0
				;;
			switch)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/repo/switch.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/repo/switch.sh
				pkgctl_repo_switch "$@"
				exit 0
				;;
			web)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/repo/web.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/repo/web.sh
				pkgctl_repo_web "$@"
				exit 0
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				die "invalid command: %s" "$1"
				;;
		esac
	done
}

#!/bin/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_REPO_CONFIGURE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_REPO_CONFIGURE_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh
# shellcheck source=src/lib/util/git.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/git.sh
# shellcheck source=src/lib/util/makepkg.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/makepkg.sh

source /usr/share/makepkg/util/message.sh

set -e
shopt -s nullglob


pkgctl_repo_configure_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [PATH]...

		Configure Git packaging repositories according to distro specs and
		makepkg.conf settings.

		Git author information and the used signing key is set up from
		makepkg.conf read from any valid location like /etc or XDG_CONFIG_HOME.

		The remote protocol is automatically determined from the author email
		address by choosing SSH for all official packager identities and
		read-only HTTPS otherwise.

		Git default excludes and hooks are applied to the configured repo.

		OPTIONS
		    --protocol https     Configure remote url to use https
		    -j, --jobs N         Run up to N jobs in parallel (default: $(nproc))
		    -h, --help           Show this help text

		EXAMPLES
		    $ ${COMMAND} *
_EOF_
}

get_packager_name() {
	local packager=$1
	local packager_pattern="(.+) <(.+@.+)>"
	local name

	if [[ ! $packager =~ $packager_pattern ]]; then
		return 1
	fi

	name=$(echo "${packager}"|sed -E "s/${packager_pattern}/\1/")
	printf "%s" "${name}"
}

get_packager_email() {
	local packager=$1
	local packager_pattern="(.+) <(.+@.+)>"
	local email

	if [[ ! $packager =~ $packager_pattern ]]; then
		return 1
	fi

	email=$(echo "${packager}"|sed -E "s/${packager_pattern}/\2/")
	printf "%s" "${email}"
}

is_packager_name_valid() {
	local packager_name=$1
	if [[ -z ${packager_name} ]]; then
		return 1
	elif [[ ${packager_name} == "John Doe" ]]; then
		return 1
	elif [[ ${packager_name} == "Unknown Packager" ]]; then
		return 1
	fi
	return 0
}

is_packager_email_official() {
	local packager_email=$1
	if [[ -z ${packager_email} ]]; then
		return 1
	elif [[ $packager_email =~ .+@archlinux.org ]]; then
		return 0
	fi
	return 1
}

pkgctl_repo_configure() {
	# options
	local GIT_REPO_BASE_URL=${GIT_PACKAGING_URL_HTTPS}
	local official=0
	local proto=https
	local proto_force=0
	local jobs=
	jobs=$(nproc)
	local paths=()

	# variables
	local -r command=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	local path realpath pkgbase remote_url project_path hook
	local PACKAGER GPGKEY packager_name packager_email

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_repo_configure_usage
				exit 0
				;;
			--protocol=https)
				proto_force=1
				shift
				;;
			--protocol)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if [[ $2 == https ]]; then
					proto_force=1
				else
					die "unsupported protocol: %s" "$2"
				fi
				shift 2
				;;
			-j|--jobs)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				jobs=$2
				shift 2
				;;
			--)
				shift
				break
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				paths=("$@")
				break
				;;
		esac
	done

	# check if invoked without any path from within a packaging repo
	if (( ${#paths[@]} == 0 )); then
		if [[ -f PKGBUILD ]]; then
			paths=(".")
		else
			pkgctl_repo_configure_usage
			exit 1
		fi
	fi

	# Load makepkg.conf variables to be available for packager identity
	msg "Collecting packager identity from makepkg.conf"
	makepkg_load_config
	if [[ -n ${PACKAGER} ]]; then
		if ! packager_name=$(get_packager_name "${PACKAGER}") || \
		   ! packager_email=$(get_packager_email "${PACKAGER}"); then
			die "invalid PACKAGER format '${PACKAGER}' in makepkg.conf"
		fi
		if ! is_packager_name_valid "${packager_name}"; then
			die "invalid PACKAGER '${PACKAGER}' in makepkg.conf"
		fi
		if is_packager_email_official "${packager_email}"; then
			official=1
			if (( ! proto_force )); then
				proto=ssh
				GIT_REPO_BASE_URL=${GIT_PACKAGING_URL_SSH}
			fi
		fi
	fi

	msg2 "name    : ${packager_name:-${YELLOW}undefined${ALL_OFF}}"
	msg2 "email   : ${packager_email:-${YELLOW}undefined${ALL_OFF}}"
	msg2 "gpg-key : ${GPGKEY:-${YELLOW}undefined${ALL_OFF}}"
	if [[ ${proto} == ssh ]]; then
		msg2 "protocol: ${GREEN}${proto}${ALL_OFF}"
	else
		msg2 "protocol: ${YELLOW}${proto}${ALL_OFF}"
	fi

	# parallelization
	if [[ ${jobs} != 1 ]] && (( ${#paths[@]} > 1 )); then
		if [[ -n ${BOLD} ]]; then
			export DEVTOOLS_COLOR=always
		fi

		# warm up ssh connection as it may require user input (key unlock, hostkey verification etc)
		if [[ ${proto} == ssh ]]; then
			git_warmup_ssh_connection
		fi

		if ! parallel --bar --jobs "${jobs}" "${command}" ::: "${paths[@]}"; then
			die 'Failed to configure some packages, please check the output'
			exit 1
		fi
		exit 0
	fi

	for path in "${paths[@]}"; do
		# resolve symlink for basename
		if ! realpath=$(realpath --canonicalize-existing -- "${path}"); then
			die "No such directory: ${path}"
		fi
		# skip paths that aren't directories
		if [[ ! -d "${realpath}" ]]; then
			continue
		fi

		pkgbase=$(basename "${realpath}")
		pkgbase=${pkgbase%.git}
		msg "Configuring ${pkgbase}"

		if [[ ! -d "${path}/.git" ]]; then
			die "Not a Git repository: ${path}"
		fi

		pushd "${path}" >/dev/null

		project_path=$(gitlab_project_name_to_path "${pkgbase}")
		remote_url="${GIT_REPO_BASE_URL}/${project_path}.git"
		if ! git remote add origin "${remote_url}" &>/dev/null; then
			git remote set-url origin "${remote_url}"
		fi

		# move the master branch to main
		if [[ $(git symbolic-ref --quiet --short HEAD) == master ]]; then
			git branch --move main
			git config branch.main.merge refs/heads/main
		fi

		# configure spec version and variant to avoid using development hooks in production
		git config devtools.version "${GIT_REPO_SPEC_VERSION}"
		if [[ ${_DEVTOOLS_LIBRARY_DIR} == /usr/share/devtools ]]; then
			git config devtools.variant canonical
		else
			warning "Configuring with development version of pkgctl, do not use this repo in production"
			git config devtools.variant development
		fi

		git config pull.rebase true
		git config branch.autoSetupRebase always
		git config branch.main.remote origin
		git config branch.main.rebase true

		git config transfer.fsckobjects true
		git config fetch.fsckobjects true
		git config receive.fsckobjects true

		# setup author identity
		if [[ -n ${packager_name} ]]; then
			git config user.name "${packager_name}"
			git config user.email "${packager_email}"
		fi

		# force gpg for official packagers
		if (( official )); then
			git config commit.gpgsign true
		fi

		# set custom pgp key from makepkg.conf
		if [[ -n $GPGKEY ]]; then
			git config commit.gpgsign true
			git config user.signingKey "${GPGKEY}"
			git config gpg.format openpgp
		fi

		# set default git exclude
		mkdir -p .git/info
		ln -sf "${_DEVTOOLS_LIBRARY_DIR}/git.conf.d/template/info/exclude" \
			.git/info/exclude

		# set default git hooks
		mkdir -p .git/hooks
		rm -f .git/hooks/*.sample
		for hook in "${_DEVTOOLS_LIBRARY_DIR}"/git.conf.d/template/hooks/*; do
			ln -sf "${hook}" ".git/hooks/$(basename "${hook}")"
		done

		if ! git ls-remote origin &>/dev/null; then
			warning "configured remote origin may not exist, run:"
			msg2 "pkgctl repo create ${pkgbase}"
		fi

		popd >/dev/null
	done
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_UTIL_TERM_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_UTIL_TERM_SH=1

set -eo pipefail

readonly PKGCTL_TERM_ICON_CONFIDENTIAL=⛔
export PKGCTL_TERM_ICON_CONFIDENTIAL

readonly PKGCTL_TERM_SPINNER_DOTS=Dots
export PKGCTL_TERM_SPINNER_DOTS
readonly PKGCTL_TERM_SPINNER_DOTS12=Dots12
export PKGCTL_TERM_SPINNER_DOTS12
readonly PKGCTL_TERM_SPINNER_LINE=Line
export PKGCTL_TERM_SPINNER_LINE
readonly PKGCTL_TERM_SPINNER_SIMPLE_DOTS_SCROLLING=SimpleDotsScrolling
export PKGCTL_TERM_SPINNER_SIMPLE_DOTS_SCROLLING
readonly PKGCTL_TERM_SPINNER_TRIANGLE=Triangle
export PKGCTL_TERM_SPINNER_TRIANGLE
readonly PKGCTL_TERM_SPINNER_RANDOM=Random
export PKGCTL_TERM_SPINNER_RANDOM

readonly PKGCTL_TERM_SPINNER_TYPES=(
	"${PKGCTL_TERM_SPINNER_DOTS}"
	"${PKGCTL_TERM_SPINNER_DOTS12}"
	"${PKGCTL_TERM_SPINNER_LINE}"
	"${PKGCTL_TERM_SPINNER_SIMPLE_DOTS_SCROLLING}"
	"${PKGCTL_TERM_SPINNER_TRIANGLE}"
)
export PKGCTL_TERM_SPINNER_TYPES


term_cursor_hide() {
	tput civis >&2
}

term_cursor_show() {
	tput cnorm >&2
}

term_cursor_up() {
	tput cuu1
}

term_carriage_return() {
	tput cr
}

term_erase_line() {
	tput el
}

term_erase_lines() {
	local lines=$1

	local cursor_up erase_line
	cursor_up=$(term_cursor_up)
	erase_line="$(term_carriage_return)$(term_erase_line)"

	local prefix=''
	for _ in $(seq 1 "${lines}"); do
		printf '%s' "${prefix}${erase_line}"
		prefix="${cursor_up}"
	done
}

_pkgctl_spinner_type=${PKGCTL_TERM_SPINNER_RANDOM}
term_spinner_set_type() {
	_pkgctl_spinner_type=$1
}

# takes a status directory that can be used to dynamically update the spinner
# by writing to the `status` file inside that directory atomically.
# replace the placeholder %spinner% with the currently configured spinner type
term_spinner_start() {
	local status_dir=$1
	local parent_pid=$$
	(
		local spinner_type=${_pkgctl_spinner_type}
		local spinner_offset=0
		local frame_buffer=''
		local spinner status_message line

		local status_file="${status_dir}/status"
		local next_file="${status_dir}/next"
		local drawn_file="${status_dir}/drawn"

		# assign random spinner type
		if [[ ${spinner_type} == "${PKGCTL_TERM_SPINNER_RANDOM}" ]]; then
			spinner_type=${PKGCTL_TERM_SPINNER_TYPES[$((RANDOM % ${#PKGCTL_TERM_SPINNER_TYPES[@]}))]}
		fi

		# select spinner based on the named type
		case "${spinner_type}" in
			"${PKGCTL_TERM_SPINNER_DOTS}")
				spinner=("⠋" "⠙" "⠹" "⠸" "⠼" "⠴" "⠦" "⠧" "⠇" "⠏")
				update_interval=0.08
				;;
			"${PKGCTL_TERM_SPINNER_DOTS12}")
				spinner=("⢀⠀" "⡀⠀" "⠄⠀" "⢂⠀" "⡂⠀" "⠅⠀" "⢃⠀" "⡃⠀" "⠍⠀" "⢋⠀" "⡋⠀" "⠍⠁" "⢋⠁" "⡋⠁" "⠍⠉" "⠋⠉" "⠋⠉" "⠉⠙" "⠉⠙" "⠉⠩" "⠈⢙" "⠈⡙" "⢈⠩" "⡀⢙" "⠄⡙" "⢂⠩" "⡂⢘" "⠅⡘" "⢃⠨" "⡃⢐" "⠍⡐" "⢋⠠" "⡋⢀" "⠍⡁" "⢋⠁" "⡋⠁" "⠍⠉" "⠋⠉" "⠋⠉" "⠉⠙" "⠉⠙" "⠉⠩" "⠈⢙" "⠈⡙" "⠈⠩" "⠀⢙" "⠀⡙" "⠀⠩" "⠀⢘" "⠀⡘" "⠀⠨" "⠀⢐" "⠀⡐" "⠀⠠" "⠀⢀" "⠀⡀")
				update_interval=0.08
				;;
			"${PKGCTL_TERM_SPINNER_LINE}")
				spinner=("⎯" "\\" "|" "/")
				update_interval=0.13
				;;
			"${PKGCTL_TERM_SPINNER_SIMPLE_DOTS_SCROLLING}")
				spinner=(".  " ".. " "..." " .." "  ." "   ")
				update_interval=0.2
				;;
			"${PKGCTL_TERM_SPINNER_TRIANGLE}")
				spinner=("◢" "◣" "◤" "◥")
				update_interval=0.05
				;;
		esac

		# hide the cursor while spinning
		term_cursor_hide

		# run the spinner as long as the parent process didn't terminate
		while ps -p "${parent_pid}" &>/dev/null; do
			# cache the new status template if it exists
			if mv "${status_file}" "${next_file}" &>/dev/null; then
				status_message="$(cat "$next_file")"
			elif [[ -z "${status_message}" ]]; then
				# wait until we either have a new or cached status
				sleep 0.05
			fi

			# fill the frame buffer with the current status
			local prefix=''
			while IFS= read -r line; do
				# replace spinner placeholder
				line=${line//%spinner%/${spinner[spinner_offset%${#spinner[@]}]}}

				# append the current line to the frame buffer
				frame_buffer+="${prefix}${line}"
				prefix=$'\n'
			done <<< "${status_message}"

			# print current frame buffer
			echo -n "${frame_buffer}" >&2
			mv "${next_file}" "${drawn_file}" &>/dev/null ||:

			# setup next frame buffer to clear current content
			frame_buffer=$(term_erase_lines "$(awk 'END {print NR}' <<< "${status_message}")")

			# advance the spinner animation offset
			(( ++spinner_offset ))

			# sleep for the spinner update interval
			sleep "${update_interval}"
		done
	)&
	_pkgctl_spinner_pid=$!
	disown
}

term_spinner_stop() {
	local status_dir=$1
	local frame_buffer status_file

	# kill the spinner process
	if ! kill "${_pkgctl_spinner_pid}" > /dev/null 2>&1; then
		return 1
	fi
	unset _pkgctl_spinner_pid

	# acquire last drawn status
	status_file="${status_dir}/drawn"
	if [[ ! -f ${status_file} ]]; then
		return 0
	fi

	# clear terminal based on last status line
	frame_buffer=$(term_erase_lines "$(awk 'END {print NR}' < "${status_file}")")
	echo -n "${frame_buffer}" >&2

	# show the cursor after stopping the spinner
	term_cursor_show
}

prompt() {
	local message=$1
	local answer

	read -r -p "${message} (y/N) " answer

	case "${answer}" in
		y|Y|yes|Yes|YES)
			true
			;;
		*)
			false
			;;
	esac
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_ISSUE_CREATE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_ISSUE_CREATE_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh
# shellcheck source=src/lib/util/term.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/term.sh

set -eo pipefail


pkgctl_issue_create_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS]

		The create command is used to create a new issue for an Arch Linux package.
		This command is suitable for reporting bugs, regressions, feature requests, or
		any other issues related to a package. It provides a flexible way to document
		and track new issues within the project's issue tracking system.

		By default, the command operates within the current directory, but users can
		specify a different package base if needed.

		Users can provide a title for the issue directly through the command line.
		The command allows setting various labels and attributes for the issue, such as
		confidentiality, priority, scope, severity, and status.

		In case of a failed run, the command can automatically recover to ensure that
		the issue creation process is completed without losing any data.

		OPTIONS
		    -p, --package PKGBASE    Interact with <pkgbase> instead of the current directory
		    -t, --title TITLE        Use the provided title for the issue
		    -F, --file FILE          Take issue description from <file>
		    -e, --edit               Edit the issue description using an editor
		    -w, --web                Continue issue creation with the web interface
		    --recover                Automatically recover from a failed run
		    --confidentiality TYPE   Set the issue confidentiality
		    --priority PRIORITY      Set the priority label
		    --scope SCOPE            Set the scope label
		    --severity SEVERITY      Set the severity label
		    --status STATUS          Set the status label
		    -h, --help               Show this help text

		EXAMPLES
		    $ ${COMMAND} --package linux --title "some very informative title"
_EOF_
}

pkgctl_issue_create() {
	if (( $# < 1 )); then
		pkgctl_issue_create_usage
		exit 0
	fi

	local pkgbase=""
	local title_placeholder="PLACEHOLDER"
	local title="${title_placeholder}"
	local description=""
	local labels=()
	local msgfile=""
	local edit=0
	local web=0
	local recover=0
	local confidential=0

	local issue_template_url="https://gitlab.archlinux.org/archlinux/packaging/templates/-/raw/master/.gitlab/issue_templates/Default.md"
	local issue_template
	local recovery_home=${XDG_DATA_HOME:-$HOME/.local/share}/devtools/recovery
	local recovery_file
	local issue_url
	local project_path
	local result
	local iid
	local message
	local editor

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_issue_create_usage
				exit 0
				;;
			-t|--title)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				title=$2
				shift 2
				;;
			-p|--package)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				pkgbase=$2
				shift 2
				;;
			-F|--file)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				msgfile=$2
				shift 2
				;;
			-e|--edit)
				edit=1
				shift
				;;
			-w|--web)
				web=1
				shift
				;;
			--recover)
				recover=1
				shift
				;;
			--confidentiality)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! in_array "$2" "${DEVTOOLS_VALID_ISSUE_CONFIDENTIALITY[@]}"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				if [[ $2 == confidential ]]; then
					confidential=1
				fi
				shift 2
				;;
			--priority)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(priority_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				labels+=("$label")
				shift 2
				;;
			--scope)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(scope_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				labels+=("$label")
				shift 2
				;;
			--severity)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(severity_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				labels+=("$label")
				shift 2
				;;
			--status)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(status_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				labels+=("$label")
				shift 2
				;;
			*)
				die "invalid argument: %s" "$1"
				;;
		esac
	done

	if [[ -z ${pkgbase} ]]; then
		if ! [[ -f PKGBUILD ]]; then
			die "missing --package option or PKGBUILD in current directory"
		fi
		pkgbase=$(realpath --canonicalize-existing .)
	fi
	pkgbase=$(basename "${pkgbase}")
	project_path=$(gitlab_project_name_to_path "${pkgbase}")
	recovery_file="${recovery_home}/issue_create_${pkgbase}.md"

	# spawn web browser
	if (( web )); then
		if ! command -v xdg-open &>/dev/null; then
			die "The web option requires 'xdg-open'"
		fi
		issue_url="${GIT_PACKAGING_URL_HTTPS}/${project_path}/-/issues/new"
		echo "Opening ${issue_url} in your browser."
		xdg-open "${issue_url}"
		return
	fi

	# check existence of recovery file
	if [[ -f ${recovery_file} ]]; then
		if (( ! recover )); then
			msg_warn "Recovery file already exists: ${recovery_file}"
			if prompt "${GREEN}${BOLD}?${ALL_OFF} Do you want to recover?"; then
				msgfile=${recovery_file}
				recover=1
				edit=1
			fi
		fi
	fi

	# check existence of msgfile
	if [[ -n ${msgfile} ]]; then
		if [[ ! -f ${msgfile} ]]; then
			msg_error "File does not exist: ${msgfile}${ALL_OFF}"
			exit 1
		fi
	else
		# prepare msgfile and fetch the issue template
		if ! issue_template=$(curl --url "${issue_template_url}" --silent); then
			msg_error "Failed to fetch issue template${ALL_OFF}"
			exit 1
		fi
		# populate message file
		msgfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-issue-create.XXXXXXXXXX.md)
		edit=1
		printf "# Title: %s\n\n" "${title}" >> "${msgfile}"
		printf "%s\n" "${issue_template}" >> "${msgfile}"
	fi

	# spawn editor
	if (( edit )); then
		if [[ -n $VISUAL ]]; then
			editor=${VISUAL}
		elif [[ -n $EDITOR ]]; then
			editor=${EDITOR}
		else
			die "No usable editor found (tried \$VISUAL, \$EDITOR)."
		fi

		if ! ${editor} "${msgfile}"; then
			message=$(< "${msgfile}")
			pkgctl_issue_write_recovery_file "${pkgbase}" "${message}" "${recovery_file}" "${recover}"
		fi
	fi

	# check if the file contains a title
	message=$(< "${msgfile}")
	description=${message}
	if [[ ${message} == "# Title: "* ]]; then
		title=$(head --lines 1 <<< "${message}")
		title=${title//# Title: /}
		description=$(tail --lines +2 <<< "${message}")
		if [[ ${description} == $'\n'* ]]; then
			description=$(tail --lines +3 <<< "${message}")
		fi
	fi

	# validate title
	if [[ ${title} == 'PLACEHOLDER' ]]; then
		msg_error "Invalid issue title: ${title}${ALL_OFF}"
		pkgctl_issue_write_recovery_file "${pkgbase}" "${message}" "${recovery_file}" "${recover}"
		exit 1
	fi

	# create the issue
	if ! result=$(gitlab_project_issue_create "${pkgbase}" "${title}" "${description}" "${confidential}" "${labels[@]}"); then
		msg_error "Failed to create issue in ${BOLD}${pkgbase}${ALL_OFF}"
		pkgctl_issue_write_recovery_file "${pkgbase}" "${message}" "${recovery_file}" "${recover}"
		exit 1
	fi

	# delete old recovery file if we succeeded
	if [[ -f ${recovery_file} ]]; then
		rm --force "${recovery_file}"
	fi

	# read issue iid
	{ read -r iid; } < <(
		jq --raw-output ".iid" <<< "${result}"
	)
	issue_url="${GIT_PACKAGING_URL_HTTPS}/${project_path}/-/issues/${iid}"

	msg_success "Created new issue ${BOLD}#${iid}${ALL_OFF}"
	printf "%sView this issue on GitLab: %s%s\n" "${GRAY}" "${issue_url}" "${ALL_OFF}"
}

pkgctl_issue_write_recovery_file() {
	local pkgbase=$1
	local message=$2
	local recovery_file=$3
	local recover=$4

	if [[ -f ${recovery_file} ]] && (( ! recover )); then
		msg_warn "Recovery file already exists: ${recovery_file}"
		if ! prompt "${YELLOW}${BOLD}?${ALL_OFF} Are you sure you want to overwrite it?"; then
			return 1
		fi
	fi

	mkdir -p "$(dirname "${recovery_file}")"
	printf "%s\n" "${message}" > "${recovery_file}"

	printf "Created recovery file: %s\n" "${recovery_file}"
	return 0
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_BUILD_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_BUILD_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/build/offload.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/build/offload.sh
# shellcheck source=src/lib/db/update.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/db/update.sh
# shellcheck source=src/lib/release.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/release.sh
# shellcheck source=src/lib/util/git.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/git.sh
# shellcheck source=src/lib/util/makepkg.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/makepkg.sh
# shellcheck source=src/lib/util/srcinfo.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/srcinfo.sh
# shellcheck source=src/lib/util/pacman.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/pacman.sh
# shellcheck source=src/lib/util/pkgbuild.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/pkgbuild.sh
# shellcheck source=src/lib/valid-build-install.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/valid-build-install.sh
# shellcheck source=src/lib/valid-repos.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/valid-repos.sh
# shellcheck source=src/lib/valid-tags.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/valid-tags.sh
# shellcheck source=src/lib/valid-inspect.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/valid-inspect.sh

source /usr/share/makepkg/util/message.sh

set -eo pipefail


pkgctl_build_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
    cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [PATH]...

		Build packages inside a clean chroot

		Build packages in clean chroot environment, offering various options
		and functionalities to customize the package building process.

		By default, chroot environments are located in /var/lib/archbuild/.

		BUILD OPTIONS
		    --arch ARCH          Specify architectures to build for (disables auto-detection)
		    --repo REPO          Specify target repository for new packages not in any official repo
		    -s, --staging        Build against the staging counterpart of the auto-detected repo
		    -t, --testing        Build against the testing counterpart of the auto-detected repo
		    -o, --offload        Build on a remote server and transfer artifacts afterwards
		    --offload-host HOST  Specify offload target machine (default: build.archlinux.org)
		    -c, --clean          Recreate the chroot before building
		    --inspect WHEN       Spawn an interactive shell to inspect the chroot (never, always, failure)
		    -w, --worker SLOT    Name of the worker slot, useful for concurrent builds (disables automatic names)
		    --nocheck            Do not run the check() function in the PKGBUILD

		INSTALL OPTIONS
		    -I, --install-to-chroot FILE   Install a package to the working copy of the chroot
		    -i, --install-to-host MODE     Install the built package to the host system, possible modes are 'all' and 'auto'

		PKGBUILD OPTIONS
		    --pkgver=PKGVER      Set pkgver, reset pkgrel and update checksums
		    --pkgrel=PKGREL      Set pkgrel to a given value
		    --rebuild            Increment the current pkgrel variable
		    --update-checksums   Force computation and update of the checksums (disables auto-detection)
		    -e, --edit           Edit the PKGBUILD before building

		RELEASE OPTIONS
		    -r, --release        Automatically commit, tag and release after building
		    -m, --message MSG    Use the given <msg> as the commit message
		    -u, --db-update      Automatically update the pacman database as last action

		OPTIONS
		    -h, --help           Show this help text

		EXAMPLES
		    $ ${COMMAND}
		    $ ${COMMAND} --rebuild --staging --release --message 'libyay 0.42 rebuild' libfoo libbar
		    $ ${COMMAND} --pkgver=1.42 --release --db-update
_EOF_
}

pkgctl_build_check_option_group_repo() {
	local option=$1
	local repo=$2
	local testing=$3
	local staging=$4
	if [[ -n "${repo}" ]] || (( testing )) || (( staging )); then
		die "The argument '%s' cannot be used with one or more of the other specified arguments" "${option}"
		exit 1
	fi
	return 0
}

pkgctl_build_check_option_group_ver() {
	local option=$1
	local pkgver=$2
	local pkgrel=$3
	local rebuild=$4
	if [[ -n "${pkgver}" ]] || [[ -n "${pkgrel}" ]] || (( rebuild )); then
		die "The argument '%s' cannot be used with one or more of the other specified arguments" "${option}"
		exit 1
	fi
	return 0
}

# TODO: import pgp keys
pkgctl_build() {
	if (( $# < 1 )) && [[ ! -f PKGBUILD ]]; then
		pkgctl_build_usage
		exit 1
	fi

	local UPDATE_CHECKSUMS=0
	local EDIT=0
	local REBUILD=0
	local OFFLOAD=0
	local OFFLOAD_HOST=build.archlinux.org
	local STAGING=0
	local TESTING=0
	local RELEASE=0
	local DB_UPDATE=0
	local INSTALL_TO_HOST=none

	local REPO=
	local PKGVER=
	local PKGREL=
	local MESSAGE=

	local paths=()
	local BUILD_ARCH=()
	local BUILD_OPTIONS=()
	local MAKECHROOT_OPTIONS=()
	local RELEASE_OPTIONS=()
	local MAKEPKG_OPTIONS=()
	local INSTALL_HOST_PACKAGES=()

	local WORKER=
	local WORKER_SLOT=

	# variables
	local _arch path pkgbase pkgrepo source pkgbuild_checksum current_checksum

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_build_usage
				exit 0
				;;
			--repo)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				pkgctl_build_check_option_group_repo '--repo' "${REPO}" "${TESTING}" "${STAGING}"
				REPO="${2}"
				RELEASE_OPTIONS+=("--repo" "${REPO}")
				shift 2
				;;
			--arch)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if [[ ${2} == all ]]; then
					BUILD_ARCH=("${DEVTOOLS_VALID_ARCHES[@]::${#DEVTOOLS_VALID_ARCHES[@]}-1}")
				elif [[ ${2} == any ]]; then
					BUILD_ARCH=("${DEVTOOLS_VALID_ARCHES[0]}")
				elif ! in_array "${2}" "${BUILD_ARCH[@]}"; then
					if ! in_array "${2}" "${DEVTOOLS_VALID_ARCHES[@]}"; then
						warning 'invalid architecture: %s' "${2}"
					fi
					BUILD_ARCH+=("${2}")
				fi
				shift 2
				;;
			--pkgver=*)
				pkgctl_build_check_option_group_ver '--pkgver' "${PKGVER}" "${PKGREL}" "${REBUILD}"
				PKGVER="${1#*=}"
				PKGREL=1
				UPDATE_CHECKSUMS=1
				shift
				;;
			--pkgrel=*)
				pkgctl_build_check_option_group_ver '--pkgrel' "${PKGVER}" "${PKGREL}" "${REBUILD}"
				PKGREL="${1#*=}"
				shift
				;;
			--update-checksums)
				UPDATE_CHECKSUMS=1
				shift
				;;
			--rebuild)
				# shellcheck source=src/lib/util/git.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/git.sh
				pkgctl_build_check_option_group_ver '--rebuild' "${PKGVER}" "${PKGREL}" "${REBUILD}"
				REBUILD=1
				shift
				;;
			-e|--edit)
				EDIT=1
				shift
				;;
			-o|--offload)
				OFFLOAD=1
				shift
				;;
			--offload-host)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				OFFLOAD_HOST="$2"
				shift 2
				;;
			-s|--staging)
				pkgctl_build_check_option_group_repo '--staging' "${REPO}" "${TESTING}" "${STAGING}"
				STAGING=1
				RELEASE_OPTIONS+=("--staging")
				shift
				;;
			-t|--testing)
				pkgctl_build_check_option_group_repo '--testing' "${REPO}" "${TESTING}" "${STAGING}"
				TESTING=1
				RELEASE_OPTIONS+=("--testing")
				shift
				;;
			-c|--clean)
				BUILD_OPTIONS+=("-c")
				shift
				;;
			-I|--install-to-chroot)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if (( OFFLOAD )); then
					MAKECHROOT_OPTIONS+=("-I" "$2")
				else
					MAKECHROOT_OPTIONS+=("-I" "$(realpath "$2")")
				fi
				warning 'installing packages to the chroot may break reproducible builds, use with caution!'
				shift 2
				;;
			-i|--install-to-host)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! in_array "$2" "${DEVTOOLS_VALID_BUILD_INSTALL[@]}"; then
					die 'invalid install mode: %s' "${2}"
				fi
				INSTALL_TO_HOST=$2
				shift 2
				;;
			--nocheck)
				MAKEPKG_OPTIONS+=("--nocheck")
				warning 'not running checks is disallowed for official packages, except for bootstrapping. Please rebuild after bootstrapping is completed!'
				shift
				;;
			--inspect)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! in_array "${2}" "${DEVTOOLS_VALID_INSPECT_MODES[@]}"; then
					die "Invalid inspect mode: %s" "${2}"
				fi
				MAKECHROOT_OPTIONS+=("-x" "${2}")
				shift 2
				;;
			-r|--release)
				# shellcheck source=src/lib/release.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/release.sh
				RELEASE=1
				shift
				;;
			-m|--message)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				MESSAGE=$2
				RELEASE_OPTIONS+=("--message" "${MESSAGE}")
				shift 2
				;;
			-u|--db-update)
				DB_UPDATE=1
				shift
				;;
			-w|--worker)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				WORKER_SLOT=$2
				shift 2
				;;
			--)
				shift
				break
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				paths=("$@")
				break
				;;
		esac
	done

	# check if any release specific options were specified without releasing
	if (( ! RELEASE )); then
		if (( DB_UPDATE )); then
			die "cannot use --db-update without --release"
		fi
		if [[ -n "${MESSAGE}" ]]; then
			die "cannot use --message without --release"
		fi
	fi

	# check if invoked without any path from within a packaging repo
	if (( ${#paths[@]} == 0 )); then
		if [[ -f PKGBUILD ]]; then
			paths=(".")
		else
			pkgctl_build_usage
			exit 1
		fi
	fi

	# assign default worker slot
	if [[ -z ${WORKER_SLOT} ]] && ! WORKER_SLOT="$(tty | sed 's|/dev/pts/||')"; then
		WORKER_SLOT=$(( RANDOM % $(nproc) + 1 ))
	fi
	WORKER="${USER}-${WORKER_SLOT}"

	# Update pacman cache for auto-detection
	if [[ -z ${REPO} ]]; then
		update_pacman_repo_cache multilib
	# Check valid repos if not resolved dynamically
	elif ! in_array "${REPO}" "${DEVTOOLS_VALID_REPOS[@]}"; then
		die "Invalid repository target: %s" "${REPO}"
	fi

	for path in "${paths[@]}"; do
		# skip paths that are not directories
		if [[ ! -d "${path}" ]]; then
			continue
		fi
		pushd "${path}" >/dev/null

		if [[ ! -f PKGBUILD ]]; then
			die 'PKGBUILD not found in %s' "${path}"
		fi

		source=()
		# shellcheck source=contrib/makepkg/PKGBUILD.proto
		. ./PKGBUILD
		pkgbase=${pkgbase:-$pkgname}
		pkgrepo=${REPO}
		pkgbuild_checksum=$(b2sum PKGBUILD | awk '{print $1}')
		msg "Building ${pkgbase}"

		# auto-detect target repository
		if ! repo=$(get_pacman_repo_from_pkgbuild PKGBUILD); then
			die 'Failed to query pacman repo'
		fi

		# fail if an existing package specifies --repo
		if [[ -n "${repo}" ]] && [[ -n ${pkgrepo} ]]; then
			# allow unstable to use --repo
			if [[ ${pkgrepo} == *unstable ]]; then
				repo=${pkgrepo}
			else
				die 'Using --repo for packages that exist in official repositories is disallowed'
			fi
		fi

		# assign auto-detected target repository
		if [[ -n ${repo} ]]; then
			pkgrepo=${repo}
		# fallback to extra for unreleased packages
		elif [[ -z ${pkgrepo} ]]; then
			pkgrepo=extra
		fi

		# special cases to resolve final build target
		if (( TESTING )); then
			pkgrepo="${pkgrepo}-testing"
		elif (( STAGING )); then
			pkgrepo="${pkgrepo}-staging"
		elif [[ $pkgrepo == core ]]; then
			pkgrepo="${pkgrepo}-testing"
		fi

		# auto-detection of build architecture
		if [[ $pkgrepo = multilib* ]]; then
			BUILD_ARCH=("")
		elif (( ${#BUILD_ARCH[@]} == 0 )); then
			if in_array any "${arch[@]}"; then
				BUILD_ARCH=("${DEVTOOLS_VALID_ARCHES[0]}")
			else
				for _arch in "${arch[@]}"; do
					if in_array "${_arch}" "${DEVTOOLS_VALID_ARCHES[@]}"; then
						BUILD_ARCH+=("$_arch")
					else
						warning 'invalid architecture, not building for: %s' "${_arch}"
					fi
				done
			fi
		fi

		# print gathered build modes
		msg2 "  repo: ${pkgrepo}"
		msg2 "  arch: ${BUILD_ARCH[*]}"
		msg2 "worker: ${WORKER}"

		# increment pkgrel on rebuild
		if (( REBUILD )); then
			# try to figure out if pkgrel has been changed
			if ! old_pkgrel=$(git_diff_tree HEAD PKGBUILD | grep --perl-regexp --only-matching --max-count=1 '^-pkgrel=\K\w+'); then
				old_pkgrel=${pkgrel}
			fi
			# check if pkgrel conforms expectations
			[[ ${pkgrel/.*} =~ ^[0-9]+$ ]] || die "Non-standard pkgrel declaration"
			[[ ${old_pkgrel/.*} =~ ^[0-9]+$ ]] || die "Non-standard pkgrel declaration"
			# increment pkgrel if it hasn't been changed yet
			if [[ ${pkgrel} = "${old_pkgrel}" ]]; then
				PKGREL=$((${pkgrel/.*}+1))
			else
				warning 'ignoring --rebuild as pkgrel has already been incremented from %s to %s' "${old_pkgrel}" "${pkgrel}"
			fi
		fi

		# update pkgver
		if [[ -n ${PKGVER} ]]; then
			msg "Bumping pkgver to ${PKGVER}"
			pkgbuild_set_pkgver "${PKGVER}"
		fi

		# update pkgrel
		if [[ -n ${PKGREL} ]]; then
			msg "Bumping pkgrel to ${PKGREL}"
			pkgbuild_set_pkgrel "${PKGREL}"
		fi

		# edit PKGBUILD
		if (( EDIT )); then
			stat_busy 'Editing PKGBUILD'
			if [[ -n $GIT_EDITOR ]]; then
				$GIT_EDITOR PKGBUILD || die
			elif [[ -n $VISUAL ]]; then
				$VISUAL PKGBUILD || die
			elif [[ -n $EDITOR ]]; then
				$EDITOR PKGBUILD || die
			elif giteditor=$(git config --get core.editor); then
				$giteditor PKGBUILD || die
			else
				die "No usable editor found (tried \$GIT_EDITOR, \$VISUAL, \$EDITOR, git config [core.editor])."
			fi
			stat_done
		fi

		# update checksums if any sources are declared
		if (( UPDATE_CHECKSUMS )) && (( ${#source[@]} >= 1 )); then
			if ! result=$(pkgbuild_update_checksums /dev/stderr); then
				die "${result}"
			fi
		fi

		# re-source the PKGBUILD if it changed
		current_checksum="$(b2sum PKGBUILD | awk '{print $1}')"
		if [[ ${pkgbuild_checksum} != "${current_checksum}" ]]; then
			pkgbuild_checksum=${current_checksum}
			# shellcheck source=contrib/makepkg/PKGBUILD.proto
			. ./PKGBUILD
		fi

		# execute build
		for arch in "${BUILD_ARCH[@]}"; do
			if [[ -n $arch ]]; then
				msg "Building ${pkgbase} for [${pkgrepo}] (${arch})"
				BUILDTOOL="${pkgrepo}-${arch}-build"
			else
				msg "Building ${pkgbase} for [${pkgrepo}]"
				BUILDTOOL="${pkgrepo}-build"
			fi

			if (( OFFLOAD )); then
				pkgctl_build_offload_client "${pkgbase}" "${pkgrepo}" "${arch}" "${OFFLOAD_HOST}" "${BUILD_OPTIONS[@]}" -- "${MAKECHROOT_OPTIONS[@]}" -l "${WORKER}" -- "${MAKEPKG_OPTIONS[@]}"
			else
				"${BUILDTOOL}" "${BUILD_OPTIONS[@]}" -- "${MAKECHROOT_OPTIONS[@]}" -l "${WORKER}" -- "${MAKEPKG_OPTIONS[@]}"
			fi
		done

		# re-source the PKGBUILD if it changed
		current_checksum="$(b2sum PKGBUILD | awk '{print $1}')"
		if [[ ${pkgbuild_checksum} != "${current_checksum}" ]]; then
			pkgbuild_checksum=${current_checksum}
			# shellcheck source=contrib/makepkg/PKGBUILD.proto
			. ./PKGBUILD
		fi

		# auto generate .SRCINFO
		# shellcheck disable=SC2119
		write_srcinfo_file

		version=$(get_full_version)
		msg "Finished building %s %s" "${pkgbase}" "${version}"

		# test-install (some of) the produced packages
		if [[ ${INSTALL_TO_HOST} == auto ]] || [[ ${INSTALL_TO_HOST} == all ]]; then
			makepkg_load_config

			# this is inspired by print_all_package_names from libmakepkg
			local version pkg_architecture pkg pkgfile
			version=$(get_full_version)

			for pkg in "${pkgname[@]}"; do
				pkg_architecture=$(get_pkg_arch "$pkg")
				pkgfile=$(realpath "$(printf "%s/%s-%s-%s%s\n" "${PKGDEST:-.}" "$pkg" "$version" "$pkg_architecture" "$PKGEXT")")

				# check if we install all packages or if the (split-)package is already installed
				if [[ ${INSTALL_TO_HOST} == all ]] || ( [[ ${INSTALL_TO_HOST} == auto ]] && pacman -Qq -- "$pkg" &>/dev/null ); then
					INSTALL_HOST_PACKAGES+=("$pkgfile")
				fi
			done
		fi

		# release the build
		if (( RELEASE )); then
			pkgctl_release "${RELEASE_OPTIONS[@]}"
		fi

		# reset common PKGBUILD variables
		unset pkgbase pkgname arch pkgrepo source pkgver pkgrel validpgpkeys
		popd >/dev/null
	done

	# install all collected packages to the host system
	if (( ${#INSTALL_HOST_PACKAGES[@]} )); then
		msg "Installing built packages to the host system"
		sudo pacman -U -- "${INSTALL_HOST_PACKAGES[@]}"
	fi

	# update the binary package repo db as last action
	if (( RELEASE )) && (( DB_UPDATE )); then
		# shellcheck disable=2119
		pkgctl_db_update
	fi
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_ISSUE_CLOSE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_ISSUE_CLOSE_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh

set -eo pipefail


pkgctl_issue_close_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [IID]

		This command is used to close an issue in Arch Linux packaging projects. It
		finalizes the issue by marking it as resolved and optionally providing a reason
		for its closure.

		By default, the command operates within the current directory, but users have
		the option to specify a different package base.

		Users can provide a message directly through the command line to explain the
		reason for closing the issue. Additionally, a specific resolution label can be
		set to categorize the closure reason, with the default label being "completed."

		OPTIONS
		    -p, --package PKGBASE    Interact with <pkgbase> instead of the current directory
		    -m, --message MSG        Use the provided message as the reason for closing
		    -e, --edit               Edit the reason for closing using an editor
		    -r, --resolution REASON  Set a specific resolution label (default: completed)
		    -h, --help               Show this help text

		EXAMPLES
		    $ ${COMMAND} 42
		    $ ${COMMAND} --edit --package linux 42
_EOF_
}

pkgctl_issue_close() {
	if (( $# < 1 )); then
		pkgctl_issue_close_usage
		exit 0
	fi

	local iid=""
	local pkgbase=""
	local message=""
	local edit=0
	local labels=()
	local resolution="completed"

	local issue note
	local params="state_event=close"

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_issue_close_usage
				exit 0
				;;
			-m|--message)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				message=$2
				shift 2
				;;
			-p|--package)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				pkgbase=$2
				shift 2
				;;
			-e|--edit)
				edit=1
				shift
				;;
			-r|--resolution)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(resolution_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				params+="&add_labels=${label}"
				shift 2
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				iid=$1
				shift
				;;
		esac
	done

	if [[ -z ${iid} ]]; then
		die "missing issue iid argument"
	fi

	if [[ -z ${pkgbase} ]]; then
		if ! [[ -f PKGBUILD ]]; then
			die "missing --package option or PKGBUILD in current directory"
		fi
		pkgbase=$(realpath --canonicalize-existing .)
	fi
	pkgbase=$(basename "${pkgbase}")

	# spawn editor
	if (( edit )); then
		msgfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-issue-note.XXXXXXXXXX.md)
		printf "%s\n" "${message}" >> "${msgfile}"
		if [[ -n $VISUAL ]]; then
			$VISUAL "${msgfile}" || die
		elif [[ -n $EDITOR ]]; then
			$EDITOR "${msgfile}" || die
		else
			die "No usable editor found (tried \$VISUAL, \$EDITOR)."
		fi
		message=$(cat "${msgfile}")
	fi

	# comment on issue
	if [[ -n ${message} ]]; then
		if ! note=$(gitlab_create_project_issue_note "${pkgbase}" "${iid}" "${message}"); then
			msg_error "Failed to comment on issue ${BOLD}#${iid}${ALL_OFF}"
			exit 1
		fi
		msg_success "Commented on issue ${BOLD}#${iid}${ALL_OFF}"
	fi

	# close issue
	if ! issue=$(gitlab_project_issue_edit "${pkgbase}" "${iid}" "${params}"); then
		msg_error "Failed to close issue ${BOLD}#${iid}${ALL_OFF}"
		exit 1
	fi
	msg_success "Closed issue ${BOLD}#${iid}${ALL_OFF}"
	echo

	{ read -r iid; read -r title; read -r state; read -r created_at; read -r author; } < <(
		jq --raw-output ".iid, .title, .state, .created_at, .author.username" <<< "${issue}"
	)
	mapfile -t labels < <(
		jq --raw-output ".labels[]" <<< "${issue}"
	)

	severity="$(gitlab_severity_from_labels "${labels[@]}")"
	severity_color="$(gitlab_severity_color "${severity}")"
	created_at=$(relative_date_unit "${created_at}")
	state_color="$(gitlab_issue_state_color "${state}")"
	state="$(gitlab_issue_state_display "${state}")"
	status="$(gitlab_issue_status_from_labels "${labels[@]}")"
	status_color="$(gitlab_issue_status_color "${status}")"

	scope="$(gitlab_scope_from_labels "${labels[@]}")"
	scope_color="$(gitlab_scope_color "${scope}")"
	scope_label=""
	if [[ ${scope} != unknown ]]; then
		scope_label="${scope_color}${scope}${ALL_OFF} ${GRAY}•${ALL_OFF} "
	fi

	resolution_label=""
	if resolution="$(gitlab_resolution_from_labels "${labels[@]}")"; then
		resolution_color="$(gitlab_resolution_color "${resolution}")"
		resolution_label="${resolution_color}${resolution}${ALL_OFF} ${GRAY}•${ALL_OFF} "
	fi

	printf "%s%s • %s%sseverity %s • %s • %s%sopened by %s %s ago%s\n" \
		"${state_color}${state}${ALL_OFF}" "${GRAY}" "${resolution_label}" "${severity_color}" "${severity}${GRAY}" \
		"${status_color}${status}${GRAY}" "${scope_label}" "${GRAY}" "${author}" "${created_at}" "${ALL_OFF}"
	printf "%s %s\n" "${BOLD}${title}${ALL_OFF}" "${GRAY}#${iid}${ALL_OFF}"

	# show comment
	if [[ -n ${note} ]]; then
		{ read -r created_at; read -r author; } < <(
			jq --raw-output ".created_at, .author.username" <<< "${note}"
		)
		body=$(jq --raw-output ".body" <<< "${note}")
		created_at=$(relative_date_unit "${created_at}")

		echo
		echo "${BOLD}Comments / Notes${ALL_OFF}"
		printf -v spaces '%*s' $(( COLUMNS - 2 )) ''
		printf '%s\n\n' "${spaces// /─}"

		printf "%s commented%s %s ago%s\n" "${author}" "${GRAY}" "${created_at}" "${ALL_OFF}"
		echo "${body}" | glow
	fi
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_UTIL_SRCINFO_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_UTIL_SRCINFO_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh

source /usr/share/makepkg/util/util.sh
source /usr/share/makepkg/srcinfo.sh

set -eo pipefail


print_srcinfo() {
	local pkgpath=${1:-.}
	local outdir pkg pid
	local pids=()

	# source the PKGBUILD
	# shellcheck source=contrib/makepkg/PKGBUILD.proto
	. "${pkgpath}"/PKGBUILD

	# run without parallelization for single packages
	if (( ${#pkgname[@]} == 1 )); then
		write_srcinfo_content
		return 0
	fi

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	outdir=$(mktemp --directory --tmpdir="${WORKDIR}" pkgctl-srcinfo.XXXXXXXXXX)

	# fork workload for each split pkgname
	for pkg in "${pkgname[@]}"; do
		(
			# deactivate errexit to avoid makepkg abort on grep_function
			set +e
			srcinfo_write_package "$pkg" > "${outdir}/${pkg}"
		)&
		pids+=($!)
	done

	# join workload
	for pid in "${pids[@]}"; do
		if ! wait "${pid}"; then
			return 1
		fi
	done

	# collect output
	srcinfo_write_global
	for pkg in "${pkgname[@]}"; do
		srcinfo_separate_section
		cat "${outdir}/${pkg}"
	done
}

write_srcinfo_file() {
	local pkgpath=${1:-.}
	stat_busy 'Generating .SRCINFO'
	if ! print_srcinfo "${pkgpath}" > "${pkgpath}"/.SRCINFO; then
		error 'Failed to write .SRCINFO file'
		return 1
	fi
	stat_done
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_UTIL_GIT_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_UTIL_GIT_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh


git_diff_tree() {
	local commit=$1
	local path=$2
	git \
		--no-pager \
		diff \
		--color=never \
		--color-moved=no \
		--unified=0 \
		--no-prefix \
		--no-ext-diff \
		"${commit}" \
		-- "${path}"
}

git_warmup_ssh_connection() {
	msg 'Establishing ssh connection to git@%s' "${GITLAB_HOST}"
	if ! ssh -T "git@${GITLAB_HOST}" >/dev/null; then
		die 'Failed to establish ssh connection to git@%s' "${GITLAB_HOST}"
	fi
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_ISSUE_REOPEN_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_ISSUE_REOPEN_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh

set -eo pipefail


pkgctl_issue_reopen_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [IID]

		The reopen command is used to reopen a previously closed issue in Arch Linux
		packaging projects. This command is useful when an issue needs to be revisited
		or additional work is required after it was initially closed.

		By default, the command operates within the current directory, but users can
		specify a different package base if needed.

		Users can provide a message directly through the command line to explain the
		reason for reopening the issue.

		OPTIONS
		    -p, --package PKGBASE  Interact with <pkgbase> instead of the current directory
		    -m, --message MSG      Use the provided message as the comment
		    -e, --edit             Edit the comment using an editor
		    -h, --help             Show this help text

		EXAMPLES
		    $ ${COMMAND} 42
		    $ ${COMMAND} --package linux 42
_EOF_
}

pkgctl_issue_reopen() {
	if (( $# < 1 )); then
		pkgctl_issue_reopen_usage
		exit 0
	fi

	local iid=""
	local pkgbase=""
	local message=""
	local edit=0

	local issue note result resolution labels
	local params="state_event=reopen"

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_issue_reopen_usage
				exit 0
				;;
			-p|--package)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				pkgbase=$2
				shift 2
				;;
			-m|--message)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				message=$2
				shift 2
				;;
			-e|--edit)
				edit=1
				shift
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				iid=$1
				shift
				;;
		esac
	done

	if [[ -z ${iid} ]]; then
		die "missing issue iid argument"
	fi

	if [[ -z ${pkgbase} ]]; then
		if ! [[ -f PKGBUILD ]]; then
			die "missing --package option or PKGBUILD in current directory"
		fi
		pkgbase=$(realpath --canonicalize-existing .)
	fi
	pkgbase=$(basename "${pkgbase}")


	# spawn editor
	if (( edit )); then
		msgfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-issue-note.XXXXXXXXXX.md)
		printf "%s\n" "${message}" >> "${msgfile}"
		if [[ -n $VISUAL ]]; then
			$VISUAL "${msgfile}" || die
		elif [[ -n $EDITOR ]]; then
			$EDITOR "${msgfile}" || die
		else
			die "No usable editor found (tried \$VISUAL, \$EDITOR)."
		fi
		message=$(< "${msgfile}")
	fi

	# query issue details
	if ! result=$(gitlab_project_issue "${pkgbase}" "${iid}"); then
		die "Failed to fetch issue ${pkgbase} #${iid}"
	fi
	mapfile -t labels < <(
		jq --raw-output ".labels[]" <<< "${result}"
	)
	if resolution=$(gitlab_resolution_from_labels "${labels[@]}"); then
		resolution=$(resolution_as_gitlab_label "${resolution}")
		params+="&remove_labels=${resolution}"
	fi

	# comment on issue
	if [[ -n ${message} ]]; then
		if ! note=$(gitlab_create_project_issue_note "${pkgbase}" "${iid}" "${message}"); then
			msg_error "Failed to comment on issue ${BOLD}#${iid}${ALL_OFF}"
			exit 1
		fi
		msg_success "Commented on issue ${BOLD}#${iid}${ALL_OFF}"
	fi

	# reopen issue
	if ! issue=$(gitlab_project_issue_edit "${pkgbase}" "${iid}" "${params}"); then
		msg_error "Failed to reopen issue ${BOLD}#${iid}${ALL_OFF}"
		exit 1
	fi
	msg_success "Reopened issue ${BOLD}#${iid}${ALL_OFF}"
	echo

	{ read -r iid; read -r title; read -r state; read -r created_at; read -r author; } < <(
		jq --raw-output ".iid, .title, .state, .created_at, .author.username" <<< "${issue}"
	)
	mapfile -t labels < <(
		jq --raw-output ".labels[]" <<< "${issue}"
	)

	severity="$(gitlab_severity_from_labels "${labels[@]}")"
	severity_color="$(gitlab_severity_color "${severity}")"
	created_at=$(relative_date_unit "${created_at}")
	state_color="$(gitlab_issue_state_color "${state}")"
	state="$(gitlab_issue_state_display "${state}")"
	status="$(gitlab_issue_status_from_labels "${labels[@]}")"
	status_color="$(gitlab_issue_status_color "${status}")"

	scope="$(gitlab_scope_from_labels "${labels[@]}")"
	scope_color="$(gitlab_scope_color "${scope}")"
	scope_label=""
	if [[ ${scope} != unknown ]]; then
		scope_label="${scope_color}${scope}${ALL_OFF} ${GRAY}•${ALL_OFF} "
	fi

	printf "%s%s • %sseverity %s • %s • %s%sopened by %s %s ago%s\n" \
		"${state_color}${state}${ALL_OFF}" "${GRAY}" "${severity_color}" "${severity}${GRAY}" \
		"${status_color}${status}${GRAY}" "${scope_label}" "${GRAY}" "${author}" "${created_at}" "${ALL_OFF}"
	printf "%s %s\n" "${BOLD}${title}${ALL_OFF}" "${GRAY}#${iid}${ALL_OFF}"

	# show comment
	if [[ -n ${note} ]]; then
		{ read -r created_at; read -r author; } < <(
			jq --raw-output ".created_at, .author.username" <<< "${note}"
		)
		body=$(jq --raw-output ".body" <<< "${note}")
		created_at=$(relative_date_unit "${created_at}")

		echo
		echo "${BOLD}Comments / Notes${ALL_OFF}"
		printf -v spaces '%*s' $(( COLUMNS - 2 )) ''
		printf '%s\n\n' "${spaces// /─}"

		printf "%s commented%s %s ago%s\n" "${author}" "${GRAY}" "${created_at}" "${ALL_OFF}"
		echo "${body}" | glow
	fi
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_ISSUE_VIEW_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_ISSUE_VIEW_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh
# shellcheck source=src/lib/util/term.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/term.sh

set -eo pipefail


pkgctl_issue_view_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [IID]

		This command is designed to display detailed information about a specific issue
		in Arch Linux packaging projects. It gathers and pretty prints all relevant
		data about the issue, providing a comprehensive view that includes the issue's
		description, status as well as labels and creation date.

		By default, the command operates within the current directory, but users have
		the option to specify a different package base. Additionally, users can choose
		to view the issue in a web browser for a more interactive experience.

		OPTIONS
		    -p, --package PKGBASE  Interact with <pkgbase> instead of the current directory
		    -c, --comments         Show issue comments and activities
		    -w, --web              Open issue in a browser
		    -h, --help             Show this help text

		EXAMPLES
		    $ ${COMMAND} 4
		    $ ${COMMAND} --web --package linux 4
_EOF_
}

pkgctl_issue_view() {
	if (( $# < 1 )); then
		pkgctl_issue_view_usage
		exit 0
	fi

	local web=0
	local comments=0
	local pkgbase=""
	local iid=""

	local project_path

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_issue_view_usage
				exit 0
				;;
			-p|--package)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				pkgbase=$2
				shift 2
				;;
			-w|--web)
				web=1
				shift
				;;
			-c|--comments)
				comments=1
				shift
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				iid=$1
				shift
				;;
		esac
	done

	if [[ -z ${iid} ]]; then
		die "missing issue iid argument"
	fi

	if [[ -z ${pkgbase} ]]; then
		if ! [[ -f PKGBUILD ]]; then
			die "missing --package option or PKGBUILD in current directory"
		fi
		pkgbase=$(realpath --canonicalize-existing .)
	fi
	pkgbase=$(basename "${pkgbase}")

	project_path=$(gitlab_project_name_to_path "${pkgbase}")

	if ! result=$(gitlab_project_issue "${pkgbase}" "${iid}"); then
		die "Failed to view issue ${pkgbase} #${iid}"
	fi

	{ read -r iid; read -r title; read -r state; read -r created_at; read -r closed_at; read -r author; } < <(
		jq --raw-output ".iid, .title, .state, .created_at, .closed_at, .author.username" <<< "${result}"
	)
	{ read -r upvotes; read -r downvotes; read -r user_notes_count; read -r confidential; } < <(
		jq --raw-output ".upvotes, .downvotes, .user_notes_count, .confidential" <<< "${result}"
	)
	description=$(jq --raw-output ".description" <<< "${result}")
	mapfile -t labels < <(
		jq --raw-output ".labels[]" <<< "${result}"
	)
	mapfile -t assignees < <(
		jq --raw-output ".assignees[].username" <<< "${result}"
	)
	if [[ ${closed_at} != null ]]; then
		closed_by=$(jq --raw-output ".closed_by.username" <<< "${result}")
	fi

	issue_url="${GIT_PACKAGING_URL_HTTPS}/${project_path}/-/issues/${iid}"
	if (( web )); then
		if ! command -v xdg-open &>/dev/null; then
			die "The web option requires 'xdg-open'"
		fi
		echo "Opening ${issue_url} in your browser."
		xdg-open "${issue_url}"
		return
	fi

	severity="$(gitlab_severity_from_labels "${labels[@]}")"
	severity_color="$(gitlab_severity_color "${severity}")"
	created_at=$(relative_date_unit "${created_at}")
	state_color="$(gitlab_issue_state_color "${state}")"
	state="$(gitlab_issue_state_display "${state}")"
	status="$(gitlab_issue_status_from_labels "${labels[@]}")"
	status_color="$(gitlab_issue_status_color "${status}")"

	scope="$(gitlab_scope_from_labels "${labels[@]}")"
	scope_color="$(gitlab_scope_color "${scope}")"
	scope_label=""
	if [[ ${scope} != unknown ]]; then
		scope_label="${scope_color}${scope}${ALL_OFF} ${GRAY}•${ALL_OFF} "
	fi

	resolution_label=""
	if resolution="$(gitlab_resolution_from_labels "${labels[@]}")"; then
		resolution_color="$(gitlab_resolution_color "${resolution}")"
		resolution_label="${resolution_color}${resolution}${ALL_OFF} ${GRAY}•${ALL_OFF} "
	fi

	confidential_label=""
	if [[ ${confidential} == true ]]; then
		confidential_label="${YELLOW}${PKGCTL_TERM_ICON_CONFIDENTIAL} CONFIDENTIAL${ALL_OFF} ${GRAY}•${ALL_OFF} "
	fi

	printf "%s%s • %s%s%sseverity %s • %s • %s%sopened by %s %s ago%s\n" \
		"${state_color}${state}${ALL_OFF}" "${GRAY}" "${confidential_label}" "${resolution_label}" "${severity_color}" "${severity}${ALL_OFF}${GRAY}" \
		"${status_color}${status}${ALL_OFF}${GRAY}" "${scope_label}" "${GRAY}" "${author}" "${created_at}" "${ALL_OFF}"
	printf "%s %s\n\n" "${BOLD}${title}${ALL_OFF}" "${GRAY}#${iid}${ALL_OFF}"
	printf "%s\n" "${description}" | glow
	printf "\n\n"
	printf "%s%s upvotes • %s downvotes • %s comments%s\n" "${GRAY}" "${upvotes}" "${downvotes}" "${user_notes_count}" "${ALL_OFF}"
	printf "%s %s\n" "${BOLD}Labels:${ALL_OFF}" "$(join_by ", " "${labels[@]}")"
	printf "%s %s\n" "${BOLD}Assignees:${ALL_OFF}" "$(join_by ", " "${assignees[@]}")"
	if [[ ${closed_at} != null ]]; then
		closed_at=$(relative_date_unit "${closed_at}")
		printf "%s %s %s ago\n" "${BOLD}Closed by:${ALL_OFF}" "${closed_by}" "${closed_at}"
	fi

	if (( comments )); then
		printf "\n\n"
		echo "${BOLD}Comments / Notes${ALL_OFF}"
		printf -v spaces '%*s' $(( COLUMNS - 2 )) ''
		printf '%s\n' "${spaces// /─}"
		printf "\n\n"

		status_dir=$(mktemp --tmpdir="${WORKDIR}" --directory pkgctl-gitlab-api.XXXXXXXXXX)
		printf "📡 Querying GitLab issue notes API..." > "${status_dir}/status"
		term_spinner_start "${status_dir}"
		if ! output=$(gitlab_project_issue_notes "${project_path}" "${iid}" "${status_dir}/status" "sort=asc&order_by=created_at"); then
			term_spinner_stop "${status_dir}"
			msg_error "Failed to fetch comments"
			exit 1
		fi
		term_spinner_stop "${status_dir}"

		# pretty print each result
		while read -r result; do
			{ read -r created_at; read -r author; } < <(
				jq --raw-output ".created_at, .author.username" <<< "${result}"
			)
			body=$(jq --raw-output ".body" <<< "${result}")
			created_at=$(relative_date_unit "${created_at}")

			printf "%s commented%s %s ago%s\n" "${author}" "${GRAY}" "${created_at}" "${ALL_OFF}"
			echo "${body}" | glow
			echo
		done < <(jq --compact-output '.[]' <<< "${output}")

		echo "$output" > /tmp/notes.json
	fi

	echo
	printf "%sView this issue on GitLab: %s%s\n" "${GRAY}" "${issue_url}" "${ALL_OFF}"
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later
:

# shellcheck disable=2034
DEVTOOLS_VALID_REPOS=(
	core core-staging core-testing
	extra extra-staging extra-testing
	multilib multilib-staging multilib-testing
	gnome-unstable
	kde-unstable
)

# shellcheck disable=2034
DEVTOOLS_VALID_BUILDREPOS=(
	core-staging core-testing
	extra extra-staging extra-testing
	multilib multilib-staging multilib-testing
	gnome-unstable
	kde-unstable
)

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_AUR_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_AUR_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

set -eo pipefail


pkgctl_aur_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [COMMAND] [OPTIONS]

		Interact with the Arch User Repository (AUR).

		Provides a suite of tools designed for managing and interacting with the Arch
		User Repository (AUR). It simplifies various tasks related to AUR, including
		importing repositories, managing packages, and transitioning packages between
		the official repositories and the AUR.

		COMMANDS
		    drop-from-repo    Drop a package from the official repository to the AUR

		OPTIONS
		    -h, --help        Show this help text

		EXAMPLES
		    $ ${COMMAND} drop-from-repo libfoo
_EOF_
}

pkgctl_aur() {
	if (( $# < 1 )); then
		pkgctl_aur_usage
		exit 0
	fi

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_aur_usage
				exit 0
				;;
			drop-from-repo)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/aur/drop-from-repo.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/aur/drop-from-repo.sh
				pkgctl_aur_drop_from_repo "$@"
				exit 0
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				die "invalid command: %s" "$1"
				;;
		esac
	done
}

#!/hint/bash
#
# This may be included with or without `set -euE`
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_AUTH_LOGIN_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_AUTH_LOGIN_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/config.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/config.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh

set -e


pkgctl_auth_login_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS]

		Interactively authenticate with the GitLab instance.

		The minimum required scopes for the token are: 'api', 'write_repository'.

		The GitLab API token can either be stored in a plaintext file, or
		supplied via the DEVTOOLS_GITLAB_TOKEN environment variable using a
		vault, see pkgctl-auth-login(1) for details.

		OPTIONS
		    -g, --gen-access-token   Open the URL to generate a new personal access token
		    -h, --help               Show this help text

		EXAMPLES
		    $ ${COMMAND}
		    $ ${COMMAND} --gen-access-token
_EOF_
}


pkgctl_auth_login() {
	local token personal_access_token_url
	local GEN_ACESS_TOKEN=0

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_auth_login_usage
				exit 0
				;;
			-g|--gen-access-token)
				GEN_ACESS_TOKEN=1
				shift
				;;
			*)
				die "invalid argument: %s" "$1"
				;;
		esac
	done

	personal_access_token_url="https://${GITLAB_HOST}/-/user_settings/personal_access_tokens?name=pkgctl+token&scopes=api,write_repository"

    cat <<- _EOF_
	Logging into ${BOLD}${GITLAB_HOST}${ALL_OFF}

	Tip: you can generate a Personal Access Token here ${personal_access_token_url}
	The minimum required scopes are 'api' and 'write_repository'.

	If you do not want to store the token in a plaintext file, you can abort
	the following prompt and supply the token via the DEVTOOLS_GITLAB_TOKEN
	environment variable using a vault, see pkgctl-auth-login(1) for details.
_EOF_

	if (( GEN_ACESS_TOKEN )); then
		xdg-open "${personal_access_token_url}" 2>/dev/null
	fi

	# read token from stdin
	read -s -r -p "${GREEN}?${ALL_OFF} ${BOLD}Paste your authentication token:${ALL_OFF} " token
	echo

	if [[ -z ${token} ]]; then
		msg_error "  No token provided"
		exit 1
	fi

	# check if the passed token works
	GITLAB_TOKEN="${token}"
	if ! result=$(gitlab_api_get_user); then
		printf "%s\n" "$result"
		exit 1
	fi

	msg_success "  Logged in as ${BOLD}${result}${ALL_OFF}"
	save_devtools_config
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_ISSUE_COMMENT_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_ISSUE_COMMENT_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh

set -eo pipefail


pkgctl_issue_comment_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [IID]

		This command allows users to add comments to an issue in Arch Linux packaging
		projects. This command is useful for providing feedback, updates, or any
		additional information related to an issue directly within the project's issue
		tracking system.

		By default, the command interacts with the current directory, but users can
		specify a different package base if needed.

		OPTIONS
		    -p, --package PKGBASE  Interact with <pkgbase> instead of the current directory
		    -m, --message MSG      Use the provided message as the comment
		    -e, --edit             Edit the comment using an editor
		    -h, --help             Show this help text

		EXAMPLES
		    $ ${COMMAND} --message "I've attached some logs" 42
		    $ ${COMMAND} --package linux 42
		    $ ${COMMAND} 42
_EOF_
}

pkgctl_issue_comment() {
	if (( $# < 1 )); then
		pkgctl_issue_comment_usage
		exit 0
	fi

	local iid=""
	local pkgbase=""
	local message=""
	local edit=0

	local note

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_issue_comment_usage
				exit 0
				;;
			-m|--message)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				message=$2
				shift 2
				;;
			-p|--package)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				pkgbase=$2
				shift 2
				;;
			-e|--edit)
				edit=1
				shift
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				iid=$1
				shift
				;;
		esac
	done

	if [[ -z ${iid} ]]; then
		die "missing issue iid argument"
	fi

	if [[ -z ${pkgbase} ]]; then
		if ! [[ -f PKGBUILD ]]; then
			die "missing --package option or PKGBUILD in current directory"
		fi
		pkgbase=$(realpath --canonicalize-existing .)
	fi
	pkgbase=$(basename "${pkgbase}")

	# spawn editor
	if (( edit )) || [[ -z ${message} ]]; then
		msgfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-issue-note.XXXXXXXXXX.md)
		printf "%s\n" "${message}" >> "${msgfile}"
		if [[ -n $VISUAL ]]; then
			$VISUAL "${msgfile}" || die
		elif [[ -n $EDITOR ]]; then
			$EDITOR "${msgfile}" || die
		else
			die "No usable editor found (tried \$VISUAL, \$EDITOR)."
		fi
		message=$(< "${msgfile}")
	fi

	# comment on issue
	if ! note=$(gitlab_create_project_issue_note "${pkgbase}" "${iid}" "${message}"); then
		msg_error "Failed to comment on issue ${BOLD}#${iid}${ALL_OFF}"
		exit 1
	fi
	msg_success "Commented on issue ${BOLD}#${iid}${ALL_OFF}"
	echo

	{ read -r created_at; read -r author; } < <(
		jq --raw-output ".created_at, .author.username" <<< "${note}"
	)
	body=$(jq --raw-output ".body" <<< "${note}")
	created_at=$(relative_date_unit "${created_at}")

	printf "%s commented%s %s ago%s\n" "${author}" "${GRAY}" "${created_at}" "${ALL_OFF}"
	echo "${body}" | glow
}

#!/bin/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_SEARCH_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_SEARCH_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh
# shellcheck source=src/lib/valid-search.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/valid-search.sh
# shellcheck source=src/lib/util/term.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/term.sh

source /usr/share/makepkg/util/util.sh
source /usr/share/makepkg/util/message.sh

set -eo pipefail


pkgctl_search_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] QUERY

		Search for an expression across the GitLab packaging group.

		To use a filter, include it in your query. You may use wildcards (*) to
		use glob matching.

		Available filters for the blobs scope: path, extension

		Every usage of the search command must be authenticated. Consult the
		'pkgctl auth' command to authenticate with GitLab or view the
		authentication status.

		SEARCH TIPS
		    Syntax  Description    Example
		    ───────────────────────────────────────
		    "       Exact search   "gem sidekiq"
		    ~       Fuzzy search   J~ Doe
		    |       Or             display | banner
		    +       And            display +banner
		    -       Exclude        display -banner
		    *       Partial        bug error 50*
		    \\       Escape         \\*md
		    #       Issue ID       #23456
		    !       Merge request  !23456

		OPTIONS
		    -h, --help            Show this help text

		FILTER OPTIONS
		    --no-default-filter   Do not apply default filter (like -path:keys/pgp/*.asc)

		OUTPUT OPTIONS
		    --json                Enable printing in JSON; Shorthand for '--format json'
		    -F, --format FORMAT   Controls the formatting of the results; FORMAT is 'pretty',
		                          'plain', or 'json' (default: pretty)
		    -N, --no-line-number  Don't show line numbers when formatting results

		EXAMPLES
		    $ ${COMMAND} linux
		    $ ${COMMAND} --json '"pytest -v" +PYTHONPATH'
_EOF_
}

pkgctl_search_check_option_group_format() {
	local option=$1
	local output_format=$2
	if [[ -n ${output_format} ]]; then
		die "The argument '%s' cannot be used with one or more of the other specified arguments" "${option}"
		exit 1
	fi
	return 0
}

pkgctl_search() {
	if (( $# < 1 )); then
		pkgctl_search_usage
		exit 0
	fi

	# options
	local search
	local output_format=
	local use_default_filter=1
	local line_numbers=1

	# variables
	local bat_style="header,grid"
	local default_filter="-path:keys/pgp/*.asc"
	local output result project_name_lookup project_ids project_id project_name
	local path startline currentline data line

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_search_usage
				exit 0
				;;
			--no-default-filter)
				use_default_filter=0
				shift
				;;
			--json)
				pkgctl_search_check_option_group_format "$1" "${output_format}"
				output_format=json
				shift
				;;
			-F|--format)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				pkgctl_search_check_option_group_format "$1" "${output_format}"
				output_format="${2}"
				if ! in_array "${output_format}" "${valid_search_output_format[@]}"; then
					die "Unknown output format: %s" "${output_format}"
				fi
				shift 2
				;;
			-N|--no-line-number)
				line_numbers=0
				shift
				;;
			--)
				shift
				break
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				break
				;;
		esac
	done

	if (( $# == 0 )); then
		pkgctl_search_usage
		exit 1
	fi

	# assign search parameter
	search="${*}"
	if (( use_default_filter )); then
		search+=" ${default_filter}"
	fi

	# assign default output format
	if [[ -z ${output_format} ]]; then
		output_format=pretty
	fi

	# check for optional dependencies
	if [[ ${output_format} == pretty ]] && ! command -v bat &>/dev/null; then
		warning "Failed to find optional dependency 'bat': falling back to plain output"
		output_format=plain
	fi

	# populate line numbers option
	if (( line_numbers )); then
		bat_style="numbers,${bat_style}"
	fi

	# call the gitlab search API
	status_dir=$(mktemp --tmpdir="${WORKDIR}" --directory pkgctl-gitlab-api.XXXXXXXXXX)
	printf "📡 Querying GitLab search API..." > "${status_dir}/status"
	term_spinner_start "${status_dir}"
	output=$(gitlab_api_search "${search}" "${status_dir}/status")
	term_spinner_stop "${status_dir}"
	msg_success "Querying GitLab search API"

	mapfile -t project_ids < <(
		jq --raw-output '[.[].project_id] | unique[]' <<< "${output}")

	printf "📡 Querying GitLab project names..." > "${status_dir}/status"
	term_spinner_start "${status_dir}"

	# read project_id to name mapping from cache
	declare -A project_name_lookup=()
	while read -r project_id project_name; do
		project_name_lookup[${project_id}]=${project_name}
	done < <(gitlab_lookup_project_names "${status_dir}/status" "${project_ids[@]}")

	term_spinner_stop "${status_dir}"
	msg_success "Querying GitLab project names"

	# output mode JSON
	if [[ ${output_format} == json ]]; then
		jq --from-file <(
			for project_id in $(jq '.[].project_id' <<< "${output}"); do
				project_name=${project_name_lookup[${project_id}]}
				printf 'map(if .project_id == %s then . + {"project_name": "%s"} else . end) | ' \
					"${project_id}" "${project_name}"
			done
			printf .
		) <<< "${output}"
		exit 0
	fi

	# pretty print each result
	while read -r result; do
		# read properties from search result
		mapfile -t data < <(jq --raw-output ".data" <<< "${result}")
		{ read -r project_id; read -r path; read -r startline; } < <(
			jq --raw-output ".project_id, .path, .startline" <<< "${result}"
		)
		project_name=${project_name_lookup[${project_id}]}

		# remove trailing newline for multiline results
		if (( ${#data[@]} > 1 )) && [[ ${data[-1]} == "" ]]; then
			unset "data[${#data[@]}-1]"
		fi

		# output mode plain
		if [[ ${output_format} == plain ]]; then
			printf "%s%s%s\n" "${PURPLE}" "${project_name}/${path}" "${ALL_OFF}"

			currentline=${startline}
			for line in "${data[@]}"; do
				if (( line_numbers )); then
					line="${DARK_GREEN}${currentline}${ALL_OFF}: ${line}"
					currentline=$(( currentline + 1 ))
				fi
				printf "%s\n" "${line}"
			done
			printf "\n"

			continue
		fi

		# prepend empty lines to match startline
		if (( startline > 1 )); then
			mapfile -t data < <(
				printf '%.0s\n' $(seq 1 "$(( startline - 1 ))")
				printf "%s\n" "${data[@]}"
			)
		fi

		bat \
			--file-name="${project_name}/${path}" \
			--line-range "${startline}:" \
			--paging=never \
			--force-colorization \
			--style "${bat_style}" \
			--map-syntax "PKGBUILD:Bourne Again Shell (bash)" \
			--map-syntax ".SRCINFO:INI" \
			--map-syntax "*install:Bourne Again Shell (bash)" \
			--map-syntax "*sysusers*:Bourne Again Shell (bash)" \
			--map-syntax "*tmpfiles*:Bourne Again Shell (bash)" \
			--map-syntax "*.hook:INI" \
			<(printf "%s\n" "${data[@]}")
	done < <(jq --compact-output '.[]' <<< "${output}")
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_RELEASE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_RELEASE_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/db/update.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/db/update.sh
# shellcheck source=src/lib/util/pacman.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/pacman.sh
# shellcheck source=src/lib/valid-repos.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/valid-repos.sh

source /usr/share/makepkg/util/util.sh

set -e


pkgctl_release_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [PATH]...

		Release step to commit, tag and upload build artifacts

		Modified version controlled files will first be staged for commit,
		afterwards a Git tag matching the pkgver will be created and finally
		all build artifacts will be uploaded.

		By default the target pacman repository will be auto-detected by querying
		the repo it is currently released in. When initially adding a new package
		to the repositories, the target repo must be specified manually.

		OPTIONS
		    -m, --message MSG   Use the given <msg> as the commit message
		    -r, --repo REPO     Specify target repository for new packages not in any official repo
		    -s, --staging       Release to the staging counterpart of the auto-detected repo
		    -t, --testing       Release to the testing counterpart of the auto-detected repo
		    -u, --db-update     Automatically update the pacman database after uploading
		    -h, --help          Show this help text

		EXAMPLES
		    $ ${COMMAND}
		    $ ${COMMAND} --staging --message 'libyay 0.42 rebuild' libfoo libbar
		    $ ${COMMAND} --repo extra --db-update new-package
_EOF_
}

pkgctl_release_check_option_group() {
	local option=$1
	local repo=$2
	local testing=$3
	local staging=$4
	if [[ -n "${repo}" ]] || (( testing )) || (( staging )); then
		die "The argument '%s' cannot be used with one or more of the other specified arguments" "${option}"
		exit 1
	fi
	return 0
}

pkgctl_release() {
	if (( $# < 1 )) && [[ ! -f PKGBUILD ]]; then
		pkgctl_release_usage
		exit 1
	fi

	local MESSAGE=""
	local PKGBASES=()
	local REPO=""
	local TESTING=0
	local STAGING=0
	local DB_UPDATE=0

	local path pkgbase pkgnames repo repos

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_release_usage
				exit 0
				;;
			-m|--message)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				MESSAGE=$2
				shift 2
				;;
			-r|--repo)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				pkgctl_release_check_option_group '--repo' "${REPO}" "${TESTING}" "${STAGING}"
				REPO=$2
				shift 2
				;;
			-s|--staging)
				pkgctl_release_check_option_group '--staging' "${REPO}" "${TESTING}" "${STAGING}"
				STAGING=1
				shift
				;;
			-t|--testing)
				pkgctl_release_check_option_group '--testing' "${REPO}" "${TESTING}" "${STAGING}"
				TESTING=1
				shift
				;;
			-u|--db-update)
				DB_UPDATE=1
				shift
				;;
			-*)
				die "invalid option: %s" "$1"
				;;
			*)
				PKGBASES+=("$@")
				break
				;;
		esac
	done

	# Resolve package from current working directory
	if (( 0 == ${#PKGBASES[@]} )); then
		PKGBASES=("$PWD")
	fi

	# Update pacman cache for auto-detection
	if [[ -z ${REPO} ]]; then
		update_pacman_repo_cache multilib
	# Check valid repos if not resolved dynamically
	elif ! in_array "${REPO}" "${DEVTOOLS_VALID_REPOS[@]}"; then
		die "Invalid repository target: %s" "${REPO}"
	fi

	for path in "${PKGBASES[@]}"; do
		pushd "${path}" >/dev/null
		pkgbase=$(basename "${path}")

		# auto-detect target repository
		if ! repo=$(get_pacman_repo_from_pkgbuild PKGBUILD); then
			die 'Failed to query pacman repo'
		fi

		# fail if an existing package specifies --repo
		if [[ -n "${repo}" ]] && [[ -n ${REPO} ]]; then
			# allow unstable to use --repo
			if [[ ${REPO} == *unstable ]]; then
				repo=${REPO}
			else
				die 'Using --repo for packages that exist in official repositories is disallowed'
			fi
		fi

		# fail if a new package does not specify --repo
		if [[ -z "${repo}" ]]; then
			if [[ -z ${REPO} ]]; then
				die 'Specify --repo for packages that do not yet exist in official repositories'
			fi
			repo=${REPO}
		fi

		# output a warning if .nvchecker.toml does not exists
		if [[ ! -f ".nvchecker.toml" ]]; then
			warning "Nvchecker integration is not set, run 'pkgctl version setup --help' to see how to automate the creation of the '.nvchecker.toml' configuration file"
		fi

		if (( TESTING )); then
			repo="${repo}-testing"
		elif (( STAGING )); then
			repo="${repo}-staging"
		elif [[ $repo == core ]]; then
			repo="${repo}-testing"
		fi

		msg "Releasing ${pkgbase} to ${repo}"
		commitpkg "${repo}" "${MESSAGE}"

		unset repo
		popd >/dev/null
	done

	if (( DB_UPDATE )); then
		# shellcheck disable=2119
		pkgctl_db_update
	fi
}

#!/hint/bash
#
# This may be included with or without `set -euE`
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_AUTH_STATUS_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_AUTH_STATUS_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh

set -e


pkgctl_auth_status_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS]

		Verifies and displays information about your authentication state of
		services like the GitLab instance and reports issues if any.

		OPTIONS
		    -t, --show-token   Display the auth token
		    -h, --help         Show this help text

		EXAMPLES
		    $ ${COMMAND}
		    $ ${COMMAND} --show-token
_EOF_
}

pkgctl_auth_status() {
	local SHOW_TOKEN=0
	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_auth_status_usage
				exit 0
				;;
			-t|--show-token)
				SHOW_TOKEN=1
				shift
				;;
			*)
				die "invalid argument: %s" "$1"
				;;
		esac
	done

	printf "%s\n" "${BOLD}${GITLAB_HOST}${ALL_OFF}"
	# shellcheck disable=2119
	if ! username=$(gitlab_api_get_user); then
		printf "%s\n" "${username}"
		exit 1
	fi

	msg_success "  Logged in as ${BOLD}${username}${ALL_OFF}"
	if (( SHOW_TOKEN )); then
		msg_success "  Token: ${GITLAB_TOKEN}"
	else
		msg_success "  Token: **************************"
	fi
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_ISSUE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_ISSUE_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

set -eo pipefail


pkgctl_issue_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [COMMAND] [OPTIONS]

		Work with GitLab packaging issues.

		COMMANDS
		    close     Close an issue
		    comment   Comment on an issue
		    create    Create a new issue
		    edit      Edit and modify an issue
		    list      List project or group issues
		    move      Move an issue to another project
		    reopen    Reopen a closed issue
		    view      Display information about an issue

		OPTIONS
		    -h, --help    Show this help text

		EXAMPLES
		    $ ${COMMAND} list libfoo libbar
		    $ ${COMMAND} view 4
_EOF_
}

pkgctl_issue() {
	if (( $# < 1 )); then
		pkgctl_issue_usage
		exit 0
	fi

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_issue_usage
				exit 0
				;;
			close)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/issue/close.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/issue/close.sh
				pkgctl_issue_close "$@"
				exit 0
				;;
			create)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/issue/create.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/issue/create.sh
				pkgctl_issue_create "$@"
				exit 0
				;;
			edit|update)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/issue/edit.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/issue/edit.sh
				pkgctl_issue_edit "$@"
				exit 0
				;;
			list)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/issue/list.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/issue/list.sh
				pkgctl_issue_list "$@"
				exit 0
				;;
			comment|note)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/issue/comment.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/issue/comment.sh
				pkgctl_issue_comment "$@"
				exit 0
				;;
			move)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/issue/move.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/issue/move.sh
				pkgctl_issue_move "$@"
				exit 0
				;;
			reopen)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/issue/reopen.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/issue/reopen.sh
				pkgctl_issue_reopen "$@"
				exit 0
				;;
			view)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/issue/view.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/issue/view.sh
				pkgctl_issue_view "$@"
				exit 0
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				die "invalid command: %s" "$1"
				;;
		esac
	done
}

#!/bin/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
[[ -z ${DEVTOOLS_INCLUDE_LICENSE_SETUP_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_LICENSE_SETUP_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/license/check.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/license/check.sh

source /usr/share/makepkg/util/message.sh

set -eo pipefail
shopt -s nullglob


pkgctl_license_setup_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [PKGBASE]...

		Automate the creation of a basic REUSE configuration by analyzing the
		license array specified in the PKGBUILD file of a package.

		If no PKGBASE is specified, the command defaults to using the current
		working directory.

		OPTIONS
		    -f, --force   Overwrite existing REUSE configuration
		    -h, --help    Show this help text
		    --no-check    Do not run license check after setup

		EXAMPLES
		    $ ${COMMAND} neovim vim
_EOF_
}

pkgctl_license_setup() {
	local pkgbases=()
	local force=0
	local run_check=1

	local path exit_code
	local checks=()

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_license_setup_usage
				exit 0
				;;
			-f|--force)
				force=1
				shift
				;;
			--no-check)
				run_check=0
				shift
				;;
			--)
				shift
				break
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				pkgbases=("$@")
				break
				;;
		esac
	done

	if ! command -v reuse &>/dev/null; then
		die "The \"$_DEVTOOLS_COMMAND\" command requires the 'reuse' CLI tool"
	fi

	# Check if used without pkgbases in a packaging directory
	if (( ${#pkgbases[@]} == 0 )); then
		if [[ -f PKGBUILD ]]; then
			pkgbases=(".")
		else
			pkgctl_license_setup_usage
			exit 1
		fi
	fi

	exit_code=0
	for path in "${pkgbases[@]}"; do
		# skip paths that are not directories
		if [[ ! -d "${path}" ]]; then
			continue
		fi

		pushd "${path}" >/dev/null
		if license_setup "${path}" "${force}"; then
			checks+=("${path}")
		else
			exit_code=1
		fi
		popd >/dev/null
	done

	# run checks on the setup targets
	if (( run_check )) && (( ${#checks[@]} >= 1 )); then
		echo
		echo 📡 Running checks...
		pkgctl_license_check "${checks[@]}" || true
	fi

	return "$exit_code"
}

license_setup() {
	local path=$1
	local force=$2
	local pkgbase pkgname license

	if [[ ! -f PKGBUILD ]]; then
		msg_error "${BOLD}${path}:${ALL_OFF} no PKGBUILD found"
		return 1
	fi

	# shellcheck source=contrib/makepkg/PKGBUILD.proto
	if ! . ./PKGBUILD; then
		msg_error "${BOLD}${path}:${ALL_OFF} failed to source PKGBUILD"
		return 1
	fi
	pkgbase=${pkgbase:-$pkgname}

	# setup LICENSE file
	if ! license_file_setup "${pkgbase}" "${force}"; then
		return 1
	fi

	# setup REUSE.toml
	if ! reuse_setup "${pkgbase}" "${force}" "${license[@]}"; then
		return 1
	fi
}

is_valid_license() {
	local license_name=$1
	local supported_licenses
	supported_licenses=$(reuse supported-licenses | awk '{print $1}')
	if grep --quiet "^${license_name}$" <<< "$supported_licenses"; then
		return 0
	else
		return 1
	fi
}

license_file_setup() {
	local pkgbase=$1
	local force=$2

	local license_text
	license_text=$(< "${_DEVTOOLS_LIBRARY_DIR}"/data/LICENSE)

	# Write LICENSE file, or check if it mismatches
	if (( force )) || [[ ! -f LICENSE ]]; then
		printf "%s\n" "${license_text}" > LICENSE
		msg_success "${BOLD}${pkgbase}:${ALL_OFF} successfully configured LICENSE"
	elif [[ -f LICENSE ]]; then
		# if there is a license file, check whether it has the text we expect
		existing_license="$(< LICENSE)"
		if [[ "${license_text}" != "${existing_license}" ]]; then
			msg_error "${BOLD}${pkgbase}:${ALL_OFF} existing LICENSE file doesn't have expected content, use --force to overwrite"
			return 1
		fi
	fi

	# make sure the LICENSE file is found by REUSE
	mkdir --parents LICENSES
	ln --symbolic --force ../LICENSE LICENSES/0BSD.txt
}

reuse_default_annotations() {
	cat << EOF
version = 1

[[annotations]]
path = [
    "PKGBUILD",
    "README.md",
    "keys/**",
    ".SRCINFO",
    ".gitignore",
    ".nvchecker.toml",
    "*.install",
    "*.sysusers",
    "*sysusers.conf",
    "*.tmpfiles",
    "*tmpfiles.conf",
    "*.logrotate",
    "*.pam",
    "*.service",
    "*.socket",
    "*.timer",
    "*.desktop",
    "*.hook",
]
SPDX-FileCopyrightText = "Arch Linux contributors"
SPDX-License-Identifier = "0BSD"
EOF
}

reuse_setup() {
	local pkgbase=$1
	local force=$2
	shift 2
	local license=("$@")

	# Check if REUSE.toml already exists
	if (( ! force )) && [[ -f REUSE.toml ]]; then
		msg_error "${BOLD}${pkgbase}:${ALL_OFF} REUSE.toml already exists, use --force to overwrite"
		return 1
	fi

	reuse_default_annotations > REUSE.toml

	local warning_occurred=0
	local patches=(*.patch)
	# If there are patches and there's only a single well-known license listed in the package,
	# we can generate the annotations for the patches, otherwise we will fail to do so and warn
	# the user.
	if (( ${#patches} )); then
		# If there are multiple licenses, we can't make a good guess about which license the
		# patches should have. In case the first element contains a space, we are dealing with
		# a complex SPDX license identifier.
		if (( ${#license[@]} > 1 )) || [[ ${license[0]} =~ [[:space:]] ]]; then
			msg_warn "${BOLD}${pkgbase}:${ALL_OFF} .patch files were found but couldn't automatically guess a suitable license because PKGBUILD has multiple licenses"
			patch_annotations "${pkgbase}" "TODO-Choose-a-license" "${patches[@]}" >> REUSE.toml
			warning_occurred=1
		elif ! is_valid_license "${license[0]}"; then
			msg_warn "${BOLD}${pkgbase}:${ALL_OFF} .patch files were found but couldn't automatically guess a suitable license because the PKGBUILD license '${license[0]}' is not a recognized SPDX license"
			patch_annotations "${pkgbase}" "TODO-Choose-a-license" "${patches[@]}" >> REUSE.toml
			warning_occurred=1
		else
			patch_annotations "${pkgbase}" "${license[0]}" "${patches[@]}" >> REUSE.toml
		fi
	fi

	if (( warning_occurred )); then
		msg_warn "${BOLD}${pkgbase}:${ALL_OFF} configured REUSE but a warning occurred, manually edit and fix REUSE.toml"
		return 1
	else
		reuse download --all
		msg_success "${BOLD}${pkgbase}:${ALL_OFF} successfully configured REUSE.toml"
	fi
}

patch_annotations() {
	local pkgbase=$1
	local license_identifier=$2
	shift 2
	local patches=("$@")

	local annotations
	annotations+="\n[[annotations]]\n"
	annotations+="path = [\n"
	for patch in "${patches[@]}"; do
		annotations+="    \"$(basename "${patch}")\",\n"
	done
	annotations+="]\n"
	annotations+="SPDX-FileCopyrightText = \"${pkgbase} contributors\"\n"
	annotations+="SPDX-License-Identifier = \"${license_identifier}\""
	echo -e "${annotations}"
}

#!/bin/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
[[ -z ${DEVTOOLS_INCLUDE_VERSION_SETUP_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_VERSION_SETUP_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/version/check.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/version/check.sh

source /usr/share/makepkg/util/message.sh
source /usr/share/makepkg/util/source.sh

set -eo pipefail


pkgctl_version_setup_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [PKGBASE]...

		Automate the creation of a basic nvchecker configuration file by
		analyzing the source array specified in the PKGBUILD file of a package.

		If no PKGBASE is specified, the command defaults to using the current
		working directory.

		OPTIONS
		    -f, --force            Overwrite existing nvchecker configuration
		    --prefer-platform-api  Prefer platform specific GitHub/GitLab API for complex cases
		    --url URL              Derive check target from URL instead of source array
		    --no-check             Do not run version check after setup
		    --no-upstream          Setup a blank config for packages without upstream sources
		    -h, --help             Show this help text

		EXAMPLES
		    $ ${COMMAND} neovim vim
_EOF_
}

pkgctl_version_setup() {
	local pkgbases=()
	local override_url=
	local run_check=1
	local force=0
	local prefer_platform_api=0
	local no_upstream=0

	local path ret
	local checks=()

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_version_setup_usage
				exit 0
				;;
			-f|--force)
				force=1
				shift
				;;
			--prefer-platform-api)
				prefer_platform_api=1
				shift
				;;
			--url)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				override_url=$2
				shift 2
				;;
			--no-check)
				run_check=0
				shift
				;;
			--no-upstream)
				no_upstream=1
				shift
				;;
			--)
				shift
				break
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				pkgbases=("$@")
				break
				;;
		esac
	done

	# Check if used without pkgbases in a packaging directory
	if (( ${#pkgbases[@]} == 0 )); then
		if [[ -f PKGBUILD ]]; then
			pkgbases=(".")
		else
			pkgctl_version_setup_usage
			exit 1
		fi
	fi

	ret=0
	for path in "${pkgbases[@]}"; do
		# skip paths that are not directories
		if [[ ! -d "${path}" ]]; then
			continue
		fi

		pushd "${path}" >/dev/null
		if nvchecker_setup "${path}" "${force}" "${prefer_platform_api}" "${override_url}" "${no_upstream}"; then
			checks+=("${path}")
		else
			ret=1
		fi
		popd >/dev/null
	done

	# run checks on the setup targets
	if (( run_check )) && (( ${#checks[@]} >= 1 )); then
		echo
		pkgctl_version_check --verbose "${checks[@]}" || true
	fi

	return $ret
}

nvchecker_setup() {
	local path=$1
	local force=$2
	local prefer_platform_api=$3
	local override_url=$4
	local no_upstream=$5
	local pkgbase pkgname source source_url proto domain url_parts section body

	if [[ ! -f PKGBUILD ]]; then
		msg_error "${BOLD}${path}:${ALL_OFF} no PKGBUILD found"
		return 1
	fi

	unset body pkgbase pkgname source url
	# shellcheck source=contrib/makepkg/PKGBUILD.proto
	if ! . ./PKGBUILD; then
		msg_error "${BOLD}${path}:${ALL_OFF} failed to source PKGBUILD"
		return 1
	fi
	pkgbase=${pkgbase:-$pkgname}

	# try to guess from url as last try
	if [[ -n ${url} ]]; then
		source+=("${url}")
	fi

	# handle overwrite of existing config
	if [[ -f .nvchecker.toml ]] && (( ! force )); then
		msg_warn "${BOLD}${pkgbase}:${ALL_OFF} nvchecker already configured"
		return 1
	fi

	# override the source array with a passed URL
	if [[ -n ${override_url} ]]; then
		source=("${override_url}")
	fi

	# skip empty source array
	if (( ${#source[@]} == 0 )) && (( ! no_upstream )); then
		msg_error "${BOLD}${pkgbase}:${ALL_OFF} PKGBUILD has no source array"
		return 1
	fi

	for source_url in "${source[@]}"; do
		# Strips out filename::http for example
		source_url=$(get_url "${source_url}")
		# discard query fragments
		source_url=${source_url%\?*}
		source_url=${source_url%#*}

		# skip patches
		if [[ ${source_url} == *.patch ]]; then
			continue
		fi
		# skip signatures
		if [[ ${source_url} == *.asc ]] || [[ ${source_url} == *.sig ]]; then
			continue
		fi
		# skip local files
		if [[ ${source_url} != *://* ]]; then
			continue
		fi

		# split URL segments while avoiding empty element after protocol and newline at the end
		mapfile -td / url_parts <<< "${source_url/:\/\//\/}/"
		unset "url_parts[-1]"

		# extract protocol and domain to select the configuration type
		proto=${url_parts[0]}
		domain=${url_parts[1]}

		case "${domain}" in
			gitlab.*)
				if (( prefer_platform_api )); then
					body=$(nvchecker_setup_gitlab "${url_parts[@]}")
				else
					body=$(nvchecker_setup_git "${url_parts[@]}")
				fi
				break
				;;
			github.com)
				if (( prefer_platform_api )); then
					body=$(nvchecker_setup_github "${url_parts[@]}")
				else
					body=$(nvchecker_setup_git "${url_parts[@]}")
				fi
				break
				;;
			codeberg.org)
				body=$(nvchecker_setup_git "${url_parts[@]}")
				break
				;;
			pypi.org|pypi.io|files.pythonhosted.org)
				body=$(nvchecker_setup_pypi "${url_parts[@]}")
				break
				;;
			hackage.haskell.org)
				body=$(nvchecker_setup_hackage "${url_parts[@]}")
				break
				;;
			registry.npmjs.org|npmjs.com|www.npmjs.com)
				body=$(nvchecker_setup_npm "${url_parts[@]}")
				break
				;;
			rubygems.org)
				body=$(nvchecker_setup_rubygems "${url_parts[@]}")
				break
				;;
			*.cpan.org|*.mcpan.org|*.metacpan.org)
				body=$(nvchecker_setup_cpan "${url_parts[@]}")
				break
				;;
			crates.io|*.crates.io)
				body=$(nvchecker_setup_crates_io "${url_parts[@]}")
				break
				;;
			*)
				if [[ ${proto} == git ]] || [[ ${proto} == git+https ]]; then
					body=$(nvchecker_setup_git "${url_parts[@]}")
				fi
				;;
		esac
	done

	if (( no_upstream )); then
		body='source = "manual"'
	fi

	if [[ -z "${body}" ]]; then
		msg_error "${BOLD}${pkgbase}:${ALL_OFF} unable to automatically setup nvchecker"
		return 1
	fi

	# escape the section if it contains toml subsection chars
	section="${pkgbase}"
	if [[ ${section} == *.* ]] || [[ ${section} == *+* ]]; then
		section="\"${section}\""
	fi

	msg_success "${BOLD}${pkgbase}:${ALL_OFF} successfully configured nvchecker"
	cat > .nvchecker.toml << EOF
[${section}]
${body}
EOF
}

get_git_url_from_parts() {
	local url_parts=("$@")
	local proto=${url_parts[0]#*+}
	local domain=${url_parts[1]}
	local url
	url="${proto}://$(join_by / "${url_parts[@]:1}")"

	case "${domain}" in
		gitlab.*)
			url=${url%/-/*/*}
			[[ ${url} != *.git ]] && url+=.git
			;;
		github.com|codeberg.org)
			url="${proto}://$(join_by / "${url_parts[@]:1:3}")"
			[[ ${url} != *.git ]] && url+=.git
			;;
	esac

	printf '%s' "${url}"
}

# PyPI
#
# As Arch python packages don't necessarily match the pypi name, when the
# provided source url comes from pypi.io or pypi.org try to extract the package
# name from the (predictable) tarball download url for example:
#
# https://pypi.io/packages/source/p/pyflakes/pyflakes-3.1.0.tar.gz
# https://pypi.io/packages/source/p/pyflakes
# https://pypi.org/packages/source/b/bleach
# https://files.pythonhosted.org/packages/source/p/pyflakes
# https://pypi.org/project/SQLAlchemy/
nvchecker_setup_pypi() {
	local url_parts=("$@")
	local pypi

	if [[ ${url_parts[2]} == packages ]]; then
		pypi=${url_parts[5]}
	elif [[ ${url_parts[2]} == project ]]; then
		pypi=${url_parts[3]}
	else
		return 1
	fi

	cat << EOF
source = "pypi"
pypi = "${pypi}"
EOF
}

# Git
#
# Set up a generic Git source, while removing the proto specific part from makepkg
#
# git+https://github.com/prometheus/prometheus.git
# https://git.foobar.com/some/path/group/project.git
# https://gitlab.com/sub/group/project/-/archive/8.0.0/packages-8.0.0.tar.gz
nvchecker_setup_git() {
	local url_parts=("$@")
	local url
	url=$(get_git_url_from_parts "${url_parts[@]}")

	cat << EOF
source = "git"
git = "${url}"
EOF

	# best effort check if the tags are prefixed with v
	if git_tags_have_version_prefix "${url}"; then
		echo 'prefix = "v"'
	fi
}

git_tags_have_version_prefix() {
	local url=$1
	# best effort check if the tags are prefixed with v
	if ! grep --max-count=1 --quiet --extended-regex 'refs/tags/v[0-9]+[\.0-9]*$' \
		<(GIT_TERMINAL_PROMPT=0 git ls-remote --quiet --tags "${url}" 2>/dev/null); then
		return 1
	fi
	return 0
}

# Github
#
# We want to know the $org/$project name from the url
#
# https://github.com/prometheus/prometheus/archive/v2.49.1.tar.gz
nvchecker_setup_github() {
	local url_parts=("$@")
	local url project
	if ! url=$(get_git_url_from_parts "${url_parts[@]}"); then
		return 1
	fi
	project=${url#*://*/}
	project=${project%.git}

	cat << EOF
source = "github"
github = "${project}"
use_max_tag = true
EOF

	# best effort check if the tags are prefixed with v
	if git_tags_have_version_prefix "${url}"; then
		echo 'prefix = "v"'
	fi
}

# GitLab
#
# We want to know the $org/$project name from the url
#
# git+https://gitlab.com/inkscape/inkscape.git#tag=091e20ef0f204eb40ecde54436e1ef934a03d894
nvchecker_setup_gitlab() {
	local url_parts=("$@")
	local url project host
	if ! url=$(get_git_url_from_parts "${url_parts[@]}"); then
		return 1
	fi
	project=${url#*://*/}
	project=${project%.git}
	cat << EOF
source = "gitlab"
gitlab = "${project}"
EOF

	host=${url#*://}
	host=${host%%/*}
	if [[ ${host} != gitlab.com ]]; then
		echo "host = \"${host}\""
	fi

	echo "use_max_tag = true"

	# best effort check if the tags are prefixed with v
	if git_tags_have_version_prefix "${url}"; then
		echo 'prefix = "v"'
	fi
}

# Hackage
#
# We want to know the project name
#
# https://hackage.haskell.org/package/xmonad
# https://hackage.haskell.org/package/xmonad-0.18.0/xmonad-0.18.0.tar.gz
# https://hackage.haskell.org/packages/archive/digits/0.3.1/digits-0.3.1.tar.gz
nvchecker_setup_hackage() {
	local url_parts=("$@")
	local hackage

	if [[ ${url_parts[2]} == packages ]]; then
		hackage=${url_parts[4]}
	elif [[ ${url_parts[2]} == package ]] && (( ${#url_parts[@]} == 4 )); then
		hackage=${url_parts[3]}
	elif [[ ${url_parts[2]} == package ]] && (( ${#url_parts[@]} >= 5 )); then
		hackage=${url_parts[3]%-*}
	else
		return 1
	fi

	cat << EOF
source = "hackage"
hackage = "${hackage}"
EOF
}

# NPM
#
# We want to know the project name
#
# https://registry.npmjs.org/eslint_d/-/eslint_d-12.1.0.tgz
# https://www.npmjs.com/package/node-gyp
nvchecker_setup_npm() {
	local url_parts=("$@")
	local npm

	if [[ ${url_parts[1]} == registry.npmjs.org ]]; then
		npm=${url_parts[2]}
	elif [[ ${url_parts[2]} == package ]] && (( ${#url_parts[@]} == 4 )); then
		npm=${url_parts[3]}
	else
		return 1
	fi

	cat << EOF
source = "npm"
npm = "${npm}"
EOF
}

# RubyGems
#
# We want to know the project name
#
# https://rubygems.org/downloads/polyglot-0.3.5.gem
# https://rubygems.org/gems/diff-lcs
nvchecker_setup_rubygems() {
	local url_parts=("$@")
	local gem

	if [[ ${url_parts[2]} == downloads ]]; then
		gem=${url_parts[-1]%-*}
	elif [[ ${url_parts[2]} == gems ]]; then
		gem=${url_parts[3]}
	else
		return 1
	fi

	cat << EOF
source = "gems"
gems = "${gem}"
EOF
}

# CPAN
#
# We want to know the project name
#
# source = https://search.cpan.org/CPAN/authors/id/C/CO/COSIMO/Locale-PO-1.2.3.tar.gz
nvchecker_setup_cpan() {
	local url_parts=("$@")
	local cpan=${url_parts[-1]}
	cpan=${cpan%-*}

	cat << EOF
source = "cpan"
cpan = "${cpan}"
EOF
}

# crates.io
#
# We want to know the crate name
#
# https://crates.io/api/v1/crates/${pkgname}/${pkgver}/download
# https://static.crates.io/crates/${pkgname}/$pkgname-$pkgver.crate
# https://crates.io/crates/git-smash
nvchecker_setup_crates_io() {
	local url_parts=("$@")
	local crate

	if [[ ${url_parts[2]} == crates ]]; then
		crate=${url_parts[3]}
	elif [[ ${url_parts[4]} == crates ]]; then
		crate=${url_parts[5]}
	else
		return 1
	fi


	for i in "${!url_parts[@]}"; do
		if [[ ${url_parts[i]} == crates ]]; then
			crate=${url_parts[(( i + 1 ))]}
		fi
	done

	cat << EOF
source = "cratesio"
cratesio = "${crate}"
EOF
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_API_GITLAB_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_API_GITLAB_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/cache.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/cache.sh
# shellcheck source=src/lib/config.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/config.sh
# shellcheck source=src/lib/valid-issue.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/valid-issue.sh

set -e

graphql_api_call() {
	local outfile=$1
	local request=$2
	local node_type=$3
	local data=$4
	local hasNextPage cursor

	# empty token
	if [[ -z "${GITLAB_TOKEN}" ]]; then
		msg_error "  api call failed: No token provided"
		return 1
	fi

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	api_workdir=$(mktemp --tmpdir="${WORKDIR}" --directory pkgctl-gitlab-api.XXXXXXXXXX)

	# normalize graphql data and prepare query
	data="${data//\"/\\\"}"
	data='{
		"query": "'"${data}"'"
	}'
	data="${data//$'\t'/ }"
	data="${data//$'\n'/}"

	cursor=""
	hasNextPage=true
	while [[ ${hasNextPage} == true ]]; do
		data=$(sed -E 's|after: \\"[a-zA-Z0-9]*\\"|after: \\"'"${cursor}"'\\"|' <<< "${data}")
		result="${api_workdir}/result.${cursor}"

		if ! curl --request "${request}" \
				--url "https://${GITLAB_HOST}/api/graphql" \
				--header "Authorization: Bearer ${GITLAB_TOKEN}" \
				--header "Content-Type: application/json" \
				--data "${data}" \
				--output "${result}" \
				--silent; then
			msg_error "  api call failed: $(cat "${outfile}")"
			return 1
		fi

		hasNextPage=$(jq --raw-output ".data | .${node_type} | .pageInfo | .hasNextPage" < "${result}")
		cursor=$(jq --raw-output ".data | .${node_type} | .pageInfo | .endCursor" < "${result}")

		cp "${result}" "${api_workdir}/tmp"
		jq ".data.${node_type}.nodes" "${api_workdir}/tmp" > "${result}"
	done

	jq --slurp add "${api_workdir}"/result.* > "${outfile}"
	return 0
}

gitlab_api_call() {
	local outfile=$1
	local request=$2
	local endpoint=$3
	local data=${4:-}

	# empty token
	if [[ -z "${GITLAB_TOKEN}" ]]; then
		msg_error "  api call failed: No token provided"
		return 1
	fi

	if ! curl --request "${request}" \
			--url "https://${GITLAB_HOST}/api/v4/${endpoint}" \
			--header "PRIVATE-TOKEN: ${GITLAB_TOKEN}" \
			--header "Content-Type: application/json" \
			--data "${data}" \
			--output "${outfile}" \
			--silent; then
		msg_error "  api call failed: $(cat "${outfile}")"
		return 1
	fi

	if ! gitlab_check_api_errors "${outfile}"; then
		return 1
	fi

	return 0
}

gitlab_api_call_paged() {
	local outfile=$1
	local status_file=$2
	local request=$3
	local endpoint=$4
	local data=${5:-}
	local result header

	# empty token
	if [[ -z "${GITLAB_TOKEN}" ]]; then
		msg_error "  api call failed: No token provided"
		return 1
	fi

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	api_workdir=$(mktemp --tmpdir="${WORKDIR}" --directory pkgctl-gitlab-api.XXXXXXXXXX)
	tmp_file=$(mktemp --tmpdir="${api_workdir}" spinner.tmp.XXXXXXXXXX)

	local next_page=1
	local total_pages=1
	local known_total_pages=1
	local percentage=100

	while [[ -n "${next_page}" ]]; do
		percentage=$(( 100 * next_page / total_pages ))
		printf "📡 Querying GitLab: %s/%s [%s] %%spinner%%" \
			"${BOLD}${next_page}" "${known_total_pages}" "${percentage}%${ALL_OFF}"  \
			> "${tmp_file}"
		mv "${tmp_file}" "${status_file}"

		result="${api_workdir}/result.${next_page}"
		header="${api_workdir}/header"
		if ! curl --request "${request}" \
				--get \
				--url "https://${GITLAB_HOST}/api/v4/${endpoint}&per_page=100&page=${next_page}" \
				--header "PRIVATE-TOKEN: ${GITLAB_TOKEN}" \
				--header "Content-Type: application/json" \
				--data-urlencode "${data}" \
				--dump-header "${header}" \
				--output "${result}" \
				--silent; then
			msg_error "  api call failed: $(cat "${result}")"
			return 1
		fi

		if ! gitlab_check_api_errors "${result}"; then
			return 1
		fi

		next_page=$(grep "x-next-page" "${header}" | tr -d '\r' | awk '{ print $2 }')
		total_pages=$(grep "x-total-pages" "${header}" | tr -d '\r' | awk '{ print $2 }')
		# The api is not guaranteed to return x-total-pages for larger query results
		# https://gitlab.com/gitlab-org/gitlab-foss/-/merge_requests/23931
		# https://gitlab.com/gitlab-org/gitlab/-/issues/436373
		if (( total_pages == 0 )); then
			total_pages=${next_page}
			known_total_pages="?"
		else
			known_total_pages=${total_pages}
		fi
	done

	jq --slurp add "${api_workdir}"/result.* > "${outfile}"
	return 0
}

gitlab_check_api_errors() {
	local file=$1
	local error

	# search API only returns an array, no errors
	if [[ $(jq --raw-output 'type' < "${file}") == "array" ]]; then
		return 0
	fi

	# check for general purpose api error
	if error=$(jq --raw-output --exit-status '.error' < "${file}"); then
		msg_error "  api call failed: ${error}"
		return 1
	fi

	# check for api specific error messages
	if ! jq --raw-output --exit-status '.id' < "${file}" >/dev/null; then
		if jq --raw-output --exit-status '.message | keys[]' < "${file}" &>/dev/null; then
			while read -r error; do
				msg_error "  api call failed: ${error}"
			done < <(jq --raw-output --exit-status '.message|to_entries|map("\(.key) \(.value[])")[]' < "${file}")
		elif error=$(jq --raw-output --exit-status '.message' < "${file}"); then
			msg_error "  api call failed: ${error}"
		fi
		return 1
	fi
	return 0
}

graphql_check_api_errors() {
	local file=$1
	local error

	# early exit if we do not have errors
	if ! jq --raw-output --exit-status '.errors[]' < "${file}" &>/dev/null; then
		return 0
	fi

	# check for api specific error messages
	while read -r error; do
		msg_error "  api call failed: ${error}"
	done < <(jq --raw-output --exit-status '.errors[].message' < "${file}")
	return 1
}

gitlab_api_get_user() {
	local outfile username

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	outfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-gitlab-api.XXXXXXXXXX)

	# query user details
	if ! gitlab_api_call "${outfile}" GET "user/"; then
		msg_warn "  Invalid token provided?"
		exit 1
	fi

	# extract username from details
	if ! username=$(jq --raw-output --exit-status '.username' < "${outfile}"); then
		msg_error "  failed to query username: $(cat "${outfile}")"
		return 1
	fi

	printf "%s" "${username}"
	return 0
}

gitlab_api_get_project_name_mapping() {
	local query=$1
	local outfile

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	outfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-gitlab-api.XXXXXXXXXX)

	# query user details
	if ! graphql_api_call "${outfile}" POST projects "${query}"; then
		msg_warn "  Invalid token provided?"
		exit 1
	fi

	cat "${outfile}"
	return 0
}

gitlab_lookup_project_names() {
	local status_file=$1; shift
	local project_ids=("$@")
	local graphql_lookup_batch=200

	local project_name_cache_file tmp_file from length percentage
	local project_slice query projects mapping_output

	# collect project ids whose name needs to be looked up
	project_name_cache_file=$(get_cache_file gitlab/project_id_to_name)
	lock 11 "${project_name_cache_file}" "Locking project name cache"

	# early exit if there is nothing new to look up
	if (( ! ${#project_ids[@]} )); then
		cat "${project_name_cache_file}"
		# close project name cache lock
		lock_close 11
		return
	fi

	# reduce project_ids to uncached entries
	mapfile -t project_ids < <(
		printf "%s\n" "${project_ids[@]}" | \
			grep --invert-match --file <(awk '{ print $1 }' < "${project_name_cache_file}" ))

	# look up project names
	tmp_file=$(mktemp --tmpdir="${WORKDIR}" pkgctl-gitlab-api-spinner.tmp.XXXXXXXXXX)
	local entries="${#project_ids[@]}"
	local until=0
	while (( until < entries )); do
		from=${until}
		until=$(( until + graphql_lookup_batch ))
		if (( until > entries )); then
			until=${entries}
		fi
		length=$(( until - from ))

		percentage=$(( 100 * until / entries ))
		printf "📡 Querying GitLab project names: %s/%s [%s] %%spinner%%" \
			"${BOLD}${until}" "${entries}" "${percentage}%${ALL_OFF}"  \
			> "${tmp_file}"
		mv "${tmp_file}" "${status_file}"

		project_slice=("${project_ids[@]:${from}:${length}}")
		printf -v projects '"gid://gitlab/Project/%s",' "${project_slice[@]}"
		query='{
			projects(after: "" ids: ['"${projects}"']) {
				pageInfo {
					startCursor
					endCursor
					hasNextPage
				}
				nodes {
					id
					name
				}
			}
		}'
		mapping_output=$(gitlab_api_get_project_name_mapping "${query}")

		# update cache
		while read -r project_id project_name; do
			printf "%s %s\n" "${project_id}" "${project_name}" >> "${project_name_cache_file}"
		done < <(jq --raw-output \
			'.[] | "\(.id | rindex("/") as $lastSlash | .[$lastSlash+1:]) \(.name)"' \
			<<< "${mapping_output}")
	done

	cat "${project_name_cache_file}"

	# close project name cache lock
	lock_close 11
}

longest_package_name_from_ids() {
	local project_ids=("$@")
	local longest=0

	# collect project ids whose name needs to be looked up
	project_name_cache_file=$(get_cache_file gitlab/project_id_to_name)
	lock 11 "${project_name_cache_file}" "Locking project name cache"

	# read project_id to name mapping from cache
	while read -r project_id project_name; do
		if (( ${#project_name} > longest )) && in_array "${project_id}" "${project_ids[@]}"; then
			longest="${#project_name}"
		fi
	done < "${project_name_cache_file}"

	# close project name cache lock
	lock_close 11

	printf "%s" "${longest}"
}

# Convert arbitrary project names to GitLab valid path names.
#
# GitLab has several limitations on project and group names and also maintains
# a list of reserved keywords as documented on their docs.
# https://docs.gitlab.com/ee/user/reserved_names.html
#
# 1. replace single '+' between word boundaries with '-'
# 2. replace any other '+' with literal 'plus'
# 3. replace any special chars other than '_', '-' and '.' with '-'
# 4. replace consecutive '_-' chars with a single '-'
# 5. replace 'tree' with 'unix-tree' due to GitLab reserved keyword
gitlab_project_name_to_path() {
	local name=$1
	printf "%s" "${name}" \
		| sed -E 's/([a-zA-Z0-9]+)\+([a-zA-Z]+)/\1-\2/g' \
		| sed -E 's/\+/plus/g' \
		| sed -E 's/[^a-zA-Z0-9_\-\.]/-/g' \
		| sed -E 's/[_\-]{2,}/-/g' \
		| sed -E 's/^tree$/unix-tree/g'
}

gitlab_api_create_project() {
	local pkgbase=$1
	local outfile data path project_path

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	outfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-gitlab-api.XXXXXXXXXX)

	project_path=$(gitlab_project_name_to_path "${pkgbase}")

	# create GitLab project
	data='{
		"name": "'"${pkgbase}"'",
		"path": "'"${project_path}"'",
		"namespace_id": "'"${GIT_PACKAGING_NAMESPACE_ID}"'",
		"request_access_enabled": "false"
	}'
	if ! gitlab_api_call "${outfile}" POST "projects/" "${data}"; then
		return 1
	fi

	if ! path=$(jq --raw-output --exit-status '.path' < "${outfile}"); then
		msg_error "  failed to query path: $(cat "${outfile}")"
		return 1
	fi

	printf "%s" "${path}"
	return 0
}

# TODO: parallelize
# https://docs.gitlab.com/ee/api/search.html#scope-blobs
gitlab_api_search() {
	local search=$1
	local status_file=$2
	local outfile

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	outfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-gitlab-api.XXXXXXXXXX)

	if ! gitlab_api_call_paged "${outfile}" "${status_file}" GET "/groups/archlinux%2fpackaging%2fpackages/search?scope=blobs" "search=${search}"; then
		return 1
	fi

	cat "${outfile}"

	return 0
}

# https://docs.gitlab.com/ee/api/projects.html#get-single-project
gitlab_project() {
	local project=$1
	local outfile project_path

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	outfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-gitlab-api.XXXXXXXXXX)

	project_path=$(gitlab_project_name_to_path "${project}")

	if ! gitlab_api_call "${outfile}" GET "projects/archlinux%2fpackaging%2fpackages%2f${project_path}/"; then
		return 1
	fi

	cat "${outfile}"

	return 0
}

# TODO: parallelize
# https://docs.gitlab.com/ee/api/issues.html#list-project-issues
gitlab_projects_issues_list() {
	local project=$1
	local status_file=$2
	local params=${3:-}
	local data=${4:-}
	local outfile

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	outfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-gitlab-api.XXXXXXXXXX)

	if ! gitlab_api_call_paged "${outfile}" "${status_file}" GET "/projects/archlinux%2fpackaging%2fpackages%2f${project}/issues?${params}" "${data}"; then
		return 1
	fi

	cat "${outfile}"

	return 0
}

# TODO: parallelize
# https://docs.gitlab.com/ee/api/issues.html#list-project-issues
gitlab_group_issue_list() {
	local group=$1
	local status_file=$2
	local params=${3:-}
	local data=${4:-}
	local outfile

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	outfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-gitlab-api.XXXXXXXXXX)

	group=${group//\//%2f}
	params=${params//\[/%5B}
	params=${params//\]/%5D}

	if ! gitlab_api_call_paged "${outfile}" "${status_file}" GET "/groups/${group}/issues?${params}" "${data}"; then
		return 1
	fi

	cat "${outfile}"
}

# https://docs.gitlab.com/ee/api/issues.html#single-project-issue
gitlab_project_issue() {
	local pkgbase=$1
	local iid=$2
	local outfile data path project_path

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	outfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-gitlab-api.XXXXXXXXXX)

	project_path=$(gitlab_project_name_to_path "${pkgbase}")

	if ! gitlab_api_call "${outfile}" GET "projects/archlinux%2fpackaging%2fpackages%2f${project_path}/issues/${iid}"; then
		return 1
	fi

	if ! path=$(jq --raw-output --exit-status '.title' < "${outfile}"); then
		msg_error "  failed to query path: $(cat "${outfile}")"
		return 1
	fi

	cat "${outfile}"
	return 0
}

gitlab_project_issue_create() {
	local pkgbase=$1
	local title=$2
	local description=$3
	local confidential=$4
	shift 4
	local labels=("${@}")
	local outfile data iid project_path

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	outfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-gitlab-api.XXXXXXXXXX)

	project_path=$(gitlab_project_name_to_path "${pkgbase}")

	data=$(jq --null-input \
		--arg title "${title}" \
		--arg description "${description}" \
		--arg confidential "${confidential}" \
		--arg labels "$(join_by , "${labels[@]}")" \
		'$ARGS.named')
	if ! gitlab_api_call "${outfile}" POST "/projects/archlinux%2fpackaging%2fpackages%2f${project_path}/issues" "${data}"; then
		return 1
	fi

	if ! iid=$(jq --raw-output --exit-status '.iid' < "${outfile}"); then
		msg_error "  failed to query note: $(cat "${outfile}")"
		return 1
	fi

	cat "${outfile}"
	return 0
}

# TODO: parallelize
# https://docs.gitlab.com/ee/api/notes.html#list-project-issue-notes
gitlab_project_issue_notes() {
	local project=$1
	local iid=$2
	local status_file=$3
	local params=${4:-}
	local outfile

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	outfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-gitlab-api.XXXXXXXXXX)

	if ! gitlab_api_call_paged "${outfile}" "${status_file}" GET "/projects/archlinux%2fpackaging%2fpackages%2f${project}/issues/${iid}/notes?${params}"; then
		return 1
	fi

	cat "${outfile}"
	return 0
}

# https://docs.gitlab.com/ee/api/issues.html#edit-an-issue
gitlab_project_issue_edit() {
	local pkgbase=$1
	local iid=$2
	local params=$3
	local data=${4:-}
	local outfile data path project_path

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	outfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-gitlab-api.XXXXXXXXXX)

	project_path=$(gitlab_project_name_to_path "${pkgbase}")

	if ! gitlab_api_call "${outfile}" PUT "/projects/archlinux%2fpackaging%2fpackages%2f${project_path}/issues/${iid}?${params}" "${data}"; then
		return 1
	fi

	if ! path=$(jq --raw-output --exit-status '.title' < "${outfile}"); then
		msg_error "  failed to query path: $(cat "${outfile}")"
		return 1
	fi

	cat "${outfile}"
	return 0
}

gitlab_create_project_issue_note() {
	local pkgbase=$1
	local iid=$2
	local body=$3
	local outfile data path project_path

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	outfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-gitlab-api.XXXXXXXXXX)

	project_path=$(gitlab_project_name_to_path "${pkgbase}")

	data=$(jq --null-input --arg body "${body}" '$ARGS.named')
	if ! gitlab_api_call "${outfile}" POST "/projects/archlinux%2fpackaging%2fpackages%2f${project_path}/issues/${iid}/notes" "${data}"; then
		return 1
	fi

	if ! path=$(jq --raw-output --exit-status '.body' < "${outfile}"); then
		msg_error "  failed to query note: $(cat "${outfile}")"
		return 1
	fi

	cat "${outfile}"
	return 0
}

gitlab_project_issue_move() {
	local pkgbase=$1
	local iid=$2
	local to_project_id=$3
	local outfile path project_path

	[[ -z ${WORKDIR:-} ]] && setup_workdir
	outfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-gitlab-api.XXXXXXXXXX)

	project_path=$(gitlab_project_name_to_path "${pkgbase}")

	if ! gitlab_api_call "${outfile}" POST "/projects/archlinux%2fpackaging%2fpackages%2f${project_path}/issues/${iid}/move?to_project_id=${to_project_id}"; then
		return 1
	fi

	if ! path=$(jq --raw-output --exit-status '.title' < "${outfile}"); then
		msg_error "  failed to move issue: $(cat "${outfile}")"
		return 1
	fi

	cat "${outfile}"
	return 0
}

gitlab_severity_from_labels() {
	local labels=("$@")
	local severity="unknown"
	local label
	for label in "${labels[@]}"; do
		if [[ ${label} == severity::* ]]; then
			severity="${label#*-}"
		fi
	done
	printf "%s" "${severity}"
}

severity_as_gitlab_label() {
	local severity=$1
	case "${severity}" in
		lowest)
			printf "severity::5-%s" "${severity}" ;;
		low)
			printf "severity::4-%s" "${severity}" ;;
		medium)
			printf "severity::3-%s" "${severity}" ;;
		high)
			printf "severity::2-%s" "${severity}" ;;
		critical)
			printf "severity::1-%s" "${severity}" ;;
		*)
			return 1 ;;
	esac
	return 0
}

gitlab_priority_from_labels() {
	local labels=("$@")
	local priority="normal"
	local label
	for label in "${labels[@]}"; do
		if [[ ${label} == priority::* ]]; then
			priority="${label#*-}"
		fi
	done
	printf "%s" "${priority}"
}

priority_as_gitlab_label() {
	local priority=$1
	case "${priority}" in
		low)
			printf "priority::4-%s" "${priority}" ;;
		normal)
			printf "priority::3-%s" "${priority}" ;;
		high)
			printf "priority::2-%s" "${priority}" ;;
		urgent)
			printf "priority::1-%s" "${priority}" ;;
		*)
			return 1 ;;
	esac
	return 0
}

gitlab_scope_from_labels() {
	local labels=("$@")
	local scope="unknown"
	local label
	for label in "${labels[@]}"; do
		if [[ ${label} == scope::* ]]; then
			scope="${label#*::}"
		fi
	done
	printf "%s" "${scope}"
}

scope_as_gitlab_label() {
	local scope=$1
	if ! in_array "${scope}" "${DEVTOOLS_VALID_ISSUE_SCOPE[@]}"; then
		return 1
	fi
	printf "scope::%s" "${scope}"
}

gitlab_scope_short() {
	local scope=$1
	case "${scope}" in
		regression)
			scope=regress ;;
		enhancement)
			scope=enhance ;;
		documentation)
			scope=doc ;;
		reproducibility)
			scope=repro ;;
		out-of-date)
			scope=ood ;;
	esac
	printf "%s" "${scope}"
}

gitlab_scope_color() {
	local scope=$1
	local color="${GRAY}"

	case "${scope}" in
		bug)
			color="${DARK_RED}" ;;
		feature)
			color="${DARK_BLUE}" ;;
		security)
			color="${RED}" ;;
		question)
			color="${PURPLE}" ;;
		regression)
			color="${DARK_RED}" ;;
		enhancement)
			color="${DARK_BLUE}" ;;
		documentation)
			color="${ALL_OFF}" ;;
		reproducibility)
			color="${DARK_GREEN}" ;;
		out-of-date)
			color="${DARK_YELLOW}" ;;
	esac

	printf "%s" "${color}"
}

status_as_gitlab_label() {
	local status=$1
	if ! in_array "${status}" "${DEVTOOLS_VALID_ISSUE_STATUS[@]}"; then
		return 1
	fi
	printf "status::%s" "${status}"
	return 0
}

gitlab_issue_state_display() {
	local state=$1
	if [[ ${state} == opened ]]; then
		state=open
	fi
	printf "%s" "${state}"
}

gitlab_issue_status_from_labels() {
	local labels=("$@")
	local status=unconfirmed
	local label
	for label in "${labels[@]}"; do
		if [[ ${label} == status::* ]]; then
			status="${label#*::}"
		fi
	done
	printf "%s" "${status}"
}

gitlab_issue_status_short() {
	local status=$1
	if [[ ${status} == waiting-* ]]; then
		status=waiting
	fi
	printf "%s" "${status}"
}

gitlab_issue_status_color() {
	local status=$1
	local color="${GRAY}"

	case "${status}" in
		confirmed)
			color="${GREEN}" ;;
		in-progress)
			color="${YELLOW}" ;;
		in-review)
			color="${PURPLE}" ;;
		on-hold|unconfirmed)
			color="${GRAY}" ;;
		waiting-input|waiting-upstream)
			color="${DARK_BLUE}" ;;
	esac

	printf "%s" "${color}"
}

resolution_as_gitlab_label() {
	local resolution=$1
	if ! in_array "${resolution}" "${DEVTOOLS_VALID_ISSUE_RESOLUTION[@]}"; then
		return 1
	fi
	printf "resolution::%s" "${resolution}"
}

gitlab_resolution_from_labels() {
	local labels=("$@")
	local label
	for label in "${labels[@]}"; do
		if [[ ${label} == resolution::* ]]; then
			printf "%s" "${label#*::}"
			return 0
		fi
	done
	return 1
}

gitlab_resolution_color() {
	local resolution=$1
	local color=""

	case "${resolution}" in
		cant-reproduce)
			color="${DARK_YELLOW}" ;;
		completed)
			color="${GREEN}" ;;
		duplicate)
			color="${GRAY}" ;;
		invalid)
			color="${DARK_YELLOW}" ;;
		not-a-bug)
			color="${GRAY}" ;;
		upstream)
			color="${PURPLE}" ;;
		wont-fix)
			color="${DARK_BLUE}" ;;
	esac

	printf "%s" "${color}"
}

gitlab_severity_color() {
	local severity=$1
	local color="${PURPLE}"

	case "${severity}" in
		lowest)
			color="${DARK_GREEN}" ;;
		low)
			color="${GREEN}" ;;
		medium)
			color="${YELLOW}" ;;
		high)
			color="${RED}" ;;
		critical)
			color="${RED}${UNDERLINE}" ;;
	esac

	printf "%s" "${color}"
}

gitlab_priority_color() {
	local priority=$1
	local color="${PURPLE}"

	case "${priority}" in
		low)
			color="${DARK_GREEN}" ;;
		normal)
			color="${GREEN}" ;;
		high)
			color="${YELLOW}" ;;
		urgent)
			color="${RED}" ;;
	esac

	printf "%s" "${color}"
}

gitlab_issue_state_color() {
	local state=$1
	local state_color="${DARK_GREEN}"

	if [[ ${state} == closed ]]; then
		state_color="${DARK_RED}"
	fi
	printf "%s" "${state_color}"
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_DB_UPDATE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_DB_UPDATE_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh

set -e


pkgctl_db_update_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS]

		Update the binary repository as final release step for packages that
		have been transfered and staged on ${PACKAGING_REPO_RELEASE_HOST}.

		OPTIONS
		    -h, --help    Show this help text

		EXAMPLES
		    $ ${COMMAND}
_EOF_
}

pkgctl_db_update() {
	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_db_update_usage
				exit 0
				;;
			*)
				die "invalid argument: %s" "$1"
				;;
		esac
	done

	ssh "${PACKAGING_REPO_RELEASE_HOST}" db-update
}

#!/bin/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_REPO_CREATE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_REPO_CREATE_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh
# shellcheck source=src/lib/repo/clone.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/repo/clone.sh
# shellcheck source=src/lib/repo/configure.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/repo/configure.sh

set -e


pkgctl_repo_create_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [PKGBASE]...

		Create a new Git packaging repository in the canonical GitLab namespace.

		This command requires a valid GitLab API authentication. To setup a new
		GitLab token or check the currently configured one please consult the
		'auth' subcommand for further instructions.

		If invoked without a parameter, try to create a packaging repository
		based on the PKGBUILD from the current working directory and configure
		the local repository afterwards.

		OPTIONS
		    -c, --clone   Clone the Git repository after creation
		    -h, --help    Show this help text

		EXAMPLES
		    $ ${COMMAND} libfoo
_EOF_
}

pkgctl_repo_create() {
	# options
	local pkgbases=()
	local pkgbase
	local clone=0
	local configure=0

	# variables
	local path

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_repo_create_usage
				exit 0
				;;
			-c|--clone)
				clone=1
				shift
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				pkgbases=("$@")
				break
				;;
		esac
	done

	# check if invoked without any path from within a packaging repo
	if (( ${#pkgbases[@]} == 0 )); then
		if [[ -f PKGBUILD ]]; then
			if ! path=$(realpath -e .); then
				die "failed to read path from current directory"
			fi
			pkgbases=("$(basename "${path}")")
			clone=0
			configure=1
		else
			pkgctl_repo_create_usage
			exit 1
		fi
	fi

	# create projects
	for pkgbase in "${pkgbases[@]}"; do
		if ! gitlab_api_create_project "${pkgbase}" >/dev/null; then
			die "failed to create project: ${pkgbase}"
		fi
		msg_success "Successfully created ${pkgbase}"
		if (( clone )); then
			pkgctl_repo_clone "${pkgbase}"
		elif (( configure )); then
			pkgctl_repo_configure
		fi
	done

	# some convenience hints if not in auto clone/configure mode
	if (( ! clone )) && (( ! configure )); then
		cat <<- _EOF_

		For new clones:
			$(msg2 "pkgctl repo clone ${pkgbases[*]}")
		For existing clones:
			$(msg2 "pkgctl repo configure ${pkgbases[*]}")
		_EOF_
	fi
}

#!/bin/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_REPO_CLONE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_REPO_CLONE_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/api/archweb.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/archweb.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh
# shellcheck source=src/lib/repo/configure.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/repo/configure.sh
# shellcheck source=src/lib/util/git.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/git.sh

source /usr/share/makepkg/util/message.sh

set -e
set -o pipefail


pkgctl_repo_clone_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [PKGBASE]...

		Clone Git packaging repositories from the canonical namespace.

		The configure command is subsequently invoked to synchronize the distro
		specs and makepkg.conf settings. The protocol option can be used
		for cloning packaging repositories without SSH access using read-only
		HTTPS.

		OPTIONS
		    -m, --maintainer=NAME  Clone all packages of the named maintainer
		    --protocol https       Clone the repository over https
		    --switch VERSION       Switch the current working tree to a specified version
		    --universe             Clone all existing packages, useful for cache warming
		    -j, --jobs N           Run up to N jobs in parallel (default: $(nproc))
		    -h, --help             Show this help text

		EXAMPLES
		    $ ${COMMAND} libfoo linux libbar
		    $ ${COMMAND} --maintainer mynickname
		    $ ${COMMAND} --switch 1:1.0-2 libfoo
_EOF_
}

pkgctl_repo_clone() {
	if (( $# < 1 )); then
		pkgctl_repo_clone_usage
		exit 0
	fi

	# options
	local protocol=ssh
	local GIT_REPO_BASE_URL=${GIT_PACKAGING_URL_SSH}
	local CLONE_ALL=0
	local MAINTAINER=
	local VERSION=
	local CONFIGURE_OPTIONS=()
	local jobs=
	jobs=$(nproc)
	local -a pkgbases

	# variables
	local command=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	local project_path

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_repo_clone_usage
				exit 0
				;;
			--protocol=https)
				GIT_REPO_BASE_URL=${GIT_PACKAGING_URL_HTTPS}
				protocol=https
				CONFIGURE_OPTIONS+=("$1")
				shift
				;;
			--protocol)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if [[ $2 == https ]]; then
					GIT_REPO_BASE_URL=${GIT_PACKAGING_URL_HTTPS}
				else
					die "unsupported protocol: %s" "$2"
				fi
				protocol="$2"
				CONFIGURE_OPTIONS+=("$1" "$2")
				shift 2
				;;
			-m|--maintainer)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				MAINTAINER="$2"
				shift 2
				;;
			--maintainer=*)
				MAINTAINER="${1#*=}"
				shift
				;;
			--switch)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				# shellcheck source=src/lib/repo/switch.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/repo/switch.sh
				VERSION="$2"
				shift 2
				;;
			--switch=*)
				# shellcheck source=src/lib/repo/switch.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/repo/switch.sh
				VERSION="${1#*=}"
				shift
				;;
			--universe)
				CLONE_ALL=1
				shift
				;;
			-j|--jobs)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				jobs=$2
				shift 2
				;;
			--)
				shift
				break
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				pkgbases=("$@")
				break
				;;
		esac
	done

	# Query packages of a maintainer
	if [[ -n ${MAINTAINER} ]]; then
		mapfile -t pkgbases < <(archweb_query_maintainer_packages "${MAINTAINER}")
		if ! wait $!; then
			die "Failed to query maintainer packages"
		fi
	fi

	# Query all released packages
	if (( CLONE_ALL )); then
		mapfile -t pkgbases < <(archweb_query_all_packages)
		if ! wait $!; then
			die "Failed to query all packages"
		fi
	fi

	# parallelization
	if [[ ${jobs} != 1 ]] && (( ${#pkgbases[@]} > 1 )); then
		# force colors in parallel if parent process is colorized
		if [[ -n ${BOLD} ]]; then
			export DEVTOOLS_COLOR=always
		fi
		# assign command options
		if [[ -n "${VERSION}" ]]; then
			command+=" --switch '${VERSION}'"
		fi

		# warm up ssh connection as it may require user input (key unlock, hostkey verification etc)
		if [[ ${protocol} == ssh ]]; then
			git_warmup_ssh_connection
		fi

		if ! parallel --bar --jobs "${jobs}" "${command}" ::: "${pkgbases[@]}"; then
			die 'Failed to clone some packages, please check the output'
			exit 1
		fi
		exit 0
	fi

	for pkgbase in "${pkgbases[@]}"; do
		if [[ ! -d ${pkgbase} ]]; then
			msg "Cloning ${pkgbase} ..."
			project_path=$(gitlab_project_name_to_path "${pkgbase}")
			remote_url="${GIT_REPO_BASE_URL}/${project_path}.git"
			if ! git clone --origin origin "${remote_url}" "${pkgbase}"; then
				die 'failed to clone %s' "${pkgbase}"
			fi
		else
			warning "Skip cloning ${pkgbase}: Directory exists"
		fi

		pkgctl_repo_configure "${CONFIGURE_OPTIONS[@]}" "${pkgbase}"

		if [[ -n "${VERSION}" ]]; then
			pkgctl_repo_switch "${VERSION}" "${pkgbase}"
		fi
	done
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_API_ARCHWEB_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_API_ARCHWEB_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh

set -e
set -o pipefail


archweb_query_all_packages() {
	local -a pkgbases

	[[ -z ${WORKDIR:-} ]] && setup_workdir

	stat_busy "Query all released packages"
	mapfile -t pkgbases < <(
		curl --location --show-error --no-progress-meter --fail --retry 3 --retry-delay 3 \
			"${PKGBASE_MAINTAINER_URL}" 2> "${WORKDIR}/error" \
			| jq --raw-output --exit-status 'keys[]' 2> "${WORKDIR}/error"
	)
	if ! wait $!; then
		stat_failed
		print_workdir_error
		return 1
	fi
	stat_done

	printf "%s\n" "${pkgbases[@]}"
	return 0
}


archweb_query_maintainer_packages() {
	local maintainer=$1
	local -a pkgbases

	[[ -z ${WORKDIR:-} ]] && setup_workdir

	stat_busy "Query maintainer packages"
	mapfile -t pkgbases < <(
		curl --location --show-error --no-progress-meter --fail --retry 3 --retry-delay 3 \
			"${PKGBASE_MAINTAINER_URL}" 2> "${WORKDIR}/error" \
			| jq --raw-output --exit-status '. as $parent | keys[] | select(. as $key | $parent[$key] | index("'"${maintainer}"'"))' 2> "${WORKDIR}/error"
	)
	if ! wait $!; then
		stat_failed
		print_workdir_error
		return 1
	fi
	stat_done

	printf "%s\n" "${pkgbases[@]}"
	return 0
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_ISSUE_LIST_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_ISSUE_LIST_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh
# shellcheck source=src/lib/util/term.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/term.sh

set -eo pipefail


pkgctl_issue_list_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [PKGBASE]

		The pkgctl issue list command is used to list issues associated with a specific
		packaging project or the entire packaging subgroup in Arch Linux. This command
		facilitates efficient issue management by allowing users to list and filter
		issues based on various criteria.

		Results can also be displayed directly in a web browser for easier navigation
		and review.

		OPTIONS
		    -g, --group             Get issues from the whole packaging subgroup
		    -w, --web               View results in a browser
		    -h, --help              Show this help text

		FILTER
		    -A, --all               Get all issues including closed
		    -c, --closed            Get only closed issues
		    -U, --unconfirmed       Shorthand to filter by unconfirmed status label
		    --search SEARCH         Search <string> in the fields defined by --in
		    --in LOCATION           Search in title or description (default: all)
		    -l, --label NAME        Filter issue by label <name>
		    --confidentiality TYPE  Filter by confidentiality
		    --priority PRIORITY     Shorthand to filter by priority label
		    --resolution REASON     Shorthand to filter by resolution label
		    --scope SCOPE           Shorthand to filter by scope label
		    --severity SEVERITY     Shorthand to filter by severity label
		    --status STATUS         Shorthand to filter by status label
		    --assignee USERNAME     Filter issues assigned to the given username
		    --assigned-to-me        Shorthand to filter issues assigned to you
		    --author USERNAME       Filter issues authored by the given username
		    --created-by-me         Shorthand to filter issues created by you

		EXAMPLES
		    $ ${COMMAND} libfoo libbar
		    $ ${COMMAND} --group --unconfirmed
_EOF_
}

pkgctl_issue_list() {
	if (( $# < 1 )) && [[ ! -f PKGBUILD ]]; then
		pkgctl_issue_list_usage
		exit 0
	fi

	local paths path project_path params web_params label username issue_url

	local group=0
	local web=0
	local confidential=0
	local state=opened
	local request_data=""
	local search_in="all"
	local labels=()
	local assignee=
	local author=
	local scope=all
	local confidentiality=

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_issue_list_usage
				exit 0
				;;
			-A|--all)
				state=all
				shift
				;;
			-c|--closed)
				state=closed
				shift
				;;
			-U|--unconfirmed)
				labels+=("$(status_as_gitlab_label unconfirmed)")
				shift
				;;
			-g|--group)
				group=1
				shift
				;;
			-w|--web)
				web=1
				shift
				;;
			--in)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				search_in=$2
				shift 2
				;;
			--search)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				request_data="search=$2"
				web_params+="&search=$2"
				shift 2
				;;
			-l|--label)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				labels+=("$2")
				shift 2
				;;
			--confidentiality)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				confidentiality=$2
				if ! in_array "${confidentiality}" "${DEVTOOLS_VALID_ISSUE_CONFIDENTIALITY[@]}"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				shift 2
				;;
			--priority)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(priority_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				labels+=("$label")
				shift 2
				;;
			--resolution)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(resolution_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				labels+=("$label")
				shift 2
				;;
			--scope)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(scope_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				labels+=("$label")
				shift 2
				;;
			--severity)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(severity_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				labels+=("$label")
				shift 2
				;;
			--status)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(status_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				labels+=("$label")
				shift 2
				;;
			--assignee)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				assignee="$2"
				shift 2
				;;
			--assigned-to-me)
				scope=assigned_to_me
				shift
				;;
			--author)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				author="$2"
				shift 2
				;;
			--created-by-me)
				scope=created_by_me
				shift
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				paths=("$@")
				break
				;;
		esac
	done

	if [[ ${search_in} == all ]]; then
		search_in="title,description"
	else
		web_params+="&in=${search_in^^}"
	fi
	params+="&in=${search_in}"

	if [[ ${state} != all ]]; then
		params+="&state=${state}"
	fi
	web_params+="&state=${state}"

	if (( ${#labels} )); then
		params+="&labels=$(join_by , "${labels[@]}")"
		web_params+="&label_name[]=$(join_by "&label_name[]=" "${labels[@]}")"
	fi

	if [[ -n ${scope} ]]; then
		params+="&scope=${scope}"
		if (( web )); then
			if ! username=$(gitlab_api_get_user); then
				exit 1
			fi
			case "${scope}" in
				created_by_me) author=${username} ;;
				assigned_to_me) assignee=${username} ;;
			esac
		fi
	fi

	if [[ -n ${assignee} ]]; then
		params+="&assignee_username=${assignee}"
		web_params+="&assignee_username=${assignee}"
	fi

	if [[ -n ${author} ]]; then
		params+="&author_username=${author}"
		web_params+="&author_username=${author}"
	fi

	if [[ -n ${confidentiality} ]]; then
		if [[ ${confidentiality} == confidential ]]; then
			params+="&confidential=true"
			web_params+="&confidential=yes"
		else
			params+="&confidential=false"
			web_params+="&confidential=no"
		fi
	fi

	# check if invoked without any path from within a packaging repo
	if (( ${#paths[@]} == 0 )); then
		if [[ -f PKGBUILD ]] && (( ! group )); then
			paths=("$(realpath --canonicalize-existing .)")
		elif (( ! group )); then
			pkgctl_issue_list_usage
			exit 1
		fi
	fi

	if (( web )) && ! command -v xdg-open &>/dev/null; then
		die "The web option requires 'xdg-open'"
	fi

	local separator="	"

	for path in "${paths[@]}"; do
		# skip paths from a glob that aren't directories
		if [[ -e "${path}" ]] && [[ ! -d "${path}" ]]; then
			continue
		fi

		pkgbase=$(basename "${path}")
		project_path=$(gitlab_project_name_to_path "${pkgbase}")

		echo "${UNDERLINE}${pkgbase}${ALL_OFF}"

		if (( web )); then
			issue_url="${GIT_PACKAGING_URL_HTTPS}/${project_path}/-/issues/?${web_params}"
			echo "Opening ${issue_url} in your browser."
			xdg-open "${issue_url}"
			continue
		fi

		status_dir=$(mktemp --tmpdir="${WORKDIR}" --directory pkgctl-gitlab-api.XXXXXXXXXX)
		printf "📡 Querying GitLab issues API..." > "${status_dir}/status"
		term_spinner_start "${status_dir}"
		if ! output=$(gitlab_projects_issues_list "${project_path}" "${status_dir}/status" "${params}" "${request_data}"); then
			term_spinner_stop "${status_dir}"
			echo
			continue
		fi
		term_spinner_stop "${status_dir}"

		issue_count=$(jq --compact-output 'length' <<< "${output}")
		if (( issue_count == 0 )); then
			echo "No open issues match your search"
			echo
			continue
		else
			echo "Showing ${issue_count} issues that match your search"
		fi

		print_issue_list "${output}"
	done

	if (( group )); then
		if (( web )); then
			issue_url="https://${GITLAB_HOST}/groups/${GIT_PACKAGING_NAMESPACE}/-/issues/?${web_params}"
			echo "Opening ${issue_url} in your browser."
			xdg-open "${issue_url}"
			return
		fi

		status_dir=$(mktemp --tmpdir="${WORKDIR}" --directory pkgctl-gitlab-api.XXXXXXXXXX)
		printf "📡 Querying GitLab issues API..." > "${status_dir}/status"
		term_spinner_start "${status_dir}"
		if ! output=$(gitlab_group_issue_list "${GIT_PACKAGING_NAMESPACE_ID}" "${status_dir}/status" "${params}" "${request_data}"); then
			term_spinner_stop "${status_dir}"
			exit 1
		fi
		term_spinner_stop "${status_dir}"

		print_issue_list "${output}"
	fi
}

print_issue_list() {
	local output=$1
	local limit=${2:-100}
	local i=0
	local status_dir
	local longest_pkgname

	# limit results
	output=$(jq ".[:${limit}]" <<< "${output}")

	mapfile -t project_ids < <(
		jq --raw-output '[.[].project_id] | unique[]' <<< "${output}")

	status_dir=$(mktemp --tmpdir="${WORKDIR}" --directory pkgctl-gitlab-api.XXXXXXXXXX)
	printf "📡 Querying GitLab project names..." > "${status_dir}/status"
	term_spinner_start "${status_dir}"

	# read project_id to name mapping from cache
	declare -A project_name_lookup=()
	while read -r project_id project_name; do
		project_name_lookup[${project_id}]=${project_name}
	done < <(gitlab_lookup_project_names "${status_dir}/status" "${project_ids[@]}")
	longest_pkgname=$(longest_package_name_from_ids "${project_ids[@]}")

	term_spinner_stop "${status_dir}"

	result_file=$(mktemp --tmpdir="${WORKDIR}" pkgctl-issue-list.XXXXXXXXXX)
	printf "📡 Collecting issue information %%spinner%%" > "${status_dir}/status"
	term_spinner_start "${status_dir}"

	local columns="ID,Title,Scope,Status,Severity,Age"
	if (( group )); then
		columns="ID,Package,Title,Scope,Status,Severity,Age"
	fi

	# pretty print each result
	while read -r result; do
		if (( i > limit )); then
			break
		fi
		i=$(( ++i ))

		{ read -r project_id; read -r iid; read -r title; read -r state; read -r created_at; read -r confidential; } < <(
			jq --raw-output ".project_id, .iid, .title, .state, .created_at, .confidential" <<< "${result}"
		)
		mapfile -t labels < <(
			jq --raw-output ".labels[]" <<< "${result}"
		)

		pkgbase=${project_name_lookup[${project_id}]}
		created_at=$(relative_date_unit "${created_at}")
		severity="$(gitlab_severity_from_labels "${labels[@]}")"
		severity_color="$(gitlab_severity_color "${severity}")"
		state_color="$(gitlab_issue_state_color "${state}")"
		state="$(gitlab_issue_state_display "${state}")"
		status="$(gitlab_issue_status_from_labels "${labels[@]}")"
		status_color="$(gitlab_issue_status_color "${status}")"
		status="$(gitlab_issue_status_short "${status}")"
		scope="$(gitlab_scope_from_labels "${labels[@]}")"
		scope_color="$(gitlab_scope_color "${scope}")"
		scope="$(gitlab_scope_short "${scope}")"

		title_space=$(( COLUMNS - 7 - 10 - 15 - 12 - 10 ))
		if (( group )); then
			title_space=$(( title_space - longest_pkgname ))
		fi
		if [[ ${confidential} == true ]]; then
			title_space=$(( title_space - 2 ))
		fi
		title=$(trim_string "${title_space}" "${title}")
		# gum is silly and doesn't allow double quotes
		title=${title//\"/}
		if [[ ${confidential} == true ]]; then
			title="${YELLOW}${PKGCTL_TERM_ICON_CONFIDENTIAL} ${title}${ALL_OFF}"
		fi

		if (( group )); then
			printf "%s\n" "${state_color}#$iid${ALL_OFF}${separator}${BOLD}${pkgbase}${separator}${ALL_OFF}${title}${separator}${scope_color}${scope}${ALL_OFF}${separator}${status_color}${status}${separator}${severity_color}${severity}${ALL_OFF}${separator}${GRAY}${created_at}${ALL_OFF}" \
				>> "${result_file}"
		else
			printf "%s\n" "${state_color}#$iid${ALL_OFF}${separator}${title}${separator}${scope_color}${scope}${ALL_OFF}${separator}${status_color}${status}${separator}${severity_color}${severity}${ALL_OFF}${separator}${GRAY}${created_at}${ALL_OFF}" \
				>> "${result_file}"
		fi
	done < <(jq --compact-output '.[]' <<< "${output}")

	term_spinner_stop "${status_dir}"

	gum table --print --border="none" --columns="${columns}" \
		--separator="${separator}" --file "${result_file}"
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later
:

# shellcheck disable=2034
CHROOT_VERSION='v6'

##
#  usage : check_root $keepenv
##
check_root() {
	local keepenv=$1
	shift
	local orig_argv=("$@")

	(( EUID == 0 )) && return
	if type -P sudo >/dev/null; then
		exec sudo --preserve-env="${keepenv}" -- "${orig_argv[@]}"
	else
		exec su root -c "$(printf ' %q' "${orig_argv[@]}")"
	fi
}

##
#  usage : is_zfs( $path )
# return : whether $path is on zfs
##
is_zfs() {
	[[ -e "$1" && "$(stat -f -c %T "$1")" == zfs ]]
}

##
#  usage : is_zfs_dataset( $path [, $dataset_name [, $dataset_mountpoint]] )
# return : whether $path is the root of a (mounted) zfs dataset.
#    out : ${$dataset_name}: the name of the zfs dataset mounted onto $path.
#    out : ${$dataset_mountpoint}: the currently visible mountpoint of the zfs dataset mounted onto $path.
##
is_zfs_dataset() {
	local _dataset _fsroot _mountpoint

	# NB: use `findmnt -o TARGET` instead of `zfs list -Ho mountpoint`
	#     because the latter is not namespace-aware
	#     (we might be in a container)
	# && _mountpoint="$(zfs list -Ho mountpoint "$_dataset")" \

	# shellcheck disable=SC2034
	[[ -e "$1" ]] \
	&& is_zfs "$1" \
	&& mountpoint -q "$1" \
	&& _fsroot="$(findmnt --list --noheading --first-only --nofsroot -o FSROOT "$1")" \
	&& [[ "$_fsroot" == / ]] \
	&& _dataset="$(findmnt --list --noheading --first-only --nofsroot -o SOURCE "$1")" \
	&& [[ "$_dataset" && "$_dataset" == */* && "$_dataset" != /* ]] \
	&& _mountpoint="$(findmnt --list --noheading --first-only --nofsroot -o TARGET "$1")" \
	&& [[ "$1" -ef "$_mountpoint" ]] \
	&& printf -v "${2-_}" "%s" "$_dataset" \
	&& printf -v "${3-_}" "%s" "$_mountpoint" \
	# EOL
}

##
#  usage : is_zfs_dataset_or_nothing( $path [, $dataset_name [, $dataset_mountpoint]] )
# return : whether $path is the root of a (mounted) zfs dataset, or true if $path does not exist.
#    out : ${$dataset_name}: the name of the zfs dataset mounted onto $path.
#    out : ${$dataset_mountpoint}: the canonical mountpoint of the zfs dataset mounted onto $path.
##
is_zfs_dataset_or_nothing() {
	[[ ! -e "$1" ]] || is_zfs_dataset "$@"
}

##
#  usage : is_zfs_dataset_or_nothing( $path [, $dataset_name] )
# return : whether $path is the root of a (mounted) zfs dataset, or true if $path does not exist.
#    out : ${$dataset_name}: the name of the zfs dataset mounted onto $path.
#    out : ${$dataset_mountpoint}: the currently visible mountpoint of the zfs dataset mounted onto $path.
##
is_zfs_dataset_subdir() {
	local _dataset _fsroot _mountpoint

	# NB: use `findmnt -o TARGET` instead of `zfs list -Ho mountpoint`
	#     because the latter is not namespace-aware
	#     (we might be in a container)
	# && _mountpoint="$(zfs list -Ho mountpoint "$_dataset")" \

	[[ -e "$1" ]] \
	&& is_zfs "$1" \
	&& _fsroot="$(findmnt --list --noheading --first-only --nofsroot -o FSROOT --target "$1")" \
	&& [[ "$_fsroot" == / ]] \
	&& _dataset="$(findmnt --list --noheading --first-only --nofsroot -o SOURCE --target "$1")" \
	&& [[ "$_dataset" && "$_dataset" == */* && "$_dataset" != /* ]] \
	&& _mountpoint="$(findmnt --list --noheading --first-only --nofsroot -o TARGET --target "$1")" \
	&& [[ "$1" == "$_mountpoint" || "$1" == "$_mountpoint"/* ]] \
	&& printf -v "${2-_}" "%s" "$_dataset" \
	&& printf -v "${3-_}" "%s" "$_mountpoint" \
	# EOL
}

zfs_dataset_clone() {
	local src="$1" dest="$2"
	local dataset clone clonename

	is_zfs_dataset "$src" dataset || return
	clone="$(dirname "$dataset")/$(basename "$dest")"
	clonename="$(basename "$clone")"

	zfs snapshot "$dataset@clone-$clonename" || return
	zfs clone -u "$dataset@clone-$clonename" "$clone" || return
	mount --mkdir -t zfs "$clone" "$dest" -o zfsutil || return
}

zfs_dataset_delete() {
	local src="$1"
	local dataset origin

	[[ -e "$src" ]] || return 0
	is_zfs_dataset "$src" dataset || return
	origin="$(zfs list -Ho origin "$dataset")" || return

	if [[ $origin == *@clone-$(basename "$dataset") ]]; then
		zfs destroy -v -R "$origin" || return
	else
		zfs destroy -v -r "$dataset" || return
	fi
	rm -fd "$src"
}

zfs_dataset_create_at() {
	local dest="$1"
	local dataset mountpoint subdir

	# if $dest is already a zfs dataset, do nothing
	is_zfs_dataset "$dest" dataset && return 0

	is_zfs_dataset_subdir "$dest" dataset mountpoint || return
	# $dest cannot be a dataset itself, it must be a subdirectory
	[[ "$dest" == "$mountpoint"/* ]] || return
	subdir="${dest#"$mountpoint/"}"

	rm -fd "$dest" || return
	zfs create -u -pp "$dataset/$subdir" || return
	mount --mkdir -t zfs "$dataset/$subdir" "$dest" -o zfsutil || return
}

##
#  usage : is_btrfs( $path )
# return : whether $path is on a btrfs
##
is_btrfs() {
	[[ -e "$1" && "$(stat -f -c %T "$1")" == btrfs ]]
}

##
#  usage : is_subvolume( $path )
# return : whether $path is a the root of a btrfs subvolume (including
#          the top-level subvolume).
##
is_subvolume() {
	[[ -e "$1" && "$(stat -f -c %T "$1")" == btrfs && "$(stat -c %i "$1")" == 256 ]]
}

##
#  usage : subvolume_delete_recursive( $path )
#
#    Find all btrfs subvolumes under and including $path and delete them.
##
subvolume_delete_recursive() {
	local subvol

	is_subvolume "$1" || return 0

	while IFS= read -d $'\0' -r subvol; do
		if ! subvolume_delete_recursive "$subvol"; then
			return 1
		fi
	done < <(find "$1" -mindepth 1 -xdev -depth -inum 256 -print0)
	if ! btrfs subvolume delete "$1" &>/dev/null; then
		error "Unable to delete subvolume %s" "$subvol"
		return 1
	fi

	return 0
}

##
#  $1: name of the output array
#  $2...: mirror URL
#  out: ${$1}: extracted cache directories
#
#    Extract potential cache directories from a list of mirror URIs.
##
_extract_cache_dirs () {
	declare -n cache_dirs__=$1
	local mirror m
	shift

	for mirror; do
		if [[ $mirror =~ ^file://(/.+)/\$repo/os/\$arch$ ]]; then
			# full pool mirrors
			mirror="${BASH_REMATCH[1]}"
			for m in "$mirror"/pool/*/; do
				in_array "$m" "${cache_dirs__[@]}" || cache_dirs__+=("$m")
			done
		elif [[ $mirror =~ ^file://(/.+)/?$ ]]; then
			# other arbitrary file:// sources
			mirror="${BASH_REMATCH[1]}"
			in_array "$mirror" "${cache_dirs__[@]}" || cache_dirs__+=("$mirror")
		fi
	done
}
# }}}

##
#  $1: path to pacman.conf
#  $2: name of the output array for R/O cache dirs
#      (inferred from file:/// mirrors)
#  $3: name of the output array for primary cache dirs
#      (if non-empty, not overwritten)
#  out: ${$2}, ${$3}: extracted cache directories
#  out: $host_mirrors, $host_cachemirrors: set according to host pacman.conf
#
#    Extract potential cache directories from a pacman.conf.
#    NOTE: R/O cache dirs must have priority over primary cache dirs
##
get_cache_dirs () {
	local pac_conf=$1
	declare -n cache_dirs_ro_=$2 cache_dirs_rw_=$3
	local use_host_mirrors
	local -a cache_dirs_host
	local repo
	local -a mirrors

	# extract primary cache dirs if it was not set
	if (( ${#cache_dirs_rw_[@]} == 0 )); then
		mapfile -t cache_dirs_rw_ < <(pacman-conf --config "$pac_conf" CacheDir)
	fi

	# check if the primary cache dir is equal to host's
	# (i.e. we are building for the same Arch flavor as the host)
	mapfile -t cache_dirs_host < <(pacman-conf CacheDir)
	if [[ "${cache_dirs_rw_[-1]}" -ef /var/cache/pacman/pkg \
	   && "${cache_dirs_host[-1]}" -ef /var/cache/pacman/pkg ]]; then
	   	use_host_mirrors=1
	fi

	# first, analyze target mirrors (for all repos)
	# we need this to add local repo file:/// mirrors to target's CacheDirs
	while read -r repo; do
		# (note `sed -n 's###p'` is NOT used)
		mapfile -t mirrors < <( \
			pacman-conf --config "$pac_conf" --repo "$repo" Server 2>/dev/null \
			| sed -r "s#(.*/)${repo}/(os/)?$(uname -m)#\1\$repo\2\$arch#" \
		)
		_extract_cache_dirs cache_dirs_ro_ "${mirrors[@]}"
	done < <(pacman-conf --config "$pac_conf" --repo-list)

	# if target's primary cache dir is same as host's, also analyze host mirrors
	# (use [extra] as an exemplar)
	# 1. to pull whole-pool file:/// mirrors to add to target's CacheDirs
	# 2. to pull any CacheServers to add to target's mirrorlist
	if (( use_host_mirrors )); then
		# extract and generalize host mirrors
		# (note `sed -n 's###p'` is used as we only want generalizable mirrors)
		mapfile -t host_mirrors < <( \
			pacman-conf --repo extra Server 2>/dev/null \
			| sed -nr "s#(.*/)extra/(os/)?$(uname -m)#\1\$repo/\2\$arch#p" \
		)
		_extract_cache_dirs cache_dirs_ro_ "${host_mirrors[@]}"

		# ...and cachemirrors
		mapfile -t host_cachemirrors < <( \
			pacman-conf --repo extra CacheServer 2>/dev/null \
			| sed -nr "s#(.*/)extra/(os/)?$(uname -m)#\1\$repo/\2\$arch#p" \
		)
		# also try to extract CacheDirs for completeness
		_extract_cache_dirs cache_dirs_ro_ "${host_cachemirrors[@]}"
	fi

	# final cache_dirs_{ro,rw} looks like this:
	#   cache_dirs_ro=(chroot-conf-mirrors host-conf-mirrors)
	#   cache_dirs_rw=(                                      global-cachedirs)
	# this way, the most relevant directories are listed first (in pacman.conf priority order)
	# this guarantees that we will pick the right cached package even if repos have duplicates
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_DB_REMOVE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_DB_REMOVE_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/util/pacman.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/pacman.sh
# shellcheck source=src/lib/util/term.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/term.sh
# shellcheck source=src/lib/valid-repos.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/valid-repos.sh

set -e


pkgctl_db_remove_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [REPO] [PKGBASE]...

		Remove packages from pacman repositories. By default passing a pkgbase removes
		all split packages, debug packages as well as entries from the state repo for
		all existing architectures.

		Beware when using the --partial option, as it may most likely lead to
		undesired effects by leaving debug packages behind as well as dangling entries
		in the state repository.

		OPTIONS
		    -a, --arch    Remove only one specific architecture (disables auto-detection)
		    --partial     Remove only partial pkgnames from a split package. This leaves
		                  debug packages behind and pkgbase entries in the state repo.
		    --noconfirm   Bypass any confirmation messages, should only be used with caution
		    -h, --help    Show this help text

		EXAMPLES
		    $ ${COMMAND} core-testing libfoo libbar
		    $ ${COMMAND} --arch x86_64 core libyay
_EOF_
}

pkgctl_db_remove() {
	local REPO=""
	local PKGBASES=()
	local pkgnames=()
	local partial=0
	local confirm=1
	local dbscripts_options=()
	local lookup_repo=multilib
	local pkgname

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_db_remove_usage
				exit 0
				;;
			--partial)
				partial=1
				dbscripts_options+=(--partial)
				shift
				;;
			-a|--arch)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				dbscripts_options+=(--arch "$2")
				shift 2
				;;
			--noconfirm)
				confirm=0
				shift
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				break
				;;
		esac
	done

	if (( $# < 2 )); then
		pkgctl_db_remove_usage
		exit 1
	fi

	REPO=$1
	shift
	PKGBASES+=("$@")
	pkgnames=("${PKGBASES[@]}")

	# check if the target repo is valid
	if ! in_array "${REPO}" "${DEVTOOLS_VALID_REPOS[@]}"; then
		die "Invalid repository target: %s" "${REPO}"
	fi

	# update pacman cache to query all pkgnames
	if (( ! partial )); then
		case ${REPO} in
			*-unstable)
				update_pacman_repo_cache unstable
				;;
			*-staging)
				update_pacman_repo_cache multilib-staging
				;;
			*-testing)
				update_pacman_repo_cache multilib-testing
				;;
			*)
				update_pacman_repo_cache multilib
				;;
		esac

		# fetch the pkgnames of all pkgbase as present in the repo
		mapfile -t pkgnames < <(get_pkgnames_from_repo_pkgbase "${REPO}" "${PKGBASES[@]}")
		echo

		if (( ! ${#pkgnames[@]} )); then
			error "Packages not found in %s" "${REPO}"
			exit 1
		fi
	fi

	# print list of packages
	printf "%sRemoving packages from %s:%s\n" "${RED}" "${REPO}" "${ALL_OFF}"
	for pkgname in "${pkgnames[@]}"; do
		printf "• %s\n" "${pkgname}"
	done

	# print explenation about partial removal
	if (( partial )); then
		echo
		msg_warn "${YELLOW}Removing only partial pkgnames from a split package.${ALL_OFF}"
		msg_warn "${YELLOW}This leaves debug packages and pkgbase entries in the state repo!${ALL_OFF}"
	fi

	# ask for confirmation
	if (( confirm )); then
		echo
		if ! prompt "${GREEN}${BOLD}?${ALL_OFF} Are you sure this is correct?"; then
			exit 1
		fi
	fi

	echo
	# shellcheck disable=SC2029
	ssh "${PACKAGING_REPO_RELEASE_HOST}" db-remove "${dbscripts_options[@]}" "${REPO}" "${PKGBASES[@]}"
}

#!/hint/bash
#
# This may be included with or without `set -euE`
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_COMMON_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_COMMON_SH="$(set +o|grep nounset)"

set +u +o posix
# shellcheck disable=1091
. /usr/share/makepkg/util.sh
$DEVTOOLS_INCLUDE_COMMON_SH

# Avoid any encoding problems
export LANG=C.UTF-8

# Avoid systemd trying to color the terminal on systemd-nspawn
export SYSTEMD_TINT_BACKGROUND=no

# Avoid diffoscope looking at remote debug info through readelf
unset DEBUGINFOD_URLS

# Set buildtool properties
export BUILDTOOL=devtools
export BUILDTOOLVER=1:1.5.1-1-any

# Set common properties
export PACMAN_KEYRING_DIR=/etc/pacman.d/gnupg
export GITLAB_HOST=gitlab.archlinux.org
export GIT_REPO_SPEC_VERSION=2
export GIT_PACKAGING_NAMESPACE=archlinux/packaging/packages
export GIT_PACKAGING_NAMESPACE_ID=11323
export GIT_PACKAGING_URL_SSH="git@${GITLAB_HOST}:${GIT_PACKAGING_NAMESPACE}"
export GIT_PACKAGING_URL_HTTPS="https://${GITLAB_HOST}/${GIT_PACKAGING_NAMESPACE}"
export PACKAGING_REPO_RELEASE_HOST=repos.archlinux.org
export PKGBASE_MAINTAINER_URL=https://archlinux.org/packages/pkgbase-maintainer
export AUR_URL_SSH=aur@aur.archlinux.org

# Create or reuse a shared SSH control socket with ControlMaster=auto. The
# connection is initialized on the first use and persisted for some time, so
# multiple invokations of devtools can share it.
# shellcheck disable=SC2016
export SSH_OPTS=(
  -o ControlMaster=auto
  -o ControlPersist=60s
  -o ControlPath='${XDG_RUNTIME_DIR}/devtools-%r@%h:%p'
)

export RSYNC_OPTS=(
  --rsh="ssh ${SSH_OPTS[*]}"
  --checksum
  --copy-links
  --human-readable
  --progress
  --partial
  # suffix the partial dir with the PID in order to avoid clashes
  --partial-dir=.partial.$$
  --delay-updates
)

# ensure TERM is set with a fallback to dumb
export TERM=${TERM:-dumb}

# check if messages are to be printed using color
if [[ -t 2 && "$TERM" != dumb ]] || [[ ${DEVTOOLS_COLOR} == always ]]; then
	colorize
	if tput setaf 0 &>/dev/null; then
		PURPLE="$(tput setaf 5)"
		DARK_GREEN="$(tput setaf 2)"
		DARK_RED="$(tput setaf 1)"
		DARK_BLUE="$(tput setaf 4)"
		DARK_YELLOW="$(tput setaf 3)"
		UNDERLINE="$(tput smul)"
		GRAY=$(tput setaf 242)
	else
		PURPLE="\e[35m"
		DARK_GREEN="\e[32m"
		DARK_RED="\e[31m"
		DARK_BLUE="\e[34m"
		DARK_YELLOW="\e[33m"
		UNDERLINE="\e[4m"
		GRAY=""
	fi
else
	# shellcheck disable=2034
	declare -gr ALL_OFF='' BOLD='' BLUE='' GREEN='' RED='' YELLOW='' PURPLE='' DARK_RED='' DARK_GREEN='' DARK_BLUE='' DARK_YELLOW='' UNDERLINE='' GRAY=''
fi

stat_busy() {
	local mesg=$1; shift
	# shellcheck disable=2059
	printf "${GREEN}==>${ALL_OFF}${BOLD} ${mesg}...${ALL_OFF}" "$@" >&2
}

stat_progress() {
	# shellcheck disable=2059
	printf "${BOLD}.${ALL_OFF}" >&2
}

stat_done() {
	# shellcheck disable=2059
	printf "${BOLD}done${ALL_OFF}\n" >&2
}

stat_failed() {
	# shellcheck disable=2059
	printf "${BOLD}${RED}failed${ALL_OFF}\n" >&2
}

msg_success() {
	local msg=$1
	local padding
	padding=$(echo "${msg}"|sed -E 's/( *).*/\1/')
	msg=$(echo "${msg}"|sed -E 's/ *(.*)/\1/')
	printf "%s %s\n" "${padding}${GREEN}✓${ALL_OFF}" "${msg}" >&2
}

msg_error() {
	local msg=$1
	local padding
	padding=$(echo "${msg}"|sed -E 's/( *).*/\1/')
	msg=$(echo "${msg}"|sed -E 's/ *(.*)/\1/')
	printf "%s %s\n" "${padding}${RED}x${ALL_OFF}" "${msg}" >&2
}

msg_warn() {
	local msg=$1
	local padding
	padding=$(echo "${msg}"|sed -E 's/( *).*/\1/')
	msg=$(echo "${msg}"|sed -E 's/ *(.*)/\1/')
	printf "%s %s\n" "${padding}${YELLOW}!${ALL_OFF}" "${msg}" >&2
}

print_workdir_error() {
	if [[ ! -f "${WORKDIR}"/error ]]; then
		return
	fi
	while read -r LINE; do
		error '%s' "${LINE}"
	done < "${WORKDIR}/error"
}

_setup_workdir=false
# Ensure that there is no outside value for WORKDIR leaking in
unset WORKDIR
setup_workdir() {
	[[ -z ${WORKDIR:-} ]] && WORKDIR=$(mktemp -d --tmpdir "${0##*/}.XXXXXXXXXX")
	_setup_workdir=true
	trap 'trap_abort' INT QUIT TERM HUP
	trap 'trap_exit' EXIT
}

cleanup() {
	if [[ -n ${WORKDIR:-} ]] && $_setup_workdir; then
		rm -rf "$WORKDIR"
	fi
	if tput setaf 0 &>/dev/null; then
		tput cnorm >&2
	fi
	exit "${1:-0}"
}

abort() {
	error 'Aborting...'
	cleanup 255
}

trap_abort() {
	trap - EXIT INT QUIT TERM HUP
	abort
}

trap_exit() {
	local r=$?
	trap - EXIT INT QUIT TERM HUP
	cleanup $r
}

die() {
	(( $# )) && error "$@"
	cleanup 255
}

##
#  usage : lock( $fd, $file, $message, [ $message_arguments... ] )
##
lock() {
	# Only reopen the FD if it wasn't handed to us
	if ! [[ "/dev/fd/$1" -ef "$2" ]]; then
		mkdir -p -- "$(dirname -- "$2")"
		eval "exec $1>>"'"$2"'
	fi

	if ! flock -n "$1"; then
		stat_busy "${@:3}"
		flock "$1"
		stat_done
	fi
}

##
#  usage : slock( $fd, $file, $message, [ $message_arguments... ] )
##
slock() {
	# Only reopen the FD if it wasn't handed to us
	if ! [[ "/dev/fd/$1" -ef "$2" ]]; then
		mkdir -p -- "$(dirname -- "$2")"
		eval "exec $1>"'"$2"'
	fi

	if ! flock -sn "$1"; then
		stat_busy "${@:3}"
		flock -s "$1"
		stat_done
	fi
}

##
#  usage : lock_close( $fd )
##
lock_close() {
	local fd=$1
	# https://github.com/koalaman/shellcheck/issues/862
	# shellcheck disable=2034
	exec {fd}>&-
}

##
# usage: pkgver_equal( $pkgver1, $pkgver2 )
##
pkgver_equal() {
	if [[ $1 = *-* && $2 = *-* ]]; then
		# if both versions have a pkgrel, then they must be an exact match
		[[ $1 = "$2" ]]
	else
		# otherwise, trim any pkgrel and compare the bare version.
		[[ ${1%%-*} = "${2%%-*}" ]]
	fi
}

##
#  usage: find_cached_package( $pkgname, $pkgver, $arch )
#
#    $pkgver can be supplied with or without a pkgrel appended.
#    If not supplied, any pkgrel will be matched.
##
shopt -s extglob
find_cached_package() {
	local searchdirs=("$PWD" "$PKGDEST") results=()
	local targetname=$1 targetver=$2 targetarch=$3
	local dir pkg packages pkgbasename name ver rel arch r results

	for dir in "${searchdirs[@]}"; do
		[[ -d $dir ]] || continue

		shopt -s extglob nullglob
		mapfile -t packages < <(printf "%s\n" "$dir"/"${targetname}"-"${targetver}"-*"${targetarch}".pkg.tar?(.!(sig|*.*)))
		shopt -u extglob nullglob

		for pkg in "${packages[@]}"; do
			[[ -f $pkg ]] || continue

			# avoid adding duplicates of the same inode
			for r in "${results[@]}"; do
				[[ $r -ef $pkg ]] && continue 2
			done

			# split apart package filename into parts
			pkgbasename=${pkg##*/}
			pkgbasename=${pkgbasename%.pkg.tar*}

			arch=${pkgbasename##*-}
			pkgbasename=${pkgbasename%-"$arch"}

			rel=${pkgbasename##*-}
			pkgbasename=${pkgbasename%-"$rel"}

			ver=${pkgbasename##*-}
			name=${pkgbasename%-"$ver"}

			if [[ $targetname = "$name" && $targetarch = "$arch" ]] &&
					pkgver_equal "$targetver" "$ver-$rel"; then
				results+=("$pkg")
			fi
		done
	done

	case ${#results[*]} in
		0)
			return 1
			;;
		1)
			printf '%s\n' "${results[0]}"
			return 0
			;;
		*)
			error 'Multiple packages found:'
			printf '\t%s\n' "${results[@]}" >&2
			return 1
	esac
}
shopt -u extglob

check_package_validity(){
	local pkgfile=$1
	if grep -q "packager = Unknown Packager" <(bsdtar -xOqf "$pkgfile" .PKGINFO); then
		die "PACKAGER was not set when building package"
	fi
	hashsum=sha256sum
	pkgbuild_hash=$(awk -v"hashsum=$hashsum" -F' = ' '$1 == "pkgbuild_"hashsum {print $2}' <(bsdtar -xOqf "$pkgfile" .BUILDINFO))
	if [[ "$pkgbuild_hash" != "$($hashsum PKGBUILD|cut -d' ' -f1)" ]]; then
		die "PKGBUILD $hashsum mismatch: expected $pkgbuild_hash"
	fi
}


# usage: grep_pkginfo pkgfile pattern
grep_pkginfo() {
	local _ret=()
	mapfile -t _ret < <(bsdtar -xOqf "$1" ".PKGINFO" | grep "^${2} = ")
	printf '%s\n' "${_ret[@]#${2} = }"
}


# Get the package name
getpkgname() {
	local _name

	_name="$(grep_pkginfo "$1" "pkgname")"
	if [[ -z $_name ]]; then
		error "Package '%s' has no pkgname in the PKGINFO. Fail!" "$1"
		exit 1
	fi

	echo "$_name"
}


# Get the package base or name as fallback
getpkgbase() {
	local _base

	_base="$(grep_pkginfo "$1" "pkgbase")"
	if [[ -z $_base ]]; then
		getpkgname "$1"
	else
		echo "$_base"
	fi
}


getpkgdesc() {
	local _desc

	_desc="$(grep_pkginfo "$1" "pkgdesc")"
	if [[ -z $_desc ]]; then
		error "Package '%s' has no pkgdesc in the PKGINFO. Fail!" "$1"
		exit 1
	fi

	echo "$_desc"
}


get_tag_from_pkgver() {
	local pkgver=$1
	local tag=${pkgver}

	tag=${tag/:/-}
	tag=${tag//~/.}
	echo "${tag}"
}


is_debug_package() {
	local pkgfile=${1} pkgbase pkgname pkgdesc
	pkgbase="$(getpkgbase "${pkgfile}")"
	pkgname="$(getpkgname "${pkgfile}")"
	pkgdesc="$(getpkgdesc "${pkgfile}")"
	[[ ${pkgdesc} == "Detached debugging symbols for "* && ${pkgbase}-debug = "${pkgname}" ]]
}

# Proxy function to check if a file exists. Using [[ -f ... ]] directly is not
# always wanted because we might want to expand bash globs first. This way we
# can pass unquoted globs to is_globfile() and have them expanded as function
# arguments before being checked.
is_globfile() {
	[[ -f $1 ]]
}

join_by() {
	local IFS="	"
	local sep=$1
	local split
	shift
	split=$(printf "%s" "$*")
	echo "${split//${IFS}/"${sep}"}"
}

trim_string() {
	local max_length=$1
	local string=$2

	if (( ${#string} > max_length )); then
		# Subtract 3 from max_length to accommodate "..."
		max_length=$((max_length - 3))
		string="${string:0:max_length}..."
	fi

	printf "%s" "${string}"
}

relative_date_unit() {
	local target_date=$1
	local now diff value units unit names

	target_date=$(date -d "$1" +%s)
	now=$(date +%s)
	diff=$((now - target_date))

	local names=(year month week day hour minute second)
	declare -A units=(
		[year]=$((60 * 60 * 24 * 365))
		[month]=$((60 * 60 * 24 * 30))
		[week]=$((60 * 60 * 24 * 7))
		[day]=$((60 * 60 * 24))
		[hour]=$((60 * 60))
		[minute]=60
		[second]=1
	)

	for unit in "${names[@]}"; do
		local value=$((diff / ${units[${unit}]}))
		if (( value > 1 )); then
			printf "%s %ss" "${value}" "${unit}"
			return
		elif (( value == 1 )); then
			printf "%s %s" "${value}" "${unit}"
			return
		fi
	done
	printf "1 second"
}


# escapes regex metacharacters in a given string
regex_escape() {
	# shellcheck disable=SC2001,SC2016
	sed 's/[\^.\[$()|*+?{\\]/\\&/g' <<<"$1"
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_DB_MOVE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_DB_MOVE_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh

set -e


pkgctl_db_move_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [SOURCE_REPO] [TARGET_REPO] [PKGBASE]...

		Move packages between binary repositories.

		OPTIONS
		    -h, --help         Show this help text

		EXAMPLES
		    $ ${COMMAND} extra-staging extra-testing libfoo libbar
		    $ ${COMMAND} extra core libfoo libbar
_EOF_
}

pkgctl_db_move() {
	local SOURCE_REPO=""
	local TARGET_REPO=""
	local PKGBASES=()

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_db_move_usage
				exit 0
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				break
				;;
		esac
	done

	if (( $# < 3 )); then
		pkgctl_db_move_usage
		exit 1
	fi

	SOURCE_REPO=$1
	TARGET_REPO=$2
	shift 2
	PKGBASES+=("$@")

	# shellcheck disable=SC2029
	ssh "${PACKAGING_REPO_RELEASE_HOST}" db-move "${SOURCE_REPO}" "${TARGET_REPO}" "${PKGBASES[@]}"
}

#!/bin/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_AUR_DROP_FROM_REPO_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_AUR_DROP_FROM_REPO_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/db/remove.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/db/remove.sh
# shellcheck source=src//lib/util/pacman.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/pacman.sh

source /usr/share/makepkg/util/message.sh

set -eo pipefail


pkgctl_aur_drop_from_repo_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [PATH]...

		Drops a specified package from the official repositories to the Arch
		User Repository.

		This command requires a local Git clone of the package repository. It
		reconfigures the repository for AUR compatibility and pushes it to the
		AUR. Afterwards, the package is removed from the official repository.

		By default, the package is automatically disowned in the AUR.

		OPTIONS
		    --no-disown    Do not disown the package on the AUR
		    -f, --force    Force push to the AUR overwriting the remote repository
		    -h, --help     Show this help text

		EXAMPLES
		    $ ${COMMAND} foo
		    $ ${COMMAND} --no-disown --force
_EOF_
}

pkgctl_aur_drop_from_repo() {
	# options
	local paths=()
	local DISOWN=1
	local FORCE=0

	# variables
	local path realpath pkgbase pkgrepo remote_url

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_aur_drop_from_repo_usage
				exit 0
				;;
			--no-disown)
				DISOWN=0
				shift
				;;
			-f|--force)
				FORCE=1
				shift
				;;
			--)
				shift
				break
				;;
			-*)
				die "Invalid argument: %s" "$1"
				;;
			*)
				paths=("$@")
				break
				;;
		esac
	done

	# check if invoked without any path from within a packaging repo
	if (( ${#paths[@]} == 0 )); then
		if [[ -f PKGBUILD ]]; then
			paths=(".")
		else
			pkgctl_aur_drop_from_repo_usage
			exit 1
		fi
	fi

	for path in "${paths[@]}"; do
		# resolve symlink for basename
		if ! realpath=$(realpath --canonicalize-existing -- "${path}"); then
			die "No such directory: ${path}"
		fi
		# skip paths that are not directories
		if [[ ! -d "${realpath}" ]]; then
			continue
		fi

		pkgbase=$(basename "${realpath}")
		pkgbase=${pkgbase%.git}

		if [[ ! -d "${realpath}/.git" ]]; then
			die "Not a Git repository: ${path}"
		fi

		pushd "${path}" >/dev/null

		if [[ ! -f PKGBUILD ]]; then
			die 'PKGBUILD not found in %s' "${path}"
		fi

		msg "Dropping ${pkgbase} to the AUR"

		remote_url="${AUR_URL_SSH}:${pkgbase}.git"
		if ! git remote add origin "${remote_url}" &>/dev/null; then
			git remote set-url origin "${remote_url}"
		fi

		# move the main branch to master
		if [[ $(git symbolic-ref --quiet --short HEAD) == main ]]; then
			git branch --move master
			git config branch.master.merge refs/heads/master
		fi

		# auto generate .SRCINFO if not already present
		if [[ -z "$(git ls-tree -r HEAD --name-only .SRCINFO)" ]]; then
			stat_busy 'Generating .SRCINFO'
			makepkg --printsrcinfo > .SRCINFO
			stat_done

			git add --force -- .SRCINFO
			git commit --quiet --message "Adding .SRCINFO" -- .SRCINFO
		fi

		msg "Pushing ${pkgbase} to the AUR"
		if (( FORCE )); then
			AUR_OVERWRITE=1 \
				GIT_SSH_COMMAND="ssh -o SendEnv=AUR_OVERWRITE" \
				git push --force --no-follow-tags origin master
		else
			git push --no-follow-tags origin master
		fi

		# update the local default branch in case this clone is used in the future
		git remote set-head origin master

		if (( DISOWN )); then
			msg "Disowning ${pkgbase} on the AUR"
			# shellcheck disable=SC2029
			ssh "${AUR_URL_SSH}" disown "${pkgbase}"
		fi

		# auto-detection of the repo to remove from
		if ! pkgrepo=$(get_pacman_repo_from_pkgbuild PKGBUILD); then
			die 'Failed to get pacman repo'
		fi

		msg "Deleting ${pkgbase} from the official repository"
		if [[ -z "${pkgrepo}" ]]; then
			warning 'Did not find %s in any repository, please delete manually' "${pkgbase}"
		else
			msg2 "  repo: ${pkgrepo}"
			pkgctl_db_remove --noconfirm "${pkgrepo}" "${pkgbase}"
		fi

		popd >/dev/null
	done
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_UTIL_PKGBUILD_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_UTIL_PKGBUILD_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/util/makepkg.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/makepkg.sh

source /usr/share/makepkg/util/message.sh
source /usr/share/makepkg/util/schema.sh

set -eo pipefail


# set the pkgver variable in a PKGBUILD
# assumes that the pkgbuild is sourced to detect the presence of a pkgver function
pkgbuild_set_pkgver() {
	local new_pkgver=$1
	local pkgver=${pkgver}

	pkgver="$(regex_escape "${pkgver}")"

	if [[ $(type -t pkgver) == function ]]; then
		# TODO: check if die or warn, if we provide _commit _gitcommit setter maybe?
		warning 'setting pkgver variable has no effect if the PKGBUILD has a pkgver() function'
	fi

	if ! grep --extended-regexp --quiet --max-count=1 "^pkgver=${pkgver}$" PKGBUILD; then
		die "Non-standard pkgver declaration"
	fi
	sed --regexp-extended "s|^(pkgver=)${pkgver}$|\1${new_pkgver}|g" --in-place PKGBUILD
}

# set the pkgrel variable in a PKGBUILD
# assumes that the pkgbuild is sourced so pkgrel is present
pkgbuild_set_pkgrel() {
	local new_pkgrel=$1
	local pkgrel=${pkgrel}

	if ! grep --extended-regexp --quiet --max-count=1 "^pkgrel=${pkgrel}$" PKGBUILD; then
		die "Non-standard pkgrel declaration"
	fi
	sed --regexp-extended "s|^(pkgrel=)${pkgrel}$|\1${new_pkgrel}|g" --in-place PKGBUILD
}

pkgbuild_update_checksums() {
	local status_file=$1
	local builddir newbuildfile sumtypes newsums

	[[ -z ${WORKDIR:-} ]] && setup_workdir

	builddir=$(mktemp --tmpdir="${WORKDIR}" --directory update-checksums.XXXXXX)
	newbuildfile="${builddir}/PKGBUILD"

	# generate new integrity checksums
	if ! newsums=$(BUILDDIR=${builddir} makepkg_generate_integrity 2>"${status_file}"); then
		printf 'Failed to generate new checksums'
		return 1
	fi

	# early exit if no integrity checksums are needed
	if [[ -z ${newsums} ]]; then
		return 0
	fi

	# replace the integrity sums and write it to a temporary file
	sumtypes=$(IFS='|'; echo "${known_hash_algos[*]}")
	if ! awk --assign=sumtypes="${sumtypes}" --assign=newsums="${newsums}" '
		$0 ~"^[[:blank:]]*(" sumtypes ")sums(_[^=]+)?\\+?=", $0 ~ "\\)[[:blank:]]*(#.*)?$" {
			if (!w) {
				print newsums
				w++
			}
			next
		}

		1
		END { if (!w) print newsums }' PKGBUILD > "${newbuildfile}"; then
		printf 'Failed to replace the generated checksums'
		return 1
	fi

	# overwrite the original PKGBUILD while preserving permissions
	if ! cat -- "${newbuildfile}" > PKGBUILD; then
		printf "Failed to write to the PKGBUILD file"
		return 1
	fi

	return 0
}

#!/bin/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
[[ -z ${DEVTOOLS_INCLUDE_VERSION_CHECK_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_VERSION_CHECK_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/util/term.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/term.sh
# shellcheck source=src/lib/valid-version.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/valid-version.sh

source /usr/share/makepkg/util/message.sh

set -eo pipefail

readonly PKGCTL_VERSION_CHECK_EXIT_UP_TO_DATE=0
export PKGCTL_VERSION_CHECK_EXIT_UP_TO_DATE
readonly PKGCTL_VERSION_CHECK_EXIT_OUT_OF_DATE=2
export PKGCTL_VERSION_CHECK_EXIT_OUT_OF_DATE
readonly PKGCTL_VERSION_CHECK_EXIT_FAILURE=3
export PKGCTL_VERSION_CHECK_EXIT_FAILURE


pkgctl_version_check_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [PKGBASE]...

		Compares the versions of packages in the local packaging repository against
		their latest upstream versions.

		Upon execution, it generates a grouped list that provides detailed insights
		into each package's status. For each package, it displays the current local
		version alongside the latest version available upstream.

		Outputs a summary of up-to-date packages, out-of-date packages, and any
		check failures.

		OPTIONS
		    -h, --help           Show this help text

		FILTER OPTIONS
		    -v, --verbose        Display all results including up-to-date versions

		OUTPUT OPTIONS
		    --json               Enable printing in JSON; Shorthand for '--format json'
		    -F, --format FORMAT  Controls the output format of the results;
		                         FORMAT is 'pretty', or 'json' (default: pretty)

		EXAMPLES
		    $ ${COMMAND} neovim vim
_EOF_
}

pkgctl_version_check() {
	local pkgbases=()
	local verbose=0
	local output_format=pretty

	local path status_file path pkgbase upstream_version result

	local json_data=()
	local up_to_date=()
	local out_of_date=()
	local failure=()
	local current_item=0
	local section_separator=''
	local exit_code=${PKGCTL_VERSION_CHECK_EXIT_UP_TO_DATE}

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_version_check_usage
				exit 0
				;;
			--json)
				output_format=json
				shift
				;;
			-F|--format)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				output_format="${2}"
				if ! in_array "${output_format}" "${DEVTOOLS_VALID_VERSION_OUTPUT_FORMAT[@]}"; then
					die "Unknown output format: %s" "${output_format}"
				fi
				shift 2
				;;
			-v|--verbose)
				verbose=1
				shift
				;;
			--)
				shift
				break
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				pkgbases=("$@")
				break
				;;
		esac
	done

	if ! command -v nvchecker &>/dev/null; then
		die "The \"$_DEVTOOLS_COMMAND\" command requires 'nvchecker'"
	fi

	# Check if used without pkgbases in a packaging directory
	if (( ${#pkgbases[@]} == 0 )); then
		if [[ -f PKGBUILD ]]; then
			pkgbases=(".")
		else
			pkgctl_version_check_usage
			exit 1
		fi
	fi

	# enable verbose mode when we only have a single item to check
	if (( ${#pkgbases[@]} == 1 )); then
		verbose=1
	fi

	if [[ ${output_format} == pretty ]]; then
		# start a terminal spinner as checking versions takes time
		status_dir=$(mktemp --tmpdir="${WORKDIR}" --directory pkgctl-version-check-spinner.XXXXXXXXXX)
		term_spinner_start "${status_dir}"
	fi

	for path in "${pkgbases[@]}"; do
		# skip paths that are not directories
		if [[ ! -d "${path}" ]]; then
			continue
		fi
		pushd "${path}" >/dev/null

		if [[ ${output_format} == pretty ]]; then
			# update the current terminal spinner status
			(( ++current_item ))
			pkgctl_version_check_spinner \
				"${status_dir}" \
				"${#up_to_date[@]}" \
				"${#out_of_date[@]}" \
				"${#failure[@]}" \
				"${current_item}" \
				"${#pkgbases[@]}"
		fi

		if [[ ! -f "PKGBUILD" ]]; then
			result="no PKGBUILD found"
			pkgbase=$(basename "${path}")
			json_data+=("$(build_json_package_version_entry "${pkgbase}" failure "${result}" false null null)")
			failure+=("${BOLD}${pkgbase}${ALL_OFF}: ${result}")
			popd >/dev/null
			continue
		fi

		# reset common PKGBUILD variables
		unset pkgbase pkgname arch source pkgver pkgrel validpgpkeys
		# shellcheck source=contrib/makepkg/PKGBUILD.proto
		. ./PKGBUILD
		pkgbase=${pkgbase:-$pkgname}

		if ! result=$(get_upstream_version); then
			json_data+=("$(build_json_package_version_entry "${pkgbase}" failure "${result}" false "${pkgver}" null)")
			failure+=("${BOLD}${pkgbase}${ALL_OFF}: ${result}")
			popd >/dev/null
			continue
		fi
		upstream_version=${result}

		if ! result=$(vercmp "${upstream_version}" "${pkgver}"); then
			result="failed to compare version ${upstream_version} against ${pkgver}"
			json_data+=("$(build_json_package_version_entry "${pkgbase}" failure "${result}" false "${pkgver}" "${upstream_version}")")
			result="${BOLD}${pkgbase}${ALL_OFF}: ${result}"
			failure+=("${result}")
			popd >/dev/null
			continue
		fi

		if (( result == 0 )); then
			if (( verbose )); then
				json_data+=("$(build_json_package_version_entry "${pkgbase}" success null false "${pkgver}" "${upstream_version}")")
			fi
			result="${BOLD}${pkgbase}${ALL_OFF}: current version ${PURPLE}${pkgver}${ALL_OFF} is latest"
			up_to_date+=("${result}")
		elif (( result < 0 )); then
			if (( verbose )); then
				json_data+=("$(build_json_package_version_entry "${pkgbase}" warning "local version is newer than upstream" false "${pkgver}" "${upstream_version}")")
			fi
			result="${BOLD}${pkgbase}${ALL_OFF}: current version ${PURPLE}${pkgver}${ALL_OFF} is newer than ${DARK_GREEN}${upstream_version}${ALL_OFF}"
			up_to_date+=("${result}")
		elif (( result > 0 )); then
			json_data+=("$(build_json_package_version_entry "${pkgbase}" success null true "${pkgver}" "${upstream_version}")")
			result="${BOLD}${pkgbase}${ALL_OFF}: upgrade from version ${PURPLE}${pkgver}${ALL_OFF} to ${DARK_GREEN}${upstream_version}${ALL_OFF}"
			out_of_date+=("${result}")
		fi

		popd >/dev/null
	done

	if [[ ${output_format} == pretty ]]; then
		# stop the terminal spinner after all checks
		term_spinner_stop "${status_dir}"
	fi

	# early exit for json output
	if [[ ${output_format} == json ]]; then
		jq --null-input \
			'$ARGS.positional' \
			--jsonargs "${json_data[@]}"
		return 0
	fi

	if (( verbose )) && (( ${#up_to_date[@]} > 0 )); then
		printf "%sUp-to-date%s\n" "${section_separator}${BOLD}${UNDERLINE}" "${ALL_OFF}"
		section_separator=$'\n'
		for result in "${up_to_date[@]}"; do
			msg_success " ${result}"
		done
	fi

	if (( ${#failure[@]} > 0 )); then
		exit_code=${PKGCTL_VERSION_CHECK_EXIT_FAILURE}
		printf "%sFailure%s\n" "${section_separator}${BOLD}${UNDERLINE}" "${ALL_OFF}"
		section_separator=$'\n'
		for result in "${failure[@]}"; do
			msg_error " ${result}"
		done
	fi

	if (( ${#out_of_date[@]} > 0 )); then
		exit_code=${PKGCTL_VERSION_CHECK_EXIT_OUT_OF_DATE}
		printf "%sOut-of-date%s\n" "${section_separator}${BOLD}${UNDERLINE}" "${ALL_OFF}"
		section_separator=$'\n'
		for result in "${out_of_date[@]}"; do
			msg_warn " ${result}"
		done
	fi

	# Show summary when processing multiple packages
	if (( ${#pkgbases[@]} > 1 )); then
		printf '%s' "${section_separator}"
		pkgctl_version_check_summary \
			"${#up_to_date[@]}" \
			"${#out_of_date[@]}" \
			"${#failure[@]}"
	fi

	# return status based on results
	return "${exit_code}"
}

build_json_package_version_entry() {
	local pkgbase=$1
	local status=$2
	local message=$3
	local is_out_of_date=$4
	local local_version=$5
	local upstream_version=$6

	jq --null-input \
		--arg pkgbase "${pkgbase}" \
		--arg status "${status}" \
		--arg message "${message}" \
		--arg local_version "${local_version}" \
		--arg upstream_version "${upstream_version}" \
		--argjson out_of_date "${is_out_of_date}" \
		'$ARGS.named | walk(if type == "string" and (. == "null" or . == "") then null else . end)'
}

get_upstream_version() {
	local config=.nvchecker.toml
	local output errors upstream_version
	local output
	local opts=()
	local keyfile="${XDG_CONFIG_HOME:-${HOME}/.config}/nvchecker/keyfile.toml"

	# check nvchecker config file
	if ! errors=$(nvchecker_check_config "${config}"); then
		printf "%s" "${errors}"
		return 1
	fi

	# populate keyfile to nvchecker opts
	if [[ -f ${keyfile} ]]; then
		opts+=(--keyfile "${keyfile}")
	fi

	if ! output=$(GIT_TERMINAL_PROMPT=0 GIT_CONFIG_GLOBAL=/dev/null GIT_CONFIG_SYSTEM=/dev/null \
			nvchecker --file "${config}" --logger json "${opts[@]}" 2>&1 | \
			jq --raw-output 'select((.level != "debug") and (.event != "ignoring invalid version"))'); then
		printf "failed to run nvchecker: %s" "${output}"
		return 1
	fi

	if ! errors=$(nvchecker_check_error "${output}"); then
		printf "%s" "${errors}"
		return 1
	fi

	if ! output=$(jq --raw-output --exit-status 'select(.name == "'"${pkgbase}"'")' <<< "${output}"); then
		printf "failed to select pkgbase result from output"
		return 1
	fi

	if ! upstream_version=$(jq --raw-output --exit-status '.version' <<< "${output}"); then
		printf "failed to select version from result"
		return 1
	fi

	# implement none source for packages like meta and virtual packages that are always up to date
	if [[ ${upstream_version} == None ]] && grep --quiet --extended-regexp '^source *= *"manual"' "${config}" &>/dev/null; then
		printf '%s' "${pkgver:-${upstream_version}}"
		return 0
	fi

	printf "%s" "${upstream_version}"
	return 0
}

nvchecker_check_config() {
	local config=$1

	local restricted_properties=(keyfile httptoken token)
	local property

	# check if the config file exists
	if [[ ! -f ${config} ]]; then
		printf "configuration file not found: %s" "${config}"
		return 1
	fi

	# check if config contains any restricted properties like secrets
	for property in "${restricted_properties[@]}"; do
		if grep --max-count=1 --quiet "^${property}" < "${config}"; then
			printf "restricted property in %s: %s" "${config}" "${property}"
			return 1
		fi
	done

	# check if the config contains a pkgbase section
	if [[ -n ${pkgbase} ]] && ! grep --max-count=1 --extended-regexp --quiet "^\\[\"?${pkgbase//+/\\+}\"?\\]" < "${config}"; then
		printf "missing pkgbase section in %s: %s" "${config}" "${pkgbase}"
		return 1
	fi

	# check if the config contains any section other than pkgbase
	if [[ -n ${pkgbase} ]] && property=$(grep --max-count=1 --perl-regexp "^\\[(?!\"?${pkgbase//+/\\+}(:.+)?\"?\\]).+\\]" < "${config}"); then
		printf "non-pkgbase section not supported in %s: %s" "${config}" "${property}"
		return 1
	fi

	# check if the config is using the 'cmd' source
	if grep --extended-regexp --quiet '^\s*source\s*=\s*["'\'']cmd["'\''].*' "${config}"; then
		printf "using the 'cmd' source in %s is disallowed" "${config}"
		return 1
	fi
}

nvchecker_check_error() {
	local result=$1
	local errors

	if ! errors=$(jq --raw-output --exit-status \
			'select((.level == "error") and (.error != null)) | "\(.event)" + if .error then ": \(.error)" else "" end' \
			<<< "${result}"); then
		return 0
	fi

	mapfile -t errors <<< "${errors}"
	printf "%s\n" "${errors[@]}"
	return 1
}

pkgctl_version_check_summary() {
	local up_to_date_count=$1
	local out_of_date_count=$2
	local failure_count=$3

	# print nothing if all stats are zero
	if (( up_to_date_count == 0 )) && \
			(( out_of_date_count == 0 )) && \
			(( failure_count == 0 )); then
		return 0
	fi

	# print summary for all none zero stats
	printf "%sSummary%s\n" "${BOLD}${UNDERLINE}" "${ALL_OFF}"
	if (( up_to_date_count > 0 )); then
		msg_success " Up-to-date: ${BOLD}${up_to_date_count}${ALL_OFF}" 2>&1
	fi
	if (( failure_count > 0 )); then
		msg_error " Failure: ${BOLD}${failure_count}${ALL_OFF}" 2>&1
	fi
	if (( out_of_date_count > 0 )); then
		msg_warn " Out-of-date: ${BOLD}${out_of_date_count}${ALL_OFF}" 2>&1
	fi
}

pkgctl_version_check_spinner() {
	local status_dir=$1
	local up_to_date_count=$2
	local out_of_date_count=$3
	local failure_count=$4
	local current=$5
	local total=$6

	local percentage=$(( 100 * current / total ))
	local tmp_file="${status_dir}/tmp"
	local status_file="${status_dir}/status"

	# print the current summary
	pkgctl_version_check_summary \
		"${up_to_date_count}" \
		"${out_of_date_count}" \
		"${failure_count}" > "${tmp_file}"

	# print the progress status
	printf "📡 Checking: %s/%s [%s] %%spinner%%" \
		"${BOLD}${current}" "${total}" "${percentage}%${ALL_OFF}"  \
		>> "${tmp_file}"

	# swap the status file
	mv "${tmp_file}" "${status_file}"
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later
:

# shellcheck disable=2034
DEVTOOLS_VALID_BINARY_ARCHES=(
	x86_64
)

# shellcheck disable=2034
DEVTOOLS_VALID_ARCHES=(
	"${DEVTOOLS_VALID_BINARY_ARCHES[@]}"
	any
)

# shellcheck disable=2034
DEVTOOLS_VALID_TAGS=(
	core-x86_64 core-any
	core-staging-x86_64 core-staging-any
	core-testing-x86_64 core-testing-any
	extra-x86_64 extra-any
	extra-staging-x86_64 extra-staging-any
	extra-testing-x86_64 extra-testing-any
	multilib-x86_64
	multilib-testing-x86_64
	multilib-staging-x86_64
	kde-unstable-x86_64 kde-unstable-any
	gnome-unstable-x86_64 gnome-unstable-any
)

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_UTIL_PACMAN_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_UTIL_PACMAN_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh

set -e


readonly _DEVTOOLS_PACMAN_CACHE_DIR=${XDG_CACHE_DIR:-$HOME/.cache}/devtools/pacman/db
readonly _DEVTOOLS_PACMAN_CONF_DIR=${_DEVTOOLS_LIBRARY_DIR}/pacman.conf.d
readonly _DEVTOOLS_MAKEPKG_CONF_DIR=${_DEVTOOLS_LIBRARY_DIR}/makepkg.conf.d


update_pacman_repo_cache() {
	local repo=${1:-multilib}

	mkdir -p "${_DEVTOOLS_PACMAN_CACHE_DIR}"
	msg "Updating pacman database cache"
	lock 10 "${_DEVTOOLS_PACMAN_CACHE_DIR}.lock" "Locking pacman database cache"
	fakeroot -- pacman --config "${_DEVTOOLS_PACMAN_CONF_DIR}/${repo}.conf" \
		--dbpath "${_DEVTOOLS_PACMAN_CACHE_DIR}" \
		--disable-sandbox \
		-Sy
	lock_close 10
}

get_pacman_repo_from_pkgbuild() {
	local path=${1:-PKGBUILD}
	local repo=${2:-multilib}
	local -a pkgnames

	# shellcheck source=contrib/makepkg/PKGBUILD.proto
	mapfile -t pkgnames < <(source "${path}"; printf "%s\n" "${pkgname[@]}")

	if (( ${#pkgnames[@]} == 0 )); then
		die 'Failed to get pkgname from %s' "${path}"
		return
	fi

	# update the pacman repo cache if it doesn't exist yet
	if [[ ! -d "${_DEVTOOLS_PACMAN_CACHE_DIR}" ]]; then
		update_pacman_repo_cache "${repo}"
	fi

	slock 10 "${_DEVTOOLS_PACMAN_CACHE_DIR}.lock" "Locking pacman database cache"
	# query repo of passed pkgname, specify --nodeps twice to skip all dependency checks
	mapfile -t repos < <(pacman --config "${_DEVTOOLS_PACMAN_CONF_DIR}/${repo}.conf" \
		--dbpath "${_DEVTOOLS_PACMAN_CACHE_DIR}" \
		--sync \
		--nodeps \
		--nodeps \
		--print \
		--print-format '%n %r' \
		"${pkgnames[0]}" 2>/dev/null | awk '$1=="'"${pkgnames[0]}"'"{print $2}'
	)
	lock_close 10

	printf "%s" "${repos[0]}"
}

get_pkgnames_from_repo_pkgbase() {
	local repo=$1
	shift
	local pkgbases=("$@")
	local -a pkgnames

	# update the pacman repo cache if it doesn't exist yet
	if [[ ! -d "${_DEVTOOLS_PACMAN_CACHE_DIR}" ]]; then
		update_pacman_repo_cache universe
	fi

	slock 10 "${_DEVTOOLS_PACMAN_CACHE_DIR}.lock" "Locking pacman database cache"
	# query pkgnames of passed pkgbase inside a repo
	mapfile -t pkgnames < <(expac --config <(sed "s|#DBPath.*|DBPath = $(realpath "${_DEVTOOLS_PACMAN_CACHE_DIR}")|" < "${_DEVTOOLS_PACMAN_CONF_DIR}/universe.conf") \
		--sync '%r %e %n' 2>/dev/null \
		| sort | awk -v pkgbase="${pkgbases[*]}" \
			'BEGIN { split(pkgbase, array); for (item in array) filter[array[item]]=1 } $1=="'"${repo}"'" && $2 in filter {print $3}'
	)
	lock_close 10

	if (( ! ${#pkgnames[@]} )); then
		return 1
	fi

	printf "%s\n" "${pkgnames[@]}"
	return 0
}

#!/bin/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_REPO_CLEAN_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_REPO_CLEAN_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh

source /usr/share/makepkg/util/message.sh

set -eo pipefail


pkgctl_repo_clean_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTION] [PATH]...

		Cleans the working tree by recursively removing files that are not under
		version control, starting from the current directory.

		Files unknown to Git as well as ignored files are removed. This can, for
		example, be useful to remove all build products.

		OPTIONS
		    -i, --interactive   Show what would be done and clean files interactively
		    -n, --dry-run       Don't remove anything, just show what would be done
		    -h, --help          Show this help text

		EXAMPLES
		    $ ${COMMAND} libfoo linux libbar
		    $ ${COMMAND} --interactive libfoo linux libbar
		    $ ${COMMAND} --dry-run *
_EOF_
}

pkgctl_repo_clean() {
	# options
	local git_clean_options=()
	local paths

	local path pkgbase

	while (( $# )); do
		case $1 in
			-i|--interactive)
				git_clean_options+=("$1")
				shift
				;;
			-n|--dry-run)
				git_clean_options+=("$1")
				shift
				;;
			-h|--help)
				pkgctl_repo_clean_usage
				exit 0
				;;
			--)
				shift
				break
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				paths=("$@")
				break
				;;
		esac
	done

	# check if invoked without any path from within a packaging repo
	if (( ${#paths[@]} == 0 )); then
		paths=(".")
	fi

	# print message about the work chunk
	printf "🗑️ Removing untracked files from %s working trees\n" "${BOLD}${#paths[@]}${ALL_OFF}"

	for path in "${paths[@]}"; do
		# skip paths that are not directories
		if [[ ! -d "${path}" ]]; then
			continue
		fi

		if [[ ! -f "${path}/PKGBUILD" ]]; then
			msg_error "Not a package repository: ${path}"
			continue
		fi

		if [[ ! -d "${path}/.git" ]]; then
			msg_error "Not a Git repository: ${path}"
			continue
		fi

		pkgbase=$(basename "$(realpath "${path}")")
		pkgbase=${pkgbase%.git}

		# run dry mode to see if git would clean any files
		if [[ ! $(git -C "${path}" clean -x -d --dry-run 2>&1) ]]; then
			continue
		fi

		# git clean untracked files
		msg_success "Cleaning ${BOLD}${pkgbase}${ALL_OFF}"
		if ! git -C "${path}" clean -x -d --force "${git_clean_options[@]}"; then
			msg_error "Failed to remove untracked files"
		fi
		echo
	done
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_CONFIG_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_CONFIG_SH=1

set -e

readonly XDG_DEVTOOLS_DIR="${XDG_CONFIG_HOME:-$HOME/.config}/devtools"
readonly XDG_DEVTOOLS_GITLAB_CONFIG="${XDG_DEVTOOLS_DIR}/gitlab.conf"

# default config variables
export GITLAB_TOKEN=""

load_devtools_config() {
	# temporary permission fixup
	if [[ -d "${XDG_DEVTOOLS_DIR}" ]]; then
		chmod 700 "${XDG_DEVTOOLS_DIR}"
	fi
	if [[ -f "${XDG_DEVTOOLS_GITLAB_CONFIG}" ]]; then
		chmod 600 "${XDG_DEVTOOLS_GITLAB_CONFIG}"
	fi
	if [[ -n "${DEVTOOLS_GITLAB_TOKEN}" ]]; then
		GITLAB_TOKEN="${DEVTOOLS_GITLAB_TOKEN}"
		return
	fi
	if [[ -f "${XDG_DEVTOOLS_GITLAB_CONFIG}" ]]; then
		GITLAB_TOKEN=$(grep GITLAB_TOKEN "${XDG_DEVTOOLS_GITLAB_CONFIG}"|cut -d= -f2|cut -d\" -f2)
		return
	fi
	GITLAB_TOKEN=""
}

save_devtools_config() {
	# temporary permission fixup
	if [[ -d "${XDG_DEVTOOLS_DIR}" ]]; then
		chmod 700 "${XDG_DEVTOOLS_DIR}"
	fi
	if [[ -f "${XDG_DEVTOOLS_GITLAB_CONFIG}" ]]; then
		chmod 600 "${XDG_DEVTOOLS_GITLAB_CONFIG}"
	fi
	(
		umask 0077
		mkdir -p "${XDG_DEVTOOLS_DIR}"
		printf 'GITLAB_TOKEN="%s"\n' "${GITLAB_TOKEN}" > "${XDG_DEVTOOLS_GITLAB_CONFIG}"
	)
}

#!/hint/bash
#
# This may be included with or without `set -euE`
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_AUTH_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_AUTH_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

set -e


pkgctl_auth_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [COMMAND] [OPTIONS]

		Authenticate with services like GitLab.

		COMMANDS
		    login    Authenticate with the GitLab instance
		    status   View authentication status

		OPTIONS
		    -h, --help     Show this help text

		EXAMPLES
		    $ ${COMMAND} login --gen-access-token
		    $ ${COMMAND} status
_EOF_
}

pkgctl_auth() {
	if (( $# < 1 )); then
		pkgctl_auth_usage
		exit 0
	fi

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_auth_usage
				exit 0
				;;
			login)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/auth/login.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/auth/login.sh
				pkgctl_auth_login "$@"
				exit 0
				;;
			status)
				_DEVTOOLS_COMMAND+=" $1"
				shift
				# shellcheck source=src/lib/auth/status.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/auth/status.sh
				pkgctl_auth_status "$@"
				exit 0
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				die "invalid command: %s" "$1"
				;;
		esac
	done
}

#!/bin/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
[[ -z ${DEVTOOLS_INCLUDE_VERSION_UPGRADE_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_VERSION_UPGRADE_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/version/check.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/version/check.sh
# shellcheck source=src/lib/util/pkgbuild.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/pkgbuild.sh
# shellcheck source=src/lib/util/term.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/term.sh

source /usr/share/makepkg/util/message.sh

set -e

pkgctl_version_upgrade_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [PKGBASE]...

		Streamlines the process of keeping PKGBUILD files up-to-date with the latest
		upstream versions.

		Upon execution, it automatically adjusts the PKGBUILD file, ensuring that the
		pkgver field is set to match the latest version available from the upstream
		source. In addition to updating the pkgver, this command also resets the pkgrel
		to 1 and updates checksums.

		Outputs a summary of upgraded packages, up-to-date packages, and any check
		failures.

		OPTIONS
		    --no-update-checksums  Disable computation and update of the checksums
		    -v, --verbose          Display results including up-to-date versions
		    -h, --help             Show this help text

		EXAMPLES
		    $ ${COMMAND} neovim vim
_EOF_
}

pkgctl_version_upgrade() {
	local path upstream_version result
	local pkgbases=()
	local verbose=0
	local exit_code=0
	local current_item=0
	local update_checksums=1

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_version_upgrade_usage
				exit 0
				;;
			--no-update-checksums)
				update_checksums=0
				shift
				;;
			-v|--verbose)
				verbose=1
				shift
				;;
			--)
				shift
				break
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				pkgbases=("$@")
				break
				;;
		esac
	done

	if ! command -v nvchecker &>/dev/null; then
		die "The \"$_DEVTOOLS_COMMAND\" command requires 'nvchecker'"
	fi

	# Check if used without pkgbases in a packaging directory
	if (( ${#pkgbases[@]} == 0 )); then
		if [[ -f PKGBUILD ]]; then
			pkgbases=(".")
		else
			pkgctl_version_upgrade_usage
			exit 1
		fi
	fi

	# enable verbose mode when we only have a single item to check
	if (( ${#pkgbases[@]} == 1 )); then
		verbose=1
	fi

	# start a terminal spinner as checking versions takes time
	status_dir=$(mktemp --tmpdir="${WORKDIR}" --directory pkgctl-version-check-spinner.XXXXXXXXXX)
	term_spinner_start "${status_dir}"

	for path in "${pkgbases[@]}"; do
		# skip paths that aren't directories
		if [[ ! -d "${path}" ]]; then
			continue
		fi
		pushd "${path}" >/dev/null

		(( ++current_item ))

		if [[ ! -f "PKGBUILD" ]]; then
			result="${BOLD}${path}${ALL_OFF}: no PKGBUILD found"
			failure+=("${result}")
			popd >/dev/null
			continue
		fi

		# reset common PKGBUILD variables
		unset pkgbase pkgname arch source pkgver pkgrel validpgpkeys
		# shellcheck source=contrib/makepkg/PKGBUILD.proto
		. ./PKGBUILD
		pkgbase=${pkgbase:-$pkgname}

		# update the current terminal spinner status
		pkgctl_version_upgrade_spinner \
			"${status_dir}" \
			"${#up_to_date[@]}" \
			"${#out_of_date[@]}" \
			"${#failure[@]}" \
			"${current_item}" \
			"${#pkgbases[@]}" \
			"${pkgbase}" \
			"query latest version"

		if ! result=$(get_upstream_version); then
			result="${BOLD}${pkgbase}${ALL_OFF}: ${result}"
			failure+=("${result}")
			popd >/dev/null
			continue
		fi
		upstream_version=${result}

		if ! result=$(vercmp "${upstream_version}" "${pkgver}"); then
			result="${BOLD}${pkgbase}${ALL_OFF}: failed to compare version ${upstream_version} against ${pkgver}"
			failure+=("${result}")
			popd >/dev/null
			continue
		fi

		if (( result == 0 )); then
			result="${BOLD}${pkgbase}${ALL_OFF}: current version ${PURPLE}${pkgver}${ALL_OFF} is latest"
			up_to_date+=("${result}")
		elif (( result < 0 )); then
			result="${BOLD}${pkgbase}${ALL_OFF}: current version ${PURPLE}${pkgver}${ALL_OFF} is newer than ${DARK_GREEN}${upstream_version}${ALL_OFF}"
			up_to_date+=("${result}")
		elif (( result > 0 )); then
			result="${BOLD}${pkgbase}${ALL_OFF}: upgraded from version ${PURPLE}${pkgver}${ALL_OFF} to ${DARK_GREEN}${upstream_version}${ALL_OFF}"
			out_of_date+=("${result}")

			# change the PKGBUILD
			pkgbuild_set_pkgver "${upstream_version}"
			pkgbuild_set_pkgrel 1

			# download sources and update the checksums
			if (( update_checksums )); then
				pkgctl_version_upgrade_spinner \
					"${status_dir}" \
					"${#up_to_date[@]}" \
					"${#out_of_date[@]}" \
					"${#failure[@]}" \
					"${current_item}" \
					"${#pkgbases[@]}" \
					"${pkgbase}" \
					"updating checksums"

				if ! result=$(pkgbuild_update_checksums /dev/null); then
					result="${BOLD}${pkgbase}${ALL_OFF}: failed to update checksums for version ${DARK_GREEN}${upstream_version}${ALL_OFF}"
					failure+=("${result}")
				fi
			fi
		fi

		popd >/dev/null
	done

	# stop the terminal spinner after all checks
	term_spinner_stop "${status_dir}"

	if (( verbose )) && (( ${#up_to_date[@]} > 0 )); then
		printf "%sUp-to-date%s\n" "${section_separator}${BOLD}${UNDERLINE}" "${ALL_OFF}"
		section_separator=$'\n'
		for result in "${up_to_date[@]}"; do
			msg_success " ${result}"
		done
	fi

	if (( ${#failure[@]} > 0 )); then
		exit_code=1
		printf "%sFailure%s\n" "${section_separator}${BOLD}${UNDERLINE}" "${ALL_OFF}"
		section_separator=$'\n'
		for result in "${failure[@]}"; do
			msg_error " ${result}"
		done
	fi

	if (( ${#out_of_date[@]} > 0 )); then
		printf "%sUpgraded%s\n" "${section_separator}${BOLD}${UNDERLINE}" "${ALL_OFF}"
		section_separator=$'\n'
		for result in "${out_of_date[@]}"; do
			msg_warn " ${result}"
		done
	fi

	# Show summary when processing multiple packages
	if (( ${#pkgbases[@]} > 1 )); then
		printf '%s' "${section_separator}"
		pkgctl_version_upgrade_summary \
			"${#up_to_date[@]}" \
			"${#out_of_date[@]}" \
			"${#failure[@]}"
	fi

	# return status based on results
	return "${exit_code}"
}

pkgctl_version_upgrade_summary() {
	local up_to_date_count=$1
	local out_of_date_count=$2
	local failure_count=$3

	# print nothing if all stats are zero
	if (( up_to_date_count == 0 )) && \
			(( out_of_date_count == 0 )) && \
			(( failure_count == 0 )); then
		return 0
	fi

	# print summary for all none zero stats
	printf "%sSummary%s\n" "${BOLD}${UNDERLINE}" "${ALL_OFF}"
	if (( up_to_date_count > 0 )); then
		msg_success " Up-to-date: ${BOLD}${up_to_date_count}${ALL_OFF}" 2>&1
	fi
	if (( failure_count > 0 )); then
		msg_error " Failure: ${BOLD}${failure_count}${ALL_OFF}" 2>&1
	fi
	if (( out_of_date_count > 0 )); then
		msg_warn " Upgraded: ${BOLD}${out_of_date_count}${ALL_OFF}" 2>&1
	fi
}

pkgctl_version_upgrade_spinner() {
	local status_dir=$1
	local up_to_date_count=$2
	local out_of_date_count=$3
	local failure_count=$4
	local current=$5
	local total=$6
	local pkgbase=$7
	local message=$8

	local percentage=$(( 100 * current / total ))
	local tmp_file="${status_dir}/tmp"
	local status_file="${status_dir}/status"

	# print the current summary
	pkgctl_version_upgrade_summary \
		"${up_to_date_count}" \
		"${out_of_date_count}" \
		"${failure_count}" > "${tmp_file}"

	# print the progress status
	printf "📡 %s: %s\n" \
		"${pkgbase}" "${BOLD}${message}${ALL_OFF}" >> "${tmp_file}"
	printf "⌛ Upgrading: %s/%s [%s] %%spinner%%" \
		"${BOLD}${current}" "${total}" "${percentage}%${ALL_OFF}" \
		>> "${tmp_file}"

	# swap the status file
	mv "${tmp_file}" "${status_file}"
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_ISSUE_EDIT_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_ISSUE_EDIT_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}

# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/api/gitlab.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/api/gitlab.sh
# shellcheck source=src/lib/util/term.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/term.sh

set -eo pipefail


pkgctl_issue_edit_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
	cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [IID]

		The pkgctl issue edit command is used to modify an existing issue in Arch Linux
		packaging projects. This command allows users to update the issue's title,
		description, and various attributes, ensuring that the issue information
		remains accurate and up-to-date. It also provides a streamlined facility
		for bug wranglers to categorize and prioritize issues efficiently.

		By default, the command operates within the current directory, but users can
		specify a different package base if needed.

		In case of a failed run, the command can automatically recover to ensure that
		the editing process is completed without losing any data.

		OPTIONS
		    -p, --package PKGBASE   Interact with <pkgbase> instead of the current directory
		    -t, --title TITLE       Use the provided title for the issue
		    -e, --edit              Edit the issue title and description using an editor
		    --recover               Automatically recover from a failed run
		    --confidentiality TYPE  Set the issue confidentiality
		    --priority PRIORITY     Set the priority label
		    --resolution REASON     Set the resolution label
		    --scope SCOPE           Set the scope label
		    --severity SEVERITY     Set the severity label
		    --status STATUS         Set the status label
		    -h, --help              Show this help text

		EXAMPLES
		    $ ${COMMAND} --package linux --title "some very informative title"
_EOF_
}

pkgctl_issue_edit() {
	if (( $# < 1 )); then
		pkgctl_issue_edit_usage
		exit 0
	fi

	local pkgbase=""
	local title=""
	local description=""
	local labels=()
	local confidential=""
	local msgfile=""
	local edit=0
	local recover=0

	local recovery_home=${XDG_DATA_HOME:-$HOME/.local/share}/devtools/recovery
	local recovery_file
	local issue_url
	local project_path
	local result
	local iid
	local message

	# option checking
	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_issue_edit_usage
				exit 0
				;;
			-p|--package)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				pkgbase=$2
				shift 2
				;;
			-t|--title)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				title=$2
				shift 2
				;;
			-e|--edit)
				edit=1
				shift
				;;
			--recover)
				recover=1
				shift
				;;
			--confidentiality)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! in_array "$2" "${DEVTOOLS_VALID_ISSUE_CONFIDENTIALITY[@]}"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				if [[ $2 == public ]]; then
					confidential=false
				else
					confidential=true
				fi
				shift 2
				;;
			--priority)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(priority_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				labels+=("$label")
				shift 2
				;;
			--resolution)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(resolution_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				labels+=("$label")
				shift 2
				;;
			--scope)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(scope_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				labels+=("$label")
				shift 2
				;;
			--severity)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(severity_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				labels+=("$label")
				shift 2
				;;
			--status)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if ! label="$(status_as_gitlab_label "$2")"; then
					die "invalid argument for %s: %s" "$1" "$2"
				fi
				labels+=("$label")
				shift 2
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				iid=$1
				shift
				;;
		esac
	done

	if [[ -z ${iid} ]]; then
		die "missing issue iid argument"
	fi

	if [[ -z ${pkgbase} ]]; then
		if ! [[ -f PKGBUILD ]]; then
			die "missing --package option or PKGBUILD in current directory"
		fi
		pkgbase=$(realpath --canonicalize-existing .)
	fi
	pkgbase=$(basename "${pkgbase}")
	project_path=$(gitlab_project_name_to_path "${pkgbase}")
	recovery_file="${recovery_home}/issue_edit_${pkgbase}.md"

	# load current issue data
	if ! result=$(gitlab_project_issue "${pkgbase}" "${iid}"); then
		die "Failed to query issue ${pkgbase} #${iid}"
	fi
	{ read -r current_title; read -r current_confidential; } < <(
		jq --raw-output ".title, .confidential" <<< "${result}"
	)
	current_description=$(jq --raw-output ".description" <<< "${result}")

	# check existence of recovery file
	if [[ -f ${recovery_file} ]]; then
		if (( ! recover )); then
			msg_warn "Recovery file already exists: ${recovery_file}"
			if prompt "${GREEN}${BOLD}?${ALL_OFF} Do you want to recover?"; then
				msgfile=${recovery_file}
				recover=1
				edit=1
			fi
		fi
	fi

	# assign data to msgfile
	if [[ -n ${msgfile} ]]; then
		# check existence of msgfile
		if [[ ! -f ${msgfile} ]]; then
			msg_error "File does not exist: ${msgfile}${ALL_OFF}"
			exit 1
		fi
	fi

	# spawn editor
	if (( edit )); then
		if [[ -z ${msgfile} ]]; then
			msgfile=$(mktemp --tmpdir="${WORKDIR}" pkgctl-issue-create.XXXXXXXXXX.md)
			if [[ -n ${title} ]]; then
				printf "# Title: %s\n\n" "${title}" >> "${msgfile}"
			else
				printf "# Title: %s\n\n" "${current_title}" >> "${msgfile}"
			fi
			printf "%s\n" "${current_description}" >> "${msgfile}"
		fi

		if [[ -n $VISUAL ]]; then
			editor=${VISUAL}
		elif [[ -n $EDITOR ]]; then
			editor=${EDITOR}
		else
			die "No usable editor found (tried \$VISUAL, \$EDITOR)."
		fi

		if ! ${editor} "${msgfile}"; then
			message=$(< "${msgfile}")
			pkgctl_issue_write_recovery_file "${pkgbase}" "${message}" "${recovery_file}" "${recover}"
			return 1
		fi
	fi

	# check if the file contains a title
	if [[ -n ${msgfile} ]]; then
		message=$(< "${msgfile}")
		description=${message}
		if [[ ${message} == "# Title: "* ]]; then
			title=$(head --lines 1 <<< "${message}")
			title=${title//# Title: /}
			description=$(tail --lines +2 <<< "${message}")
			if [[ ${description} == $'\n'* ]]; then
				description=$(tail --lines +3 <<< "${message}")
			fi
		fi
	fi

	# prepare changes
	data='{}'
	if [[ -n ${title} ]] && [[ ${title} != "${current_title}" ]]; then
		result=$(jq --null-input \
			--arg title "${title}" \
			'$ARGS.named')
		data=$(jq --slurp '.[0] * .[1]' <(echo "${data}") <(echo "${result}"))
	fi
	if [[ -n ${description} ]] && [[ ${description} != "${current_description}" ]]; then
		result=$(jq --null-input \
			--arg description "${description}" \
			'$ARGS.named')
		data=$(jq --slurp '.[0] * .[1]' <(echo "${data}") <(echo "${result}"))
	fi
	if [[ -n ${confidential} ]] && [[ ${confidential} != "${current_confidential}" ]]; then
		result=$(jq --null-input \
			--arg confidential "${confidential}" \
			'$ARGS.named')
		data=$(jq --slurp '.[0] * .[1]' <(echo "${data}") <(echo "${result}"))
	fi
	if (( ${#labels[@]} )); then
		result=$(jq --null-input \
			--arg add_labels "$(join_by , "${labels[@]}")" \
			'$ARGS.named')
		data=$(jq --slurp '.[0] * .[1]' <(echo "${data}") <(echo "${result}"))
	fi

	# edit the issue
	if ! result=$(gitlab_project_issue_edit "${pkgbase}" "${iid}" "${params}" "${data}"); then
		msg_error "Failed to edit issue ${BOLD}${pkgbase}${ALL_OFF} #${iid}"
		pkgctl_issue_write_recovery_file "${pkgbase}" "${message}" "${recovery_file}" "${recover}"
		exit 1
	fi

	# delete old recovery file if we succeeded
	if [[ -f ${recovery_file} ]]; then
		rm --force "${recovery_file}"
	fi

	issue_url="${GIT_PACKAGING_URL_HTTPS}/${project_path}/-/issues/${iid}"
	msg_success "Updated issue ${BOLD}#${iid}${ALL_OFF}"
	printf "%sView this issue on GitLab: %s%s\n" "${GRAY}" "${issue_url}" "${ALL_OFF}"
}

pkgctl_issue_write_recovery_file() {
	local pkgbase=$1
	local message=$2
	local recovery_file=$3

	if [[ -f ${recovery_file} ]]; then
		msg_warn "Recovery file already exists: ${recovery_file}"
		if ! prompt "${YELLOW}${BOLD}?${ALL_OFF} Are you sure you want to overwrite it?"; then
			return 1
		fi
	fi

	mkdir -p "$(dirname "${recovery_file}")"
	printf "%s\n" "${message}" > "${recovery_file}"

	printf "Created recovery file: %s\n" "${recovery_file}"
	return 0
}

#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_UTIL_MAKEPKG_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_UTIL_MAKEPKG_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/share/devtools}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/util/srcinfo.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/srcinfo.sh


set -e

makepkg_source_package() {
	if (( EUID != 0 )); then
		[[ -z ${WORKDIR:-} ]] && setup_workdir
		(
			export WORKDIR DEVTOOLS_INCLUDE_COMMON_SH SRCPKGDEST
			fakeroot -- bash -$- -c "source '${BASH_SOURCE[0]}' && ${FUNCNAME[0]}"
		)
		return
	fi
	(
		# shellcheck disable=SC2030 disable=SC2031
		export LIBMAKEPKG_LINT_PKGBUILD_SH=1
		lint_pkgbuild() { :; }

		# shellcheck disable=SC2030 disable=SC2031
		export LIBMAKEPKG_SRCINFO_SH=1
		write_srcinfo() { print_srcinfo; }

		# explicitly instruct makepkg to not sign the source package, even when
		# the BUILDENV array in makepkg.conf contains 'sign'
		set +e -- -F --source --nosign
		# shellcheck source=/usr/bin/makepkg
		source "$(command -v makepkg)"
	)
}

makepkg_generate_integrity() {
	if [[ -z ${DEVTOOLS_GENERATE_INTEGRITY} ]]; then
		[[ -z ${WORKDIR:-} ]] && setup_workdir
		export WORKDIR DEVTOOLS_INCLUDE_COMMON_SH
		bash -$- -c "DEVTOOLS_GENERATE_INTEGRITY=1; source '${BASH_SOURCE[0]}' && ${FUNCNAME[0]}"
		return
	fi
	(
		# shellcheck disable=SC2030 disable=SC2031
		export LIBMAKEPKG_LINT_PKGBUILD_SH=1
		lint_pkgbuild() { :; }

		set +e -- --geninteg
		# shellcheck source=/usr/bin/makepkg
		source "$(command -v makepkg)"
	)
}

# Load makepkg.conf once to make variables available
_makepkg_config_loaded=false
makepkg_load_config() {
	if $_makepkg_config_loaded; then
		return
	fi
	_makepkg_config_loaded=true
	source /usr/share/makepkg/util/config.sh
	# shellcheck disable=2119
	load_makepkg_config
}

---
title: "Depending on a development version"
vignette: >
  %\VignetteIndexEntry{Depending on a development version}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

Sometimes you need your package to depend on a development version of another package that isn't yet available on [CRAN](https://cran.r-project.org) or [Bioconductor](https://bioconductor.org/).
The `Remotes` field in `DESCRIPTION` lets you specify where to install such a dependency from.
**`Remotes` is not a standard `DESCRIPTION` field**.
It is not acceptable when releasing a package on CRAN.
The `Remotes` field is understood by devtools, [pak](https://pak.r-lib.org/), and related tools, and is meant to be a temporary measure during development.

This vignette covers the mechanics of the `Remotes` field.
For more context, see the [Nonstandard dependencies](https://r-pkgs.org/dependencies-in-practice.html#nonstandard-dependencies) section of the R Packages book.

## The `Remotes` field

You can mark any dependency listed in `Depends`, `Imports`, `Suggests`, or `Enhances` as being installed from a non-standard source by adding a package reference to `Remotes` in your `DESCRIPTION` file.

The general form is:

```
Remotes: [type::]<source>, [type2::]<source2>
```

Multiple remote dependencies are separated by commas, just like regular dependencies elsewhere in `DESCRIPTION`.

It is important to remember that you **must always declare the dependency in the usual way**, i.e. in `Depends`, `Imports`, `Suggests`, or `Enhances`.
The `Remotes` field only provides instructions about *where* to install the dependency from.
For example, note how rlang appears in both `Imports` and `Remotes`:

```
Package: mypackage
Title: What the Package Does (One Line, Title Case)
Version: 0.0.0.9000
Imports:
    rlang (>= 1.1.0.9000)
Remotes:
    r-lib/rlang
```

You can use `usethis::use_dev_package()` to add or update a development dependency.
It takes care of modifying both the `Imports` (or `Suggests`) and `Remotes` fields.

## GitHub

GitHub is the most commonly used source for development packages, and the default when no type prefix is specified:

```yaml
Remotes: r-lib/rlang
```

You can request a specific branch, tag, commit (SHA), or pull request:

```yaml
Remotes: r-lib/rlang@some-branch,
  r-lib/rlang@v1.0.0,
  r-lib/rlang@84be6207,
  r-lib/rlang#142
```

A `github::` prefix is accepted but not required:

```yaml
Remotes: github::r-lib/rlang
```

## Other sources

There are many other supported source types:

```yaml
# GitLab
Remotes: gitlab::user/repo

# Git (any host)
Remotes: git::https://github.com/r-lib/rlang.git

# Bioconductor
Remotes: bioc::SummarizedExperiment

# URL (package archive)
Remotes: url::https://example.com/package-0.1.0.tar.gz

# Local
Remotes: local::/path/to/package
```

See the [pak documentation on package sources](https://pak.r-lib.org/reference/pak_package_sources.html) for a complete list.

## CRAN submission

When you submit your package to CRAN, all of its dependencies must also be available on CRAN or Bioconductor.
You need to remove the `Remotes` field from your `DESCRIPTION` before submission.
This means having a `Remotes` field is a temporary development state: once the dependency you need is released to CRAN or Bioconductor, you should update your minimum version requirement and drop the `Remotes` entry.

# github_ functions are deprecated

    Code
      . <- github_pull(1)
    Condition
      Warning:
      `github_pull()` was deprecated in devtools 2.5.0.
      i Please use `remotes::github_pull()` instead.
    Code
      . <- github_release()
    Condition
      Warning:
      `github_release()` was deprecated in devtools 2.5.0.
      i Please use `remotes::github_release()` instead.

# useful errors if too few or too many

    Code
      build_readme(pkg)
    Condition
      Error in `build_readme()`:
      ! Can't find 'README.qmd' or 'README.Rmd', at the top-level or below 'inst/'.

---

    Code
      build_readme(pkg)
    Condition
      Error in `build_readme()`:
      ! Found multiple executable READMEs: 'README.Rmd' and 'inst/README.Rmd'. There can only be one.

# errors if both README.qmd and README.Rmd exist

    Code
      build_readme(pkg)
    Condition
      Error in `build_readme()`:
      ! Found multiple executable READMEs: 'README.qmd' and 'README.Rmd'. There can only be one.

# build_readme() aborts when a dep is missing

    Code
      build_readme(pkg)
    Condition
      Error in `build_readme()`:
      ! 1 {PACKAGE} dependency is not installed.
      i Install it with `pak::local_install_dev_deps()`.
        missingpkg (missing)

# build_rmd() is deprecated

    Code
      suppressMessages(build_rmd("README.Rmd", path = pkg, quiet = TRUE))
    Condition
      Warning:
      `build_rmd()` was deprecated in devtools 2.5.0.
      i Please use `build_readme()` instead.

# fails when NEWS is missing or improperly formatted

    Code
      show_news(pkg)
    Condition
      Error in `find_news()`:
      ! Failed to find NEWS file.

---

    Code
      show_news(pkg)
    Condition
      Error in `show_news()`:
      ! Failed to parse NEWS file.
      Caused by error:
      ! NEWS.Rd: Sections \title, and \name must exist and be unique in Rd files

# load_all() errors if called recursively

    Code
      load_all()
    Condition
      Error in `load_all()`:
      ! Recursive loading detected.
      i Did you accidentally include `load_all()` in an R source file?

# can determine when to document

    Code
      result <- can_document("1.0.0", installed = "2.0.0")
    Output
      == Documenting =================================================================
    Message
      i Installed roxygen2 version (2.0.0) doesn't match declared (1.0.0)
      x `check()` will not re-document this package.
      i Do you need to re-run `document()`?

# check_for_rstudio_updates

    Code
      writeLines(check_for_rstudio_updates("windows", "haha-no-wut", TRUE))
    Output
      Unable to check for RStudio updates (you're using haha-no-wut).

---

    Code
      writeLines(check_for_rstudio_updates("darwin", "0.0.1", TRUE))
    Output
      RStudio {VERSION} is now available (you're using 0.0.1).
      Download at: https://www.rstudio.com/products/rstudio/download/

---

    Code
      writeLines(check_for_rstudio_updates("windows", "1.4.1717", TRUE))
    Output
      RStudio {VERSION} is now available (you're using 1.4.1717).
      Download at: https://www.rstudio.com/products/rstudio/download/

---

    Code
      writeLines(check_for_rstudio_updates("darwin", "2021.09.1+372", TRUE))
    Output
      RStudio {VERSION} is now available (you're using 2021.09.1+372).
      Download at: https://www.rstudio.com/products/rstudio/download/

---

    Code
      writeLines(check_for_rstudio_updates("windows", "2021.09.0-daily+328", TRUE))
    Output
      RStudio {VERSION} is now available (you're using 2021.09.0-daily+328).
      Download at: https://www.rstudio.com/products/rstudio/download/

# build_vignettes() and clean_vignettes() are deprecated

    Code
      build_vignettes(pkg)
    Condition
      Warning:
      `build_vignettes()` was deprecated in devtools 2.5.0.
    Code
      clean_vignettes(pkg)
    Condition
      Warning:
      `clean_vignettes()` was deprecated in devtools 2.5.0.

# gist with multiple files uses first with warning

    Code
      source_gist("605a984e764f9ed358556b4ce48cbd08", sha1 = "f176f5e1fe0", local = environment())
    Message
      2 .R files in gist, using first
      i Sourcing gist "605a984e764f9ed358556b4ce48cbd08"

# errors with bad id

    Code
      source_gist("xxxx")
    Condition
      Error in `source_gist()`:
      ! Invalid gist id specification "xxxx"

# error if file doesn't exist or no files

    Code
      find_gist("605a984e764f9ed358556b4ce48cbd08", 1)
    Condition
      Error:
      ! `filename` must be `NULL`, or a single filename ending in .R/.r
    Code
      find_gist("605a984e764f9ed358556b4ce48cbd08", "c.r")
    Condition
      Error:
      ! 'c.r' not found in gist
    Code
      find_gist("c535eee2d02e5f47c8e7642811bc327c")
    Condition
      Error:
      ! No R files found in gist

# check_sha1() checks or reports sha1 as needed

    Code
      check_sha1(path, NULL)
    Message
      i SHA-1 hash of file is "9f7efafc467018e11a7efc4bb7089ff0e5bff371"
    Code
      check_sha1(path, "f")
    Condition
      Error in `check_sha1()`:
      ! `sha1` must be at least 6 characters, not 1.
    Code
      check_sha1(path, "ffffff")
    Condition
      Error in `check_sha1()`:
      ! `sha1` ("ffffff") doesn't match SHA-1 hash of downloaded file ("9f7efa")

# package_file() gives useful errors

    Code
      package_file(path = 1)
    Condition
      Error in `package_file()`:
      ! `path` must be a single string, not the number 1.
    Code
      package_file(path = "doesntexist")
    Condition
      Error in `package_file()`:
      ! 'doesntexist' is not a directory.
    Code
      package_file(path = "/")
    Condition
      Error in `package_file()`:
      ! Could not find package root.
      i Is '/' inside a package?

# create argument is deprecated

    Code
      x <- as.package(path, create = TRUE)
    Condition
      Warning:
      The `create` argument of `as.package()` is deprecated as of devtools 2.5.0.

# build_manual() shows stderr on failure

    Code
      build_manual(pkg)
    Message
      Saving manual to '<pkgdir>_0.0.0.9000.pdf'
    Condition
      Error in `build_manual()`:
      ! Failed to build manual.
      ! LaTeX Error: File `inconsolata.sty' not found.
      Caused by error in `rd2pdf()`:
      ! System command 'R' failed

# check_doc_fields output - missing fields

    Code
      check_doc_fields(pkg)
    Message
      ! Missing `\value` section in 1 file:
      * man/foo.Rd
      ! Missing `\examples` section in 1 file:
      * man/bar.Rd

# check_doc_fields output - all present

    Code
      check_doc_fields(pkg)
    Message
      v All Rd files have a `\value` section.
      v All Rd files have a `\examples` section.

# check_man works

    Code
      check_man(pkg)
    Message
      i Updating {package} documentation
      i Loading {package}
      i Checking documentation...
    Output
      Rd files without \alias:
        'foo.Rd'
      Rd files without \description:
        'foo.Rd'
      Argument items with no description in Rd file 'foo.Rd':
        'foo'
      Undocumented arguments in Rd file 'foo.Rd'
        'x'
      

# test_coverage_active_file() computes coverage

    Code
      test_coverage_active_file(file.path(pkg, "R", "math.R"), report = "zero")
    Message
      Uncovered lines in 'R/math.R':
      * `compute()`: 4-5
      * `multiply()`: 2

# test_coverage_active_file() reports full coverage

    Code
      test_coverage_active_file(file.path(pkg, "R", "math.R"), report = "zero")
    Message
      v All lines covered!

# report_default() does its job

    Code
      report_default("bad")
    Condition
      Error:
      ! `report` must be one of "silent", "zero", or "html", not "bad".

# release() is deprecated

    Code
      . <- release()
    Condition
      Warning:
      `release()` was deprecated in devtools 2.5.0.
      i Please use `usethis::use_release_issue()` instead.
      Error in `release()`:
      ! Interactive session required.

# find_active_file() gives useful error if no RStudio

    Code
      find_active_file()
    Condition
      Error:
      ! `file` is absent but must be supplied.

# fails if can't find tests

    Code
      find_test_file("R/foo.blah")
    Condition
      Error:
      ! Don't know how to find tests associated with the active file 'foo.blah'
    Code
      find_test_file("R/foo.R")
    Condition
      Error:
      ! No test files found

# change_maintainer_email checks fields

    Code
      change_maintainer_email(path, "x@example.com")
    Condition
      Error:
      ! DESCRIPTION must use Authors@R field when changing `email`

---

    Code
      change_maintainer_email(path, "x@example.com")
    Condition
      Error:
      ! DESCRIPTION can't use Maintainer field when changing `email`

# email confirmation gives useful advice

    Code
      confirm_maintainer_email("hadley@posit.co")
    Message
      Email results to hadley@posit.co?
    Condition
      Error:
      ! User declined upload.
      i Use `email = "your email"` to override.

## Test environments
* local OS X install, R 3.3.1
* ubuntu 12.04 (on travis-ci), R 3.3.1
* win-builder (devel and release)

## R CMD check results

0 errors | 0 warnings | 1 note

* This is a new release.

## Reverse dependencies

This is a new release, so there are no reverse dependencies.

---

* I have run R CMD check on the NUMBER downstream dependencies.
  (Summary at ...). 
  
* FAILURE SUMMARY

* All revdep maintainers were notified of the release on RELEASE DATE.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package-deps.r
\name{check_dep_version}
\alias{check_dep_version}
\title{Check that the version of an imported package satisfies the requirements}
\usage{
check_dep_version(dep_name, dep_ver = NA, dep_compare = NA)
}
\arguments{
\item{dep_name}{The name of the package with objects to import}

\item{dep_ver}{The version of the package}

\item{dep_compare}{The comparison operator to use to check the version}
}
\description{
Check that the version of an imported package satisfies the requirements
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check.R
\name{check}
\alias{check}
\alias{check_built}
\title{Build and check a package}
\usage{
check(
  pkg = ".",
  document = NULL,
  build_args = NULL,
  ...,
  manual = FALSE,
  cran = TRUE,
  remote = FALSE,
  incoming = remote,
  force_suggests = FALSE,
  run_dont_test = FALSE,
  args = "--timings",
  env_vars = c(NOT_CRAN = "true"),
  quiet = FALSE,
  check_dir = NULL,
  cleanup = deprecated(),
  vignettes = TRUE,
  error_on = c("never", "error", "warning", "note")
)

check_built(
  path = NULL,
  cran = TRUE,
  remote = FALSE,
  incoming = remote,
  force_suggests = FALSE,
  run_dont_test = FALSE,
  manual = FALSE,
  args = "--timings",
  env_vars = NULL,
  check_dir = tempdir(),
  quiet = FALSE,
  error_on = c("never", "error", "warning", "note")
)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{document}{By default (\code{NULL}) will document if your installed
roxygen2 version matches the version declared in the \code{DESCRIPTION}
file. Use \code{TRUE} or \code{FALSE} to override the default.}

\item{build_args}{Additional arguments passed to \verb{R CMD build}}

\item{...}{Additional arguments passed on to \code{\link[pkgbuild:build]{pkgbuild::build()}}.}

\item{manual}{If \code{FALSE}, don't build and check manual (\code{--no-manual}).}

\item{cran}{if \code{TRUE} (the default), check using the same settings as CRAN
uses. Because this is a moving target and is not uniform across all of
CRAN's machine, this is on a "best effort" basis. It is more complicated
than simply setting \code{--as-cran}.}

\item{remote}{Sets \verb{_R_CHECK_CRAN_INCOMING_REMOTE_} env var. If \code{TRUE},
performs a number of CRAN incoming checks that require remote access.}

\item{incoming}{Sets \verb{_R_CHECK_CRAN_INCOMING_} env var. If \code{TRUE}, performs a
number of CRAN incoming checks.}

\item{force_suggests}{Sets \verb{_R_CHECK_FORCE_SUGGESTS_}. If \code{FALSE} (the
default), check will proceed even if all suggested packages aren't found.}

\item{run_dont_test}{Sets \code{--run-donttest} so that examples surrounded in
\verb{\donttest{}} are also run. When \code{cran = TRUE}, this only affects R 3.6 and
earlier; in R 4.0, code in \verb{\donttest{}} is always run as part of CRAN
submission.}

\item{args}{Character vector of arguments to pass to \verb{R CMD check}. Pass each
argument as a single element of this character vector (do not use spaces to
delimit arguments like you would in the shell). For example, to skip
running of examples and tests, use \code{args = c("--no-examples", "--no-tests")} and not \code{args = "--no-examples --no-tests"}. (Note that
instead of the \code{--output} option you should use the \code{check_dir} argument,
because  \code{--output} cannot deal with spaces and other special characters on
Windows.)}

\item{env_vars}{Environment variables set during \verb{R CMD check}}

\item{quiet}{if \code{TRUE} suppresses output from this function.}

\item{check_dir}{Path to a directory where the check is performed.
If this is not \code{NULL}, then the a temporary directory is used, that
is cleaned up when the returned object is garbage collected.}

\item{cleanup}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} See \code{check_dir} for details.}

\item{vignettes}{If \code{FALSE}, do not build or check vignettes, equivalent to
using \code{args = '--ignore-vignettes'} and \code{build_args = '--no-build-vignettes'}.}

\item{error_on}{Whether to throw an error on \verb{R CMD check} failures. Note
that the check is always completed (unless a timeout happens), and the
error is only thrown after completion.

\code{error_on} is passed through to \code{\link[rcmdcheck:rcmdcheck]{rcmdcheck::rcmdcheck()}}, which is the
definitive source for what the different values mean. If not specified by
the user, both \code{check()} and \code{check_built()} default to \code{error_on = "never"} in interactive use and \code{"warning"} in a non-interactive setting.}

\item{path}{Path to built package.}
}
\value{
An object containing errors, warnings, notes, and more.
}
\description{
\code{check()} automatically builds and checks a source package, using all known
best practices. \code{check_built()} checks an already-built package.

Passing \verb{R CMD check} is essential if you want to submit your package to
CRAN: you must not have any ERRORs or WARNINGs, and you want to ensure that
there are as few NOTEs as possible.  If you are not submitting to CRAN, at
least ensure that there are no ERRORs or WARNINGs: these typically represent
serious problems.

\code{check()} automatically builds a package before calling \code{check_built()}, as
this is the recommended way to check packages.  Note that this process runs
in an independent R session, so nothing in your current workspace will affect
the process. Under-the-hood, \code{check()} and \code{check_built()} rely on
\code{\link[pkgbuild:build]{pkgbuild::build()}} and \code{\link[rcmdcheck:rcmdcheck]{rcmdcheck::rcmdcheck()}}.
}
\section{Environment variables}{


Devtools does its best to set up an environment that combines best practices
with how check works on CRAN. This includes:
\itemize{
\item The standard environment variables set by devtools:
\code{\link[=r_env_vars]{r_env_vars()}}. Of particular note for package tests is the \code{NOT_CRAN} env
var, which lets you know that your tests are running somewhere other than
CRAN, and hence can take a reasonable amount of time.
\item Debugging flags for the compiler, set by
\code{\link[pkgbuild:compiler_flags]{compiler_flags(FALSE)}}.
\item If \code{aspell} is found, \verb{_R_CHECK_CRAN_INCOMING_USE_ASPELL_}
is set to \code{TRUE}. If no spell checker is installed, a warning is issued.
\item Environment variables, controlled by arguments \code{incoming}, \code{remote} and
\code{force_suggests}.
}
}

\seealso{
\code{\link[=release]{release()}} if you want to send the checked package to
CRAN.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build-site.R
\name{build_site}
\alias{build_site}
\title{Run \code{pkgdown::build_site()}}
\usage{
build_site(path = ".", ...)
}
\arguments{
\item{path}{Path to the package to build the static HTML.}

\item{...}{Additional arguments passed to \code{\link[pkgdown:build_site]{pkgdown::build_site()}}.}
}
\description{
This is a thin wrapper around \code{\link[pkgdown:build_site]{pkgdown::build_site()}}, used for generating
static HTML documentation. Learn more at \url{https://pkgdown.r-lib.org}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/R.r
\name{RCMD}
\alias{RCMD}
\title{Run R CMD xxx from within R}
\usage{
RCMD(cmd, options, path = tempdir(), env_vars = character(), ...)
}
\arguments{
\item{cmd}{one of the R tools available from the R CMD interface.}

\item{options}{a character vector of options to pass to the command}

\item{path}{the directory to run the command in.}

\item{env_vars}{environment variables to set before running the command.}

\item{...}{additional arguments passed to \code{\link{system_check}}}
}
\value{
\code{TRUE} if the command succeeds, throws an error if the command
fails.
}
\description{
Run R CMD xxx from within R
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build-github-devtools.r
\name{build_github_devtools}
\alias{build_github_devtools}
\title{Build the development version of devtools from GitHub.}
\usage{
build_github_devtools(outfile = NULL)
}
\arguments{
\item{outfile}{The name of the output file. If NULL (the default), it uses
./devtools.tgz (Mac and Linux), or ./devtools.zip (Windows).}
}
\value{
a string giving the location (including file name) of the built
 package
}
\description{
This function is especially useful for Windows users who want to upgrade
their version of devtools to the development version hosted on on GitHub.
In Windows, it's not possible to upgrade devtools while the package is loaded
because there is an open DLL, which in Windows can't be overwritten. This
function allows you to build a binary package of the development version of
devtools; then you can restart R (so that devtools isn't loaded) and install
the package.
}
\details{
Mac and Linux users don't need this function; they can use
\code{\link{install_github}} to install devtools directly, without going
through the separate build-restart-install steps.

This function requires a working development environment. On Windows, it
needs \url{https://cran.r-project.org/bin/windows/Rtools/}.
}
\examples{
\dontrun{
library(devtools)
build_github_devtools()

#### Restart R before continuing ####
install.packages("./devtools.zip", repos = NULL)

# Remove the package after installation
unlink("./devtools.zip")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test.R
\name{devtools-defunct}
\alias{devtools-defunct}
\alias{test_file}
\alias{test_coverage_file}
\title{Defunct functions}
\usage{
test_file(file = find_active_file(), ...)

test_coverage_file(file = find_active_file(), ...)
}
\description{
These functions are defunct and will be removed in a future version of
devtools.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install-github.r
\name{github_pull}
\alias{github_pull}
\alias{github_release}
\title{GitHub references}
\usage{
github_pull(pull)

github_release()
}
\arguments{
\item{pull}{The pull request to install}
}
\description{
Use as \code{ref} parameter to \code{\link{install_github}}.
Allows installing a specific pull request or the latest release.
}
\seealso{
\code{\link{install_github}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sitrep.R
\name{dev_sitrep}
\alias{dev_sitrep}
\title{Report package development situation}
\usage{
dev_sitrep(pkg = ".", debug = FALSE)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{debug}{If \code{TRUE}, will print out extra information useful for
debugging. If \code{FALSE}, it will use result cached from a previous run.}
}
\value{
A named list, with S3 class \code{dev_sitrep} (for printing purposes).
}
\description{
Call this function if things seem weird and you're not sure
what's wrong or how to fix it. It reports:
\itemize{
\item If R is up to date.
\item If RStudio or Positron is up to date.
\item If compiler build tools are installed and available for use.
\item If devtools and its dependencies are up to date.
\item If the package's dependencies are up to date.
}
}
\examples{
\dontrun{
dev_sitrep()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{devtools-deprecated}
\alias{devtools-deprecated}
\title{Deprecated Functions}
\description{
These functions are Deprecated in this release of devtools, they will be
marked as Defunct and removed in a future version.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.r
\name{is.package}
\alias{is.package}
\title{Is the object a package?}
\usage{
is.package(x)
}
\description{
Is the object a package?
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run-examples.R
\name{run_examples}
\alias{run_examples}
\title{Run all examples in a package}
\usage{
run_examples(
  pkg = ".",
  start = NULL,
  show = deprecated(),
  run_donttest = FALSE,
  run_dontrun = FALSE,
  fresh = FALSE,
  document = TRUE,
  run = deprecated(),
  test = deprecated()
)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{start}{Where to start running the examples: this can either be the
name of \code{Rd} file to start with (with or without extensions), or
a topic name. If omitted, will start with the (lexicographically) first
file. This is useful if you have a lot of examples and don't want to
rerun them every time you fix a problem. To run only one example, use
\code{\link[pkgload:dev_example]{pkgload::run_example()}}.}

\item{show}{DEPRECATED.}

\item{run_donttest}{if \code{TRUE}, do run \verb{\\donttest} sections in the Rd files.}

\item{run_dontrun}{if \code{TRUE}, do run \verb{\\dontrun} sections in the Rd files.}

\item{fresh}{if \code{TRUE}, will be run in a fresh R session. This has
the advantage that there's no way the examples can depend on anything in
the current session, but interactive code (like \code{\link[=browser]{browser()}})
won't work.}

\item{document}{if \code{TRUE}, \code{\link[=document]{document()}} will be run to ensure
examples are updated before running them.}

\item{run, test}{Deprecated, see \code{run_dontrun} and \code{run_donttest} above.}
}
\description{
One of the most frustrating parts of \verb{R CMD check} is getting all of your
examples to pass - whenever one fails you need to fix the problem and then
restart the whole process.  This function makes it a little easier by
making it possible to run all examples from an R function.
}
\seealso{
\code{\link[pkgload:dev_example]{pkgload::run_example()}} to run a single example.
}
\concept{example functions}
\keyword{programming}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/release.R
\name{release}
\alias{release}
\title{Release package to CRAN}
\usage{
release(pkg = ".", check = FALSE, args = NULL)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{check}{if \code{TRUE}, run checking, otherwise omit it.  This
is useful if you've just checked your package and you're ready to
release it.}

\item{args}{An optional character vector of additional command
line arguments to be passed to \verb{R CMD build}.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{release()} is deprecated in favour of \code{\link[usethis:use_release_issue]{usethis::use_release_issue()}}.
We no longer feel confident recommending \code{release()} because we don't use it
ourselves, so there's no guarantee that it will track best practices as
they evolve over time.

If you want to programmatical submit to CRAN, you can continue to use
\code{\link[=submit_cran]{submit_cran()}}.
}
\seealso{
\code{\link[usethis:use_release_issue]{usethis::use_release_issue()}} to create a checklist of release
tasks that you can use in addition to or in place of \code{release}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-doc.R
\name{check_man}
\alias{check_man}
\title{Check documentation, as \verb{R CMD check} does.}
\usage{
check_man(pkg = ".")
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}
}
\value{
Nothing. This function is called purely for it's side effects: if
}
\description{
This function attempts to run the documentation related checks in the
same way that \verb{R CMD check} does. Unfortunately it can't run them
all because some tests require the package to be loaded, and the way
they attempt to load the code conflicts with how devtools does it.
}
\examples{
\dontrun{
check_man("mypkg")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infrastructure-git.R
\name{use_github}
\alias{use_github}
\title{Connect a local repo with GitHub.}
\usage{
use_github(auth_token = github_pat(), private = FALSE, pkg = ".",
  host = "https://api.github.com", protocol = c("ssh", "https"),
  credentials = NULL)
}
\arguments{
\item{auth_token}{Provide a personal access token (PAT) from
\url{https://github.com/settings/tokens}. Defaults to the \code{GITHUB_PAT}
environment variable.}

\item{private}{If \code{TRUE}, creates a private repository.}

\item{pkg}{Path to package. See \code{\link{as.package}} for more
information.}

\item{host}{GitHub API host to use. Override with the endpoint-root for your
GitHub enterprise instance, for example,
"https://github.hostname.com/api/v3".}

\item{protocol}{transfer protocol, either "ssh" (the default) or "https"}

\item{credentials}{A \code{\link[git2r]{cred_ssh_key}} specifying specific
ssh credentials or NULL for default ssh key and ssh-agent behaviour.
Default is NULL.}
}
\description{
If the current repo does not use git, calls \code{\link{use_git}}
automatically. \code{\link{use_github_links}} is called to populate the
\code{URL} and \code{BugReports} fields of DESCRIPTION.
}
\section{Authentication}{


  A new GitHub repo will be created via the GitHub API, therefore you must
  provide a GitHub personal access token (PAT) via the argument
  \code{auth_token}, which defaults to the value of the \code{GITHUB_PAT}
  environment variable. Obtain a PAT from
  \url{https://github.com/settings/tokens}. The "repo" scope is required
  which is one of the default scopes for a new PAT.

  The argument \code{protocol} reflects how you wish to authenticate with
  GitHub for this repo in the long run. For either \code{protocol}, a remote
  named "origin" is created, an initial push is made using the specified
  \code{protocol}, and a remote tracking branch is set. The URL of the
  "origin" remote has the form \code{git@github.com:<USERNAME>/<REPO>.git}
  (\code{protocol = "ssh"}, the default) or
  \code{https://github.com/<USERNAME>/<REPO>.git} (\code{protocol =
  "https"}). For \code{protocol = "ssh"}, it is assumed that public and
  private keys are in the default locations, \code{~/.ssh/id_rsa.pub} and
  \code{~/.ssh/id_rsa}, respectively, and that \code{ssh-agent} is configured
  to manage any associated passphrase.  Alternatively, specify a
  \code{\link[git2r]{cred_ssh_key}} object via the \code{credentials}
  parameter.
}

\examples{
\dontrun{
## to use default ssh protocol
create("testpkg")
use_github(pkg = "testpkg")

## or use https
create("testpkg2")
use_github(pkg = "testpkg2", protocol = "https")
}
}
\seealso{
Other git infrastructure: \code{\link{use_git_hook}},
  \code{\link{use_github_links}}, \code{\link{use_git}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vignettes.R
\name{clean_vignettes}
\alias{clean_vignettes}
\title{Clean built vignettes}
\usage{
clean_vignettes(pkg = ".")
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{clean_vignettes()} is deprecated because \code{\link[=build_vignettes]{build_vignettes()}} is deprecated.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infrastructure.R
\name{use_readme_rmd}
\alias{use_readme_rmd}
\alias{use_readme_md}
\title{Create README files.}
\usage{
use_readme_rmd(pkg = ".")

use_readme_md(pkg = ".")
}
\arguments{
\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information}
}
\description{
Creates skeleton README files with sections for
\itemize{
\item a high-level description of the package and its goals
\item R code to install from GitHub, if GitHub usage detected
\item a basic example
}
Use \code{Rmd} if you want a rich intermingling of code and data. Use
\code{md} for a basic README. \code{README.Rmd} will be automatically
added to \code{.Rbuildignore}. The resulting README is populated with default
YAML frontmatter and R fenced code blocks (\code{md}) or chunks (\code{Rmd}).
}
\examples{
\dontrun{
use_readme_rmd()
use_readme_md()
}
}
\seealso{
Other infrastructure: \code{\link{infrastructure}},
  \code{\link{use_build_ignore}},
  \code{\link{use_data_raw}}, \code{\link{use_data}},
  \code{\link{use_news_md}}, \code{\link{use_package}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build-readme.R
\name{build_readme}
\alias{build_readme}
\title{Build README}
\usage{
build_readme(path = ".", quiet = TRUE, ...)
}
\arguments{
\item{path}{Path to the top-level directory of the source package.}

\item{quiet}{If \code{TRUE}, suppresses most output. Set to \code{FALSE}
if you need to debug.}

\item{...}{Additional arguments passed to \code{\link[rmarkdown:render]{rmarkdown::render()}}, in the
case of \code{README.Rmd}. Not used for \code{README.qmd}}
}
\description{
Renders an executable README, i.e. \code{README.qmd} or \code{README.Rmd}, to
\code{README.md}. Specifically, \code{build_readme()}:
\itemize{
\item Installs a copy of the package's current source to a temporary library
\item Renders the README in a clean R session
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-devtools.R
\name{release_checks}
\alias{release_checks}
\title{Custom devtools release checks.}
\usage{
release_checks(pkg = ".", built_path = NULL)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}
}
\description{
This function performs additional checks prior to release. It is called
automatically by \code{\link[=release]{release()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install-svn.r
\name{install_svn}
\alias{install_svn}
\title{Install a package from a SVN repository}
\usage{
install_svn(url, subdir = NULL, branch = NULL, args = character(0), ...,
  revision = NULL, quiet = FALSE)
}
\arguments{
\item{url}{Location of package. The url should point to a public or
private repository.}

\item{subdir}{A sub-directory withing a svn repository that may contain the
package we are interested in installing. By default, this
points to the 'trunk' directory.}

\item{branch}{Name of branch or tag to use, if not trunk.}

\item{args}{A character vector providing extra arguments to pass on to}

\item{...}{Other arguments passed on to \code{\link{install}}}

\item{revision}{svn revision, if omitted updates to latest}

\item{quiet}{if \code{TRUE} suppresses output from this function.}
}
\description{
This function requires \code{svn} to be installed on your system in order to
be used.
}
\details{
It is vectorised so you can install multiple packages with
a single command.
}
\examples{
\dontrun{
install_svn("https://github.com/hadley/stringr")
install_svn("https://github.com/hadley/httr", branch = "oauth")
}
}
\seealso{
Other package installation: \code{\link{install_bioc}},
  \code{\link{install_bitbucket}},
  \code{\link{install_cran}}, \code{\link{install_github}},
  \code{\link{install_git}}, \code{\link{install_url}},
  \code{\link{install_version}}, \code{\link{install}},
  \code{\link{uninstall}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spell-check.R
\name{spell_check}
\alias{spell_check}
\title{Spell checking}
\usage{
spell_check(pkg = ".", vignettes = TRUE, use_wordlist = TRUE)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{vignettes}{also check all \code{rmd} and \code{rnw} files in the pkg \code{vignettes} folder}

\item{use_wordlist}{ignore words in the package \link[spelling:wordlist]{WORDLIST} file}
}
\description{
Runs a spell check on text fields in the package description file, manual
pages, and optionally vignettes. Wraps the \link[spelling:spell_check_package]{spelling}
package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reload.R
\name{reload}
\alias{reload}
\title{Unload and reload package}
\usage{
reload(pkg = ".", quiet = FALSE)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{quiet}{if \code{TRUE} suppresses output from this function.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{reload()} is deprecated because we no longer use or recommend this
workflow. Instead, we recommend \code{\link[=load_all]{load_all()}} to load a package for
interactive development.
}
\seealso{
\code{\link[=load_all]{load_all()}} to load a package for interactive development.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-meta.r
\name{dev_meta}
\alias{dev_meta}
\title{Return devtools metadata environment}
\usage{
dev_meta(name)
}
\arguments{
\item{name}{The name of a loaded package}
}
\description{
If the package was not loaded with devtools, returns \code{NULL}.
}
\examples{
dev_meta("stats") # NULL

if (has_tests()) {
# Load the test package in directory "testLoadHooks"
load_all(devtest("testLoadHooks"))

# Get metdata for the package
x <- dev_meta("testLoadHooks")
as.list(x)

# Clean up.
unload(devtest("testLoadHooks"))
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-win.R
\name{check_win}
\alias{check_win}
\alias{check_win_devel}
\alias{check_win_release}
\alias{check_win_oldrelease}
\title{Check a package on Windows}
\usage{
check_win_devel(
  pkg = ".",
  args = NULL,
  manual = TRUE,
  email = NULL,
  quiet = FALSE,
  webform = FALSE,
  ...
)

check_win_release(
  pkg = ".",
  args = NULL,
  manual = TRUE,
  email = NULL,
  quiet = FALSE,
  webform = FALSE,
  ...
)

check_win_oldrelease(
  pkg = ".",
  args = NULL,
  manual = TRUE,
  email = NULL,
  quiet = FALSE,
  webform = FALSE,
  ...
)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{args}{An optional character vector of additional command
line arguments to be passed to \verb{R CMD build} if \code{binary = FALSE},
or \verb{R CMD install} if \code{binary = TRUE}.}

\item{manual}{Should the manual be built?}

\item{email}{An alternative email address to use. If \code{NULL}, the default is
to use the package maintainer's email.}

\item{quiet}{If \code{TRUE}, suppresses output.}

\item{webform}{If \code{TRUE}, uses web form instead of passive FTP upload.}

\item{...}{Additional arguments passed to \code{\link[pkgbuild:build]{pkgbuild::build()}}.}
}
\description{
This function first bundles a source package, then uploads it to
\url{https://win-builder.r-project.org/}. Once the service has built and checked
the package, an email is sent to address of the maintainer listed in
\code{DESCRIPTION}. This usually takes around 30 minutes. The email contains a
link to a directory with the package binary and check logs, which will be
deleted after a couple of days.
}
\section{Functions}{
\itemize{
\item \code{check_win_devel()}: Check package on the development version of R.

\item \code{check_win_release()}: Check package on the released version of R.

\item \code{check_win_oldrelease()}: Check package on the previous major release version of R.

}}
\seealso{
Other build functions: 
\code{\link{check_mac_release}()}
}
\concept{build functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/namespace-env.r
\name{ns_env}
\alias{ns_env}
\title{Return the namespace environment for a package.}
\usage{
ns_env(pkg = ".")
}
\arguments{
\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information}
}
\description{
Contains all (exported and non-exported) objects, and is a descendent of
\code{R_GlobalEnv}. The hierarchy is \code{<namespace:pkg>},
\code{<imports:pkg>}, \code{<namespace:base>}, and then
\code{R_GlobalEnv}.
}
\details{
If the package is not loaded, this function returns \code{NULL}.
}
\seealso{
\code{\link{pkg_env}} for the attached environment that
  contains the exported objects.

\code{\link{imports_env}} for the environment that contains
  imported objects for the package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install-version.r
\name{install_version}
\alias{install_version}
\title{Install specified version of a CRAN package.}
\usage{
install_version(package, version = NULL, repos = getOption("repos"),
  type = getOption("pkgType"), ..., quiet = FALSE)
}
\arguments{
\item{package}{package name}

\item{version}{If the specified version is NULL or the same as the most
recent version of the package, this function simply calls
\code{\link{install}}. Otherwise, it looks at the list of
archived source tarballs and tries to install an older version instead.}

\item{repos}{
    character vector, the base URL(s) of the repositories
    to use, e.g., the URL of a CRAN mirror such as
    \code{"https://cloud.r-project.org"}.  For more details on
    supported URL schemes see \code{\link{url}}.

    Can be \code{NULL} to install from local files, directories or URLs:
    this will be inferred by extension from \code{pkgs} if of length one.
  }

\item{type}{character, indicating the type of package to download and
    install.  Will be \code{"source"} except on Windows and some macOS
    builds: see the section on \sQuote{Binary packages} for those.
  }

\item{...}{Other arguments passed on to \code{\link{install}}.}

\item{quiet}{
    logical: if true, reduce the amount of output.
  }
}
\description{
If you are installing an package that contains compiled code, you will
need to have an R development environment installed.  You can check
if you do by running \code{\link{has_devel}}.
}
\seealso{
Other package installation: \code{\link{install_bioc}},
  \code{\link{install_bitbucket}},
  \code{\link{install_cran}}, \code{\link{install_github}},
  \code{\link{install_git}}, \code{\link{install_svn}},
  \code{\link{install_url}}, \code{\link{install}},
  \code{\link{uninstall}}
}
\author{
Jeremy Stephens
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/revdep.R
\name{revdep}
\alias{revdep}
\alias{revdep_maintainers}
\title{Reverse dependency tools.}
\usage{
revdep(
  pkg,
  dependencies = c("Depends", "Imports", "Suggests", "LinkingTo"),
  recursive = FALSE,
  ignore = NULL,
  bioconductor = FALSE
)

revdep_maintainers(pkg = ".")
}
\arguments{
\item{pkg}{Package name. This is unlike most devtools packages which
take a path because you might want to determine dependencies for a package
that you don't have installed. If omitted, defaults to the name of the
current package.}

\item{dependencies}{A character vector listing the types of dependencies
to follow.}

\item{recursive}{If \code{TRUE} look for full set of recursive dependencies.}

\item{ignore}{A character vector of package names to ignore. These packages
will not appear in returned vector.}

\item{bioconductor}{If \code{TRUE} also look for dependencies amongst
Bioconductor packages.}
}
\description{
Tools to check and notify maintainers of all CRAN and Bioconductor
packages that depend on the specified package.
}
\details{
The first run in a session will be time-consuming because it must download
all package metadata from CRAN and Bioconductor. Subsequent runs will
be faster.
}
\examples{
\dontrun{
revdep("ggplot2")

revdep("ggplot2", ignore = c("xkcd", "zoo"))
}
}
\seealso{
The \href{https://github.com/r-lib/revdepcheck}{revdepcheck} package can
be used to run R CMD check on all reverse dependencies.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-info.r
\name{loaded_packages}
\alias{loaded_packages}
\title{Return a vector of names of attached packages}
\usage{
loaded_packages()
}
\value{
A data frame with columns package and path, giving the name of
  each package and the path it was loaded from.
}
\description{
Return a vector of names of attached packages
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wd.R
\name{wd}
\alias{wd}
\title{Set working directory}
\usage{
wd(pkg = ".", path = "")
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{path}{path within package. Leave empty to change working directory
to package directory.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{wd()} is deprecated because we no longer use or recommend this workflow.
Set working directory
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-git.R
\name{git_checks}
\alias{git_checks}
\title{Git checks.}
\usage{
git_checks(pkg = ".")
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}
}
\description{
This function performs Git checks checks prior to release. It is called
automatically by \code{\link[=release]{release()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-example.r
\name{dev_example}
\alias{dev_example}
\title{Run a examples for an in-development function.}
\usage{
dev_example(topic)
}
\arguments{
\item{topic}{Name or topic (or name of Rd) file to run examples for}
}
\description{
Run a examples for an in-development function.
}
\examples{
\dontrun{
# Runs installed example:
library("ggplot2")
example("ggplot")

# Runs develoment example:
load_all("ggplot2")
dev_example("ggplot")
}
}
\seealso{
Other example functions: \code{\link{run_examples}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test.r
\name{devtest}
\alias{devtest}
\title{Return the path to one of the packages in the devtools test dir}
\usage{
devtest(package)
}
\arguments{
\item{package}{Name of the test package.}
}
\description{
Devtools comes with some simple packages for testing. This function
returns the path to them.
}
\examples{
if (has_tests()) {
devtest("testData")
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-results.R
\name{check_failures}
\alias{check_failures}
\title{Parses R CMD check log file for ERRORs, WARNINGs and NOTEs}
\usage{
check_failures(path, error = TRUE, warning = TRUE, note = TRUE)
}
\arguments{
\item{path}{check path, e.g., value of the \code{check_dir} argument in a
call to \code{\link[=check]{check()}}}

\item{error, warning, note}{logical, indicates if errors, warnings and/or
notes should be returned}
}
\value{
a character vector with the relevant messages, can have length zero
if no messages are found
}
\description{
Extracts check messages from the \verb{00check.log} file generated by
\verb{R CMD check}.
}
\seealso{
\code{\link[=check]{check()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infrastructure.R
\name{use_data}
\alias{use_data}
\title{Use data in a package.}
\usage{
use_data(..., pkg = ".", internal = FALSE, overwrite = FALSE,
  compress = "bzip2")
}
\arguments{
\item{...}{Unquoted names of existing objects to save.}

\item{pkg}{Package where to store data. Defaults to package in working
directory.}

\item{internal}{If \code{FALSE}, saves each object in individual
\code{.rda} files in the \code{data/} directory. These are available
whenever the package is loaded. If \code{TRUE}, stores all objects in
a single \code{R/sysdata.rda} file. These objects are only available
within the package.}

\item{overwrite}{By default, \code{use_data} will not overwrite existing
files. If you really want to do so, set this to \code{TRUE}.}

\item{compress}{Choose the type of compression used by \code{\link{save}}.
Should be one of "gzip", "bzip2" or "xz".}
}
\description{
This function makes it easy to save package data in the correct format.
}
\examples{
\dontrun{
x <- 1:10
y <- 1:100

use_data(x, y) # For external use
use_data(x, y, internal = TRUE) # For internal use
}
}
\seealso{
Other infrastructure: \code{\link{infrastructure}},
  \code{\link{use_build_ignore}},
  \code{\link{use_data_raw}}, \code{\link{use_news_md}},
  \code{\link{use_package}}, \code{\link{use_readme_rmd}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remotes.R
\name{install-deprecated}
\alias{install-deprecated}
\alias{install_bioc}
\alias{install_bitbucket}
\alias{install_cran}
\alias{install_dev}
\alias{install_git}
\alias{install_github}
\alias{install_gitlab}
\alias{install_local}
\alias{install_svn}
\alias{install_url}
\alias{install_version}
\alias{update_packages}
\alias{dev_package_deps}
\alias{github_pull}
\alias{github_release}
\title{Deprecated package installation functions}
\usage{
install_bioc(...)

install_bitbucket(...)

install_cran(...)

install_dev(...)

install_git(...)

install_github(...)

install_gitlab(...)

install_local(...)

install_svn(...)

install_url(...)

install_version(...)

update_packages(...)

dev_package_deps(...)

github_pull(...)

github_release(...)
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

These functions have been deprecated in favor of \href{https://pak.r-lib.org/}{pak},
as that is what we now recommend for package installation. There are a few
functions which have no pak equivalent, where you can instead call the
old remotes functions directly.
\subsection{Migration guide}{\tabular{ll}{
   devtools function \tab Replacement \cr
   \code{install_bioc("pkg")} \tab \code{pak::pak("bioc::pkg")} \cr
   \code{install_bitbucket("user/repo")} \tab \code{remotes::install_bitbucket("user/repo")} \cr
   \code{install_cran("pkg")} \tab \code{pak::pak("pkg")} \cr
   \code{install_dev("pkg")} \tab \code{remotes::install_dev("pkg")} \cr
   \code{install_git("url")} \tab \code{pak::pak("git::url")} \cr
   \code{install_github("user/repo")} \tab \code{pak::pak("user/repo")} \cr
   \code{install_gitlab("user/repo")} \tab \code{pak::pak("gitlab::user/repo")} \cr
   \code{install_local("path")} \tab \code{pak::pak("local::path")} \cr
   \code{install_svn("url")} \tab \code{remotes::install_svn("url")} \cr
   \code{install_url("url")} \tab \code{pak::pak("url::url")} \cr
   \code{install_version("pkg", "1.0.0")} \tab \code{pak::pak("pkg@1.0.0")} \cr
   \code{update_packages("pkg")} \tab \code{pak::pak("pkg")} \cr
   \code{dev_package_deps()} \tab \code{pak::local_dev_deps()} \cr
   \code{github_pull("123")} \tab \code{remotes::github_pull("123")} \cr
   \code{github_release()} \tab \code{remotes::github_release()} \cr
}

}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test.R
\name{test}
\alias{test}
\alias{test_active_file}
\alias{test_coverage}
\alias{test_coverage_active_file}
\title{Execute testthat tests in a package}
\usage{
test(pkg = ".", filter = NULL, stop_on_failure = FALSE, export_all = TRUE, ...)

test_active_file(file = find_active_file(), ...)

test_coverage(pkg = ".", report = NULL, ...)

test_coverage_active_file(
  file = find_active_file(),
  filter = TRUE,
  report = NULL,
  export_all = TRUE,
  ...
)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{filter}{If not \code{NULL}, only tests with file names matching this
regular expression will be executed. Matching is performed on the file
name after it's stripped of \code{"test-"} and \code{".R"}.}

\item{stop_on_failure}{If \code{TRUE}, throw an error if any tests fail.}

\item{export_all}{If \code{TRUE} (the default), export all objects.
If \code{FALSE}, export only the objects that are listed as exports
in the NAMESPACE file.}

\item{...}{additional arguments passed to wrapped functions.}

\item{file}{One or more source or test files. If a source file the
corresponding test file will be run. The default is to use the active file
in RStudio (if available).}

\item{report}{How to display the coverage report.
\itemize{
\item \code{"html"} opens an interactive report in the browser.
\item \code{"zero"} prints uncovered lines to the console.
\item \code{"silent"} returns the coverage object without display.
}

Defaults to \code{"html"} if interactive; otherwise to \code{"zero"}.}
}
\description{
\itemize{
\item \code{test()} runs all tests in a package. It's a shortcut for
\code{\link[testthat:test_dir]{testthat::test_dir()}}
\item \code{test_active_file()} runs \code{test()} on the active file.
\item \code{test_coverage()} computes test coverage for your package. It's a
shortcut for \code{\link[covr:package_coverage]{covr::package_coverage()}} plus \code{\link[covr:report]{covr::report()}}.
\item \code{test_coverage_active_file()} computes test coverage for the active file. It's a
shortcut for \code{\link[covr:file_coverage]{covr::file_coverage()}} plus \code{\link[covr:report]{covr::report()}}.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infrastructure.R
\name{use_package}
\alias{use_package}
\title{Use specified package.}
\usage{
use_package(package, type = "Imports", pkg = ".")
}
\arguments{
\item{package}{Name of package to depend on.}

\item{type}{Type of dependency: must be one of "Imports", "Depends", 
"Suggests", "Enhances", or "LinkingTo" (or unique abbreviation)}

\item{pkg}{package description, can be path or package name. See
\code{\link{as.package}} for more information.}
}
\description{
This adds a dependency to DESCRIPTION and offers a little advice
about how to best use it.
}
\examples{
\dontrun{
use_package("ggplot2")
use_package("dplyr", "suggests")

}
}
\seealso{
Other infrastructure: \code{\link{infrastructure}},
  \code{\link{use_build_ignore}},
  \code{\link{use_data_raw}}, \code{\link{use_data}},
  \code{\link{use_news_md}}, \code{\link{use_readme_rmd}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/release.R
\name{submit_cran}
\alias{submit_cran}
\title{Submit a package to CRAN}
\usage{
submit_cran(pkg = ".", args = NULL)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{args}{An optional character vector of additional command
line arguments to be passed to \verb{R CMD build}.}
}
\description{
This submits your package to CRAN using the web-form submission process.
To complete the submission you will need respond to the email sent to the
maintainer email address.

We generally recommend using this part of the process defined by
\code{\link[usethis:use_release_issue]{usethis::use_release_issue()}}; this process maximizes the chances of a successful
submission.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/has-tests.R
\name{has_tests}
\alias{has_tests}
\title{Was devtools installed with tests?}
\usage{
has_tests()
}
\description{
Was devtools installed with tests?
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infrastructure-git.R
\name{use_github_links}
\alias{use_github_links}
\title{Add GitHub links to DESCRIPTION.}
\usage{
use_github_links(pkg = ".", auth_token = github_pat(),
  host = "https://api.github.com")
}
\arguments{
\item{pkg}{Path to package. See \code{\link{as.package}} for more
information.}

\item{auth_token}{Provide a personal access token (PAT) from
\url{https://github.com/settings/tokens}. Defaults to the \code{GITHUB_PAT}
environment variable.}

\item{host}{GitHub API host to use. Override with the endpoint-root for your
GitHub enterprise instance, for example,
"https://github.hostname.com/api/v3".}
}
\description{
Populates the URL and BugReports fields of DESCRIPTION with
\code{https://github.com/<USERNAME>/<REPO>} AND
\code{https://github.com/<USERNAME>/<REPO>/issues}, respectively, unless
those fields already exist.
}
\seealso{
Other git infrastructure: \code{\link{use_git_hook}},
  \code{\link{use_github}}, \code{\link{use_git}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgload.R
\name{load_all}
\alias{load_all}
\title{Load complete package}
\usage{
load_all(
  path = ".",
  reset = TRUE,
  recompile = FALSE,
  export_all = TRUE,
  helpers = TRUE,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{path}{Path to a package, or within a package.}

\item{reset}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} This is no longer supported
because preserving the namespace requires unlocking its environment, which
is no longer possible in recent versions of R.}

\item{recompile}{DEPRECATED. force a recompile of DLL from source code, if
present. This is equivalent to running \code{\link[pkgbuild:clean_dll]{pkgbuild::clean_dll()}} before
\code{load_all()}}

\item{export_all}{If \code{TRUE} (the default), export all objects.
If \code{FALSE}, export only the objects that are listed as exports
in the NAMESPACE file.}

\item{helpers}{if \code{TRUE} loads \pkg{testthat} test helpers.}

\item{quiet}{if \code{TRUE} suppresses output from this function.}

\item{...}{Additional arguments passed to \code{\link[pkgload:load_all]{pkgload::load_all()}}.}
}
\description{
\code{load_all()} loads a package. It roughly simulates what happens
when a package is installed and loaded with \code{\link[=library]{library()}}, without
having to first install the package. It:
\itemize{
\item Loads all data files in \verb{data/}. See \code{\link[pkgload:load_data]{load_data()}} for more
details.
\item Sources all R files in the R directory, storing results in
environment that behaves like a regular package namespace. See
\code{\link[pkgload:load_code]{load_code()}} for more details.
\item Adds a shim from \code{\link[pkgload:system.file]{system.file()}} to \code{\link[pkgload:shim_system.file]{shim_system.file()}} in
the imports environment of the package. This ensures that \code{system.file()}
works with both development and installed packages despite their differing
directory structures.
\item Adds shims from \code{help()} and \verb{?} to \code{\link[pkgload:shim_help]{shim_help()}} and \code{\link[pkgload:shim_question]{shim_question()}}
to make it easier to preview development documentation.
\item Compiles any C, C++, or Fortran code in the \verb{src/} directory and
connects the generated DLL into R. See \code{\link[pkgbuild:compile_dll]{pkgbuild::compile_dll()}}
for more details.
\item Loads any compiled translations in \code{inst/po}.
\item Runs \code{.onAttach()}, \code{.onLoad()} and \code{.onUnload()} functions at
the correct times.
\item If you use \pkg{testthat}, will load all test helpers so you can
access them interactively. devtools sets the \code{DEVTOOLS_LOAD}
environment variable to the package name to let you check whether the
helpers are run during package loading.
}

\code{is_loading()} returns \code{TRUE} when it is called while \code{load_all()}
is running. This may be useful e.g. in \code{.onLoad} hooks.
A package loaded with \code{load_all()} can be identified with
\code{\link[pkgload:is_dev_package]{is_dev_package()}}.
}
\section{Differences to regular loading}{

\code{load_all()} tries its best to reproduce the behaviour of
\code{\link[=loadNamespace]{loadNamespace()}} and \code{\link[=library]{library()}}. However it deviates from normal
package loading in several ways.
\itemize{
\item \code{load_all()} doesn't install the package to a library, so \code{\link[pkgload:system.file]{system.file()}}
doesn't work. pkgload fixes this for the package itself installing a shim,
\code{\link[pkgload:shim_system.file]{shim_system.file()}}. However, this shim is not visible to third party
packages, so they will fail if they attempt to find files within your
package. One potential workaround is to use \code{\link[fs:path_package]{fs::path_package()}}
instead of \code{system.file()}, since that understands the mechanisms that
devtools uses to load packages.
\item \code{load_all()} loads all packages referenced in \code{Imports} at load time,
but \code{loadNamespace()} and \code{library()} only load package dependencies as
they are needed.
\item \code{load_all()} copies all objects (not just the ones listed as exports)
into the package environment. This is useful during development because
it makes internal objects easy to access.  To export only the objects
listed as exports, use \code{export_all = FALSE}. This more closely simulates
behavior when loading an installed package with \code{\link[=library]{library()}}, and can
be useful for checking for missing exports.
}

}

\section{Controlling the debug compiler flags}{

\code{load_all()} delegates to \code{\link[pkgbuild:compile_dll]{pkgbuild::compile_dll()}} to perform the actual
compilation, during which by default some debug compiler flags are
appended. If you would like to produce an optimized build instead, you can
opt out by either using \code{debug = FALSE}, setting the \code{pkg.build_extra_flags}
option to \code{FALSE}, or setting the \code{PKG_BUILD_EXTRA_FLAGS} environment variable
to \code{FALSE}. For further details see the Details section in \code{\link[pkgbuild:compile_dll]{pkgbuild::compile_dll()}}.

}

\examples{
\dontrun{
# Load the package in the current directory
load_all("./")

# Running again loads changed files
load_all("./")

# With export_all=FALSE, only objects listed as exports in NAMESPACE
# are exported
load_all("./", export_all = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-help.r
\name{dev_help}
\alias{dev_help}
\title{Read the in-development help for a package loaded with devtools.}
\usage{
dev_help(topic, stage = "render", type = getOption("help_type"))
}
\arguments{
\item{topic}{name of help to search for.}

\item{stage}{at which stage ("build", "install", or "render") should
\\Sexpr macros be executed? This is only important if you're using
\\Sexpr macro's in your Rd files.}

\item{type}{of html to produce: \code{"html"} or \code{"text"}. Defaults to
your default documentation type.}
}
\description{
Note that this only renders a single documentation file, so that links
to other files within the package won't work.
}
\examples{
\dontrun{
library("ggplot2")
help("ggplot") # loads installed documentation for ggplot

load_all("ggplot2")
dev_help("ggplot") # loads development documentation for ggplot
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.r
\name{package_file}
\alias{package_file}
\title{Find file in a package.}
\usage{
package_file(..., path = ".")
}
\arguments{
\item{...}{Components of the path.}

\item{path}{Place to start search for package directory.}
}
\description{
It always starts by finding by walking up the path until it finds the
root directory, i.e. a directory containing \code{DESCRIPTION}. If it
cannot find the root directory, or it can't find the specified path, it
will throw an error.
}
\examples{
\dontrun{
package_file("figures", "figure_1")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/show-news.R
\name{show_news}
\alias{show_news}
\title{Show package news}
\usage{
show_news(pkg = ".", latest = TRUE, ...)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{latest}{if \code{TRUE}, only show the news for the most recent
version.}

\item{...}{other arguments passed on to \code{news}}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{show_news()} is deprecated because we no longer use or recommend this
workflow.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.R
\name{as.package}
\alias{as.package}
\title{Coerce input to a package}
\usage{
as.package(x = NULL, create = deprecated())
}
\arguments{
\item{x}{object to coerce to a package}

\item{create}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Hasn't worked for some time.}
}
\description{
Possible specifications of package:
\itemize{
\item path
\item package object
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run-source.R
\name{source_url}
\alias{source_url}
\title{Run a script through some protocols such as http, https, ftp, etc.}
\usage{
source_url(url, ..., sha1 = NULL)
}
\arguments{
\item{url}{url}

\item{...}{other options passed to \code{\link[=source]{source()}}}

\item{sha1}{The (prefix of the) SHA-1 hash of the file at the remote URL.}
}
\description{
If a SHA-1 hash is specified with the \code{sha1} argument, then this
function will check the SHA-1 hash of the downloaded file to make sure it
matches the expected value, and throw an error if it does not match. If the
SHA-1 hash is not specified, it will print a message displaying the hash of
the downloaded file. The purpose of this is to improve security when running
remotely-hosted code; if you have a hash of the file, you can be sure that
it has not changed. For convenience, it is possible to use a truncated SHA1
hash, down to 6 characters, but keep in mind that a truncated hash won't be
as secure as the full hash.
}
\examples{
\dontrun{

source_url("https://gist.github.com/hadley/6872663/raw/hi.r")

# With a hash, to make sure the remote file hasn't changed
source_url("https://gist.github.com/hadley/6872663/raw/hi.r",
  sha1 = "54f1db27e60bb7e0486d785604909b49e8fef9f9")

# With a truncated hash
source_url("https://gist.github.com/hadley/6872663/raw/hi.r",
  sha1 = "54f1db27e60")
}
}
\seealso{
\code{\link[=source_gist]{source_gist()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/revdep-email.R
\name{revdep_email}
\alias{revdep_email}
\title{Experimental email notification system.}
\usage{
revdep_email(pkg = ".", date, version, author = getOption("devtools.name"),
  draft = TRUE, unsent = NULL, template = "revdep/email.md",
  only_problems = TRUE)
}
\arguments{
\item{pkg}{Path to package. Defaults to current directory.}

\item{date}{Date package will be submitted to CRAN}

\item{version}{Version which will be used for the CRAN submission (usually
different from the current package version)}

\item{author}{Name used to sign email}

\item{draft}{If \code{TRUE}, creates as draft email; if \code{FALSE},
sends immediately.}

\item{unsent}{If some emails fail to send, in a previous}

\item{template}{Path of template to use}

\item{only_problems}{Only inform authors with problems?}
}
\description{
This currently assumes that you use github and gmail, and you have a
\code{revdep/email.md} email template.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vignettes.R
\name{build_vignettes}
\alias{build_vignettes}
\title{Build package vignettes}
\usage{
build_vignettes(
  pkg = ".",
  dependencies = "VignetteBuilder",
  clean = TRUE,
  upgrade = "never",
  quiet = FALSE,
  install = TRUE,
  keep_md = TRUE
)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{dependencies}{Which dependencies do you want to check?
Can be a character vector (selecting from "Depends", "Imports",
"LinkingTo", "Suggests", or "Enhances"), or a logical vector.

\code{TRUE} is shorthand for "Depends", "Imports", "LinkingTo" and
"Suggests". \code{NA} is shorthand for "Depends", "Imports" and "LinkingTo"
and is the default. \code{FALSE} is shorthand for no dependencies (i.e.
just check this package, not its dependencies).

The value "soft" means the same as \code{TRUE}, "hard" means the same as \code{NA}.

You can also specify dependencies from one or more additional fields,
common ones include:
\itemize{
\item Config/Needs/website - for dependencies used in building the pkgdown site.
\item Config/Needs/coverage for dependencies used in calculating test coverage.
}}

\item{clean}{Remove all files generated by the build, even if there
    were copies there before.}

\item{upgrade}{Should package dependencies be upgraded? One of "default", "ask", "always", or "never". "default"
respects the value of the \code{R_REMOTES_UPGRADE} environment variable if set,
and falls back to "ask" if unset. "ask" prompts the user for which out of
date packages to upgrade. For non-interactive sessions "ask" is equivalent
to "always". \code{TRUE} and \code{FALSE} are also accepted and correspond to
"always" and "never" respectively.}

\item{quiet}{If \code{TRUE}, suppresses most output. Set to \code{FALSE}
if you need to debug.}

\item{install}{If \code{TRUE}, install the package before building
vignettes.}

\item{keep_md}{If \code{TRUE}, move md intermediates as well as rendered
outputs. Most useful when using the \code{keep_md} YAML option for Rmarkdown
outputs.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{build_vignettes()} is deprecated because we no longer recommend that you
build vignettes in this way, because it leaves build artifacts in your
development directory. Instead, use \code{\link[pkgdown:build_articles]{pkgdown::build_article()}} to
render articles locally for preview and polishing.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/save-all.R
\name{save_all}
\alias{save_all}
\title{Save all documents in an active IDE session.}
\usage{
save_all()
}
\description{
Helper function wrapping IDE-specific calls to save all documents in the
active session. In this form, callers of \code{save_all()} don't need to
execute any IDE-specific code. This function can be extended to include
other IDE implementations of their equivalent
\code{rstudioapi::documentSaveAll()} methods.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install-bioc.r
\name{install_bioc}
\alias{install_bioc}
\title{Install a package from a Bioconductor repository}
\usage{
install_bioc(repo, mirror = getOption("BioC_svn",
  "https://hedgehog.fhcrc.org/bioconductor"), ..., quiet = FALSE)
}
\arguments{
\item{repo}{Repository address in the format
\code{[username:password@][release/]repo[#revision]}. Valid values for
the release are \sQuote{devel} (the default if none specified),
\sQuote{release} or numeric release numbers (e.g. \sQuote{3.3}).}

\item{mirror}{The bioconductor SVN mirror to use}

\item{...}{Other arguments passed on to \code{\link{install}}}

\item{quiet}{if \code{TRUE} suppresses output from this function.}
}
\description{
This function requires \code{svn} to be installed on your system in order to
be used.
}
\details{
It is vectorised so you can install multiple packages with
a single command.

'
}
\examples{
\dontrun{
install_bioc("SummarizedExperiment")
install_bioc("user@SummarizedExperiment")
install_bioc("user:password@release/SummarizedExperiment")
install_bioc("user:password@3.3/SummarizedExperiment")
install_bioc("user:password@3.3/SummarizedExperiment#117513")
}
}
\seealso{
Other package installation: \code{\link{install_bitbucket}},
  \code{\link{install_cran}}, \code{\link{install_github}},
  \code{\link{install_git}}, \code{\link{install_svn}},
  \code{\link{install_url}}, \code{\link{install_version}},
  \code{\link{install}}, \code{\link{uninstall}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/github.R
\name{github_pat}
\alias{github_pat}
\title{Retrieve GitHub personal access token.}
\usage{
github_pat(quiet = TRUE)
}
\description{
A GitHub personal access token
Looks in env var \code{GITHUB_PAT}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-mac.R
\name{check_mac_release}
\alias{check_mac_release}
\alias{check_mac_devel}
\title{Check a package on macOS}
\usage{
check_mac_release(
  pkg = ".",
  dep_pkgs = character(),
  args = NULL,
  manual = TRUE,
  quiet = FALSE,
  ...
)

check_mac_devel(
  pkg = ".",
  dep_pkgs = character(),
  args = NULL,
  manual = TRUE,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{dep_pkgs}{Additional custom dependencies to install prior to checking
the package.}

\item{args}{An optional character vector of additional command
line arguments to be passed to \verb{R CMD build} if \code{binary = FALSE},
or \verb{R CMD install} if \code{binary = TRUE}.}

\item{manual}{Should the manual be built?}

\item{quiet}{If \code{TRUE}, suppresses output.}

\item{...}{Additional arguments passed to \code{\link[pkgbuild:build]{pkgbuild::build()}}.}
}
\value{
The url with the check results (invisibly)
}
\description{
Check on either the released or development versions of R, using
\url{https://mac.r-project.org/macbuilder/submit.html}.
}
\seealso{
Other build functions: 
\code{\link{check_win}()}
}
\concept{build functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/devtools-package.R
\docType{package}
\name{devtools-package}
\alias{devtools}
\alias{devtools-package}
\title{devtools: Tools to Make Developing R Packages Easier}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Collection of package development tools.
}
\section{Package options}{


Devtools uses the following \code{\link[=options]{options()}} to configure behaviour:
\itemize{
\item \code{devtools.install.args}: a string giving extra arguments passed to
\verb{R CMD install} by \code{\link[=install]{install()}}.
\item \code{devtools.path}: Deprecated. Path used by the now-deprecated \code{\link[=dev_mode]{dev_mode()}}
function.
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://devtools.r-lib.org/}
  \item \url{https://github.com/r-lib/devtools}
  \item Report bugs at \url{https://github.com/r-lib/devtools/issues}
}

}
\author{
\strong{Maintainer}: Jennifer Bryan \email{jenny@posit.co} (\href{https://orcid.org/0000-0002-6983-2759}{ORCID})

Authors:
\itemize{
  \item Hadley Wickham
  \item Jim Hester
  \item Winston Chang
}

Other contributors:
\itemize{
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create.R
\name{create}
\alias{create}
\title{Create a package}
\usage{
create(path, ..., open = FALSE)
}
\arguments{
\item{path}{A path. If it exists, it is used. If it does not exist, it is
created, provided that the parent path exists.}

\item{...}{Additional arguments passed to \code{\link[usethis:create_package]{usethis::create_package()}}}

\item{open}{If \code{TRUE}, \link[usethis:proj_activate]{activates} the new project:
\itemize{
\item If using RStudio or Positron, the new project is opened in a new session,
window, or browser tab, depending on the product (RStudio or Positron)
and context (desktop or server).
\item Otherwise, the working directory and active project of the current R
session are changed to the new project.
}}
}
\value{
The path to the created package, invisibly.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{create()} is deprecated. Please use \code{\link[usethis:create_package]{usethis::create_package()}} directly
instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install.R
\name{install_deps}
\alias{install_deps}
\alias{install_dev_deps}
\title{Install package dependencies if needed.}
\usage{
install_deps(
  pkg = ".",
  dependencies = NA,
  repos = getOption("repos"),
  type = getOption("pkgType"),
  upgrade = c("default", "ask", "always", "never"),
  quiet = FALSE,
  build = TRUE,
  build_opts = c("--no-resave-data", "--no-manual", " --no-build-vignettes"),
  ...
)

install_dev_deps(
  pkg = ".",
  dependencies = TRUE,
  repos = getOption("repos"),
  type = getOption("pkgType"),
  upgrade = c("default", "ask", "always", "never"),
  quiet = FALSE,
  build = TRUE,
  build_opts = c("--no-resave-data", "--no-manual", " --no-build-vignettes"),
  ...
)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{dependencies}{Which dependencies do you want to check?
Can be a character vector (selecting from "Depends", "Imports",
"LinkingTo", "Suggests", or "Enhances"), or a logical vector.

\code{TRUE} is shorthand for "Depends", "Imports", "LinkingTo" and
"Suggests". \code{NA} is shorthand for "Depends", "Imports" and "LinkingTo"
and is the default. \code{FALSE} is shorthand for no dependencies (i.e.
just check this package, not its dependencies).

The value "soft" means the same as \code{TRUE}, "hard" means the same as \code{NA}.

You can also specify dependencies from one or more additional fields,
common ones include:
\itemize{
\item Config/Needs/website - for dependencies used in building the pkgdown site.
\item Config/Needs/coverage for dependencies used in calculating test coverage.
}}

\item{repos}{A character vector giving repositories to use.}

\item{type}{Type of package to \code{update}.}

\item{upgrade}{One of "default", "ask", "always", or "never". "default"
respects the value of the \code{R_REMOTES_UPGRADE} environment variable if set,
and falls back to "ask" if unset. "ask" prompts the user for which out of
date packages to upgrade. For non-interactive sessions "ask" is equivalent
to "always". \code{TRUE} and \code{FALSE} are also accepted and correspond to
"always" and "never" respectively.}

\item{quiet}{If \code{TRUE}, suppress output.}

\item{build}{if \code{TRUE} \code{\link[pkgbuild:build]{pkgbuild::build()}}s the package first:
this ensures that the installation is completely clean, and prevents any
binary artefacts (like \file{.o}, \code{.so}) from appearing in your local
package directory, but is considerably slower, because every compile has
to start from scratch.}

\item{build_opts}{Options to pass to \verb{R CMD build}, only used when \code{build}}

\item{...}{additional arguments passed to \code{\link[remotes:install_deps]{remotes::install_deps()}}
when installing dependencies.}
}
\description{
\code{install_deps()} will install the
user dependencies needed to run the package, \code{install_dev_deps()} will also
install the development dependencies needed to test and build the package.
}
\examples{
\dontrun{install_deps(".")}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/has-devel.r
\name{has_devel}
\alias{has_devel}
\title{Check if you have a development environment installed.}
\usage{
has_devel()
}
\value{
TRUE if your development environment is correctly set up, otherwise
  returns an error.
}
\description{
Thanks to the suggestion of Simon Urbanek.
}
\examples{
has_devel()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgbuild.R
\name{build}
\alias{build}
\title{Build package}
\usage{
build(
  pkg = ".",
  path = NULL,
  binary = FALSE,
  vignettes = TRUE,
  manual = FALSE,
  args = NULL,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{path}{Path in which to produce package.  If \code{NULL}, defaults to
the parent directory of the package.}

\item{binary}{Produce a binary (\code{--binary}) or source (
\code{--no-manual --no-resave-data}) version of the package.}

\item{vignettes, manual}{For source packages: if \code{FALSE}, don't build PDF
vignettes (\code{--no-build-vignettes}) or manual (\code{--no-manual}).}

\item{args}{An optional character vector of additional command
line arguments to be passed to \verb{R CMD build} if \code{binary = FALSE},
or \verb{R CMD install} if \code{binary = TRUE}.}

\item{quiet}{if \code{TRUE} suppresses output from this function.}

\item{...}{Additional arguments passed to \link[pkgbuild:build]{pkgbuild::build}.}
}
\value{
a string giving the location (including file name) of the built
package
}
\description{
Building converts a package source directory into a single bundled file.
If \code{binary = FALSE} this creates a \code{tar.gz} package that can
be installed on any platform, provided they have a full development
environment (although packages without source code can typically be
installed out of the box). If \code{binary = TRUE}, the package will have
a platform specific extension (e.g. \code{.zip} for windows), and will
only be installable on the current platform, but no development
environment is needed.
}
\details{
\subsection{Configuration}{
\subsection{\code{DESCRIPTION} entries}{
\itemize{
\item \code{Config/build/clean-inst-doc} can be set to \code{FALSE} to avoid cleaning up
\code{inst/doc} when building a source package. Set it to \code{TRUE} to force a
cleanup. See the \code{clean_doc} argument.
\item \code{Config/build/copy-method} can be used to avoid copying large
directories in \verb{R CMD build}. It works by copying (or linking) the
files of the package to a temporary directory, leaving out the
(possibly large) files that are not part of the package. Possible
values:
\itemize{
\item \code{none}: pkgbuild does not copy the package tree. This is the default.
\item \code{copy}: the package files are copied to a temporary directory before
\verb{ R CMD build}.
\item \code{link}: the package files are symbolic linked to a temporary
directory before \verb{R CMD build}. Windows does not have symbolic
links, so on Windows this is equivalent to \code{copy}.
}

You can also use the \code{pkg.build_copy_method} option or the
\code{PKG_BUILD_COPY_METHOD} environment variable to set the copy method.
The option is consulted first, then the \code{DESCRIPTION} entry, then the
environment variable.
\item \code{Config/build/extra-sources} can be used to define extra source files
for pkgbuild to decide whether a package DLL needs to be recompiled in
\code{needs_compile()}. The syntax is a comma separated list of file names,
or globs. (See \code{\link[utils:glob2rx]{utils::glob2rx()}}.) E.g. \verb{src/rust/src/*.rs} or
\verb{configure*}.
\item \code{Config/build/bootstrap} can be set to \code{TRUE} to run
\verb{Rscript bootstrap.R} in the source directory prior to running subsequent
build steps.
\item \code{Config/build/never-clean} can be set to \code{TRUE} to never add \code{--preclean}
to \verb{R CMD INSTALL}, e.g., when header files have changed.
This helps avoiding rebuilds that can take long for very large C/C++ codebases
and can lead to build failures if object files are out of sync with header files.
Control the dependencies between object files and header files
by adding \verb{include file.d} to \code{Makevars} for each \code{file.c} or \code{file.cpp} source file.
}
}

\subsection{Options}{
\itemize{
\item \code{pkg.build_copy_method}: use this option to avoid copying large
directories when building a package. See possible values above, at the
\code{Config/build/copy-method} \code{DESCRIPTION} entry.
\item \code{pkg.build_stop_for_warnings}: if it is set to \code{TRUE}, then pkgbuild
will stop for \verb{R CMD build} errors. It takes precedence over the
\code{PKG_BUILD_STOP_FOR_WARNINGS} environment variable.
}
}

\subsection{Environment variables}{
\itemize{
\item \code{PKG_BUILD_COLOR_DIAGNOSTICS}: set it to \code{false} to opt out of colored
compiler diagnostics. Set it to \code{true} to force colored compiler
diagnostics.
\item \code{PKG_BUILD_COPY_METHOD}: use this environment variable to avoid copying
large directories when building a package. See possible values above,
at the \code{Config/build/copy-method} \code{DESCRIPTION} entry.
}

will stop for \verb{R CMD build} errors. The \code{pkg.build_stop_for_warnings}
option takes precedence over this environment variable.
}

}
}
\note{
The default \code{manual = FALSE} is not suitable for a CRAN
submission, which may require \code{manual = TRUE}.  Even better, use
\code{\link[=submit_cran]{submit_cran()}} or \code{\link[=release]{release()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/doctor.R
\name{dr_devtools}
\alias{dr_devtools}
\title{Diagnose potential devtools issues}
\usage{
dr_devtools()
}
\description{
This checks to make sure you're using the latest release of R,
the released version of RStudio (if you're using it as your gui),
and the latest version of devtools and its dependencies.
}
\examples{
\dontrun{
dr_devtools()
}
}
\seealso{
Other doctors: \code{\link{dr_github}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/R.R
\name{r_env_vars}
\alias{r_env_vars}
\title{Environment variables to set when calling R}
\usage{
r_env_vars()
}
\value{
a named character vector
}
\description{
Devtools sets a number of environmental variables to ensure consistent
between the current R session and the new session, and to ensure that
everything behaves the same across systems. It also suppresses a common
warning on windows, and sets \code{NOT_CRAN} so you can tell that your
code is not running on CRAN. If \code{NOT_CRAN} has been set externally, it
is not overwritten.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\docType{package}
\name{devtools}
\alias{devtools}
\alias{devtools-package}
\title{Package development tools for R.}
\description{
Collection of package development tools.
}
\section{Package options}{


Devtools uses the following \code{\link[=options]{options()}} to configure behaviour:

\itemize{
\item \code{devtools.path}: path to use for \code{\link[=dev_mode]{dev_mode()}}

\item \code{devtools.name}: your name, used when signing draft
emails.

\item \code{devtools.install.args}: a string giving extra arguments passed
to \verb{R CMD install} by \code{\link[=install]{install()}}.

\item \code{devtools.desc.author}: a string providing a default Authors@R
string to be used in new \file{DESCRIPTION}s.  Should be a R code, and
look like \code{"Hadley Wickham <h.wickham@gmail.com> [aut, cre]"}. See
\code{\link[utils:person]{utils::as.person()}} for more details.

\item \code{devtools.desc.license}: a default license string to use for
new packages.

\item \code{devtools.desc.suggests}: a character vector listing packages to
to add to suggests by defaults for new packages.
\item \code{devtools.desc}: a named list listing any other
extra options to add to \file{DESCRIPTION}

}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://devtools.r-lib.org/}
  \item \url{https://github.com/r-lib/devtools}
  \item Report bugs at \url{https://github.com/r-lib/devtools/issues}
}

}
\author{
\strong{Maintainer}: Jim Hester \email{jim.hester@rstudio.com}

Authors:
\itemize{
  \item Hadley Wickham
  \item Winston Chang
}

Other contributors:
\itemize{
  \item RStudio [copyright holder]
  \item R Core team (Some namespace and vignette code extracted from base R) [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/missing-s3.R
\name{missing_s3}
\alias{missing_s3}
\title{Find missing s3 exports}
\usage{
missing_s3(pkg = ".")
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{missing_s3()} is deprecated because roxygen2 now provides the same
functionality. Run \code{devtools::document()} and look for
\code{"Missing documentation for S3 method"} warnings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run-source.r
\name{source_gist}
\alias{source_gist}
\title{Run a script on gist}
\usage{
source_gist(id, ..., filename = NULL, sha1 = NULL, quiet = FALSE)
}
\arguments{
\item{id}{either full url (character), gist ID (numeric or character of
numeric).}

\item{...}{other options passed to \code{\link{source}}}

\item{filename}{if there is more than one R file in the gist, which one to
source (filename ending in '.R')? Default \code{NULL} will source the
first file.}

\item{sha1}{The SHA-1 hash of the file at the remote URL. This is highly
recommend as it prevents you from accidentally running code that's not
what you expect. See \code{\link{source_url}} for more information on
using a SHA-1 hash.}

\item{quiet}{if \code{FALSE}, the default, prints informative messages.}
}
\description{
\dQuote{Gist is a simple way to share snippets and pastes with others.
  All gists are git repositories, so they are automatically versioned,
  forkable and usable as a git repository.}
\url{https://gist.github.com/}
}
\examples{
\dontrun{
# You can run gists given their id
source_gist(6872663)
source_gist("6872663")

# Or their html url
source_gist("https://gist.github.com/hadley/6872663")
source_gist("gist.github.com/hadley/6872663")

# It's highly recommend that you run source_gist with the optional
# sha1 argument - this will throw an error if the file has changed since
# you first ran it
source_gist(6872663, sha1 = "54f1db27e60")
# Wrong hash will result in error
source_gist(6872663, sha1 = "54f1db27e61")

#' # You can speficy a particular R file in the gist
source_gist(6872663, filename = "hi.r")
source_gist(6872663, filename = "hi.r", sha1 = "54f1db27e60")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-doc.R
\name{check_doc_fields}
\alias{check_doc_fields}
\title{Check for missing documentation fields}
\usage{
check_doc_fields(pkg = ".", fields = c("value", "examples"))
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{fields}{A character vector of Rd field names to check for.}
}
\value{
A named list of character vectors, one for each field, containing
the names of Rd files missing that field. Returned invisibly.
}
\description{
Checks all Rd files in \verb{man/} and looks for any that have a \verb{\\usage} section
(i.e. a function) but that \emph{don't} have \verb{\\value} and \verb{\\examples} sections.
These missing fields are flagged by CRAN on initial submission.
}
\examples{
\dontrun{
check_doc_fields(".")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remotes.R
\name{remote-reexports}
\alias{remote-reexports}
\alias{install_bioc}
\alias{install_bitbucket}
\alias{install_cran}
\alias{install_dev}
\alias{install_git}
\alias{install_github}
\alias{install_gitlab}
\alias{install_local}
\alias{install_svn}
\alias{install_url}
\alias{install_version}
\alias{update_packages}
\alias{dev_package_deps}
\title{Functions re-exported from the remotes package}
\usage{
install_bioc(
  repo,
  mirror = getOption("BioC_git", download_url("git.bioconductor.org/packages")),
  git = c("auto", "git2r", "external"),
  dependencies = NA,
  upgrade = c("default", "ask", "always", "never"),
  force = FALSE,
  quiet = FALSE,
  build = TRUE,
  build_opts = c("--no-resave-data", "--no-manual", "--no-build-vignettes"),
  build_manual = FALSE,
  build_vignettes = FALSE,
  repos = getOption("repos"),
  type = getOption("pkgType"),
  ...
)

install_bitbucket(
  repo,
  ref = "HEAD",
  subdir = NULL,
  auth_user = bitbucket_user(),
  password = bitbucket_password(),
  host = "api.bitbucket.org/2.0",
  dependencies = NA,
  upgrade = c("default", "ask", "always", "never"),
  force = FALSE,
  quiet = FALSE,
  build = TRUE,
  build_opts = c("--no-resave-data", "--no-manual", "--no-build-vignettes"),
  build_manual = FALSE,
  build_vignettes = FALSE,
  repos = getOption("repos"),
  type = getOption("pkgType"),
  ...
)

install_cran(
  pkgs,
  repos = getOption("repos"),
  type = getOption("pkgType"),
  dependencies = NA,
  upgrade = c("default", "ask", "always", "never"),
  force = FALSE,
  quiet = FALSE,
  build = TRUE,
  build_opts = c("--no-resave-data", "--no-manual", "--no-build-vignettes"),
  build_manual = FALSE,
  build_vignettes = FALSE,
  ...
)

install_dev(package, cran_url = getOption("repos")[["CRAN"]], ...)

install_git(
  url,
  subdir = NULL,
  ref = NULL,
  branch = NULL,
  credentials = git_credentials(),
  git = c("auto", "git2r", "external"),
  dependencies = NA,
  upgrade = c("default", "ask", "always", "never"),
  force = FALSE,
  quiet = FALSE,
  build = TRUE,
  build_opts = c("--no-resave-data", "--no-manual", "--no-build-vignettes"),
  build_manual = FALSE,
  build_vignettes = FALSE,
  repos = getOption("repos"),
  type = getOption("pkgType"),
  ...
)

install_github(
  repo,
  ref = "HEAD",
  subdir = NULL,
  auth_token = github_pat(quiet),
  host = "api.github.com",
  dependencies = NA,
  upgrade = c("default", "ask", "always", "never"),
  force = FALSE,
  quiet = FALSE,
  build = TRUE,
  build_opts = c("--no-resave-data", "--no-manual", "--no-build-vignettes"),
  build_manual = FALSE,
  build_vignettes = FALSE,
  repos = getOption("repos"),
  type = getOption("pkgType"),
  ...
)

install_gitlab(
  repo,
  subdir = NULL,
  auth_token = gitlab_pat(quiet),
  host = "gitlab.com",
  dependencies = NA,
  upgrade = c("default", "ask", "always", "never"),
  force = FALSE,
  quiet = FALSE,
  build = TRUE,
  build_opts = c("--no-resave-data", "--no-manual", "--no-build-vignettes"),
  build_manual = FALSE,
  build_vignettes = FALSE,
  repos = getOption("repos"),
  type = getOption("pkgType"),
  ...
)

install_local(
  path = ".",
  subdir = NULL,
  dependencies = NA,
  upgrade = c("default", "ask", "always", "never"),
  force = FALSE,
  quiet = FALSE,
  build = !is_binary_pkg(path),
  build_opts = c("--no-resave-data", "--no-manual", "--no-build-vignettes"),
  build_manual = FALSE,
  build_vignettes = FALSE,
  repos = getOption("repos"),
  type = getOption("pkgType"),
  ...
)

install_svn(
  url,
  subdir = NULL,
  args = character(0),
  revision = NULL,
  dependencies = NA,
  upgrade = c("default", "ask", "always", "never"),
  force = FALSE,
  quiet = FALSE,
  build = TRUE,
  build_opts = c("--no-resave-data", "--no-manual", "--no-build-vignettes"),
  build_manual = FALSE,
  build_vignettes = FALSE,
  repos = getOption("repos"),
  type = getOption("pkgType"),
  ...
)

install_url(
  url,
  subdir = NULL,
  dependencies = NA,
  upgrade = c("default", "ask", "always", "never"),
  force = FALSE,
  quiet = FALSE,
  build = TRUE,
  build_opts = c("--no-resave-data", "--no-manual", "--no-build-vignettes"),
  build_manual = FALSE,
  build_vignettes = FALSE,
  repos = getOption("repos"),
  type = getOption("pkgType"),
  ...
)

install_version(
  package,
  version = NULL,
  dependencies = NA,
  upgrade = c("default", "ask", "always", "never"),
  force = FALSE,
  quiet = FALSE,
  build = FALSE,
  build_opts = c("--no-resave-data", "--no-manual", "--no-build-vignettes"),
  build_manual = FALSE,
  build_vignettes = FALSE,
  repos = getOption("repos"),
  type = "source",
  ...
)

update_packages(
  packages = TRUE,
  dependencies = NA,
  upgrade = c("default", "ask", "always", "never"),
  force = FALSE,
  quiet = FALSE,
  build = TRUE,
  build_opts = c("--no-resave-data", "--no-manual", "--no-build-vignettes"),
  build_manual = FALSE,
  build_vignettes = FALSE,
  repos = getOption("repos"),
  type = getOption("pkgType"),
  ...
)

dev_package_deps(
  pkgdir = ".",
  dependencies = NA,
  repos = getOption("repos"),
  type = getOption("pkgType")
)
}
\description{
These functions are re-exported from the remotes package. They differ only
that the ones in devtools use the \link{ellipsis} package to ensure all dotted
arguments are used.
}
\details{
Follow the links below to see the documentation.
\code{\link[remotes:install_bioc]{remotes::install_bioc()}}, \code{\link[remotes:install_bitbucket]{remotes::install_bitbucket()}}, \code{\link[remotes:install_cran]{remotes::install_cran()}}, \code{\link[remotes:install_dev]{remotes::install_dev()}},
\code{\link[remotes:install_git]{remotes::install_git()}}, \code{\link[remotes:install_github]{remotes::install_github()}}, \code{\link[remotes:install_gitlab]{remotes::install_gitlab()}}, \code{\link[remotes:install_local]{remotes::install_local()}},
\code{\link[remotes:install_svn]{remotes::install_svn()}}, \code{\link[remotes:install_url]{remotes::install_url()}}, \code{\link[remotes:install_version]{remotes::install_version()}}, \code{\link[remotes:update_packages]{remotes::update_packages()}},
\code{\link[remotes:package_deps]{remotes::dev_package_deps()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bash.R
\name{bash}
\alias{bash}
\title{Open bash shell in package directory}
\usage{
bash(pkg = ".")
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{bash()} is deprecated because we no longer use or recommend this workflow.
Open bash shell in package directory
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/with-debug.r
\name{with_debug}
\alias{with_debug}
\title{Temporarily set debugging compilation flags.}
\usage{
with_debug(code, CFLAGS = NULL, CXXFLAGS = NULL, FFLAGS = NULL,
  FCFLAGS = NULL, debug = TRUE)
}
\arguments{
\item{code}{to execute.}

\item{CFLAGS}{flags for compiling C code}

\item{CXXFLAGS}{flags for compiling C++ code}

\item{FFLAGS}{flags for compiling Fortran code.}

\item{FCFLAGS}{flags for Fortran 9x code.}

\item{debug}{If \code{TRUE} adds \code{-g -O0} to all flags
(Adding \env{FFLAGS} and \env{FCFLAGS}}
}
\description{
Temporarily set debugging compilation flags.
}
\examples{
flags <- names(compiler_flags(TRUE))
with_debug(Sys.getenv(flags))

\dontrun{
install("mypkg")
with_debug(install("mypkg"))
}
}
\seealso{
Other debugging flags: \code{\link{compiler_flags}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build-readme.R
\name{build_rmd}
\alias{build_rmd}
\title{Build Rmarkdown files}
\usage{
build_rmd(files, path = ".", output_options = list(), ..., quiet = TRUE)
}
\arguments{
\item{files}{The Rmarkdown files to be rendered.}

\item{path}{path to the package to build the readme.}

\item{output_options}{List of output options that can override the options
specified in metadata (e.g. could be used to force \code{self_contained} or
\code{mathjax = "local"}). Note that this is only valid when the output
format is read from metadata (i.e. not a custom format object passed to
\code{output_format}).}

\item{...}{additional arguments passed to \code{\link[rmarkdown:render]{rmarkdown::render()}}}

\item{quiet}{If \code{TRUE}, suppress output.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{build_rmd()} is deprecated, as it is a low-level helper for internal use. To
render your package's \code{README.qmd} or \code{README.Rmd}, use \code{\link[=build_readme]{build_readme()}}. To
preview a vignette or article, use functions like \code{\link[pkgdown:build_site]{pkgdown::build_site()}} or
\code{\link[pkgdown:build_articles]{pkgdown::build_article()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lint.R
\name{lint}
\alias{lint}
\title{Lint all source files in a package}
\usage{
lint(pkg = ".", cache = TRUE, ...)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{cache}{Store the lint results so repeated lints of the same content use
the previous results. Consult the lintr package to learn more about its
caching behaviour.}

\item{...}{Additional arguments passed to \code{\link[lintr:lint]{lintr::lint_package()}}.}
}
\description{
The default linters correspond to the style guide at
\url{https://style.tidyverse.org/}, however it is possible to override any or all
of them using the \code{linters} parameter.
}
\seealso{
\code{\link[lintr:lint]{lintr::lint_package()}}, \code{\link[lintr:lint]{lintr::lint()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run-loadhooks.r
\name{run_pkg_hook}
\alias{run_pkg_hook}
\alias{run_user_hook}
\title{Run user and package hooks.}
\usage{
run_pkg_hook(pkg, hook)

run_user_hook(pkg, hook)
}
\arguments{
\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information}

\item{hook}{hook name: one of "load", "unload", "attach", or "detach"}
}
\description{
Run user and package hooks.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imports-env.r
\name{imports_env}
\alias{imports_env}
\title{Return imports environment for a package}
\usage{
imports_env(pkg = ".")
}
\arguments{
\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information.}
}
\description{
Contains objects imported from other packages. Is the parent of the
package namespace environment, and is a child of <namespace:base>,
which is a child of R_GlobalEnv.
}
\seealso{
\code{\link{ns_env}} for the namespace environment that
  all the objects (exported and not exported).

\code{\link{pkg_env}} for the attached environment that contains
  the exported objects.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-mode.R
\name{dev_mode}
\alias{dev_mode}
\title{Activate and deactivate development mode}
\usage{
dev_mode(on = NULL, path = getOption("devtools.path"))
}
\arguments{
\item{on}{turn dev mode on (\code{TRUE}) or off (\code{FALSE}).  If omitted will guess
based on whether or not \code{path} is in \code{\link[=.libPaths]{.libPaths()}}}

\item{path}{directory to library.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

We no longer recommend \code{dev_mode()} and it will be removed in a future
release of devtools. Instead, we now rely on \code{\link[=load_all]{load_all()}} to test drive an
in-development package. If you really like the idea of corralling
experimental packages in a special library, you might enjoy
\code{withr::local_libpaths()}. If you are concerned about different projects
interfering with each other through the use of a shared library, consider
using the \href{https://rstudio.github.io/renv/}{renv package}.

Original description: When activated, \code{dev_mode} creates a new library for
storing installed packages. This new library is automatically created when
\code{dev_mode} is activated if it does not already exist. This allows you to test
development packages in a sandbox, without interfering with the other
packages you have installed.
}
\examples{
\dontrun{
dev_mode()
dev_mode()
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build-manual.R
\name{build_manual}
\alias{build_manual}
\title{Create package pdf manual}
\usage{
build_manual(pkg = ".", path = NULL)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{path}{path in which to produce package manual.
If \code{NULL}, defaults to the parent directory of the package.}
}
\description{
Create package pdf manual
}
\seealso{
\code{\link[=Rd2pdf]{Rd2pdf()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install.R
\name{install}
\alias{install}
\title{Install a local development package}
\usage{
install(
  pkg = ".",
  reload = TRUE,
  quick = FALSE,
  build = !quick,
  args = getOption("devtools.install.args"),
  quiet = FALSE,
  dependencies = NA,
  upgrade = FALSE,
  build_vignettes = FALSE,
  keep_source = getOption("keep.source.pkgs") || !build,
  force = deprecated()
)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{reload}{if \code{TRUE} (the default), will automatically attempt to reload
the package after installing. Reloading is not always completely possible
so see \code{\link[pkgload:unload]{pkgload::unregister()}} for caveats.}

\item{quick}{if \code{TRUE}, skips some optional steps (e.g. help
pre-rendering and multi-arch builds) to make installation as fast
as possible.}

\item{build}{If \code{TRUE} (the default), first \code{\link[pkgbuild:build]{pkgbuild::build()}}s the
package. This ensures that the installation is completely clean, and
prevents any binary artefacts (like \file{.o}, \code{.so}) from appearing
in your local package directory, but is considerably slower, because
every compile has to start from scratch.

One downside of installing from a built tarball is that the package is
installed from a temporary location. This means that any source references
will point to dangling locations and debuggers won't have direct access to
the source for step-debugging. For development purposes, \code{build = FALSE} is
often the better choice.

If \code{FALSE}, the package is installed directly from its source
directory. This is faster and can be favorable for preserving source
references for debugging (see \code{keep_source}).}

\item{args}{An optional character vector of additional command line
arguments to be passed to \verb{R CMD INSTALL}. This defaults to the
value of the option \code{"devtools.install.args"}.}

\item{quiet}{If \code{TRUE}, suppress output.}

\item{dependencies}{What kinds of dependencies to install. Most commonly
one of the following values:
\itemize{
\item \code{NA}: only required (hard) dependencies,
\item \code{TRUE}: required dependencies plus optional and development
dependencies,
\item \code{FALSE}: do not install any dependencies. (You might end up with a
non-working package, and/or the installation might fail.)
See \link[pak]{Package dependency types} for other possible values and more
information about package dependencies.
}}

\item{upgrade}{When \code{FALSE}, the default, pak does the minimum amount
of work to give you the latest version(s) of \code{pkg}. It will only upgrade
dependent packages if \code{pkg}, or one of their dependencies explicitly
require a higher version than what you currently have. It will also
prefer a binary package over to source package, even it the binary
package is older.

When \code{upgrade = TRUE}, pak will ensure that you have the latest
version(s) of \code{pkg} and all their dependencies.}

\item{build_vignettes}{if \code{TRUE}, will build vignettes. Normally it is
\code{build} that's responsible for creating vignettes; this argument makes
sure vignettes are built even if a build never happens (i.e. because
\code{build = FALSE}).}

\item{keep_source}{If \code{TRUE} will keep the srcrefs from an installed
package. This is useful for debugging (especially inside of RStudio or
Positron). Defaults to \code{getOption("keep.source.pkgs") || !build},
since srcrefs are most useful when the package is installed from its
source directory, i.e. when \code{build = FALSE}.}

\item{force}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} No longer used.}
}
\description{
Uses \verb{R CMD INSTALL} to install the package, after installing needed
dependencies with \code{\link[pak:local_install_deps]{pak::local_install_deps()}}.

To install to a non-default library, use \code{\link[withr:with_libpaths]{withr::with_libpaths()}}.
}
\seealso{
\code{\link[=with_debug]{with_debug()}} to install packages with debugging flags set.

Other package installation: 
\code{\link{uninstall}()}
}
\concept{package installation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/uninstall.R
\name{uninstall}
\alias{uninstall}
\title{Uninstall a local development package.}
\usage{
uninstall(pkg = ".", unload = TRUE, quiet = FALSE, lib = .libPaths()[[1]])
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{unload}{if \code{TRUE} (the default), will automatically unload the
package prior to uninstalling.}

\item{quiet}{If \code{TRUE}, suppress output.}

\item{lib}{a character vector giving the library directories to remove the
    packages from.  If missing, defaults to the first element in
    \code{\link{.libPaths}()}.}
}
\description{
Uses \code{remove.package} to uninstall the package.
To uninstall a package from a non-default library,
use \code{\link[withr:with_libpaths]{withr::with_libpaths()}}.
}
\seealso{
\code{\link[=with_debug]{with_debug()}} to install packages with debugging flags
set.

Other package installation: 
\code{\link{install}()}
}
\concept{package installation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unload.r
\name{unload}
\alias{unload}
\title{Unload a package}
\usage{
unload(pkg = ".")
}
\arguments{
\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information}
}
\description{
This function attempts to cleanly unload a package, including unloading
its namespace, deleting S4 class definitions and unloading any loaded
DLLs. Unfortunately S4 classes are not really designed to be cleanly
unloaded, and so we have to manually modify the class dependency graph in
order for it to work - this works on the cases for which we have tested
but there may be others.  Similarly, automated DLL unloading is best tested
for simple scenarios (particularly with \code{useDynLib(pkgname)} and may
fail in other cases. If you do encounter a failure, please file a bug report
at \url{http://github.com/hadley/devtools/issues}.
}
\examples{
\dontrun{
# Unload package that is in current directory
unload(".")

# Unload package that is in ./ggplot2/
unload("ggplot2/")

# Can use inst() to find the path of an installed package
# This will load and unload the installed ggplot2 package
library(ggplot2)
unload(inst("ggplot2"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inst.R, R/package-deps.R, R/pkgbuild.R,
%   R/pkgload.R, R/remotes.R, R/session-info.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{inst}
\alias{parse_deps}
\alias{check_dep_version}
\alias{with_debug}
\alias{clean_dll}
\alias{has_devel}
\alias{find_rtools}
\alias{unload}
\alias{github_pull}
\alias{github_release}
\alias{session_info}
\alias{package_info}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{pkgbuild}{\code{\link[pkgbuild]{clean_dll}}, \code{\link[pkgbuild:has_rtools]{find_rtools}}, \code{\link[pkgbuild:has_compiler]{has_devel}}, \code{\link[pkgbuild]{with_debug}}}

  \item{pkgload}{\code{\link[pkgload]{check_dep_version}}, \code{\link[pkgload]{inst}}, \code{\link[pkgload]{parse_deps}}, \code{\link[pkgload]{unload}}}

  \item{remotes}{\code{\link[remotes:github_refs]{github_pull}}, \code{\link[remotes:github_refs]{github_release}}}

  \item{sessioninfo}{\code{\link[sessioninfo]{package_info}}, \code{\link[sessioninfo]{session_info}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test.R
\name{uses_testthat}
\alias{uses_testthat}
\title{Does a package use testthat?}
\usage{
uses_testthat(pkg = ".")
}
\description{
Does a package use testthat?
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infrastructure-git.R
\name{use_git}
\alias{use_git}
\title{Initialise a git repository.}
\usage{
use_git(message = "Initial commit", pkg = ".")
}
\arguments{
\item{message}{Message to use for first commit.}

\item{pkg}{Path to package. See \code{\link{as.package}} for more
information.}
}
\description{
Initialise a git repository.
}
\examples{
\dontrun{use_git()}
}
\seealso{
Other git infrastructure: \code{\link{use_git_hook}},
  \code{\link{use_github_links}}, \code{\link{use_github}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/document.R
\name{document}
\alias{document}
\title{Use roxygen to document a package}
\usage{
document(pkg = ".", roclets = NULL, quiet = FALSE)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{roclets}{Character vector of roclet names to use with package.
The default, \code{NULL}, uses the roxygen \code{roclets} option,
which defaults to \code{c("collate", "namespace", "rd")}.}

\item{quiet}{if \code{TRUE} suppresses output from this function.}
}
\description{
This function is a wrapper for the \code{\link[roxygen2:roxygenize]{roxygen2::roxygenize()}}
function from the roxygen2 package. See the documentation and vignettes of
that package to learn how to use roxygen.
}
\seealso{
\code{\link[roxygen2:roxygenize]{roxygen2::roxygenize()}},
\code{browseVignettes("roxygen2")}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-info.r
\name{session_info}
\alias{session_info}
\title{Print session information}
\usage{
session_info(pkgs = NULL, include_base = FALSE)
}
\arguments{
\item{pkgs}{Either a vector of package names or NULL. If \code{NULL},
displays all loaded packages. If a character vector, also, includes
all dependencies of the package.}

\item{include_base}{Include base packages in summary? By default this is
false since base packages should always match the R version.}
}
\description{
This is \code{\link{sessionInfo}()} re-written from scratch to both exclude
data that's rarely useful (e.g., the full collate string or base packages
loaded) and include stuff you'd like to know (e.g., where a package was
installed from).
}
\examples{
session_info()
session_info("devtools")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r-hub.R
\name{check_rhub}
\alias{check_rhub}
\title{Run CRAN checks for package on R-hub}
\usage{
check_rhub(
  pkg = ".",
  platforms = NULL,
  email = NULL,
  interactive = TRUE,
  build_args = NULL,
  ...
)
}
\arguments{
\item{pkg}{The package to use, can be a file path to the package or a
package object.  See \code{\link[=as.package]{as.package()}} for more information.}

\item{platforms}{R-hub platforms to run the check on.}

\item{email}{email address to notify.}

\item{interactive}{whether to show the status of the build.}

\item{build_args}{Arguments passed to \verb{R CMD build}.}

\item{...}{extra arguments, passed to \code{rhub::check_for_cran()}.}
}
\value{
a \code{rhub_check} object.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

This function is deprecated since the underlying function
\code{rhub::check_for_cran()} is now deprecated and defunct. See \code{\link[rhub:rhubv2]{rhub::rhubv2}}
learn about the new check system, R-hub v2.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/with-debug.r
\name{compiler_flags}
\alias{compiler_flags}
\title{Default compiler flags used by devtools.}
\usage{
compiler_flags(debug = FALSE)
}
\arguments{
\item{debug}{If \code{TRUE} adds \code{-g -O0} to all flags
(Adding \env{FFLAGS} and \env{FCFLAGS}}
}
\description{
These default flags enforce good coding practice by ensuring that
\env{CFLAGS} and \env{CXXFLAGS} are set to \code{-Wall -pedantic}.
These tests are run by cran and are generally considered to be good practice.
}
\details{
By default \code{\link{compile_dll}} is run with \code{compiler_flags(TRUE)},
and check with \code{compiler_flags(FALSE)}.  If you want to avoid the
possible performance penalty from the debug flags, install the package.
}
\examples{
compiler_flags()
compiler_flags(TRUE)
}
\seealso{
Other debugging flags: \code{\link{with_debug}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load.r
\name{create_description}
\alias{create_description}
\title{Create a default DESCRIPTION file for a package.}
\usage{
create_description(path = ".", extra = getOption("devtools.desc"),
  quiet = FALSE)
}
\arguments{
\item{path}{path to package root directory}

\item{extra}{a named list of extra options to add to \file{DESCRIPTION}.
Arguments that take a list}

\item{quiet}{if \code{TRUE}, suppresses output from this function.}
}
\description{
Create a default DESCRIPTION file for a package.
}
\details{
To set the default author and licenses, set \code{options}
\code{devtools.desc.author} and \code{devtools.desc.license}.  I use
\code{options(devtools.desc.author = '"Hadley Wickham <h.wickham@gmail.com> [aut,cre]"',
  devtools.desc.license = "GPL-3")}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deps.R
\name{package_deps}
\alias{package_deps}
\alias{dev_package_deps}
\alias{update.package_deps}
\title{Find all dependencies of a CRAN or dev package.}
\usage{
package_deps(pkg, dependencies = NA, repos = getOption("repos"),
  type = getOption("pkgType"))

dev_package_deps(pkg = ".", dependencies = NA, repos = getOption("repos"),
  type = getOption("pkgType"), bioconductor = TRUE)

\method{update}{package_deps}(object, ..., quiet = FALSE, upgrade = TRUE)
}
\arguments{
\item{pkg}{A character vector of package names. If missing, defaults to
the name of the package in the current directory.}

\item{dependencies}{Which dependencies do you want to check?
  Can be a character vector (selecting from "Depends", "Imports",
   "LinkingTo", "Suggests", or "Enhances"), or a logical vector.

  \code{TRUE} is shorthand for "Depends", "Imports", "LinkingTo" and
  "Suggests". \code{NA} is shorthand for "Depends", "Imports" and "LinkingTo"
  and is the default. \code{FALSE} is shorthand for no dependencies (i.e.
  just check this package, not its dependencies).}

\item{repos}{A character vector giving repositories to use.}

\item{type}{Type of package to \code{update}.  If "both", will switch
automatically to "binary" to avoid interactive prompts during package
installation.}

\item{bioconductor}{Install Bioconductor dependencies if the package has a
BiocViews field in the DESCRIPTION.}

\item{object}{A \code{package_deps} object.}

\item{...}{Additional arguments passed to \code{install_packages}.}

\item{quiet}{If \code{TRUE}, suppress output.}

\item{upgrade}{If \code{TRUE}, also upgrade any of out date dependencies.}
}
\value{
A \code{data.frame} with columns:

\tabular{ll}{
\code{package} \tab The dependent package's name,\cr
\code{installed} \tab The currently installed version,\cr
\code{available} \tab The version available on CRAN,\cr
\code{diff} \tab An integer denoting whether the locally installed version
  of the package is newer (1), the same (0) or older (-1) than the version
  currently available on CRAN.\cr
}
}
\description{
Find all the dependencies of a package and determine whether they are ahead
or behind CRAN. A \code{print()} method identifies mismatches (if any)
between local and CRAN versions of each dependent package; an
\code{update()} method installs outdated or missing packages from CRAN.
}
\examples{
\dontrun{
package_deps("devtools")
# Use update to update any out-of-date dependencies
update(package_deps("devtools"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compile-dll.r
\name{compile_dll}
\alias{compile_dll}
\title{Compile a .dll/.so from source.}
\usage{
compile_dll(pkg = ".", quiet = FALSE)
}
\arguments{
\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information}

\item{quiet}{if \code{TRUE} suppresses output from this function.}
}
\description{
\code{compile_dll} performs a fake R CMD install so code that
works here should work with a regular install (and vice versa).
}
\details{
During compilation, debug flags are set with
\code{\link{compiler_flags}(TRUE)}.

Invisibly returns the names of the DLL.
}
\note{
If this is used to compile code that uses Rcpp, you will need to
  add the following line to your \code{Makevars} file so that it
  knows where to find the Rcpp headers:
  \code{PKG_CPPFLAGS=`$(R_HOME)/bin/Rscript -e 'Rcpp:::CxxFlags()'`}
}
\seealso{
\code{\link{clean_dll}} to delete the compiled files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clean.r
\name{eval_clean}
\alias{eval_clean}
\alias{evalq_clean}
\title{Evaluate code in a clean R session.}
\usage{
eval_clean(expr, quiet = TRUE)

evalq_clean(expr, quiet = TRUE)
}
\arguments{
\item{expr}{an R expression to evaluate. For \code{eval_clean} this should
already be quoted. For \code{evalq_clean} it will be quoted for you.}

\item{quiet}{if \code{TRUE}, the default, only the final result and the
any explicitly printed output will be displayed. If \code{FALSE}, all
input and output will be displayed, as if you'd copied and paste the code.}
}
\value{
An invisible \code{TRUE} on success.
}
\description{
Evaluate code in a clean R session.
}
\examples{
x <- 1
y <- 2
ls()
evalq_clean(ls())
evalq_clean(ls(), FALSE)
eval_clean(quote({
  z <- 1
  ls()
}))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-help.r
\name{help}
\alias{help}
\alias{shim_help}
\alias{?}
\alias{shim_question}
\title{Drop-in replacements for help and ? functions}
\usage{
# help(topic, package = NULL, ...)

# ?e2
# e1?e2
}
\arguments{
\item{topic}{A name or character string specifying the help topic.}

\item{package}{A name or character string specifying the package in which
to search for the help topic. If NULL, search all packages.}

\item{...}{Additional arguments to pass to \code{\link[utils]{help}}.}

\item{e1}{First argument to pass along to \code{utils::`?`}.}

\item{e2}{Second argument to pass along to \code{utils::`?`}.}
}
\description{
The \code{?} and \code{help} functions are replacements for functions of the
same name in the utils package. They are made available when a package is
loaded with \code{\link{load_all}}.
}
\details{
The \code{?} function is a replacement for \code{\link[utils]{?}} from the
utils package. It will search for help in devtools-loaded packages first,
then in regular packages.

The \code{help} function is a replacement for \code{\link[utils]{help}} from
the utils package. If \code{package} is not specified, it will search for
help in devtools-loaded packages first, then in regular packages. If
\code{package} is specified, then it will search for help in devtools-loaded
packages or regular packages, as appropriate.
}
\examples{
\dontrun{
# This would load devtools and look at the help for load_all, if currently
# in the devtools source directory.
load_all()
?load_all
help("load_all")
}

# To see the help pages for utils::help and utils::`?`:
help("help", "utils")
help("?", "utils")

\dontrun{
# Examples demonstrating the multiple ways of supplying arguments
# NB: you can't do pkg <- "ggplot2"; help("ggplot2", pkg)
help(lm)
help(lm, stats)
help(lm, 'stats')
help('lm')
help('lm', stats)
help('lm', 'stats')
help(package = stats)
help(package = 'stats')
topic <- "lm"
help(topic)
help(topic, stats)
help(topic, 'stats')
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package-env.r
\name{pkg_env}
\alias{pkg_env}
\title{Return package environment}
\usage{
pkg_env(pkg = ".")
}
\arguments{
\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information}
}
\description{
This is an environment like \code{<package:pkg>}. The package
environment contains the exported objects from a package. It is
attached, so it is an ancestor of \code{R_GlobalEnv}.
}
\details{
When a package is loaded the normal way, using \code{\link{library}},
this environment contains only the exported objects from the
namespace. However, when loaded with \code{\link{load_all}}, this
environment will contain all the objects from the namespace, unless
\code{load_all} is used with \code{export_all=FALSE}.

If the package is not attached, this function returns \code{NULL}.
}
\seealso{
\code{\link{ns_env}} for the namespace environment that
  all the objects (exported and not exported).

\code{\link{imports_env}} for the environment that contains
  imported objects for the package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infrastructure.R
\name{infrastructure}
\alias{infrastructure}
\alias{use_testthat}
\alias{use_test}
\alias{use_rstudio}
\alias{use_vignette}
\alias{use_rcpp}
\alias{use_travis}
\alias{add_travis}
\alias{use_coverage}
\alias{use_appveyor}
\alias{use_package_doc}
\alias{use_revdep}
\alias{add_travis}
\alias{use_cran_comments}
\alias{add_travis}
\alias{use_code_of_conduct}
\alias{add_travis}
\alias{use_cran_badge}
\alias{use_mit_license}
\alias{use_gpl3_license}
\alias{use_dev_version}
\title{Add useful infrastructure to a package.}
\usage{
use_testthat(pkg = ".")

use_test(name, pkg = ".")

use_rstudio(pkg = ".")

use_vignette(name, pkg = ".")

use_rcpp(pkg = ".")

use_travis(pkg = ".", browse = interactive())

use_coverage(pkg = ".", type = c("codecov", "coveralls"))

use_appveyor(pkg = ".")

use_package_doc(pkg = ".")

use_revdep(pkg = ".")

use_cran_comments(pkg = ".")

use_code_of_conduct(pkg = ".")

use_cran_badge(pkg = ".")

use_mit_license(pkg = ".", copyright_holder = getOption("devtools.name",
  "<Author>"))

use_gpl3_license(pkg = ".")

use_dev_version(pkg = ".")
}
\arguments{
\item{pkg}{package description, can be path or package name. See
\code{\link{as.package}} for more information.}

\item{name}{File name to use for new vignette. Should consist only of
numbers, letters, _ and -. I recommend using lower case.}

\item{browse}{open a browser window to enable Travis builds for the package
automatically.}

\item{type}{CI tool to use. Currently supports codecov and coverall.}

\item{copyright_holder}{The copyright holder for this package. Defaults to
\code{getOption("devtools.name")}.}
}
\description{
Add useful infrastructure to a package.
}
\section{\code{use_testthat}}{

Add testing infrastructure to a package that does not already have it.
This will create \file{tests/testthat.R}, \file{tests/testthat/} and
add \pkg{testthat} to the suggested packages. This is called
automatically from \code{\link{test}} if needed.
}

\section{\code{use_test}}{

Add a test file, also add testing infrastructure if necessary.
This will create \file{tests/testthat/test-<name>.R} with a user-specified
name for the test.  Will fail if the file exists.
}

\section{\code{use_vignette}}{

Adds needed packages to \code{DESCRIPTION}, and creates draft vignette
in \code{vignettes/}. It adds \code{inst/doc} to \code{.gitignore}
so you don't accidentally check in the built vignettes.
}

\section{\code{use_rcpp}}{

Creates \code{src/} and adds needed packages to \code{DESCRIPTION}.
}

\section{\code{use_travis}}{

Add basic travis template to a package. Also adds \code{.travis.yml} to
\code{.Rbuildignore} so it isn't included in the built package.
}

\section{\code{use_coverage}}{

Add test code coverage to basic travis template to a package.
}

\section{\code{use_appveyor}}{

Add basic AppVeyor template to a package. Also adds \code{appveyor.yml} to
\code{.Rbuildignore} so it isn't included in the built package.
}

\section{\code{use_package_doc}}{

Adds a roxygen template for package documentation
}

\section{\code{use_revdep}}{

Add \code{revdep} directory and basic check template.
}

\section{\code{use_cran_comments}}{

Add \code{cran-comments.md} template.
}

\section{\code{use_code_of_conduct}}{

Add a code of conduct to from \url{http://contributor-covenant.org}.
}

\section{\code{use_cran_badge}}{

Add a badge to show CRAN status and version number on the README
}

\section{\code{use_mit_license}}{

Adds the necessary infrastructure to declare your package as
distributed under the MIT license.
}

\section{\code{use_gpl3_license}}{

Adds the necessary infrastructure to declare your package as
distributed under the GPL v3.
}

\section{\code{use_dev_version}}{

This adds ".9000" to the package \code{DESCRIPTION}, adds a new heading to
\code{NEWS.md} (if it exists), and then checks the result into git.
}

\seealso{
Other infrastructure: \code{\link{use_build_ignore}},
  \code{\link{use_data_raw}}, \code{\link{use_data}},
  \code{\link{use_news_md}}, \code{\link{use_package}},
  \code{\link{use_readme_rmd}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install-github.r
\name{install_github}
\alias{install_github}
\title{Attempts to install a package directly from GitHub.}
\usage{
install_github(repo, username = NULL, ref = "master", subdir = NULL,
  auth_token = github_pat(quiet), host = "https://api.github.com",
  quiet = FALSE, ...)
}
\arguments{
\item{repo}{Repository address in the format
\code{username/repo[/subdir][@ref|#pull]}. Alternatively, you can
specify \code{subdir} and/or \code{ref} using the respective parameters
(see below); if both are specified, the values in \code{repo} take
precedence.}

\item{username}{User name. Deprecated: please include username in the
\code{repo}}

\item{ref}{Desired git reference. Could be a commit, tag, or branch
name, or a call to \code{\link{github_pull}}. Defaults to \code{"master"}.}

\item{subdir}{subdirectory within repo that contains the R package.}

\item{auth_token}{To install from a private repo, generate a personal
access token (PAT) in \url{https://github.com/settings/tokens} and
supply to this argument. This is safer than using a password because
you can easily delete a PAT without affecting any others. Defaults to
the \code{GITHUB_PAT} environment variable.}

\item{host}{GitHub API host to use. Override with your GitHub enterprise
hostname, for example, \code{"github.hostname.com/api/v3"}.}

\item{quiet}{if \code{TRUE} suppresses output from this function.}

\item{...}{Other arguments passed on to \code{\link{install}}.}
}
\description{
This function is vectorised on \code{repo} so you can install multiple
packages in a single command.
}
\details{
Attempting to install from a source repository that uses submodules
raises a warning. Because the zipped sources provided by GitHub do not
include submodules, this may lead to unexpected behaviour or compilation
failure in source packages. In this case, cloning the repository manually
using \code{\link{install_git}} with \code{args="--recursive"} may yield
better results.
}
\examples{
\dontrun{
install_github("klutometis/roxygen")
install_github("wch/ggplot2")
install_github(c("rstudio/httpuv", "rstudio/shiny"))
install_github(c("hadley/httr@v0.4", "klutometis/roxygen#142",
  "mfrasca/r-logging/pkg"))

# Update devtools to the latest version, on Linux and Mac
# On Windows, this won't work - see ?build_github_devtools
install_github("hadley/devtools")

# To install from a private repo, use auth_token with a token
# from https://github.com/settings/tokens. You only need the
# repo scope. Best practice is to save your PAT in env var called
# GITHUB_PAT.
install_github("hadley/private", auth_token = "abc")

}
}
\seealso{
\code{\link{github_pull}}

Other package installation: \code{\link{install_bioc}},
  \code{\link{install_bitbucket}},
  \code{\link{install_cran}}, \code{\link{install_git}},
  \code{\link{install_svn}}, \code{\link{install_url}},
  \code{\link{install_version}}, \code{\link{install}},
  \code{\link{uninstall}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compile-dll.r
\name{clean_dll}
\alias{clean_dll}
\title{Remove compiled objects from /src/ directory}
\usage{
clean_dll(pkg = ".")
}
\arguments{
\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information}
}
\description{
Invisibly returns the names of the deleted files.
}
\seealso{
\code{\link{compile_dll}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load-data.r
\name{load_data}
\alias{load_data}
\title{Load data.}
\usage{
load_data(pkg = ".")
}
\arguments{
\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information}
}
\description{
Loads all \code{.RData} files in the data subdirectory.
}
\keyword{programming}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infrastructure.R
\name{use_data_raw}
\alias{use_data_raw}
\title{Use \code{data-raw} to compute package datasets.}
\usage{
use_data_raw(pkg = ".")
}
\arguments{
\item{pkg}{Package where to create \code{data-raw}. Defaults to package in
working directory.}
}
\description{
Use \code{data-raw} to compute package datasets.
}
\seealso{
Other infrastructure: \code{\link{infrastructure}},
  \code{\link{use_build_ignore}}, \code{\link{use_data}},
  \code{\link{use_news_md}}, \code{\link{use_package}},
  \code{\link{use_readme_rmd}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load-dll.r
\name{load_dll}
\alias{load_dll}
\title{Load a compiled DLL}
\usage{
load_dll(pkg = ".")
}
\arguments{
\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information}
}
\description{
Load a compiled DLL
}
\keyword{programming}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infrastructure-git.R
\name{use_git_hook}
\alias{use_git_hook}
\title{Add a git hook.}
\usage{
use_git_hook(hook, script, pkg = ".")
}
\arguments{
\item{hook}{Hook name. One of "pre-commit", "prepare-commit-msg",
"commit-msg", "post-commit", "applypatch-msg", "pre-applypatch",
"post-applypatch", "pre-rebase", "post-rewrite", "post-checkout",
"post-merge", "pre-push", "pre-auto-gc".}

\item{script}{Text of script to run}

\item{pkg}{Path to package. See \code{\link{as.package}} for more
information.}
}
\description{
Add a git hook.
}
\seealso{
Other git infrastructure: \code{\link{use_github_links}},
  \code{\link{use_github}}, \code{\link{use_git}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgdown-test-test.R
\name{pkgdown_test_test}
\alias{pkgdown_test_test}
\title{pkgdown_test_test}
\usage{
pkgdown_test_test(x)
}
\arguments{
\item{x}{marks the spot}
}
\value{
FALSE
}
\description{
pkgdown_test_test
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path.r
\name{on_path}
\alias{on_path}
\title{Test if an object is on the path.}
\usage{
on_path(...)
}
\arguments{
\item{...}{Strings indicating the executables to check for on the path.}
}
\description{
Test if an object is on the path.
}
\examples{
on_path("R")
on_path("gcc")
on_path("foo", "bar")  # FALSE in most cases
withr::with_path(tempdir(), on_path("gcc"))
}
\seealso{
Other path: \code{\link{path}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install-git.r
\name{install_git}
\alias{install_git}
\title{Install a package from a git repository}
\usage{
install_git(url, subdir = NULL, branch = NULL, credentials = NULL,
  quiet = FALSE, ...)
}
\arguments{
\item{url}{Location of package. The url should point to a public or
private repository.}

\item{subdir}{A sub-directory within a git repository that may
contain the package we are interested in installing.}

\item{branch}{Name of branch or tag to use, if not master.}

\item{credentials}{A git2r credentials object passed through
to \code{\link[git2r]{clone}}.}

\item{quiet}{if \code{TRUE} suppresses output from this function.}

\item{...}{passed on to \code{\link{install}}}
}
\description{
It is vectorised so you can install multiple packages with
a single command. You do not need to have git installed.
}
\examples{
\dontrun{
install_git("git://github.com/hadley/stringr.git")
install_git("git://github.com/hadley/stringr.git", branch = "stringr-0.2")
}
}
\seealso{
Other package installation: \code{\link{install_bioc}},
  \code{\link{install_bitbucket}},
  \code{\link{install_cran}}, \code{\link{install_github}},
  \code{\link{install_svn}}, \code{\link{install_url}},
  \code{\link{install_version}}, \code{\link{install}},
  \code{\link{uninstall}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/foofoo.r
\name{foofoo}
\alias{foofoo}
\title{Test function for help}
\usage{
foofoo()
}
\description{
The purpose of this function is to test out \code{help} and \code{?} from
devtools.
}
\examples{
stopifnot(foofoo() == 'You called foofoo.')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deps.R
\name{update_packages}
\alias{update_packages}
\title{Update packages that are missing or out-of-date.}
\usage{
update_packages(pkgs = NULL, dependencies = NA,
  repos = getOption("repos"), type = getOption("pkgType"))
}
\arguments{
\item{pkgs}{Character vector of packages to update. IF \code{TRUE} all
installed packages are updated. If \code{NULL} user is prompted to
confirm update of all installed packages.}

\item{dependencies}{Which dependencies do you want to check?
  Can be a character vector (selecting from "Depends", "Imports",
   "LinkingTo", "Suggests", or "Enhances"), or a logical vector.

  \code{TRUE} is shorthand for "Depends", "Imports", "LinkingTo" and
  "Suggests". \code{NA} is shorthand for "Depends", "Imports" and "LinkingTo"
  and is the default. \code{FALSE} is shorthand for no dependencies (i.e.
  just check this package, not its dependencies).}

\item{repos}{A character vector giving repositories to use.}

\item{type}{Type of package to \code{update}.  If "both", will switch
automatically to "binary" to avoid interactive prompts during package
installation.}
}
\description{
Works similarly to \code{install.packages()} but doesn't install packages
that are already installed, and also upgrades out dated dependencies.
}
\examples{
\dontrun{
update_packages("ggplot2")
update_packages(c("plyr", "ggplot2"))
}
}
\seealso{
\code{\link{package_deps}} to see which packages are out of date/
  missing.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rtools.r
\name{setup_rtools}
\alias{setup_rtools}
\alias{find_rtools}
\title{Find rtools.}
\usage{
setup_rtools(cache = TRUE, debug = FALSE)
}
\arguments{
\item{cache}{if \code{TRUE} will used cached version of RTools.}

\item{debug}{if \code{TRUE} prints a lot of additional information to
help in debugging.}
}
\value{
Either a visible \code{TRUE} if rtools is found, or an invisible
  \code{FALSE} with a diagnostic \code{\link{message}}.
  As a side-effect the internal package variable \code{rtools_path} is
  updated to the paths to rtools binaries.
}
\description{
To build binary packages on windows, Rtools (found at
\url{https://cran.r-project.org/bin/windows/Rtools/}) needs to be on
the path. The default installation process does not add it, so this
script finds it (looking first on the path, then in the registry).
It also checks that the version of rtools matches the version of R.
}
\section{Acknowledgements}{

  This code borrows heavily from RStudio's code for finding rtools.
  Thanks JJ!
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/revdep-summarise.R, R/revdep.R
\name{revdep_check_save_summary}
\alias{revdep_check_save_summary}
\alias{revdep_check_print_problems}
\alias{revdep_check}
\alias{revdep_check_resume}
\alias{revdep_check_reset}
\title{Run R CMD check on all downstream dependencies.}
\usage{
revdep_check_save_summary(pkg = ".")

revdep_check_print_problems(pkg = ".")

revdep_check(pkg = ".", recursive = FALSE, ignore = NULL,
  dependencies = c("Depends", "Imports", "Suggests", "LinkingTo"),
  skip = character(), libpath = getOption("devtools.revdep.libpath"),
  srcpath = libpath, bioconductor = FALSE, type = getOption("pkgType"),
  threads = getOption("Ncpus", 1), env_vars = NULL, check_dir = NULL,
  install_dir = NULL, quiet_check = TRUE)

revdep_check_resume(pkg = ".", ...)

revdep_check_reset(pkg = ".")
}
\arguments{
\item{pkg}{Path to package. Defaults to current directory.}

\item{recursive}{If \code{TRUE} look for full set of recursive dependencies.}

\item{ignore}{A character vector of package names to ignore. These packages
will not appear in returned vector. This is used in
\code{\link{revdep_check}} to avoid packages with installation problems
or extremely long check times.}

\item{dependencies}{A character vector listing the types of dependencies
to follow.}

\item{skip}{A character vector of package names to exclude from the
checks.}

\item{libpath}{Path to library to store dependencies packages - if you
you're doing this a lot it's a good idea to pick a directory and stick
with it so you don't have to download all the packages every time.}

\item{srcpath}{Path to directory to store source versions of dependent
packages - again, this saves a lot of time because you don't need to
redownload the packages every time you run the package.}

\item{bioconductor}{If \code{TRUE} also look for dependencies amongst
bioconductor packages.}

\item{type}{binary Package type to test (source, mac.binary etc). Defaults
to the same type as \code{\link{install.packages}()}.}

\item{threads}{Number of concurrent threads to use for checking.
It defaults to the option \code{"Ncpus"} or \code{1} if unset.}

\item{env_vars}{Environment variables set during \code{R CMD check}}

\item{check_dir}{A temporary directory to hold the results of the package
checks. This should not exist as after the revdep checks complete
successfully this directory is blown away.}

\item{install_dir}{Directory to store check and installation
results.}

\item{quiet_check}{If \code{TRUE}, suppresses individual \code{R CMD
check} output and only prints summaries. Set to \code{FALSE} for
debugging.}

\item{...}{Optionally, override original value of arguments to
\code{revdep_check}. Use with care.}
}
\value{
An invisible list of results. But you'll probably want to look
  at the check results on disk, which are saved in \code{check_dir}.
  Summaries of all ERRORs and WARNINGs will be stored in
  \code{check_dir/00check-summary.txt}.
}
\description{
Use \code{revdep_check()} to run \code{\link{check_cran}()} on all downstream
dependencies. Summarises the results with \code{revdep_check_summary()} and
see problems with \code{revdep_check_print_problems()}.
}
\details{
Revdep checks are resumable - this is very helpful if somethings goes
wrong (like you run out of power or you lose your internet connection) in
the middle of a check. You can resume a partially completed check with
\code{revdep_check_resume()}, or blow away the cached result so you can
start afresh with \code{revdep_check_reset()}.
}
\section{Check process}{

\enumerate{
\item Install \code{pkg} (in special library, see below).
\item Find all CRAN packages that depend on \code{pkg}.
\item Install those packages, along with their dependencies.
\item Run \code{R CMD check} on each package.
\item Uninstall \code{pkg} (so other reverse dependency checks don't
  use the development version instead of the CRAN version)
}
}

\section{Package library}{

By default \code{revdep_check} uses a temporary library to store any packages
that are required by the packages being tested. This ensures that they don't
interfere with your default library, but means that if you restart R
between checks, you'll need to reinstall all the packages. If you're
doing reverse dependency checks frequently, I recommend that you create
a directory for these packages and set \code{options(devtools.revdep.libpath)}.
}

\examples{
\dontrun{
# Run R CMD check on all downstream dependencies
revdep_check()
revdep_check_save_summary()
revdep_check_print_problems()
}
}
\seealso{
\code{\link{revdep_maintainers}()} to get a list of all revdep
  maintainers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shims.r
\name{system.file}
\alias{system.file}
\alias{shim_system.file}
\title{Replacement version of system.file}
\usage{
# system.file(..., package = "base", lib.loc = NULL, mustWork = FALSE)
}
\arguments{
\item{...}{character vectors, specifying subdirectory and file(s)
    within some package.  The default, none, returns the
    root of the package.  Wildcards are not supported.}

\item{package}{a character string with the name of a single package.
    An error occurs if more than one package name is given.}

\item{lib.loc}{a character vector with path names of \R libraries.
    See \sQuote{Details} for the meaning of the default value of \code{NULL}.}

\item{mustWork}{logical.  If \code{TRUE}, an error is given if there
    are no matching files.}
}
\description{
This function is meant to intercept calls to \code{\link[base]{system.file}},
so that it behaves well with packages loaded by devtools. It is made
available when a package is loaded with \code{\link{load_all}}.
}
\details{
When \code{system.file} is called from the R console (the global
envrironment), this function detects if the target package was loaded with
\code{\link{load_all}}, and if so, it uses a customized method of searching
for the file. This is necessary because the directory structure of a source
package is different from the directory structure of an installed package.

When a package is loaded with \code{load_all}, this function is also inserted
into the package's imports environment, so that calls to \code{system.file}
from within the package namespace will use this modified version. If this
function were not inserted into the imports environment, then the package
would end up calling \code{base::system.file} instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install-local.r
\name{install_local}
\alias{install_local}
\title{Install a package from a local file}
\usage{
install_local(path, subdir = NULL, ..., quiet = FALSE)
}
\arguments{
\item{path}{path to local directory, or compressed file (tar, zip, tar.gz
tar.bz2, tgz2 or tbz)}

\item{subdir}{subdirectory within url bundle that contains the R package.}

\item{...}{Other arguments passed on to \code{\link{install}}.}

\item{quiet}{if \code{TRUE} suppresses output from this function.}
}
\description{
This function is vectorised so you can install multiple packages in
a single command.
}
\examples{
\dontrun{
dir <- tempfile()
dir.create(dir)
pkg <- download.packages("testthat", dir, type = "source")
install_local(pkg[, 2])
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install-bitbucket.r
\name{install_bitbucket}
\alias{install_bitbucket}
\title{Install a package directly from bitbucket}
\usage{
install_bitbucket(repo, username, ref = "master", subdir = NULL,
  quiet = FALSE, auth_user = NULL, password = NULL, ...)
}
\arguments{
\item{repo}{Repository address in the format
\code{username/repo[/subdir][@ref|#pull]}. Alternatively, you can
specify \code{subdir} and/or \code{ref} using the respective parameters
(see below); if both are specified, the values in \code{repo} take
precedence.}

\item{username}{User name. Deprecated: please include username in the
\code{repo}}

\item{ref}{Desired git reference; could be a commit, tag, or branch name.
Defaults to master.}

\item{subdir}{subdirectory within repo that contains the R package.}

\item{quiet}{if \code{TRUE} suppresses output from this function.}

\item{auth_user}{your account username if you're attempting to install
a package hosted in a private repository (and your username is different
to \code{username})}

\item{password}{your password}

\item{...}{Other arguments passed on to \code{\link{install}}.}
}
\description{
This function is vectorised so you can install multiple packages in
a single command.
}
\examples{
\dontrun{
install_bitbucket("sulab/mygene.r@default")
install_bitbucket("dannavarro/lsr-package")
}
}
\seealso{
Bitbucket API docs:
  \url{https://confluence.atlassian.com/bitbucket/use-the-bitbucket-cloud-rest-apis-222724129.html}

Other package installation: \code{\link{install_bioc}},
  \code{\link{install_cran}}, \code{\link{install_github}},
  \code{\link{install_git}}, \code{\link{install_svn}},
  \code{\link{install_url}}, \code{\link{install_version}},
  \code{\link{install}}, \code{\link{uninstall}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path.r
\name{path}
\alias{path}
\alias{get_path}
\alias{set_path}
\alias{add_path}
\title{Get/set the PATH variable.}
\usage{
get_path()

set_path(path)

add_path(path, after = Inf)
}
\arguments{
\item{path}{character vector of paths}

\item{after}{for \code{add_path}, the place on the PATH where the new paths
should be added}
}
\value{
\code{set_path} invisibly returns the old path.
}
\description{
Get/set the PATH variable.
}
\examples{
path <- get_path()
length(path)
old <- add_path(".")
length(get_path())
set_path(old)
length(get_path())
}
\seealso{
\code{\link[withr]{with_path}} to temporarily set the path for a block
  of code

Other path: \code{\link{on_path}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build.r
\name{build_win}
\alias{build_win}
\title{Build windows binary package.}
\usage{
build_win(pkg = ".", version = c("R-release", "R-devel"), args = NULL,
  quiet = FALSE)
}
\arguments{
\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information}

\item{version}{directory to upload to on the win-builder, controlling
which version of R is used to build the package. Possible options are
listed on \url{http://win-builder.r-project.org/}. Defaults to R-devel.}

\item{args}{An optional character vector of additional command
line arguments to be passed to \code{R CMD build} if \code{binary = FALSE},
or \code{R CMD install} if \code{binary = TRUE}.}

\item{quiet}{if \code{TRUE} suppresses output from this function.}
}
\description{
This function works by bundling source package, and then uploading to
\url{http://win-builder.r-project.org/}.  Once building is complete you'll
receive a link to the built package in the email address listed in the
maintainer field.  It usually takes around 30 minutes. As a side effect,
win-build also runs \code{R CMD check} on the package, so \code{build_win}
is also useful to check that your package is ok on windows.
}
\seealso{
Other build functions: \code{\link{build}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/namespace-env.r
\name{parse_ns_file}
\alias{parse_ns_file}
\title{Parses the NAMESPACE file for a package}
\usage{
parse_ns_file(pkg = ".")
}
\arguments{
\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information}
}
\description{
Parses the NAMESPACE file for a package
}
\examples{
if (has_tests()) {
parse_ns_file(devtest("testLoadHooks"))
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install-cran.r
\name{install_cran}
\alias{install_cran}
\title{Attempts to install a package from CRAN.}
\usage{
install_cran(pkgs, repos = getOption("repos"), type = getOption("pkgType"),
  ..., quiet = FALSE)
}
\arguments{
\item{pkgs}{Character vector of packages to install.}

\item{repos}{A character vector giving repositories to use.}

\item{type}{Type of package to \code{update}.  If "both", will switch
automatically to "binary" to avoid interactive prompts during package
installation.}

\item{...}{Additional arguments passed to \code{install_packages}.}

\item{quiet}{If \code{TRUE}, suppress output.}
}
\description{
This function is vectorised on \code{pkgs} so you can install multiple
packages in a single command.
}
\examples{
\dontrun{
install_cran("ggplot2")
install_cran(c("httpuv", "shiny")
}
}
\seealso{
Other package installation: \code{\link{install_bioc}},
  \code{\link{install_bitbucket}},
  \code{\link{install_github}}, \code{\link{install_git}},
  \code{\link{install_svn}}, \code{\link{install_url}},
  \code{\link{install_version}}, \code{\link{install}},
  \code{\link{uninstall}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imports-env.r
\name{load_imports}
\alias{load_imports}
\title{Load all of the imports for a package}
\usage{
load_imports(pkg = ".")
}
\description{
The imported objects are copied to the imports environment, and are not
visible from R_GlobalEnv. This will automatically load (but not attach)
the dependency packages.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/topic-index.r
\name{find_topic}
\alias{find_topic}
\title{Find the rd file that documents a topic.}
\usage{
find_topic(topic)
}
\arguments{
\item{topic}{The topic, a string.}
}
\value{
A named string. The values gives the path to file; the name gives
  the path to package.
}
\description{
Only packages loaded by devtools are searched.
}
\examples{
find_topic("help")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package-deps.r
\name{parse_deps}
\alias{parse_deps}
\title{Parse package dependency strings.}
\usage{
parse_deps(string)
}
\arguments{
\item{string}{to parse. Should look like \code{"R (>= 3.0), ggplot2"} etc.}
}
\value{
list of two character vectors: \code{name} package names,
  and \code{version} package versions. If version is not specified,
  it will be stored as NA.
}
\description{
Parse package dependency strings.
}
\examples{
parse_deps("httr (< 2.1),\\nRCurl (>= 3)")
# only package dependencies are returned
parse_deps("utils (== 2.12.1),\\ntools,\\nR (>= 2.10),\\nmemoise")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infrastructure.R
\name{use_news_md}
\alias{use_news_md}
\title{Use NEWS.md}
\usage{
use_news_md(pkg = ".")
}
\arguments{
\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information}
}
\description{
This creates \code{NEWS.md} from a template.
}
\seealso{
Other infrastructure: \code{\link{infrastructure}},
  \code{\link{use_build_ignore}},
  \code{\link{use_data_raw}}, \code{\link{use_data}},
  \code{\link{use_package}}, \code{\link{use_readme_rmd}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-cran.r
\name{check_cran}
\alias{check_cran}
\title{Check a package from CRAN.}
\usage{
check_cran(pkgs, libpath = file.path(tempdir(), "R-lib"), srcpath = libpath,
  check_libpath = libpath, bioconductor = FALSE,
  type = getOption("pkgType"), threads = getOption("Ncpus", 1),
  check_dir = tempfile("check_cran"),
  install_dir = tempfile("check_cran_install"), env_vars = NULL,
  quiet_check = TRUE)
}
\arguments{
\item{pkgs}{Vector of package names - note that unlike other \pkg{devtools}
functions this is the name of a CRAN package, not a path.}

\item{libpath}{Path to library to store dependencies packages - if you
you're doing this a lot it's a good idea to pick a directory and stick
with it so you don't have to download all the packages every time.}

\item{srcpath}{Path to directory to store source versions of dependent
packages - again, this saves a lot of time because you don't need to
redownload the packages every time you run the package.}

\item{check_libpath}{Path to library used for checking, should contain
the top-level library from \code{libpath}.}

\item{bioconductor}{Include bioconductor packages in checking?}

\item{type}{binary Package type to test (source, mac.binary etc). Defaults
to the same type as \code{\link{install.packages}()}.}

\item{threads}{Number of concurrent threads to use for checking.
It defaults to the option \code{"Ncpus"} or \code{1} if unset.}

\item{check_dir, install_dir}{Directory to store check and installation
results.}

\item{env_vars}{Environment variables set during \code{R CMD check}}

\item{quiet_check}{If \code{TRUE}, suppresses individual \code{R CMD
check} output and only prints summaries. Set to \code{FALSE} for
debugging.}
}
\value{
Returns (invisibly) the directory where check results are stored.
}
\description{
Internal function used to power \code{\link{revdep_check}()}.
}
\details{
This function does not clean up after itself, but does work in a
session-specific temporary directory, so all files will be removed
when your current R session ends.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infrastructure.R
\name{use_build_ignore}
\alias{use_build_ignore}
\alias{add_build_ignore}
\title{Add a file to \code{.Rbuildignore}}
\usage{
use_build_ignore(files, escape = TRUE, pkg = ".")
}
\arguments{
\item{files}{Name of file.}

\item{escape}{If \code{TRUE}, the default, will escape \code{.} to
\code{\\.} and surround with \code{^} and \code{$}.}

\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information}
}
\value{
Nothing, called for its side effect.
}
\description{
\code{.Rbuildignore} has a regular expression on each line, but it's
usually easier to work with specific file names. By default, will (crudely)
turn a filename into a regular expression that will only match that
path. Repeated entries will be silently removed.
}
\seealso{
Other infrastructure: \code{\link{infrastructure}},
  \code{\link{use_data_raw}}, \code{\link{use_data}},
  \code{\link{use_news_md}}, \code{\link{use_package}},
  \code{\link{use_readme_rmd}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load-code.r
\name{load_code}
\alias{load_code}
\title{Load R code.}
\usage{
load_code(pkg = ".")
}
\arguments{
\item{pkg}{package description, can be path or package name.  See
\code{\link{as.package}} for more information}
}
\description{
Load all R code in the \code{R} directory. The first time the code is
loaded, \code{.onLoad} will be run if it exists.
}
\keyword{programming}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/doctor.R
\name{dr_github}
\alias{dr_github}
\title{Diagnose potential GitHub issues}
\usage{
dr_github(path = ".")
}
\arguments{
\item{path}{Path to repository to check. Defaults to current working
directory}
}
\description{
Diagnose potential GitHub issues
}
\examples{
\donttest{
dr_github()
}
}
\seealso{
Other doctors: \code{\link{dr_devtools}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inst.r
\name{inst}
\alias{inst}
\title{Get the installation path of a package}
\usage{
inst(name)
}
\arguments{
\item{name}{the name of a package.}
}
\description{
Given the name of a package, this returns a path to the installed
copy of the package, which can be passed to other devtools functions.
}
\details{
It searches for the package in \code{\link{.libPaths}()}. If multiple
dirs are found, it will return the first one.
}
\examples{
inst("devtools")
inst("grid")
\dontrun{
# Can be passed to other devtools functions
unload(inst("ggplot2"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-info.r
\name{dev_packages}
\alias{dev_packages}
\title{Return a vector of names of packages loaded by devtools}
\usage{
dev_packages()
}
\description{
Return a vector of names of packages loaded by devtools
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clean.r
\name{clean_source}
\alias{clean_source}
\title{Sources an R file in a clean environment.}
\usage{
clean_source(path, quiet = FALSE)
}
\arguments{
\item{path}{path to R script}

\item{quiet}{If \code{FALSE}, the default, all input and output will be
displayed, as if you'd copied and paste the code.  If \code{TRUE}
only the final result and the any explicitly printed output will be
displayed.}
}
\description{
Opens up a fresh R environment and sources file, ensuring that it works
independently of the current working environment.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install-url.r
\name{install_url}
\alias{install_url}
\title{Install a package from a url}
\usage{
install_url(url, subdir = NULL, config = list(), ..., quiet = FALSE)
}
\arguments{
\item{url}{location of package on internet. The url should point to a
zip file, a tar file or a bzipped/gzipped tar file.}

\item{subdir}{subdirectory within url bundle that contains the R package.}

\item{config}{additional configuration argument (e.g. proxy,
authentication) passed on to \code{\link[httr]{GET}}.}

\item{...}{Other arguments passed on to \code{\link{install}}.}

\item{quiet}{if \code{TRUE} suppresses output from this function.}
}
\description{
This function is vectorised so you can install multiple packages in
a single command.
}
\examples{
\dontrun{
install_url("https://github.com/hadley/stringr/archive/master.zip")
}
}
\seealso{
Other package installation: \code{\link{install_bioc}},
  \code{\link{install_bitbucket}},
  \code{\link{install_cran}}, \code{\link{install_github}},
  \code{\link{install_git}}, \code{\link{install_svn}},
  \code{\link{install_version}}, \code{\link{install}},
  \code{\link{uninstall}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system.r
\name{system_output}
\alias{system_output}
\title{Run a system command and capture the output.}
\usage{
system_output(cmd, args = character(), env_vars = character(), path = ".",
  quiet = FALSE, ...)
}
\arguments{
\item{cmd}{Command to run. Will be quoted by \code{\link{shQuote}()}.}

\item{args}{A character vector of arguments.}

\item{env_vars}{A named character vector of environment variables.}

\item{path}{Path in which to execute the command}

\item{quiet}{If \code{FALSE}, the command to be run will be echoed.}

\item{...}{additional arguments passed to \code{\link[base]{system}}}
}
\value{
command output if the command succeeds, an error will be thrown if
  the command fails.
}
\description{
Run a system command and capture the output.
}
\keyword{internal}
